/*
 * cmapl_wrapper_file_version.c
 *
 *  Created on: Mar 27, 2014
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>
#include "cmapl_wrapper_header.h"

/**
 * Routine to get archiver(ccma) file version.
 * ccma_file_path : path to ccma file
 * Returns cmapl_wrapper_file_version which has file version indicated by major and minor numbers
 */
cmapl_wrapper_file_version get_cmapl_file_version(const char* ccma_file_path, char ** error){
	// On UNIX to enable syslog, add to /etc/syslog.conf or /etc/rsyslog.conf
	// local1.*         /var/log/mylog
	// restart syslogd
	// /etc/init.d/syslog restart (or) /etc/init.d/rsyslog restart
	setlogmask (LOG_UPTO (LOG_NOTICE));
	openlog ("cmapl_wrapper_get_timestamps", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	*error = NULL; // error message to be propagated to JNA interface

	uint32_t FILE_VERSION_ERROR = -1;
	//Open ccma file to read
	cmapl_file_handle_t* file;
	const char* meta_path;
	const char* data_path;

	// Get the cmapl file_handle
	cmapl_error_t err = cmapl_open_data_file(ccma_file_path, "r", &file, &meta_path, &data_path);
	if(err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Error opening file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error;
	}

	// declare file-version
	cmapl_wrapper_file_version file_version;

	// Get file info routine give the following information, Current version of wrapper is only
	// interested in major and minor versions.
	//	uint32_t		major_v;		/**< major version number */
	//	uint32_t                minor_v;                /**< minor version number */
	//	char*			file_path;              /**< path to the generic archive file or perfstat file*/
	//	char*			meta_path;              /**< path to the .meta meta file; only used for archive file */
	//	char*			data_path;              /**< path to the .gz data file; only used for archive file */
	//	uint32_t		file_index;		/**< index number of the file */
	//	uint64_t                start_time;             /**< start UTC time of the file, in milliseconds */
	//	uint64_t                end_time;               /**< end UTC time of the file, in milliseconds */
	//	cmapl_file_type_t       file_type;

	cmapl_file_def_t cmapl_file_def;
	err = cmapl_get_file_info (file, &cmapl_file_def);
	if (err != CMAPL_SUCCESS) {
		syslog(LOG_ERR, "Error getting file info, error : %s\n", cmapl_error_labels[err]);
		goto error;
	}

	//set file_version
	file_version.major_v = cmapl_file_def.major_v;
	file_version.minor_v = cmapl_file_def.minor_v;

	err = cmapl_close_data_file(file);
	if (err != CMAPL_SUCCESS){
		syslog(LOG_ERR, "Error closing file: %s, error : %s\n", ccma_file_path, cmapl_error_labels[err]);
		file = NULL;
		goto error;
	}
	closelog();
	return file_version;

error:
	// Any error caught until this point should be propagated to the JNA interface
    if (err != CMAPL_SUCCESS) {
    	*error = malloc(strlen(cmapl_error_labels[err]) + 1);
    	strcpy(*error, cmapl_error_labels[err]);
    }

	// All information about the ccma data file still in memory will be de-allocated
	if(file != NULL) {
		err = cmapl_close_data_file(file);
		if (err != CMAPL_SUCCESS){
			syslog(LOG_ERR, "\nCould not close ccma data file handler, error : %s for file : %s",cmapl_error_labels[err], ccma_file_path);
		}
		file = NULL;
	}

	file_version.major_v = FILE_VERSION_ERROR;
	file_version.minor_v = FILE_VERSION_ERROR;
	closelog();
	return file_version;
}

/**
 * We dont have any dynamic memory for this in this version, so just de-allocate the error message.
 */
void free_cmapl_file_version(char * error) {
	// Also de-allocate error message if there were any
	if(error != NULL) {
		free(error);
		error = NULL;
	}
}
