/*
 * $Id: //depot/prod/newport_news/main/modules/crypto-util/src/main/java/com/netapp/crypto/KeyRotator.java#1 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.crypto;

import java.io.IOException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.crypto.SecretKey;

import net.jcip.annotations.ThreadSafe;



/**
 * Manages Rotation of keys.<p>
 *
 * One "current" key always exists (see {@link #getCurrentKey()}).
 * This key should be used for all new encryptions.<p>
 *
 * Many other keys also exist that can be used for decryption.<p>
 *
 * The current key index (see {@link KeyEntry#getIndex()})
 * should be stored with any ciphertext.<p>
 *
 * When you need to decrypt something, you should retrieve the key via {@link #getKey(long)}
 * (using the key index previously stored with the ciphertext).<p>
 *
 * When you want to rotate the keys, call {@link #generateNextKey()}.<p>
 */
@ThreadSafe
class KeyRotator {

    /**
     * The prefix for the rotating key aliases.
     *
     * key_${cipherDelegateId}_${keyId}
     */
    private static final String KEY_ALIAS_FORMAT = "key_%s_%s";
    private static final Pattern KEY_ALIAS_PATTERN = Pattern.compile(String.format(KEY_ALIAS_FORMAT, "(\\d+)", "(\\d+)"));

    private final KeyStoreWrapper keyStoreWrapper;

    private SymmetricCipherDelegate cipherDelegate;

    private KeyEntry currentKeyEntry;

    private Map<Long, KeyEntry> cachedKeyEntries = new HashMap<Long, KeyEntry>();

    /**
     * Represents an entry within the key rotation. <p>
     *
     * If you encrypt anything with the key ({@link #getKey()}),
     * then you need to store the index ({@link #getIndex()})
     * along with the ciphertext. <p>
     *
     * Then, when you need to decrypt, you can retrieve the key
     * via {@link KeyRotator#getKey(long)}.  <p>
     *
     * This class is threadsafe
     */
    public static final class KeyEntry {
        private final long index;
        private final SecretKey key;

        protected KeyEntry(long index, SecretKey key) {
            this.index = index;
            this.key = key;
        }

        /**
         * Gets the index of this key within the rotation.
         * Store this value with the ciphertext, so you
         * can retrieve the key at decryption time via
         * {@link KeyRotator#getKey(long)}
         */
        public long getIndex() {
            return index;
        }

        public SecretKey getKey() {
            return key;
        }

    }

    KeyRotator(KeyStoreWrapper keyStoreWrapper, SymmetricCipherDelegate cipherDelegate) throws UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {
        this.keyStoreWrapper = keyStoreWrapper;
        this.cipherDelegate = cipherDelegate;
        initializeCurrentKeyEntry();
    }

    /**
     * Finds and sets the current key. <p>
     *
     * If no current key exists in the KeyStoreManager, then
     * a new key is generated.
     */
    private synchronized void initializeCurrentKeyEntry() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException, UnrecoverableKeyException {
        long maxKeyIndex = -1;
        for (Enumeration<String> e = this.keyStoreWrapper.aliases(); e.hasMoreElements();) {
            String alias = e.nextElement();

            Long keyIndex = getKeyIndex(alias);
            if (keyIndex != null) {
                maxKeyIndex = Math.max(keyIndex.longValue(), maxKeyIndex);
            }
        }

        if (maxKeyIndex == -1) {
            /*
             * Key doesn't exist, create it.
             */
            generateNextKey();
        } else {
            /*
             * Key already exists, retrieve it.
             */
            this.currentKeyEntry = getKey(maxKeyIndex);
        }
    }

    private Long getKeyIndex(String alias) {
        Matcher matcher = KEY_ALIAS_PATTERN.matcher(alias);
        if (matcher.matches()) {
            if (Integer.parseInt(matcher.group(1)) == this.cipherDelegate.getId()) {
                return Long.valueOf(matcher.group(2));
            }
        }
        return null;
    }

    private String getKeyAlias(long index) {
        return String.format(KEY_ALIAS_FORMAT, this.cipherDelegate.getId(), index);
    }

    /**
     * Gets the current (most recently generated) key. <p>
     *
     * The current key should be used for all new encryptions. <p>
     *
     * Note that when you encrypt something with this key, you should store
     * the key index {@link KeyEntry#getIndex()}, so you can later
     * retrieve the key at decryption time via {@link #getKey(long)}
     */
    public synchronized KeyEntry getCurrentKey() {
        return this.currentKeyEntry;
    }

    /**
     * Gets the key at the given index.
     */
    public synchronized KeyEntry getKey(long index) throws UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException {
        KeyEntry keyEntry = this.cachedKeyEntries.get(index);
        if (keyEntry == null) {
            keyEntry = new KeyEntry(
                    index,
                    (SecretKey) this.keyStoreWrapper.getKey(getKeyAlias(index)));
            this.cachedKeyEntries.put(index, keyEntry);
        }
        return keyEntry;
    }

    /**
     * Generates a new key and sets the current key to this newly generated key.
     */
    public synchronized void generateNextKey() throws NoSuchAlgorithmException, KeyStoreException, CertificateException, IOException {

        String nextCurrentKeyAlias = generateNextKeyAlias();

        SecretKey nextCurrentKey = this.cipherDelegate.generateKey();

        this.keyStoreWrapper.setKeyEntry(nextCurrentKeyAlias, nextCurrentKey);

        this.currentKeyEntry = new KeyEntry(
                getKeyIndex(nextCurrentKeyAlias),
                nextCurrentKey);

        this.cachedKeyEntries.put(this.currentKeyEntry.getIndex(), this.currentKeyEntry);
    }

    private synchronized String generateNextKeyAlias() {
        if (this.currentKeyEntry == null) {
            return getKeyAlias(0L);
        } else {
            return getKeyAlias(this.currentKeyEntry.getIndex() + 1);
        }
    }

}
