/*
 * $Id: //depot/prod/newport_news/main/modules/crypto-util/src/main/java/com/netapp/crypto/KeyStoreWrapper.java#2 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.crypto;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;

import javax.crypto.SecretKey;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

/**
 * The KeyStoreWrapper wraps a {@link KeyStore} and provides
 * file persistence, and other conveniences.
 * <p>
 * Create instances using one of the {@code createWrapper} methods.
 */
public class KeyStoreWrapper {

    private final KeyStore keyStore;
    private final File keyStoreFile;
    private final char[] keyStorePassword;
    private final Set<KeyStoreWrapperListener> listeners = new HashSet<KeyStoreWrapperListener>();

    private KeyStoreWrapper(
            KeyStore keyStore,
            File keyStoreFile,
            char[] keyStorePassword) {

        this.keyStore = keyStore;
        this.keyStoreFile = keyStoreFile;
        this.keyStorePassword = keyStorePassword.clone();
    }

    public File getKeyStoreFile() {
        return this.keyStoreFile;
    }

    public KeyStore getKeyStore() {
        return this.keyStore;
    }

    public String getKeyStoreType() {
        return this.keyStore.getType();
    }

    public char[] getKeyStorePassword() {
        return this.keyStorePassword.clone();
    }

    public char[] getKeyPassword() {
        /*
         * Use the same password for all keys in the keystore
         */
        return this.keyStorePassword.clone();
    }

    public boolean addListener(KeyStoreWrapperListener listener) {
        return this.listeners.add(listener);
    }
    public boolean removeListener(KeyStoreWrapperListener listener) {
        return this.listeners.remove(listener);
    }

    private synchronized void saveKeyStore() throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        OutputStream keyStoreOutputStream = null;
        try {
            if (!this.keyStoreFile.getCanonicalFile().getParentFile().exists()) {

                if (!this.keyStoreFile.getCanonicalFile().getParentFile().mkdirs()) {
                    throw new IOException("Unable to make directory: " + this.keyStoreFile.getCanonicalFile().getParentFile());
                }
            }
            keyStoreOutputStream = new BufferedOutputStream(new FileOutputStream(this.keyStoreFile));

            this.keyStore.store(keyStoreOutputStream, this.keyStorePassword);
        } finally {
            IOUtils.closeQuietly(keyStoreOutputStream);
        }

        for (KeyStoreWrapperListener listener : this.listeners) {
            listener.onSave(this);
        }
    }

    public synchronized Enumeration<String> aliases() throws KeyStoreException {
        return this.keyStore.aliases();
    }

    public synchronized Key getKey(String alias) throws KeyStoreException, NoSuchAlgorithmException, UnrecoverableKeyException {
        return this.keyStore.getKey(alias, getKeyPassword());
    }

    public synchronized Certificate getCertificate(String alias) throws KeyStoreException {
        return this.keyStore.getCertificate(alias);
    }

    /**
     * Sets the secret key entry in the keystore, then saves the keystore to disk
     *
     * @see KeyStore#setKeyEntry(String, Key, char[], java.security.cert.Certificate[])
     */
    public synchronized void setKeyEntry(String alias, SecretKey key) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        this.keyStore.setKeyEntry(alias, key, getKeyPassword(), null);
        saveKeyStore();
    }

    /**
     * Sets the private key entry in the keystore, then saves the keystore to disk
     *
     * @see KeyStore#setKeyEntry(String, Key, char[], java.security.cert.Certificate[])
     */
    public synchronized void setKeyEntry(String alias, PrivateKey key, X509Certificate certificate) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        this.keyStore.setKeyEntry(alias, key, getKeyPassword(), new Certificate[] {certificate});
        saveKeyStore();
    }

    /**
     * Sets the certificate entry in the keystore, then saves the keystore to disk.
     * Generally used for truststores (i.e. don't have private/shared keys)
     *
     * @see KeyStore#setCertificateEntry(String, Certificate)
     */
    public synchronized void setCertificateEntry(String alias, X509Certificate certificate) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        this.keyStore.setCertificateEntry(alias, certificate);
        saveKeyStore();
    }

    /**
     * Deletes the key entry from the keystore, then saves the keystore to disk.
     *
     * After deleting a key, you will no longer be able to decrypt anything
     * that was encrypted with the key.  So ensure all ciphertext encrypted
     * with the key being deleted has been re-encrypted with a newer key.
     *
     * @see KeyStore#deleteEntry(String)
     */
    public synchronized void deleteEntry(String alias) throws KeyStoreException, NoSuchAlgorithmException, CertificateException, IOException {
        this.keyStore.deleteEntry(alias);
        saveKeyStore();
    }

    /**
     * Returns true if the underlying keystore has a PrivateKey.
     */
    public synchronized boolean hasPrivateKey() throws KeyStoreException {
        for (Enumeration<String> e = this.keyStore.aliases(); e.hasMoreElements();) {
            String alias = e.nextElement();

            if (this.keyStore.entryInstanceOf(alias, KeyStore.PrivateKeyEntry.class)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Returns true if the underlying keystore has a certificate entry.
     */
    public synchronized boolean hasCertificate() throws KeyStoreException {
        for (Enumeration<String> e = this.keyStore.aliases(); e.hasMoreElements();) {
            String alias = e.nextElement();

            if (this.keyStore.entryInstanceOf(alias, KeyStore.TrustedCertificateEntry.class)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Creates a {@link KeyStoreWrapper} instance for a "jceks" keystore using a password file that is
     * the keystore file with ".password" appended.
     */
    public static KeyStoreWrapper createWrapper(File keyStoreFile) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {

        return createWrapper(
            keyStoreFile,
            "jceks",
            new File(keyStoreFile.getAbsolutePath() + ".password"));
    }

    /**
     * Creates a {@link KeyStoreWrapper} instance.
     *
     * @param keyStoreFile The java keystore file
     * @param keyStoreType The type of keystore
     * @param keyStorePasswordFile The file containing the cleartext keystore password
     */
    public static KeyStoreWrapper createWrapper(
            File keyStoreFile,
            String keyStoreType,
            File keyStorePasswordFile) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {

        char[] keyStorePassword = readKeyStorePassword(keyStorePasswordFile);
        return createWrapper(
            keyStoreFile,
            keyStoreType,
            keyStorePassword);
    }

    /**
     * Creates a {@link KeyStoreWrapper} instance.
     *
     * @param keyStoreFile The java keystore file
     * @param keyStoreType The type of keystore
     * @param keyStorePassword The keystore password
     */
    public static KeyStoreWrapper createWrapper(
            File keyStoreFile,
            String keyStoreType,
            char[] keyStorePassword) throws KeyStoreException, IOException, NoSuchAlgorithmException, CertificateException {
        KeyStore keyStore = KeyStore.getInstance(keyStoreType);

        InputStream keyStoreInputStream = null;
        try {
            boolean existingKeyStore = keyStoreFile.exists();
            if (existingKeyStore) {
                keyStoreInputStream = new BufferedInputStream(new FileInputStream(keyStoreFile));
            }

            keyStore.load(keyStoreInputStream, keyStorePassword);

            KeyStoreWrapper wrapper = new KeyStoreWrapper(
                keyStore,
                keyStoreFile,
                keyStorePassword);

            return wrapper;
        } finally {
            IOUtils.closeQuietly(keyStoreInputStream);
        }
    }

    private static char[] readKeyStorePassword(File keyStorePasswordFile) throws IOException {

        if (keyStorePasswordFile.exists()) {
            return FileUtils.readFileToString(keyStorePasswordFile).trim().toCharArray();
        } else {
            /*
             * TODO perhaps come up with better way of setting default password?
             */
            return new char[] {'c', 'h', 'a', 'n', 'g', 'e', 'm', 'e'};
        }
    }
}
