/*
 * Copyright (c) 2012 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.volume.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.volume.interfaces.data.VolumeMeasurements;
import com.netapp.oci.netapp.performance.volume.interfaces.local.VolumePerformanceQueryLocal;
import com.netapp.oci.netapp.performance.volume.interfaces.remote.VolumePerformanceQueryRemote;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import com.onaro.sanscreen.util.Constant;
import org.apache.log4j.Logger;
import org.jboss.ejb3.annotation.LocalBinding;
import org.jboss.ejb3.annotation.RemoteBinding;
import org.jboss.ejb3.annotation.RemoteBindings;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Anton Sergeev
 */
@Stateless
@RemoteBindings({
        @RemoteBinding(jndiBinding = VolumePerformanceQueryRemote.JNDI_NAME),
        @RemoteBinding(jndiBinding = Constant.HTTPS_JNDI_BINDING_PREFIX + VolumePerformanceQueryRemote.JNDI_NAME)})
@LocalBinding(jndiBinding = VolumePerformanceQueryLocal.JNDI_NAME)
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class VolumePerformanceQueryBean implements VolumePerformanceQueryRemote, VolumePerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = Logger.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<VolumeMeasurements>>> getSamples(Set<Long> volumeIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + volumeIds.size() + "volume samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<VolumeMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<VolumeMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(volumeIds);

            Map<Long, Set<Long>> vserverIdToVolumeIds = NetAppModel.INSTANCE.getVserverIdByVolumeIds(connection, volumeIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToVolumeIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, fromTime, toTime, ids));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find volume with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query volume samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<VolumeMeasurements>>> getSamples(Long containerId, Set<Long> volumeIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + volumeIds.size() + "volume samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, fromTime, toTime, volumeIds);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query volume samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, SampleInfo<VolumeMeasurements>> getLatestSample(Long vServerId, Set<Long> volumeIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + volumeIds.size() + "volume samples for vServer:" + vServerId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getVolumePerformanceComponentsFactory().createPersistenceFacade(connection, vServerId)
                            .getLatestSample(volumeIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public List<SampleInfo<VolumeMeasurements>> getSamples(Long containerId, Long volumeId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(volumeId), fromTime, toTime).get(volumeId);

    }

    private Map<Long, List<SampleInfo<VolumeMeasurements>>> getSamples(Connection connection, Long containerId, Date fromTime, Date toTime, Set<Long> ids) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getVolumePerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(ids, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, List<Summary<VolumeMeasurements>>> getSummaries(Set<Long> volumeIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + volumeIds.size() + "volume summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<VolumeMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<>(volumeIds);

            Map<Long, Set<Long>> vserverIdToVolumeIds = NetAppModel.INSTANCE.getVserverIdByVolumeIds(connection, volumeIds);
            for (Map.Entry<Long, Set<Long>> entry : vserverIdToVolumeIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<VolumeMeasurements>> summaries = ComponentsFactoryRepository.getVolumePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find volume with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<VolumeMeasurements>>> getSummaries(Long vServerId, Set<Long> volumeIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + volumeIds.size() + "volume summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<VolumeMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<VolumeMeasurements>> summaries = ComponentsFactoryRepository.getVolumePerformanceComponentsFactory()
                        .createPersistenceFacade(connection, vServerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), volumeIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }


    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "volume has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getVolumePerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if volume has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}
