/*
 * Copyright (c) 2011 NetApp
 * All rights reserved
 */

package com.netapp.oci.netapp.performance.vserver.interfaces.impl;

import com.netapp.oci.netapp.db.NetAppModel;
import com.netapp.oci.netapp.interfaces.data.NetappInterfaceException;
import com.netapp.oci.netapp.performance.PerformanceConstants;
import com.netapp.oci.netapp.performance.shared.ComponentsFactoryRepository;
import com.netapp.oci.netapp.performance.vserver.interfaces.data.VolumeVserverMeasurements;
import com.netapp.oci.netapp.performance.vserver.interfaces.data.VolumeVserverMeasurements;
import com.netapp.oci.netapp.performance.vserver.interfaces.local.VolumeVserverPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.vserver.interfaces.local.VolumeVserverPerformanceQueryLocal;
import com.netapp.oci.netapp.performance.vserver.interfaces.remote.VolumeVserverPerformanceQueryRemote;
import com.netapp.oci.platform.common.ServerInterfaceInterceptor;
import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;
import com.netapp.oci.platform.common.interfaces.data.performance.SampleInfo;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.framework.performance.data.Summary;
import com.netapp.oci.platform.framework.performance.util.PerformanceUtils;
import com.onaro.commons.util.CollectionPrinter;
import com.onaro.sanscreen.util.Constant;
import org.apache.log4j.Logger;
import org.jboss.ejb3.annotation.LocalBinding;
import org.jboss.ejb3.annotation.RemoteBinding;
import org.jboss.ejb3.annotation.RemoteBindings;

import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.ejb.TransactionAttributeType;
import javax.interceptor.Interceptors;
import javax.sql.DataSource;
import java.sql.Connection;
import java.util.*;

/**
 * @author Nancy Walia
 *         Date: Jan 17, 2012
 */
@Stateless
@RemoteBindings({
        @RemoteBinding(jndiBinding = VolumeVserverPerformanceQueryRemote.JNDI_NAME),
        @RemoteBinding(jndiBinding = Constant.HTTPS_JNDI_BINDING_PREFIX + VolumeVserverPerformanceQueryRemote.JNDI_NAME)})
@LocalBinding(jndiBinding = VolumeVserverPerformanceQueryLocal.JNDI_NAME)
@TransactionAttribute(TransactionAttributeType.REQUIRES_NEW)
@RolesAllowed({"guest", "user", "admin", "scheduler", "acquisition"})
@Interceptors(ServerInterfaceInterceptor.class)
public class VolumeVserverPerformanceQueryBean implements VolumeVserverPerformanceQueryRemote, VolumeVserverPerformanceQueryLocal {
    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    private static final Logger logger = Logger.getLogger(PerformanceConstants.QUERY_PACKAGE_LOGGER);

    @Override
    public Map<Long, List<SampleInfo<VolumeVserverMeasurements>>> getSamples(Set<Long> vserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + vserverIds.size() + "vserver samples from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<SampleInfo<VolumeVserverMeasurements>>> containerIdToSamples = new HashMap<Long, List<SampleInfo<VolumeVserverMeasurements>>>();
            Set<Long> remainingIds = new HashSet<Long>(vserverIds);

            Map<Long, Set<Long>> clusterIdByWorkloadIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vserverIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByWorkloadIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                containerIdToSamples.putAll(getSamples(connection, containerId, vserverIds, fromTime, toTime));
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find volume vserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }

            return containerIdToSamples;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume vserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query volume vserver samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public Map<Long, List<SampleInfo<VolumeVserverMeasurements>>> getSamples(Long containerId, Set<Long> vserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + vserverIds.size() + "wolume vserver samples from " + fromTime + " to " + toTime + " for containerId:" + containerId);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            return getSamples(connection, containerId, vserverIds, fromTime, toTime);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume vserver samples", e);
        } finally {
            ConnectionHelper.close(connection);
            logger.info("Total Execution time to query volume vserver samples: " + (System.currentTimeMillis() - startTime));

        }
    }

    @Override
    public List<SampleInfo<VolumeVserverMeasurements>> getSamples(Long containerId, Long vserverId, Date fromTime, Date toTime) throws NetappInterfaceException {
        return getSamples(containerId, Collections.singleton(vserverId), fromTime, toTime).get(vserverId);

    }

    private Map<Long, List<SampleInfo<VolumeVserverMeasurements>>> getSamples(Connection connection, Long containerId, Set<Long> vserverIds, Date fromTime, Date toTime) throws PerformanceException {
        return PerformanceUtils.toSampleInfos(
                ComponentsFactoryRepository.getVolumeVserverPerformanceComponentsFactory().createPersistenceFacade(connection, containerId)
                        .getSamples(vserverIds, new TimeSpan(fromTime, toTime)));
    }

    @Override
    public Map<Long, SampleInfo<VolumeVserverMeasurements>> getLatestSample(Long clusterId, Set<Long> vServerIds) throws NetappInterfaceException {
        long startTime = System.currentTimeMillis();
        logger.info("Query " + vServerIds.size() + "Volume vServer samples for cluster:" + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return PerformanceUtils.toSampleInfo(
                    ComponentsFactoryRepository.getVolumeVserverPerformanceComponentsFactory().createPersistenceFacade(connection, clusterId)
                            .getLatestSample(vServerIds));
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Volume vServer samples", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<VolumeVserverMeasurements>>> getSummaries(Set<Long> vserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vserverIds.size() + "volume vserver summaries from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            Map<Long, List<Summary<VolumeVserverMeasurements>>> idToSummaries = new HashMap<>();
            Set<Long> remainingIds = new HashSet<Long>(vserverIds);

            Map<Long, Set<Long>> clusterIdByVserverIds = NetAppModel.INSTANCE.getClusterIdByVserverIds(connection, vserverIds);
            for (Map.Entry<Long, Set<Long>> entry : clusterIdByVserverIds.entrySet()) {
                Long containerId = entry.getKey();
                Set<Long> ids = entry.getValue();

                remainingIds.removeAll(ids);
                List<Summary<VolumeVserverMeasurements>> summaries = ComponentsFactoryRepository.getVolumeVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, containerId)
                        .getSummaries(new TimeSpan(fromTime, toTime), ids);
                PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            }

            if (!remainingIds.isEmpty()) {
                throw new NetappInterfaceException("Failed to find vserver with id " + CollectionPrinter.LONG.toString(remainingIds));
            }
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get Volume vserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    @Override
    public Map<Long, List<Summary<VolumeVserverMeasurements>>> getSummaries(Long clusterId, Set<Long> vserverIds, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query " + vserverIds.size() + "Volume vserver summaries from " + fromTime + " to " + toTime
                + " for cluster " + clusterId);
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            Map<Long, List<Summary<VolumeVserverMeasurements>>> idToSummaries = new HashMap<>();
            List<Summary<VolumeVserverMeasurements>> summaries = ComponentsFactoryRepository.getVolumeVserverPerformanceComponentsFactory()
                        .createPersistenceFacade(connection, clusterId)
                        .getSummaries(new TimeSpan(fromTime, toTime), vserverIds);
            PerformanceUtils.addSummariesToMap(idToSummaries, summaries);
            return idToSummaries;
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to get volume vserver summaries", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }


    @Override
    public Boolean areStatisticsAvailable(Long containerId, Long objId, Date fromTime, Date toTime) throws NetappInterfaceException {
        logger.info("Query if " + objId + "Volume vserver has counters from " + fromTime + " to " + toTime);

        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);
            return ComponentsFactoryRepository.getVolumeVserverPerformanceComponentsFactory()
                    .createPersistenceFacade(connection, containerId).doSamplesExist(new TimeSpan(fromTime, toTime), objId);
        } catch (Exception e) {
            throw new NetappInterfaceException("Failed to assert if volume vserver has counters ", e);
        } finally {
            ConnectionHelper.close(connection);
        }
    }
}

