package com.netapp.oci.platform.framework.performance;

import com.netapp.oci.platform.common.TimeSpan;
import com.netapp.oci.platform.framework.performance.data.Sample;
import com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException;

import java.util.List;
import java.sql.Connection;
import java.util.Map;
import java.util.Set;

/**
 * Facade for sample queries
 *
 * @author Anton Sergeev
 */
public interface SampleFacade<C> {

    /**
     * Retrieves samples from persistence layer.
     *
     * @param connection - database connection
     * @param id         identifies the object for which to retrieve samples (objId)
     * @param timeSpan   start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if failed to retrieve samples
     */
    List<Sample<C>> getSamples(Connection connection, Long id, TimeSpan timeSpan) throws PerformanceException;

    /**
     * Retrieves samples from persistence layer.
     *
     * @param connection - database connection
     * @param ids        identifies the objects for which to retrieve samples (objId)
     * @param timeSpan   start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if failed to retrieve samples
     */
    Map<Long, List<Sample<C>>> getSamples(Connection connection, Set<Long> ids, TimeSpan timeSpan) throws PerformanceException;

    /**
     * Retrieves the latest sample for the given object ids from persistence layer.
     *
     * @param connection - database connection
     * @param ids        identifies the objects for which to retrieve samples (objId)
     * @return Unordered samples
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if failed to retrieve samples
     */
    Map<Long, Sample<C>> getLatestSample(Connection connection, Set<Long> ids) throws PerformanceException;

    /**
     * Retrieves samples from persistence layer group by time.
     *
     * @param connection - database connection
     * @param ids        identifies the object for which to retrieve samples (objId)
     * @param timeSpan   start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if failed to retrieve samples
     */
    List<Sample<C>> getSamplesGroupByTime(Connection connection, Set<Long> ids, TimeSpan timeSpan) throws PerformanceException;

    /**
     * Retrieves samples from persistence layer group by time.
     *
     * @param connection - database connection
     * @param timeSpan   start and end times filtering the samples
     * @return the samples ordered by report time
     * @throws com.netapp.oci.platform.common.interfaces.data.performance.PerformanceException
     *          if failed to retrieve samples
     */
    List<Sample<C>> getSamplesGroupByTime(Connection connection, TimeSpan timeSpan) throws PerformanceException;


    /**
     * returns true if the samples exist over a period of time.
     *
     * @param timeSpan the period in which samples sould be retrieved
     * @param id
     * @return true if samples exist
     * @throws com.netapp.oci.platform.common.table.TableException
     *          if failed to retrieve the samples
     */
    public Boolean doSamplesExist(Connection connection, TimeSpan timeSpan, Long id) throws PerformanceException;
}
