package com.netapp.oci.platform.framework.performance.tables;

import com.netapp.oci.platform.common.table.ObjectParser;
import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.TimeSpan;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Summary table that knows how to insert summaries based on values from related sample table
 * Usually when performance package uses this kind of summary table it means that
 * a) summarization process is simple and can be discribed using SQL functions like MAX, SUM, AVG etc
 * b) SummaryCalculator is not implemented for such package since there is no need for it
 *
 * @author Anton Sergeev
 */
public class InsertSelectSummaryTable<S> extends SummaryTable<S> {
	private final static Logger LOG = LoggerFactory.getLogger(InsertSelectSummaryTable.class);
	
    private ObjectParser<S> objectParser;

    public InsertSelectSummaryTable(Connection connection, String suffixName, Column[] columns, ObjectParser<S> objectParser) {
        super(connection, suffixName, columns);
        this.objectParser = objectParser;
    }

    @Override
    protected void prepareSummaryDataForInsert(PreparedStatementWrapper preparedStatement, S summaryData) throws TableException {
        objectParser.prepareObjectDataForInsert(preparedStatement, summaryData);
    }

    public void insert(SampleTable sampleTable, TimeSpan timeSpan, SqlMethod [][] summarizationMethods) throws TableException {
        StringBuilder sql = new StringBuilder();

        sql.append("INSERT INTO ").append(getFullName()).append("\n");

        sql.append("SELECT objid,").append(timeSpan.getFromTime()).append(",").append(timeSpan.getToTime());
        sql.append(SqlMethod.getWrappedColumns("", sampleTable.getSampleColumns(), summarizationMethods));
        sql.append("\n");

        sql.append("FROM ").append(sampleTable.getFullName()).append("\n");
        sql.append("WHERE empty = 0 GROUP BY objid \n");
        sql.append(getOnDuplicateKeyUpdate("", sampleTable.getSampleColumns(), summarizationMethods));

        LOG.debug("Execute sql:\n" + sql.toString());
        
        getPrepareStatementEx(sql.toString()).executeUpdateEx();
    }

    protected S parseSummaryData(ResultSet resultSet, int startIndex) throws SQLException {
        return objectParser.parseObjectData(resultSet, startIndex);
    }
    
    private String getOnDuplicateKeyUpdate(String prefix, Column[] columns, SqlMethod[][] methods) {
        StringBuilder sb = new StringBuilder();
        sb.append("ON DUPLICATE KEY UPDATE \n");
        for (int i = 0; i < columns.length; i++) {
            if (i != 0) {
                sb.append(",\r\n    ");
            }
            Column column = columns[i];
            SqlMethod[] columnMethods = methods[i];
            for (int j = 0; j < columnMethods.length; j++) {
                if (j != 0) {
                    sb.append(",\r\n    ");
                }
                sb.append(prefix + column.getName()).append(" = ");
                sb.append("VALUES(").append(column.getName()).append(")");
            }
        }
        
        return sb.toString();
    }
    
}
