/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.platform.originator;

import com.netapp.oci.platform.common.table.PreparedStatementWrapper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.common.table.TableResultParserDefault;
import com.netapp.oci.platform.originator.db.DbOperation;
import com.netapp.oci.platform.originator.db.DeleteOperation;
import com.netapp.oci.platform.originator.db.RunnerOperation;
import com.netapp.oci.platform.originator.handler.XmlHandler;
import com.netapp.oci.platform.originator.key.KeyCache;
import com.netapp.oci.platform.originator.runner.DbRunner;
import com.netapp.oci.platform.originator.runner.UnlinkObjectRunner;
import com.netapp.oci.platform.plugin.data.PlatformPlugin;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import org.apache.log4j.Logger;

import java.sql.Connection;
import java.util.*;

/**
 * @author Iti Behari
 *         Date: 6/4/13
 */
public class SoftDeletedObjectsPurger {
    private static final Logger logger = Logger.getLogger(SoftDeletedObjectsPurger.class);

    public Map<String, Set<Long>> purge(Connection connection, PlatformPlugin plugin, long softDeletedObjsLifetime, long originId, long updateTime) throws TableException, ServerInterfaceException {
        logger.info("Begin purging for originator " + originId);
        long cutoffTime = updateTime - softDeletedObjsLifetime;
        Long rootId = Long.valueOf(plugin.getRootId(connection, originId));
        List<String> tables = getTablesInDeletionOrder(plugin);
        Map<String, String> tableToTagLookupMap = buildTableToTagLookupMap(plugin);

        Map<String, Set<Long>> tagToIdsBeingDeleted = new HashMap<String, Set<Long>>();
        int totalObjsBeingDeleted = 0;
        for (String tableName : tables) {
            Set<Long> objIdsToBeDeleted = queryObjIdsToBeDeletedForTable(connection, originId, tableName, cutoffTime);
            if (!objIdsToBeDeleted.isEmpty()) {
                totalObjsBeingDeleted += objIdsToBeDeleted.size();
                String tagName = tableToTagLookupMap.get(tableName);
                deleteDataFromTable(connection, plugin, tagName, tableName, objIdsToBeDeleted, updateTime);
                tagToIdsBeingDeleted.put(tagName, objIdsToBeDeleted);

                String subTableName = plugin.getRegistry().getSubTable(tableName);
                if (subTableName != null) {
                    String subTagName = tableToTagLookupMap.get(subTableName);
                    deleteDataFromTable(connection, plugin, subTagName, subTableName, objIdsToBeDeleted, updateTime);
                }
            }
        }

        if (!tagToIdsBeingDeleted.isEmpty()) {
            insertChangeRecord(connection, plugin, rootId, tagToIdsBeingDeleted, updateTime);
            clearCache(plugin, originId, tagToIdsBeingDeleted);
        }
        logger.info("Finished purging for originator  " + originId + ", total objects purged:" + totalObjsBeingDeleted);
        logger.debug("Tags that were affected by Purging : " + tagToIdsBeingDeleted.keySet());
        return tagToIdsBeingDeleted;
    }

    private void insertChangeRecord(Connection connection, PlatformPlugin plugin, long rootId, Map<String, Set<Long>> tagToIdsBeingDeleted, long updateTime) throws ServerInterfaceException, TableException {
        plugin.logObjectsPurgeJob(connection, rootId, tagToIdsBeingDeleted, updateTime);
    }


    private List<String> getTablesInDeletionOrder(PlatformPlugin plugin) {
        List<String> reversedTableOrder = new ArrayList<String>(plugin.getTableOrder());
        Collections.reverse(reversedTableOrder);
        return reversedTableOrder;
    }


    private void clearCache(PlatformPlugin plugin, Long originId, Map<String, Set<Long>> tagsToIdsToBeDeleted) {
        KeyCache.get().purgeSoftDeletedObjects(originId, tagsToIdsToBeDeleted);
    }

    private Map<String, String> buildTableToTagLookupMap(PlatformPlugin plugin) {
        Map<String, String> tableToTagName = new HashMap<String, String>();
        Map<String, XmlHandler> tagToXmlHandler = plugin.getTagToXmlHandlers();
        for (Map.Entry<String, XmlHandler> entry : tagToXmlHandler.entrySet()) {
            tableToTagName.put(entry.getValue().getTableName(), entry.getKey());
        }
        return tableToTagName;
    }

    private void deleteDataFromTable(Connection connection, PlatformPlugin plugin, String currentTagName, String tableName, Set<Long> idsToBeDeleted, long updateTime) throws TableException {
        // Queue up any runner cleanup operations. These will execute before the standard insert/update/delete xml handler
        // operations run.
        Map<String, DbRunner> tagToDbRunner = plugin.getTagToDbRunners();
        DbRunner dbRunner = tagToDbRunner.get(currentTagName);
        if (dbRunner != null) {
            String tempDatabase = tableName.substring(0, tableName.indexOf('.'));
            RunnerOperation runnerOperation = dbRunner.getDeleteExec(tempDatabase, idsToBeDeleted);
            runnerOperation.execute(connection, updateTime);
        }
        Map<String, Map<String, List<String>>> tagToColumnsToUnlinkOnDelete = plugin.getRegistry().getTagToColumnsToUnlinkOnDelete();

        // Queue up unlink object operations
        Map<String, List<String>> objectsToUnlink = tagToColumnsToUnlinkOnDelete.containsKey(currentTagName) ? tagToColumnsToUnlinkOnDelete.get(currentTagName) : Collections.<String, List<String>>emptyMap();
        if (!objectsToUnlink.isEmpty()) {
            new UnlinkObjectRunner().getRunnerOp("management", currentTagName, idsToBeDeleted, objectsToUnlink).execute(connection, updateTime);
        }
        DbOperation operation = new DeleteOperation(tableName, idsToBeDeleted);
        operation.execute(connection, updateTime);

    }

    private Set<Long> queryObjIdsToBeDeletedForTable(Connection connection, Long originId, String tableName, long cutOffTime) throws TableException {
        final String sql = String.format("SELECT objId FROM %s WHERE originId = ? AND updateTime < ? AND objState = 'DEAD' ", tableName);
        List<Long> ids = PreparedStatementWrapper.create(connection, sql, TableResultParserDefault.LONG)
                .setNext(originId)
                .setNext(cutOffTime)
                .executeQueryEx();
        return new HashSet<Long>(ids);
    }
}
