package com.netapp.oci.tool.replay;

import java.io.BufferedReader;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.PollingPackage;
import com.onaro.sanscreen.acquisition.framework.datasource.config.ConfigurationFactory;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationReplayListener;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.util.xml.XMLException;
import com.onaro.sanscreen.util.xml.XMLUtil;

public class FoundationReplay {
    private static long REPLAY_DATA_SOURCE_ID = 99999;
    Long originatorID;

    public FoundationReplay(Long originatorID) {   
        this.originatorID = originatorID;
    }

    private void runFile(File file) throws Exception {
        try {
            if (file.getName().endsWith(".zip")) {
                replayZip(file);
            }
            else {
                System.out.print(file.getName() + "is not a valid compressed replay file");
            }
        }
        catch (FrameworkException e) {
            System.err.println("FoundationReplay failed");
            e.printStackTrace(System.out);
        }
        catch (IOException e) {
            System.err.println("IO exception");
            e.printStackTrace(System.out);
        }
    }

    private void runDir(final File dir) throws Exception {
        final Pattern recording = Pattern.compile(".*_(\\d+_\\d+_\\d+_\\d+_\\d+_\\d+_\\d+).zip");

        File[] zipFiles = dir.listFiles(new FilenameFilter() {

            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".zip");
            }
        });
        Arrays.sort(zipFiles, new Comparator<File>() {

            @Override
            public int compare(File o1, File o2) {
                Matcher matcher1 = recording.matcher(o1.getName());
                Matcher matcher2 = recording.matcher(o2.getName());
                if (matcher1.matches() && matcher2.matches()) {
                    return matcher1.group(1).compareTo(matcher2.group(1));
                }
                else {
                    return o1.getName().compareTo(o2.getName());
                }
            }
        });

        for (File zip : zipFiles) {
            replayZip(zip);
        }
    }

    private void replayZip(File zip) throws Exception {
        System.out.println("Processing recording: " + zip);
        Long origID = originatorID;
        
        if (origID == null) {
            origID  = extractOriginatorID(zip);
            
            if (origID == null) {
                throw new Exception("No originator ID is provided.");
            }
        }
        
        DSSessionFileReader recording = new DSSessionFileReader(zip);
        String dataSourceName = stripRedundantPrefix(recording.getFile().getName());

        // Instantiate the configuration object...
        ConfigurationFactory factory = new ConfigurationFactory(dataSourceName, FoundationReplay.REPLAY_DATA_SOURCE_ID);
        DataSourceConfiguration dsConfiguration = factory.createEmpty(recording.getDataSourceType());

        PackageConfiguration<?> packageConfiguration = factory.createPackageForUnitTestSimulation(dsConfiguration,
                recording, null);

        NetAppFoundationConfiguration<?> foundationConfiguration = (NetAppFoundationConfiguration<?>) packageConfiguration;
        dsConfiguration.registerPackageConfiguraiton(foundationConfiguration);

        BaseDataSource dataSource = new BaseDataSource(dsConfiguration);

        //PollingPackage pollingPackage = foundationConfiguration.instantiatePackage(dataSource);
        NetAppFoundationReplayListener listener = new NetAppFoundationReplayListener(foundationConfiguration, origID);
        PollingPackage<?> pollingPackage = foundationConfiguration.instantiatePackage(dataSource, listener, listener, listener, listener, listener);
        dataSource.registerPackage(pollingPackage);

        System.out.println(String.format("Polling from ZIP file: %s\n ", recording.getFile().getAbsolutePath()));

        // Now execute the foundation package
        dataSource.start();
        dataSource.run(NetAppFoundationPackage.PACKAGE);
        dataSource.stop();
    }

    private String stripRedundantPrefix(String name) {
        while (name.startsWith("netappfoundation_")) {
            name = name.substring("netappfoundation_".length());
        }
        
        Pattern pattern  = Pattern.compile("(.*)_\\d+_\\d+_\\d+_\\d+_\\d+_\\d+_\\d+_\\d+.zip");
        Matcher matcher = pattern.matcher(name);
        
        if (matcher.find()) {
            return matcher.group(1);
        }
        else {
            return name;
        }
    }
    
    private long extractOriginatorID (File zip) throws IOException, XMLException {  
        Long originatorID = null;
        
        ZipFile zipFile = new ZipFile(zip);

        Enumeration<? extends ZipEntry> entries = zipFile.entries();
        while (entries.hasMoreElements()) {
            ZipEntry zipEntry = entries.nextElement();
            if (zipEntry.getName().matches("\\d+(/|\\\\)netappdata_netappfoundation.*\\.xml")) {
                System.out.println("Extract originator ID from: " + zipEntry.getName());
                StringBuilder sb = new StringBuilder();

                InputStream is = zipFile.getInputStream(zipEntry);
                InputStreamReader isr = new InputStreamReader(is);
                BufferedReader reader = new BufferedReader(isr);
                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    sb.append(line);
                }
                reader.close();
                isr.close();
                is.close();
                
                originatorID = XMLUtil.getLongElementVal(sb.toString(), "Originator");
                break;
            }
        }
        
        zipFile.close();
        return originatorID;
        
    }

    public static void main(String[] args) {
        Long originatorID = null;
       
        // quiet log4j
        BasicConfigurator.configure(new NullAppender());

        Options options = new Options().addOption("o", "originatorID", true, "Import originator ID")
                                       .addOption("f", "file", true, "Import zip file")
                                       .addOption("d", "dir", true, "Import directory containing recordings");
        try {
            CommandLine commandLine = new GnuParser().parse(options, args);
            ServiceManager.getInstance().start();
            
            if (commandLine.hasOption("o")) {
                originatorID = Long.valueOf(commandLine.getOptionValue("o"));
            }

            if (commandLine.hasOption('f')) {
                File file = new File(commandLine.getOptionValue('f'));
                if (file.exists() && file.isFile() && file.getName().endsWith(".zip")) {
                    new FoundationReplay(originatorID).runFile(file);
                }
                else {
                    System.out.println("Can't read " + file);
                }
            }
            else if (commandLine.hasOption('d')) {
                File dir = new File(commandLine.getOptionValue('d'));
                if (dir.isDirectory()) {
                    new FoundationReplay(originatorID).runDir(dir);
                }
                else {
                    System.out.println(dir + " is not a directory");
                }
            }
            else {
                new HelpFormatter().printHelp("fdnreplay", options);
            }
        }
        catch (Exception e) {
            e.printStackTrace(System.out);
        }
        finally {
            ServiceManager.getInstance().stop();  
        }    
    }
}

