package com.netapp.oci.tool.replay;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.PollingPackage;
import com.onaro.sanscreen.acquisition.framework.datasource.config.ConfigurationFactory;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSInputStream;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader.Sample;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationReplayListener;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant.NetAppFoundationDependantPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.dependant.config.NetAppFoundationDependantConfiguration;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceListener;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceReplayListener;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.config.NetAppStoragePerformanceConfiguration;
import com.onaro.sanscreen.util.xml.XMLException;
import com.onaro.sanscreen.util.xml.XMLUtil;

public class PerformanceReplay {
    private static long REPLAY_DATA_SOURCE_ID = 99999;
    Long originatorID;
    private List<File> temporaryFiles = new LinkedList<File>();

    public PerformanceReplay(Long originatorID) {
        this.originatorID = originatorID;
    }

    private void runFile(File file) throws Exception {
        try {
            if (file.getName().endsWith(".zip")) {
                replayZip(file);
            }
            else {
                System.out.print(file.getName() + "is not a valid compressed replay file");
            }
        }
        catch (FrameworkException e) {
            System.err.println("PerformanceReplay failed");
            e.printStackTrace(System.out);
        }
        catch (IOException e) {
            System.err.println("IO exception");
            e.printStackTrace(System.out);
        }
    }

    private void runDir(final File dir) throws Exception {
        final Pattern recording = Pattern.compile(".*_(\\d+_\\d+_\\d+_\\d+_\\d+_\\d+_\\d+).zip");

        File[] zipFiles = dir.listFiles(new FilenameFilter() {

            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".zip");
            }
        });
        Arrays.sort(zipFiles, new Comparator<File>() {

            @Override
            public int compare(File o1, File o2) {
                Matcher matcher1 = recording.matcher(o1.getName());
                Matcher matcher2 = recording.matcher(o2.getName());
                if (matcher1.matches() && matcher2.matches()) {
                    return matcher1.group(1).compareTo(matcher2.group(1));
                }
                else {
                    return o1.getName().compareTo(o2.getName());
                }
            }
        });

        for (File zip : zipFiles) {
            replayZip(zip);
        }
    }

    private void replayZip(File zip) throws Exception {
        System.out.println("Processing recording: " + zip);
        Long origID = originatorID;
        DSSessionFileReader recording = new DSSessionFileReader(zip);
        
        try {
            // instantiate the configuration object.
            final ConfigurationFactory factory = new ConfigurationFactory(recording.getFile().getName(), PerformanceReplay.REPLAY_DATA_SOURCE_ID);

            final DataSourceConfiguration dsConfiguration = factory.createEmpty(recording.getDataSourceType());

            // need to find if a foundation recording embedded in the first sample, and assume there is always one.
            DSSessionFileReader.Sample firstDependentSample = recording.getSamples().iterator().next();
            if (firstDependentSample == null) {
                throw new Exception("The recording does not have any samples.");
            }

            final DSSessionFileReader firstFoundationRecording = retrieveFoundationRecordingFromSample(firstDependentSample);
            if (firstFoundationRecording == null) {
                throw new Exception("Could not find a foundation recording in first sample!");
            }
            
            if (origID == null) {
                origID  = extractOriginatorID(firstFoundationRecording.getFile());
                
                if (origID == null) {
                    throw new Exception("No originator ID is provided.");
                }
            }

            // create the foundation package configuration
            final PackageConfiguration packageConfiguration = factory.createPackageForUnitTestSimulation(dsConfiguration, firstFoundationRecording, null);
            final NetAppFoundationConfiguration foundationConfiguration = (NetAppFoundationConfiguration) packageConfiguration;
            dsConfiguration.registerPackageConfiguraiton(foundationConfiguration);

            // create the dependent package configuration
            final PackageConfiguration dependentPackageConfiguration = factory.createPackageForUnitTestSimulation(dsConfiguration, recording, null);
            final NetAppFoundationDependantConfiguration performanceConfiguration = (NetAppFoundationDependantConfiguration) dependentPackageConfiguration;
            dsConfiguration.registerPackageConfiguraiton(performanceConfiguration);

            // create DS
            final BaseDataSource dataSource = new BaseDataSource(dsConfiguration);

            // create the foundation package
            NetAppFoundationReplayListener listener = new NetAppFoundationReplayListener(foundationConfiguration, origID);
            PollingPackage<?> pollingPackage = foundationConfiguration.instantiatePackage(dataSource, listener, listener, listener, listener, listener);
            dataSource.registerPackage(pollingPackage);

            // create the dependent package
            final NetAppStoragePerformanceConfiguration storagePerformanceConfiguration = (NetAppStoragePerformanceConfiguration) dependentPackageConfiguration;
            // create the storage performance package
            NetAppStoragePerformanceListener storagePerformanceListener = new NetAppStoragePerformanceReplayListener(storagePerformanceConfiguration);
            final NetAppFoundationDependantPackage dependentPackage = storagePerformanceConfiguration.instantiatePackage(
                    dataSource, storagePerformanceListener, storagePerformanceListener, storagePerformanceListener);
            dataSource.registerPackage(dependentPackage);
            
            // start the data source
            dataSource.start();

            // run first foundation run, then close out temporary recording.
            dataSource.run(NetAppFoundationPackage.PACKAGE);
            firstFoundationRecording.close();

            // simulate dependent samples.
            for (int i = 0; i < recording.getSamples().size(); ++i) {
                // see if current sample has another foundation update.
                // already ran foundation for first sample.
                if (i > 0) {

                    final DSSessionFileReader foundationRecording = retrieveFoundationRecordingFromSample(recording.getCurrentSample());
                    if (foundationRecording != null) {

                        // reset the simulation session for the foundation package.
                        pollingPackage.getConfiguration().setSimulationSession(foundationRecording);

                        // re-run foundation.
                        dataSource.run(NetAppFoundationPackage.PACKAGE);

                        // close out temporary recording.
                        foundationRecording.close();
                    }
                }

                dataSource.run(dependentPackage.getPackageName());
            }

            dataSource.stop();  
        }
        catch (Throwable e) {
            e.printStackTrace();
            throw new Exception("Failed to run the replay", e);
        }
        finally {
            // temporary files contains temporary copies of foundation zips we made.  Delete them.
            for (final File f : temporaryFiles) {
                f.delete();
            }
        }      
      
    }
    
    /**
     * Determine if passed recorded sample has an embedded foundation zip and retrieve it if so.   Returns NULL
     * if there is no foundation updated in this sample.
     *
     * @param sample the sample to check and retrieve foundation from.
     *
     * @return the recording session representation of the foundation zip if present.  NULL if not in this sample.
     *
     * @throws Exception for any problem dealing with the recording or retrieval.
     */
    private DSSessionFileReader retrieveFoundationRecordingFromSample(final DSSessionFileReader.Sample sample)
            throws Exception {
        final String foundationRecordingFile = sample.getDescriptionProperty("foundation", "recording");
        if (foundationRecordingFile == null) {
            return null;
        }

        DSSessionFileReader foundationZip = null;
        for (DSInputStream stream : sample.getStreams()) {
            if (stream.getFileName().equals(foundationRecordingFile)) {
                // make a copy of this outside the zip file for use.
                File outputZip = new File(foundationRecordingFile);
                temporaryFiles.add(outputZip);

                BufferedOutputStream out = null;
                try {
                    out = new BufferedOutputStream(new FileOutputStream(outputZip));

                    final int bufferSize = 8 * 1024;
                    byte[] buffer = new byte[bufferSize];
                    int count = stream.read(buffer);
                    while (count > 0) {
                        out.write(buffer, 0, count);
                        count = stream.read(buffer);
                    }
                    out.flush();

                }
                catch (IOException ioe) {
                    //ignore
                }
                finally {
                    if (out != null) {
                        try {
                            out.close();
                        }
                        catch (IOException e) {

                            //ignore.
                        }
                    }
                }

                foundationZip = new DSSessionFileReader(outputZip);

            }
        }

        return foundationZip;
    }
    
    private long extractOriginatorID (File zip) throws IOException, XMLException {  
        Long originatorID = null;
        
        ZipFile zipFile = new ZipFile(zip);

        Enumeration<? extends ZipEntry> entries = zipFile.entries();
        while (entries.hasMoreElements()) {
            ZipEntry zipEntry = entries.nextElement();
            if (zipEntry.getName().matches("\\d+(/|\\\\)netappdata_netappfoundation.*\\.xml")) {
                System.out.println("Extract originator ID from: " + zipEntry.getName());
                StringBuilder sb = new StringBuilder();

                InputStream is = zipFile.getInputStream(zipEntry);
                InputStreamReader isr = new InputStreamReader(is);
                BufferedReader reader = new BufferedReader(isr);
                for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                    sb.append(line);
                }
                reader.close();
                isr.close();
                is.close();
                
                originatorID = XMLUtil.getLongElementVal(sb.toString(), "Originator");
                break;
            }
        }
        
        zipFile.close();
        return originatorID;
        
    }

    public static void main(String[] args) {
        Long originatorID = null;
       
        // quiet log4j
        BasicConfigurator.configure(new NullAppender());

        Options options = new Options().addOption("o", "originatorID", true, "Import originator ID")
                                       .addOption("f", "file", true, "Import a recording zip file")
                                       .addOption("d", "dir", true, "Import directory containing recordings");
        try {
            CommandLine commandLine = new GnuParser().parse(options, args);
            ServiceManager.getInstance().start();
            
            if (commandLine.hasOption("o")) {
                originatorID = Long.valueOf(commandLine.getOptionValue("o"));
            }

            if (commandLine.hasOption('f')) {
                File file = new File(commandLine.getOptionValue('f'));
                if (file.exists() && file.isFile() && file.getName().endsWith(".zip")) {
                    new PerformanceReplay(originatorID).runFile(file);
                }
                else {
                    System.out.println("Can't read " + file);
                }
            }
            else if (commandLine.hasOption('d')) {
                File dir = new File(commandLine.getOptionValue('d'));
                if (dir.isDirectory()) {
                    new PerformanceReplay(originatorID).runDir(dir);
                }
                else {
                    System.out.println(dir + " is not a directory");
                }
            }
            else {
                new HelpFormatter().printHelp("perfreplay.sh", options);
            }
        }
        catch (Exception e) {
            e.printStackTrace(System.out);
        }
        finally {
            ServiceManager.getInstance().stop();  
        }    
    }
}

