/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.netapp.oci.tool.replay;

import com.netapp.oci.server.UpdateTaskException;
import com.onaro.sanscreen.server.originatorupdate.OriginatorCacheChanges;
import com.onaro.sanscreen.server.sessions.OriginatorUpdateSession;
import com.onaro.sanscreen.util.Schema;
import com.onaro.util.sessions.DefaultClientLogin;
import com.onaro.util.sessions.SessionCache;
import com.onaro.util.sessions.SessionCacheException;
import org.apache.commons.cli.*;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.varia.NullAppender;

import java.io.*;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

/**
 * @author Al Velella
 *         Date: 8/19/13
 *         Time: 11:33 AM
 */
public class Replay {

    private SessionCache sessionCache = null;
    private OriginatorUpdateSession updateSession;

    public Replay(String address) throws SessionCacheException {
        System.setProperty("java.naming.factory.initial", "org.jnp.interfaces.NamingContextFactory");
        System.setProperty("java.naming.provider.url", address + ":1099");
        System.setProperty("java.naming.factory.url.pkgs", "org.jboss.naming");
        System.setProperty("javax.net.ssl.trustStore", getTrustStore());
        System.setProperty("javax.net.ssl.trustStorePassword", getTrustStorePassword());
        this.sessionCache = new SessionCache(Replay.class.getClassLoader(), new DefaultClientLogin("admin", "admin123"), 0);
        this.updateSession = createSession(OriginatorUpdateSession.JNDI_NAME);
    }

    @SuppressWarnings({"RedundantCast", "unchecked"})
    public synchronized <E> E createSession(final String sessionJndiName) throws SessionCacheException {
        return (E) sessionCache.getSession(sessionJndiName);
    }

    private String getTrustStore() {
        if (isLinux()) {
            return "/opt/netapp/essentials/jboss/server/onaro/cert/server.keystore";
        } else {
            return "C:\\Program Files\\SANscreen\\jboss\\server\\onaro\\cert\\server.keystore";
        }
    }

    private String getTrustStorePassword() {
        return "changeit";
    }

    private void runFile(File file) {
        try {
            if (file.getName().endsWith(".zip")) {
                replayZip(file);
            } else if (file.getName().endsWith(".xml")) {
                replayFile(file);
            }
        } catch (UpdateTaskException e) {
            System.err.println("Error sending report to server");
            e.printStackTrace();
        } catch (IOException e) {
            System.err.println("IO exception");
            e.printStackTrace();
        }
    }

    private void runDir(final File dir, final boolean all) {
        final Pattern recording = Pattern.compile(".*_(\\d+_\\d+_\\d+_\\d+_\\d+_\\d+_\\d+).zip");

        File[] zipFiles = dir.listFiles(new FilenameFilter() {
            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".zip") && (all || !name.endsWith("_copy.zip"));
            }
        });
        Arrays.sort(zipFiles, new Comparator<File>() {
            @Override
            public int compare(File o1, File o2) {
                Matcher matcher1 = recording.matcher(o1.getName());
                Matcher matcher2 = recording.matcher(o2.getName());
                if (matcher1.matches() && matcher2.matches()) {
                    return matcher1.group(1).compareTo(matcher2.group(1));
                } else {
                    return o1.getName().compareTo(o2.getName());
                }
            }
        });

        try {
            for (File zip : zipFiles) {
                replayZip(zip);
            }
        } catch (UpdateTaskException e) {
            System.err.println("Error sending report to server");
            e.printStackTrace();
        } catch (FileNotFoundException e) {
            System.err.println("File not found" + dir.getName());
        } catch (ZipException e) {
            System.err.println("Error accessing zip");
            e.printStackTrace();
        } catch (IOException e) {
            System.err.println("IO exception");
            e.printStackTrace();        }
    }

    private void replayZip(File zip) throws IOException, UpdateTaskException {
        System.out.println("Processing recording: " + zip);
        ZipFile zipFile = new ZipFile(zip);

        Enumeration<? extends ZipEntry> entries = zipFile.entries();
        while (entries.hasMoreElements()) {
            OriginatorCacheChanges changes = null;
            ZipEntry zipEntry = entries.nextElement();
            if (zipEntry.getName().matches("\\d+(/|\\\\)netappdata_netappfoundation.*\\.xml")) {
                try {
                    System.out.print("Replay " + zipEntry.getName() + " ... ");

                    StringBuilder netappData = new StringBuilder();

                    InputStream is = zipFile.getInputStream(zipEntry);
                    InputStreamReader isr = new InputStreamReader(is);
                    BufferedReader reader = new BufferedReader(isr);
                    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                        netappData.append(line);
                    }
                    reader.close();
                    isr.close();
                    is.close();
                    changes = updateSession.getUpdate(netappData.toString(), Schema.NETAPP);
                    System.out.println(changes.getChangesCount() == 1 ? "1 change" : (changes.getChangesCount() + " changes"));
                } finally {
                    if (changes == null) {
                        System.out.println("failed");
                    }
                }
            }
        }
        zipFile.close();
    }

    private void replayFile(File file) throws IOException, UpdateTaskException {
        OriginatorCacheChanges changes = null;
        try {
            System.out.print("Replay: ... ");

            FileReader fr = new FileReader(file);
            BufferedReader reader = new BufferedReader(fr);

            StringBuilder netappData = new StringBuilder();
            for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                netappData.append(line);
            }

            reader.close();
            fr.close();
            changes = updateSession.getUpdate(netappData.toString(), Schema.NETAPP);
            System.out.println(changes.getChangesCount() == 1 ? "1 change" : (changes.getChangesCount() + " changes"));
        } finally {
            if (changes == null) {
                System.out.println(" failed");
            }
        }
    }

    private void removeOriginator(Long originatorId) {
        try {
            System.out.print("Replay: delete originator " + originatorId + " ... ");

            updateSession.removeOriginUpdate(originatorId);
            System.out.println("done");
        } catch (UpdateTaskException e) {
            System.out.println(" failed");
            System.err.println("Error removing originator " + originatorId);
            e.printStackTrace();
        }
    }

    private boolean isLinux() {
        return System.getProperty("os.name").equals("Linux");
    }

    public static void main(String[] args) {
        // quiet log4j
        BasicConfigurator.configure(new NullAppender());

        String address = "localhost";

        final Options options = new Options()
                .addOption("f", "file", true, "Import file (can be a zip file or xml")
                .addOption("d", "dir", true, "Import directory containing recordings")
                .addOption("s", "server", true, "Server address (default is localhost)")
                .addOption("a", "all", false, "Don't ignore performance recordings with foundation copy")
                .addOption("r", "remove", true, "Remove originator");
        try {
            CommandLine commandLine = new GnuParser().parse(options, args);

            address = commandLine.hasOption("s") ? commandLine.getOptionValue("s") : address;
            boolean all = commandLine.hasOption("a");

            if (commandLine.hasOption('f')) {
                File file = new File(commandLine.getOptionValue('f'));
                if (file.exists() && file.isFile()) {
                    new Replay(address).runFile(file);
                } else {
                    System.out.println("Can't read " + file);
                }
            } else if (commandLine.hasOption('d')) {
                File dir = new File(commandLine.getOptionValue('d'));
                if (!dir.isDirectory()) {
                    System.out.println(dir + " is not a directory");
                }
                new Replay(address).runDir(dir, all);
            } else if (commandLine.hasOption('r')) {
                new Replay(address).removeOriginator(Long.valueOf(commandLine.getOptionValue('r')));
            } else {
                new HelpFormatter().printHelp("replay", options);
            }
        } catch (ParseException e) {
            e.printStackTrace();
        } catch (SessionCacheException e) {
            System.err.println("Failed to connect to server " + address);
            e.printStackTrace();
        }
    }
}
