package com.onaro.commons.prefs;

import java.util.prefs.*;
import java.io.NotSerializableException;


/**
 * Encapsulates an event emitted by a <code>PreferencesExt</code> node to indicate
 * that a preference has been added, removed or has had its value changed.<p>
 *
 * Note, that although PreferenceChangeEvent inherits Serializable interface
 * from EventObject, it is not intended to be Serializable. Appropriate
 * serialization methods are implemented to throw NotSerializableException.
 */
public class PreferenceExtChangeEvent extends java.util.EventObject {

    private static final long serialVersionUID = 1L;
    /**
     * The encapsulated event.
     */
    private PreferenceChangeEvent event;


    /**
     * Constructs a new <code>PreferenceChangeEvent</code> instance encapsulating
     * a <code>PreferenceChangeEvent</code>.
     *
     * @param event the encapsulated event
     */
    public PreferenceExtChangeEvent(PreferenceChangeEvent event) {
        super(event.getSource());
        this.event = event;
    }

    /**
     * Returns the preference node that emitted the event.
     *
     * @return  The preference node that emitted the event.
     */
    public PreferencesExt getNode() {
        return PreferencesExt.factory.get(event.getNode());
    }


    /**
     * Returns the key of the preference that was changed.
     *
     * @return  The key of the preference that was changed.
     */
    public String getKey() {
        return event.getKey();
    }


    /**
     * Returns the new value for the preference.
     *
     * @return  The new value for the preference, or <tt>null</tt> if the
     *          preference was removed.
     */
    public String getNewValue() {
        return event.getNewValue();
    }


    /**
     * Throws NotSerializableException, since NodeChangeEvent objects
     * are not intended to be serializable.
     *
     * @param out
     * @throws NotSerializableException
     */
     private void writeObject(java.io.ObjectOutputStream out)
                                               throws NotSerializableException {
         throw new NotSerializableException("Not serializable."); //$NON-NLS-1$
     }


    /**
     * Throws NotSerializableException, since PreferenceChangeEvent objects
     * are not intended to be serializable.
     *
     * @param in
     * @throws NotSerializableException
     */
     private void readObject(java.io.ObjectInputStream in)
                                               throws NotSerializableException {
         throw new NotSerializableException("Not serializable."); //$NON-NLS-1$
     }
}
