package com.onaro.commons.prefs;

import java.util.prefs.*;
import java.util.*;
import java.io.*;

/**
 * An implementation of the <code>PreferencesExtFactory</code> that uses
 * <code>PreferencesExtImpl</code>.
 */
public class PreferencesExtFactoryImpl implements PreferencesExtFactory {

    /**
     * Maps <code>Preferences</code> instances to <code>PreferencesExtImpl</code>
     * instances.
     */
    private Map prefToImpl = new HashMap();


    /**
     * Gets a <code>PreferencesExt</code> object for the specified <code>Preferences</code> node.
     *
     * @param node the node
     * @return a <code>PreferencesExt</code> encapsulating this node
     */
    public synchronized PreferencesExt get(Preferences node) {
        PreferencesExt pref = (PreferencesExt)prefToImpl.get(node);
        if (pref == null) {
            pref = new PreferencesExtImpl(node);
            prefToImpl.put(node, pref);
        }
        return pref;
    }


    /**
     * Gets the user preferences for the package of the specified class.
     *
     * @param c the class
     * @return the user preferences associated with the pacakge of the specified class
     */
    public PreferencesExt userNodeForPackage(Class c) {
        return get(Preferences.userNodeForPackage(c));
    }


    /**
     * Gets the system preferences for the package of the specified class.
     *
     * @param c the class
     * @return the system preferences associated with the pacakge of the specified class
     */
    public PreferencesExt systemNodeForPackage(Class c) {
        return get(Preferences.systemNodeForPackage(c));
    }


    /**
     * Gets the root node of the user preferences.
     *
     * @return the root node of the user preferences
     */
    public PreferencesExt userRoot() {
        return get(Preferences.userRoot());
    }


    /**
     * Gets the root node of the system preferences.
     *
     * @return the root node of the system preferences
     */
    public PreferencesExt systemRoot() {
        return get(Preferences.systemRoot());
    }


    /**
     * Imports user preferences from an XML file.
     *
     * @param is an input stream reading the XML file
     * @throws IOException if failed to read
     * @throws InvalidPreferencesFormatException if the input file isn't a valid
     *         XML preferences
     */
    public void importPreferences(InputStream is)
    throws IOException, InvalidPreferencesFormatException {
        Preferences.importPreferences(is);
    }
}
