package com.onaro.sanscreen.acquisition.cli;

import com.onaro.sanscreen.acquisition.cli.commands.Command;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PBE;
import com.onaro.util.sessions.ClientLogin;
import com.onaro.util.sessions.DefaultClientLogin;
import com.onaro.util.sessions.SessionCache;
import com.onaro.util.sessions.SessionCacheException;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLSession;

import java.net.MalformedURLException;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;

/**
 * Provides CLI execution functionality
 *
 * @see CommandLineMode For running this as a line-by-line commands
 */
public class CLI {

    private SessionCache sessionCache = null;
    private String user;

    /**
     * Executes the command
     *
     * @param command Command to execute
     * @param arguments arguments to pass on to the command
     *
     * @throws IllegalArgumentException On any argument parsing issue
     * @throws Exception                On any execution error
     */
    public void run(final Command command, final Map<String, String> arguments) throws Exception {
        command.run(this, arguments);
    }

    /**
     * Utility method to parse arguments into a map of key/value. If a flag is included, its value will be null.
     *
     * @param args Arguments
     * @param startFrom Start index to parse from
     *
     * @return Map of key/value of arguments parsed
     *
     * @throws IllegalArgumentException On any parsing error
     */
    public Map<String, String> parseArguments(final String[] args, final int startFrom) throws Exception {
        final Map<String, String> arguments = new HashMap<String, String>();

        for (int i = startFrom; i < args.length; i++) {
            if (!args[i].startsWith("-")) {
                throw new IllegalArgumentException("All arguments have to start with '-'");
            }

            final String remaining = args[i].substring(1);

            // Check out if this is a flag or a name/value pair
            int ind = remaining.indexOf('=');
            if (ind == -1) {
                // this is a flag
                arguments.put(remaining, null);
            }
            else {
                final String name = remaining.substring(0, ind);
                final String value = remaining.substring(ind + 1).trim();
                if (value.length() == 0) {
                    throw new IllegalArgumentException("No value assigned for argument '" + name + "'");
                }

                // Clean out passwords
                if(name.contains(".password")  && !"netappfoundation.password".equals(name)) {
                    System.out.println("encrypting " + name);
                    arguments.put(name, PBE.encrypt(value));
                }
                else {
                    arguments.put(name, value);
                }
            }
        }

        return arguments;
    }

    /**
     * Checks if certain attributes were included in the given attribute list
     *
     * @param arguments List of attributes (map of name/value)
     * @param checkValues If true, this method will also check that each mandatory attribute also has a non empty value
     * @param mandatoryArguments Set of mandatory attributes to check for
     *
     * @throws IllegalArgumentException On any missing attribute
     */
    public void checkMandatoryArguments(final Map<String, String> arguments, final boolean checkValues, final String... mandatoryArguments) {
        for (final String mandatoryArgument : mandatoryArguments) {
            if (!arguments.containsKey(mandatoryArgument)) {
                throw new IllegalArgumentException("Missing mandatory argument '" + mandatoryArgument + "'");
            }

            if (checkValues && arguments.get(mandatoryArgument) == null) {
                throw new IllegalArgumentException("Missing value for argument '" + mandatoryArgument + "'");
            }
        }
    }

    public void login(final String user, final String password) throws RemoteException, MalformedURLException {
        this.user = user;

        System.setProperty("com.onaro.rmiOverHttp", Boolean.TRUE.toString());

        final String jndiFactory = "https://localhost:443/invoker/JNDIFactorySSL";

        System.setProperty("java.naming.provider.url", jndiFactory);
        System.setProperty("java.naming.factory.initial", "org.jboss.naming.HttpNamingContextFactory");
        System.setProperty("java.naming.factory.url.pkgs", "org.jboss.naming:org.jnp.interfaces");
        System.setProperty("com.onaro.http.invoker", "https://localhost:443/invoker/JMXInvokerServlet");

        String home = System.getenv("SANSCREEN_HOME");
        // Reuse the certificate store from acquisition. This relies on acquisition service already running (and downloading
        // the certificates from the server).
        final String acqHome = System.getProperty("acq.home", home + "/acq");
        System.setProperty("javax.net.ssl.trustStore", acqHome + "/conf/cert/client.keystore");
        System.setProperty("javax.net.ssl.trustStorePassword", "changeit");

        // HEADSUP: This violates HTTPS host verification, since we aren't setting up SSL certs right.
        final HostnameVerifier hv = new HostnameVerifier() {
            @Override
            public boolean verify(final String urlHostName, final SSLSession session) {
                return true;
            }
        };

        HttpsURLConnection.setDefaultHostnameVerifier(hv);

        new DefaultClientLogin(user, password).login();
    }

    public synchronized <E> E getSession(final String sessionJndiName) throws SessionCacheException {
        if (sessionCache == null) {
            sessionCache = connectToServer();
        }

        return (E) sessionCache.getSession(sessionJndiName);
    }

    /**
     * Connects to the SANscreen server and gets the initial context.
     *
     * @return sessionCache object
     *
     * @throws SessionCacheException On any error
     */
    private SessionCache connectToServer() throws SessionCacheException {
        return new SessionCache(this.getClass().getClassLoader(), new ClientLogin() {
            @Override
            public void login() {
            }

            @Override
            public boolean isLoggedIn() {
                return true;
            }

            @Override
            public void clean() {
            }

            @Override
            public String getUser() {
                return user;
            }
        }, 2000);
    }

}
