package com.onaro.sanscreen.acquisition.cli;

import com.onaro.sanscreen.acquisition.cli.commands.Command;
import com.onaro.sanscreen.acquisition.cli.commands.CommandType;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Map;

/**
 * Implements the script use case (login once into a session and then run through a given script of commands)
 */
public class ScriptMode {
    public static void main(final String[] args) {
        final CLI cli = new CLI();
        BufferedReader reader = null;

        try {
            // First parse the argument... they have to include user and password
            final Map<String, String> arguments = cli.parseArguments(args, 0);

            if (arguments.size() != 3)
                throw new IllegalArgumentException("3 arguments (user, password and script are required)");

            cli.checkMandatoryArguments(arguments, true, "user", "password", "script");

            final String user = arguments.get("user");
            final String password = arguments.get("password");
            reader = new BufferedReader(new FileReader(arguments.get("script")));

            // Login (once) into the server
            cli.login(user, password);
        }
        catch (IllegalArgumentException e) {
            if (e != null) {
                System.out.println(e.getMessage());
                System.out.println();
            }

            System.out.println("Usage:");
            System.out.println("======");
            System.out.println("acqcli_script -user=<user> -password=<password> -script=<path_no_spaces>");

            System.exit(1);
        }
        catch (final Exception e) {
            System.err.println("Error: " + e.getMessage());
            System.exit(1);
        }

        assert(reader != null);
        System.out.println("Loading script...");

        int lineCount = 1;

        try {
            // Now start the loop of the command line parsing...
            boolean stopped = false;
            while (!stopped) {
                try {
                    final String line = reader.readLine().trim();

                    // Check end of file
                    if (line == null)
                        break;

                    if (line.length() == 0)
                        continue;

                    final String[] params = line.split("\\s+");

                    // First argument has to be a command
                    if (params.length < 1)
                        throw new IllegalArgumentException("No arguments provided (Line: " + lineCount + ")");

                    final Command command = CommandType.fromCommandString(params[0]).getCommand();
                    final Map<String, String> arguments = cli.parseArguments(params, 1);

                    cli.run(command, arguments);
                    System.out.println();
                }
                catch (final Exception e) {
                    System.err.println("Error: " + e.getMessage() + " (Line: " + lineCount + ")");
                    System.exit(1);
                }

                lineCount++;
            }

            System.out.println("Done.");
        }
        finally {
            try {
                reader.close();
            }
            catch (IOException e) {
                // Fall through...
            }
        }
    }
}
