package com.onaro.sanscreen.acquisition.datasource.netapp_ocie;

import com.netapp.oci.platform.acq.interfaces.data.PackagePollCompleteStatus;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollStartStatus;
import com.netapp.oci.platform.acq.interfaces.remote.AcquisitionStatusRemote;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.ClusterBuilder;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppOCIEConfiguration;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeCluster;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeClusterRoot;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll.PartialPollContext;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transform.ocie.NetAppOCIETransform;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppClockSynchronizer;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationUpdatesListener;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppOriginatorProvider;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.RootFactory;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.StorageRoot;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionUnitException;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class NetAppOCIEDataSource extends NetAppFoundationPackage<NetAppOCIEConfiguration> implements NetAppOCIEDataSourceMBean {

    private static final Logger LOG = LoggerFactory.getLogger(NetAppOCIEDataSource.class);
    private PartialPollContext partialPollContext;

    /**
     * For extracting fields from server generated error message for duplicate data source.
     * Like:
     * Another datasource was found acquiring the same device (deviceKey - 14477215-1f6b-11e1-a744-123478563412).
     * To correct the problem, please remove one of the duplicate datasources: [ocn2, ocn]
     */
    private static final Pattern DUPLICATE_DS_MESSAGE_PATTERN = Pattern.compile(
            "^Another datasource was found acquiring the same device \\(deviceKey - (\\p{Alnum}{8}-\\p{Alnum}{4}-\\p{Alnum}{4}-\\p{Alnum}{4}-\\p{Alnum}{12})\\). " +
            "To correct the problem, please remove one of the duplicate datasources: \\[(.*)\\].*");

    /**
     * Flag to enable/disable debug of the transport layer (Zapi)
     */
    private boolean debugTransport = false;

    /**
     * Create a new NetAppOCIDataSource instance
     *
     * @param dataSource                  the parent data source
     * @param configuration               the data source configuration
     * @param dataSourceLifeCycleListener A listener to receive DS life cycle events (to write audits, etc...)
     * @param dataSourceSampleListener    A listener to receive sample events (to write audits, etc...)
     * @param foundationUpdatesListener   A listener to receive foundation updates
     * @param originatorProvider          Originator ID provider
     */
    public NetAppOCIEDataSource(BaseDataSource dataSource, NetAppOCIEConfiguration configuration,
            DataSourceLifeCycleListener dataSourceLifeCycleListener, DataSourceSampleListener dataSourceSampleListener,
            NetAppFoundationUpdatesListener foundationUpdatesListener, NetAppOriginatorProvider originatorProvider,
            NetAppClockSynchronizer clockSynchronizer)
    {
        super(dataSource, configuration, dataSourceLifeCycleListener,
                dataSourceSampleListener, foundationUpdatesListener,
                originatorProvider, clockSynchronizer);

      //Create the partial polling context
      String baseDir = ServiceManager.getInstance().getProperty( "PARTIAL_POLLING_BASE_DIR" );
      this.partialPollContext = new PartialPollContext( dataSource.getDataSourceName(), baseDir  );

      //Enabled/Disable partial polling from the general properties file.
      String partialPollingEnabled = ServiceManager.getInstance().getProperty( "SUPPORT_PARTIAL_POLLING" );
      partialPollContext.setGloballyEnabled( partialPollingEnabled == null ? true : Boolean.valueOf( partialPollingEnabled ) );
    }

    @Override
    protected void doStart() throws FrameworkException {

    }

    @Override
    protected void doStop() {

    }

    @Override
    public boolean isDebugTransport() {
        return debugTransport;
    }

    @Override
    public void setDebugTransport(final boolean debugTransport) {
        this.debugTransport = debugTransport;
    }

    @Override
    protected void doFoundation(final BaseDataSource dataSource, final Properties properties) throws SchemaException, FrameworkException, DataSourceErrorException {
        sendPollStart();
        final Long datasourceID = this.getDataSource().getDataSourceID();
        final String datasourceName = this.getDataSource().getDataSourceName();
        final String packageName = PackageContext.getPackage().getPackageName();
        final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus =
                new NetAppFoundationConfiguration.AcquisitionPollStatus(datasourceID, datasourceName, packageName, -1);
        this.getConfiguration().putAcquisitionPollStatus(packageName, pollStatus);
        long zapiDuration = -1;
        final long pollStartTime = System.currentTimeMillis();
        try {
            CModeClusterRoot clusterModelRoot = new CModeClusterRoot();
            final ClusterBuilder clusterBuilder = new ClusterBuilder( dataSource, partialPollContext, properties );
            final CModeCluster cModeCluster = clusterBuilder.validateTargetCluster(clusterModelRoot);
            final Long originatorID = getOriginatorID(cModeCluster.getClusterUUID());
            zapiDuration = clusterBuilder.executeAndBuild(clusterModelRoot);
            recordModelData(clusterModelRoot);
            final String dsName = getConfiguration().getDataSourceName();
            final RootFactory rootFactory = new RootFactory(dsName);
            final StorageRoot storageRoot = rootFactory.createStorageRoot(originatorID, dsName);
            final NetAppOCIETransform ocieTransform = new NetAppOCIETransform(clusterModelRoot, storageRoot);
            final NetAppOCIEConfiguration config = getConfiguration();
            ocieTransform.transform(config);

            if (getConfiguration().isCrawling()) {
                try {
                    final String custom = Long.toString(storageRoot.getOriginator().getOriginatorID());
                    final DSFileCRCWriter dsFileCRCWriter = getRecordingSample().createTextualOutputCRCFile(getConfiguration(), DataSourceOutputFileType.netappdata, custom, "xml");
                    marshalAndCRCNetAppData(storageRoot, dsFileCRCWriter);
                    dsFileCRCWriter.close();
                } catch (IOException ioe) {
                    LOG.error("{} - Failed to record Netapp data", PackageContext.getPackage().getLogPrefix(), ioe);
                }
                return;
            }

            //noinspection UnusedAssignment
            clusterModelRoot = null;            // we don't need this anymore and would like to free resources
            // report the data
            reportNetAppData(storageRoot, dsName);

        } catch (final DataSourceErrorException dsee) {
            reportDataSourceErrorException(pollStatus, dsee);
            throw dsee;
        } catch (FrameworkException fe) {
            reportFrameworkException(pollStatus, fe);
            throw fe;
        } finally {

            // Record poll completion details so listeners may report end of poll status to OCIE server
            final long pollEndTime = System.currentTimeMillis();
            final int pollDuration = (pollEndTime == -1) ? -1 : (int) (pollEndTime - pollStartTime);
            this.getConfiguration().putAcquisitionPollStatus(packageName, pollStatus);
            pollStatus.setPollDuration(pollDuration);
            pollStatus.setTotalZAPITime(zapiDuration);
            final NetAppFoundationPackage<?> netappFoundationPackage = (NetAppFoundationPackage<?>)PackageContext.getPackage();
            pollStatus.setPollStartTime(netappFoundationPackage.getCachedServerTime());
            pollStatus.setTotalReportTime(getOcieReportDuration());
            recordPollStatus(pollStatus);
        }
    }

    private void recordPollStatus(final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus) {

        final Package<?> thePackage = PackageContext.getPackage();
        final String packageName = thePackage.getPackageName();
        this.getConfiguration().putAcquisitionPollStatus(packageName, pollStatus);
        final String dsName = thePackage.getConfiguration().getDataSourceName();
        DSFileWriter pollStatusWriter = null;
        try {
            pollStatusWriter = thePackage.getRecordingSession().getCurrentSample().createTextualOutputFile(thePackage.getConfiguration(),
                    DataSourceOutputFileType.log,
                    "poll_time", "txt");
            pollStatusWriter.writeln("datasource, package, poll start, poll duration, zapi time, report time");
            pollStatusWriter.writeln(", ", dsName, packageName,
                    pollStatus.getPollStartTime(), pollStatus.getPollDuration(),
                    pollStatus.getTotalZAPITime(), pollStatus.getTotalReportTime());
        }
        catch (IOException ioe) {
            LOG.error("{} - Failed to record ZAPI response times", thePackage.getLogPrefix(), ioe);
        }
        finally {
            if (pollStatusWriter != null) {
                try {
                    pollStatusWriter.close();
                }
                catch (IOException ignore) {}
            }
        }
    }

    private void reportDataSourceErrorException(final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus,
            final DataSourceErrorException dsee) {

        if (dsee instanceof AggregatedDataSourceErrorsException) {
            final AggregatedDataSourceErrorsException aggregateException = (AggregatedDataSourceErrorsException) dsee;
            for (final DataSourceErrorException idsee : aggregateException.getExceptions()) {
                reportDataSourceErrorException(pollStatus, idsee);
            }
            return;
        }
        final NetAppOCIEConfiguration config = getConfiguration();
        final String managementIP = config.getIP();
        final Integer port = config.getPort();
        final String message = dsee.getLogMessage();
        final Map<String, String> errorProperties = new HashMap<String, String>();
        final Map<String, String> errorParameters = dsee.getTemplateConverterParameters();
        final String zapiCommand = errorParameters.get("command");
        errorProperties.put("ClusterManagementIP", managementIP);
        errorProperties.put("Port", String.valueOf(port));
        errorProperties.put("Message", message);
        errorProperties.put("Zapi", zapiCommand);
        final DataSourceErrorType errorType = dsee.getType();
        final PackagePollCompleteStatus.Error error;
        switch (errorType) {
            case CONNECTING:
                if (message.startsWith("target lif") && message.contains("is not a CLUSTER_MGMT")) {
                    error = PackagePollCompleteStatus.Error.BAD_LIF_ERROR;
                } else {
                    error = PackagePollCompleteStatus.Error.NETWORK_ACCESS_FAILURE;
                }
                break;

            case INVALID_LOGIN_CREDENTIALS:
                error = PackagePollCompleteStatus.Error.CREDENTIAL_FAILURE;
                final String userName = config.getUser();
                errorProperties.put("Username", userName);
                break;

            case RETRIEVE_DATA_FROM_DEVICE:
                error = PackagePollCompleteStatus.Error.BAD_ZAPI_RESPONSE;
                break;

            case INTERNAL_ERROR:
                error = PackagePollCompleteStatus.Error.INTERNAL_ERROR;
                break;

            default:
                error = PackagePollCompleteStatus.Error.INTERNAL_ERROR;
                break;

        }

        final Map<PackagePollCompleteStatus.Error, Map<String, String>> pollError = new HashMap<PackagePollCompleteStatus.Error, Map<String, String>>();
        pollError.put(error, errorProperties);
        pollStatus.addDataSourceError(pollError);
    }

    private void reportFrameworkException(final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus,
            final FrameworkException fe) {

        final String message;
        final Throwable cause = fe.getCause();
        if (cause instanceof AcquisitionUnitException) {
            // We have to parse this server-side generated message
            final AcquisitionUnitException aue = (AcquisitionUnitException) cause;
            message = aue.getMessage();
        } else {
            message = fe.getMessage();
        }

        final NetAppOCIEConfiguration config = getConfiguration();
        final String managementIP = config.getIP();
        final Integer port = config.getPort();
        final PackagePollCompleteStatus.Error error;
        final Map<String, String> errorProperties = new HashMap<String, String>();
        errorProperties.put("ClusterManagementIP", managementIP);
        errorProperties.put("Port", String.valueOf(port));
        errorProperties.put("Message", message);

        final Matcher messageMatch = DUPLICATE_DS_MESSAGE_PATTERN.matcher(message);
        if (messageMatch.matches()) {
            error = PackagePollCompleteStatus.Error.DUPLICATE_DATASOURCE;
            final String deviceKey = messageMatch.group(1);
            errorProperties.put("DeviceKey", deviceKey);
            final String duplicatesCSV = messageMatch.group(2);
            errorProperties.put("Duplicates", duplicatesCSV);
        } else {
            error = PackagePollCompleteStatus.Error.INTERNAL_ERROR;
        }

        final Map<PackagePollCompleteStatus.Error, Map<String, String>> pollError = new HashMap<PackagePollCompleteStatus.Error, Map<String, String>>();
        pollError.put(error, errorProperties);
        pollStatus.addDataSourceError(pollError);
    }

    private void sendPollStart() {
        final ServiceManager serviceManager = ServiceManager.getInstance();
        if (serviceManager.isStandalone()) {
            return;
        }
        final NetAppFoundationPackage<?> netappFoundationPackage = (NetAppFoundationPackage<?>)PackageContext.getPackage();
        final BaseDataSource thisDataSource = this.getDataSource();
        final Long datasourceID = thisDataSource.getDataSourceID();
        final String datasourceName = thisDataSource.getDataSourceName();
        final String packageName = netappFoundationPackage.getPackageName();
        final long cachedServerTime = netappFoundationPackage.getCachedServerTime();
        final PackagePollStartStatus pollStartStatus = new PackagePollStartStatus(datasourceName, datasourceID, packageName, cachedServerTime);
        try {
            final AcquisitionStatusRemote acqStatusSession = serviceManager.getCommunicationManager().getSession(AcquisitionStatusRemote.JNDI_NAME);
            if (acqStatusSession == null) {
                LOG.error("Failed to update AcquisitionStatusRemote packageStarted {} {}",
                        dataSource.getDataSourceName(), netappFoundationPackage.getPackageName());
                return;
            }
            acqStatusSession.packageStarted(NetAppFoundationConfiguration.AcquisitionPollStatus.schema, pollStartStatus);
        }
        catch (FrameworkException fe) {
            LOG.error("Failed to update AcquisitionStatusRemote packageStarted {} {}",
                    dataSource.getDataSourceName(), netappFoundationPackage.getPackageName());
        }
        catch (ServerInterfaceException sie) {
            LOG.error("Failed to update AcquisitionStatusRemote packageStarted {} {}",
                    dataSource.getDataSourceName(), netappFoundationPackage.getPackageName());
        }
    }
}
