package com.onaro.sanscreen.acquisition.datasource.netapp_ocie;

import com.netapp.oci.platform.acq.interfaces.data.PackagePollCompleteStatus;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollStartStatus;
import com.netapp.oci.platform.acq.interfaces.remote.AcquisitionStatusRemote;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.PerformanceArchiveManager;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance.*;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppOCIEPerformanceConfiguration;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.common.SyntheticCounterManager;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnectionFactory;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppClockSynchronizer;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformancePackage;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceReporter;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformanceUpdatesListener;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.GlobalScratchPad;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

public class NetAppOCIEPerformancePackage extends NetAppStoragePerformancePackage<NetAppOCIEPerformanceConfiguration, NetAppDefaultStorageContext> implements NetAppStoragePerformanceReporter {
    private static final Logger LOG = LoggerFactory.getLogger(NetAppOCIEPerformancePackage.class);
    private final VolumeMetricsBuilder volumeMetricsBuilder;
    private final LunMetricsBuilder lunMetricsBuilder;
    private final CifsVserverMetricsBuilder cifsVserverMetricsBuilder;
    private final LifVserverMetricsBuilder lifVserverMetricsBuilder;
    private final Nfsv3MetricsBuilder nfsv3MetricsBuilder;
    private final Nfsv4MetricsBuilder nfsv4MetricsBuilder;
    private final Nfsv41MetricsBuilder nfsv41MetricsBuilder;
    private final NodeMetricsBuilder nodeMetricsBuilder;
    private final AggregateMetricsBuilder aggregateMetricsBuilder;
    private final DiskMetricsBuilder diskMetricsBuilder;
    private final FcpLifMetricsBuilder fcpLifMetricsBuilder;
    private final FcpLifVserverMetricsBuilder fcpLifVserverMetricsBuilder;
    private final FcpPortMetricsBuilder fcpPortMetricsBuilder;
    private final IpPortMetricsBuilder ipPortMetricsBuilder;
    private final NicCommonMetricsBuilder nicCommonMetricsBuilder;
    private final VolumeVserverMetricsBuilder volumeVserverMetricsBuilder;
    private final LifMetricsBuilder lifMetricsBuilder;
    private final ProcessorMetricsBuilder processorMetricsBuilder;
    private final ISCSILIFMetricsBuilder iSCSILifMetricsBuilder;
    private final IscsiLifVserverMetricsBuilder iscsiLifVserverMetricsBuilder;
    private final QosWorkloadDetailMetricsBuilder qosWorkloadDetailMetricsBuilder;
    private final PerformanceArchiveManager performanceArchiveManager;
    private final QosWorkloadConstituentMetricsBuilder qosWorkloadConstituentMetricsBuilder;
    private final QosServiceCenterMetricsBuilder qosServiceCenterMetricsBuilder;
    private final QosWorkloadMetricsBuilder qosWorkloadMetricsBuilder;
    private final QosWorkloadQueueNBladeMetricsBuilder qosWorkloadQueueNBladeMetricsBuilder;
    private final QosWorkloadQueueDBladeMetricsBuilder qosWorkloadQueueDBladeMetricsBuilder;
    private final WaflHyaAggrMetricsBuilder waflHyaAggrMetricsBuilder;

    private final CIFSNodeMetricsBuilder cifsNodeMetricsBuilder;
    private final NFSv3NodeMetricsBuilder nfSv3NodeMetricsBuilder;
    private final NFSv4NodeMetricsBuilder nfSv4NodeMetricsBuilder;
    private final NFSv41NodeMetricsBuilder nfSv41NodeMetricsBuilder;
    private final SMB1NodeMetricsBuilder smb1NodeMetricsBuilder;
    private final SMB2NodeMetricsBuilder smb2NodeMetricsBuilder;
    private final WaflMetricsBuilder waflMetricsBuilder;

    private final HostadapterMetricsBuilder hostadapterMetricsBuilder;
    
    private final ClusterMetricsBuilder clusterMetricsBuilder;

    private final SyntheticCounterManager syntheticCounterManager;

    private final boolean metricsDebugEnabled;
    

    private long sampleTime;
    
    private DataSourceErrorException acquisitionFailedException;

    private final GlobalScratchPad globalScratchPad;

    public NetAppOCIEPerformancePackage(final BaseDataSource dataSource, final NetAppOCIEPerformanceConfiguration configuration, final DataSourceSampleListener dataSourceSampleListener, final NetAppStoragePerformanceUpdatesListener storagePerformanceListener, final NetAppClockSynchronizer clockSynchronizer) {
        super(dataSource, configuration, dataSourceSampleListener, storagePerformanceListener, clockSynchronizer);
        metricsDebugEnabled = configuration.isMetricsDebugEnabled();
        volumeMetricsBuilder = new VolumeMetricsBuilder(this);
        lunMetricsBuilder = new LunMetricsBuilder(this);
        nodeMetricsBuilder = new NodeMetricsBuilder(this);
        cifsVserverMetricsBuilder = new CifsVserverMetricsBuilder(this);
        lifVserverMetricsBuilder = new LifVserverMetricsBuilder(this);
        nfsv3MetricsBuilder = new Nfsv3MetricsBuilder(this);
        nfsv4MetricsBuilder = new Nfsv4MetricsBuilder(this);
        nfsv41MetricsBuilder = new Nfsv41MetricsBuilder(this);
        aggregateMetricsBuilder = new AggregateMetricsBuilder(this);
        diskMetricsBuilder = new DiskMetricsBuilder(this);
        fcpLifMetricsBuilder = new FcpLifMetricsBuilder(this);
        fcpLifVserverMetricsBuilder = new FcpLifVserverMetricsBuilder(this);
        fcpPortMetricsBuilder = new FcpPortMetricsBuilder(this);
        ipPortMetricsBuilder = new IpPortMetricsBuilder(this);
        nicCommonMetricsBuilder = new NicCommonMetricsBuilder(this);
        volumeVserverMetricsBuilder = new VolumeVserverMetricsBuilder(this);
        lifMetricsBuilder = new LifMetricsBuilder(this);
        processorMetricsBuilder = new ProcessorMetricsBuilder(this);
        iSCSILifMetricsBuilder = new ISCSILIFMetricsBuilder(this);
        iscsiLifVserverMetricsBuilder = new IscsiLifVserverMetricsBuilder(this);
        waflHyaAggrMetricsBuilder = new WaflHyaAggrMetricsBuilder(this);
        performanceArchiveManager = new PerformanceArchiveManager(this);
        qosWorkloadDetailMetricsBuilder = new QosWorkloadDetailMetricsBuilder(this);
        // Service Center counters contain sums of workload detail counters, therefore they are
        // dependent on workload detail counters.
        qosServiceCenterMetricsBuilder = new QosServiceCenterMetricsBuilder(this, qosWorkloadDetailMetricsBuilder);
        
        // volume workload counters contain sums of workload constituent counters, therefore they are
        // dependent on workload constituent counters.
        qosWorkloadConstituentMetricsBuilder = new QosWorkloadConstituentMetricsBuilder(this);
        qosWorkloadMetricsBuilder = new QosWorkloadMetricsBuilder(this, qosWorkloadConstituentMetricsBuilder);
        
        qosWorkloadQueueNBladeMetricsBuilder = new QosWorkloadQueueNBladeMetricsBuilder(this,qosWorkloadDetailMetricsBuilder, qosWorkloadConstituentMetricsBuilder);
        qosWorkloadQueueDBladeMetricsBuilder = new QosWorkloadQueueDBladeMetricsBuilder(this, qosWorkloadDetailMetricsBuilder,qosWorkloadConstituentMetricsBuilder);
        cifsNodeMetricsBuilder = new CIFSNodeMetricsBuilder(this);
        nfSv3NodeMetricsBuilder = new NFSv3NodeMetricsBuilder(this);
        nfSv4NodeMetricsBuilder = new NFSv4NodeMetricsBuilder(this);
        nfSv41NodeMetricsBuilder = new NFSv41NodeMetricsBuilder(this);
        smb1NodeMetricsBuilder = new SMB1NodeMetricsBuilder(this);
        smb2NodeMetricsBuilder = new SMB2NodeMetricsBuilder(this);

        globalScratchPad = new GlobalScratchPad();
        syntheticCounterManager = new SyntheticCounterManager(this);
        waflMetricsBuilder = new WaflMetricsBuilder(this);
        
        hostadapterMetricsBuilder = new HostadapterMetricsBuilder(this);
        clusterMetricsBuilder = new ClusterMetricsBuilder(this, nodeMetricsBuilder);
    }

    @Override
    protected NetAppDefaultStorageContext createStorageContext(final NetAppStoragePerformancePackage thePackage, final Cluster.ClusterKey key) {
        return new NetAppDefaultStorageContext(key);
    }

    @Override
    protected void doStart() throws FrameworkException, DataSourceErrorException {
    }

    @Override
    protected void doStop() throws FrameworkException, DataSourceErrorException {
    }

    @Override
    protected void doArrayPerformance(final BaseDataSource dataSource) throws FrameworkException, DataSourceErrorException {
        final Package thePackage = PackageContext.getPackage();
        sendPollStart();
        ZAPIConnection zapiConnection = null;
        acquisitionFailedException = null;
        final Long datasourceID = this.getDataSource().getDataSourceID();
        final String datasourceName = this.getDataSource().getDataSourceName();
        final String packageName = PackageContext.getPackage().getPackageName();
        final long pollStartTime = System.currentTimeMillis();
        final NetAppOCIEPerformanceConfiguration config = getConfiguration();
        final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus = new NetAppFoundationConfiguration.AcquisitionPollStatus(datasourceID, datasourceName, packageName, -1);
        this.getConfiguration().getNetAppFoundationConfiguration().putAcquisitionPollStatus(packageName, pollStatus);

        // Create recorded log file report times so each reportSampleXXX call made by
        // NetAppStoragePerformanceListener can be recorded.
        DSFileWriter reportTimesWriter = null;
        try {
            reportTimesWriter = thePackage.getRecordingSession().getCurrentSample().createTextualOutputFile(thePackage.getConfiguration(), DataSourceOutputFileType.log, "ocie_report_times", "txt");
            reportTimesWriter.writeln("datasource, api, parent UUID, poll start time, poll start time, report start time, report duration, sample size");
        }
        catch (IOException ioe) {
            LOG.warn("Failed to initialize ocie_report_times log", ioe);
        }

        try {
            final NetAppDefaultStorageContext context = getNetAppDefaultStorageContext();
            if (context == null) {
                // No previous foundation data, don't attempt performance
                return;
            }
            boolean isQosDiscoveryEnable = !context.getQosVolumeWorkloadInstanceUUIDs().isEmpty();
            
            // Need to get the data source attributes so we can see if some of the 
            // flags are enabled.
            //
            // In particular we want to look at the:
            //
            //  o - cookedpersistence.enabled flag
            //  o - rawpersistence.enabled flag
           //
            context.setCookedPersistenceEnabled(config.isCookedStatisticsPersistenceEnabled());
            context.setRawPersistenceEnabled(config.isRawStatisticsPersistenceEnabled());

            zapiConnection = ZAPIConnectionFactory.createConnection();
            
            // if the cluster is Fullsteam or above then sample time is the time on the cluster (read from the archiver file) 
            // else it is simply the current server time.  
            try{
                setSampleTime(context, zapiConnection, isQosDiscoveryEnable);
            }catch(DataSourceErrorException dsee){
                clearAllPreviousValues();
                throw dsee;
            }
            
        	getAndReportPerformanceMetrics(waflHyaAggrMetricsBuilder, context, zapiConnection);
        	getAndReportPerformanceMetrics(ipPortMetricsBuilder, context, zapiConnection);
            getAndReportPerformanceMetrics(cifsNodeMetricsBuilder,context, zapiConnection);
            getAndReportPerformanceMetrics(nfSv3NodeMetricsBuilder,context, zapiConnection);
            getAndReportPerformanceMetrics(nfSv4NodeMetricsBuilder,context, zapiConnection);
            getAndReportPerformanceMetrics(nfSv41NodeMetricsBuilder,context, zapiConnection);
            getAndReportPerformanceMetrics(smb1NodeMetricsBuilder,context, zapiConnection);
            getAndReportPerformanceMetrics(smb2NodeMetricsBuilder,context, zapiConnection);

            //these counters does not exist in FS.0 archiver; only available in FS.1 archiver.
            //Therefore for FS.0 and SN, ZAPI call is required.  In OPMPerformanceArchive, these
            // are called only for FS.1 and up.
            if(!context.isGreaterThanFullSteam0()){
            	getAndReportPerformanceMetrics(waflMetricsBuilder, context, zapiConnection);
              	getAndReportPerformanceMetrics(processorMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(diskMetricsBuilder, context, zapiConnection);
                //AggregateBuilder should be after diskBuilder
                getAndReportPerformanceMetrics(aggregateMetricsBuilder,context, zapiConnection);
                getAndReportPerformanceMetrics(hostadapterMetricsBuilder, context, zapiConnection);
              	//wafl, disk and processor should always be before nodeBuilder, since nodeMetricsBuilder uses these scratchPads.
              	getAndReportPerformanceMetrics(nodeMetricsBuilder, context, zapiConnection);
                clusterMetricsBuilder.getAndReportPerformance(context);
            }

            if(context.isLessThanFullSteam()){
              	getAndReportPerformanceMetrics(volumeMetricsBuilder, context, zapiConnection);
              	getAndReportPerformanceMetrics(lifMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(nicCommonMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(iSCSILifMetricsBuilder, context, zapiConnection);
                // Do not change the fcpLif and fcpPort order as fcpPort depends on fcpLif from global scratchpad.
                getAndReportPerformanceMetrics(fcpLifMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(fcpPortMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(lunMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(cifsVserverMetricsBuilder, context, zapiConnection);
            	getAndReportPerformanceMetrics(lifVserverMetricsBuilder, context, zapiConnection);
            	getAndReportPerformanceMetrics(iscsiLifVserverMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(nfsv3MetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(nfsv4MetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(nfsv41MetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(fcpLifVserverMetricsBuilder, context, zapiConnection);
                getAndReportPerformanceMetrics(volumeVserverMetricsBuilder, context, zapiConnection);
            } else {
                performanceArchiveManager.getAndReportPerformance(context, zapiConnection);
            }
            

            if (isQosDiscoveryEnable) {
                if(!context.isGreaterThanFullSteam0()){
                    getAndReportPerformanceMetrics(qosWorkloadDetailMetricsBuilder,context, zapiConnection);
                }
                if (context.isLessThanFullSteam()) {
                    getAndReportPerformanceMetrics(qosWorkloadConstituentMetricsBuilder,context, zapiConnection);
                    /*
                     * Qos Service Center counters contain sums of workload detail counters, therefore we first
                     *  record workload details counters before reporting service center counters
                     */
                    getAndReportPerformanceMetrics(qosServiceCenterMetricsBuilder,context, zapiConnection);
                    /*
                     * Qos Workload counters contain sums of workload constituent counters, therefore we first
                     *  record workload constituent counters before reporting volume workload counters
                     */
                    qosWorkloadMetricsBuilder.getAndReportPerformance(context);
                    /*
                     * Both Qos workload queue dblade and workload queue nblade counters are derived from workload 
                     * detail and workload constituent counters, therefore we first record workload detail and workload 
                     * constituent counters before reporting workload queue dblade and workload queue nblade counters.
                     */
                    qosWorkloadQueueNBladeMetricsBuilder.getAndReportPerformance(context);
                    qosWorkloadQueueDBladeMetricsBuilder.getAndReportPerformance(context);
                } 
            }
            syntheticCounterManager.getAndReportPerformance(context, zapiConnection);

            zapiConnection.recordAttributes();
            if(acquisitionFailedException != null){
                throw acquisitionFailedException;
            }
        }
            catch (DataSourceErrorException dsee) {
            reportDataSourceErrorException(pollStatus, dsee);
            throw dsee;
        }
        finally {
            //Poll complete now clear the global scratchpad.
            globalScratchPad.clear();
            if (reportTimesWriter != null) {
                try {
                    reportTimesWriter.close();
                }
                catch (IOException ioe) {
                    // ignore
                }
            }
            if (zapiConnection != null) {
                zapiConnection.close();
            }
            final long pollEndTime = System.currentTimeMillis();
            final int pollDuration = (pollEndTime == -1) ? -1 : (int) (pollEndTime - pollStartTime);
            pollStatus.setPollDuration(pollDuration);
            final long zapiDuration = (zapiConnection != null) ? zapiConnection.getTotalZAPILatency() : -1;
            pollStatus.setTotalZAPITime(zapiDuration);
            pollStatus.setPollStartTime(getCachedServerTime());
            pollStatus.setSampleTime(getSampleTime());
            recordPollStatus(pollStatus);
        }
    }

    private void setSampleTime(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection, final boolean qosDiscoveryEnable) throws DataSourceErrorException {
        if (context.isLessThanFullSteam() || !qosDiscoveryEnable) {
            this.sampleTime = super.getSampleTime();
        } else {
            try {
                this.sampleTime = performanceArchiveManager.getArchiveSampleTime(context, zapiConnection);
            } catch (DataSourceErrorException dataSourceExe) {
                this.sampleTime = super.getSampleTime();
                LOG.equals("Unable to retrieve samples from the performance archive DataSourceErrorException: " + dataSourceExe.getMessage());
                throw dataSourceExe;
            }
        }
    }

    public long getSampleTime() {
        return sampleTime;
    }

    private void recordPollStatus(final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus) {
        final Package thePackage = PackageContext.getPackage();
        final String packageName = thePackage.getPackageName();
        final String dsName = thePackage.getConfiguration().getDataSourceName();
        DSFileWriter pollStatusWriter = null;
        try {
            pollStatusWriter = thePackage.getRecordingSession().getCurrentSample().createTextualOutputFile(thePackage.getConfiguration(),
                                                                                                           DataSourceOutputFileType.log,
                                                                                                           "poll_time", "txt");
            pollStatusWriter.writeln("datasource, package, poll start, poll duration, zapi time, report time");
            pollStatusWriter.writeln(dsName + ", "
                                     + packageName + ", "
                                     + pollStatus.getPollStartTime() + ", "
                                     + pollStatus.getPollDuration() + ", "
                                     + pollStatus.getTotalZAPITime() + ", "
                                     + pollStatus.getTotalReportTime());
        }
        catch (IOException ioe) {
            LOG.error(thePackage.getLogPrefix() + " - Failed to record ZAPI response times: " + ioe, ioe);
        }
        finally {
            if (pollStatusWriter != null) {
                try {
                    pollStatusWriter.close();
                }
                catch (IOException ioe) {
                    // ignore
                }
            }
        }
    }

    private void reportDataSourceErrorException(final NetAppFoundationConfiguration.AcquisitionPollStatus pollStatus,
                                                final DataSourceErrorException dsee) {

        if (dsee instanceof AggregatedDataSourceErrorsException) {
            final AggregatedDataSourceErrorsException aggregateException = (AggregatedDataSourceErrorsException) dsee;
            for (final DataSourceErrorException idsee : aggregateException.getExceptions()) {
                reportDataSourceErrorException(pollStatus, idsee);
            }
            return;
        }
        final NetAppOCIEPerformanceConfiguration config = getConfiguration();
        final String managementIP = config.getIP();
        final Integer port = config.getPort();
        final String message = dsee.getLogMessage();
        final Map<String, String> errorProperties = new HashMap<String, String>();
        errorProperties.put("ClusterManagementIP", managementIP);
        errorProperties.put("Port", String.valueOf(port));
        errorProperties.put("Message", message);
        final DataSourceErrorType errorType = dsee.getType();
        final PackagePollCompleteStatus.Error error;
        switch (errorType) {
            case CONNECTING:
                if (message.startsWith("target lif") && message.contains("is not a CLUSTER_MGMT")) {
                    error = PackagePollCompleteStatus.Error.BAD_LIF_ERROR;
                }
                else {
                    error = PackagePollCompleteStatus.Error.NETWORK_ACCESS_FAILURE;
                }
                break;

            case INVALID_LOGIN_CREDENTIALS:
                error = PackagePollCompleteStatus.Error.CREDENTIAL_FAILURE;
                final String userName = config.getUser();
                errorProperties.put("Username", userName);
                break;

            case RETRIEVE_DATA_FROM_DEVICE:
                error = PackagePollCompleteStatus.Error.BAD_ZAPI_RESPONSE;
                break;

            case INTERNAL_ERROR:
                error = PackagePollCompleteStatus.Error.INTERNAL_ERROR;
                break;

            default:
                error = PackagePollCompleteStatus.Error.INTERNAL_ERROR;
                break;

        }

        final Map<PackagePollCompleteStatus.Error, Map<String, String>> pollError = new HashMap<PackagePollCompleteStatus.Error, Map<String, String>>();
        pollError.put(error, errorProperties);
        pollStatus.addDataSourceError(pollError);
    }

    private void sendPollStart() {
        final ServiceManager serviceManager = ServiceManager.getInstance();
        if (serviceManager.isStandalone()) {
            return;
        }
        final NetAppStoragePerformancePackage netappStoragePerformancePackage = (NetAppStoragePerformancePackage) PackageContext.getPackage();
        final BaseDataSource thisDataSource = this.getDataSource();
        final Long datasourceID = thisDataSource.getDataSourceID();
        final String datasourceName = thisDataSource.getDataSourceName();
        final String packageName = netappStoragePerformancePackage.getPackageName();
        final long pollStartTime = netappStoragePerformancePackage.getCachedServerTime();
        final PackagePollStartStatus pollStartStatus = new PackagePollStartStatus(datasourceName, datasourceID, packageName, pollStartTime);
        try {
            final AcquisitionStatusRemote acqStatusSession = serviceManager.getCommunicationManager().getSession(AcquisitionStatusRemote.JNDI_NAME);
            if (acqStatusSession == null) {
                LOG.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageStarted "
                             + dataSource.getDataSourceName() + " " + netappStoragePerformancePackage.getPackageName());
                return;
            }
            acqStatusSession.packageStarted(NetAppFoundationConfiguration.AcquisitionPollStatus.schema, pollStartStatus);
        }
        catch (FrameworkException fe) {
            LOG.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageStarted "
                         + dataSource.getDataSourceName() + " " + netappStoragePerformancePackage.getPackageName());
        }
        catch (ServerInterfaceException sie) {
            LOG.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageStarted "
                         + dataSource.getDataSourceName() + " " + netappStoragePerformancePackage.getPackageName());
        }
    }

    public NetAppDefaultStorageContext getNetAppDefaultStorageContext() {
        final Collection<NetAppDefaultStorageContext> storageContexts = getStorageContexts().values();
        if (storageContexts.isEmpty()) {
            return null;
        }
        final NetAppDefaultStorageContext context = storageContexts.iterator().next();
        if (context == null) {
            return null;
        }
        return context;
    }

    public boolean isMetricsDebugEnabled() {
        return metricsDebugEnabled;
    }
    
    /**
     * Clears off the cached previous counters values from all performance metric builders.
     * We only need to do this for those builders which fetch counter values directly from zapi\archiver file source and uses scratchPad to 
     * derive current values (delta from previous values) .     
     */
    private void clearAllPreviousValues(){
        volumeMetricsBuilder.clearPreviousValues();
        lunMetricsBuilder.clearPreviousValues();
        nodeMetricsBuilder.clearPreviousValues();
        cifsVserverMetricsBuilder.clearPreviousValues();
        lifVserverMetricsBuilder.clearPreviousValues();
        nfsv3MetricsBuilder.clearPreviousValues();
        nfsv4MetricsBuilder.clearPreviousValues();
        nfsv41MetricsBuilder.clearPreviousValues();
        aggregateMetricsBuilder.clearPreviousValues();
        diskMetricsBuilder.clearPreviousValues();
        fcpLifMetricsBuilder.clearPreviousValues();
        fcpLifVserverMetricsBuilder.clearPreviousValues();
        fcpPortMetricsBuilder.clearPreviousValues();
        ipPortMetricsBuilder.clearPreviousValues();
        nicCommonMetricsBuilder.clearPreviousValues();
        volumeVserverMetricsBuilder.clearPreviousValues();
        lifMetricsBuilder.clearPreviousValues();
        processorMetricsBuilder.clearPreviousValues();
        iSCSILifMetricsBuilder.clearPreviousValues();
        iscsiLifVserverMetricsBuilder.clearPreviousValues();
        waflHyaAggrMetricsBuilder.clearPreviousValues();
        performanceArchiveManager.clearAllPreviousValues();
        qosWorkloadDetailMetricsBuilder.clearPreviousValues();
        qosServiceCenterMetricsBuilder.clearPreviousValues();
        qosWorkloadConstituentMetricsBuilder.clearPreviousValues();
        cifsNodeMetricsBuilder.clearPreviousValues();
        nfSv3NodeMetricsBuilder.clearPreviousValues();
        nfSv4NodeMetricsBuilder.clearPreviousValues();
        nfSv41NodeMetricsBuilder.clearPreviousValues();
        smb1NodeMetricsBuilder.clearPreviousValues();
        smb2NodeMetricsBuilder.clearPreviousValues();
        hostadapterMetricsBuilder.clearPreviousValues();
        LOG.info("{} - Cleared cached previous counter values from all performance metric builders.", dataSource.getDataSourceName());
    }

    /**
     * Given a performance metrics builder, zapi connection instance and context, parse zapi response and report performance statistics
     * @param performanceMetricsBuilder
     * @param context
     * @param zapiConnection
     */
    private void getAndReportPerformanceMetrics(MetricsBuilderBase performanceMetricsBuilder, NetAppDefaultStorageContext context, ZAPIConnection zapiConnection){
        try{
            
            performanceMetricsBuilder.getAndReportPerformance(context, zapiConnection);
            
        }catch(DataSourceErrorException dsee){
            // we need to only remember the first thrown Exception as that is the root cause error
            // this exception is remembered until all performance previous value scratchPads are cleared and then we fail the poll against this remembered Exception.
            if(acquisitionFailedException == null){
                acquisitionFailedException = dsee;
            }
            // The current poll has failed. We do not want to carry forward stale counter values from previous poll and deviate from the 5-minute sample norm.
            // This is done so that the delta and rated counter values are consistently calculated for uniform time intervals.
            performanceMetricsBuilder.clearPreviousValues();
            LOG.info("{} - Cleared counter values cached from previous poll for {}.", dataSource.getDataSourceName(), performanceMetricsBuilder.getClass().getName());
        }
    }

    public GlobalScratchPad getGlobalScratchPad() {
        return globalScratchPad;
    }
}
