package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.aggregate.interfaces.data.AggregateMeasurements;
import com.netapp.oci.netapp.performance.aggregate.interfaces.data.AggregateRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * Builder class for processing and reporting aggregate metrics from archiver
 */
public class AggregateMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        AggregateMeasurements, AggregateRawMeasurements, AggregateMetricsBuilder.AggregateCounter> {

    public static enum AggregateCounter implements ArchiveCounterMetaData {

        //Number of blocks read per second during a CP on the aggregate
        cp_read_blocks,
        //Number of reads per second done during a CP to the aggregate
        cp_reads,
        //Total number of transfers per second serviced by the aggregate
        total_transfers,
        //Number of blocks read per second on the aggregate
        user_read_blocks,
        //Number of user reads per second to the aggregate
        user_reads,
        //Number of blocks written per second to the aggregate
        user_write_blocks,
        //Number of user writes per second to the aggregate
        user_writes,
        //Total number of transfers per second serviced by the aggregate HDD disks
        total_transfers_hdd,
        //Total number of transfers per second serviced by the aggregate SSD disks
        total_transfers_ssd,
        //Number of blocks read per second on the aggregate HDD disks
        user_read_blocks_hdd,
        //Number of blocks read per second on the aggregate SSD disks
        user_read_blocks_ssd,
        //Number of user reads per second to the aggregate HDD disks
        user_reads_hdd,
        //Number of user reads per second to the aggregate SSD disks
        user_reads_ssd,
        //Number of blocks written per second to the aggregate HDD disks
        user_write_blocks_hdd,
        //Number of blocks written per second to the aggregate SSD disks
        user_write_blocks_ssd,
        //Number of user writes per second to the aggregate HDD disks
        user_writes_hdd,
        //Number of user writes per second to the aggregate SSD disks
        user_writes_ssd;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private AggregateCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private AggregateCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        private AggregateCounter(int onTapCounterIndex) {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = onTapCounterIndex;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }

    }

    /**
     * constructor
     *
     * @param reportingPackage
     */

    public AggregateMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "aggregate", AggregateCounter.class);
    }

     @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, final String nodeName,
                                                           final Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,Collection<RawSample<AggregateMeasurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<AggregateMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<AggregateMeasurements>> aggregateMeasurements = measurementEntry.getValue();
            reportingPackage.reportAggregatePerformance(key, aggregateMeasurements);
        }
    }

    @Override
    public AggregateCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return AggregateCounter.values();
    }


    @Override
    protected AggregateMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<AggregateCounter> currentScratchPad,
                                                                  final NetAppDefaultStorageContext context) {
        final AggregateMeasurements aggregateMeasurements = new AggregateMeasurements();
        float userReadBlocks = currentScratchPad.getCounterRatePerSec(AggregateCounter.user_read_blocks);
        float cpReadBlocks =  currentScratchPad.getCounterRatePerSec(AggregateCounter.cp_read_blocks);
        float userWriteBlocks = currentScratchPad.getCounterRatePerSec(AggregateCounter.user_write_blocks);

        float opmAggrUtilization = MetricsRowSample.NULL_FLOAT;
        float opmAggrUserReadLatency = MetricsRowSample.NULL_FLOAT;
        float opmAggrUserWriteLatency = MetricsRowSample.NULL_FLOAT;
        float opmAggrCpReadLatency = MetricsRowSample.NULL_FLOAT;
        float diskUserReads = MetricsRowSample.NULL_FLOAT;
        float diskUserWrites = MetricsRowSample.NULL_FLOAT;
        float diskCpReads = MetricsRowSample.NULL_FLOAT;

        Map<String, List<String>> aggrUUIDToDiskUUIDList = context.getAggregateUUIDToDiskUUIDList();
        List<String> diskUUIDList = aggrUUIDToDiskUUIDList.get(currentScratchPad.getPerfInstanceUUID());
        if(diskUUIDList!=null && !diskUUIDList.isEmpty())  {
            List<DiskMetricsBuilder.DiskCounter> numeratorCounterNamesLatency = new ArrayList<>();
            numeratorCounterNamesLatency.add(DiskMetricsBuilder.DiskCounter.user_read_latency);
            numeratorCounterNamesLatency.add(DiskMetricsBuilder.DiskCounter.user_write_latency);
            numeratorCounterNamesLatency.add(DiskMetricsBuilder.DiskCounter.cp_read_latency);
            List<DiskMetricsBuilder.DiskCounter> denominatorCounterNamesLatency = new ArrayList<>();
            denominatorCounterNamesLatency.add(DiskMetricsBuilder.DiskCounter.user_reads);
            denominatorCounterNamesLatency.add(DiskMetricsBuilder.DiskCounter.user_writes);
            denominatorCounterNamesLatency.add(DiskMetricsBuilder.DiskCounter.cp_reads);
            Map<DiskMetricsBuilder.DiskCounter,MetricsRowScratchPad<DiskMetricsBuilder.DiskCounter>> opmLatencyMap = reportingPackage.getGlobalScratchPad().getScratchPadsForMaximumLatencyOps(diskUUIDList, "disk", numeratorCounterNamesLatency, denominatorCounterNamesLatency);
            MetricsRowScratchPad<DiskMetricsBuilder.DiskCounter> opmAggrUserReadLatencyPad = opmLatencyMap.get(DiskMetricsBuilder.DiskCounter.user_read_latency);
            MetricsRowScratchPad<DiskMetricsBuilder.DiskCounter> opmAggrUserWriteLatencyPad = opmLatencyMap.get(DiskMetricsBuilder.DiskCounter.user_write_latency);
            MetricsRowScratchPad<DiskMetricsBuilder.DiskCounter> opmAggrCpReadLatencyPad = opmLatencyMap.get(DiskMetricsBuilder.DiskCounter.cp_read_latency);
            if(opmAggrUserReadLatencyPad!=null) {
                opmAggrUserReadLatency = opmAggrUserReadLatencyPad.getLatencyPerOp(DiskMetricsBuilder.DiskCounter.user_read_latency, DiskMetricsBuilder.DiskCounter.user_reads);
                diskUserReads = opmAggrUserReadLatencyPad.getCounterRatePerSec(DiskMetricsBuilder.DiskCounter.user_reads);
            }
            if(opmAggrUserWriteLatencyPad!=null) {
                opmAggrUserWriteLatency = opmAggrUserWriteLatencyPad.getLatencyPerOp(DiskMetricsBuilder.DiskCounter.user_write_latency, DiskMetricsBuilder.DiskCounter.user_writes);
                diskUserWrites = opmAggrUserWriteLatencyPad.getCounterRatePerSec(DiskMetricsBuilder.DiskCounter.user_writes);
            }
            if(opmAggrCpReadLatencyPad!=null) {
                opmAggrCpReadLatency = opmAggrCpReadLatencyPad.getLatencyPerOp(DiskMetricsBuilder.DiskCounter.cp_read_latency, DiskMetricsBuilder.DiskCounter.cp_reads);
                diskCpReads = opmAggrUserWriteLatencyPad.getCounterRatePerSec(DiskMetricsBuilder.DiskCounter.cp_reads);
            }

            List<DiskMetricsBuilder.DiskCounter> numeratorCounterNamesUtil = new ArrayList<>();
            numeratorCounterNamesUtil.add(DiskMetricsBuilder.DiskCounter.disk_busy);
            List<DiskMetricsBuilder.DiskCounter> denominatorCounterNamesUtil = new ArrayList<>();
            denominatorCounterNamesUtil.add(DiskMetricsBuilder.DiskCounter.base_for_disk_busy);
            Map<DiskMetricsBuilder.DiskCounter,Float> opmUtilMap = reportingPackage.getGlobalScratchPad().getMaximumOfPercentDeltas(diskUUIDList, "disk", numeratorCounterNamesUtil, denominatorCounterNamesUtil);
            opmAggrUtilization = opmUtilMap.get(DiskMetricsBuilder.DiskCounter.disk_busy);
        }

        //to be used to calculate weighted average of Latency
        List<Float> latencyList = new ArrayList<>();
        latencyList.add(opmAggrUserReadLatency);
        latencyList.add(opmAggrUserWriteLatency);
        latencyList.add(opmAggrCpReadLatency);
        List<Float>  opList = new ArrayList<>();
        opList.add(diskUserReads);
        opList.add(diskUserWrites);
        opList.add(diskCpReads);

        aggregateMeasurements.setCpReadBlocks(cpReadBlocks);
        aggregateMeasurements.setCpReads(currentScratchPad.getCounterRatePerSec(AggregateCounter.cp_reads));
        aggregateMeasurements.setTotalTransfers(currentScratchPad.getCounterRatePerSec(AggregateCounter.total_transfers));
        aggregateMeasurements.setUserReadBlocks(userReadBlocks);
        aggregateMeasurements.setUserReads(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_reads));
        aggregateMeasurements.setUserWriteBlocks(userWriteBlocks);
        aggregateMeasurements.setUserWrites(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_writes));
        aggregateMeasurements.setTotalTransfersHdd(currentScratchPad.getCounterRatePerSec(AggregateCounter.total_transfers_hdd));
        aggregateMeasurements.setTotalTransfersSsd(currentScratchPad.getCounterRatePerSec(AggregateCounter.total_transfers_ssd));
        aggregateMeasurements.setUserReadBlocksHdd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_read_blocks_hdd));
        aggregateMeasurements.setUserReadBlocksSsd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_read_blocks_ssd));
        aggregateMeasurements.setUserReadsHdd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_reads_hdd));
        aggregateMeasurements.setUserReadsSsd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_reads_ssd));
        aggregateMeasurements.setUserWriteBlocksHdd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_write_blocks_hdd));
        aggregateMeasurements.setUserWriteBlocksSsd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_write_blocks_ssd));
        aggregateMeasurements.setUserWritesHdd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_writes_hdd));
        aggregateMeasurements.setUserWritesSsd(currentScratchPad.getCounterRatePerSec(AggregateCounter.user_writes_ssd));
        aggregateMeasurements.setOpmAggrThroughput(MetricsRowScratchPad.add(userReadBlocks, userWriteBlocks, cpReadBlocks) * 4096);   //blocks to bytes
        aggregateMeasurements.setOpmAggrUserReadLatency(opmAggrUserReadLatency);
        aggregateMeasurements.setOpmAggrUserWriteLatency(opmAggrUserWriteLatency);
        aggregateMeasurements.setOpmAggrCpReadLatency(opmAggrCpReadLatency);
        aggregateMeasurements.setOpmAggrUtilization(opmAggrUtilization);
        aggregateMeasurements.setOpmAggrLatency(MetricsRowScratchPad.averageLatency(latencyList, opList));
        return  aggregateMeasurements;
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey,Map<Cluster.ClusterKey, Collection<RawSample<AggregateRawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<AggregateRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<AggregateRawMeasurements>> aggregateRawMeasurements = measurementEntry.getValue();
            reportingPackage.reportAggregateRawPerformance(key, aggregateRawMeasurements);
        }
    }

    @Override
    protected AggregateRawMeasurements rawMetrics(ArchiveMetricsScratchPad<AggregateCounter> currentScratchPad,NetAppDefaultStorageContext context) {
        final AggregateRawMeasurements m = new AggregateRawMeasurements();
        Map<AggregateCounter,BigInteger> counterValues = currentScratchPad.getCounterValues();
        m.setCpReadBlocks(counterValues.get(AggregateCounter.cp_read_blocks));
        m.setCpReads(counterValues.get(AggregateCounter.cp_reads));
        m.setTotalTransfers(counterValues.get(AggregateCounter.total_transfers));
        m.setUserWrites(counterValues.get(AggregateCounter.user_writes));
        m.setUserWriteBlocks(counterValues.get(AggregateCounter.user_write_blocks));
        m.setUserReadBlocks(counterValues.get(AggregateCounter.user_read_blocks));
        m.setUserReads(counterValues.get(AggregateCounter.user_reads));
        m.setTotalTransfersHdd(counterValues.get(AggregateCounter.total_transfers_hdd));
        m.setTotalTransfersSsd(counterValues.get(AggregateCounter.total_transfers_ssd));
        m.setUserReadBlocksHdd(counterValues.get(AggregateCounter.user_read_blocks_hdd));
        m.setUserReadBlocksSsd(counterValues.get(AggregateCounter.user_read_blocks_ssd));
        m.setUserWriteBlocksHdd(counterValues.get(AggregateCounter.user_write_blocks_hdd));
        m.setUserWriteBlocksSsd(counterValues.get(AggregateCounter.user_write_blocks_ssd));
        m.setUserReadsHdd(counterValues.get(AggregateCounter.user_reads_hdd));
        m.setUserReadsSsd(counterValues.get(AggregateCounter.user_reads_ssd));
        m.setUserWritesHdd(counterValues.get(AggregateCounter.user_writes_hdd));
        m.setUserWritesSsd(counterValues.get(AggregateCounter.user_writes_ssd));

        return m;
    }

}
