package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.common.UUID;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.parser.CounterTable;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import com.onaro.sanscreen.acquisition.schema.Key;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.*;

/**
 * Base class to get, transform, normalize and report to server sets of performance metrics from the performance archive
 *
 * @param <K> the foundation UUID parent Key to identify performance measurements
 * @param <S> the normalized performance Sample to report to the server
 * @param <R> the raw performance Sample to report to the server
 * @param <Z> An Enumeration, defined by derived classes, that identifies the set of counter/column names requested from Archive
 */
public abstract class ArchiveMetricsBuilderBase<K extends Key, S, R, Z extends Enum<Z> & ArchiveCounterMetaData> {
    private static final Logger logger = LoggerFactory.getLogger(ArchiveMetricsBuilderBase.class);

    /**
     * vserver uuid identifier in archiver file
     */
    protected static final String VSERVER_UUID = "vserver_uuid";
    
    protected final NetAppOCIEPerformancePackage reportingPackage;

    private final String counterObjectName;

    private final Class<Z> metricEnumClass;

    private Map<String, ArchiveMetricsScratchPad<Z>> performanceUUIDToPreviousScratchPad;

    protected ArchiveMetricsBuilderBase(final NetAppOCIEPerformancePackage reportingPackage, final String counterObjectName,
                                        final Class<Z> metricEnumClass) {
        this.reportingPackage = reportingPackage;
        this.counterObjectName = counterObjectName;
        this.metricEnumClass = metricEnumClass;
        performanceUUIDToPreviousScratchPad = new HashMap<>();
    }

    /**
     * Get the parent foundation UUID key to use to report the samples under
     *
     * @param context          the foundation context used to determine the parent foundation UUID key
     * @param perfInstanceUUID the performance instance UUID to find the parent for
     * @param nodeName 	Name of the node to which the archiver file belongs to
     * @param counterRow row from the archiver file, which contains values of all the counters
     * @return the parent inventory/foundation UUID key to use to report the samples under
     * 
     */
    protected abstract K getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID, final String nodeName,
            final Map<String, Object> counterRow);
    
    /**
     * Get the foundation UUID key for the given performance UUID key
     *
     * @param context                 the foundation context used to determine the foundation UUID key
     * @param performanceInstanceUUID the performance UUID to find the foundation UUID for
     * @param nodeName 	Name of the node to which the archiver file belongs to
     * @param counterRow row from the archiver file, which contains values of all the counters
     * @return the inventory/foundation UUID key for the givent performance UUID key
     */
    protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context, final String performanceInstanceUUID, final String nodeName,
            final Map<String, Object> counterRow) {
        return performanceInstanceUUID;
    }

    /**
     * Report the given samples to the server
     *
     * @param measurements the samples to report
     * @throws DataSourceErrorException when there is a problem reporting samples to the server
     */
    public abstract void reportPerformance(Cluster.ClusterKey clusterKey, Map<K, Collection<RawSample<S>>> measurements) throws DataSourceErrorException;
    
    /**
     * Report raw samples to the server
     * 
     * @param clusterKey the cluster key of the sample
     * @param rawMeasurements the samples to report
     * @throws DataSourceErrorException
     */
    public abstract void reportRawPerformance(Cluster.ClusterKey clusterKey, final Map<K, Collection<RawSample<R>>> rawMeasurements) throws DataSourceErrorException;   

    /**
     * This method returns the list of OnTap counter names. These names will be used to identify counter in the
     * performance archive file.
     *
     * @return Array of enum counter names.
     */
    protected abstract Z[] getCountersMetaData(final NetAppDefaultStorageContext context);

    /**
     * Returns the current scratch pad when called after the builder processes the current polling data. Returns the
     * previous scratch pad when called before the builder processes the current polling data.
     *
     * @return The Scratch Pad
     */
    public Map<String, ArchiveMetricsScratchPad<Z>> getScratchPad() {
        return Collections.unmodifiableMap(performanceUUIDToPreviousScratchPad);
    }

    /**
     * Normalized this metrics instance against the previous metrics instance, and wrap in the
     * measurements class expected by the server
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return normalized metric values wrapped in the class expected by the server
     */
    protected abstract S normalizeMetrics(ArchiveMetricsScratchPad<Z> currentScratchPad, NetAppDefaultStorageContext context);

    
    /**
     * Raw metrics
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return raw metric values wrapped in the class expected by the server
     */
    protected abstract R rawMetrics(ArchiveMetricsScratchPad<Z> currentScratchPad, NetAppDefaultStorageContext context);
    
    /**
     * Returns the Measurement object to report if a wrap reset is detected.
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return normalized metric values wrapped in the class expected by the server
     */
    protected S getWrapResetSample(final ArchiveMetricsScratchPad<Z> currentScratchPad, final NetAppDefaultStorageContext context) {
        return null;
    }

    /**
     * This method returns the default debug writer - {@link ArchiveMetricsDebugWriter}. The derived class can override
     * this method and return its own implementation of the debug writter.
     */
    protected ArchiveMetricsDebugWriter getArchiveMetricsDebugWriter() {
        return  new ArchiveMetricsDebugWriter(reportingPackage, counterObjectName, metricEnumClass);
    }

    /**
     * Normalize performance measurements
     *
     * @param context       the foundation context used to determine the performance UUID keys
     * @param counterTables List of counter tables, one for each node in the cluster.
     * @throws DataSourceErrorException when there is a problem acquiring or reporting samples to the server
     */
    public Map<K, Collection<RawSample<S>>> processCounterData(final NetAppDefaultStorageContext context, final List<CounterTable> counterTables) throws DataSourceErrorException {
        final ArchiveMetricsDebugWriter archiveMetricsDebugWriter = getArchiveMetricsDebugWriter();
        Map<String, ArchiveMetricsScratchPad<Z>> performanceUUIDToCurrentScratchPad = new HashMap<>();
        Map<String, K>  performanceUUIDToParentInventoryUUIDs  = new HashMap<>();
        getCounterData(context, counterTables, performanceUUIDToCurrentScratchPad, performanceUUIDToParentInventoryUUIDs);
        final Map<K, Collection<RawSample<S>>> allMeasurementsToReport = new HashMap<>();
        final Map<K, Collection<RawSample<R>>> rawMetricsToReport = new HashMap<>();

        boolean rawPersistenceEnabled = context.isRawPersistenceEnabled();
        try {
            for(final Map.Entry<String, ArchiveMetricsScratchPad<Z>> currentScratchPadEntry : performanceUUIDToCurrentScratchPad.entrySet())  {
                String performanceUUID = currentScratchPadEntry.getKey();
                ArchiveMetricsScratchPad<Z> currentScratchPad = currentScratchPadEntry.getValue();
                if(performanceUUID != null)  {
                    final K parentInventoryKey = performanceUUIDToParentInventoryUUIDs.get(performanceUUID);
                    if (parentInventoryKey != null) {
                        final ArchiveMetricsScratchPad<Z> previousScratchPad = this.performanceUUIDToPreviousScratchPad.get(performanceUUID);
                        if (!allMeasurementsToReport.containsKey(parentInventoryKey)) {
                            allMeasurementsToReport.put(parentInventoryKey, new LinkedList<RawSample<S>>());
                        }
                        if (previousScratchPad == null) {
                            archiveMetricsDebugWriter.writeMetricsDebugCSV(currentScratchPad);
                            continue;// first poll, nothing to report
                        }

                        // Compare current values to previous
                        final RawSample<S> measurementToReport = getNormalizedSample(currentScratchPad, context);
                        // Now we finally have the final value for debug CSV
                        archiveMetricsDebugWriter.writeMetricsDebugCSV(currentScratchPad);
                        final Collection<RawSample<S>> measurementsToReportForParentKey = allMeasurementsToReport.get(parentInventoryKey);
                        if (measurementToReport != null) {
                            measurementsToReportForParentKey.add(measurementToReport);
                        }

                        if(rawPersistenceEnabled){
                            if(!rawMetricsToReport.containsKey(parentInventoryKey)) {
                                rawMetricsToReport.put(parentInventoryKey, new LinkedList<RawSample<R>>());
                            }

                            final Collection<RawSample<R>> countersForReport = rawMetricsToReport.get(parentInventoryKey);
                            countersForReport.add(getRawSample(currentScratchPad, context));
                        }
                    }
                }
            }
            if(rawPersistenceEnabled){
                reportRawPerformance(context.getKey(), rawMetricsToReport);
            }

        } finally {
            //Clear previous and replace with current.
            this.performanceUUIDToPreviousScratchPad.clear(); //I should not do this. This is O(n), can the GC take care of it ?
            this.performanceUUIDToPreviousScratchPad = performanceUUIDToCurrentScratchPad;
            archiveMetricsDebugWriter.closeMetricDebug();
        }
        this.reportingPackage.getGlobalScratchPad().put(this.counterObjectName, Collections.unmodifiableMap(this.performanceUUIDToPreviousScratchPad));

        return allMeasurementsToReport;
    }

    /**
     *
     * @param context Polling context with data from the foundation poll
     * @param counterTables Raw data form the archiver
     * @param performanceUUIDToCurrentScratchPad   Map of String, ArchiveMetricsScratchPad
     * @param performanceUUIDToParentInventoryUUIDs Map of performanceUUID, parentInventoryUUID
     * @throws DataSourceErrorException
     */
    private void getCounterData(final NetAppDefaultStorageContext context, final List<CounterTable> counterTables, final Map<String, ArchiveMetricsScratchPad<Z>> performanceUUIDToCurrentScratchPad, final Map<String, K>  performanceUUIDToParentInventoryUUIDs) throws DataSourceErrorException{
        if(counterTables != null) {
            boolean nullParentInventoryUUIDReported = false;
            boolean nullInventoryUUIDReported = false;
            boolean nullPerformanceUUIDReported = false;
            for (final CounterTable counterTable : counterTables) {
                while (counterTable.hasNext()) {
                    final Map<String, Object> counterRow = counterTable.next();
                    final String performanceUUID = getPerformanceUUID(context, counterTable.getNodeName(), counterRow);
                    if (performanceUUID != null) {
                        final K parentInventoryKey = getParentInventoryUUIDKey(context, performanceUUID, counterTable.getNodeName(), counterRow);
                        if(parentInventoryKey!=null)  {
                            performanceUUIDToParentInventoryUUIDs.put(performanceUUID, parentInventoryKey);
                        } else{
                            nullParentInventoryUUIDReported = true;
                        }
                        final String inventoryUUID = getInventoryUUIDFromPerformanceUUID(context, performanceUUID, counterTable.getNodeName(), counterRow);
                        if (inventoryUUID != null) {
                             final ArchiveMetricsScratchPad<Z> previousScratchPad = this.performanceUUIDToPreviousScratchPad.get(performanceUUID);

                            Number archiveUUIDLo = null;
                            Number archiveUUIDHi = null;
                            try {
                                archiveUUIDLo = (Number) counterRow.get("archive_uuid_lo");
                                archiveUUIDHi = (Number) counterRow.get("archive_uuid_hi");
                            } catch (ClassCastException classCaseExe) {
                                logger.warn("ClassCastException : Unable to determine UUID, invalid type returned for archive_uuid_lo & archive_uuid_hi. archive_uuid_lo : {} archive_uuid_hi : {}",
                                        counterRow.get("archive_uuid_lo").getClass().getName(),
                                        counterRow.get("archive_uuid_hi").getClass().getName());
                            }
                            final ArchiveMetricsScratchPad<Z> currentScratchPad;
                            if (archiveUUIDHi != null || archiveUUIDLo != null) {
                                currentScratchPad = new ArchiveMetricsScratchPad<>(this.metricEnumClass,
                                        archiveUUIDHi, archiveUUIDLo, performanceUUID, inventoryUUID,
                                        this.reportingPackage.getSampleTime(), counterTable.getArchiveTimeStamp(),
                                        previousScratchPad, reportingPackage.isMetricsDebugEnabled());
                            } else {
                                currentScratchPad = new ArchiveMetricsScratchPad<>(this.metricEnumClass,
                                        performanceUUID, inventoryUUID, this.reportingPackage.getSampleTime(),
                                        counterTable.getArchiveTimeStamp(), previousScratchPad,
                                        reportingPackage.isMetricsDebugEnabled());
                            }
                            if(currentScratchPad.getCollectionPeriod() >= (2 * reportingPackage.getConfiguration().getPollTimeMillis())){
                                // ignore this stat as OPM needs only 5-minute samples
                                logger.warn("Collection period {}ms was higher than {}ms. Ignoring the stats with perfUUID: {} ", new Object[]{currentScratchPad.getCollectionPeriod(), reportingPackage.getConfiguration().getPollTimeMillis(), performanceUUID});
                                continue;
                            }
                            populateCurrentScratchPad(context, counterRow, currentScratchPad);
                            //check to see if this performanceUUID is already created
                            if(performanceUUIDToCurrentScratchPad.containsKey(performanceUUID))  {
                                ArchiveMetricsScratchPad<Z> existingScratchPad = performanceUUIDToCurrentScratchPad.get(performanceUUID);
                                aggregateScratchPads(existingScratchPad, currentScratchPad, context);
                            }
                            performanceUUIDToCurrentScratchPad.put(performanceUUID, currentScratchPad);
                        }
                        else {
                            nullInventoryUUIDReported = true;
                        }
                    }
                    else {
                        nullPerformanceUUIDReported = true;
                    }
                }
            }
            if(nullParentInventoryUUIDReported) {
                logger.info("Unable to determine parent inventory UUID from Builder {}", this.getClass().getSimpleName());
            }
            if(nullInventoryUUIDReported) {
                logger.info("Unable to determine inventory UUID from Builder {}", this.getClass().getSimpleName());
            }
            if(nullPerformanceUUIDReported) {
                logger.info("Unable to determine performance UUID from Builder {}", this.getClass().getSimpleName());
            }

        }

    }


    /**
     * This is to aggregate the values for two scratchpad and saved in currentScratchPad
     * @param previousScratchPad Previous scratch pad
     * @param currentScratchPad Current scratch pad
     * @param context Context with data from foundation poll
     */
    private void aggregateScratchPads(final ArchiveMetricsScratchPad<Z> previousScratchPad, final ArchiveMetricsScratchPad<Z> currentScratchPad, final NetAppDefaultStorageContext context) {
        final EnumMap<Z, BigInteger> currentCounterValues = currentScratchPad.getCounterValues();
        final EnumMap<Z, BigInteger> previousCounterValues = previousScratchPad.getCounterValues();
        for (final Z counterMetaData : getCountersMetaData(context))  {
            final BigInteger currentCounterValue = currentCounterValues.get(counterMetaData);
            final BigInteger previousCounterValue = previousCounterValues.get(counterMetaData);
            if (previousCounterValue!=null) {
                BigInteger result = previousCounterValue;
                if(currentCounterValue != null) {
                    result = MetricsRowScratchPad.add(currentCounterValue, previousCounterValue);
                }
                currentCounterValues.put(counterMetaData, result);
            }
        }
    }

    private void populateCurrentScratchPad(final NetAppDefaultStorageContext context, final Map<String, Object> counterRow,
                                           final ArchiveMetricsScratchPad<Z> currentScratchPad) {
        final EnumMap<Z, BigInteger> currentCounterValues = currentScratchPad.getCounterValues();
        final EnumMap<Z, String> currentRawValues = currentScratchPad.getRawValues();
        for (final Z counterMetaData : getCountersMetaData(context)) {
            final Object counterValue = counterRow.get(counterMetaData.getOnTapCounterName());
            if (counterValue != null) {
                //Counter can be Integer, Long, String or a List of Integer, Long or String.Most are Integer or Long.
                BigInteger counterBigIntValue = MetricsRowSample.NULL_BIG_INT;
                if (counterValue instanceof Number) {
                    counterBigIntValue = BigInteger.valueOf(((Number) counterValue).longValue());
                } else if (counterValue instanceof List) {
                    final int onTapCounterIndex = counterMetaData.getOnTapCounterIndex();
                    final int counterValueArraySize = ((List) counterValue).size();
                    if (onTapCounterIndex != -1 && onTapCounterIndex < ((List) counterValue).size()) {
                        final Object counterValueAtIndex = ((List) counterValue).get(counterMetaData.getOnTapCounterIndex());
                        if (counterValueAtIndex instanceof Number) {
                            counterBigIntValue = BigInteger.valueOf(((Number) counterValueAtIndex).longValue());
                        } else {
                            //Populate currentRawValues for counters that's not a number.
                            currentRawValues.put(counterMetaData, counterValueAtIndex.toString());
                        }
                    }
                    else if (onTapCounterIndex == -1){
                    	counterBigIntValue = BigInteger.ZERO;
                        final StringBuffer counterStringValue = new StringBuffer();
                    	for (int i=0; i < counterValueArraySize ; i++ ){
                    		final Object counterValueAtEachIndex = ((List) counterValue).get(i);
                            if(counterValueAtEachIndex instanceof Number) {
                                counterBigIntValue = counterBigIntValue.add(BigInteger.valueOf(((Number)counterValueAtEachIndex).longValue()));
                            } else {
                                // If the counter type is not a number then populate the currentRawValues with a CSV
                                // of the counter values.
                                counterStringValue.append(',').append(counterValueAtEachIndex.toString());
                            }
                    	}
                        if(counterStringValue.length() != 0) {
                            currentRawValues.put(counterMetaData, counterStringValue.substring(1));
                        }
                    }
                } else {
                    //Populate currentRawValues for counters that's not a number.
                    currentRawValues.put(counterMetaData, counterValue.toString());
                }
                currentCounterValues.put(counterMetaData, counterBigIntValue);
            }
        }
    }

    /**
     * Normalize the current sample by comparing to the previous sample and converting counters to rates, etc
     *
     * @param currentScratchPad the current set of metrics
     * @return the normalized raw sample
     */
    protected RawSample<S> getNormalizedSample(final ArchiveMetricsScratchPad<Z> currentScratchPad,
                                               final NetAppDefaultStorageContext context) {
        final String inventoryUUID = currentScratchPad.getInventoryUUID();
        final S measurements = normalizeMetrics(currentScratchPad, context);
        RawSample<S> rawSample = new RawSample<>(inventoryUUID, measurements);
        if (currentScratchPad.isZeroSuppressDetected()) {
            // Make no report at all if zero suppression has been detected.
            rawSample = null;
        } else if (currentScratchPad.isWrapResetDetected()) {
            // detected Wrap/Reset, contract with server requires reporting the UUID with a null payload.
            rawSample = new RawSample<>(inventoryUUID, getWrapResetSample(currentScratchPad, context));
        }
        return rawSample;
    }
    
    /**
     * Get the raw counters out of the current metric.
     * @param currentScratchPad Current Scratch pad
     * @param context Context with data from foundation poll
     * @return Raw Sample
     */
    final RawSample<R> getRawSample(final ArchiveMetricsScratchPad<Z> currentScratchPad,
            final NetAppDefaultStorageContext context) {
        final String inventoryUUID = currentScratchPad.getInventoryUUID();
        final R measurements = rawMetrics(currentScratchPad, context);
        return new RawSample<>(inventoryUUID, measurements);
    }

    /**
     * This method returns the UUID of the counter instance. If the object return instance_uuid, then uuid = instance_uuid else uuid is constructed using the archive_uuid_hi and
     * archive_uuid_lo counters.
     *
     * @param context    The foundation context
     * @param counterRow Raw counter data form the performance archive
     * @param nodeName   Node name from where the archiver was retrieved
     * @return UUID from the archive file.
     */

    protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
                                        final Map<String, Object> counterRow) {
        String uuid = (String)counterRow.get("instance_uuid");
        
        if(uuid == null) {
            uuid = getUUIDFromHiLo(counterRow,  "archive_uuid_hi", "archive_uuid_lo");
        }
        return uuid;
    }

    /**
     * This method is a general instance to return uuid.
     * @param counterRow     Raw counter data form the performance archive
     * @param counterHi     the high order 8 bytes of the uuid
     * @param counterLo     the low order 8 bytes of the uuid
     * @return
     */
    protected String getUUIDFromHiLo(final Map<String, Object> counterRow, String counterHi, String counterLo) {
        String uuid = null;
        Number archiveUUIDLo = null;
        Number archiveUUIDHi = null;

        try {
            archiveUUIDLo = (Number) counterRow.get(counterLo);
            archiveUUIDHi = (Number) counterRow.get(counterHi);
        } catch (ClassCastException classCaseExe) {
            logger.warn(new StringBuilder("Unable to determine UUID, invalid type returned for " + counterHi + " ")
                    .append("& archive_uuid_hi. archive_uuid_lo : ")
                    .append(counterRow.get(counterLo).getClass().getName())
                    .append(". " + counterHi + ": ")
                    .append(counterRow.get(counterHi).getClass().getName()).toString());
        }

        if (archiveUUIDHi != null && archiveUUIDLo != null) {
            uuid = new UUID(archiveUUIDHi.longValue(), archiveUUIDLo.longValue()).toString();
        }

        return uuid;
    }

    public void clearPreviousValues(){
        performanceUUIDToPreviousScratchPad.clear();
    }
}
