package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.cifsvserver.interfaces.data.CifsVserverMeasurements;
import com.netapp.oci.netapp.performance.cifsvserver.interfaces.data.CifsVserverRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

public class CifsVserverMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        CifsVserverMeasurements, CifsVserverRawMeasurements, CifsVserverMetricsBuilder.CifsVserverCounter> {

    private static final Logger logger = LoggerFactory.getLogger(FcpLifMetricsBuilder.class);

    public static enum CifsVserverCounter implements ArchiveCounterMetaData {

        instance_name,
        vserver_name,
        vserver_id,
        cifs_ops,
        cifs_latency,
        cifs_latency_base,
        cifs_read_ops,
        cifs_write_ops,
        cifs_read_latency,
        cifs_write_latency,
        read_data,
        write_data; 


        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private CifsVserverCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private CifsVserverCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }


    public CifsVserverMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "cifs", CifsVserverCounter.class);
    }

    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
                                        final Map<String, Object> counterRow) {
        String vServerUUID = null;
        Map<String, String> cifsVserverIDToUUIDMap = context.getCifsVserverIDToUUIDMap();
        String cifsVserverId = (String)counterRow.get(CifsVserverCounter.vserver_id.name());
        vServerUUID = cifsVserverIDToUUIDMap.get(cifsVserverId);
        if (vServerUUID == null) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerName for cifs " + nodeName);
        }
        // Performance UUID will be vServerUUID
        return vServerUUID;
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(NetAppDefaultStorageContext context, String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public CifsVserverCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return CifsVserverCounter.values();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey,
                                  Map<Cluster.ClusterKey, Collection<RawSample<CifsVserverMeasurements>>> measurements)
            throws DataSourceErrorException {
        if(measurements == null ) {
            return;
        }

        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<CifsVserverMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<CifsVserverMeasurements>> cifsVserverMeasurements = measurementEntry.getValue();
            reportingPackage.reportCifsVserverPerformance(key, cifsVserverMeasurements);
        }
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey,
                                     Map<Cluster.ClusterKey, Collection<RawSample<CifsVserverRawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<CifsVserverRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<CifsVserverRawMeasurements>> cifsVserverRawMeasurements = measurementEntry.getValue();
            reportingPackage.reportCifsVserverRawPerformance(key, cifsVserverRawMeasurements);
        }
    }

    @Override
    protected CifsVserverMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<CifsVserverCounter> currentScratchPad,
                                                  final NetAppDefaultStorageContext context) {
        final CifsVserverMeasurements cifsVserverMeasurements = new CifsVserverMeasurements();

        cifsVserverMeasurements.setCifsLatency(currentScratchPad.getLatencyPerOp(CifsVserverCounter.cifs_latency, CifsVserverCounter.cifs_ops));
        cifsVserverMeasurements.setCifsLatencyBase(currentScratchPad.getCounterDeltaFinal(CifsVserverCounter.cifs_latency_base));
        cifsVserverMeasurements.setCifsOps(currentScratchPad.getCounterRatePerSec(CifsVserverCounter.cifs_ops));
        cifsVserverMeasurements.setCifsReadLatency(currentScratchPad.getLatencyPerOp(CifsVserverCounter.cifs_read_latency, CifsVserverCounter.cifs_read_ops));
        cifsVserverMeasurements.setCifsReadOps(currentScratchPad.getCounterRatePerSec(CifsVserverCounter.cifs_read_ops));
        cifsVserverMeasurements.setCifsReadThroughput(currentScratchPad.getCounterRatePerSec(CifsVserverCounter.read_data));
        cifsVserverMeasurements.setCifsWriteLatency(currentScratchPad.getLatencyPerOp(CifsVserverCounter.cifs_write_latency, CifsVserverCounter.cifs_write_ops));
        cifsVserverMeasurements.setCifsWriteOps(currentScratchPad.getCounterRatePerSec(CifsVserverCounter.cifs_write_ops));
        cifsVserverMeasurements.setCifsWriteThroughput(currentScratchPad.getCounterRatePerSec(CifsVserverCounter.write_data));
        cifsVserverMeasurements.setCifsThroughput(MetricsRowScratchPad.add(cifsVserverMeasurements.getCifsReadThroughput(),
                                                cifsVserverMeasurements.getCifsWriteThroughput()));
        return cifsVserverMeasurements;
    }


    @Override
    protected CifsVserverRawMeasurements rawMetrics(ArchiveMetricsScratchPad<CifsVserverCounter> currentScratchPad, NetAppDefaultStorageContext context) {
        final CifsVserverRawMeasurements cifsVserverRawMeasurements = new CifsVserverRawMeasurements();
        Map<CifsVserverCounter, BigInteger> counterValue = currentScratchPad.getCounterValues();

        cifsVserverRawMeasurements.setCifsLatency(counterValue.get(CifsVserverCounter.cifs_latency));
        cifsVserverRawMeasurements.setCifsLatencyBase(counterValue.get(CifsVserverCounter.cifs_latency_base));
        cifsVserverRawMeasurements.setCifsOps(counterValue.get(CifsVserverCounter.cifs_ops));
        cifsVserverRawMeasurements.setCifsReadLatency(counterValue.get(CifsVserverCounter.cifs_read_latency));
        cifsVserverRawMeasurements.setCifsReadOps(counterValue.get(CifsVserverCounter.cifs_read_ops));
        cifsVserverRawMeasurements.setCifsReadThroughput(counterValue.get(CifsVserverCounter.read_data));
        cifsVserverRawMeasurements.setCifsWriteOps(counterValue.get(CifsVserverCounter.cifs_write_ops));
        cifsVserverRawMeasurements.setCifsWriteThroughput(counterValue.get(CifsVserverCounter.write_data));
        cifsVserverRawMeasurements.setCifsWriteLatency(counterValue.get(CifsVserverCounter.cifs_write_latency));
        cifsVserverRawMeasurements.setCifsThroughput(MetricsRowScratchPad.add(cifsVserverRawMeasurements.getCifsReadThroughput(),
                                                    cifsVserverRawMeasurements.getCifsWriteThroughput()));
        return cifsVserverRawMeasurements;
    }
}