package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.fcplif.interfaces.data.FcpLifMeasurements;
import com.netapp.oci.netapp.performance.fcplif.interfaces.data.FcpLifRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

public class FcpLifMetricsBuilder extends ArchiveMetricsBuilderBase<VServer.VServerKey,
        FcpLifMeasurements, FcpLifRawMeasurements, FcpLifMetricsBuilder.FcpLIFCounter> {

    private static final Logger logger = LoggerFactory.getLogger(FcpLifMetricsBuilder.class);

    public static enum FcpLIFCounter implements ArchiveCounterMetaData {

        avg_other_latency,
        avg_read_latency,
        avg_write_latency,
        avg_latency,
        instance_uuid,
        instance_name,
        other_ops,
        port_id,
        port_name,
        read_data,
        read_ops,
        total_ops,
        vserver_name,
        vserver_uuid,
        write_data,
        write_ops;


        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private FcpLIFCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private FcpLIFCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }


    public FcpLifMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "fcp_lif", FcpLIFCounter.class);
    }

    @Override
    protected VServer.VServerKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID,
                                                           String nodeName, Map<String, Object> counterRow) {
        final String vServerUUID = (String) counterRow.get(FcpLIFCounter.vserver_uuid.name());
        if (vServerUUID == null || vServerUUID.isEmpty()) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for iscsi lif " + perfInstanceUUID);
            return null;
        }
        return new VServer.VServerKey(context.getKey().getID(Cluster.ClusterKey.Field.cluster_uuid), vServerUUID);
    }

    @Override
    protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context,
                                                         final String performanceInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        Map<String, String> fcpLifNameToFoundationUUID = context.getFcpLifNameToFoundationUUIDMap();
        String vServerName = (String)counterRow.get(FcpLIFCounter.vserver_name.name());
        String fcplifName = (String)counterRow.get(FcpLIFCounter.instance_name.name());
        String fcpFoundationLif = null ;
        if (vServerName != null && fcplifName != null){
            fcpFoundationLif = vServerName + "." + fcplifName ;
        }

        String fcpLifFoundationUUID = fcpLifNameToFoundationUUID.get(fcpFoundationLif);

        if(fcpLifFoundationUUID == null){
            logger.warn("{} - Failed to find UUID for fcp lif {}", new Object[] {PackageContext.getLogPrefix(), fcplifName});
        }

        return fcpLifFoundationUUID;
    }

    @Override
    public FcpLIFCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return FcpLIFCounter.values();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey,
                                  Map<VServer.VServerKey, Collection<RawSample<FcpLifMeasurements>>> measurements)
            throws DataSourceErrorException {
        if(measurements == null ) {
            return;
        }

        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<FcpLifMeasurements>>> measurementEntry : measurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<FcpLifMeasurements>> fcpLifMeasurements = measurementEntry.getValue();
            reportingPackage.reportFcpLifPerformance(clusterKey, key, fcpLifMeasurements);
        }
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey,
                                     Map<VServer.VServerKey, Collection<RawSample<FcpLifRawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<FcpLifRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<FcpLifRawMeasurements>> fcpLifrawMeasurements = measurementEntry.getValue();
            reportingPackage.reportFcpLifRawPerformance(clusterKey, key, fcpLifrawMeasurements);
        }
    }

    @Override
    protected FcpLifMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<FcpLIFCounter> currentScratchPad,
                                                    final NetAppDefaultStorageContext context) {
        final FcpLifMeasurements fcpLifMeasurements = new FcpLifMeasurements();




        fcpLifMeasurements.setAvgLatency(currentScratchPad.getLatencyPerOp(FcpLIFCounter.avg_latency, FcpLIFCounter.total_ops));
        fcpLifMeasurements.setAvgOtherLatency(currentScratchPad.getLatencyPerOp(FcpLIFCounter.avg_other_latency, FcpLIFCounter.other_ops));
        fcpLifMeasurements.setAvgReadLatency(currentScratchPad.getLatencyPerOp(FcpLIFCounter.avg_read_latency, FcpLIFCounter.read_ops));
        fcpLifMeasurements.setAvgWriteLatency(currentScratchPad.getLatencyPerOp(FcpLIFCounter.avg_write_latency, FcpLIFCounter.write_ops));
        fcpLifMeasurements.setOtherOps(currentScratchPad.getCounterRatePerSec(FcpLIFCounter.other_ops));
        fcpLifMeasurements.setReadOps(currentScratchPad.getCounterRatePerSec(FcpLIFCounter.read_ops));
        fcpLifMeasurements.setTotalOps(currentScratchPad.getCounterRatePerSec(FcpLIFCounter.total_ops));
        fcpLifMeasurements.setWriteOps(currentScratchPad.getCounterRatePerSec(FcpLIFCounter.write_ops));

        float readData = currentScratchPad.getCounterRatePerSec(FcpLIFCounter.read_data);
        float writeData = currentScratchPad.getCounterRatePerSec(FcpLIFCounter.write_data);

        fcpLifMeasurements.setReadData(readData);
        fcpLifMeasurements.setWriteData(writeData);
        fcpLifMeasurements.setThroughput(MetricsRowScratchPad.add(readData,writeData));
        return fcpLifMeasurements;
    }


    @Override
    protected FcpLifRawMeasurements rawMetrics(ArchiveMetricsScratchPad<FcpLIFCounter> currentScratchPad, NetAppDefaultStorageContext context) {
        final FcpLifRawMeasurements fcpLifrawMeasurements = new FcpLifRawMeasurements();
        Map<FcpLIFCounter, BigInteger> counterValue = currentScratchPad.getCounterValues();

        fcpLifrawMeasurements.setAvgLatency(counterValue.get(FcpLIFCounter.avg_latency));
        fcpLifrawMeasurements.setAvgOtherLatency(counterValue.get(FcpLIFCounter.avg_other_latency));
        fcpLifrawMeasurements.setAvgReadLatency(counterValue.get(FcpLIFCounter.avg_read_latency));
        fcpLifrawMeasurements.setAvgWriteLatency(counterValue.get(FcpLIFCounter.avg_write_latency));
        fcpLifrawMeasurements.setOtherOps(counterValue.get(FcpLIFCounter.other_ops));
        fcpLifrawMeasurements.setReadOps(counterValue.get(FcpLIFCounter.read_ops));
        fcpLifrawMeasurements.setWriteOps(counterValue.get(FcpLIFCounter.write_ops));
        fcpLifrawMeasurements.setTotalOps(counterValue.get(FcpLIFCounter.total_ops));
        fcpLifrawMeasurements.setReadData(counterValue.get(FcpLIFCounter.read_data));
        fcpLifrawMeasurements.setWriteData(counterValue.get(FcpLIFCounter.write_data));
        fcpLifrawMeasurements.setThroughput(MetricsRowScratchPad.add(
                fcpLifrawMeasurements.getReadData(),fcpLifrawMeasurements.getWriteData()));
        return fcpLifrawMeasurements;
    }
}