package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.fcpport.interfaces.data.FcpPortMeasurements;
import com.netapp.oci.netapp.performance.fcpport.interfaces.data.FcpPortRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.FcpPort;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.math.BigInteger;
import java.util.Collection;
import java.util.List;
import java.util.Map;


public class FcpPortMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
        FcpPortMeasurements, FcpPortRawMeasurements, FcpPortMetricsBuilder.FcpPortCounter> {

    private static final long MEGABIT_1 = 1000000;

    public static enum FcpPortCounter implements ArchiveCounterMetaData {

        avg_other_latency,
        avg_read_latency,
        avg_write_latency,
        instance_name,
        instance_uuid,
        link_speed,
        node_name,
        node_uuid,
        other_ops,
        port_id,
        read_ops,
        write_ops;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private FcpPortCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private FcpPortCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }

    public FcpPortMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "fcp_port", FcpPortCounter.class);
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID,
                                                           final String nodeName,
                                                           final Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context,
                                                         final String performanceInstanceUUID, String nodeName, Map<String, Object> counterRow) {
        final String normalizedPerfUUID = BuilderUtil.normalizeWWN(performanceInstanceUUID);
        return context.getFcpPortPerfUUIDToUUIDMap().get(normalizedPerfUUID);
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey, Map<Cluster.ClusterKey,
            Collection<RawSample<FcpPortMeasurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<FcpPortMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<FcpPortMeasurements>> fcpPortMeasurements = measurementEntry.getValue();
            reportingPackage.reportFcpPortPerformance(key, fcpPortMeasurements);
        }
    }

    @Override
    public FcpPortCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return FcpPortCounter.values();
    }


    @Override
    protected FcpPortMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<FcpPortCounter> currentScratchPad,
                                                                  final NetAppDefaultStorageContext context) {
        final FcpPortMeasurements fcpPortMeasurements = new FcpPortMeasurements();
        fcpPortMeasurements.setAvgOtherLatency(currentScratchPad.getLatencyPerOp(FcpPortCounter.avg_other_latency, FcpPortCounter.other_ops));
        fcpPortMeasurements.setAvgReadLatency(currentScratchPad.getLatencyPerOp(FcpPortCounter.avg_read_latency, FcpPortCounter.read_ops));
        fcpPortMeasurements.setAvgWriteLatency(currentScratchPad.getLatencyPerOp(FcpPortCounter.avg_write_latency, FcpPortCounter.write_ops));
        fcpPortMeasurements.setReadData(0f); // TODO Missing counters from archvier
        fcpPortMeasurements.setWriteData(0f); // TODO MIssing counters from archiver
        fcpPortMeasurements.setOtherOps(currentScratchPad.getCounterRatePerSec(FcpPortCounter.other_ops));
        fcpPortMeasurements.setReadOps(currentScratchPad.getCounterRatePerSec(FcpPortCounter.read_ops));
        fcpPortMeasurements.setWriteOps(currentScratchPad.getCounterRatePerSec(FcpPortCounter.write_ops));
        fcpPortMeasurements.setTotalOps(MetricsRowScratchPad.add(
                fcpPortMeasurements.getReadOps(), fcpPortMeasurements.getWriteOps(), fcpPortMeasurements.getOtherOps()));

        //  Compute (read/write)_throughput at this Port
        //  We pass the fcpLifuuid list to addCounters method which aggregates the counters  at globalScratchPad
        //  and returns its sum
        String fcpPortWWN = BuilderUtil.normalizeWWN(currentScratchPad.getRawValues().get(FcpPortCounter.instance_uuid));
        List<String> fcpLifUUIDList = context.getFcpPortWWNToLifUUIDMap().get(fcpPortWWN);
        Map<FcpLifMetricsBuilder.FcpLIFCounter, Float> sumResults = reportingPackage.getGlobalScratchPad()
                .addCounterRatePerSec("fcp_lif", fcpLifUUIDList,
                        FcpLifMetricsBuilder.FcpLIFCounter.read_data,
                        FcpLifMetricsBuilder.FcpLIFCounter.write_data);

        float read_throughput = sumResults.get(FcpLifMetricsBuilder.FcpLIFCounter.read_data);
        float write_throughput = sumResults.get(FcpLifMetricsBuilder.FcpLIFCounter.write_data);

        fcpPortMeasurements.setReadThroughput(read_throughput);
        fcpPortMeasurements.setWriteThroughput(write_throughput);
        fcpPortMeasurements.setThroughput(MetricsRowScratchPad.add(fcpPortMeasurements.getReadThroughput(), fcpPortMeasurements.getWriteThroughput()));

        //Fetch fcpPort utilization from fcpPortToLinkSpeedMap
        fcpPortMeasurements.setUtilization(calculateFcpPortUtilization(fcpPortMeasurements.getThroughput(), context.getFcpPortToLinkSpeedMap().get(currentScratchPad.getInventoryUUID())));
        return fcpPortMeasurements;
    }

    @Override
    protected FcpPortRawMeasurements rawMetrics(ArchiveMetricsScratchPad<FcpPortCounter> currentScratchPad, NetAppDefaultStorageContext context) {
        Map<FcpPortCounter,BigInteger> counterValues = currentScratchPad.getCounterValues();
        final FcpPortRawMeasurements fcpPortRawMeasurements = new FcpPortRawMeasurements();
        fcpPortRawMeasurements.setAvgOtherLatency(counterValues.get(FcpPortCounter.avg_other_latency));
        fcpPortRawMeasurements.setAvgReadLatency(counterValues.get(FcpPortCounter.avg_read_latency));
        fcpPortRawMeasurements.setAvgWriteLatency(counterValues.get(FcpPortCounter.avg_write_latency));
        fcpPortRawMeasurements.setReadOps(counterValues.get(FcpPortCounter.read_ops));
        fcpPortRawMeasurements.setOtherOps(counterValues.get(FcpPortCounter.other_ops));
        fcpPortRawMeasurements.setWriteOps(counterValues.get(FcpPortCounter.write_ops));
        fcpPortRawMeasurements.setTotalOps(MetricsRowScratchPad.add(fcpPortRawMeasurements.getReadOps(),
                fcpPortRawMeasurements.getWriteOps(), fcpPortRawMeasurements.getOtherOps()));
        //fcpPortRawMeasurements.setReadData(0); // TODO Missing counters froim archvier
        //fcpPortRawMeasurements.setWriteData(0); // TODO MIssing counters from archiver

        //  Compute (read/write)_throughput at this Port
        //  We pass the fcpLifuuid list to addCounters method which aggregates the counters  at globalScratchPad
        //  and returns its sum
        String fcpPortWWN = BuilderUtil.normalizeWWN(currentScratchPad.getRawValues().get(FcpPortCounter.instance_uuid));
        List<String> fcpLifUUIDList = context.getFcpPortWWNToLifUUIDMap().get(fcpPortWWN);
        Map<FcpLifMetricsBuilder.FcpLIFCounter, BigInteger> sumResults = reportingPackage.getGlobalScratchPad()
                .addByCountersRaw("fcp_lif", fcpLifUUIDList,
                        FcpLifMetricsBuilder.FcpLIFCounter.read_data,
                        FcpLifMetricsBuilder.FcpLIFCounter.write_data);

        BigInteger read_throughput = sumResults.get(FcpLifMetricsBuilder.FcpLIFCounter.read_data);
        BigInteger write_throughput = sumResults.get(FcpLifMetricsBuilder.FcpLIFCounter.write_data);

        fcpPortRawMeasurements.setReadThroughput(read_throughput);
        fcpPortRawMeasurements.setWriteThroughput(write_throughput);
        fcpPortRawMeasurements.setThroughput(MetricsRowScratchPad.add(fcpPortRawMeasurements.getReadThroughput(), fcpPortRawMeasurements.getWriteThroughput()));

        return fcpPortRawMeasurements;
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey, Map<Cluster.ClusterKey,
                                        Collection<RawSample<FcpPortRawMeasurements>>> rawMeasurements)
                                        throws DataSourceErrorException {
        for(final Map.Entry<Cluster.ClusterKey,
                Collection<RawSample<FcpPortRawMeasurements>>> rawMeasurementEntry : rawMeasurements.entrySet()) {
            final Cluster.ClusterKey key = rawMeasurementEntry.getKey();
            final Collection<RawSample<FcpPortRawMeasurements>> fcpPortRawMeasurements = rawMeasurementEntry.getValue();
            reportingPackage.reportFcpPortRawPerformance(key, fcpPortRawMeasurements);
        }
     }

    /**
     * @param fcpPortThroughput
     * @param linkSpeed
     * @return float fcpPortUtilization (fcpPortThroughput*8(converting bytes to bits) / link_speed) * 100
     */
    private float calculateFcpPortUtilization(float fcpPortThroughput, Integer linkSpeed){

        if(null == linkSpeed)
            return MetricsRowSample.NULL_FLOAT;
        long linkSpeedInBitsPerSec = linkSpeed * 1000 * MEGABIT_1;
        if(linkSpeedInBitsPerSec > 0){
            // multiplying by 8 to convert Bytes/sec to Bits/sec and 100 to return utilization in 0 to 100 % range
            return (float)(100 * ((8*(double)fcpPortThroughput)/linkSpeedInBitsPerSec));
        }else{
            // if we have 0 or negative values(error handling), then utilization cannot be calibrated, hence setting to null representation for float
            return MetricsRowSample.NULL_FLOAT;
        }
    }
}