package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifMeasurements;
import com.netapp.oci.netapp.performance.iscsilif.interfaces.data.IscsiLifRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

public class ISCSILIFMetricsBuilder extends ArchiveMetricsBuilderBase<VServer.VServerKey,
        IscsiLifMeasurements, IscsiLifRawMeasurements, ISCSILIFMetricsBuilder.ISCSILIFCounter> {

    private static final Logger logger = LoggerFactory.getLogger(ISCSILIFMetricsBuilder.class);

    public static enum ISCSILIFCounter implements ArchiveCounterMetaData {

        avg_latency,
        avg_other_latency,
        avg_read_latency,
        avg_write_latency,
        cmd_transfered,
        current_port,
        instance_name,
        instance_uuid,
        iscsi_other_ops,
        iscsi_read_ops,
        iscsi_write_ops,
        data_in_blocks, // TODO confirm if this exists in archiver file
        data_out_blocks, // TODO confirm if this exists in archiver file
        read_data,
        vserver_name,
        vserver_uuid,
        write_data;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private ISCSILIFCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private ISCSILIFCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }
    }


    public ISCSILIFMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "iscsi_lif", ISCSILIFCounter.class);
    }

    @Override
    protected VServer.VServerKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfInstanceUUID,
                                                           String nodeName, Map<String, Object> counterRow) {
        final String vServerUUID = (String) counterRow.get(ISCSILIFCounter.vserver_uuid.name());
        if (vServerUUID == null || vServerUUID.isEmpty()) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for iscsi lif " + perfInstanceUUID);
            return null;
        }
        return new VServer.VServerKey(context.getKey().getID(Cluster.ClusterKey.Field.cluster_uuid), vServerUUID);
    }

    @Override
    public ISCSILIFCounter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return ISCSILIFCounter.values();
    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey,
                                  Map<VServer.VServerKey, Collection<RawSample<IscsiLifMeasurements>>> measurements)
                                  throws DataSourceErrorException {
        if(measurements == null ) {
            return;
        }

        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<IscsiLifMeasurements>>> measurementEntry : measurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<IscsiLifMeasurements>> iscsiLifMeasurements = measurementEntry.getValue();
            reportingPackage.reportIscsiLifPerformance(clusterKey, key, iscsiLifMeasurements);
        }
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey,
                                     Map<VServer.VServerKey, Collection<RawSample<IscsiLifRawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<IscsiLifRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final VServer.VServerKey key = measurementEntry.getKey();
            final Collection<RawSample<IscsiLifRawMeasurements>> iscsiLifrawMeasurements = measurementEntry.getValue();
            reportingPackage.reportIscsiLifRawPerformance(clusterKey, key, iscsiLifrawMeasurements);
        }
    }

    @Override
    protected IscsiLifMeasurements normalizeMetrics(final ArchiveMetricsScratchPad<ISCSILIFCounter> currentScratchPad,
                                                   final NetAppDefaultStorageContext context) {
        final IscsiLifMeasurements iscsiLifMeasurements = new IscsiLifMeasurements();

        iscsiLifMeasurements.setAvgLatency(currentScratchPad.getLatencyPerOp(ISCSILIFCounter.avg_latency, ISCSILIFCounter.cmd_transfered));
        iscsiLifMeasurements.setAvgOtherLatency(currentScratchPad.getLatencyPerOp(ISCSILIFCounter.avg_other_latency, ISCSILIFCounter.iscsi_other_ops));
        iscsiLifMeasurements.setAvgReadLatency(currentScratchPad.getLatencyPerOp(ISCSILIFCounter.avg_read_latency, ISCSILIFCounter.iscsi_read_ops));
        iscsiLifMeasurements.setAvgWriteLatency(currentScratchPad.getLatencyPerOp(ISCSILIFCounter.avg_write_latency, ISCSILIFCounter.iscsi_write_ops));
        iscsiLifMeasurements.setDataInBlocks(currentScratchPad.getCounterDeltaFinal(ISCSILIFCounter.data_in_blocks));
        iscsiLifMeasurements.setCmdTransfered(currentScratchPad.getCounterDeltaFinal(ISCSILIFCounter.cmd_transfered));
        iscsiLifMeasurements.setDataOutBlocks(currentScratchPad.getCounterDeltaFinal(ISCSILIFCounter.data_out_blocks));

        float iscsiReadOps = currentScratchPad.getCounterRatePerSec(ISCSILIFCounter.iscsi_read_ops);
        float iscsiWriteOps = currentScratchPad.getCounterRatePerSec(ISCSILIFCounter.iscsi_write_ops);
        float iscsiOtherOps = currentScratchPad.getCounterRatePerSec(ISCSILIFCounter.iscsi_other_ops);
        float iscsiReadData = currentScratchPad.getCounterRatePerSec(ISCSILIFCounter.read_data);
        float iscsiWriteData = currentScratchPad.getCounterRatePerSec(ISCSILIFCounter.write_data);

        iscsiLifMeasurements.setIscsiReadOps(iscsiReadOps);
        iscsiLifMeasurements.setIscsiWriteOps(iscsiWriteOps);
        iscsiLifMeasurements.setIscsiOtherOps(iscsiOtherOps);
        iscsiLifMeasurements.setReadData(iscsiReadData);
        iscsiLifMeasurements.setWriteData(iscsiWriteData);
        iscsiLifMeasurements.setOps(MetricsRowScratchPad.add(iscsiReadOps,iscsiWriteOps, iscsiOtherOps));
        iscsiLifMeasurements.setThroughput(MetricsRowScratchPad.add(iscsiReadData, iscsiWriteData));

        return iscsiLifMeasurements;
    }


    @Override
    protected IscsiLifRawMeasurements rawMetrics(ArchiveMetricsScratchPad<ISCSILIFCounter> currentScratchPad, NetAppDefaultStorageContext context) {
        final IscsiLifRawMeasurements iscsiLifrawMeasurements = new IscsiLifRawMeasurements();
        Map<ISCSILIFCounter, BigInteger> counterValue = currentScratchPad.getCounterValues();

        iscsiLifrawMeasurements.setAvgLatency(counterValue.get(ISCSILIFCounter.avg_latency));
        iscsiLifrawMeasurements.setAvgOtherLatency(counterValue.get(ISCSILIFCounter.avg_other_latency));
        iscsiLifrawMeasurements.setAvgReadLatency(counterValue.get(ISCSILIFCounter.avg_read_latency));
        iscsiLifrawMeasurements.setAvgWriteLatency(counterValue.get(ISCSILIFCounter.avg_write_latency));
        iscsiLifrawMeasurements.setIscsiOtherOps(counterValue.get(ISCSILIFCounter.iscsi_other_ops));
        iscsiLifrawMeasurements.setCmdTransfered(counterValue.get(ISCSILIFCounter.cmd_transfered));
        iscsiLifrawMeasurements.setIscsiReadOps(counterValue.get(ISCSILIFCounter.iscsi_read_ops));
        iscsiLifrawMeasurements.setIscsiWriteOps(counterValue.get(ISCSILIFCounter.iscsi_write_ops));
        iscsiLifrawMeasurements.setReadData(counterValue.get(ISCSILIFCounter.read_data));
        iscsiLifrawMeasurements.setWriteData(counterValue.get(ISCSILIFCounter.write_data));
        iscsiLifrawMeasurements.setOps(MetricsRowScratchPad.add(
                iscsiLifrawMeasurements.getIscsiReadOps(), iscsiLifrawMeasurements.getIscsiWriteOps(),
                iscsiLifrawMeasurements.getIscsiOtherOps()));
        iscsiLifrawMeasurements.setThroughput(MetricsRowScratchPad.add(
                iscsiLifrawMeasurements.getReadData(), iscsiLifrawMeasurements.getWriteData()));
        return iscsiLifrawMeasurements;
    }
}