package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifMeasurements;
import com.netapp.oci.netapp.performance.networklif.interfaces.data.NetworkLifRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer.VServerKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

public class LifMetricsBuilder extends ArchiveMetricsBuilderBase<VServer.VServerKey,
NetworkLifMeasurements, NetworkLifRawMeasurements, LifMetricsBuilder.LifCounter>{
	
	 private static final Logger logger = LoggerFactory.getLogger(LifMetricsBuilder.class);
	 public static enum LifCounter implements ArchiveCounterMetaData {
		 recv_data,
		 recv_packet,
		 recv_errors,
		 sent_data,
		 sent_packet,
		 sent_errors,
		 
		 /**
		  * vserver_id is not for reporting but used for lif_vserver counter aggregation.
		  */
		 vserver_id;
		
		private final String onTapCounterName;
        private final int onTapCounterIndex;
        
        private LifCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private LifCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }


		@Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}
	 }
	 
	 public LifMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
			super(reportingPackage, "lif", LifCounter.class);
		}

	

	@Override
	protected LifCounter[] getCountersMetaData(
			NetAppDefaultStorageContext context) {
		return LifCounter.values();
	}

	@Override
	protected NetworkLifMeasurements normalizeMetrics(
			ArchiveMetricsScratchPad<LifCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		NetworkLifMeasurements lifMeasurements = new NetworkLifMeasurements();
		lifMeasurements.setRecvData(currentScratchPad.getCounterRatePerSec(LifCounter.recv_data));
		lifMeasurements.setRecvErrors(currentScratchPad.getCounterRatePerSec(LifCounter.recv_errors));
		lifMeasurements.setRecvPacket(currentScratchPad.getCounterRatePerSec(LifCounter.recv_packet));
		lifMeasurements.setSentData(currentScratchPad.getCounterRatePerSec(LifCounter.sent_data));
		lifMeasurements.setSentErrors(currentScratchPad.getCounterRatePerSec(LifCounter.sent_errors));
		lifMeasurements.setSentPacket(currentScratchPad.getCounterRatePerSec(LifCounter.sent_packet));
		lifMeasurements.setOpmLifPackets(MetricsRowScratchPad.add(lifMeasurements.getRecvPacket(), lifMeasurements.getSentPacket()));
		lifMeasurements.setOpmLifThroughput(MetricsRowScratchPad.add(lifMeasurements.getRecvData(), lifMeasurements.getSentData()));
		return lifMeasurements;
	}
	
	@Override
	public void reportPerformance(Cluster.ClusterKey clusterKey,
			Map<VServerKey, Collection<RawSample<NetworkLifMeasurements>>> measurements)
			throws DataSourceErrorException {
		if (measurements == null) {
            return;
        }
        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<NetworkLifMeasurements>>> measurementEntry : measurements.entrySet()) {
            final VServer.VServerKey vServerKey = measurementEntry.getKey();
            final Collection<RawSample<NetworkLifMeasurements>> lifMeasurements = measurementEntry.getValue();
            reportingPackage.reportLifPerformance(clusterKey, vServerKey, lifMeasurements);
        }
	}

	@Override
	protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context,
                final String performanceInstanceUUID, String nodeName, Map<String, Object> counterRow) {
		Map<String, String> networkLifNameToFoundationUUID = context.getNetworkLifNameToFoundationUUIDMap();
		String vServerName = (String)counterRow.get("vserver_name");
		String lifName = (String)counterRow.get("instance_name");
		String fqLifName = null ;
		if (vServerName != null && lifName != null){
			 fqLifName = vServerName + ":" + lifName ;
		}
		String lifFoundationUUID = networkLifNameToFoundationUUID.get(fqLifName);
		if(lifFoundationUUID != null){
			return lifFoundationUUID;
		}
		logger.warn("{} - Failed to find UUID for network lif {}", new Object[] {PackageContext.getLogPrefix(), lifName});
		return null;
	}
	
	@Override
	protected VServerKey getParentInventoryUUIDKey(
			NetAppDefaultStorageContext context, String perfInstanceUUID,
			String nodeName, Map<String, Object> counterRow) {
		Map<String, String> LifNameToVServerUUIDMap = context.getNetworkLifNameToVServerUUIDMap();
		String fqLifName = (String)counterRow.get("vserver_name") + ":" + (String)counterRow.get("instance_name");
		String vServerUUID = LifNameToVServerUUIDMap.get(fqLifName);
		if (vServerUUID == null) {
	            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for network lif " + perfInstanceUUID);
	            return null;
	     }
		
		return new VServer.VServerKey(context.getKey().getID(Cluster.ClusterKey.Field.cluster_uuid), vServerUUID);
		
	}

	@Override
	public void reportRawPerformance(Cluster.ClusterKey clusterKey,
			Map<VServerKey, Collection<RawSample<NetworkLifRawMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		for (final Map.Entry<VServer.VServerKey, Collection<RawSample<NetworkLifRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) { 
            final VServer.VServerKey key = measurementEntry.getKey(); 
            final Collection<RawSample<NetworkLifRawMeasurements>> lifMeasurements = measurementEntry.getValue(); 
            reportingPackage.reportLifRawPerformance(clusterKey, key, lifMeasurements); 
        } 
	}

	@Override
	protected NetworkLifRawMeasurements rawMetrics(
			ArchiveMetricsScratchPad<LifCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		final NetworkLifRawMeasurements m = new NetworkLifRawMeasurements();
        Map<LifCounter,BigInteger> counterValue = currentScratchPad.getCounterValues();
        m.setRecvData(counterValue.get(LifCounter.recv_data));
        m.setRecvErrors(counterValue.get(LifCounter.recv_errors));
        m.setRecvPacket(counterValue.get(LifCounter.recv_packet));        
        m.setSentData(counterValue.get(LifCounter.sent_data));
        m.setSentErrors(counterValue.get(LifCounter.sent_errors));
        m.setSentPacket(counterValue.get(LifCounter.sent_packet));
        m.setOpmLifPackets(MetricsRowScratchPad.add(m.getRecvPacket(), m.getSentPacket()));
        m.setOpmLifThroughput(MetricsRowScratchPad.add(m.getRecvData(), m.getSentData()));
        return m;
    }

	
}
