package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv3Measurements;
import com.netapp.oci.netapp.performance.nfs.interfaces.data.Nfsv3RawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

/**
 * Builder class for processing and reporting NFSv3 metrics from archiver
 */
public class Nfsv3MetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey, Nfsv3Measurements, Nfsv3RawMeasurements,
        Nfsv3MetricsBuilder.NFSv3Counter>   {

    private static final Logger logger = LoggerFactory.getLogger(Nfsv3MetricsBuilder.class);

    public static enum NFSv3Counter implements ArchiveCounterMetaData {

        nfsv3_ops,
        nfsv3_read_ops,
        nfsv3_write_ops,
        read_total,
        nfsv3_read_throughput,
        nfsv3_throughput,
        write_total,
        nfsv3_write_throughput,
        read_avg_latency,
        write_avg_latency,
        latency,
        total_ops,
        instance_uuid;

        private final String onTapCounterName;
        private final int onTapCounterIndex;

        private NFSv3Counter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private NFSv3Counter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }

        private NFSv3Counter(int onTapCounterIndex) {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = onTapCounterIndex;
        }

        @Override
        public String getOnTapCounterName() {
            return onTapCounterName;
        }

        @Override
        public int getOnTapCounterIndex() {
            return onTapCounterIndex;
        }

    }

    /**
     * constructor
     *
     * @param reportingPackage
     */

    public Nfsv3MetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "nfsv3", NFSv3Counter.class);
    }


    /**
     * Retrieve the vserver_id fromm the performance archive
     *
     * @param context    The foundation context
     * @param nodeName   Node name from where the archiver was retrieved
     * @param counterRow Raw counter data form the performance archive
     * @return Performance Instance UUID
     */
    @Override
    protected String getPerformanceUUID(final NetAppDefaultStorageContext context,
                                        final String nodeName, final Map<String, Object> counterRow) {
        String instanceUUID = (String)counterRow.get(NFSv3Counter.instance_uuid.name()); // Format : vserverID(23):processor(233)
        String vserverId = BuilderUtil.extractIdFromParens(instanceUUID, "vserverID");
        String vServerUUID = null;

        if(vserverId != null && !vserverId.isEmpty()) {
            Map<String, String> vserverIDToUUIDMap = context.getNfsv3VserverIDToUUIDMap();
            vServerUUID = vserverIDToUUIDMap.get(vserverId);
            if (vServerUUID == null) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerId for NFSv3 " + nodeName);
            }
        } else {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServerId for NFSv3 " + nodeName+ " with vserverID "+vserverId);
        }

        return vServerUUID;
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context,
                                                           final String perfInstanceUUID, final String nodeName,
                                                           final Map<String, Object> counterRow) {
        return context.getKey();
    }

    @Override
    public void reportPerformance(final Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey,
            Collection<RawSample<Nfsv3Measurements>>> measurements) throws DataSourceErrorException {
        if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<Nfsv3Measurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<Nfsv3Measurements>> nfsv3Measurements = measurementEntry.getValue();
            reportingPackage.reportNfsv3Performance(key, nfsv3Measurements);
        }
    }

    @Override
    public NFSv3Counter[] getCountersMetaData(final NetAppDefaultStorageContext context) {
        return NFSv3Counter.values();
    }


    @Override
    protected Nfsv3Measurements normalizeMetrics(final ArchiveMetricsScratchPad<NFSv3Counter> currentScratchPad,
                                                     final NetAppDefaultStorageContext context) {
        final Nfsv3Measurements nfsv3Measurements = new Nfsv3Measurements();
        nfsv3Measurements.setNfsv3Ops(currentScratchPad.getCounterRatePerSec(NFSv3Counter.nfsv3_ops));
        nfsv3Measurements.setNfsv3ReadOps(currentScratchPad.getCounterRatePerSec(NFSv3Counter.nfsv3_read_ops));
        nfsv3Measurements.setNfsv3WriteOps(currentScratchPad.getCounterRatePerSec(NFSv3Counter.nfsv3_write_ops));
        nfsv3Measurements.setNfsv3ReadThroughput(currentScratchPad.getCounterRatePerSec(NFSv3Counter.nfsv3_read_throughput));
        nfsv3Measurements.setNfsv3Throughput(currentScratchPad.getCounterRatePerSec(NFSv3Counter.nfsv3_throughput));
        nfsv3Measurements.setNfsv3WriteThroughput(currentScratchPad.getCounterRatePerSec(NFSv3Counter.nfsv3_write_throughput));
        nfsv3Measurements.setReadTotal(currentScratchPad.getCounterDeltaFinal(NFSv3Counter.read_total));
        nfsv3Measurements.setWriteTotal(currentScratchPad.getCounterDeltaFinal(NFSv3Counter.write_total));
        nfsv3Measurements.setReadAvgLatency(currentScratchPad.getLatencyPerOp(NFSv3Counter.read_avg_latency, getDenominatorForNumerator(NFSv3Counter.read_avg_latency)));
        nfsv3Measurements.setWriteAvgLatency(currentScratchPad.getLatencyPerOp(NFSv3Counter.write_avg_latency, getDenominatorForNumerator(NFSv3Counter.write_avg_latency)));
        nfsv3Measurements.setLatency(currentScratchPad.getLatencyPerOp(NFSv3Counter.latency, getDenominatorForNumerator(NFSv3Counter.latency)));
        nfsv3Measurements.setTotalOps(currentScratchPad.getCounterRatePerSec(NFSv3Counter.total_ops));
        return  nfsv3Measurements;
    }

    private NFSv3Counter getDenominatorForNumerator(final NFSv3Counter numeratorName) {
        final NFSv3Counter denominatorName;
        switch (numeratorName) {
            case read_avg_latency:
                denominatorName = NFSv3Counter.read_total;
                break;
            case write_avg_latency:
                denominatorName = NFSv3Counter.write_total;
                break;
            case latency:
                denominatorName = NFSv3Counter.total_ops;
                break;
            default:
                denominatorName = null;
        }
        return denominatorName;
    }

    @Override
    public void reportRawPerformance(Cluster.ClusterKey clusterKey, Map<Cluster.ClusterKey, Collection<RawSample<Nfsv3RawMeasurements>>> rawMeasurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<Nfsv3RawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<Nfsv3RawMeasurements>> nfsv3RawMeasurements = measurementEntry.getValue();
            reportingPackage.reportNfsv3RawPerformance(clusterKey, nfsv3RawMeasurements);
        }
    }

    @Override
    protected Nfsv3RawMeasurements rawMetrics(ArchiveMetricsScratchPad<NFSv3Counter> currentScratchPad, NetAppDefaultStorageContext context) {
        final Nfsv3RawMeasurements m = new Nfsv3RawMeasurements();
        Map<NFSv3Counter,BigInteger> counterValues = currentScratchPad.getCounterValues();
        m.setNfsv3Ops(counterValues.get(NFSv3Counter.nfsv3_ops));
        m.setNfsv3ReadOps(counterValues.get(NFSv3Counter.nfsv3_read_ops));
        m.setReadTotal(counterValues.get(NFSv3Counter.read_total));
        m.setNfsv3ReadThroughput(counterValues.get(NFSv3Counter.nfsv3_read_throughput));
        m.setNfsv3Throughput(counterValues.get(NFSv3Counter.nfsv3_throughput));
        m.setNfsv3WriteOps(counterValues.get(NFSv3Counter.nfsv3_write_ops));
        m.setNfsv3WriteThroughput(counterValues.get(NFSv3Counter.nfsv3_write_throughput));
        m.setReadAvgLatency(counterValues.get(NFSv3Counter.read_avg_latency));
        m.setWriteAvgLatency(counterValues.get(NFSv3Counter.write_avg_latency));
        m.setWriteTotal(counterValues.get(NFSv3Counter.write_total));
        m.setLatency(counterValues.get(NFSv3Counter.latency));
        m.setTotalOps(counterValues.get(NFSv3Counter.total_ops));
        return m;
    }

}
