package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.performance.nic.interfaces.data.NicMeasurements;
import com.netapp.oci.netapp.performance.nic.interfaces.data.NicRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.NetworkPort.Speed;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

/**
 * nic_common counter values are fetched from archiver for FS or higher Ontap counter collection..
 * @author DPADMANA
 *
 */
public class NicCommonMetricsBuilder extends ArchiveMetricsBuilderBase<ClusterKey, NicMeasurements,NicRawMeasurements, NicCommonMetricsBuilder.NicCommonCounter> {
	private static final Logger logger = LoggerFactory.getLogger(NicCommonMetricsBuilder.class);
	private static final long MEGABIT_1 = 1000000;

	
	public static enum NicCommonCounter implements ArchiveCounterMetaData {
		rx_bytes_rate,
		rx_errors_rate,
		tx_bytes_rate,
		tx_errors_rate;

		private final String onTapCounterName;
		private final int onTapCounterIndex;

		private NicCommonCounter(final String onTapCounterName,
				final int onTapCounterIndex) {
			this.onTapCounterName = onTapCounterName;
			this.onTapCounterIndex = onTapCounterIndex;
		}

		private NicCommonCounter() {
			this.onTapCounterName = this.name();
			this.onTapCounterIndex = -1;
		}

		@Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}

	}

	public NicCommonMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
		super(reportingPackage, "nic_common", NicCommonCounter.class);
	}

	@Override
	protected NicCommonCounter[] getCountersMetaData(
	NetAppDefaultStorageContext context) {
		return NicCommonCounter.values();
	}

	@Override
	protected NicMeasurements normalizeMetrics(ArchiveMetricsScratchPad<NicCommonCounter> currentScratchPad, NetAppDefaultStorageContext context) {
		NicMeasurements nicCommonMeasurements = new NicMeasurements();
		
		float byteRate = currentScratchPad.getCounterRatePerSec(NicCommonCounter.rx_bytes_rate);
		nicCommonMeasurements.setRxBytesRate(byteRate < 0 ? MetricsRowSample.NULL_FLOAT :byteRate); // we cannot have rx byte rate -ve, which we have observed in the raw counter file, handling that case
		byteRate = currentScratchPad.getCounterRatePerSec(NicCommonCounter.rx_errors_rate);
		nicCommonMeasurements.setRxErrorsRate(byteRate < 0 ? MetricsRowSample.NULL_FLOAT :byteRate);
		nicCommonMeasurements.setTxBytesRate(currentScratchPad.getCounterRatePerSec(NicCommonCounter.tx_bytes_rate));
		nicCommonMeasurements.setTxErrorsRate(currentScratchPad.getCounterRatePerSec(NicCommonCounter.tx_errors_rate));
		nicCommonMeasurements.setOpmNicTotalBytePerSec(MetricsRowScratchPad.add(nicCommonMeasurements.getRxBytesRate(), nicCommonMeasurements.getTxBytesRate()));
		nicCommonMeasurements.setOpmNicTotalErrorRate(MetricsRowScratchPad.add(nicCommonMeasurements.getRxErrorsRate(), nicCommonMeasurements.getTxErrorsRate()));
		nicCommonMeasurements.setOpmNicUtilization(calculateOpmNicUtilization(nicCommonMeasurements.getOpmNicTotalBytePerSec(), context.getIpPortToLinkSpeedMap().get(currentScratchPad.getInventoryUUID())));

		return nicCommonMeasurements;
	}

	
	@Override
	protected ClusterKey getParentInventoryUUIDKey(NetAppDefaultStorageContext context, String perfInstanceUUID,String nodeName, Map<String, Object> counterRow) {
		return context.getKey();
	}

	@Override
	public void reportPerformance(ClusterKey clusterKey,Map<ClusterKey, Collection<RawSample<NicMeasurements>>> measurements)throws DataSourceErrorException {
		
		if (measurements == null) {
		         return;
		}
	
		for (final Map.Entry<ClusterKey, Collection<RawSample<NicMeasurements>>> measurementEntry : measurements.entrySet()) {
			final Collection<RawSample<NicMeasurements>> nicCommonMeasurements = measurementEntry.getValue();
			reportingPackage.reportNicPerformance(clusterKey, nicCommonMeasurements);
		}		
	}

	@Override
	public void reportRawPerformance(ClusterKey clusterKey,	Map<ClusterKey, Collection<RawSample<NicRawMeasurements>>> rawMeasurements) throws DataSourceErrorException {
		for (final Map.Entry<ClusterKey, Collection<RawSample<NicRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) { 
            final ClusterKey key = measurementEntry.getKey(); 
            final Collection<RawSample<NicRawMeasurements>> nicCommonRawMeasurements = measurementEntry.getValue();
            reportingPackage.reportNicRawPerformance(key,nicCommonRawMeasurements); 
        } 
	}


	@Override
	protected NicRawMeasurements rawMetrics(ArchiveMetricsScratchPad<NicCommonCounter> currentScratchPad,NetAppDefaultStorageContext context) {
		NicRawMeasurements m = new NicRawMeasurements();
        Map<NicCommonCounter,BigInteger> counterValue = currentScratchPad.getCounterValues();

        // Ensuring we have all +ve byte rate values for the tx and rx. we have seen cases when we get negative number resulting in insertion errors.
        m.setRxBytesRate(validatedRawCounter(counterValue.get(NicCommonCounter.rx_bytes_rate), NicCommonCounter.rx_bytes_rate.name()));
        m.setRxErrorsRate(validatedRawCounter(counterValue.get(NicCommonCounter.rx_errors_rate), NicCommonCounter.rx_errors_rate.name()));
        m.setTxBytesRate(validatedRawCounter(counterValue.get(NicCommonCounter.tx_bytes_rate), NicCommonCounter.tx_bytes_rate.name()));
        m.setTxErrorsRate(validatedRawCounter(counterValue.get(NicCommonCounter.tx_errors_rate), NicCommonCounter.tx_errors_rate.name()));
		m.setOpmNicTotalBytePersec(MetricsRowScratchPad.add(m.getRxBytesRate(), m.getTxBytesRate()));
		m.setOpmNicTotalErrorRate(MetricsRowScratchPad.add(m.getRxErrorsRate(), m.getTxErrorsRate()));

		return m;
	}
	
	protected BigInteger validatedRawCounter(BigInteger br, String counterName){
		if(null != br && br.signum() >= 0){
			return br;
		}else{
			if(br != null && br.signum() < 0){
				logger.error("Ontap opm-archiver file value Error: counter ("+counterName+"), has negative value which is totally wrong: "+ br);
			}
			return MetricsRowSample.NULL_BIG_INT;
		}
	}
	
	/**
	 * Since there is no straight way to uniquely identifying a a NIC counter, we are formulating  a unique UUIs by <Node Name><":kernel:"<port name>;
	 */
	@Override
	protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
            final Map<String, Object> counterRow){
		String portName = (String)counterRow.get("instance_name");
		if(nodeName == null || nodeName.isEmpty() || portName == null || portName.isEmpty()){
			logger.warn("Failed to find Nic Common countr's instance_name for node: "+nodeName +" (or) node name: "+nodeName );
            return null;
		}
		return nodeName + ":kernel:" + portName;
		
	}

    @Override
    protected String getInventoryUUIDFromPerformanceUUID(final NetAppDefaultStorageContext context, final String performanceInstanceUUID, final String nodeName,
            final Map<String, Object> counterRow) {
    	return context.getIpPortPerfUUIDToUUIDMap().get(performanceInstanceUUID);
    }
    
    
    /**
     * 
     * @param float opmNicTotalBytePerSec
     * @param float linkSpeed
     * @return float nicUtilixation (opm_total_byte_per_sec*8(converting bytes to bits) / link_speed) * 100
     */
    private float calculateOpmNicUtilization(float opmNicTotalBytePerSec, Speed linkSpeed){
    		
    	if(null == linkSpeed)
    		return MetricsRowSample.NULL_FLOAT;
    	long linkSpeesInBitsPerSec = getLinkSpeesInBitsPerSec(linkSpeed);
    	if(linkSpeesInBitsPerSec > 0){
    		// multiplying by 8 to convert Bytes/sec to Bits/sec and 100 to return utilization in 0 to 100 % range
    		return (float)(100 * ((8*(double)opmNicTotalBytePerSec)/linkSpeesInBitsPerSec));
    	}else{
    		// if we have 0 or negative values(error handling), then utilization cannot be calibrated, hence setting to null representation for float
    		return MetricsRowSample.NULL_FLOAT;
    	}
    }
    
    /**
     *  Converting Enum NetworkPort.Speed to BigIntiger equivalent
     * @param linkSpeed
     * @return BigInteger Link speed in  Bits/Sec
     */
    private long getLinkSpeesInBitsPerSec(Speed linkSpeed){
    	long linkSpeesInBitsPerSec =0;
    	if(linkSpeed.compareTo(Speed.MEGABIT_10) == 0) linkSpeesInBitsPerSec = 10 * MEGABIT_1; // 10 Mega Bits per second
    	else if(linkSpeed.compareTo(Speed.MEGABIT_100) == 0) linkSpeesInBitsPerSec = 100 * MEGABIT_1; // 100 Mega Bit per second
    	else if(linkSpeed.compareTo(Speed.GIGABIT_1) == 0) linkSpeesInBitsPerSec = 1000 * MEGABIT_1; // 1 Giga Bits per second
    	else if(linkSpeed.compareTo(Speed.GIGABIT_10) == 0) linkSpeesInBitsPerSec = 10000 * MEGABIT_1; // 10 Giga Bits per second
    	return linkSpeesInBitsPerSec;
    }
    
}