package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;


import com.netapp.oci.netapp.performance.node.interfaces.data.NodeMeasurements;
import com.netapp.oci.netapp.performance.node.interfaces.data.NodeRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders.DiskMetricsBuilder.DiskCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders.HostadapterMetricsBuilder.HostadapterCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders.ProcessorMetricsBuilder.ProcessorCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders.WaflMetricsBuilder.WaflCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.GlobalScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

public class NodeMetricsBuilder extends ArchiveMetricsBuilderBase<Cluster.ClusterKey,
NodeMeasurements, NodeRawMeasurements, NodeMetricsBuilder.NodeCounter> {

    private static final Logger logger = LoggerFactory.getLogger(NodeMetricsBuilder.class);

	public static enum NodeCounter implements ArchiveCounterMetaData {
		avg_processor_busy,
        cpu_busy,
        cpu_elapsed_time,
        net_data_recv,
        net_data_sent,
        cifs_ops,
        fcp_ops,
        iscsi_ops,
        nfs_ops,
        other_ops,
        read_ops,
        total_latency,
        read_latency,
        write_latency,
        total_ops,
        write_ops,
        fcp_data_recv,
        fcp_data_sent;
		
		private final String onTapCounterName;
	    private final int onTapCounterIndex;
	    
	    private NodeCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private NodeCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }
		@Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}
		
	}
	
	public NodeMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
	        super(reportingPackage, "system", NodeCounter.class);
	}

	@Override
	protected ClusterKey getParentInventoryUUIDKey(
			NetAppDefaultStorageContext context, String perfInstanceUUID,
			String nodeName, Map<String, Object> counterRow) {
		return context.getKey();
	}

	@Override
	public void reportPerformance(Cluster.ClusterKey clusterKey, 
			Map<ClusterKey, Collection<RawSample<NodeMeasurements>>> measurements)
			throws DataSourceErrorException {
		if (measurements == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<NodeMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<NodeMeasurements>> nodeMeasurements = measurementEntry.getValue();
            reportingPackage.reportNodePerformance(key, nodeMeasurements);
        }	
	}

	@Override
	protected NodeCounter[] getCountersMetaData(
			NetAppDefaultStorageContext context) {
		return NodeCounter.values();
	}
	
	@Override
	protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
            final Map<String, Object> counterRow) {
		Map<String, String> nodeNameToUUIDMap = context.getNodeNameToUUIDMap();
		return nodeNameToUUIDMap.get(nodeName);
	}

	@Override
	protected NodeMeasurements normalizeMetrics(
			ArchiveMetricsScratchPad<NodeCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		final NodeMeasurements nodeMeasurements = new NodeMeasurements();
		nodeMeasurements.setAvgProcessorBusy(currentScratchPad.getPercentFromDeltas(NodeCounter.avg_processor_busy, NodeCounter.cpu_elapsed_time));
		nodeMeasurements.setCpuBusy(currentScratchPad.getPercentFromDeltas(NodeCounter.cpu_busy, NodeCounter.cpu_elapsed_time));
		nodeMeasurements.setCpuElapsedTime(currentScratchPad.getCounterDeltaFinal(NodeCounter.cpu_elapsed_time));
		nodeMeasurements.setNetDataRecv(currentScratchPad.getCounterRatePerSec(NodeCounter.net_data_recv));
		nodeMeasurements.setNetDataSent(currentScratchPad.getCounterRatePerSec(NodeCounter.net_data_sent));
		nodeMeasurements.setCifsOps(currentScratchPad.getCounterRatePerSec(NodeCounter.cifs_ops));
		nodeMeasurements.setFcpOps(currentScratchPad.getCounterRatePerSec(NodeCounter.fcp_ops));
		nodeMeasurements.setIscsiOps(currentScratchPad.getCounterRatePerSec(NodeCounter.iscsi_ops));
		nodeMeasurements.setNfsOps(currentScratchPad.getCounterRatePerSec(NodeCounter.nfs_ops));
		nodeMeasurements.setOtherOps(currentScratchPad.getCounterRatePerSec(NodeCounter.other_ops));
		nodeMeasurements.setReadOps(currentScratchPad.getCounterRatePerSec(NodeCounter.read_ops));
		nodeMeasurements.setWriteOps(currentScratchPad.getCounterRatePerSec(NodeCounter.write_ops));
		nodeMeasurements.setTotalOps(currentScratchPad.getCounterRatePerSec(NodeCounter.total_ops));
		nodeMeasurements.setSysAvgLatency(currentScratchPad.getLatencyPerOp(NodeCounter.total_latency, NodeCounter.total_ops));
		nodeMeasurements.setSysReadLatency(currentScratchPad.getLatencyPerOp(NodeCounter.read_latency, NodeCounter.read_ops));
		nodeMeasurements.setSysWriteLatency(currentScratchPad.getLatencyPerOp(NodeCounter.write_latency, NodeCounter.write_ops));
		nodeMeasurements.setFcpDataRecv(currentScratchPad.getCounterRatePerSec(NodeCounter.fcp_data_recv));
		nodeMeasurements.setFcpDataSent(currentScratchPad.getCounterRatePerSec(NodeCounter.fcp_data_sent));
		nodeMeasurements.setOpmSysReadThroughput(MetricsRowScratchPad.add(nodeMeasurements.getNetDataRecv(), nodeMeasurements.getFcpDataRecv()));
		nodeMeasurements.setOpmSysWriteThroughput(MetricsRowScratchPad.add(nodeMeasurements.getNetDataSent(), nodeMeasurements.getFcpDataSent()));
		nodeMeasurements.setOpmSysThroughput(MetricsRowScratchPad.add(nodeMeasurements.getOpmSysReadThroughput(), nodeMeasurements.getOpmSysWriteThroughput()));
		nodeMeasurements.setOpmSysUtilization(getNodeUtilization(context, currentScratchPad, nodeMeasurements));
		nodeMeasurements.setMaxAggregateUtilization(getMaxAggregateUtilization(context, currentScratchPad));
		nodeMeasurements.setPercentCleanTimeUtil(getPercentCleanTimeUtil(nodeMeasurements));
		nodeMeasurements.setDiskShelfFibreChannelLoopUtilization(getDiskShelfFibreChannelLoopUtilization(context, currentScratchPad));
		
		return nodeMeasurements;
	}

	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<ClusterKey, Collection<RawSample<NodeRawMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<NodeRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) { 
            final Cluster.ClusterKey key = measurementEntry.getKey(); 
            final Collection<RawSample<NodeRawMeasurements>> nodeMeasurements = measurementEntry.getValue(); 
            reportingPackage.reportNodeRawPerformance(key, nodeMeasurements); 
        } 
	}

	@Override
	protected NodeRawMeasurements rawMetrics(
			ArchiveMetricsScratchPad<NodeCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		Map<NodeCounter,BigInteger> counterValues = currentScratchPad.getCounterValues();
	 	final NodeRawMeasurements m = new NodeRawMeasurements();
    	m.setAvgProcessorBusy(counterValues.get(NodeCounter.avg_processor_busy));
    	m.setCpuBusy(counterValues.get(NodeCounter.cpu_busy));
    	m.setCpuElapsedTime(counterValues.get(NodeCounter.cpu_elapsed_time));
    	m.setNetDataRecv(counterValues.get(NodeCounter.net_data_recv));
    	m.setNetDataSent(counterValues.get(NodeCounter.net_data_sent));
    	m.setCifsOps(counterValues.get(NodeCounter.cifs_ops));
    	m.setFcpOps(counterValues.get(NodeCounter.fcp_ops));
    	m.setIscsiOps(counterValues.get(NodeCounter.iscsi_ops));
    	m.setNfsOps(counterValues.get(NodeCounter.nfs_ops));
    	m.setOtherOps(counterValues.get(NodeCounter.other_ops));
    	m.setReadOps(counterValues.get(NodeCounter.read_ops));
    	m.setSysAvgLatency(counterValues.get(NodeCounter.total_latency));
    	m.setSysReadLatency(counterValues.get(NodeCounter.read_latency));
    	m.setSysWriteLatency(counterValues.get(NodeCounter.write_latency));
    	m.setTotalOps(counterValues.get(NodeCounter.total_ops));
    	m.setWriteOps(counterValues.get(NodeCounter.write_ops));
    	m.setFcpDataRecv(counterValues.get(NodeCounter.fcp_data_recv));
    	m.setFcpDataSent(counterValues.get(NodeCounter.fcp_data_sent));
    	m.setOpmSysReadThroughput(MetricsRowScratchPad.add(m.getNetDataRecv(), m.getFcpDataRecv()));
    	m.setOpmSysWriteThroughput(MetricsRowScratchPad.add(m.getNetDataSent(), m.getFcpDataSent()));
    	m.setOpmSysThroughput(MetricsRowScratchPad.add(m.getOpmSysReadThroughput(), m.getOpmSysWriteThroughput()));
    	
    	return m;
	}
	
	/**
	 * Method to calculate node utilization using processor and wafl counters
	 * @param context - NetAppDefaultStorageContext 
	 * @param currentScratchPad - MetricsRowScratchPad with current values
	 * @return
	 */
	
	protected float getNodeUtilization(NetAppDefaultStorageContext context, MetricsRowScratchPad<NodeCounter> currentScratchPad, NodeMeasurements nodeMeasurements){
		final String inventoryUUID = currentScratchPad.getInventoryUUID();
		Map<String, String> nodeNameToUUIDMap = context.getNodeNameToUUIDMap();
		GlobalScratchPad globalScratchPad = reportingPackage.getGlobalScratchPad();
		Map<String, MetricsRowScratchPad<ProcessorCounter>> processorScratchPadsMap =  (Map<String, MetricsRowScratchPad<ProcessorCounter>>) globalScratchPad.get("processor");
		Map<String, MetricsRowScratchPad<WaflCounter>> waflScratchPadsMap = (Map<String, MetricsRowScratchPad<WaflCounter>>) globalScratchPad.get("wafl");
		long totalCpMSecs = 0;
		long P2Flush = 0;
		long totalKahunaBusy = 0L;
		long maxCpuElapsedTime = 0L;
		float nodeUtilization=0;
		final float KAHUNA_THRESHOLD_LIMIT = 60F;
		final float MAX_THRESHOLD_LIMIT = 100F;
		final long deltaPollTimeInMs = currentScratchPad.getCollectionPeriod(); 
		
		nodeMeasurements.setDeltaPollTime(deltaPollTimeInMs);
		
		float processorBusyPercent =  currentScratchPad.getPercentFromDeltas(NodeCounter.avg_processor_busy, NodeCounter.cpu_elapsed_time);
		
		for(String key : processorScratchPadsMap.keySet()){
			final String[] tokens = key.split(":");
			String nodeUUID = nodeNameToUUIDMap.get(tokens[0]);
			if(!nodeUUID.equals(inventoryUUID)){
				continue;
			}
			MetricsRowScratchPad<ProcessorCounter> processorScratchPad = processorScratchPadsMap.get(key);
			long kahunaBusy = processorScratchPad.getCounterDelta(ProcessorCounter.kahuna);
			long processorElapsedTime = processorScratchPad.getCounterDelta(ProcessorCounter.processor_elapsed_time);
			if(kahunaBusy != MetricsRowSample.NULL_FLOAT){
				totalKahunaBusy += kahunaBusy;
			}
			if((processorElapsedTime != MetricsRowSample.NULL_LONG) && (processorElapsedTime > maxCpuElapsedTime)){
				maxCpuElapsedTime = processorElapsedTime;
			}
			
		}
		
		nodeMeasurements.setTotalKahunaBusy(totalKahunaBusy);
		nodeMeasurements.setMaxProcessorElapsedTime(maxCpuElapsedTime);
		
		final float totalKahunaBusyPercent = (maxCpuElapsedTime == 0 || totalKahunaBusy == 0 || 
				maxCpuElapsedTime == MetricsRowSample.NULL_LONG || totalKahunaBusy == MetricsRowSample.NULL_LONG)
                ? 0
                : (float) (100d * (double) totalKahunaBusy / (double) maxCpuElapsedTime);
		
		for(String key : waflScratchPadsMap.keySet()){
			final String[] tokens = key.split(":");
			String nodeUUID = nodeNameToUUIDMap.get(tokens[0]);
			if(!nodeUUID.equals(inventoryUUID)){
				continue;
			}
			MetricsRowScratchPad<WaflCounter> waflScratchPad = waflScratchPadsMap.get(key);
			totalCpMSecs = waflScratchPad.getCounterDelta(WaflCounter.total_cp_msecs);
			P2Flush = waflScratchPad.getCounterDelta(WaflCounter.P2_FLUSH);
			break;
		}
		
		nodeMeasurements.setTotalCpMSecs(totalCpMSecs);
		nodeMeasurements.setP2Flush(P2Flush);
		
		final float waflUtilization = (totalCpMSecs == MetricsRowSample.NULL_LONG || totalCpMSecs == 0 || 
				deltaPollTimeInMs == MetricsRowSample.NULL_LONG || deltaPollTimeInMs == 0 || P2Flush == 0)
				? 0
				: (float) (((double)totalCpMSecs * (1 - ((double)P2Flush/(double)totalCpMSecs)/.8F)/deltaPollTimeInMs) * 100d);

        // Formula
        // processor_busy_percent=avg_processor_busy(delta)/cpu_elapsed_time1(delta))
        // total_cp_time=total_cp_msecs(delta)
        // p2_flush_time=cp_phase_times.P2FLUSH(delta)
        // kahuna=domain_busy.kahuna(delta)/processor_elapsed_time(delta)
        // Node_Utilization =
        //       100% when kahuna > 60% OR
        //       Max of {
        //           processor_busy_percent * 100
        //          (total_cp_time(1 - (p2_flush_time/total_cp_time)/.8) /poll time msecs) * 100
        if(totalKahunaBusyPercent > KAHUNA_THRESHOLD_LIMIT){
            nodeUtilization =  MAX_THRESHOLD_LIMIT;
        } else if (processorBusyPercent > MAX_THRESHOLD_LIMIT || waflUtilization > MAX_THRESHOLD_LIMIT) {
            // This case should not happen. Logging it as warning as it could only happen if ontap reports large values for walf cp_time.
            logger.warn(PackageContext.getLogPrefix()+" - processorBusyPercent - {} or waflUtilization - {} cannot be greater than 100%." +
                    "for node - "+currentScratchPad.getInventoryUUID()+" ", processorBusyPercent, waflUtilization);
            nodeUtilization =  MAX_THRESHOLD_LIMIT;
        } else {
            nodeUtilization = Math.max(processorBusyPercent, waflUtilization );
        }
        return nodeUtilization;
	}
	
	protected float getMaxAggregateUtilization(NetAppDefaultStorageContext context, MetricsRowScratchPad<NodeCounter> currentScratchPad){
		String inventoryUUID = currentScratchPad.getInventoryUUID();
		GlobalScratchPad globalScratchPad = reportingPackage.getGlobalScratchPad();
		Map<String, List<String>> nodeUUIDToDiskUUIDs = context.getNodeUUIDToDiskUUIDs();
		List<String> diskPerformanceUUIDs = nodeUUIDToDiskUUIDs.get(inventoryUUID);
		if(diskPerformanceUUIDs == null){
			return MetricsRowSample.NULL_LONG;
		}
		
		List<DiskCounter> numerator = new ArrayList<DiskCounter>();
		numerator.add(DiskCounter.disk_busy);
		List<DiskCounter> denominator = new ArrayList<DiskCounter>();
		denominator.add(DiskCounter.base_for_disk_busy);
		Map<DiskCounter, Float> maxPercentDelta = globalScratchPad.getMaximumOfPercentDeltas(diskPerformanceUUIDs, "disk", numerator, denominator);
		return (maxPercentDelta == null) ? MetricsRowSample.NULL_FLOAT : maxPercentDelta.get(DiskCounter.disk_busy);
	}
	
	protected float getPercentCleanTimeUtil(NodeMeasurements m){
		long totalCpMSecs = m.getTotalCpMSecs();
		long P2Flush = m.getP2Flush();
		final float DEFAULT_MAXAGGREGATE_UTILIZATION = 50F;
		float percentCleanTimeUtil = MetricsRowSample.NULL_FLOAT;
		
		if(m.getDeltaPollTime() == MetricsRowSample.NULL_LONG || m.getDeltaPollTime() == 0 || totalCpMSecs == MetricsRowSample.NULL_LONG || totalCpMSecs == 0 
				|| P2Flush == MetricsRowSample.NULL_LONG){
			return MetricsRowSample.NULL_FLOAT;
		}

		if(m.getMaxAggregateUtilization() < DEFAULT_MAXAGGREGATE_UTILIZATION){
			percentCleanTimeUtil = (float)(100d * (((double) totalCpMSecs - (double) P2Flush)/.8F)/(double)m.getDeltaPollTime());
		}else {
			percentCleanTimeUtil = (float)(100d * (double)totalCpMSecs / (double)m.getDeltaPollTime());
		}
		return percentCleanTimeUtil;
	}
	
	protected float getDiskShelfFibreChannelLoopUtilization(NetAppDefaultStorageContext context, MetricsRowScratchPad<NodeCounter> currentScratchPad){
		String inventoryUUID = currentScratchPad.getInventoryUUID();
		Map<String, String> nodeNameToUUIDMap = context.getNodeNameToUUIDMap();
		GlobalScratchPad globalScratchPad = reportingPackage.getGlobalScratchPad();
		Map<String, MetricsRowScratchPad<HostadapterCounter>> hostAdapterScratchPadsMap = (Map<String, MetricsRowScratchPad<HostadapterCounter>>) globalScratchPad.get("hostadapter");
		float diskShelfFibreChannelLoopUtilization = MetricsRowSample.NULL_FLOAT;
		
		long deltaPollTimeInMs = currentScratchPad.getCollectionPeriod();
		
		for(String key : hostAdapterScratchPadsMap.keySet()){
			final String[] tokens = key.split(":");
			String nodeUUID = nodeNameToUUIDMap.get(tokens[0]);
			if(!nodeUUID.equals(inventoryUUID)){
				continue;
			}
			MetricsRowScratchPad<HostadapterCounter> hostAdapterScratchPad = hostAdapterScratchPadsMap.get(key);
			float perHostAdapterUtilization = MetricsRowSample.NULL_FLOAT;
			float totalBytes = MetricsRowScratchPad.add(hostAdapterScratchPad.getCounterDelta(HostadapterCounter.bytes_written), 
					hostAdapterScratchPad.getCounterDelta(HostadapterCounter.bytes_read));
			long hostAdapterMaxLinkDataRate = ((hostAdapterScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate) != MetricsRowSample.NULL_BIG_INT) && (hostAdapterScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate) != null)) 
												? hostAdapterScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate).longValue() : MetricsRowSample.NULL_LONG;
			
			if(hostAdapterMaxLinkDataRate != MetricsRowSample.NULL_LONG || hostAdapterMaxLinkDataRate != 0l){
					perHostAdapterUtilization = (float) (100d * ((double)totalBytes/(double)deltaPollTimeInMs/(double)hostAdapterMaxLinkDataRate));
			}
			
			if(diskShelfFibreChannelLoopUtilization < perHostAdapterUtilization){
				diskShelfFibreChannelLoopUtilization = perHostAdapterUtilization;
			}
		}
		return diskShelfFibreChannelLoopUtilization;
	}
		
}
