package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.netapp.oci.netapp.performance.volume.interfaces.data.VolumeMeasurements;
import com.netapp.oci.netapp.performance.volume.interfaces.data.VolumeRawMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveCounterMetaData;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer.VServerKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

public class VolumeMetricsBuilder extends ArchiveMetricsBuilderBase<VServer.VServerKey,
	VolumeMeasurements, VolumeRawMeasurements, VolumeMetricsBuilder.VolumeCounter> {
	
	 private static final Logger logger = LoggerFactory.getLogger(VolumeMetricsBuilder.class);
	public static enum VolumeCounter implements ArchiveCounterMetaData {
		
		avg_latency,
		total_ops,
		read_latency,
		read_ops,
		write_latency,
		write_ops,
		other_latency,
		other_ops,
		nfs_read_latency,
		nfs_read_ops,
		nfs_write_latency,
		nfs_write_ops,
		nfs_other_latency,
		nfs_other_ops,
		cifs_read_latency,
		cifs_read_ops,
		cifs_write_latency,
		cifs_write_ops,
		cifs_other_latency,
		cifs_other_ops,
		fcp_read_latency,
		fcp_read_ops,
		fcp_write_latency,
		fcp_write_ops,
		fcp_other_latency,
		fcp_other_ops,
		iscsi_read_latency,
		iscsi_read_ops,
		iscsi_write_latency,
		iscsi_write_ops,
		iscsi_other_latency,
		iscsi_other_ops,
		san_other_latency,
		san_other_ops,
		san_read_latency,
		san_read_ops,
		san_write_latency,
		san_write_ops,
		total_protocol_write_latency,
		total_protocol_other_latency,
		total_protocol_read_latency,
		vserver_uuid;
		
		private final String onTapCounterName;
	    private final int onTapCounterIndex;
	    
	    private VolumeCounter(final String onTapCounterName, final int onTapCounterIndex) {
            this.onTapCounterName = onTapCounterName;
            this.onTapCounterIndex = onTapCounterIndex;
        }

        private VolumeCounter() {
            this.onTapCounterName = this.name();
            this.onTapCounterIndex = -1;
        }
		
	    @Override
		public String getOnTapCounterName() {
			return onTapCounterName;
		}

		@Override
		public int getOnTapCounterIndex() {
			return onTapCounterIndex;
		}
		
	}
	
	public VolumeMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "volume", VolumeCounter.class);
    }

	@Override
	protected VServer.VServerKey getParentInventoryUUIDKey(
			NetAppDefaultStorageContext context, String perfInstanceUUID, String nodeName, Map<String, Object> counterRow) {
		String clusterUUID = context.getKey().getID(Cluster.ClusterKey.Field.cluster_uuid);
		String vServerUUID = (String)counterRow.get("vserver_uuid");
		if (vServerUUID == null || vServerUUID.length() == 0) {
            logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for network lif " + perfInstanceUUID);
            return null;
     }
		return new VServer.VServerKey(clusterUUID, vServerUUID);
	}

	@Override
	public void reportPerformance(Cluster.ClusterKey clusterKey, Map<VServer.VServerKey, Collection<RawSample<VolumeMeasurements>>> measurements)
			throws DataSourceErrorException {
		 if (measurements == null) {
	            return;
	        }
	        for (final Map.Entry<VServer.VServerKey, Collection<RawSample<VolumeMeasurements>>> measurementEntry : measurements.entrySet()) {
	            final VServer.VServerKey vServerkey = measurementEntry.getKey();
	            final Collection<RawSample<VolumeMeasurements>> volumeMeasurements = measurementEntry.getValue();
	            reportingPackage.reportVolumePerformance(clusterKey, vServerkey, volumeMeasurements);
	        }
		 
	}
	
	@Override
	protected String getPerformanceUUID(final NetAppDefaultStorageContext context, final String nodeName,
            final Map<String, Object> counterRow){
		Map<String, String> vServerUUIDVolumeNameToVolumeUUIDMap = context.getvServerUUIDVolumeNameToVolumeUUIDMap();
		String vServerUUID = (String)counterRow.get("vserver_uuid");
		String volumeName = (String)counterRow.get("instance_name");
		if(vServerUUID == null){
			logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer for volume " + volumeName);
            return null;
		}
		return vServerUUIDVolumeNameToVolumeUUIDMap.get(vServerUUID+volumeName)!= null ? vServerUUIDVolumeNameToVolumeUUIDMap.get(vServerUUID+volumeName) : null;
		
	}

	@Override
	protected VolumeCounter[] getCountersMetaData(
			NetAppDefaultStorageContext context) {
		return VolumeCounter.values();
	}

	@Override
	protected VolumeMeasurements normalizeMetrics(
			ArchiveMetricsScratchPad<VolumeCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		final VolumeMeasurements m = new VolumeMeasurements();
					m.setAvgLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.avg_latency, VolumeCounter.total_ops));
					m.setCifsOtherLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.cifs_other_latency, VolumeCounter.cifs_other_ops));
					m.setCifsOtherOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.cifs_other_ops));
					m.setCifsReadLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.cifs_read_latency, VolumeCounter.cifs_read_ops));
					m.setCifsReadOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.cifs_read_ops));
					m.setCifsWriteLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.cifs_write_latency, VolumeCounter.cifs_write_ops));
					m.setCifsWriteOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.cifs_write_ops));
					m.setFcpOtherLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.fcp_other_latency, VolumeCounter.fcp_other_ops));
					m.setFcpOtherOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.fcp_other_ops));
					m.setFcpReadLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.fcp_read_latency, VolumeCounter.fcp_read_ops));
					m.setFcpReadOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.fcp_read_ops));
					m.setFcpWriteLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.fcp_write_latency, VolumeCounter.fcp_write_ops));
					m.setFcpWriteOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.fcp_write_ops));
					m.setIscsiOtherLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.iscsi_other_latency, VolumeCounter.iscsi_other_ops));
					m.setIscsiOtherOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.iscsi_other_ops));
					m.setIscsiReadLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.iscsi_read_latency, VolumeCounter.iscsi_read_ops));
					m.setIscsiReadOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.iscsi_read_ops));
					m.setIscsiWriteLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.iscsi_write_latency, VolumeCounter.iscsi_write_ops));
					m.setIscsiWriteOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.iscsi_write_ops));
					m.setNfsOtherLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.nfs_other_latency, VolumeCounter.nfs_other_ops));
					m.setNfsOtherOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.nfs_other_ops));
					m.setNfsReadLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.nfs_read_latency, VolumeCounter.nfs_read_ops));
					m.setNfsReadOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.nfs_read_ops));
					m.setNfsWriteLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.nfs_write_latency, VolumeCounter.nfs_write_ops));
					m.setNfsWriteOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.nfs_write_ops));
					m.setOtherLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.other_latency, VolumeCounter.other_ops));
					m.setOtherOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.other_ops));
					m.setReadLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.read_latency, VolumeCounter.read_ops));
					m.setReadOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.read_ops));
					m.setSanOtherLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.san_other_latency, VolumeCounter.san_other_ops));
					m.setSanOtherOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.san_other_ops));
					m.setSanReadLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.san_read_latency, VolumeCounter.san_read_ops));
					m.setSanReadOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.san_read_ops));
					m.setSanWriteLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.san_write_latency, VolumeCounter.san_write_ops));
					m.setSanWriteOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.san_write_ops));
					m.setTotalOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.total_ops));
					m.setWriteLatency(currentScratchPad.getLatencyPerOp(VolumeCounter.write_latency, VolumeCounter.write_ops));
					m.setWriteOps(currentScratchPad.getCounterRatePerSec(VolumeCounter.write_ops));
					m.setOpmVolNfsOps(MetricsRowScratchPad.add(m.getNfsReadOps(), m.getNfsWriteOps(), m.getNfsOtherOps()));
					m.setOpmVolCifsOps(MetricsRowScratchPad.add(m.getCifsReadOps(), m.getCifsWriteOps(), m.getCifsOtherOps()));
					m.setOpmVolIscsiOps(MetricsRowScratchPad.add(m.getIscsiReadOps(), m.getIscsiWriteOps(), m.getIscsiOtherOps()));
					m.setOpmVolFcpOps(MetricsRowScratchPad.add(m.getFcpReadOps(),m.getFcpWriteOps(), m.getFcpOtherOps()));
		
		return m;
	}


	@Override
	public void reportRawPerformance(
			ClusterKey clusterKey,
			Map<VServerKey, Collection<RawSample<VolumeRawMeasurements>>> rawMeasurements)
			throws DataSourceErrorException {
		for (final Map.Entry<VServer.VServerKey, Collection<RawSample<VolumeRawMeasurements>>> measurementEntry : rawMeasurements.entrySet()) { 
            final VServer.VServerKey key = measurementEntry.getKey(); 
            final Collection<RawSample<VolumeRawMeasurements>> volumeRawMeasurements = measurementEntry.getValue(); 
            reportingPackage.reportVolumeRawPerformance(clusterKey, key, volumeRawMeasurements); 
        } 		
	}

	@Override
	protected VolumeRawMeasurements rawMetrics(
			ArchiveMetricsScratchPad<VolumeCounter> currentScratchPad,
			NetAppDefaultStorageContext context) {
		 final VolumeRawMeasurements m = new VolumeRawMeasurements();
	        Map<VolumeCounter,BigInteger> counterValues = currentScratchPad.getCounterValues();
	        m.setAvgLatency(counterValues.get(VolumeCounter.avg_latency));
	        m.setCifsOtherLatency(counterValues.get(VolumeCounter.cifs_other_latency));
	        m.setCifsOtherOps(counterValues.get(VolumeCounter.cifs_other_ops));
	        m.setCifsReadLatency(counterValues.get(VolumeCounter.cifs_read_latency));
	        m.setCifsReadOps(counterValues.get(VolumeCounter.cifs_read_ops));
	        m.setCifsWriteLatency(counterValues.get(VolumeCounter.cifs_write_latency));
	        m.setCifsWriteOps(counterValues.get(VolumeCounter.cifs_write_ops));
	        m.setFcpOtherLatency(counterValues.get(VolumeCounter.fcp_other_latency));
	        m.setFcpOtherOps(counterValues.get(VolumeCounter.fcp_other_ops));
	        m.setFcpReadLatency(counterValues.get(VolumeCounter.fcp_read_latency));
	        m.setFcpReadOps(counterValues.get(VolumeCounter.fcp_read_ops));
	        m.setFcpWriteLatency(counterValues.get(VolumeCounter.fcp_write_latency));
	        m.setFcpWriteOps(counterValues.get(VolumeCounter.fcp_write_ops));
	        m.setIscsiOtherLatency(counterValues.get(VolumeCounter.iscsi_other_latency));
	        m.setIscsiOtherOps(counterValues.get(VolumeCounter.iscsi_other_ops));
	        m.setIscsiReadLatency(counterValues.get(VolumeCounter.iscsi_read_latency));
	        m.setIscsiReadOps(counterValues.get(VolumeCounter.iscsi_read_ops));
	        m.setIscsiWriteLatency(counterValues.get(VolumeCounter.iscsi_write_latency));
	        m.setIscsiWriteOps(counterValues.get(VolumeCounter.iscsi_write_ops));
	        m.setNfsOtherLatency(counterValues.get(VolumeCounter.nfs_other_latency));
	        m.setNfsOtherOps(counterValues.get(VolumeCounter.nfs_other_ops));
	        m.setNfsReadLatency(counterValues.get(VolumeCounter.nfs_read_latency));
	        m.setNfsReadOps(counterValues.get(VolumeCounter.nfs_read_ops));
	        m.setNfsWriteLatency(counterValues.get(VolumeCounter.nfs_write_latency));
	        m.setNfsWriteOps(counterValues.get(VolumeCounter.nfs_write_ops));
	        m.setOtherLatency(counterValues.get(VolumeCounter.other_latency));
	        m.setOtherOps(counterValues.get(VolumeCounter.other_ops));
	        m.setReadLatency(counterValues.get(VolumeCounter.read_latency));
	        m.setReadOps(counterValues.get(VolumeCounter.read_ops));
	        m.setWriteLatency(counterValues.get(VolumeCounter.write_latency));
	        m.setWriteOps(counterValues.get(VolumeCounter.write_ops));
	        m.setTotalOps(counterValues.get(VolumeCounter.total_ops));
	        
	        Map<VolumeCounter, String> rawValue = currentScratchPad.getRawValues();
	        m.setTotalProtocolOtherLatency(rawValue.get(VolumeCounter.total_protocol_other_latency));
	        m.setTotalProtocolReadLatency(rawValue.get(VolumeCounter.total_protocol_read_latency));
	        m.setTotalProtocolWriteLatency(rawValue.get(VolumeCounter.total_protocol_write_latency));
			m.setOpmVolNfsOps(MetricsRowScratchPad.add(m.getNfsReadOps(), m.getNfsWriteOps(), m.getNfsOtherOps()));
			m.setOpmVolCifsOps(MetricsRowScratchPad.add(m.getCifsReadOps(), m.getCifsWriteOps(), m.getCifsOtherOps()));
			m.setOpmVolIscsiOps(MetricsRowScratchPad.add(m.getIscsiReadOps(), m.getIscsiWriteOps(), m.getIscsiOtherOps()));
			m.setOpmVolFcpOps(MetricsRowScratchPad.add(m.getFcpReadOps(),m.getFcpWriteOps(), m.getFcpOtherOps()));
	        return m; 
	}

}
