package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.common;


import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.parser.ArchiveFileParser;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.parser.CounterTable;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorType;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.*;
import java.util.concurrent.*;

/**
 * Base class to download the latest performance archive files from the archiver, defined by the implementation. This
 * class also retrieves the samples from the archive, determines which sample to process and passes it to the
 * implementation class for processing.
 *
 * @param <C> Enum of the counter manager objects available in the archive.
 */
public abstract class BasePerformanceArchive<C extends Enum<C> & CounterObjectsInterface> {

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BasePerformanceArchive.class);

    protected final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage;

    protected final String archiveFileDownloadDir;

    protected List<ArchiveFileMetaData<C>> archiveFileMetaDataList;

    protected List<ArchiveDataSourceInfo> previousArchiveDataSourceInfoList;

    protected List<ArchiveDataSourceInfo> archiveDataSourceInfoList;
    
    private FileDownloader archiveFileDownloader;

    protected BasePerformanceArchive(final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage,
            final String archiveFileDownloadDir) {
        this.netAppOCIEPerformancePackage = netAppOCIEPerformancePackage;
        this.archiveFileDownloadDir = archiveFileDownloadDir;
        if (!netAppOCIEPerformancePackage.isSimulating()) {
            this.archiveFileDownloader = new ArchiveFileDownloader();
        }
        else {
            this.archiveFileDownloader = new ReplayFileDownloader();
        }
    }

    /**
     * OnTap data store name of the archiver.
     */
    public abstract String getDataStoreName();

    /**
     * OnTap process name of the archiver.
     * @return
     */
    public abstract String getProcessName();

    /**
     * List of objects available in the archiver.
     * @return
     */
    public abstract List<C> getCounterObjects();

    /**
     * return sample time from the archive.
     *
     * @param archiveTimestamps Archive time stamp in milliseconds
     * @param fileTimeStamp     File time stamp in milliseconds
     * @param counterObject     counter object
     * @return The archive time stamp.
     */
    public abstract Long determineArchivePoolTime(final List<Long> archiveTimestamps, final Long fileTimeStamp, final C counterObject);

    /**
     * Process and report the counters retrieved from the performance archive.
     * @param context {@link NetAppDefaultStorageContext} with data from the foundation poll
     * @param objectNameToCounterTableList Counter Object name to list of {@link CounterTable} map. The list contains
     *                                     a {@link CounterTable} object, one for each node in the cluster.
     * @throws DataSourceErrorException
     */
    public abstract void processAndReportPerformance(NetAppDefaultStorageContext context, Map<C, List<CounterTable>> objectNameToCounterTableList) throws DataSourceErrorException;
    
    /**
     * Retrieves the list of latest archive files, downloads the files, collects the list of sample times in the files
     * and returns the newest sample time that will be used to report the performance data.
     *
     * @param context Storage context with data from the foundation poll.
     * @param zapiConnection ZAPI Connection
     * @return Sample time from the performance archive.
     * @throws DataSourceErrorException
     */
    public long getArchiveSampleTime(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        //Retrieve the list of latest performance files.
        final ArchiveDataSourceInfoBuilder archiveDataSourceInfoBuilder = new ArchiveDataSourceInfoBuilder(this.getDataStoreName(), this.getProcessName());
        List<ArchiveDataSourceInfo> archiveDataSourceInfoOriginalList = archiveDataSourceInfoBuilder.buildModel(zapiConnection);
        //List to hold files that were not previously downloaded
        List<ArchiveDataSourceInfo> archiveDataSourceInfoList = new ArrayList<>(archiveDataSourceInfoOriginalList.size());
        if(previousArchiveDataSourceInfoList != null) {
            //Filter out perilously downloaded files
            for(final ArchiveDataSourceInfo archiveDataSourceInfo: archiveDataSourceInfoOriginalList) {
                if(!previousArchiveDataSourceInfoList.contains(archiveDataSourceInfo)) {
                    archiveDataSourceInfoList.add(archiveDataSourceInfo);
                } else {
                    logger.error("Ignoring perilously downloaded file {}" + archiveDataSourceInfo);
                }
            }
        } else {
            archiveDataSourceInfoList = archiveDataSourceInfoOriginalList;
        }
        if(archiveDataSourceInfoList.size() != archiveDataSourceInfoOriginalList.size()) {
            String errorMessage = new StringBuilder("Archive files are not available from all nodes. ")
                    .append(archiveDataSourceInfoList.size())
                    .append(" new archive files are available.").toString();
            logger.error(errorMessage);
            throw new DataSourceErrorException(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE,errorMessage, errorMessage);
        }
        previousArchiveDataSourceInfoList = archiveDataSourceInfoOriginalList;

        //Download the latest files
        try {
        	archiveFileDownloader.downloadArchiveFiles(archiveDataSourceInfoList, archiveFileDownloadDir,
                    netAppOCIEPerformancePackage.getConfiguration());
        } catch (DataSourceErrorException dataSourceError) {
            //If the download fails, cleanup any successfully downloaded files.
        	archiveFileDownloader.cleanDownloadedFiles(archiveFileDownloadDir);
            throw dataSourceError;
        }
        final List<ArchiveFileMetaData<C>> archiveFileMetaDataList = getArchiveFileMetaData(archiveDataSourceInfoList);
        //Find the newest timestamp from all the files and use that as the sample time
        final TreeSet<Long> allTimestamps = new TreeSet<Long>();
        for (final ArchiveFileMetaData<C> archiveFileMetaData : archiveFileMetaDataList) {
            allTimestamps.addAll(archiveFileMetaData.getTimestampsToProcess());
        }
        //Hold on to archive data source info list
        this.archiveDataSourceInfoList = archiveDataSourceInfoList;
        if(allTimestamps.size() > 0) {
            return allTimestamps.last();
        } else {
            return -1;
        }
    }

    public void getAndReportPerformance(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        long startTime = System.currentTimeMillis();
        final Map<C, List<CounterTable>> objectNameToCounterTableList = new HashMap<C, List<CounterTable>>();
        for (final ArchiveFileMetaData<C> archiveFileMetaData : this.archiveFileMetaDataList) {
            final Map<C, Long> objectNameToTimestampToProcess = archiveFileMetaData.getCounterObjectNameToTimestampToProcess();
            for (final Map.Entry<C, Long> objectNameToTimeStamp : objectNameToTimestampToProcess.entrySet()) {
                List<CounterTable> counterTables = objectNameToCounterTableList.get(objectNameToTimeStamp.getKey());
                if(counterTables == null) {
                    counterTables = new LinkedList<CounterTable>();
                    objectNameToCounterTableList.put(objectNameToTimeStamp.getKey(), counterTables);
                }

                // Period is multiplied by 1000 as the cmapl wrapper expects it to be in milli seconds
                CounterTable counterTable = retrieveCounterData(archiveFileMetaData.getArchiveDataSourceInfo(),
                        objectNameToTimeStamp.getKey().name(), objectNameToTimeStamp.getValue(), objectNameToTimeStamp.getKey().getPeriod() * 1000);
                if(counterTable != null) {
                    counterTables.add(counterTable);
                }
            }
        }
        logger.info(new StringBuilder(netAppOCIEPerformancePackage.getConfiguration().getDataSourceName())
                .append(" GetCounters-Total Time for all files ")
                .append(System.currentTimeMillis() - startTime).toString());
        processAndReportPerformance(context, objectNameToCounterTableList);

        //Add to recording and delete files
        if (netAppOCIEPerformancePackage.isMetricsDebugEnabled()) {
            addCCMAFilesToRecording(archiveDataSourceInfoList); //RecordingSample will delete the file after archiving.
        } else {
            archiveFileDownloader.cleanDownloadedFiles(archiveFileDownloadDir);
        }
    }

    /**
     * Read the archive files in parallel and retrieve the timestamps associated with each object type i.e disk,lun etc.
     */
    private List<ArchiveFileMetaData<C>> getArchiveFileMetaData(final List<ArchiveDataSourceInfo> archiveDataSourceInfoList)
            throws DataSourceErrorException {
        long startTime = System.currentTimeMillis();
        final List<ArchiveFileMetaData<C>> archiveFileMetaDataList =
                new ArrayList<ArchiveFileMetaData<C>>(archiveDataSourceInfoList.size());
        final ArchiveFileParser parser = ArchiveFileParser.INSTANCE;
        final List<C> counterObjects = getCounterObjects();
        for (final ArchiveDataSourceInfo archiveDataSourceInfo : archiveDataSourceInfoList) {
            final Map<C, Long> counterObjectToTimestamp = new HashMap<C, Long>(counterObjects.size());
            final String archiveFilePath = getArchiveFilePath(archiveDataSourceInfo);
            for (final C counterObject : counterObjects) {
                final Long archiveTimestamp = determineArchivePoolTime(
                        // getPeriod* 1000 as the cmapl wrapper looks for period in milli seconds
                        parser.getTimestampsForObject(archiveFilePath, counterObject.getOnTapObjectName(), counterObject.getPeriod() * 1000),
                        archiveDataSourceInfo.getTimestamp() * 1000, counterObject);
                if (archiveTimestamp != null) {
                    counterObjectToTimestamp.put(counterObject, archiveTimestamp);
                }
            }
            archiveFileMetaDataList.add(new ArchiveFileMetaData<C>(archiveDataSourceInfo, counterObjectToTimestamp));
        }
        logger.info(new StringBuilder(netAppOCIEPerformancePackage.getConfiguration().getDataSourceName())
                .append(" GetTimeStamp-Total Time for all files ")
                .append(System.currentTimeMillis() - startTime).toString());
        //Hold on to archive file metadata list
        this.archiveFileMetaDataList = archiveFileMetaDataList;

        return archiveFileMetaDataList;
    }

    private String getArchiveFilePath(final ArchiveDataSourceInfo archiveDataSourceInfo) {
        final StringBuilder archiveFilePath = new StringBuilder(archiveFileDownloadDir);
        //Append Node name to path
        archiveFilePath.append(File.separatorChar).append(archiveDataSourceInfo.getNodeName());
        // ZAPI returns 001151_000060_1386354916478_0256944.ccma.gz & 001151_000060_1386354916478_0256944.ccma.meta
        // We need 001151_000060_1386354916478_0256944.ccma
        String archiveFileName = archiveDataSourceInfo.getFileNames().iterator().next();
        archiveFileName = archiveFileName.substring(0, archiveFileName.lastIndexOf('.'));
        archiveFilePath.append(File.separatorChar).append(archiveFileName);
        return archiveFilePath.toString();
    }

    private void addCCMAFilesToRecording(final List<ArchiveDataSourceInfo> opmArchiveDataSourceInfo) {
        if (netAppOCIEPerformancePackage.isMetricsDebugEnabled()) {
            final com.onaro.sanscreen.acquisition.framework.datasource.Package perfPackage = PackageContext.getPackage();
            for (final ArchiveDataSourceInfo archiveDataSourceInfo : opmArchiveDataSourceInfo) {
                for (final String fileName : archiveDataSourceInfo.getFileNames()) {
                    final String filePath = new StringBuilder(archiveFileDownloadDir).append(File.separatorChar)
                            .append(archiveDataSourceInfo.getNodeName()).append(File.separatorChar)
                            .append(fileName).toString();
                    try {
                        perfPackage.getRecordingSession().getCurrentSample().addRecordingFile(filePath);
                    } catch (FileNotFoundException fileNotFound) {
                        logger.error(new StringBuilder("Unable to save CCMA file ")
                                .append(filePath).append(" to recording. FileNotFoundException : ")
                                .append(fileNotFound.getMessage()).toString());
                    }
                }
            }
        }
    }

    private CounterTable retrieveCounterData(final ArchiveDataSourceInfo archiveDataSourceInfo, final String objectName,
                                             final Long archiveTimestamp, final long periodValue)
            throws DataSourceErrorException {
        final ArchiveFileParser parser = ArchiveFileParser.INSTANCE;
        final String archiveFilePath = getArchiveFilePath(archiveDataSourceInfo);
        final CounterTable counterData = parser.getCountersForObjectTimestamp(archiveFilePath, archiveTimestamp, periodValue, objectName);
        if(counterData != null)
            counterData.setNodeName(archiveDataSourceInfo.getNodeName());
        return counterData;
    }

    /**
     * clear previous value scratchPads from all archiver builders
     */
	public abstract void clearAllPreviousValues();

}
