package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.common;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPINames;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIUtils;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import netapp.manage.NaElement;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;


/**
 * abstract base class for iterator based performance builders. Handles iterator tag management and 
 * allows derived classes to parse the repeated XML element. The derived class must implement buildIterInstance()
 * to parse the repeated XML element
 */
public abstract class IterBuilderBase<E> {

    private static final Logger logger = Logger.getLogger(IterBuilderBase.class);

    private final String zapiCommand;
    private final String repeatingInstanceEltName;
    private final int maxRecords;
    private final NaElement desiredAttributes;
    private final NaElement[] additionalElements;

    private static int defaultMaxRecords = 1000;

    public static int getDefaultMaxRecords() {
        return defaultMaxRecords;
    }

    public static void setDefaultMaxRecords(final int defaultMaxRecords) {
        IterBuilderBase.defaultMaxRecords = defaultMaxRecords;
    }

    /**
     * Create a new builder for a specific iterator based ZAPI
     *
     * @param zapiCommand              the iterator based ZAPI to invoke
     * @param repeatingInstanceEltName the name of the top level element that iterates (appears more than once, repeats)
     * @param maxRecords               the requested maximum record chunk size.   Usually defaults to 1000 instances (DEFAULT_MAX_RECORDS)
     * @param desiredAttributesMap     a Map where the Key represents a path to an Element of interest, and the value represents List of desired child elements.
     *                                 The Key is an Array of Strings that represents the path to the XML Element containing desired attributes.
     *                                 The Value is an Array of Strings that represents the child elements of interest.
     * @param args                     optional Ordered Map of Key, Value pairs used as additional parameters to the ZAPI.  It is important that
     *                                 this Map be ordered so that the toString of the request is consistent, and may be used as a deterministic key for lookup
     *                                 during recording playback.
     */
    protected IterBuilderBase(final String zapiCommand, final String repeatingInstanceEltName, final int maxRecords,
                              final LinkedHashMap<String[], String[]> desiredAttributesMap, final LinkedHashMap<String, String> args) {
        this(zapiCommand, repeatingInstanceEltName, maxRecords,
                ZAPIUtils.convertDesiredAttributes(desiredAttributesMap), ZAPIUtils.children(args));
    }

    protected IterBuilderBase(final String zapiCommand, final String repeatingInstanceEltName, final int maxRecords) {
        this(zapiCommand, repeatingInstanceEltName, maxRecords, null, (NaElement[]) null);
    }

    protected IterBuilderBase(final String zapiCommand, final String repeatingInstanceEltName, final int maxRecords,
                              final NaElement desiredAttributes) {
        this(zapiCommand, repeatingInstanceEltName, maxRecords, desiredAttributes, (NaElement[]) null);
    }

    protected IterBuilderBase(final String zapiCommand, final String repeatingInstanceEltName, final int maxRecords,
                              final NaElement desiredAttributes, final NaElement... additionalElements) {
        this.zapiCommand = zapiCommand;
        this.repeatingInstanceEltName = repeatingInstanceEltName;
        this.maxRecords = maxRecords;
        this.desiredAttributes = desiredAttributes;
        this.additionalElements = additionalElements;
    }

    protected IterBuilderBase(final String zapiCommand, final String repeatingInstanceEltName, final int maxRecords,
                              final LinkedHashMap<String[], String[]> desiredAttributesMap) {
        this(zapiCommand, repeatingInstanceEltName, maxRecords, desiredAttributesMap, new LinkedHashMap<String, String>());
    }

    /**
     * builds the XML request model based on the set zapi parameters, parse the zapi reponse 
     * and return the results as a list of Info objects. 
     * @param zapiConnection
     * @return
     * @throws DataSourceErrorException
     */
    public List<E> buildModel(final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        LinkedList<E> dataFromZAPI = new LinkedList<E>();
        // ZAPI iterator support keeping track of nextTag, invoking command multiple times until all results returned
        final String logPrefix = PackageContext.getLogPrefix();
        String nextTag = null;      // keeps track of next tag to request when iterating.   Not specified on the first call.
        int iterCount = 0;
        final List<Integer> iterChunkCounts = new ArrayList<Integer>(100);
        final long startTime = System.currentTimeMillis();
        long totalZapiTime = 0;
        int totalZapiInvocations = 0;
        int totalElements = 0;

        Integer maxRecordsOverride = null;

        do {
            // Build XML request
            final NaElement request = ZAPIUtils.elem(zapiCommand);
            if (additionalElements != null) {
                for (final NaElement element : additionalElements) {
                    request.addChildElem(element);
                }
            }

            if (maxRecordsOverride != null) {
                logger.info("Override max records: " + zapiCommand + " " + maxRecordsOverride);
            }

            request.addChildElem(ZAPIUtils.elem(ZAPINames.MAX_RECORDS, (maxRecordsOverride == null) ? maxRecords : maxRecordsOverride));
            if (nextTag != null) {
                // Not the first call, and there are more records to iterate.   Specify where iteration left off
                request.addChildElem(ZAPIUtils.elem(ZAPINames.TAG, nextTag));
            }
            if (desiredAttributes != null) {
                request.addChildElem(desiredAttributes);
            }
            // Get next chunk of iteration elements
            final ZAPIConnection.ZapiResponse response = zapiConnection.invokeElem(request);
            iterCount++;
            totalZapiTime += response.getTotalTime();
            totalZapiInvocations += response.getInvocationCount();
            final Integer overrideCandidate = response.getMaxRecordsOverride();
            if (overrideCandidate != null) {
                // A success after a failure will NOT populate maxRecordsOverride, but we STILL
                // want to keep the override.  If we set it back to null, we will again get the default!
                maxRecordsOverride = overrideCandidate;
            }

            final Element resultsElt = response.getElement();
            if (resultsElt == null) {
                //Handling a case where the ZAPI returns a bad response
                iterChunkCounts.add(0);
                break;
            }
            final Element attributesListElt = resultsElt.getChild("attributes-list");
            if (attributesListElt == null) {
                iterChunkCounts.add(0);
                break;
            }
            // Now iterate through returned elements
            @SuppressWarnings({"unchecked"})
            final List<Element> iterationElts = attributesListElt.getChildren(repeatingInstanceEltName);
            final int chunkSizeReturned = iterationElts.size();
            iterChunkCounts.add(chunkSizeReturned);
            totalElements += chunkSizeReturned;
            for (final Element iterationElt : iterationElts) {
                E parsedData = buildIterInstance(iterationElt);
                if (parsedData != null) {
                    dataFromZAPI.add(parsedData);
                }
            }
            // Are there more records to iterate?
            final Element nextTagElt = resultsElt.getChild("next-tag");   // won't exist if no more records
            if (nextTagElt == null) {
                nextTag = null;                         // No more elements to iterate
            } else {
            	// More elements to iterate.
                // Remove whitespace, including newlines,
                // in the nextTag used in the next request,
                // in order to simplify simulation when the
                // recording platform is not the same as
                // the simulator platform.
                nextTag = nextTagElt.getTextTrim().replaceAll("\\s", "");       
            }
        }// end of do/while
        while (nextTag != null);

        final long consumedTime = System.currentTimeMillis() - startTime;

        final String fqZapiCommand = ("perf-object-instance-list-info-iter".equals(zapiCommand)) ? zapiCommand + " "
                + getAdditionalElementValue("objectname") : zapiCommand;
        if (logger.isDebugEnabled()) {
			logger.debug(new StringBuilder()
					.append(logPrefix)
					.append(" - ")
					.append(fqZapiCommand)
					.append(": max records: ")
					.append(((maxRecordsOverride == null) ? maxRecords
							: maxRecordsOverride)).append(", iterations: ")
					.append(iterCount)
					.append(", total clock time: ").append(consumedTime).append(" ms, total zapi time: ")
					.append(totalZapiTime).append(" ms, total zapi invocations ").append(totalZapiInvocations).append(", total elements: ").append(totalElements)
					.append(", buckets: ").append(getCSVStringFromIntList(iterChunkCounts).toString()));
        }

        try {
            final DSFileWriter responseTimesWriter = zapiConnection.getResponseTimesWriter();
            responseTimesWriter.writeln(PackageContext.getPackage().getConfiguration().getDataSourceName() + ", "
                    + fqZapiCommand + ", " + new Date(startTime) + ", " + ((maxRecordsOverride == null) ? maxRecords : maxRecordsOverride) + ", " + iterCount + ", "
                    + consumedTime + ", " + totalZapiTime + ", " + totalZapiInvocations + ", " + totalElements
                    + ", " + getCSVStringFromIntList(iterChunkCounts));
        } catch (IOException ioe) {
            logger.warn(logPrefix + " - IOException. Failed to record ZAPI response time for: " + zapiCommand + "(" + ioe.getMessage() + ")");
        }
        return dataFromZAPI;
    }

    protected String getAdditionalElementValue(final String name) {
        if (additionalElements != null) {
            for (final NaElement element : additionalElements) {
                if (name.equals(element.getName())) {
                    return element.getContent();
                }
            }
        }
        return null;
    }
    
    /**
     * each iteration element from the zapi response is parsed and new Info objects created
     * @param iterationElt
     * @return
     */
    protected abstract E buildIterInstance(final Element iterationElt);

    /**
     * given a list of integers returns them as a single string of CSV
     * @param list
     * @return string CSV
     */
    protected String getCSVStringFromIntList(final List<Integer> list) {
        final int listLen = list.size();
        if (listLen == 0) {
            return "";
        }
        final StringBuilder builder = new StringBuilder(listLen * 6);  // estimate for digits plus comma plus space for each entry
        builder.append(list.get(0));
        if (listLen > 1) {
            for (final Integer elt : list.subList(1, listLen)) {
                builder.append(", ");
                builder.append(elt);
            }
        }
        return builder.toString();
    }

}
