package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.parser;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorType;
import com.sun.jna.Native;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.IntByReference;
import com.sun.jna.ptr.PointerByReference;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This singleton class is a wrapper around the "cmaplwrapper" native library. The data from the native library is read
 * and the {@link com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.parser.CounterTable CounterTable} is
 * populated.
 */
public enum ArchiveFileParser {
    INSTANCE;
    private CLibrary clib;

    private static final Logger logger = LoggerFactory.getLogger(ArchiveFileParser.class);

    // Default constructor
    private ArchiveFileParser() {
        this.clib = (CLibrary) Native.loadLibrary("cmaplwrapperlib", CLibrary.class);
    }

    /**
     * Retrieves a list of counter values (Type : int, long or String) for a
     * Specific object at a specific timestamp and returns a new CounterTable instance
     * needed by MEGA
     *
     * @param ccmaFile Absolute path to the ccma file.
     * @param time Archive counter time stamp
     * @param currentObject Counter Manager counter object name
     * @return Counter data
     * @throws DataSourceErrorException
     */
    public CounterTable getCountersForObjectTimestamp(final String ccmaFile, final long time, final long timePeriod,
                                                                  final String currentObject) {
        final long startTime = System.currentTimeMillis();
        final List<String> countersHeaderList;
        final List<CounterRow> counterRowList;
        Pointer ptrFirstWrapperCntValue = null;
        int instanceCount = 0;
        int headersCount = 0;
        Pointer ptrFirstWrapperHeader = null;
        final PointerByReference error = new PointerByReference();
        Pointer pError = null;
        String errorString = null;


        try {
            final IntByReference counter_data_handle_count = new IntByReference();
            final PointerByReference wrapper_counter_data = new PointerByReference();
            final IntByReference header_count = new IntByReference();
            final PointerByReference wrapper_headers = new PointerByReference();

            final boolean ret_value = clib.get_counter_values_for_timestamp(ccmaFile, time, timePeriod, currentObject,
                    wrapper_counter_data, counter_data_handle_count, wrapper_headers, header_count, error);

            // Check if there are any errors
            if(error != null) {
                pError = error.getValue();
                errorString = (pError != null) ? pError.getString(0) : null;
            }

            if (!ret_value) {
                final StringBuilder dMessage = new StringBuilder("Could not get Counters for ").append(ccmaFile)
                        .append(", Object : ").append(currentObject).append(", timestamp ").append(time)
                        .append(", ret_val :").append(ret_value);
                if(errorString != null && !errorString.isEmpty())
                    dMessage.append(", parse error :").append(errorString);
                logger.error(dMessage.toString());
            } else {
                //Print headers array
                ptrFirstWrapperHeader = wrapper_headers.getValue();
                final CLibrary.CmaplCounterWrapperHeaders cmaplCounterWrapperHeaders = new CLibrary.CmaplCounterWrapperHeaders(ptrFirstWrapperHeader);
                // gives the no. of headers
                headersCount = header_count.getValue();
                if (headersCount != 0) {
                    final CLibrary.CmaplCounterWrapperHeaders[] cmaplCounterWrapperHeadersArray = (CLibrary.CmaplCounterWrapperHeaders[]) cmaplCounterWrapperHeaders.toArray(headersCount);
                    //Capture the Counter Header values into the List
                    countersHeaderList = new ArrayList<String>(headersCount);
                    for (final CLibrary.CmaplCounterWrapperHeaders ch : cmaplCounterWrapperHeadersArray) {
                        countersHeaderList.add(ch.counter_name);
                    }

                    ptrFirstWrapperCntValue = wrapper_counter_data.getValue();
                    final CLibrary.CmaplWrapperCounterDataT cmapl_wrapper_ref = new CLibrary.CmaplWrapperCounterDataT(ptrFirstWrapperCntValue);
                    // gives the no. of instances for counters for a specific object.
                    instanceCount = counter_data_handle_count.getValue();
                    final CLibrary.CmaplWrapperCounterDataT[] cmapl_wrapper_ref_array = (CLibrary.CmaplWrapperCounterDataT[]) cmapl_wrapper_ref.toArray(instanceCount);

                    // Create list to hold the counter data
                    counterRowList = new ArrayList<CounterRow>(instanceCount);

                    for (final CLibrary.CmaplWrapperCounterDataT cwra : cmapl_wrapper_ref_array) {

                        final Pointer ptrFirstCntDataValue = cwra.counter_values.getPointer();
                        final CLibrary.CmaplCounterDataT cmaplCounterDataT = new CLibrary.CmaplCounterDataT(ptrFirstCntDataValue);
                        final CLibrary.CmaplCounterDataT[] cmaplCounterDataTArray = (CLibrary.CmaplCounterDataT[]) cmaplCounterDataT.toArray(cwra.num_of_counters_count);
                        final List<CounterColumn> counterDataPerInstance = new ArrayList<CounterColumn>(cmaplCounterDataTArray.length);

                        int columnIndex = 0;
                        for (final CLibrary.CmaplCounterDataT ccdt : cmaplCounterDataTArray) {
                            final Pointer ptrFirstCntValue = ccdt.values.getPointer();
                            final CLibrary.CmaplCntValueT cmaplCntValueT = new CLibrary.CmaplCntValueT(ptrFirstCntValue);
                            final CLibrary.CmaplCntValueT[] cmaplCntValueTArray = (CLibrary.CmaplCntValueT[]) cmaplCntValueT.toArray(ccdt.num_values);
                            final int countCntValues = cmaplCntValueTArray.length;
                            final List<Object> counterDataPerCounter = new ArrayList<Object>(countCntValues);

                            for (final CLibrary.CmaplCntValueT ccvt : cmaplCntValueTArray) {
                                switch (ccdt.value_type) {
                                    case CLibrary.CmaplCounterDataType.CMAPL_CNT_4BYTES:
                                        cmaplCntValueT.setType(Integer.TYPE);
                                        cmaplCntValueT.read();
                                        counterDataPerCounter.add(ccvt.val_4B);
                                        break;
                                    case CLibrary.CmaplCounterDataType.CMAPL_CNT_8BYTES:
                                        cmaplCntValueT.setType(Long.TYPE);
                                        cmaplCntValueT.read();
                                        counterDataPerCounter.add(ccvt.val_8B);
                                        break;
                                    case CLibrary.CmaplCounterDataType.CMAPL_CNT_STRING:
                                        cmaplCntValueT.setType(String.class);
                                        cmaplCntValueT.read();
                                        counterDataPerCounter.add(ccvt.val_str);
                                        break;
                                }
                            }
                            counterDataPerInstance.add(new CounterColumn(countersHeaderList.get(columnIndex), counterDataPerCounter));
                            columnIndex++; //increment column index.
                        }

                        counterRowList.add(new CounterRow(counterDataPerInstance));
                    }
                    logger.debug("{} msec to parse {}", (startTime - System.currentTimeMillis()), ccmaFile);
                    return new CounterTable(time, currentObject, counterRowList);
                } else {
                    final StringBuilder message = new StringBuilder("Skipping CCMA file : ").append(ccmaFile)
                            .append(", Object : ").append(currentObject)
                            .append(", timestamp ").append(time)
                            .append(" does not contain counter Data.");
                    logger.info(message.toString(), message);
                }
            }
        } catch (Throwable allExceptions) {
            final StringBuilder message = new StringBuilder("Error parsing CCMA file : ").append(ccmaFile)
                    .append(", Object : ").append(currentObject)
                    .append(", timestamp ").append(time)
                    .append(", ").append(allExceptions.getClass().getName())
                    .append(":").append(allExceptions.getMessage());
            // adding the below line check, just in case error message has been generated by the cmapl_wrapper library, but something went wrong in JNA translation
            if(errorString != null && !errorString.isEmpty())
                message.append(", parse error :").append(errorString);
            logger.error(message.toString(), allExceptions);
        } finally {
            try {
                //Free native pointers
                clib.free_cmapl_wrapper_counter_data(ptrFirstWrapperCntValue, instanceCount, ptrFirstWrapperHeader, headersCount, pError);
            } catch (Throwable ignored) {

            }
        }
        return null; // Error parsing ccma archiver files occured. So ignore the file. Null check is handled in the calling code.
    }

    /**
     * Writes the counter values for a specific object at a specific timestamp into a file.
     * This method is currently not used, but could be useful for debugging purpose counter values in a ccma file.
     * @param ccma_file_path Absolute path to the ccma file.
     * @param output_file_path Absolute path to the output file.
     * @param time Archive counter time stamp
     * @param current_obj_char Counter Manager counter object name
     * @return true - Success, false - failure
     * @throws DataSourceErrorException
     */
    public boolean writeCounterValuesTimestampToFile(final String ccma_file_path,
                                                                  final String output_file_path,
                                                                  final long time, final long timePeriod,
                                                                  final String current_obj_char) {
        final boolean return_val;
        return_val = clib.write_counter_values_for_timestamp_to_file(ccma_file_path, output_file_path, time, timePeriod,  current_obj_char);
        if (!return_val) {
            final String message = new StringBuilder("Could not get Counters from file : ").append(ccma_file_path)
                    .append(", object : ").append(current_obj_char).append(", time ").append(time)
                    .append(". return_val").append(return_val).toString();
            logger.error(message);
        }
        return return_val;
    }


    /**
     * Gets a list of Timestamps for a specific object (resource, workload_queue_nblade...).
     *
     * @param ccma_file_path Absolute path to the ccma file.
     * @param current_object  Counter Manager counter object name
     * @return List of Time stamps
     * @throws DataSourceErrorException
     */
    public List<Long> getTimestampsForObject(final String ccma_file_path, final String current_object, final long timePeriod) {
        final PointerByReference array = new PointerByReference();
        final IntByReference number = new IntByReference();
        final PointerByReference error = new PointerByReference();
        Pointer timestampListPointer = null;
        Pointer pError = null;
        String errorString = null;
        List<Long> timestamps = Collections.EMPTY_LIST;
        boolean return_val = false;
        boolean exceptionOccured = false;
        try {
            return_val = clib.get_timestamp_list(ccma_file_path, timePeriod, current_object, array, number, error);

            // Check if there are any errors
            if(error != null) {
                pError = error.getValue();
                errorString = (pError != null) ? pError.getString(0) : null;
            }

            if (return_val) {
                final int number_timestamps = number.getValue();
                timestamps = new ArrayList<Long>(number_timestamps);
                timestampListPointer = array.getValue();
                for (int i = 0; i < number_timestamps; i++) {
                    timestamps.add(timestampListPointer.getLong(i * Native.getNativeSize(Long.TYPE)));
                }
            } else {
                final StringBuilder dMessage = new StringBuilder("Could not get timestamps from file : ").append(ccma_file_path)
                        .append(", object : ").append(current_object);
                // Add error from cmapl library to exception message
                if(errorString != null && !errorString.isEmpty())
                    dMessage.append(", parse error :").append(errorString);
                final String message = dMessage.toString();
                logger.error(message);
            }
        } catch (Throwable allExceptions) {
            final StringBuilder dMessage = new StringBuilder("Could not get timestamps from file : ").append(ccma_file_path)
                    .append(", object : ").append(current_object)
                    .append(". ").append(allExceptions.getClass().getName())
                    .append(": ").append(allExceptions.getMessage());
            // adding the below line check, just in case error message has been generated by the cmapl_wrapper library, but something went wrong in JNA translation
            if(errorString != null && !errorString.isEmpty())
                dMessage.append(", parse error :").append(errorString);
            logger.error(dMessage.toString());
            exceptionOccured = true;
        } finally {
            try {
                clib.free_cmapl_wrapper_timestamps_data(timestampListPointer, pError); // de-allocate native timestampList & error which was malloced
            } catch (Throwable ignore) {
            }
        }
        return timestamps;
    }


    /**
     * Write the timestamps for a specific object (resource, workload_queue_nblade) into file.
     * This method is currently not used, but could be useful for debugging timestamps values in a ccma file.
     * @param ccma_file_path  Absolute path to the ccma file.
     * @param output_file_path Absolute path to the output file.
     * @param current_object  Counter Manager counter object name
     * @throws DataSourceErrorException
     */
    public boolean writeTimestampsForObjectToFile(final String ccma_file_path, final String output_file_path,
                                                    final String current_object, final long timePeriod) {
        final boolean return_val;
        return_val = clib.write_timestamps_to_file(ccma_file_path, output_file_path, timePeriod, current_object);
        if (!return_val) {
            logger.error(new StringBuilder("Could not get timestamps from file : ").append(ccma_file_path)
                    .append(", object : ").append(current_object).toString());
        }
        return return_val;
    }
}
