package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppCommonConfig;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapper;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.*;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll.PartialPollContext;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll.PartialPollException;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.*;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Lif;
import com.onaro.util.VersionComparator;

import netapp.manage.NaElement;

import org.jdom.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIUtils.children;
import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIUtils.elem;

/**
 * Builds the model of the Cluster by querying the Cluster itself and by delegating to sub builders.
 */
public class ClusterBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(ClusterBuilder.class);
    private final Properties properties;
    private final PartialPollContext partialPollContext;

    private CModeCluster cModeCluster = null;
    private ZAPIConnection zapiConnection = null;
    private final BaseDataSource dataSource;

    /**
     * Indicates whether or not OCI performance is enabled.  OCI does not share DS config
     * with OCIE, but delegates to OCIE for collection.   This flag allows OCI to enable
     * performance specific functionality that must occur during the foundation poll.
     */
    private boolean isOCIPerfEnabledByOCI = false;

  /**
     * Allows OCI to enable performance without using OCIE configuration.
     * Appears unused because it is called from code that lives in a different stream
     * @param OCIPerfEnabled boolean indicating whether to invoke specific performance
     */
    public void setOCIPerfEnabled(final boolean OCIPerfEnabled) {
        isOCIPerfEnabledByOCI = OCIPerfEnabled;
    }

    public ClusterBuilder(final BaseDataSource dataSource, final PartialPollContext partialPollContext, final Properties properties) {
        this.dataSource = dataSource;
      this.properties = properties;
      this.partialPollContext = partialPollContext;
    }

    /**
     * Collection of accumulated DataSourceErrorsExceptions.  When a ZAPI interaction fails in a non-fatal way,
     * we collect the exception and continue acquisition in order to create a complete recording of cluster discovery.
     * All encountered exceptions are bulk reported when all requests have been attempted.
     */
    private final AggregatedDataSourceErrorsException allDSErrExceptions = new AggregatedDataSourceErrorsException(1);

    public CModeCluster validateTargetCluster(final CModeClusterRoot clusterRoot) throws FrameworkException, DataSourceErrorException {
        return getCModeCluster(clusterRoot);
    }

    /**
     * Execute ZAPI queries and build the model from the responses
     *
     * @param clusterRoot the root of the cluster model object graph
     * @return total amount of time spent waiting for ZAPI responses
     * @throws FrameworkException       on duplicate data source
     * @throws DataSourceErrorException when building the model fails
     */
    public long executeAndBuild(final CModeClusterRoot clusterRoot) throws FrameworkException, DataSourceErrorException {

        final ZAPIConnection zapiConnection = getZAPIConnection( properties );
        final CModeCluster cluster = getCModeCluster(clusterRoot);

        final String numThreadsStr = PackageContext.getPackage().getConfiguration().getProperty("numBuilderThreads", "2");
        final int numBuilderThreads;
        if ("nodes".equalsIgnoreCase(numThreadsStr)) {
            //"Nodes" is a flag telling us to create a number of threads equal to the number of cluster nodes.
            //The Node version array is the only meta-data hook we have for this count at this point, so use this
            //as the node count, if available.
            numBuilderThreads = cluster.getCModeNodeVersions().isEmpty() ? 1 : cluster.getCModeNodeVersions().size();
        } else {
            numBuilderThreads = Integer.parseInt(numThreadsStr);
        }

        final String defaultIteratorMaxRecordsStr = PackageContext.getPackage().getConfiguration().getProperty("default.iterator.maxrecords", "1000");
        final int defaultIteratorMaxRecords = Integer.parseInt(defaultIteratorMaxRecordsStr);
        IterBuilderBase.setDefaultMaxRecords(defaultIteratorMaxRecords);

        LOG.info("{} - Number of threads: {}", PackageContext.getLogPrefix(), numBuilderThreads);

        final ExecutorService builderExecutor = Executors.newFixedThreadPool(numBuilderThreads);

        /*
         * Build an ordered list of callable builders.   The order here is significant.  Generally, we want
         * longer builders first in the list so they are started at the beginning.
         *
         * TODO - Review the suitability of zapiConnection for multi-threading.
         */
        final List<CallableBuilder> builders = new LinkedList<CallableBuilder>();
        builders.add(new CallableBuilder(new DiagnosisStatusBuilder(), cluster, zapiConnection));
        // JSD story added & removed, probably coming back someday..
        // builders.add(new CallableBuilder(new LowOnSpareBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new VolumeBuilder(), cluster, zapiConnection));
        if (isSnapshotDiscoveryEnabled()) {
            builders.add(new CallableBuilder(new SnapshotBuilder(), cluster, zapiConnection));
        }
        builders.add(new CallableBuilder(new LUNBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new QTreeBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new SnapMirrorBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new QuotaBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new VServerBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new SnapMirrorPolicyBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new SnapMirrorHistoryBuilder(), cluster, zapiConnection));

        // TA32687 discover QoS Policy Groups even if QoS is off
        builders.add(new CallableBuilder(new QosPolicyGroupBuilder(), cluster, zapiConnection));
        if (isQosDiscoveryEnabled()) {
            builders.add(new CallableBuilder(new QosVolumeWorkloadBuilder(), cluster, zapiConnection));
            builders.add(new CallableBuilder(new QosServiceCenterBuilder(), cluster, zapiConnection));
        }
        builders.add(new CallableBuilder(new AggregateBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new AggregateSpareBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new StoragePoolBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new StoragePoolAvailableSizeBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new StoragePoolAggregateBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new DiskBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new AggregateStatusInfoBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NodeBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new VolumeMoveBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new LUNMapBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new CIFSBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new CIFSACLBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new CIFSOptionsBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new ExportsBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NFSBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new JobScheduleBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new SisPolicyBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new SnapshotPolicyBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new IGroupBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new PortSetBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NetPortBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NetworkIpSpaceBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NetPortBroadcastDomainBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NetworkFailoverGroupBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NetworkSubnetBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new FcpBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new FcpInterfaceBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new FlexCacheBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new VServerPeerBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new StorageAdapterBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new FibreChannelBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new ISCSIBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(cluster.isPreSierraNevada()
                ? new LicenseInterimBuilder()
                : new LicenseV2Builder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new NetRoutingBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new KerberorsBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new StorageServiceBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new TimezoneBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new VServerNameMapBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new ApplicationRecordBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new QuotaStatusBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new LunImportBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new MetroClusterOperationBuilder(), cluster, zapiConnection));
        
        builders.add(new CallableBuilder(new ClusterPeerBuilder(), cluster, zapiConnection));

        /*
                fix of burt 804059. May use this code later.
         */
        //builders.add(new CallableBuilder(new StorageSwitchBuilder(), cluster, zapiConnection));
        builders.add(new CallableBuilder(new SystemHealthAlertBuilder(), cluster, zapiConnection));
        if (isPerformanceEnabled()) {
            final boolean ociPerfEnabled = isOCIPerformanceEnabledByProp() || isOCIPerfEnabledByOCI;
            builders.add(new CallableBuilder(new NodeStatsUUIDBuilder(ociPerfEnabled), cluster, zapiConnection));
            
            builders.add(new CallableBuilder(new VServerStatsIDMapBuilder(), cluster, zapiConnection));
        }

      //Runtime builders
      builders.add(new CallableBuilder(new VolumeRuntimeBuilder(), cluster, zapiConnection));

      //invoke all builders and wait for them to complete
        try {
            final List<Future<CallableBuilder.BuilderResult>> futures = builderExecutor.invokeAll(builders);
            for (final Future<CallableBuilder.BuilderResult> future : futures) {
                try {
                    final CallableBuilder.BuilderResult result = future.get();
                    LOG.debug("{}Builder {} completed in {}s.", new Object[] {
                            PackageContext.getLogPrefix(), result.getName(), result.getTime() });
                } catch (ExecutionException exp) {
                    final Throwable t = exp.getCause();
                    if (t instanceof DataSourceErrorException) {
                        allDSErrExceptions.addException((DataSourceErrorException) t, null);
                    } else {
                        allDSErrExceptions.addException(new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Internal error", "Internal Error", t), null);
                    }
                }
            }
        } catch (InterruptedException ie) {
            LOG.error("Builders interrupted", ie);
        }

        //clean up thread pool
        builderExecutor.shutdown();

        //clean up zapis state
        zapiConnection.recordAttributes();
        zapiConnection.close();

        //Mark the poll as completed
        partialPollContext.completed();

        //throw exceptions if error conditions
        if (allDSErrExceptions.hasExceptions()) {
            throw allDSErrExceptions;
        }
        return zapiConnection.getTotalZAPILatency();
    }

    private ZAPIConnection getZAPIConnection() throws DataSourceErrorException {
      return getZAPIConnection( null );
    }

    private ZAPIConnection getZAPIConnection( final Properties properties ) throws DataSourceErrorException {
        if (this.zapiConnection == null) {
            this.zapiConnection = ZAPIConnectionFactory.createConnection( properties );
        }
        return this.zapiConnection;
    }

    private CModeCluster getCModeCluster(final CModeClusterRoot clusterRoot) throws FrameworkException, DataSourceErrorException {

        if (this.cModeCluster != null) {
            return this.cModeCluster;
        }

        // We need to assert a few things before we start collecting lots of data.
        // 1)     duplicate data source - is the cluster UUID already monitored
        //        by another data source instance?
        // 2)     are we talking to the correct lif?   Role must be cluster_mgmt
        //        For KFS simulator we can't get the role, we allow this (don't fail) or testing
        // 3)     If snapshot discovery is enabled, assert extra long poll period

        // Get the cluster ID create the parent cluster model object from it.
        // This will assert we are not a duplicate data source
        final ZAPIConnection zapiConnection = getZAPIConnection( properties );
        final CModeCluster cluster = buildClusterModel(clusterRoot, zapiConnection);

        //initialize partial polling and set it on the zapi connection
	      initPartialPollContext( cluster, zapiConnection );
        zapiConnection.setPartialPollContext( partialPollContext );

        // Build the LIFs first so we can assert the target LIF role
        try {
            new NetInterfaceBuilder().buildModel(cluster, zapiConnection);
        } catch (DataSourceErrorException dsee) {
            allDSErrExceptions.addException(dsee, null);
        }

        // Assert we are talking to a cluster_mgmt LIF
        assertTargetLIFRole(cluster, zapiConnection);

        // Check for snapshot discovery enabled.   If enabled, the poll period must be
        // at least 6 hours, otherwise we should fail the poll.
        if (isSnapshotDiscoveryEnabled()) {
            // Assert foundation poll period is at least 6 hours
            assertEnabledSnapshotDiscoveryPollPeriod();
        }
        this.cModeCluster = cluster;
        return this.cModeCluster;
    }

  /**
   * Create an initialized partial poll context for the supplied cluster
   *
   * @param cluster the partial poll is for
   * @param zapiConnection a zapi connection to the cluster the supplied in cluster parameter
   *
   * @throws DataSourceErrorException
   */
  private void initPartialPollContext(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException
  {
    try
    {
      //reinitialize the context for this poll execution
      partialPollContext.initialize( cluster.getVersionString(), properties );

      //Iterate through the results to find which zapis have changes that require updates
      if( partialPollContext.isEnabled() )
      {
        SystemAPIChangeBuilder builder = new SystemAPIChangeBuilder( partialPollContext );
        builder.buildModel( partialPollContext, zapiConnection );
      }
    }
    catch( PartialPollException e )
    {
      //If there is an issue log the error and disable the partial polling
      partialPollContext.setEnabled( false );
      LOG.error( "Could not initialize partial polling", e );
    }
  }

  private CModeCluster buildClusterModel(final CModeClusterRoot clusterRoot, final ZAPIConnection zapiConnection) throws FrameworkException, DataSourceErrorException
  {
    // Assert we are talking to a valid cluster
    final Element systemVersionResponseElt = getSystemVersionResponseElt( zapiConnection );

    // Get the cluster ID create the parent cluster model object from it
    final CModeCluster cModeCluster = buildAndAddCluster( clusterRoot, zapiConnection );

    // Extract more cluster attributes
    extractNodeVersions( zapiConnection, systemVersionResponseElt, cModeCluster );
    extractClusterVersion( zapiConnection, systemVersionResponseElt, cModeCluster );

    return cModeCluster;
  }

    /**
     * Build the cluster model from cluster-identity-get results and add it to the cluster model root.
     *
     * @param clusterRoot    the cluster model root
     * @param zapiConnection established authenticated connection to ZAPI server used to get cluster-identity-get results
     * @return the cluster model from cluster-identity-get results (added to the cluster model root)
     * @throws DataSourceErrorException when building the model fails
     */
    private CModeCluster buildAndAddCluster(final CModeClusterRoot clusterRoot, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        // Get cluster ID info
        final ZAPIConnection.ZapiDesiredAttributes clusterIDAttributesArg = new ZAPIConnection.ZapiDesiredAttributes("cluster-identity-info", new LinkedHashMap<String[], String[]>() {{
            put(new String[]{"cluster-identity-info"}, new String[]{"cluster-contact", "cluster-location", "cluster-name", "cluster-serial-number", "cluster-uuid"});
        }});
        final ZAPIConnection.ZapiResponse clusterIDResponse = zapiConnection.invokeCommandDesiredAttributes("cluster-identity-get", null, clusterIDAttributesArg);
        final Element clusterIDResponseElt = clusterIDResponse.getElement();
        final Element attributesElt = clusterIDResponseElt.getChild("attributes");
        final Element clusterIDInfoElt = attributesElt.getChild("cluster-identity-info");
        final String clusterContact = BuilderUtil.getChildTextTrim(clusterIDInfoElt, "cluster-contact");
        final String clusterLocation = BuilderUtil.getChildTextTrim(clusterIDInfoElt, "cluster-location");
        final String clusterName = BuilderUtil.getChildTextTrim(clusterIDInfoElt, "cluster-name");
        final String clusterSerialNum = BuilderUtil.getChildTextTrim(clusterIDInfoElt, "cluster-serial-number");
        final String clusterUUID = BuilderUtil.getChildTextTrim(clusterIDInfoElt, "cluster-uuid");
        final NetAppCommonConfig config = (NetAppCommonConfig) PackageContext.getPackage().getConfiguration();
        final String managementIP = config.getIP();
        final CModeCluster cModeCluster = clusterRoot.createCluster(clusterUUID);
        cModeCluster.setClusterName(clusterName);
        cModeCluster.setManagementIP(managementIP);
        cModeCluster.setSerialNum(clusterSerialNum);
        cModeCluster.setContact(clusterContact);
        cModeCluster.setLocation(clusterLocation);

        populateClusterHAAttributes(cModeCluster, zapiConnection);
        populateIsMetroclusterAttributes(cModeCluster, zapiConnection);
        if(cModeCluster.getIsMetrocluster())  {
            populateMetroClusterAttributes(cModeCluster, zapiConnection);
        }
        else {  //if not MCC, set up null values.
            cModeCluster.setMtConfigurationState(new EnumWrapper<Cluster.MtConfigurationState>(null, null));
            cModeCluster.setMtMode(new EnumWrapper<Cluster.MtMode>(null, null));
            cModeCluster.setMtRemoteConfigurationState(new EnumWrapper<Cluster.MtConfigurationState>(null, null));
            cModeCluster.setMtRemoteMode(new EnumWrapper<Cluster.MtMode>(null, null));
        }

        populateSslFipsEnabledAttributes(cModeCluster, zapiConnection);
        
        clusterRoot.getCModeClusters().add(cModeCluster);

        return cModeCluster;
    }

    private void populateClusterHAAttributes(final CModeCluster cModeCluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final ZAPIConnection.ZapiDesiredAttributes desiredAttributes = new ZAPIConnection.ZapiDesiredAttributes("cluster-ha-info", new LinkedHashMap<String[], String[]>() {{
            put(new String[]{"cluster-ha-info"}, new String[]{"ha-configured"});
        }});
        final Element response = zapiConnection.invokeCommandDesiredAttributes("cluster-ha-get", null, desiredAttributes).getElement();
        if (response != null) {

            final Element attributesElt = response.getChild("attributes");
            if (attributesElt != null) {
                final Element clusterInfoElt = attributesElt.getChild("cluster-ha-info");
                cModeCluster.setHaConfigured(BuilderUtil.parseBooleanFromChild(clusterInfoElt, "ha-configured"));
            }
        }
    }

    private void populateIsMetroclusterAttributes(final CModeCluster cModeCluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final Element response = zapiConnection.invokeCommand("metrocluster-is-configured", null).getElement();
        cModeCluster.setIsMetrocluster(BuilderUtil.parseBooleanFromChildDefaultFalse(response, "is-mcc-config"));
    }
    
    private void populateSslFipsEnabledAttributes(final CModeCluster cModeCluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final ZAPIConnection.ZapiDesiredAttributes desiredAttributes = new ZAPIConnection.ZapiDesiredAttributes("system-services-web-info", new LinkedHashMap<String[], String[]>() {{
            put(new String[]{"system-services-web-info"}, new String[]{"ssl-fips-enabled"});
        }});
        final Element response = zapiConnection.invokeCommandDesiredAttributes("system-services-web-get", null, desiredAttributes).getElement();
        if (response != null) {

            final Element attributesElt = response.getChild("attributes");
            if (attributesElt != null) {
                final Element clusterInfoElt = attributesElt.getChild("system-services-web-info");
                cModeCluster.setSslFipsEnabled(BuilderUtil.parseBooleanFromChild(clusterInfoElt, "ssl-fips-enabled"));
            }
        }
    }

    private void populateMetroClusterAttributes(final CModeCluster cModeCluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final ZAPIConnection.ZapiDesiredAttributes desiredAttributes = new ZAPIConnection.ZapiDesiredAttributes("metrocluster-info", new LinkedHashMap<String[], String[]>() {{
            put(new String[]{"metrocluster-info"}, new String[]{"local-configuration-state", "local-mode", "configuration-type", "remote-configuration-state", "remote-mode"});
        }});
        final Element response = zapiConnection.invokeCommandDesiredAttributes("metrocluster-get", null, desiredAttributes).getElement();

        EnumWrapper<Cluster.MtConfigurationState> mtConfigurationState = new EnumWrapper<Cluster.MtConfigurationState>(null, null);
        EnumWrapper<Cluster.MtMode> mtMode = new EnumWrapper<Cluster.MtMode>(null, null);
        String mtConfigurationType = null;
        EnumWrapper<Cluster.MtConfigurationState> mtRemoteConfigurationState = new EnumWrapper<Cluster.MtConfigurationState>(null, null);
        EnumWrapper<Cluster.MtMode> mtRemoteMode = new EnumWrapper<Cluster.MtMode>(null, null);

        if (response != null) {
            final Element attributesElt = response.getChild("attributes");
            if (attributesElt != null) {
                final Element metroclusterInfoElt = attributesElt.getChild("metrocluster-info");
                if (metroclusterInfoElt != null) {
                    mtConfigurationState = BuilderUtil.getEnumWrapperByValue(
                            Cluster.MtConfigurationState.class, metroclusterInfoElt, "local-configuration-state", Cluster.MtConfigurationState.NOT_MAPPED);
                    mtMode = BuilderUtil.getEnumWrapperByValue(
                            Cluster.MtMode.class, metroclusterInfoElt, "local-mode", Cluster.MtMode.NOT_MAPPED);
                    mtRemoteConfigurationState = BuilderUtil.getEnumWrapperByValue(
                            Cluster.MtConfigurationState.class, metroclusterInfoElt, "remote-configuration-state", Cluster.MtConfigurationState.NOT_MAPPED);
                    mtRemoteMode = BuilderUtil.getEnumWrapperByValue(
                            Cluster.MtMode.class, metroclusterInfoElt, "remote-mode", Cluster.MtMode.NOT_MAPPED);
                    mtConfigurationType = BuilderUtil.getChildText(metroclusterInfoElt, "configuration-type");
                }
            }
        }

        cModeCluster.setMtConfigurationState(mtConfigurationState);
        cModeCluster.setMtMode(mtMode);
        cModeCluster.setMtRemoteConfigurationState(mtRemoteConfigurationState);
        cModeCluster.setMtRemoteMode(mtRemoteMode);
        cModeCluster.setMtConfigurationType(mtConfigurationType);
    }

    /**
     * Get the system-get-version response XML Element
     *
     * @param zapiConnection established authenticated connection to ZAPI server
     * @return the system-get-version response XML Element
     * @throws DataSourceErrorException on communication errors, 7 Mode Filers, and unsupported cluster versions
     */
    private Element getSystemVersionResponseElt(final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        // Make sure this is a valid Cluster, if not fail the poll immediately.
        final ZAPIConnection.ZapiResponse systemVersionResponse = zapiConnection.invokeCommand("system-get-version", null);
        final Element systemVersionResponseElt = systemVersionResponse.getElement();
        final String versionString = systemVersionResponseElt.getChildTextTrim("version");
        final String isClusteredString = systemVersionResponseElt.getChildTextNormalize("is-clustered");
        final Boolean isClustered = Boolean.parseBoolean(isClusteredString);
        if (!isClustered) {
            final String errorMessage = "7 Mode filers are not supported.";
            final String targetIP = zapiConnection.getIP();
            final EnhancedErrorDetail eed = new EnhancedErrorDetail("notCMode.html", targetIP, errorMessage);
            eed.addParam("IP", targetIP);
            eed.addParam("FilerVersion", versionString);
            throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, errorMessage, eed, null);
        }

        // Cluster must be at least version 8.1
        final String minimumSupportedVersion = "8.1";
        if (VersionComparator.compareVersions(minimumSupportedVersion, versionString) > 0) {
            final String errorMessage = "Cluster versions before " + minimumSupportedVersion + " are not supported.";
            final String targetIP = zapiConnection.getIP();
            final EnhancedErrorDetail eed = new EnhancedErrorDetail("not81.html", targetIP, errorMessage);
            eed.addParam("IP", targetIP);
            eed.addParam("ClusterVersion", versionString);
            throw DataSourceErrorException.createWithEnhanced(DataSourceErrorType.RETRIEVE_DATA_FROM_DEVICE, errorMessage, eed, null);
        }

        return systemVersionResponseElt;
    }

    /**
     * Extract the map of node UUID to node version from the system-get-version response XML Element
     *
     * @param zapiConnection           connection used to get the response, for manually recording desired attributes only (ZAPI does not support desired attributes)
     * @param systemVersionResponseElt the system-get-version response XML Element
     * @param cModeCluster             the model of the cModeCluster to populate
     */
    private void extractNodeVersions(final ZAPIConnection zapiConnection, final Element systemVersionResponseElt, final CModeCluster cModeCluster) {
        // manually record attributes because we can not specify desired attributes
        zapiConnection.recordRequestedAttribute("system-get-version.node-version-details.node-version-detail-info.node-uuid");
        zapiConnection.recordRequestedAttribute("system-get-version.node-version-details.node-version-detail-info.version");

        // Extract the version of each Node in the cluster from the response
        final Map<String, String> nodeVersions = new LinkedHashMap<String, String>();
        final Element nodeVersionDetailsElt = systemVersionResponseElt.getChild("node-version-details");
        if (nodeVersionDetailsElt != null) {
            @SuppressWarnings({"unchecked"})
            final List<Element> nodeVersionDetailElts = nodeVersionDetailsElt.getChildren("node-version-detail-info");
            for (final Element nodeVersionDetailElt : nodeVersionDetailElts) {
                final String nodeUUID = BuilderUtil.getChildTextTrim(nodeVersionDetailElt, "node-uuid");
                final String nodeVersion = BuilderUtil.getChildTextTrim(nodeVersionDetailElt, "version");
                nodeVersions.put(nodeUUID, nodeVersion);
            }
        }
        cModeCluster.setCModeNodeVersions(nodeVersions);
    }

    /**
     * Extract cluster version consisting of Generation, Major, Minor as an Integer array of length 3 from the system-get-version response XML Element
     *
     * @param zapiConnection           connection used to get the response, for manually recording desired attributes only (ZAPI does not support desired attributes)
     * @param systemVersionResponseElt the system-get-version response XML Element
     * @param cModeCluster             the model of the cModeCluster to populate
     */
    private void extractClusterVersion(final ZAPIConnection zapiConnection, final Element systemVersionResponseElt, final CModeCluster cModeCluster) {
        // manually record attributes because we can not specify desired attributes
        zapiConnection.recordRequestedAttribute("system-get-version.version");
        zapiConnection.recordRequestedAttribute("system-get-version.version-tuple.system-version-tuple.generation");
        zapiConnection.recordRequestedAttribute("system-get-version.version-tuple.system-version-tuple.major");
        zapiConnection.recordRequestedAttribute("system-get-version.version-tuple.system-version-tuple.minor");

        final String versionString = BuilderUtil.getChildTextTrim(systemVersionResponseElt, "version");
        cModeCluster.setVersionString(versionString);

        final Element versionTupleElt = systemVersionResponseElt.getChild("version-tuple");
        final Element systemVersionTupleElt = versionTupleElt.getChild("system-version-tuple");
        final Integer generation = BuilderUtil.parseIntFromString(systemVersionTupleElt.getChildTextTrim("generation"));
        final Integer major = BuilderUtil.parseIntFromString(systemVersionTupleElt.getChildTextTrim("major"));
        final Integer minor = BuilderUtil.parseIntFromString(systemVersionTupleElt.getChildTextTrim("minor"));

        cModeCluster.setGeneration(generation);
        cModeCluster.setMajor(major);
        cModeCluster.setMinor(minor);
    }

    private boolean assertTargetLIFRole(final CModeCluster cModeCluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        boolean validatedLif = false;
        final String targetIPString = zapiConnection.getTargetIPAddressString();
        for (final CModeLif candidateLif : cModeCluster.getLifs().values()) {
            if (targetIPString.equals(candidateLif.getIpAddress())
                    || convertToStandardForm(targetIPString).equals(convertToStandardForm(candidateLif.getIpAddress()))) {
                // This is the LIF we are talking to.   Make sure the Role is "cluster_mgmt".
                final Lif.Role lifRole = candidateLif.getRole().getEnumValue();
                if (Lif.Role.CLUSTER_MGMT != lifRole && Lif.Role.NODE_MGMT != lifRole) {
                    final String errorMessage = "target lif " + targetIPString + " role " + lifRole + " is not a "
                            + Lif.Role.CLUSTER_MGMT + " lif.";
                    throw new DataSourceErrorException(DataSourceErrorType.CONNECTING, errorMessage, errorMessage);
                }
                validatedLif = true;
                break;
            }
        }
        
        if (!validatedLif) {
            // How can this be?  KFS simulators don't give a role to validate.
            // We want to allow discovery of KFS simulators
            LOG.info("{} - Failed to validate role of target LIF {}",
                    PackageContext.getLogPrefix(), targetIPString);
        }
        
        return validatedLif;
    }
    
    private static String convertToStandardForm(String address){
        if (address == null || !address.contains(":")) {
            return address;
        }
        
        StringBuilder sb = new StringBuilder();
        String[] splitted = address.trim().toLowerCase().split(":");
        for (String str : splitted) {
            if ("".equals(str)) {
                for (int i = 0; i < 8 - splitted.length + 1; i++) {
                    sb.append("0000:");
                }
            }
            else {
                while (str.length() < 4) {
                    str = "0" + str;
                }
                sb.append(str).append(":");
            }
        }
        
        return sb.substring(0, sb.length()-1);
    }


    /**
     * If snapshot discovery is enabled then the configured poll period must be at least 6 hours, otherwise
     * fail the poll.
     *
     * @throws DataSourceErrorException when snapshot discovery is enabled then the configured poll period
     *                                  is less than 6 hours
     */
    private void assertEnabledSnapshotDiscoveryPollPeriod() throws DataSourceErrorException {
        final NetAppCommonConfig config = getConfig();
        final int pollPeriodMs = config.getPollTimeMillis();
        final int sixHoursInMs = 6 * 60 * 60 * 1000;
        if (pollPeriodMs < sixHoursInMs) {
            final String errorMessage = "Configuration poll period must be at least 6 hours when snapshot discovery is enabled.";
            throw new DataSourceErrorException(DataSourceErrorType.CONFIGURATION, errorMessage, errorMessage);
        }
    }

    public static boolean isSnapshotDiscoveryEnabled() {
        final NetAppCommonConfig config = getConfig();
        return config.isSnapshotDiscoveryEnabled();
    }

    public static boolean isQosDiscoveryEnabled() {
        final NetAppCommonConfig config = getConfig();
        return config.isQOSDiscoveryEnabled();
    }

    static NetAppCommonConfig getConfig() {
        final Package thePackage = PackageContext.getPackage();
        return (NetAppCommonConfig) thePackage.getConfiguration();
    }

    private static final class DiagnosisStatusBuilder extends BuilderBase<CModeCluster> {

        private static final String ZAPI_NAME_DIAGNOSIS_STATUS = "diagnosis-status";

        @SuppressWarnings({"unchecked"})
        @Override
        public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
            final NaElement request = elem("diagnosis-status-get",
                    elem(ZAPINames.DESIRED_ATTRIBUTES,
                            elem(ZAPI_NAME_DIAGNOSIS_STATUS, children(
                                    "status"))));
            final Element resultsElem = zapiConnection.invokeElem(request).getElement();

            final Element attributesElem = resultsElem.getChild("attributes");
            if (attributesElem == null) {
                return;
            }

            final Element diagnosisStatusElem = attributesElem.getChild(ZAPI_NAME_DIAGNOSIS_STATUS);
            final EnumWrapper<Cluster.HmStatus> diagnosisStatus;
            if (diagnosisStatusElem != null) {
                diagnosisStatus = BuilderUtil.getEnumWrapperByValue(
                        Cluster.HmStatus.class, diagnosisStatusElem, "status", Cluster.HmStatus.NOT_MAPPED);
            } else {
                diagnosisStatus = new EnumWrapper<Cluster.HmStatus>(null, null);
            }

            cluster.setDiagnosisStatus(diagnosisStatus);
        }
    }

    @SuppressWarnings({"UnusedDeclaration"})
    private static final class LowOnSpareBuilder extends BuilderBase<CModeCluster> {
        @Override
		public void buildModel(final CModeCluster cModeCluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
            final Element isSpareLowElt = zapiConnection.invokeCommand("aggr-check-spare-low", null).getElement();
            final Boolean isSpareLow = BuilderUtil.parseBooleanFromChild(isSpareLowElt, "is-spare-low");
            cModeCluster.setSpareLow(isSpareLow);
        }
    }

    private boolean isPerformanceEnabled() {
        if (isOCIPerfEnabledByOCI) {
            return true;
        }
        final DataSourceConfiguration dsConfig = dataSource.getConfiguration();
        final PackageConfiguration perfPackageConfig = dsConfig.getPackageConfiguration("netappstorageperformance");
        return perfPackageConfig != null && perfPackageConfig.isEnabled();
    }

    private boolean isOCIPerformanceEnabledByProp() {
        final DataSourceConfiguration dsConfig = dataSource.getConfiguration();
        final PackageConfiguration perfPackageConfig = dsConfig.getPackageConfiguration("netappstorageperformance");
        if (perfPackageConfig == null) {
            return false;
        }
        final String ociePerfEnabledFromProp = perfPackageConfig.getProperties().getProperty("oci.performance.enabled", "false");
        return Boolean.parseBoolean(ociePerfEnabledFromProp);
    }
}
