package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;


import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeCluster;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * A bunch of Node statistics have non-deterministic UUIDs, we build the map
 * of name to UUID here, only when performance is enabled.   The top level class
 * collects UUIDs for "cifs:node", and calls other classes to collect UUIDs
 * for other object names (smb1:node, smb2:node ...)
 */
public class NodeStatsUUIDBuilder extends PerfInstanceBuilder {

    private final boolean ociPerfEnabled;

    public NodeStatsUUIDBuilder(final boolean ociPerfEnabled) {
        super("cifs:node");
        this.ociPerfEnabled = ociPerfEnabled;
    }

    @Override
    public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        if (ociPerfEnabled) {
            // Only build protocol metric UUIDs if OCI performance is enabled
            super.buildModel(cluster, zapiConnection);
            cluster.getCifsNodeUUIDNameMap().putAll(getUUIDNameMap());

            final SMB1NodeStatsBuilder smb1NodeStatsBuilder = new SMB1NodeStatsBuilder();
            smb1NodeStatsBuilder.buildModel(cluster, zapiConnection);
            cluster.getSmb1NodeUUIDNameMap().putAll(smb1NodeStatsBuilder.getUUIDNameMap());

            final SMB2NodeStatsBuilder smb2NodeStatsBuilder = new SMB2NodeStatsBuilder();
            smb2NodeStatsBuilder.buildModel(cluster, zapiConnection);
            cluster.getSmb2NodeUUIDNameMap().putAll(smb2NodeStatsBuilder.getUUIDNameMap());

            final NVSv3NodeStatsBuilder nfsv3NodeStatsBuilder = new NVSv3NodeStatsBuilder();
            nfsv3NodeStatsBuilder.buildModel(cluster, zapiConnection);
            cluster.getNfsv3NodeUUIDNameMap().putAll(nfsv3NodeStatsBuilder.getUUIDNameMap());

            final NVSv4NodeStatsBuilder nfsv4NodeStatsBuilder = new NVSv4NodeStatsBuilder();
            nfsv4NodeStatsBuilder.buildModel(cluster, zapiConnection);
            cluster.getNfsv4NodeUUIDNameMap().putAll(nfsv4NodeStatsBuilder.getUUIDNameMap());

            final NVSv4_1NodeStatsBuilder nfsv4_1NodeStatsBuilder = new NVSv4_1NodeStatsBuilder();
            nfsv4_1NodeStatsBuilder.buildModel(cluster, zapiConnection);
            cluster.getNfsv4_1NodeUUIDNameMap().putAll(nfsv4_1NodeStatsBuilder.getUUIDNameMap());
        }

        // Always build CPU columns
        new CpuColumnBuilder().buildModel(cluster, zapiConnection);
        new WaflCpPhaseTimesColumnBuilder().buildModel(cluster, zapiConnection);
    }

    public static final class SMB1NodeStatsBuilder extends PerfInstanceBuilder {
        public SMB1NodeStatsBuilder() {
            super("smb1:node");
        }
    }

    public static final class SMB2NodeStatsBuilder extends PerfInstanceBuilder {
        public SMB2NodeStatsBuilder() {
            super("smb2:node");
        }
    }

    public static final class NVSv3NodeStatsBuilder extends PerfInstanceBuilder {
        public NVSv3NodeStatsBuilder() {
            super("nfsv3:node");
        }
    }

    public static final class NVSv4NodeStatsBuilder extends PerfInstanceBuilder {
        public NVSv4NodeStatsBuilder() {
            super("nfsv4:node");
        }
    }


    public static final class NVSv4_1NodeStatsBuilder extends PerfInstanceBuilder {
        public NVSv4_1NodeStatsBuilder() {
            super("nfsv4_1:node");
        }
    }

    static final class CpuColumnBuilder extends BuilderBase<CModeCluster> {
        @Override
        public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
            final Map<String, String> args = new LinkedHashMap<String, String>() {{
                put("objectname", "processor");
            }};
            final ZAPIConnection.ZapiResponse response = zapiConnection.invokeCommand("perf-object-counter-list-info", args);
            final Element resultsElt = response.getElement();
            final Element countersElt = resultsElt.getChild("counters");
            if (countersElt == null) {
                return;
            }
            @SuppressWarnings("unchecked")
            final List<Element> counterElts = countersElt.getChildren("counter-info");
            for (final Element counterElt : counterElts) {
                final String counterName = BuilderUtil.getChildTextTrim(counterElt, "name");
                if (!"domain_busy".equals(counterName)) {
                    continue;
                }
                final Element labelsElt = counterElt.getChild("labels");
                if (labelsElt == null) {
                    break;
                }
                final String cpuColumnsCSV = BuilderUtil.getChildTextTrim(labelsElt, "label-info");
                if (cpuColumnsCSV == null) {
                    continue;
                }
                final List<String> cpuColumnsList = Arrays.asList(cpuColumnsCSV.split(","));
                cluster.setCpuColumns(cpuColumnsList);
            }
        }
    }
    
    static final class WaflCpPhaseTimesColumnBuilder extends BuilderBase<CModeCluster> {
        @Override
        public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
            final Map<String, String> args = new LinkedHashMap<String, String>() {{
                put("objectname", "wafl");
            }};
            final ZAPIConnection.ZapiResponse response = zapiConnection.invokeCommand("perf-object-counter-list-info", args);
            final Element resultsElt = response.getElement();
            final Element countersElt = resultsElt.getChild("counters");
            if (countersElt == null) {
                return;
            }
            @SuppressWarnings("unchecked")
            final List<Element> counterElts = countersElt.getChildren("counter-info");
            for (final Element counterElt : counterElts) {
                final String counterName = BuilderUtil.getChildTextTrim(counterElt, "name");
                if (!"cp_phase_times".equals(counterName)) {
                    continue;
                }
                final Element labelsElt = counterElt.getChild("labels");
                if (labelsElt == null) {
                    break;
                }
                final String WaflCpPhaseTimesColumnsCSV = BuilderUtil.getChildTextTrim(labelsElt, "label-info");
                if (WaflCpPhaseTimesColumnsCSV == null) {
                    continue;
                }
                final List<String> waflCpPhaseTimesColumnsList = Arrays.asList(WaflCpPhaseTimesColumnsCSV.split(","));
                cluster.setWaflCpPhaseTimesColumns(waflCpPhaseTimesColumnsList);
            }
        }
    }
}
