package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapper;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeCluster;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeQosVolumeWorkload;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.QosVolumeWorkload;
import com.onaro.util.VersionComparator;
import org.apache.log4j.Logger;
import org.jdom.Element;

import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Collections;

/**
 * Builder class used to call the qos-workload-get-iter ZAPI and create {@link CModeQosVolumeWorkload} objects.
 */
public class QosVolumeWorkloadBuilder extends BuilderBase<CModeCluster> {

    private static final Logger logger = Logger.getLogger(QosVolumeWorkloadBuilder.class);

    public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        new QosWorkloadIterBuilder().buildModel(cluster, zapiConnection);
        new ReadIOColumnBuilder().buildModel(cluster, zapiConnection);
    }

    private class QosWorkloadIterBuilder extends IterBuilderBase {
        public QosWorkloadIterBuilder() {
            super("qos-workload-get-iter", "qos-workload-info", getDefaultMaxRecords(),
                    new LinkedHashMap<String[], String[]>() {
                        {
                            put(new String[] { 
                                    "qos-workload-info" }, 
                                new String[] { 
                                    "workload-class", 
                                    "workload-name",
                                    "workload-uuid",
                                    "volume",
                                    "vserver",
                                    "lun",
                                    "file",
                                    "wid"});
                        }
                    });


        }

        /**
         * This method will be called for each qos-workload-info nodes returned by the ZAPI. Process and create
         * {@link CModeQosVolumeWorkload} objects. This method also filters out unwanted workloads returned by the
         * ZAPI.
         */
        @Override
        protected void buildIterInstance(final CModeCluster cluster, final Element qosWorkloadAttributesElt) {
            final EnumWrapper<QosVolumeWorkload.QosWorkloadClassType> workloadTypeWrapper =
                    BuilderUtil.getEnumWrapperByValue(QosVolumeWorkload.QosWorkloadClassType.class,
                            qosWorkloadAttributesElt, "workload-class",
                            QosVolumeWorkload.QosWorkloadClassType.NOT_MAPPED);
            final String workloadName = BuilderUtil.getChildTextTrim(qosWorkloadAttributesElt, "workload-name");
            final String workloadUuid = BuilderUtil.getChildTextTrim(qosWorkloadAttributesElt, "workload-uuid");
            final String workloadVolume = BuilderUtil.getChildTextTrim(qosWorkloadAttributesElt, "volume");
            final String workloadLun = BuilderUtil.getChildTextTrim(qosWorkloadAttributesElt, "lun");
            final String workloadVserver = BuilderUtil.getChildTextTrim(qosWorkloadAttributesElt, "vserver");
            final String workloadFile = BuilderUtil.getChildTextTrim(qosWorkloadAttributesElt, "file");
            final Long workloadId = BuilderUtil.parseLongFromChild(qosWorkloadAttributesElt, "wid");

            //Workloads starting with _WID are internal workloads that OPM is not interested in.
            if (workloadName.startsWith("_WID")) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Skipping uninteresting QoS workload: " + workloadName);
                }
                return;
            }

            if (workloadUuid == null) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to create QoS workload: " + workloadName
                        + " due to missing uuid field");
                return;
            }

            // Filter LUN related work loads.
            if(workloadLun != null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Skipping LUN related QoS workload: " + workloadName);
                }
                return;
            }
            // Filter VServer related work loads.
            if(workloadVolume == null &&  workloadVserver != null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Skipping VServer related QoS workload: " + workloadName);
                }
                return;
            }
            // Filter File related work loads.
            if(workloadFile != null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Skipping File related QoS workload: " + workloadName);
                }
                return;
            }

            final VersionComparator versionComparator = new VersionComparator();
            final String clusterVersion = cluster.getVersionString();
            final boolean isFullSteam = (versionComparator.compare(clusterVersion, Cluster.FULL_STEAM) > 0);
            // workloads that have wids less than 100 and not equal to 1(System-Default) or 2(User-Default) are identified as diagnostic-workloads or internal workloads
            // skip these internal workloads that are not in INTERNAL_WORKLOADS.
            if(isFullSteam && workloadId != null && (workloadId < 100 && workloadId > 2)
                    && !INTERNAL_WORKLOADS.contains(workloadName)) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Skipping redundant system diagnostic (or internal) QoS workload: " + workloadName + " [wid: " + workloadId.toString() + "]");
                }
                return;
            }

            // Create and add the volume
            final CModeQosVolumeWorkload qosWorkload = cluster.createQosVolumeWorkload(workloadUuid);
            qosWorkload.setWorkloadTypeWrapper(workloadTypeWrapper);
            qosWorkload.setName(workloadName);
            qosWorkload.setVolumeName(workloadVolume);
            qosWorkload.setVserverName(workloadVserver);

            cluster.getQosVolumeWorkloads().add(qosWorkload);
        }

        /**
         * Process the workloads returned by OnTap and add dd the internal workloads not returned by OnTap.
         */
        @Override
        public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
            super.buildModel(cluster, zapiConnection);
            final VersionComparator versionComparator = new VersionComparator();
            final String clusterVersion = cluster.getVersionString();
            final boolean isLessThanFullSteam = (versionComparator.compare(clusterVersion, Cluster.FULL_STEAM) < 0);
            if (isLessThanFullSteam) {
                buildHardCodedWorkloads(cluster);
            }
        }

        /**
         * List of internal workloads not returned by OnTap 8.2. These workloads will be added to the ZAPI results.
         */
        public final List<String> INTERNAL_WORKLOADS = Collections.unmodifiableList(Arrays.asList(new String[]{
                "_RAID",
                "_STORAGE",
                "_WAFL",
                "_WAFL_CP",
                "_TAPE-Backup",
                "_WAFL_SCAN",
                "_SIS_CLONE_SPLIT",
                "_SNAPMIRROR_SLO_PREMIUM",
                "_SNAPMIRROR_SLO_STANDARD",
                "_VOL_MOVE",
                "_BACKGROUND_DELETE",
                "_RAIDIO_READ_DONE",
                "_RA_STRIPE_X",
                "_DEDUP",
                "_SPARSE",
                "_USERSPACE_APPS",
                "_SNAPMIRROR",
        }));

        /**
         * Create and add {@link CModeQosVolumeWorkload} objects for internal workloads, not returned by OnTap.
         */
        private void buildHardCodedWorkloads(final CModeCluster cluster) {
            final EnumWrapper<QosVolumeWorkload.QosWorkloadClassType> workloadTypeWrapper = new EnumWrapper<QosVolumeWorkload.QosWorkloadClassType>("present",
                    QosVolumeWorkload.QosWorkloadClassType.PRESET);
            for (final String internalWorkloadName : INTERNAL_WORKLOADS) {
                final CModeQosVolumeWorkload internalQosWorkload = cluster.createQosVolumeWorkload(internalWorkloadName);
                internalQosWorkload.setName(internalWorkloadName);
                internalQosWorkload.setWorkloadTypeWrapper(workloadTypeWrapper);
                cluster.getQosVolumeWorkloads().add(internalQosWorkload);
            }
        }
    }

    /**
     * Read IO counter has a list of values. This builder calls to perf-object-counter-list-info ZAPI to retrieve
     * the definition for Raid IO Counters.
     */
    private static final class ReadIOColumnBuilder extends BuilderBase<CModeCluster> {
        @Override
        public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
            final Map<String, String> args = new LinkedHashMap<String, String>() {{
                put("objectname", "workload");
            }};
            final ZAPIConnection.ZapiResponse response = zapiConnection.invokeCommand("perf-object-counter-list-info", args);
            final Element resultsElt = response.getElement();
            final Element countersElt = resultsElt.getChild("counters");
            if (countersElt == null) {
                return;
            }
            @SuppressWarnings({"unchecked"})
            final List<Element> counterElts = countersElt.getChildren("counter-info");
            for (final Element counterElt : counterElts) {
                final String counterName = BuilderUtil.getChildTextTrim(counterElt, "name");
                if (!"read_io_type".equals(counterName)) {
                    continue;
                }
                final Element labelsElt = counterElt.getChild("labels");
                if (labelsElt == null) {
                    break;
                }
                final String readIOColumnsCSV = BuilderUtil.getChildTextTrim(labelsElt, "label-info");
                if (readIOColumnsCSV == null) {
                    break;
                }
                final List<String> cpuColumnsList = Arrays.asList(readIOColumnsCSV.split(","));
                cluster.setVolumeWorkloadReadIOColumns(cpuColumnsList);
                break;
            }
        }
    }

}
