package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapper;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.*;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import org.apache.log4j.Logger;
import org.jdom.Element;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;

public class VServerBuilder extends IterBuilderBase {

    @SuppressWarnings({"UnusedDeclaration"})
    private static final Logger logger = Logger.getLogger(VServerBuilder.class);

    @Override
    public void buildModel(final CModeCluster cluster, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        super.buildModel(cluster, zapiConnection);                              // calls vserver-get-iter
        new NetDNSBuilder().buildModel(cluster, zapiConnection);                // calls net-dns-get-iter
        new NetDDNSBuilder().buildModel(cluster, zapiConnection);               // calls net-ddns-get-iter
        new NisBuilder().buildModel(cluster, zapiConnection);                   // calls nis-get-iter
        new LDAPBuilder().buildModel(cluster, zapiConnection);                  // calls ldap-config-get-iter
        new CifsSecurityBuilder().buildModel(cluster, zapiConnection);          // calls cifs-security-get-iter
    }

    public VServerBuilder() {
        super("vserver-get-iter", "vserver-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
            put(new String[]{"vserver-info",}, new String[]{"vserver-name", "vserver-type", "uuid", "root-volume", "state", "operational-state", "max-volumes", "language", "nis-domain", "qos-policy-group", "is-repository-vserver", "snapshot-policy", "vserver-subtype", "ipspace", "is-config-locked-for-changes"});
            put(new String[]{"vserver-info", "allowed-protocols"}, new String[]{"protocol",});
            put(new String[]{"vserver-info", "disallowed-protocols"}, new String[]{"protocol",});
            put(new String[]{"vserver-info", "name-server-switch"}, new String[]{"nsswitch",});
            put(new String[]{"vserver-info", "aggr-list"}, new String[]{"aggr-name",});
        }});
    }

    protected void buildIterInstance(final CModeCluster cluster, final Element vServerInfoElt) {
        final String vServerName = BuilderUtil.getChildTextTrim(vServerInfoElt, "vserver-name");
        final EnumWrapper<VServer.Type> vServerType = BuilderUtil.getEnumWrapperByValue(VServer.Type.class, vServerInfoElt, "vserver-type", VServer.Type.NOT_MAPPED);
        final EnumWrapper<VServer.State> vServerState = BuilderUtil.getEnumWrapperByValue(VServer.State.class, vServerInfoElt, "state", VServer.State.NOT_MAPPED);
        final EnumWrapper<VServer.OperationalState> vServerOperationalState = BuilderUtil.getEnumWrapperByValue(VServer.OperationalState.class, vServerInfoElt, "operational-state", VServer.OperationalState.NOT_MAPPED);
        final EnumWrapper<VServer.Subtype> vServerSubtype = BuilderUtil.getEnumWrapperByValue(VServer.Subtype.class, vServerInfoElt, "vserver-subtype", VServer.Subtype.NOT_MAPPED);
        final String maxVolumesString = BuilderUtil.getChildTextTrim(vServerInfoElt, "max-volumes");

        // Note: defining this as a tertiary results in NPE when (maxVolumesString == null)!
        final Integer maxVolumes;
        if (maxVolumesString != null && "unlimited".equalsIgnoreCase(maxVolumesString)) {
            maxVolumes = -1;
        }
        else {
            maxVolumes = BuilderUtil.parseIntFromChild(vServerInfoElt, "max-volumes");
        }

        final Boolean isRepository = BuilderUtil.parseBooleanFromChild(vServerInfoElt, "is-repository-vserver");
        final String languageRaw = BuilderUtil.getChildTextTrim(vServerInfoElt, "language");
        final EnumWrapper<VServer.Language> language = new EnumWrapper<VServer.Language>(languageRaw, VServer.Language.fromString(languageRaw));
        final String nisDomain = BuilderUtil.getChildTextTrim(vServerInfoElt, "nis-domain");
        final String qosPolicyGroupName = BuilderUtil.getChildTextTrim(vServerInfoElt, "qos-policy-group");
        final String uuID = BuilderUtil.getChildTextTrim(vServerInfoElt, "uuid");
        final String rootVolumeName = BuilderUtil.getChildTextTrim(vServerInfoElt, "root-volume");
        final String snapshotPolicyName = BuilderUtil.getChildTextTrim(vServerInfoElt, "snapshot-policy");
        final String ipSpaceName = BuilderUtil.getChildTextTrim(vServerInfoElt, "ipspace");
        final Boolean isConfigLockedForChanges = BuilderUtil.parseBooleanFromChild(vServerInfoElt, "is-config-locked-for-changes");
        
        final List<String> allowedProtoStrings = BuilderUtil.getListFromGrandChildren(vServerInfoElt, "allowed-protocols", "protocol");
        Collections.sort(allowedProtoStrings);
        final List<EnumWrapper<VServer.Protocol>> allowedProtos = getProtocolListFromStringList(allowedProtoStrings);

        final List<String> disAllowedProtoStrings = BuilderUtil.getListFromGrandChildren(vServerInfoElt, "disallowed-protocols", "protocol");
        Collections.sort(disAllowedProtoStrings);
        final List<EnumWrapper<VServer.Protocol>> disAllowedProtos = getProtocolListFromStringList(disAllowedProtoStrings);

        final List<String> nameServerSwitches = BuilderUtil.getListFromGrandChildren(vServerInfoElt, "name-server-switch", "nsswitch");
        final String nameServerSwitchesCSV = BuilderUtil.getListAsCSVString(nameServerSwitches);

        final List<String> vServerAggrs = BuilderUtil.getListFromGrandChildren(vServerInfoElt, "aggr-list", "aggr-name");
        final String aggregatesCSV = BuilderUtil.getListAsCSVString(vServerAggrs);

        final CModeVServer vServer = cluster.createVServer(uuID);
        vServer.setName(vServerName);
        vServer.setState(vServerState);
        vServer.setOperationState(vServerOperationalState);
        vServer.setType(vServerType);
        vServer.setSubtype(vServerSubtype);
        vServer.setAllowedProtos(allowedProtos);
        vServer.setDisAllowedProtos(disAllowedProtos);
        vServer.setRootVolumeName(rootVolumeName);
        vServer.setMaxVolumes(maxVolumes);
        vServer.setRepository(isRepository);
        vServer.setLanguage(language);
        vServer.setNisDomain(nisDomain);
        vServer.setNameServerSwitches(nameServerSwitchesCSV);
        vServer.setAggregatesCSV(aggregatesCSV);
        vServer.setSnapshotPolicy(snapshotPolicyName);
        vServer.setQosPolicyGroupName(qosPolicyGroupName);
        vServer.setNetworkIpSpaceName(ipSpaceName);
        vServer.setIsConfigLockedForChanges(isConfigLockedForChanges);

        cluster.getVServers().add(vServer);
    }

    private List<EnumWrapper<VServer.Protocol>> getProtocolListFromStringList(final List<String> rawProtocols) {
        final List<EnumWrapper<VServer.Protocol>> protocols = new ArrayList<EnumWrapper<VServer.Protocol>>(rawProtocols.size());
        for (final String rawProto : rawProtocols) {
            protocols.add(BuilderUtil.getEnumWrapperByValue(VServer.Protocol.class, rawProto, VServer.Protocol.NOT_MAPPED));
        }
        return protocols;
    }

    // Get some additional vServer info
    private static final class CifsSecurityBuilder extends IterBuilderBase {
        private CifsSecurityBuilder() {
            super("cifs-security-get-iter", "cifs-security", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"cifs-security",}, new String[]{"is-smb-encryption-required","vserver",});
            }});
        }

        protected void buildIterInstance(final CModeCluster cluster, final Element cifsSecurityElt) {

            final String vServer = BuilderUtil.getChildTextTrim(cifsSecurityElt, "vserver");

            if(vServer != null && vServer.length() > 0) {
                final CModeVServerCIFSSecurityInfo cifsSecInfo = cluster.createVServerCIFSSecurityInfo(vServer);
                String elem = BuilderUtil.getChildText(cifsSecurityElt,"is-smb-encryption-required");
                cifsSecInfo.setSmbEncryptionRequired(elem == null ? null : BuilderUtil.parseBooleanFromString(elem));
                cluster.getVServerCIFSSecInfos().add(cifsSecInfo);
            }
        }
    }

    // Get some additional vServer info
    private static final class NetDNSBuilder extends IterBuilderBase {
        private NetDNSBuilder() {
            super("net-dns-get-iter", "net-dns-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"net-dns-info",}, new String[]{"dns-state", "vserver-name",});
                put(new String[]{"net-dns-info", "domains",}, new String[]{"string",});
                put(new String[]{"net-dns-info", "name-servers",}, new String[]{"ip-address",});
            }});
        }

        protected void buildIterInstance(final CModeCluster cluster, final Element netDnsInfoElt) {
            final String vServerString = BuilderUtil.getChildTextTrim(netDnsInfoElt, "vserver-name");
            final EnumWrapper<VServer.DNSState> dnsState = BuilderUtil.getEnumWrapperByValue(VServer.DNSState.class, netDnsInfoElt, "dns-state", VServer.DNSState.NOT_MAPPED);
            final List<String> domainStrings = BuilderUtil.getListFromGrandChildren(netDnsInfoElt, "domains", "string");
            final String domainsCSV = BuilderUtil.getListAsCSVString(domainStrings);
            final List<String> nameServerStrings = BuilderUtil.getListFromGrandChildren(netDnsInfoElt, "name-servers", "ip-address");
            final String nameServerCSV = BuilderUtil.getListAsCSVString(nameServerStrings);
            final CModeVServerDNSInfo dnsInfo = cluster.createVServerDNSInfo(vServerString);
            dnsInfo.setDnsState(dnsState);
            dnsInfo.setDomainNames(domainsCSV);
            dnsInfo.setNameServers(nameServerCSV);
            cluster.getVServerDNSInfos().add(dnsInfo);
        }
    }

    // Get some additional vServer info
    private static final class NisBuilder extends IterBuilderBase {
        private NisBuilder() {
            super("nis-get-iter", "nis-domain-config-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"nis-domain-config-info",}, new String[]{"is-active", "nis-domain", "vserver",});
                put(new String[]{"nis-domain-config-info", "nis-servers",}, new String[]{"ip-address",});
            }});
        }

        protected void buildIterInstance(final CModeCluster cluster, final Element nisInfoElt) {
            final String vServerString = BuilderUtil.getChildTextTrim(nisInfoElt, "vserver");
            final Boolean isActive = BuilderUtil.parseBooleanFromChild(nisInfoElt, "is-active");
            final String nisDomain = BuilderUtil.getChildTextTrimDefaultEmpty(nisInfoElt, "nis-domain");
            final List<String> nisServerIPs = BuilderUtil.getListFromGrandChildren(nisInfoElt, "nis-servers", "ip-address");
            final String nisServerIpsCSV = BuilderUtil.getListAsCSVString(nisServerIPs);
            final CModeVServerNISInfo nisInfo = cluster.createVServerNISInfo(vServerString, nisDomain);
            nisInfo.setActive(isActive);
            nisInfo.setNisServerIPs(nisServerIpsCSV);
            cluster.getVServerNISInfos().add(nisInfo);
        }
    }

    private static  final class LDAPBuilder extends  IterBuilderBase {
        private LDAPBuilder() {
            super("ldap-config-get-iter", "ldap-config", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                    put(new String[]{"ldap-config",}, new String[]{"client-config", "client-enabled", "vserver",});
                }});
            }

        protected void buildIterInstance(final CModeCluster cluster, final Element nisInfoElt) {
            final String vServerString = BuilderUtil.getChildTextTrim(nisInfoElt, "vserver");
            final Boolean clientEnabled = BuilderUtil.parseBooleanFromChild(nisInfoElt, "client-enabled");
            final String clientConfig = BuilderUtil.getChildTextTrim(nisInfoElt, "client-config");

            if(vServerString != null && vServerString.length() > 0){
                final CModeVServerLDAPInfo ldapInfo = cluster.createVServerLDAPInfo(vServerString);
                ldapInfo.setClientConfig(clientConfig);
                ldapInfo.setClientEnabled(clientEnabled);
                cluster.getVServerLDAPInfos().add(ldapInfo);
            }
        }
    }

    // Get some additional vServer info
    private static final class NetDDNSBuilder extends IterBuilderBase {
        private NetDDNSBuilder() {
            super("net-ddns-get-iter", "net-ddns-info", getDefaultMaxRecords(), new LinkedHashMap<String[], String[]>() {{
                put(new String[]{"net-ddns-info",}, new String[]{"domain-name", "is-enabled", "ttl", "use-secure", "vserver-name",});;
            }});
        }

        protected void buildIterInstance(final CModeCluster cluster, final Element netDDNSInfoElt) {
            final String vServerString = BuilderUtil.getChildTextTrim(netDDNSInfoElt, "vserver-name");
            final String domainName = BuilderUtil.getChildTextTrimDefaultEmpty(netDDNSInfoElt, "domain-name");
            final Boolean isEnabled = BuilderUtil.parseBooleanFromChild(netDDNSInfoElt, "is-enabled");
            final Long ttl = BuilderUtil.parseLongFromChild(netDDNSInfoElt, "ttl");
            final Boolean useSecure = BuilderUtil.parseBooleanFromChild(netDDNSInfoElt, "use-secure");
            final CModeVServerDDNSInfo ddnsInfo = cluster.createVServerDDNSInfo(vServerString);
            ddnsInfo.setDomainName(domainName);
            ddnsInfo.setIsEnabled(isEnabled);
            ddnsInfo.setUseSecure(useSecure);
            ddnsInfo.setTtl(ttl);
            cluster.getVServerDDNSInfos().add(ddnsInfo);
        }
    }

}
