package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.netapp.oci.netapp.performance.qosworkloaddetail.interfaces.data.QosWorkloadDetailMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappQosWorkloadDetailMetrics;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class QosWorkloadDetailMetricsBuilder extends MetricsBuilderBase<Cluster.ClusterKey,
                                                                        QosWorkloadDetailMeasurements,
                                                                        QosWorkloadDetailMeasurements,
                                                                        NetappQosWorkloadDetailMetrics,
                                                                        NetappQosWorkloadDetailMetrics.QosWorkloadDetailCounter,
                                                                        NetappQosWorkloadDetailMetrics.QosWorkloadDetailCounter> {

    private static final Logger LOG = LoggerFactory.getLogger(QosWorkloadDetailMetricsBuilder.class);
    
    private Map<String, QosWorkloadDetailMeasurements> qosWorkloadDetailMeasurements;

    
    public QosWorkloadDetailMetricsBuilder(final NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage, "workload_detail", NetappQosWorkloadDetailMetrics.QosWorkloadDetailCounter.class);
        qosWorkloadDetailMeasurements = new HashMap<String, QosWorkloadDetailMeasurements>();
    }

    @Override
    List<String> getTargetPerfUUIDs(final NetAppDefaultStorageContext context, final ZAPIConnection connection) {
        return context.getQosWorkloadDetailInstanceUUIDs();
    }

    @Override
    NetappQosWorkloadDetailMetrics createMetricsFromValues(final MetricsRowScratchPad<NetappQosWorkloadDetailMetrics.QosWorkloadDetailCounter> currentValues, final MetricsRowScratchPad<NetappQosWorkloadDetailMetrics.QosWorkloadDetailCounter> previousValues) {
        return new NetappQosWorkloadDetailMetrics(currentValues, previousValues);
    }

    @Override
    String getIventoryUUIDFromPerfUUID(final NetAppDefaultStorageContext context, final String perfInstanceUUID) {
        return context.getKey().getID(Cluster.ClusterKey.Field.uuid) + '.' + perfInstanceUUID;
    }


    @Override
    Cluster.ClusterKey getParentInventoryUUIDKey(final NetAppDefaultStorageContext context, final String perfUUID) {
        return context.getKey();
    }
    
    
    public Map<String, QosWorkloadDetailMeasurements> getQosWorkloadDetailMeasurements(){
        return qosWorkloadDetailMeasurements;
    }

    @Override
    public void getAndReportPerformance(final NetAppDefaultStorageContext context, final ZAPIConnection zapiConnection) throws DataSourceErrorException {
        final Cluster.ClusterKey clusterKey = context.getKey();
        final Map<Cluster.ClusterKey, Collection<RawSample<QosWorkloadDetailMeasurements>>> metricsToReport = getMetricsAndReportRawPerformance(context, zapiConnection);
        // clear previous measurements
        qosWorkloadDetailMeasurements.clear();
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosWorkloadDetailMeasurements>>> measurementEntry : metricsToReport.entrySet()) {
            final Collection<RawSample<QosWorkloadDetailMeasurements>> workloadDetailMeasurements = measurementEntry.getValue();
            //Subtract Exempts
            if(context.isLessThanFullSteam()){
            	subtractCpuExempt(workloadDetailMeasurements);
            }
            for (final RawSample<QosWorkloadDetailMeasurements> sample : workloadDetailMeasurements) {
                final String detailUuid = sample.getKey();     // foundation uuid
                final QosWorkloadDetailMeasurements detailMeasurements = sample.getCounters();
                qosWorkloadDetailMeasurements.put(detailUuid, detailMeasurements);
            }
            LOG.debug("Qos workload detail measurements are updated with current poll. Cluster Id: {}", measurementEntry.getKey().toString());
        }
        reportPerformance(clusterKey, metricsToReport);
    }
    
    @Override
    void reportPerformance(Cluster.ClusterKey clusterKey, final Map<Cluster.ClusterKey, Collection<RawSample<QosWorkloadDetailMeasurements>>> measurements) throws DataSourceErrorException {
    	for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosWorkloadDetailMeasurements>>> measurementEntry : measurements.entrySet()) {
    		final Cluster.ClusterKey key = measurementEntry.getKey();
    		final Collection<RawSample<QosWorkloadDetailMeasurements>> workloadDetailMeasurements = measurementEntry.getValue();
    		storagePerformanceReporter.reportQosWorkloadDetailPerformance(key, workloadDetailMeasurements);
    	} 
    }

    /**
     * DE4985 / BURT_756607: For versions less than full steam (8.3), Qos Workload Detail
     * metrics must subtract CPU_exempt service time and wait time from calculated
     * CPU_dblade service time and wait time.
     *
     * For CPU_dblade workload details, ONTAP is aggregating counters, including CPU_exempt.
     * DE4985 / BURT_756607 indicates that ONTAP should not include CPU_exempt, so to
     * workaround we subtract CPU_exempt from CPU_dblade here.
     *
     * This fix introduces a dependency between two rows of the same type of metric (same table).
     * The order of row collection is not deterministic, so this final step in metrics calculation
     * occurs after all rows have been processed and "final" values calculated for each column.
     *
     * An alternative approach is to subtract the raw CPU_exempt values from raw CPU_dblade values.
     * Issues with this approach include:
     *      - counter deltas cooked to rates in acquisition layer
     *      - time skew between collection of raw values in different rows
     *      - poll start estimate requirements require "pre-processing" of raw values
     *      - interdependency between rows of the same table (qos_workload_detail)
     *
     * So we adjust for DE4985 here, after all rows are collected, and all other row processing has
     * completed
     *
     * In Addition to subtracting CPU_exempt from CPU_dblade, we must add CPU_exempt to CPU_dblade_background
     *
     * DE5072: apply same re-allocation strategy to the workloaddetail visits counter.
     *
     * @param workloadDetailMeasurements
     */
    private void subtractCpuExempt(Collection<RawSample<QosWorkloadDetailMeasurements>> workloadDetailMeasurements) {
        // For each CPU_dblade, we need to find the corresponding CPU_exempt, and subtract CPU_exempt from CPU_dblade
        // For each CPU_dblade_background, we need to find the corresponding CPU_exempt, and add CPU_exemtp to CPU_dblade_background
        // Maps of dblade & exempt details for adjusting
        final Map<String, QosWorkloadDetailMeasurements> dBladeDetailsMap = new HashMap<String, QosWorkloadDetailMeasurements>();
        final Map<String, QosWorkloadDetailMeasurements> dBladeBackgroundDetailsMap = new HashMap<String, QosWorkloadDetailMeasurements>();
        final Map<String, QosWorkloadDetailMeasurements> exemptDetailsMap = new HashMap<String, QosWorkloadDetailMeasurements>();
        for (final RawSample<QosWorkloadDetailMeasurements> sample : workloadDetailMeasurements) {
            final String detailKey = sample.getKey();
            final QosWorkloadDetailMeasurements detailMeasurements = sample.getCounters();
            if (detailKey.endsWith(".CPU_exempt")) {
                // Samples used to adjust CPU_dblade, CPU_dblade_background
                exemptDetailsMap.put(detailKey, detailMeasurements);
            } else if (detailKey.endsWith(".CPU_dblade")) {
                // CPU_dblade Samples to be adjusted
                dBladeDetailsMap.put(detailKey, detailMeasurements);
            } else if (detailKey.endsWith(".CPU_dblade_background")) {
                // CPU_dblade_background samples to be adjusted
                dBladeBackgroundDetailsMap.put(detailKey, detailMeasurements);
            }
        }

        // Subtract CPU_exempt from each CPU_dblade
        for (final Map.Entry<String, QosWorkloadDetailMeasurements> cpuDBladeEntry : dBladeDetailsMap.entrySet()) {
            final String dBladeDetailKey = cpuDBladeEntry.getKey();
            final int prefixEndIndex = dBladeDetailKey.length() - "CPU_dblade".length();
            final String cpuExemptKey = dBladeDetailKey.substring(0, prefixEndIndex) + "CPU_exempt";
            final QosWorkloadDetailMeasurements exemptMeasurements = exemptDetailsMap.get(cpuExemptKey);
            if (exemptMeasurements == null) {
                // Can't find cpuExemptKey to adjust cpuDBladeMeasurements
                continue;
            }
            final QosWorkloadDetailMeasurements cpuDBladeMeasurements = cpuDBladeEntry.getValue();
            if (cpuDBladeMeasurements == null) {
                // cpuDBladeMeasurements had a wrap/reset, so there is nothing to subtract from
                continue;
            }
            final long originalServiceTime = cpuDBladeMeasurements.getServiceTime();
            final long exemptServiceTime = exemptMeasurements.getServiceTime();
            cpuDBladeMeasurements.setServiceTime(MetricsRowScratchPad.subtractLong(originalServiceTime, exemptServiceTime));
            final long originalWaitTime = cpuDBladeMeasurements.getWaitTime();
            final long exemptWaitTime = exemptMeasurements.getWaitTime();
            cpuDBladeMeasurements.setWaitTime(MetricsRowScratchPad.subtractLong(originalWaitTime, exemptWaitTime));
        }

        // Add CPU_exempt to each CPU_dblade_background
        for (final Map.Entry<String, QosWorkloadDetailMeasurements> cpuDBladeBackgroundEntry : dBladeBackgroundDetailsMap.entrySet()) {
            final String dBladeBackgroundDetailKey = cpuDBladeBackgroundEntry.getKey();
            final int prefixEndIndex = dBladeBackgroundDetailKey.length() - "CPU_dblade_background".length();
            final String cpuExemptKey = dBladeBackgroundDetailKey.substring(0, prefixEndIndex) + "CPU_exempt";
            final QosWorkloadDetailMeasurements exemptMeasurements = exemptDetailsMap.get(cpuExemptKey);
            if (exemptMeasurements == null) {
                // Can't find cpuExemptKey to adjust cpuDBladeBackgroundMeasurements
                continue;
            }
            final QosWorkloadDetailMeasurements cpuDBladeBackgroundMeasurements = cpuDBladeBackgroundEntry.getValue();
            if (cpuDBladeBackgroundMeasurements == null) {
                // cpuDBladeBackgroundMeasurements had a wrap/reset, so there is nothing to add to
                continue;
            }
            final long originalServiceTime = cpuDBladeBackgroundMeasurements.getServiceTime();
            final long exemptServiceTime = exemptMeasurements.getServiceTime();
            cpuDBladeBackgroundMeasurements.setServiceTime(MetricsRowScratchPad.addLong(originalServiceTime, exemptServiceTime));
            final long originalWaitTime = cpuDBladeBackgroundMeasurements.getWaitTime();
            final long exemptWaitTime = exemptMeasurements.getWaitTime();
            cpuDBladeBackgroundMeasurements.setWaitTime(MetricsRowScratchPad.addLong(originalWaitTime, exemptWaitTime));
        }
    }
}
