package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappstorageperformance;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;



import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappQosWorkloadConstituentMetrics;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappQosWorkloadConstituentMetrics.WorkloadConstituentRowMetricsScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosVolumeWorkloadMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

/**
 * Builder class for computing cluster level workload stats. The data for
 * cluster level workload stats will be aggregated from the workload_constituent
 * object
 */
public class QosWorkloadMetricsBuilder {

    private static final Logger LOG = LoggerFactory
            .getLogger(QosWorkloadMetricsBuilder.class);
    private final QosWorkloadConstituentMetricsBuilder qosWorkloadConstituentMetricsBuilder;
    private final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage;

    public QosWorkloadMetricsBuilder(
            final NetAppOCIEPerformancePackage netAppOCIEPerformancePackage,
            final QosWorkloadConstituentMetricsBuilder qosWorkloadConstituentMetricsBuilder) {
        this.netAppOCIEPerformancePackage = netAppOCIEPerformancePackage;
        this.qosWorkloadConstituentMetricsBuilder = qosWorkloadConstituentMetricsBuilder;
    }
    
    /**
     * Aggregate the data from the workload constituent objects to create cluster level
     * workload measurements.
     * @param context
     * @return collection of volume workload measurements
     */
    public List<RawSample<QosVolumeWorkloadMeasurements>> getMetricsFromConstituents(NetAppDefaultStorageContext context){
        final List<RawSample<QosVolumeWorkloadMeasurements>> rawSamples = new LinkedList<RawSample<QosVolumeWorkloadMeasurements>>();
        
        final Map<String, MetricsRowScratchPad<NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter>> constituentScratchPadsByUUID = qosWorkloadConstituentMetricsBuilder
                .getPerfUUIDToPreviousValues();

        final Map<String, QosWorkloadMetricsScratchPad> workloadScratchPadsByUUID = new HashMap<String, QosWorkloadMetricsScratchPad>();

        // sum up constituent counters on all nodes - grouping by workload and create corresponding workloadMetricsScratchPad  
        for (final Map.Entry<String, MetricsRowScratchPad<NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter>> constituentScratchPadByUUID : constituentScratchPadsByUUID.entrySet()) {

            final String qosWorkloadConstituentUUID = constituentScratchPadByUUID.getKey();
            final WorkloadConstituentRowMetricsScratchPad constituentScratchPad = (WorkloadConstituentRowMetricsScratchPad)constituentScratchPadByUUID.getValue();

            // Check if previous poll data is available. If not available skip aggregation
            if (!constituentScratchPad.isPreviousPollValuesAvailable()) {
                continue;
            }

            final QosWorkloadMetricsScratchPad workloadMetricsScratchPad = getWorkloadMetricsScratchPad(
                    workloadScratchPadsByUUID, qosWorkloadConstituentUUID,
                    constituentScratchPad.getCollectionPeriod());

            if (workloadMetricsScratchPad == null) {
                continue;
            }

            final long readLatency = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.read_latency);
            final long writeLatency = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.write_latency);
            final long latency = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.latency);
            final long readOps = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.read_ops);
            final long writeOps = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.write_ops);
            final long ops = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.ops);
            final long readData = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.read_data);
            final long writeData = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.write_data);
            final long totalData = constituentScratchPad
                    .getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.total_data);
            final long readIOTypeBase = constituentScratchPad
            		.getCounterDelta(NetappQosWorkloadConstituentMetrics.QosWorkloadConstituentCounter.read_io_type_base);
            final long sumOfDiskAndHyaHDD = constituentScratchPad.getSumOfDiskAndHyaHDD();
            
            
            workloadMetricsScratchPad.addReadLatency(readLatency);
            workloadMetricsScratchPad.addWriteLatency(writeLatency);
            workloadMetricsScratchPad.addLatency(latency);
            workloadMetricsScratchPad.addReadOps(readOps);
            workloadMetricsScratchPad.addWriteOps(writeOps);
            workloadMetricsScratchPad.addOps(ops);
            workloadMetricsScratchPad.addReadData(readData);
            workloadMetricsScratchPad.addWriteData(writeData);
            workloadMetricsScratchPad.addTotalData(totalData);
            workloadMetricsScratchPad.addReadIOTypeBase(readIOTypeBase);
            workloadMetricsScratchPad.addSumOfDiskAndHyaHDD(sumOfDiskAndHyaHDD);
            
        }

        // iterate over all of workloadMetricsScratchPads, compute workload metrics data and populate workload measurement objects 
        for (final Map.Entry<String, QosWorkloadMetricsScratchPad> workloadScratchPadByUUID : workloadScratchPadsByUUID
                .entrySet()) {
            final QosWorkloadMetricsScratchPad workloadScratchPad = workloadScratchPadByUUID
                    .getValue();
            if (workloadScratchPad.isZeroSuppressDetected()) {
                continue;
            }
            final String inventoryUUID = context.getKey().getID(
                    Cluster.ClusterKey.Field.uuid)
                    + '.' + workloadScratchPadByUUID.getKey();
            
            final QosVolumeWorkloadMeasurements qosVolumeWorkloadMeasurements = 
                    new QosVolumeWorkloadMeasurements().setReadOps(workloadScratchPad.getReadOpsPerSecond())
                    .setWriteOps(workloadScratchPad.getWriteOpsPerSecond())
                    .setOps(workloadScratchPad.getOpsPerSecond())
                    .setReadLatency(workloadScratchPad.getReadLatencyPerReadOps())
                    .setWriteLatency(workloadScratchPad.getWriteLatencyPerWriteOps())
                    .setLatency(workloadScratchPad.getLatencyPerOps())
                    .setReadData(workloadScratchPad.getReadDataPerSecond())
                    .setWriteData(workloadScratchPad.getWriteDataPerSecond())
                    .setTotalData(workloadScratchPad.getTotalDataPerSecond())
                    .setCacheMissRate(workloadScratchPad.getCacheMissRate())
                    .setCacheMissRateBase(workloadScratchPad.getReadIOTypeBase());

            rawSamples.add(new RawSample<QosVolumeWorkloadMeasurements>(
                    inventoryUUID, qosVolumeWorkloadMeasurements));
        }
        
        return rawSamples;
    }

    public void getAndReportPerformance(NetAppDefaultStorageContext context)
            throws DataSourceErrorException {

        final List<RawSample<QosVolumeWorkloadMeasurements>> rawSamples = getMetricsFromConstituents(context);
        final Map<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> allMeasurementsToReport = new HashMap<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>>(1);
        allMeasurementsToReport.put(context.getKey(), rawSamples);

        reportPerformance(allMeasurementsToReport);
    }

    /**
     * report to server all QosVolumeWorkloadPerformance stats mapped to a cluster key
     * @param qosWorkloadMetrics
     * @throws DataSourceErrorException
     */
    public void reportPerformance(
            final Map<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> qosWorkloadMetrics)
                    throws DataSourceErrorException {
        if (qosWorkloadMetrics == null) {
            return;
        }
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<QosVolumeWorkloadMeasurements>>> measurementEntry : qosWorkloadMetrics
                .entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<QosVolumeWorkloadMeasurements>> workloadMeasurements = measurementEntry
                    .getValue();
            netAppOCIEPerformancePackage.reportQosVolumeWorkloadPerformance(
                    key, workloadMeasurements);
        }

    }

    /**
     * Retrieve the workload uuid from the node workload uuid
     * (node_uuid.workload_uuid) and create a QosWorkloadMetricsScratchPad if
     * not already created.
     */
    private QosWorkloadMetricsScratchPad getWorkloadMetricsScratchPad(
            final Map<String, QosWorkloadMetricsScratchPad> workloadScratchPadByUUID,
            final String nodeWorkloadUUID, final long collectionPeriod) {

        final String nodeWorkloadUUIDBroken[] = nodeWorkloadUUID.split("\\;");
        if (nodeWorkloadUUIDBroken.length < 2) {
            return null;
        }
        final String workloadUUID = nodeWorkloadUUIDBroken[1];
        // get the old workloadMetricsScratchPad if the workload was already
        // found on another node
        QosWorkloadMetricsScratchPad workloadMetricsScratchPad = workloadScratchPadByUUID
                .get(workloadUUID);
        // if we are dealing with a new workload then create new workloadMetricsScratchPad and save it in a hashmap
        if (workloadMetricsScratchPad == null) {
            workloadMetricsScratchPad = new QosWorkloadMetricsScratchPad(
                    collectionPeriod);
            workloadScratchPadByUUID.put(workloadUUID,
                    workloadMetricsScratchPad);
        }
        // Use the longest collection period.
        if (collectionPeriod > workloadMetricsScratchPad.getCollectionPeriod()) {
            workloadMetricsScratchPad.setCollectionPeriod(collectionPeriod);
        }
        return workloadMetricsScratchPad;
    }

    /**
     * QosWorkloadMetricsScratchPad to hold workload raw values while
     * summarizing and normalize the values.
     */
    private static class QosWorkloadMetricsScratchPad {
        private long latency;
        private long readLatency;
        private long writeLatency;
        private long ops;
        private long readOps;
        private long writeOps;
        private long readData;
        private long writeData;
        private long totalData;
        private long readIOTypeBase;
        private long sumOfDiskAndHyaHDD; 
        private long collectionPeriod;

        private QosWorkloadMetricsScratchPad(final long collectionPeriod) {
            this.collectionPeriod = collectionPeriod;
        }

        private void addLatency(final long latency) {
            this.latency = (latency == MetricsRowSample.NULL_LONG) ? this.latency
                    : this.latency + latency;
        }

        private void addOps(final long ops) {
            this.ops = (ops == MetricsRowSample.NULL_LONG) ? this.ops
                    : this.ops + ops;
        }

        private void addTotalData(final long totalData) {
            this.totalData = (totalData == MetricsRowSample.NULL_LONG) ? this.totalData
                    : this.totalData + totalData;
        }

        private long getCollectionPeriod() {
            return collectionPeriod;
        }

        private void setCollectionPeriod(final long collectionPeriod) {
            this.collectionPeriod = collectionPeriod;
        }

        private float getReadLatencyPerReadOps() {
            return getLatencyPerOps(this.readLatency, this.readOps);
        }
        
        private float getCacheMissRate(){
        	return getCacheMissRate(this.sumOfDiskAndHyaHDD, this.readIOTypeBase);
        }

        private void addReadLatency(final long readLatency) {
            this.readLatency = (readLatency == MetricsRowSample.NULL_LONG) ? this.readLatency
                    : this.readLatency + readLatency;
        }

        private float getWriteLatencyPerWriteOps() {
            return getLatencyPerOps(this.writeLatency, this.writeOps);
        }

        private void addWriteLatency(final long writeLatency) {
            this.writeLatency = (writeLatency == MetricsRowSample.NULL_LONG) ? this.writeLatency
                    : this.writeLatency + writeLatency;
        }

        private float getLatencyPerOps() {
            return getLatencyPerOps(this.latency, this.ops);
        }

        private float getReadOpsPerSecond() {
            return getRatePerSecond(readOps);
        }

        private void addReadOps(final long readOps) {
            this.readOps = (readOps == MetricsRowSample.NULL_LONG) ? this.readOps
                    : this.readOps + readOps;
        }

        private float getWriteOpsPerSecond() {
            return getRatePerSecond(writeOps);
        }

        private void addWriteOps(final long writeOps) {
            this.writeOps = (writeOps == MetricsRowSample.NULL_LONG) ? this.writeOps
                    : this.writeOps + writeOps;
        }

        private float getOpsPerSecond() {
            return getRatePerSecond(ops);
        }

        private float getReadDataPerSecond() {
            return getRatePerSecond(this.readData);
        }

        private void addReadData(final long readData) {
            this.readData = (readData == MetricsRowSample.NULL_LONG) ? this.readData
                    : this.readData + readData;
        }

        private float getWriteDataPerSecond() {
            return getRatePerSecond(writeData);
        }

        private void addWriteData(final long writeData) {
            this.writeData = (writeData == MetricsRowSample.NULL_LONG) ? this.writeData
                    : this.writeData + writeData;
        }
        
        private long getReadIOTypeBase(){
        	return readIOTypeBase;
        }
        
        private void addReadIOTypeBase(final long readIOTypeBase){
        	this.readIOTypeBase = (readIOTypeBase == MetricsRowSample.NULL_LONG) ? this.readIOTypeBase
        			: this.readIOTypeBase + readIOTypeBase;
        }
        
        private long getSumOfDiskAndHyaHDD(){
        	return sumOfDiskAndHyaHDD;
        }
        
        private void addSumOfDiskAndHyaHDD(final long sumOfDiskAndHyaHDD){
        	this.sumOfDiskAndHyaHDD = (sumOfDiskAndHyaHDD == MetricsRowSample.NULL_LONG) ? this.sumOfDiskAndHyaHDD
        			: this.sumOfDiskAndHyaHDD + sumOfDiskAndHyaHDD;
        }
        
        public float getTotalDataPerSecond() {
            return getRatePerSecond(totalData);
        }

        private float getRatePerSecond(final long counterDelta) {
            if (counterDelta == MetricsRowSample.NULL_LONG
                    || this.collectionPeriod == MetricsRowSample.NULL_LONG) {
                return MetricsRowSample.NULL_FLOAT;
            }
            final double deltaPerMSDouble = ((double) counterDelta)
                    / ((double) (collectionPeriod));
            final double deltaPerSDouble = deltaPerMSDouble * 1000d;
            return (float) deltaPerSDouble;
        }

        private float getLatencyPerOps(final long latency, final long ops) {
            final float latencyPerOp;
            if (ops > 0) {
                if (latency == MetricsRowSample.NULL_LONG) {
                    latencyPerOp = MetricsRowSample.NULL_FLOAT;
                } else {
                    final double deltaLatencyPerOpDouble = ((double) latency)
                            / ((double) ops);
                    latencyPerOp = (float) deltaLatencyPerOpDouble;
                }
            } else if (ops == 0) {
                latencyPerOp = 0;
            } else {
                latencyPerOp = MetricsRowSample.NULL_FLOAT;
            }
            return latencyPerOp;
        }
        
        private float getCacheMissRate(final long sumOfDiskAndHyaHDD, final long read_io_type_base){
                final float percent = (read_io_type_base == 0 || sumOfDiskAndHyaHDD == 0)
                                      ? 0
                                      : (read_io_type_base > 0 && sumOfDiskAndHyaHDD > 0)
                                            ? (float) (100d * (double) sumOfDiskAndHyaHDD / (double) read_io_type_base)
                                            : MetricsRowSample.NULL_FLOAT;
                   	
        	return percent;
       
        }

        public boolean isZeroSuppressDetected() {
            return !(readOps != 0 || writeOps != 0 || ops != 0
                    || readLatency != 0 || writeLatency != 0 || latency != 0
                    || readData != 0 || writeData != 0 || totalData != 0 || readIOTypeBase != 0 || sumOfDiskAndHyaHDD != 0);
        }
    }

}
