package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEDataSource;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;

import java.util.Properties;

/**
 * Class to represent SANscreen configuration for the NetApp data source
 */
public class NetAppOCIEConfiguration extends NetAppFoundationConfiguration<NetAppOCIEDataSource> implements NetAppCommonConfig {
    /**
     * The port to use to connect to the filer
     */
    private final int port;

    /**
     * The type of connection to use
     */
    private final String connectionType;

    /**
     * Max ZAPI retries
     */
    private final int maxRetries;

    private final boolean snapshotDiscoveryEnabled;

    private final boolean isQOSDiscoveryEnabled;

    private final String numBuilderThreads;

    private final int timeout;

    /**
     * Whether or not this datasource should store cooked data.
     */
    private final boolean isCookedStatisticsPersistenceEnabled;

    /**
     * Whether or not this datasource should store raw data.
     */
    private final boolean isRawStatisticsPersistenceEnabled;
    
    /**
     * Whether or not Qos Volume Workload UUID conversion is needed for this datasource.
     */
    private final boolean isQosVolumeWorkloadUUIDConversionEnabled;

    /**
     * Create a new NetApp Configuration instance
     *
     * @param dataSourceConfiguration the base data source configuration
     * @param metaData                the package meta data
     * @param dataSourceName          the data source name
     * @param dataSourceID            the data source ID
     * @param props                   the set of configuration properties
     *
     * @throws FrameworkException when there is a problem creating the NetApp configuration instance
     */
    public NetAppOCIEConfiguration(final DataSourceConfiguration dataSourceConfiguration, final PackageMetaData metaData, final String dataSourceName, final long dataSourceID, final Properties props) throws FrameworkException {
        super(dataSourceConfiguration, metaData, dataSourceName, dataSourceID, props);
        connectionType = properties.getProperty("connectionType", "HTTP");
        port = Integer.parseInt(properties.getProperty("port", "80"));
        maxRetries = Integer.parseInt(properties.getProperty("maxRetries", "5"));
        snapshotDiscoveryEnabled = Boolean.parseBoolean(properties.getProperty("snapshot.enabled", "false"));
        isQOSDiscoveryEnabled = Boolean.parseBoolean(properties.getProperty("qos.enabled", "false"));
        numBuilderThreads = properties.getProperty("numBuilderThreads", "1");
        timeout = Integer.parseInt(properties.getProperty("timeout", "300"));
        isCookedStatisticsPersistenceEnabled = Boolean.parseBoolean(properties.getProperty("cookedpersistence.enabled", "true"));
        isRawStatisticsPersistenceEnabled = Boolean.parseBoolean(properties.getProperty("rawpersistence.enabled", "false"));
        isQosVolumeWorkloadUUIDConversionEnabled = Boolean.parseBoolean(properties.getProperty("upgrader.1_2_0_0", "false"));
    }

    @Override
    public boolean isMOID() {
        // this is Not actually a MOID data source, however we want to retire unreported originators,
        // which only works when isMOID returns true.
        // The issue here is if someone takes a management LIF IP address down on one cluster,
        // then brings it up on another cluster, we want the old cluster to be removed from
        // the database.
        return true;
    }

    @Override
    public boolean isCrawlSupported() {
        return true;
    }

    /**
     * Get the port to use to connect to the filer
     *
     * @return the port to use to connect to the filer
     */
    @Override
    public int getPort() {
        return port;
    }

    /**
     * Get the type of connection to use
     *
     * @return the type of connection to use
     */
    @Override
    public String getConnectionType() {
        return connectionType;
    }

    @Override
    public int getMaxRetries() {
        return maxRetries;
    }

    @Override
    public boolean isSnapshotDiscoveryEnabled() {
        return snapshotDiscoveryEnabled;
    }

    @Override
    public boolean isQOSDiscoveryEnabled() {
        return isQOSDiscoveryEnabled;
    }

    public String getNumBuilderThreads() {
        return numBuilderThreads;
    }

    public int getNumBuilderThreadsAsInt() {
        try {
            return Integer.parseInt(numBuilderThreads);
        } catch (NumberFormatException nfe) {
            return -1;
        }
    }

    public int getTimeout() {
        return timeout;
    }
    

    @Override
    public boolean isCookedStatisticsPersistenceEnabled() {
        return isCookedStatisticsPersistenceEnabled;
    }

    @Override
    public boolean isRawStatisticsPersistenceEnabled() {
        return isRawStatisticsPersistenceEnabled;
    }
    
    public boolean isQosVolumeWorkloadUUIDConversionEnabled() {
        return isQosVolumeWorkloadUUIDConversionEnabled;
    }
}
