package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import java.math.BigInteger;
import java.util.*;


public class GlobalScratchPad {

	/**
	 * Map of counter object name to a map of performance uuid to MetricsRowScratchPad e.g.
	 * "resource" => {"89898012-1212-21212"-> MetricsRowScratchPad}
	 */
	private Map<String, Map<String, ? extends MetricsRowScratchPad<? extends Enum<?>>>> globalScratchPad = new HashMap<>();

	public Map<String, ? extends MetricsRowScratchPad<? extends Enum<?>>> put(String counterObjectName,
			Map<String, ? extends MetricsRowScratchPad<? extends Enum<?>>> counterObjectScratchPad) {
		return globalScratchPad.put(counterObjectName, counterObjectScratchPad);
	}

	public Map<String, ? extends MetricsRowScratchPad<? extends Enum<?>>> get(String counterObject) {
		return globalScratchPad.get(counterObject);
	}

	/**
	 * Merging the counter values in each counter object specified by the input counterObjectNames.
	 * 
	 * @param mergeOperator      worker of the merging 
	 * @param counterObjectNames the counter objects will be merged together
	 * @return the merged result
	 */
	public <T extends Enum<T>> Map<String, MetricsRowScratchPad<T>> merge(final MergeOperator<T> mergeOperator,
			final String ... counterObjectNames) {
		final List<String> availableCounterObjNames = new ArrayList<>(Arrays.asList(counterObjectNames));
		for (String counterObjectName : counterObjectNames) {
			if (!globalScratchPad.containsKey(counterObjectName)) {
				availableCounterObjNames.remove(counterObjectName);
			}
		}
		//None of the source counters are available, return an empty map
		if (availableCounterObjNames.size() == 0) {
			return Collections.emptyMap();
		}

		final Map<String, ? extends MetricsRowScratchPad<? extends Enum<?>>> outerCounterObject = globalScratchPad.get(availableCounterObjNames.get(0));
		Map<String, MetricsRowScratchPad<T>> results = new HashMap<>(outerCounterObject.size());


		for (String counterObjectName : availableCounterObjNames) {
			mergeOperator.mergeScratchPad(results, counterObjectName, globalScratchPad.get(counterObjectName));
		}

		return results;
	}

	/**
	 * Returns raw values map of counter name -> sum of counter Names across different perfUUID's
	 * @param counterObjectName
	 * @param perfUUIDs
	 * @param counterNamesArgs
	 * @param <T>
	 * @return
	 */
	public <T extends Enum<T>> Map<T, BigInteger> addByCountersRaw(final String counterObjectName, List<String> perfUUIDs, T... counterNamesArgs) {

		final List<T> counterNames = new ArrayList<>(Arrays.asList(counterNamesArgs));
		Map<T, BigInteger> results = new HashMap<>();
		Map<String, MetricsRowScratchPad<T>> perfScratchPadMap =
				(Map<String, MetricsRowScratchPad<T>>) globalScratchPad.get(counterObjectName);

		if(perfUUIDs != null && perfScratchPadMap != null) {
			// For each perUUID we get its scratchpad and compute its sum for a specific counter name
			for (String perfUUID : perfUUIDs) {
				MetricsRowScratchPad<T> perfUUIDScratchPad = perfScratchPadMap.get(perfUUID);

				if(perfUUIDScratchPad != null) {
					for (T counterName : counterNames) {
						BigInteger counterValue = perfUUIDScratchPad.getCounterValues().get(counterName);
						BigInteger value = results.get(counterName);
						results.put(counterName, MetricsRowScratchPad.add(value, counterValue));
					}
				}
			}
		}
		return results;
	}

	/**
	 * Returns a map of counter name -> sum of counter Names across different perfUUID's
	 * @param counterObjectName
	 * @param perfUUIDs
	 * @param counterNamesArgs
	 * @param <T>
	 * @return
	 */
	public <T extends Enum<T>> Map<T, Float> addCounterRatePerSec(final String counterObjectName, List<String> perfUUIDs, T... counterNamesArgs) {

		final List<T> counterNames = new ArrayList<>(Arrays.asList(counterNamesArgs));
		Map<T, Float> results = new HashMap<>();
		Map<String, MetricsRowScratchPad<T>> perfScratchPadMap =
				(Map<String, MetricsRowScratchPad<T>>)globalScratchPad.get(counterObjectName);


		// Set results to null
		for(T counterName : counterNames) {
			results.put(counterName, MetricsRowSample.NULL_FLOAT);
		}

		if(perfUUIDs != null && perfScratchPadMap != null) {
			// For each perUUID we get its scratchpad and compute its sum for a specific counter name
			for (String perfUUID : perfUUIDs) {
				MetricsRowScratchPad<T> perfUUIDScratchPad =  perfScratchPadMap.get(perfUUID);

				if (perfUUIDScratchPad != null && perfUUIDScratchPad.getPreviousScratchPad() != null) {
					for (T counterName : counterNames) {
						float counterDelta = perfUUIDScratchPad.getCounterRatePerSec(counterName);
						Float value = results.get(counterName);
						results.put(counterName,
								MetricsRowScratchPad.add(value, counterDelta));
					}
				}
			}
		}

		return results;
	}

	/**
	 * Retrieve scratchPads specified by the input counterObjectName, then group the scratchPads by the input counterName.
	 * Roll up the counter values in each group and return them as the merged result.
	 * 
	 * @param counterClass      the counter class
	 * @param counterObjectName the counter object name where the merging retrieves data from
	 * @param counterName       the counter name that merging is based on 
	 * @return the merged result
	 */
	@SuppressWarnings("unchecked")
	public <T extends Enum<T>> Map<String, MetricsRowScratchPad<T>> merge(Class<T> counterClass, final String counterObjectName, T counterName) {

		final Map<String, MetricsRowScratchPad<T>> scratchPads = (Map<String, MetricsRowScratchPad<T>>)globalScratchPad.get(counterObjectName);
		if (scratchPads == null) {
			return Collections.emptyMap();
		}

		Map<String, MetricsRowScratchPad<T>> results = new HashMap<>();
		for(MetricsRowScratchPad<T> scratchPad : scratchPads.values()) {
			String keyValue = scratchPad.rawValues.get(counterName);
			if(!results.containsKey(keyValue)) {
				// Create new and add to results
				final MetricsRowScratchPad<T> prevScratchPad;
				if (scratchPad.previousValues != null) {
					prevScratchPad = new MetricsRowScratchPad<>(counterClass, scratchPad.previousValues.getPerfInstanceUUID(),
							scratchPad.previousValues.getInventoryUUID(), scratchPad.previousValues.getServerPollStartTimeStamp(),
                            scratchPad.previousValues.getPollStartTimeStamp(), null, scratchPad.previousValues.metricsDebugEnabled);
				} else {
					prevScratchPad = null;
				}
				final MetricsRowScratchPad<T> currScratchPad = new MetricsRowScratchPad<>(counterClass,
						scratchPad.getPerfInstanceUUID(), scratchPad.getInventoryUUID(), scratchPad.getServerPollStartTimeStamp(),
                        scratchPad.getPollStartTimeStamp(), prevScratchPad, scratchPad.metricsDebugEnabled);
				results.put(keyValue, currScratchPad);

			}

			// populate previous values to scratchPad from individual scratchPad
			final MetricsRowScratchPad<T> prevScratchPad = results.get(keyValue).getPreviousScratchPad();
			if (prevScratchPad != null && scratchPad.previousValues != null) {
				prevScratchPad.addToScratchPad(scratchPad.previousValues);	
			}            

			// populate current values to scratchPad from individual scratchPad    		
			results.get(keyValue).addToScratchPad(scratchPad);
		}

		return results;    
	}

	/**
	 * The following methods return the maximum value of the counter after applying the recipe on it.
	 * @param performanceUUIDs - List of performanceUUIDs, whose counters we need to compare
	 * @param counterObjectName - Counter object name whose scratchpads have the counter
	 * @param counterNames - list of counter names whose maximum is to be returned
	 * @return - Map<T,Long> - A map containing the counter and maximum value of the counter.
	 */

	public <T extends Enum<T>>Map<T,Long> getMaximumOfCounterDeltas(final List<String> performanceUUIDs, final String counterObjectName, T...counterNames){
		Map<String, MetricsRowScratchPad<T>> scratchPadsMap =  (Map<String, MetricsRowScratchPad<T>>) globalScratchPad.get(counterObjectName);
		Map<T, Long> counterNameToMaxValue = new HashMap<T,Long>();
		if(scratchPadsMap != null){
			for(T eachCounter : counterNames){
				Long max = MetricsRowSample.NULL_LONG;
				for(String eachPerformanceUUID : performanceUUIDs){
					MetricsRowScratchPad<T> eachScratchPadMap = scratchPadsMap.get(eachPerformanceUUID);
					if(eachScratchPadMap == null){
						continue;
					}
					Long currentDelta = eachScratchPadMap.getCounterDelta(eachCounter);
					max = (currentDelta > max) ? currentDelta : max;
				}
				counterNameToMaxValue.put(eachCounter, max);
			}
		}
		return counterNameToMaxValue;
	}

	public <T extends Enum<T>>Map<T,Float> getMaximumOfCounterRates(final List<String> performanceUUIDs, final String counterObjectName, T...counterNames){
		Map<String, MetricsRowScratchPad<T>> scratchPadsMap =  (Map<String, MetricsRowScratchPad<T>>) globalScratchPad.get(counterObjectName);
		Map<T, Float> counterNameToMaxValue = new HashMap<T,Float>();
		if(scratchPadsMap != null){
			for(T eachCounter : counterNames){
				Float max = MetricsRowSample.NULL_FLOAT;
				for(String eachPerformanceUUID : performanceUUIDs){
					MetricsRowScratchPad<T> eachScratchPadMap = scratchPadsMap.get(eachPerformanceUUID);
					if(eachScratchPadMap == null){
						continue;
					}
					Float currentRate = eachScratchPadMap.getCounterRatePerSec(eachCounter);
					max = (currentRate > max) ? currentRate : max;
				}
				counterNameToMaxValue.put(eachCounter, max);
			}    	
		}
		return counterNameToMaxValue;
	}


	public <T extends Enum<T>>Map<T,Float> getMaximumOfPercentDeltas(final List<String> performanceUUIDs, final String counterObjectName, List<T> numeratorCounterNames, 
			List<T> denominatorCounterNames){
		Map<String, MetricsRowScratchPad<T>> scratchPadsMap =  (Map<String, MetricsRowScratchPad<T>>) globalScratchPad.get(counterObjectName);
		Map<T, Float> counterNameToMaxValue = new HashMap<T,Float>();
		if(scratchPadsMap != null && numeratorCounterNames.size() == denominatorCounterNames.size()){
			for(int i=0 ; i < numeratorCounterNames.size() ; i++){
				Float max = MetricsRowSample.NULL_FLOAT;
				for(String eachPerformanceUUID : performanceUUIDs){
					MetricsRowScratchPad<T> eachScratchPadMap = scratchPadsMap.get(eachPerformanceUUID);
					if(eachScratchPadMap == null){
						continue;
					}
					Float currentPercent = eachScratchPadMap.getPercentFromDeltas(numeratorCounterNames.get(i), denominatorCounterNames.get(i));
					max = (currentPercent > max) ? currentPercent : max;
				}
				counterNameToMaxValue.put(numeratorCounterNames.get(i), max);
			}
		}
		return counterNameToMaxValue;
	}



	public <T extends Enum<T>>Map<T,Float> getMaximumOfLatencyOps(final List<String> performanceUUIDs, final String counterObjectName, List<T> numeratorCounterNames, 
			List<T> denominatorCounterNames){
		Map<String, MetricsRowScratchPad<T>> scratchPadsMap =  (Map<String, MetricsRowScratchPad<T>>) globalScratchPad.get(counterObjectName);
		Map<T, Float> counterNameToMaxValue = new HashMap<T,Float>();
		if(scratchPadsMap != null && numeratorCounterNames.size() == denominatorCounterNames.size()){
			for(int i=0 ; i < numeratorCounterNames.size() ; i++){
				Float max = MetricsRowSample.NULL_FLOAT;
				for(String eachPerformanceUUID : performanceUUIDs){
					MetricsRowScratchPad<T> eachScratchPadMap = scratchPadsMap.get(eachPerformanceUUID);
					if(eachScratchPadMap == null){
						continue;
					}
					Float currentLatencyPerOp = eachScratchPadMap.getLatencyPerOp(numeratorCounterNames.get(i), denominatorCounterNames.get(i));
					max = (currentLatencyPerOp > max) ? currentLatencyPerOp : max;
				}
				counterNameToMaxValue.put(numeratorCounterNames.get(i), max);
			}
		}
		return counterNameToMaxValue;
	}

    /**
     *  This method returns Map<perfUuid, ScratchPad> for the counter which has Max Latency/OP.
     * @param performanceUUIDs       List of performanceUUIDs, whose counters we need to compare
     * @param counterObjectName     Counter object name whose scratchpads have the counter
     * @param numeratorCounterNames  list of counter names showing as numerator
     * @param denominatorCounterNames  list of counter names showing as denominator
     * @param <T>
     * @return  Map<T,MetricsRowScratchPad<T>> - A map containing its scratchPad having the counter and maximum value of the counter.
     */
    public <T extends Enum<T>> Map<T,MetricsRowScratchPad<T>> getScratchPadsForMaximumLatencyOps(final List<String> performanceUUIDs, final String counterObjectName, List<T> numeratorCounterNames,
                                                                  List<T> denominatorCounterNames){
        Map<String, MetricsRowScratchPad<T>> scratchPadsMap =  (Map<String, MetricsRowScratchPad<T>>) globalScratchPad.get(counterObjectName);
        Map<T, MetricsRowScratchPad<T>> counterNameToMaxScratchPad = new HashMap<>();
        if(scratchPadsMap != null && numeratorCounterNames.size() == denominatorCounterNames.size()){
            for(int i=0 ; i < numeratorCounterNames.size() ; i++){
                Float max = MetricsRowSample.NULL_FLOAT;
                MetricsRowScratchPad<T> scratchPad = null;
                for(String eachPerformanceUUID : performanceUUIDs){
                    MetricsRowScratchPad<T> eachScratchPad = scratchPadsMap.get(eachPerformanceUUID);
                    if(eachScratchPad == null){
                        continue;
                    }
                    Float currentLatencyPerOp = eachScratchPad.getLatencyPerOp(numeratorCounterNames.get(i), denominatorCounterNames.get(i));
                    if(currentLatencyPerOp > max)  {
                        max = currentLatencyPerOp;
                        scratchPad = eachScratchPad;
                    }
                }
                counterNameToMaxScratchPad.put(numeratorCounterNames.get(i), scratchPad);
            }
        }
        return counterNameToMaxScratchPad;
    }


	public void clear() {
		globalScratchPad.clear();
	}
}
