package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import java.util.Map;

/**
 * MergeOperator is a call back interface to determine how to merge scratchpads and the counters in the scratchpad. These
 * are used by the merge methods in the GlobalScratchPad
 *
 * @param <T> The type of the merge result scratchpad. This is a Enum with the list of counters in the result scratchpad.
 */
public interface MergeOperator<T extends Enum<T>> {
    /**
     * This method will be called while executing a merge. Here the implementer will determine which scratchpads
     * from mergeResults & mergeWithScratchPad will be merged. This method will be called once for each counter object
     * in the merge. The implementer will call the mergeCounters method to merge the individual counter values in the
     * scratchpad.
     *
     * For e.g. when merging three counter objects o1, o2 & o3 with scratch pads s1[], s2[] & s3[] this method will be
     * called three times. In the first call an empty mergeResults and s1[] will be passed. The implementer will
     * determine witch scratch pads in s1[] will be included in the merge results. New result scratch pad will be
     * created and the mergeCounters will be called to populate the new scratch pad. The new scratch pad will then
     * be add to the mergeResults. In the second call the mergeResults populated during the first call and s2[] will be
     * passed and so on.
     *
     * @param mergeResults               Contains the merge results.
     * @param mergeWithCounterObjectName Name of the counter object to merge with.
     * @param mergeWithScratchPad        Scratchpad to merge with.
     */
    public void mergeScratchPad(Map<String, MetricsRowScratchPad<T>> mergeResults, String mergeWithCounterObjectName,
                               Map<String, ? extends MetricsRowScratchPad<? extends Enum<?>>> mergeWithScratchPad);

    /**
     * This method will be called to merge the merge results. The implementer will use mergeWithScratchPad and
     * update the resultScratchPad.
     *
     * @param resultScratchPad The scratch pad to be updated with the merge results.
     * @param mergeWithCounterObjectName Name of the counter object to merge with.
     * @param mergeWithScratchPad Scratchpad to merge with.
     * @return The merged scratch pad.
     */
    public MetricsRowScratchPad<T> mergeCounters(MetricsRowScratchPad<T> resultScratchPad, String mergeWithCounterObjectName,
                             MetricsRowScratchPad<? extends Enum<?>> mergeWithScratchPad);
}
