package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappDiskMetrics.DiskCounter.*;

import java.math.BigInteger;
import java.util.Map;

import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskMeasurements;
import com.netapp.oci.netapp.performance.disk.interfaces.data.DiskRawMeasurements;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.builders.DiskMetricsBuilder.DiskCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.archiver.metrics.ArchiveMetricsScratchPad;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

public class NetappDiskMetrics extends MetricsRowSample<DiskMeasurements, DiskRawMeasurements, NetappDiskMetrics.DiskCounter,        // raw
                                                                          NetappDiskMetrics.DiskCounter> {      // final
    
    public static enum DiskCounter {
        base_for_disk_busy,
        cp_read_blocks,
        cp_read_latency,
        cp_reads,
        disk_busy,
        guarenteed_read_blocks,
        guarenteed_read_latency,
        guarenteed_reads,
        guarenteed_write_blocks,
        guarenteed_write_latency,
        guarenteed_writes,
        total_transfers,
        user_read_blocks,
        user_read_latency,
        user_reads,
        user_write_blocks,
        user_write_latency,
        user_writes, }

    public NetappDiskMetrics(final MetricsRowScratchPad<DiskCounter> currentValues, final MetricsRowScratchPad<DiskCounter> previousValues) {
        super(DiskCounter.class, currentValues, previousValues);
    }

    @Override
    DiskCounter getDenominatorForNumerator(final DiskCounter numeratorName) {
        final DiskCounter denominatorName;
        switch (numeratorName) {
            case disk_busy:
                denominatorName = base_for_disk_busy;
                break;
            case cp_read_latency:
                denominatorName = cp_reads;
                break;
            case guarenteed_read_latency:
                denominatorName = guarenteed_reads;
                break;
            case guarenteed_write_latency:
                denominatorName = guarenteed_writes;
                break;
            case user_read_latency:
                denominatorName = user_reads;
                break;
            case user_write_latency:
                denominatorName = user_writes;
                break;
            default:
                denominatorName = null;
        }
        return denominatorName;
    }

    @Override
    float getNormalizedFloatMetric(final DiskCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case cp_read_blocks:
            //case cp_read_chain:
            case cp_reads:
            case guarenteed_read_blocks:
            case guarenteed_reads:
            case guarenteed_write_blocks:
            case guarenteed_writes:
            case total_transfers:
            case user_read_blocks:
            //case user_read_chain:
            case user_reads:
            case user_write_blocks:
            //case user_write_chain:
            case user_writes:
                normalizedMetric = currentValues.getCounterRatePerSec(counterName);
                break;
            case cp_read_latency:
            case guarenteed_read_latency:
            case guarenteed_write_latency:
            case user_read_latency:
            case user_write_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName,
                                                                 getDenominatorForNumerator(counterName));
                break;
            case disk_busy:
                normalizedMetric = currentValues.getPercentFromDeltas(counterName,
                                                                      getDenominatorForNumerator(counterName));
                break;
            default:
                normalizedMetric = NULL_FLOAT;
        }
        return normalizedMetric;
    }

    @Override
    public DiskMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        final DiskMeasurements m = new DiskMeasurements();
        m.setCpReadBlocks(getNormalizedFloatMetric(cp_read_blocks));
        m.setCpReadLatency(getNormalizedFloatMetric(cp_read_latency));
        m.setCpReads(getNormalizedFloatMetric(cp_reads));
        m.setBaseForDiskBusy(currentValues.getCounterDeltaFinal(base_for_disk_busy));
        m.setDiskBusy(getNormalizedFloatMetric(disk_busy));
        m.setGuaranteedReadBlocks(getNormalizedFloatMetric(guarenteed_read_blocks));
        m.setGuaranteedReadLatency(getNormalizedFloatMetric(guarenteed_read_latency));
        m.setGuaranteedReads(getNormalizedFloatMetric(guarenteed_reads));
        m.setGuaranteedWriteBlocks(getNormalizedFloatMetric(guarenteed_write_blocks));
        m.setGuaranteedWriteLatency(getNormalizedFloatMetric(guarenteed_write_latency));
        m.setGuaranteedWrites(getNormalizedFloatMetric(guarenteed_writes));
        m.setTotalTransfers(getNormalizedFloatMetric(total_transfers));
        m.setUserReadBlocks(getNormalizedFloatMetric(user_read_blocks));
        m.setUserReadLatency(getNormalizedFloatMetric(user_read_latency));
        m.setUserReads(getNormalizedFloatMetric(user_reads));
        m.setUserWriteBlocks(getNormalizedFloatMetric(user_write_blocks));
        m.setUserWriteLatency(getNormalizedFloatMetric(user_write_latency));
        m.setUserWrites(getNormalizedFloatMetric(user_writes));
        m.setTotalOps(MetricsRowScratchPad.add(m.getUserReads(), m.getUserWrites(), m.getCpReads()));
		final float totalOpsDelta = MetricsRowScratchPad.add(currentValues.getCounterDelta(DiskCounter.user_reads), currentValues.getCounterDelta(DiskCounter.user_writes), 
				currentValues.getCounterDelta(DiskCounter.cp_reads));
        m.setOpmDiskThroughput(MetricsRowScratchPad.add(m.getUserReadBlocks(), m.getUserWriteBlocks(), m.getCpReadBlocks()));
        m.setOpmDiskLatency(getLatency(DiskCounter.user_read_latency, DiskCounter.user_write_latency, DiskCounter.cp_read_latency, totalOpsDelta));
        return m;
    }

    @Override
    public DiskRawMeasurements rawMetrics(final NetAppDefaultStorageContext context) {
        Map<DiskCounter,BigInteger> counterValues = getCurrentValues().getCounterValues();
        return createDiskRawMeasurements(counterValues);
    }
    
    protected static DiskRawMeasurements createDiskRawMeasurements(Map<DiskCounter,BigInteger> counterValues){
        final DiskRawMeasurements m = new DiskRawMeasurements();
        m.setBaseForDiskBusy(counterValues.get(base_for_disk_busy));
        m.setCpReadBlocks(counterValues.get(cp_read_blocks));
        m.setCpReadLatency(counterValues.get(cp_read_latency));
        m.setCpReads(counterValues.get(cp_reads));
        m.setDiskBusy(counterValues.get(disk_busy));
        m.setGuarenteedReadBlocks(counterValues.get(guarenteed_read_blocks));
        m.setGuarenteedReadLatency(counterValues.get(guarenteed_read_latency));
        m.setGuarenteedReads(counterValues.get(guarenteed_reads));
        m.setGuarenteedWriteBlocks(counterValues.get(guarenteed_write_blocks));
        m.setGuarenteedWriteLatency(counterValues.get(guarenteed_write_latency));
        m.setGuarenteedWrites(counterValues.get(guarenteed_writes));
        m.setTotalTransfers(counterValues.get(total_transfers));
        m.setUserReadBlocks(counterValues.get(user_read_blocks));
        m.setUserReadLatency(counterValues.get(user_read_latency));
        m.setUserReads(counterValues.get(user_reads));
        m.setUserWriteBlocks(counterValues.get(user_write_blocks));
        m.setUserWriteLatency(counterValues.get(user_write_latency));
        m.setUserWrites(counterValues.get(user_writes));
        m.setOpmDiskThroughput(MetricsRowScratchPad.add(m.getUserReadBlocks(), m.getUserWriteBlocks(), m.getCpReadBlocks()));
        m.setTotalOps(MetricsRowScratchPad.add(m.getUserReads(), m.getUserWrites(), m.getCpReads()));
        m.setOpmDiskLatency(MetricsRowScratchPad.add(m.getUserReadLatency(), m.getUserWriteLatency(), m.getCpReadLatency()));
        return m;
    }
    
	public float getLatency(DiskCounter userReadLatency, 
			DiskCounter userWriteLatency, DiskCounter cpReadLatency, float totalOps){
		
		long userReadLatencyDelta = currentValues.getCounterDelta(userReadLatency);
		long userWriteLatencyDelta = currentValues.getCounterDelta(userWriteLatency);
		long cpReadLatencyDelta = currentValues.getCounterDelta(cpReadLatency);
		
		long totalLatency = MetricsRowScratchPad.add(userReadLatencyDelta, userWriteLatencyDelta, cpReadLatencyDelta);

		float weightedLatency = 0;
		if(totalLatency == MetricsRowSample.NULL_LONG || totalOps == MetricsRowSample.NULL_FLOAT || totalOps == 0){
			return MetricsRowSample.NULL_FLOAT;
		}

		weightedLatency = (float) ((double)totalLatency / (double)totalOps) ;
		return weightedLatency;

	}
}
