package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.netapp.oci.netapp.performance.fcpport.interfaces.data.FcpPortMeasurements;
import com.netapp.oci.netapp.performance.fcpport.interfaces.data.FcpPortRawMeasurements;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.math.BigInteger;
import java.util.List;
import java.util.Map;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappFcpPortMetrics.FcpPortCounter.*;

public class NetappFcpPortMetrics extends MetricsRowSample<FcpPortMeasurements, FcpPortRawMeasurements, NetappFcpPortMetrics.FcpPortCounter,     // raw
                                                                                NetappFcpPortMetrics.FcpPortCounter> {   // final

    private static final long MEGABIT_1 = 1000000;

    public static enum FcpPortCounter {
    	avg_other_latency,
    	avg_read_latency,
    	avg_write_latency,
        other_ops,
        read_data,
        read_ops,
        write_data,
        write_ops, }

    private NetAppOCIEPerformancePackage reportingPackage;

    public NetappFcpPortMetrics(final MetricsRowScratchPad<FcpPortCounter> currentValues,
                                final MetricsRowScratchPad<FcpPortCounter> previousValues,
                                NetAppOCIEPerformancePackage reportingPackage) {
        super(FcpPortCounter.class, currentValues, previousValues);
        this.reportingPackage = reportingPackage;
    }

    FcpPortCounter getDenominatorForNumerator(final FcpPortCounter numeratorName) {
        final FcpPortCounter denominatorName;
        switch (numeratorName) {
            case avg_other_latency:
                denominatorName = other_ops;
                break;
            case avg_read_latency:
                denominatorName = read_ops;
                break;
            case avg_write_latency:
                denominatorName = write_ops;
                break;
            default:
                denominatorName = null;
        }

        return denominatorName;
    }
    
    @Override
    float getNormalizedFloatMetric(final FcpPortCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case read_data:
            case write_data:
            case other_ops:
            case read_ops:
            case write_ops:
                normalizedMetric = currentValues.getCounterRatePerSec(counterName);
                break;
            case avg_other_latency:
            case avg_read_latency:
            case avg_write_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName,
                                                                 getDenominatorForNumerator(counterName));
                break;                
            default:
                normalizedMetric = NULL_FLOAT;
        }
        return normalizedMetric;
    }

    @Override
    public FcpPortMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        final FcpPortMeasurements m = new FcpPortMeasurements();
        m.setOtherOps(getNormalizedFloatMetric(other_ops));
        m.setReadData(getNormalizedFloatMetric(read_data));
        m.setReadOps(getNormalizedFloatMetric(read_ops));
        m.setWriteData(getNormalizedFloatMetric(write_data));
        m.setWriteOps(getNormalizedFloatMetric(write_ops));
        m.setAvgOtherLatency(getNormalizedFloatMetric(avg_other_latency));
        m.setAvgReadLatency(getNormalizedFloatMetric(avg_read_latency));
        m.setAvgWriteLatency(getNormalizedFloatMetric(avg_write_latency));
        m.setTotalOps(MetricsRowScratchPad.add(
                m.getReadOps(), m.getWriteOps(), m.getOtherOps()));

        //  Compute (read/write)_throughput at this Port
        //  We pass the fcpLifuuid list to addCounters method which aggregates the counters  at globalScratchPad
        //  and returns its sum
        final String fcpPortWWN = BuilderUtil.normalizeWWN(currentValues.getPerfInstanceUUID());
        List<String> fcpLifUUIDList = context.getFcpPortWWNToLifUUIDMap().get(fcpPortWWN);
        Map<NetappFcpLifMetrics.FcpLifCounter, Float> sumResults = reportingPackage.getGlobalScratchPad()
                .addCounterRatePerSec("fcp_lif", fcpLifUUIDList,
                        NetappFcpLifMetrics.FcpLifCounter.read_data,
                        NetappFcpLifMetrics.FcpLifCounter.write_data);

        float read_throughput = sumResults.get(NetappFcpLifMetrics.FcpLifCounter.read_data);
        float write_throughput = sumResults.get(NetappFcpLifMetrics.FcpLifCounter.write_data);

        m.setReadThroughput(read_throughput);
        m.setWriteThroughput(write_throughput);
        m.setThroughput(MetricsRowScratchPad.add(m.getReadThroughput(), m.getWriteThroughput()));

        //Fetch fcpPort utilization from fcpPortToLinkSpeedMap
        m.setUtilization(calculateFcpPortUtilization(m.getThroughput(), context.getFcpPortToLinkSpeedMap().get(currentValues.getInventoryUUID())));
        return m;
    }
    
    @Override
    public FcpPortRawMeasurements rawMetrics(final NetAppDefaultStorageContext context) {
        final FcpPortRawMeasurements m = new FcpPortRawMeasurements();
        Map<FcpPortCounter,BigInteger> counterValue = getCurrentValues().getCounterValues();
        m.setAvgOtherLatency(counterValue.get(avg_other_latency));
        m.setAvgReadLatency(counterValue.get(avg_read_latency));
        m.setAvgWriteLatency(counterValue.get(avg_write_latency));
        m.setOtherOps(counterValue.get(other_ops));
        m.setReadOps(counterValue.get(read_ops));
        m.setWriteOps(counterValue.get(write_ops));
        m.setTotalOps(MetricsRowScratchPad.add(m.getReadOps(),
                m.getWriteOps(), m.getOtherOps()));

        //  Compute (read/write)_throughput at this Port
        //  We pass the fcpLifuuid list to addCounters method which aggregates the counters  at globalScratchPad
        //  and returns its sum
        final String fcpPortWWN = BuilderUtil.normalizeWWN(currentValues.getPerfInstanceUUID());
        List<String> fcpLifUUIDList = context.getFcpPortWWNToLifUUIDMap().get(fcpPortWWN);
        Map<NetappFcpLifMetrics.FcpLifCounter, BigInteger> sumResults = reportingPackage.getGlobalScratchPad()
                .addByCountersRaw("fcp_lif", fcpLifUUIDList,
                        NetappFcpLifMetrics.FcpLifCounter.read_data,
                        NetappFcpLifMetrics.FcpLifCounter.write_data);

        BigInteger read_throughput = sumResults.get(NetappFcpLifMetrics.FcpLifCounter.read_data);
        BigInteger write_throughput = sumResults.get(NetappFcpLifMetrics.FcpLifCounter.write_data);

        m.setReadThroughput(read_throughput);
        m.setWriteThroughput(write_throughput);
        m.setThroughput(MetricsRowScratchPad.add(m.getReadThroughput(), m.getWriteThroughput()));
        return m;
    }


    /**
     * @param fcpPortThroughput
     * @param linkSpeed
     * @return float fcpPortUtilization (fcpPortThroughput*8(converting bytes to bits) / link_speed) * 100
     */
    private float calculateFcpPortUtilization(float fcpPortThroughput, Integer linkSpeed){

        if(null == linkSpeed)
            return MetricsRowSample.NULL_FLOAT;
        long linkSpeedInBitsPerSec = linkSpeed * 1000 * MEGABIT_1;
        if(linkSpeedInBitsPerSec > 0){
            // multiplying by 8 to convert Bytes/sec to Bits/sec and 100 to return utilization in 0 to 100 % range
            return (float)(100 * ((8*(double)fcpPortThroughput)/linkSpeedInBitsPerSec));
        }else{
            // if we have 0 or negative values(error handling), then utilization cannot be calibrated, hence setting to null representation for float
            return MetricsRowSample.NULL_FLOAT;
        }
    }

}
