package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;


import com.netapp.oci.netapp.performance.node.interfaces.data.NFSv41NodeMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.EnumSet;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappNFSv41NodeMetrics.NFSv41NodeCounter.*;

public class NetappNFSv41NodeMetrics extends MetricsRowSample<NFSv41NodeMeasurements, NFSv41NodeMeasurements, NetappNFSv41NodeMetrics.NFSv41NodeCounter,
                                                                                      NetappNFSv41NodeMetrics.OCIENFSv41NodeCounter>{

    /**
     * Column names requested from ZAPI
     */
    public static enum NFSv41NodeCounter {
        read_avg_latency,
        write_avg_latency,
        nfsv4_1_ops,
        read_total,
        write_total,
        access_avg_latency,
        access_total,
        commit_avg_latency,
        commit_total,
        create_avg_latency,
        create_total,
        getattr_avg_latency,
        getattr_total,
        link_avg_latency,
        link_total,
        lookup_avg_latency,
        lookup_total,
        null_avg_latency,
        null_total,
        remove_avg_latency,
        remove_total,
        setattr_avg_latency,
        setattr_total,
    }

    /**
     * ZAPI Columns contributing to other latency
     */
    public static final EnumSet<NFSv41NodeCounter> OTHER_LATENCIES = EnumSet.of(
        access_avg_latency,
        commit_avg_latency,
        create_avg_latency,
        getattr_avg_latency,
        link_avg_latency,
        lookup_avg_latency,
        null_avg_latency,
        remove_avg_latency,
        setattr_avg_latency
    );

    /**
     * Columns reported to OCIE Server
     */
    public static enum OCIENFSv41NodeCounter {
        read_avg_latency,
        write_avg_latency,
        nfsv4_1_ops,
        read_total,
        write_total,
        other_latency
    }

    public NetappNFSv41NodeMetrics(final MetricsRowScratchPad<NFSv41NodeCounter> currentValues, final MetricsRowScratchPad<NFSv41NodeCounter> previousValues) {
        super(NFSv41NodeCounter.class, currentValues, previousValues);
    }

    @Override
    float getNormalizedFloatMetric(final OCIENFSv41NodeCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case nfsv4_1_ops:
                normalizedMetric = currentValues.getCounterRatePerSec(nfsv4_1_ops);
                break;

            case read_total:
                normalizedMetric = currentValues.getCounterRatePerSec(read_total);
                break;

            case write_total:
                normalizedMetric = currentValues.getCounterRatePerSec(write_total);
                break;

            case read_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(read_avg_latency, read_total);
                break;

            case write_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(write_avg_latency, write_total);
                break;

            case other_latency:
                normalizedMetric = getNormalizedOtherLatency();
                break;

            default:
                normalizedMetric = NULL_FLOAT;
        }
        return normalizedMetric;
    }

    private float getNormalizedOtherLatency() {
        if (currentValues.isWrapResetDetected()) {
            return NULL_FLOAT;
        }
        float normalizedOtherLatency = 0.0f;
        for (final NFSv41NodeCounter latencyCounterName : OTHER_LATENCIES) {
            final float anOtherLatency = getOtherLatencyPerOp(latencyCounterName);
            if (anOtherLatency == NULL_FLOAT) {
                normalizedOtherLatency = NULL_FLOAT;
                break;
            }
            normalizedOtherLatency += anOtherLatency;
        }
        return normalizedOtherLatency;
    }

    private float getOtherLatencyPerOp(final NFSv41NodeCounter latencyCounterName) {
        final float normalizedMetric;
        switch (latencyCounterName) {
            case access_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(access_avg_latency, access_total);
                break;

            case commit_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(commit_avg_latency, commit_total);
                break;

            case create_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(create_avg_latency, create_total);
                break;

            case getattr_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(getattr_avg_latency,  getattr_total);
                break;

            case link_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(link_avg_latency, link_total);
                break;

            case lookup_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(lookup_avg_latency, lookup_total);
                break;

            case null_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(null_avg_latency, null_total);
                break;

            case remove_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(remove_avg_latency, remove_total);
                break;

            case setattr_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(setattr_avg_latency, setattr_total);
                break;


            default:
                normalizedMetric = NULL_FLOAT;
        }
        return normalizedMetric;
    }

    @Override
    public NFSv41NodeMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        final NFSv41NodeMeasurements nfSv41NodeMeasurements = new NFSv41NodeMeasurements();
        nfSv41NodeMeasurements.setWriteTotal(getNormalizedFloatMetric(OCIENFSv41NodeCounter.write_total));
        nfSv41NodeMeasurements.setReadTotal(getNormalizedFloatMetric(OCIENFSv41NodeCounter.read_total));
        nfSv41NodeMeasurements.setReadAvgLatency(getNormalizedFloatMetric(OCIENFSv41NodeCounter.read_avg_latency));
        nfSv41NodeMeasurements.setWriteAvgLatency(getNormalizedFloatMetric(OCIENFSv41NodeCounter.write_avg_latency));
        nfSv41NodeMeasurements.setNFSv41Ops(getNormalizedFloatMetric(OCIENFSv41NodeCounter.nfsv4_1_ops));
        nfSv41NodeMeasurements.setOtherLatency(getNormalizedFloatMetric(OCIENFSv41NodeCounter.other_latency));
        return nfSv41NodeMeasurements;
    }
}
