package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;


import com.netapp.oci.netapp.performance.node.interfaces.data.NFSv4NodeMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.EnumSet;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappNFSv4NodeMetrics.NFSv4NodeCounter.*;

public class NetappNFSv4NodeMetrics extends MetricsRowSample<NFSv4NodeMeasurements, NFSv4NodeMeasurements, NetappNFSv4NodeMetrics.NFSv4NodeCounter,
                                                                                    NetappNFSv4NodeMetrics.OCIENFSv4NodeCounter>{

    /**
     * Column names requested from ZAPI
     */
    public static enum NFSv4NodeCounter {
        read_avg_latency,
        read_total,
        write_avg_latency,
        write_total,
        nfsv4_ops,
        access_avg_latency,
        access_total,
        commit_avg_latency,
        commit_total,
        create_avg_latency,
        create_total,
        getattr_avg_latency,
        getattr_total,
        link_avg_latency,
        link_total,
        lookup_avg_latency,
        lookup_total,
        null_avg_latency,
        null_total,
        remove_avg_latency,
        remove_total,
        setattr_avg_latency,
        setattr_total,
    }

    /**
     * ZAPI Columns contributing to other latency
     */
    public static final EnumSet<NFSv4NodeCounter> OTHER_LATENCIES = EnumSet.of(
        access_avg_latency,
        commit_avg_latency,
        create_avg_latency,
        getattr_avg_latency,
        link_avg_latency,
        lookup_avg_latency,
        null_avg_latency,
        remove_avg_latency,
        setattr_avg_latency
    );

    /**
     * Columns reported to OCIE Server
     */
    public static enum OCIENFSv4NodeCounter {
        read_avg_latency,
        read_total,
        write_avg_latency,
        write_total,
        nfsv4_ops,
        other_latency
    }


    public NetappNFSv4NodeMetrics(final MetricsRowScratchPad<NFSv4NodeCounter> currentValues, final MetricsRowScratchPad<NFSv4NodeCounter> previousValues) {
        super(NFSv4NodeCounter.class, currentValues, previousValues);
    }

    @Override
    float getNormalizedFloatMetric(final OCIENFSv4NodeCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case nfsv4_ops:
                normalizedMetric = currentValues.getCounterRatePerSec(nfsv4_ops);
                break;

            case read_total:
                normalizedMetric = currentValues.getCounterRatePerSec(read_total);
                break;

            case write_total:
                normalizedMetric = currentValues.getCounterRatePerSec(write_total);
                break;

            case read_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(read_avg_latency, read_total);
                break;

            case write_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(write_avg_latency, write_total);
                break;

            case other_latency:
                normalizedMetric = getNormalizedOtherLatency();
                break;

            default:
                normalizedMetric = NULL_FLOAT;
        }
        return normalizedMetric;
    }


    private float getNormalizedOtherLatency() {
        if (currentValues.isWrapResetDetected()) {
            return NULL_FLOAT;
        }
        float normalizedOtherLatency = 0.0f;
        for (final NFSv4NodeCounter latencyCounterName : OTHER_LATENCIES) {
            final float anOtherLatency = getOtherLatencyPerOp(latencyCounterName);
            if (anOtherLatency == NULL_FLOAT) {
                normalizedOtherLatency = NULL_FLOAT;
                break;
            }
            normalizedOtherLatency += anOtherLatency;
        }
        return normalizedOtherLatency;
    }


    private float getOtherLatencyPerOp(final NFSv4NodeCounter latencyCounterName) {
        final float normalizedMetric;
        switch (latencyCounterName) {
            case access_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(access_avg_latency, access_total);
                break;

            case commit_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(commit_avg_latency, commit_total);
                break;

            case create_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(create_avg_latency, create_total);
                break;

            case getattr_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(getattr_avg_latency,  getattr_total);
                break;

            case link_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(link_avg_latency, link_total);
                break;

            case lookup_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(lookup_avg_latency, lookup_total);
                break;

            case null_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(null_avg_latency, null_total);
                break;

            case remove_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(remove_avg_latency, remove_total);
                break;

            case setattr_avg_latency:
                normalizedMetric = currentValues.getLatencyPerOp(setattr_avg_latency, setattr_total);
                break;

            default:
                normalizedMetric = NULL_FLOAT;
        }
        return normalizedMetric;
    }

    @Override
    public NFSv4NodeMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        final NFSv4NodeMeasurements nfSv4NodeMeasurements = new NFSv4NodeMeasurements();
        nfSv4NodeMeasurements.setWriteAvgLatency(getNormalizedFloatMetric(OCIENFSv4NodeCounter.write_avg_latency));
        nfSv4NodeMeasurements.setReadAvgLatency(getNormalizedFloatMetric(OCIENFSv4NodeCounter.read_avg_latency));
        nfSv4NodeMeasurements.setNFSv4Ops(getNormalizedFloatMetric(OCIENFSv4NodeCounter.nfsv4_ops));
        nfSv4NodeMeasurements.setReadTotal(getNormalizedFloatMetric(OCIENFSv4NodeCounter.read_total));
        nfSv4NodeMeasurements.setWriteTotal(getNormalizedFloatMetric(OCIENFSv4NodeCounter.write_total));
        nfSv4NodeMeasurements.setOtherLatency(getNormalizedFloatMetric(OCIENFSv4NodeCounter.other_latency));
        return nfSv4NodeMeasurements;
    }
}
