package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.netapp.oci.netapp.performance.node.interfaces.data.NodeMeasurements;
import com.netapp.oci.netapp.performance.node.interfaces.data.NodeRawMeasurements;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetAppHostadapterMetrics.HostadapterCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappDiskMetrics.DiskCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappProcessorMetrics.BusyDomain;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappProcessorMetrics.ProcessorCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappProcessorMetrics.ProcessorMetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappWaflMetrics.CpPhaseTimes;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappWaflMetrics.WaflCounter;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappWaflMetrics.WaflMetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappNodeMetrics.SystemCounter.*;

public class NetappNodeMetrics extends MetricsRowSample<NodeMeasurements, NodeRawMeasurements, NetappNodeMetrics.SystemCounter,
        NetappNodeMetrics.SystemCounter> {

    private static final Logger logger = LoggerFactory.getLogger(NetappNodeMetrics.class);

    public static enum SystemCounter {
        avg_processor_busy,
        cpu_busy,
        cpu_elapsed_time,
        net_data_recv,
        net_data_sent,
        cifs_ops,
        fcp_ops,
        iscsi_ops,
        nfs_ops,
        other_ops,
        read_ops,
        http_ops,
        sys_avg_latency,
        sys_read_latency,
        sys_write_latency,
        total_ops,
        write_ops,
        system_ops,
        fcp_data_recv,
        fcp_data_sent;
    }
    
    private NetAppOCIEPerformancePackage reportingPackage;

    public NetappNodeMetrics(final MetricsRowScratchPad<SystemCounter> currentValues, final MetricsRowScratchPad<SystemCounter> previousValues, 
    		NetAppOCIEPerformancePackage reportingPackage) {
        super(SystemCounter.class, currentValues, previousValues);
        this.reportingPackage = reportingPackage;
    }

    SystemCounter getDenominatorForNumerator(final SystemCounter numeratorName) {
        final SystemCounter denominatorName;
        switch (numeratorName) {
            case avg_processor_busy:
            case cpu_busy:
                denominatorName = cpu_elapsed_time;
                break;
            case sys_avg_latency:
            	denominatorName = total_ops;
            	break;
            case sys_read_latency:
            	denominatorName = read_ops;
            	break;
            case sys_write_latency:
            	denominatorName = write_ops;
            	break;            	
            default:
                denominatorName = null;
        }
        return denominatorName;
    }

    @Override
    float getNormalizedFloatMetric(final SystemCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case avg_processor_busy:
            case cpu_busy:
                normalizedMetric = currentValues.getPercentFromDeltas(counterName,
                                                                      getDenominatorForNumerator(counterName));
                break;
            case read_ops:
            case write_ops:
            case other_ops:
            case total_ops:
            case system_ops:
            case cifs_ops:
            case fcp_ops:
            case iscsi_ops:
            case nfs_ops:
            case fcp_data_recv:
            case fcp_data_sent:
            case net_data_recv:
            case net_data_sent:
                normalizedMetric = currentValues.getCounterRatePerSec(counterName);
                break;
            case sys_avg_latency:
            case sys_read_latency:
            case sys_write_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName,
                        getDenominatorForNumerator(counterName));
                break;            	
            default:
                normalizedMetric = NULL_FLOAT;
                break;
        }
        return normalizedMetric;
    }

    @Override
    public NodeMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        // NodeMeasurements is only partially populated here with node specific metrics.
        // Other metrics from other target object names (volume) are merged in later.
        final NodeMeasurements m = new NodeMeasurements();
        m.setCpuBusy(getNormalizedFloatMetric(cpu_busy));
        m.setAvgProcessorBusy(getNormalizedFloatMetric(avg_processor_busy));
        m.setCpuElapsedTime(currentValues.getCounterDeltaFinal(cpu_elapsed_time));
        m.setNetDataRecv(getNormalizedFloatMetric(net_data_recv));
        m.setNetDataSent(getNormalizedFloatMetric(net_data_sent));
        m.setCifsOps(getNormalizedFloatMetric(cifs_ops));
        m.setFcpOps(getNormalizedFloatMetric(fcp_ops));
        m.setIscsiOps(getNormalizedFloatMetric(iscsi_ops));
        m.setNfsOps(getNormalizedFloatMetric(nfs_ops));
        m.setOtherOps(getNormalizedFloatMetric(other_ops));
        m.setReadOps(getNormalizedFloatMetric(read_ops));
        m.setWriteOps(getNormalizedFloatMetric(write_ops));
        m.setSysAvgLatency(getNormalizedFloatMetric(sys_avg_latency));
        m.setSysReadLatency(getNormalizedFloatMetric(sys_read_latency));
        m.setSysWriteLatency(getNormalizedFloatMetric(sys_write_latency));
        m.setFcpDataRecv(getNormalizedFloatMetric(fcp_data_recv));
        m.setFcpDataSent(getNormalizedFloatMetric(fcp_data_sent));
        m.setOpmSysReadThroughput(MetricsRowScratchPad.add(m.getFcpDataRecv(), m.getNetDataRecv()));
        m.setOpmSysWriteThroughput(MetricsRowScratchPad.add(m.getFcpDataSent(), m.getNetDataSent()));
        m.setOpmSysThroughput(MetricsRowScratchPad.add(m.getOpmSysReadThroughput(), m.getOpmSysWriteThroughput()));
        m.setOpmSysUtilization(getNodeUtilization(context,currentValues, m));
        m.setMaxAggregateUtilization(getMaxAggregateUtilization(context, currentValues));
        m.setPercentCleanTimeUtil(getPercentCleanTimeUtil(context, currentValues, m));
        m.setDiskShelfFibreChannelLoopUtilization(getDiskShelfFibreChannelLoopUtilization(context, currentValues));
        
        
        //For FS.0 only, we use system_ops as total_ops, since total_ops always returns zero.
        if(!context.isLessThanFullSteam()) {
        	m.setTotalOps(getNormalizedFloatMetric(system_ops));
        }
        else {
        	m.setTotalOps(MetricsRowScratchPad.add(m.getCifsOps(), m.getFcpOps(), m.getIscsiOps(), m.getNfsOps(), currentValues.getCounterRatePerSec(http_ops)));
        }
        return m;
    }
    
    @Override
    public NodeRawMeasurements rawMetrics(final NetAppDefaultStorageContext context){
        Map<SystemCounter,BigInteger> counterValue = getCurrentValues().getCounterValues();
        return createNodeRawMeasurements(counterValue);
    }
    
    /**
     * Raw system_ops is not available for SN. Cooked system_ops is rolled up by the protocol counters.
     *  
     * @param counterValues
     * @return
     */
    protected static NodeRawMeasurements createNodeRawMeasurements(Map<SystemCounter,BigInteger> counterValues){
    	final NodeRawMeasurements m = new NodeRawMeasurements();
        	m.setAvgProcessorBusy(counterValues.get(avg_processor_busy));
        	m.setCpuBusy(counterValues.get(cpu_busy));
        	m.setCpuElapsedTime(counterValues.get(cpu_elapsed_time));
        	m.setNetDataRecv(counterValues.get(net_data_recv));
        	m.setNetDataSent(counterValues.get(net_data_sent));
        	m.setCifsOps(counterValues.get(cifs_ops));
        	m.setFcpOps(counterValues.get(fcp_ops));
        	m.setIscsiOps(counterValues.get(iscsi_ops));
        	m.setNfsOps(counterValues.get(nfs_ops));
        	m.setOtherOps(counterValues.get(other_ops));
        	m.setReadOps(counterValues.get(read_ops));
        	m.setSysAvgLatency(counterValues.get(sys_avg_latency));
        	m.setSysReadLatency(counterValues.get(sys_read_latency));
        	m.setSysWriteLatency(counterValues.get(sys_write_latency));
        	m.setTotalOps(counterValues.get(total_ops));
        	m.setWriteOps(counterValues.get(write_ops));
        	m.setFcpDataRecv(counterValues.get(fcp_data_recv));
        	m.setFcpDataSent(counterValues.get(fcp_data_sent));
        	m.setOpmSysReadThroughput(MetricsRowScratchPad.add(m.getNetDataRecv(), m.getFcpDataRecv()));
        	m.setOpmSysWriteThroughput(MetricsRowScratchPad.add(m.getNetDataSent(), m.getFcpDataSent()));
        	m.setOpmSysThroughput(MetricsRowScratchPad.add(m.getOpmSysReadThroughput(), m.getOpmSysWriteThroughput()));

    	return m;
    }
    
    /**
	 * Method to calculate node utilization using processor and wafl counters
	 * @param context - NetAppDefaultStorageContext 
	 * @param currentScratchPad - MetricsRowScratchPad with current values
	 * @return
	 */
	
	protected float getNodeUtilization(NetAppDefaultStorageContext context, MetricsRowScratchPad<SystemCounter> currentScratchPad, NodeMeasurements nodeMeasurements){
		final String inventoryUUID = currentScratchPad.getInventoryUUID();
		Map<String, String> nodeNameToUUIDMap = context.getNodeNameToUUIDMap();
		GlobalScratchPad globalScratchPad = reportingPackage.getGlobalScratchPad();
		Map<String, MetricsRowScratchPad<ProcessorCounter>> processorScratchPadsMap =  (Map<String, MetricsRowScratchPad<ProcessorCounter>>) globalScratchPad.get("processor");
		Map<String, MetricsRowScratchPad<WaflCounter>> waflScratchPadsMap = (Map<String, MetricsRowScratchPad<WaflCounter>>) globalScratchPad.get("wafl");
		long totalCpMSecs = 0;
		long P2Flush = 0;
		long totalKahunaBusy = 0;
		float nodeUtilization=0;
		final float KAHUNA_THRESHOLD_LIMIT = 60F;
		final float MAX_THRESHOLD_LIMIT = 100F;
		long maxCpuElapsedTime= MetricsRowSample.NULL_LONG;
		final long deltaPollTimeInMs = currentScratchPad.getCollectionPeriod(); 
		
		nodeMeasurements.setDeltaPollTime(deltaPollTimeInMs);
		
		float processorBusyPercent =  currentScratchPad.getPercentFromDeltas(SystemCounter.avg_processor_busy, SystemCounter.cpu_elapsed_time);
		
		for(String key : processorScratchPadsMap.keySet()){
			final String[] tokens = key.split(":");
			String nodeUUID = nodeNameToUUIDMap.get(tokens[0]);
			if(!nodeUUID.equals(inventoryUUID)){
				continue;
			}
			ProcessorMetricsRowScratchPad processorScratchPad = (ProcessorMetricsRowScratchPad) processorScratchPadsMap.get(key);
			MetricsRowScratchPad<BusyDomain> busyDomainScratchPad = processorScratchPad.getProcessorDomainScratchPad();
			long kahunaBusy = busyDomainScratchPad.getCounterDelta(BusyDomain.kahuna);
			long processorElapsedTime = processorScratchPad.getCounterDelta(ProcessorCounter.processor_elapsed_time);
			if(kahunaBusy != MetricsRowSample.NULL_FLOAT){
				totalKahunaBusy += kahunaBusy;
			}
			if((processorElapsedTime != MetricsRowSample.NULL_LONG) && (processorElapsedTime > maxCpuElapsedTime)){
				maxCpuElapsedTime = processorElapsedTime;
			}
			
		}
		
		nodeMeasurements.setTotalKahunaBusy(totalKahunaBusy);
		nodeMeasurements.setMaxProcessorElapsedTime(maxCpuElapsedTime);
		
		final float totalKahunaBusyPercent = (maxCpuElapsedTime == 0 || totalKahunaBusy == 0 || 
				maxCpuElapsedTime == MetricsRowSample.NULL_LONG || totalKahunaBusy == MetricsRowSample.NULL_LONG)
                ? 0
                : (float) (100d * (double) totalKahunaBusy / (double) maxCpuElapsedTime);
				
		for(String key : waflScratchPadsMap.keySet()){
			final String[] tokens = key.split(":");
			String nodeUUID = nodeNameToUUIDMap.get(tokens[0]);
			if(!nodeUUID.equals(inventoryUUID)){
				continue;
			}
			WaflMetricsRowScratchPad waflScratchPad = (WaflMetricsRowScratchPad) waflScratchPadsMap.get(key);
			MetricsRowScratchPad<CpPhaseTimes> cpPhaseTimeScratchPad = waflScratchPad.getCpPhaseTimesScratchPad();
			totalCpMSecs = waflScratchPad.getCounterDelta(WaflCounter.total_cp_msecs);
			P2Flush = cpPhaseTimeScratchPad.getCounterDelta(CpPhaseTimes.P2_FLUSH);
		}
		
		nodeMeasurements.setTotalCpMSecs(totalCpMSecs);
		nodeMeasurements.setP2Flush(P2Flush);
		
		
		final float waflUtilization = (totalCpMSecs == MetricsRowSample.NULL_LONG || totalCpMSecs == 0 || 
				deltaPollTimeInMs == MetricsRowSample.NULL_LONG || deltaPollTimeInMs == 0 || P2Flush == 0)
				? 0
				: (float) (((double)totalCpMSecs * (1 - ((double)P2Flush/(double)totalCpMSecs)/.8F)/deltaPollTimeInMs) * 100d);

        // Formula
        // processor_busy_percent=avg_processor_busy(delta)/cpu_elapsed_time1(delta))
        // total_cp_time=total_cp_msecs(delta)
        // p2_flush_time=cp_phase_times.P2FLUSH(delta)
        // kahuna=domain_busy.kahuna(delta)/processor_elapsed_time(delta)
        // Node_Utilization =
        //       100% when kahuna > 60% OR
        //       Max of {
        //           processor_busy_percent * 100
        //          (total_cp_time(1 - (p2_flush_time/total_cp_time)/.8) /poll time msecs) * 100
        if(totalKahunaBusyPercent > KAHUNA_THRESHOLD_LIMIT){
            nodeUtilization =  MAX_THRESHOLD_LIMIT;
        } else if (processorBusyPercent > MAX_THRESHOLD_LIMIT || waflUtilization > MAX_THRESHOLD_LIMIT) {
            // This case should not happen. Logging it as warning as it could only happen if ontap reports large values for walf cp_time.
            logger.warn(PackageContext.getLogPrefix()+" - processorBusyPercent - {} or waflUtilization - {} cannot be greater than 100%." +
                    "for node - "+currentScratchPad.getInventoryUUID()+" ", processorBusyPercent, waflUtilization);
            nodeUtilization =  MAX_THRESHOLD_LIMIT;
        } else {
            nodeUtilization = Math.max(processorBusyPercent, waflUtilization );
        }
        return nodeUtilization;
	}
	
	protected float getMaxAggregateUtilization(NetAppDefaultStorageContext context, MetricsRowScratchPad<SystemCounter> currentScratchPad){
		String inventoryUUID = currentScratchPad.getInventoryUUID();
		GlobalScratchPad globalScratchPad = reportingPackage.getGlobalScratchPad();
		Map<String, List<String>> nodeUUIDToDiskUUIDs = context.getNodeUUIDToDiskUUIDs();
		List<String> diskPerformanceUUIDs = nodeUUIDToDiskUUIDs.get(inventoryUUID);
		if(diskPerformanceUUIDs == null){
			return MetricsRowSample.NULL_LONG;
		}
		
		List<DiskCounter> numerator = new ArrayList<DiskCounter>();
		numerator.add(DiskCounter.disk_busy);
		List<DiskCounter> denominator = new ArrayList<DiskCounter>();
		denominator.add(DiskCounter.base_for_disk_busy);
		Map<DiskCounter, Float> maxPercentDelta = globalScratchPad.getMaximumOfPercentDeltas(diskPerformanceUUIDs, "disk", numerator, denominator);
		return (maxPercentDelta == null) ? MetricsRowSample.NULL_FLOAT : maxPercentDelta.get(DiskCounter.disk_busy);
		
	}
	
	protected float getPercentCleanTimeUtil(NetAppDefaultStorageContext context, MetricsRowScratchPad<SystemCounter> currentScratchPad, NodeMeasurements m){
		long totalCpMSecs = m.getTotalCpMSecs();
		long P2Flush = m.getP2Flush();
		final float DEFAULT_MAXAGGREGATE_UTILIZATION = 50F;
		float percentCleanTimeUtil = MetricsRowSample.NULL_FLOAT;
		
		if(m.getDeltaPollTime() == MetricsRowSample.NULL_LONG || m.getDeltaPollTime() == 0 || totalCpMSecs == MetricsRowSample.NULL_LONG || totalCpMSecs == 0 
				|| P2Flush == MetricsRowSample.NULL_LONG){
			return MetricsRowSample.NULL_FLOAT;
		}

		if(m.getMaxAggregateUtilization() < DEFAULT_MAXAGGREGATE_UTILIZATION){
			percentCleanTimeUtil = (float)((((double) totalCpMSecs - (double) P2Flush)/.8F)/(double)m.getDeltaPollTime());
		}else {
			percentCleanTimeUtil = (float)((double)totalCpMSecs / (double)m.getDeltaPollTime());
		}
		return percentCleanTimeUtil;
	}
	
	protected float getDiskShelfFibreChannelLoopUtilization(NetAppDefaultStorageContext context, MetricsRowScratchPad<SystemCounter> currentScratchPad){
		String inventoryUUID = currentScratchPad.getInventoryUUID();
		Map<String, String> nodeNameToUUIDMap = context.getNodeNameToUUIDMap();
		GlobalScratchPad globalScratchPad = reportingPackage.getGlobalScratchPad();
		Map<String, MetricsRowScratchPad<HostadapterCounter>> hostAdapterScratchPadsMap = (Map<String, MetricsRowScratchPad<HostadapterCounter>>) globalScratchPad.get("hostadapter");
		float diskShelfFibreChannelLoopUtilization = MetricsRowSample.NULL_FLOAT;
		
		long deltaPollTimeInMs = currentScratchPad.getCollectionPeriod();
		
		for(String key : hostAdapterScratchPadsMap.keySet()){
			final String[] tokens = key.split(":");
			String nodeUUID = nodeNameToUUIDMap.get(tokens[0]);
			if(!nodeUUID.equals(inventoryUUID)){
				continue;
			}
			MetricsRowScratchPad<HostadapterCounter> hostAdapterScratchPad = hostAdapterScratchPadsMap.get(key);
			float perHostAdapterUtilization = MetricsRowSample.NULL_FLOAT;
			float totalBytes = MetricsRowScratchPad.add(hostAdapterScratchPad.getCounterDelta(HostadapterCounter.bytes_written), 
					hostAdapterScratchPad.getCounterDelta(HostadapterCounter.bytes_read));
			long hostAdapterMaxLinkDataRate = ((hostAdapterScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate) != MetricsRowSample.NULL_BIG_INT) && (hostAdapterScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate) != null))
												? hostAdapterScratchPad.getCounterValues().get(HostadapterCounter.max_link_data_rate).longValue() : MetricsRowSample.NULL_LONG;
			
			if(hostAdapterMaxLinkDataRate != MetricsRowSample.NULL_LONG || hostAdapterMaxLinkDataRate != 0l){
					perHostAdapterUtilization = (float) ((double)totalBytes/(double)deltaPollTimeInMs/(double)hostAdapterMaxLinkDataRate);
			}
			
			if(diskShelfFibreChannelLoopUtilization < perHostAdapterUtilization){
				diskShelfFibreChannelLoopUtilization = perHostAdapterUtilization;
			}
		}
		return diskShelfFibreChannelLoopUtilization;
	}

}
