package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.netapp.oci.netapp.performance.processor.interfaces.data.ProcessorMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.math.BigInteger;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;


/**
 * Processor Metrics are very different from other metrics.
 *
 * Normally, for each metric reported to OCIE (MetricsRowSample.E), there is a corresponding source metric from ZAPI
 * (MetricsRowSample.Z), and a simple transformation recipe to normalize the value to OCIE expected units
 * (rate or ratio to other counter/column delta)
 *
 * The MetricsRowSample'S Enum <Z> normally defines both the metrics to request from ZAPI, and the columns to report to OCIE
 * (Z == O).
 *
 * The required OCIE Processor metrics columns <O> are orthogonal to the set of metrics requested from ZAPI <Z>,
 * so specific logic is required.
 *
 * This Metrics class uses 3 Enums:
 *      1) ZAPIProcessorCounter <Z> defines the metrics to request from ZAPI
 *      2) ProcessorBusyDomain is used to interpret domain_busy, which is a CSV list of values per domain, rather than a single value
 *      3) OCIEProcessorCounter <O> represents the columns required by OCIE, including aggregations of subsets of domain_busy values
 *
 */
public class NetappProcessorMetrics extends MetricsRowSample<ProcessorMeasurements, ProcessorMeasurements, NetappProcessorMetrics.ProcessorCounter,
                                                                                    NetappProcessorMetrics.OCIEProcessorCounter> {

    // This is 1 of 2 cases where type parameters <Z, E> are NOT the same
    // NetappQosVolumeWorkloadMetrics is the other

    @SuppressWarnings({"UnusedDeclaration"})
    private static final Logger logger = LoggerFactory.getLogger(NetappProcessorMetrics.class);

    /**
     * The processor counter fields targeted in ZAPI
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public enum ProcessorCounter {
        domain_busy,                        // Vector of CSV values per domain
        processor_busy,                     // Scalar value of processor busy time
        processor_elapsed_time              // Scalar value of total time
    }

    /**
     * The list of all processor busy domain names (for domain_busy).  SN supports all domain names, while RR supports a subset of domain names
     * See:
     *       <perf-object-counter-list-info>
     *           <objectname>processor</objectname>
     *       </perf-object-counter-list-info>
     *
     *       <labels>
     *           <label-info>idle,kahuna,storage,exempt,raid,raid_exempt,target,dnscache,cifs,wafl_exempt,wafl_xcleaner,sm_exempt,cluster,protocol,nwk_exclusive,nwk_exempt,nwk_legacy,hostOS,ssan_exempt</label-info>
     *       </labels>
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public enum BusyDomain {
        idle,
        kahuna,
        storage,
        exempt,
        raid,
        raid_exempt,
        target,
        dnscache,
        cifs,
        wafl_exempt,
        wafl_xcleaner,
        sm_exempt,
        cluster,
        protocol,
        nwk_exclusive,
        nwk_exempt,
        nwk_legacy,
        hostOS,
        ssan_exempt
    }

    /**
     * Processor busy domains that contribute to N Blade CPU utilization
     */
    private static final EnumSet<BusyDomain> NBLADE_DOMAINS = EnumSet.noneOf(BusyDomain.class);
    static {
        NBLADE_DOMAINS.add(BusyDomain.cifs);
        NBLADE_DOMAINS.add(BusyDomain.cluster);
        NBLADE_DOMAINS.add(BusyDomain.dnscache);
        NBLADE_DOMAINS.add(BusyDomain.nwk_exclusive);
        NBLADE_DOMAINS.add(BusyDomain.nwk_exempt);
        NBLADE_DOMAINS.add(BusyDomain.nwk_legacy);
        NBLADE_DOMAINS.add(BusyDomain.protocol);
        NBLADE_DOMAINS.add(BusyDomain.target);
    }

    /**
     * Processor busy domains that contribute to D Blade CPU utilization
     */
    private static final EnumSet<BusyDomain> DBLADE_DOMAINS = EnumSet.noneOf(BusyDomain.class);
    static {
        DBLADE_DOMAINS.add(BusyDomain.exempt);
        DBLADE_DOMAINS.add(BusyDomain.kahuna);
        DBLADE_DOMAINS.add(BusyDomain.raid);
        DBLADE_DOMAINS.add(BusyDomain.sm_exempt);
        DBLADE_DOMAINS.add(BusyDomain.storage);
        DBLADE_DOMAINS.add(BusyDomain.wafl_exempt);
        DBLADE_DOMAINS.add(BusyDomain.wafl_xcleaner);
    }

    /**
     * Sierra Nevada supports all known processor busy domains
     */
    @SuppressWarnings({"UnusedDeclaration"})
    private static final EnumSet<BusyDomain> SN_DOMAINS = EnumSet.allOf(BusyDomain.class);

    /**
     * Rolling Rock supports a subset of defined processor busy domains
     */
    private static final EnumSet<BusyDomain> RR_DOMAINS = EnumSet.allOf(BusyDomain.class);
    static {
        RR_DOMAINS.remove(BusyDomain.raid_exempt);
        RR_DOMAINS.remove(BusyDomain.ssan_exempt);
    }

    /**
     * The counters as required by Server
     */
    public enum OCIEProcessorCounter {
        n_blade_busy,                   // Aggregated domain_busy rate for all N Blade domains
        d_blade_busy,                   // Aggregated domain_busy rate for all D Blade domains
        processor_busy,                 // Ratio of processor_busy to processor_elapsed_time
        kahuna,
        storage,
        raid,
        target,
        cifs,
        cluster,
        protocol,
        nwk_exclusive,
        nwk_legacy,
    }

    /**
     * Get a ProcessorMetrics MetricsRowScratchPad.  Contains ZAPI counters, intermediate busy values per domain, and final values reported to OCIE.
     * @param currentValues the current raw values from ZAPI
     * @param previousValues the previous raw values from ZAPI
     */
    public NetappProcessorMetrics(final MetricsRowScratchPad<ProcessorCounter> currentValues, final MetricsRowScratchPad<ProcessorCounter> previousValues) {
        super(ProcessorCounter.class, currentValues, previousValues);
    }


    /**
     * Get the normalized float metric for report to OCIE.   This is NOT an override, because set of reported
     * metrics is derived from but not equal to the set of collected metrics
     *
     * @param counterName     The enum constant for the desired counter
     *
     * @return The measurement
     */
    float getNormalizedFloatMetric(final OCIEProcessorCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case n_blade_busy:      // processor specific recipe
                normalizedMetric = getDomainsBusyPercent(OCIEProcessorCounter.n_blade_busy, NBLADE_DOMAINS);
                break;
            case d_blade_busy:      // processor specific recipe
                normalizedMetric = getDomainsBusyPercent(OCIEProcessorCounter.d_blade_busy, DBLADE_DOMAINS);
                break;
            case processor_busy:
                normalizedMetric = currentValues.getPercentFromDeltas(ProcessorCounter.processor_busy,
                                                                      ProcessorCounter.processor_elapsed_time);
                break;

            case kahuna:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.kahuna, BusyDomain.kahuna);
                break;
            case storage:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.storage, BusyDomain.storage);
                break;
            case raid:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.raid, BusyDomain.raid);
                break;
            case target:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.target, BusyDomain.target);
                break;
            case cifs:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.cifs, BusyDomain.cifs);
                break;
            case cluster:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.cluster, BusyDomain.cluster);
                break;
            case protocol:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.protocol, BusyDomain.protocol);
                break;
            case nwk_exclusive:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.nwk_exclusive, BusyDomain.nwk_exclusive);
                break;
            case nwk_legacy:
                normalizedMetric = getDomainBusyPercent(OCIEProcessorCounter.nwk_legacy, BusyDomain.nwk_legacy);
                break;

            default:
                normalizedMetric = MetricsRowSample.NULL_FLOAT;
                break;
        }
        return normalizedMetric;
    }

    @Override
    public ProcessorMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        final ProcessorMeasurements m = new ProcessorMeasurements();
        m.setProcessorBusy(getNormalizedFloatMetric(OCIEProcessorCounter.processor_busy));
        m.setProcessorElapsedTime(currentValues.getCounterDeltaFinal(ProcessorCounter.processor_elapsed_time));
        m.setDBladeBusy(getNormalizedFloatMetric(OCIEProcessorCounter.d_blade_busy));
        m.setNBladeBusy(getNormalizedFloatMetric(OCIEProcessorCounter.n_blade_busy));
        m.setKahunaBusy(getNormalizedFloatMetric(OCIEProcessorCounter.kahuna));
        m.setStorageBusy(getNormalizedFloatMetric(OCIEProcessorCounter.storage));
        m.setRaidBusy(getNormalizedFloatMetric(OCIEProcessorCounter.raid));
        m.setTargetBusy(getNormalizedFloatMetric(OCIEProcessorCounter.target));
        m.setCifsBusy(getNormalizedFloatMetric(OCIEProcessorCounter.cifs));
        m.setClusterBusy(getNormalizedFloatMetric(OCIEProcessorCounter.cluster));
        m.setProtocolBusy(getNormalizedFloatMetric(OCIEProcessorCounter.protocol));
        m.setNwkExclusiveBusy(getNormalizedFloatMetric(OCIEProcessorCounter.nwk_exclusive));
        m.setNwkLegacyBusy(getNormalizedFloatMetric(OCIEProcessorCounter.nwk_legacy));
        return m;
    }

    /**
     * Get the percentage of time spent in the specified target domains, and store in the given counter
     *
     * @param ocieCounter the counter to hold the percent of time being calculated
     * @param targetDomains the set of processor domains to include in the calculation
     * @return aggregated Busy time as a percentage
     */
    private float getDomainsBusyPercent(final OCIEProcessorCounter ocieCounter, final EnumSet<BusyDomain> targetDomains) {
        final ProcessorMetricsRowScratchPad currentProcessorMetricsRow = (ProcessorMetricsRowScratchPad) currentValues;
        final MetricsRowScratchPad<BusyDomain> processorDomainRowScratchPad = currentProcessorMetricsRow.getProcessorDomainScratchPad();
        final MetricsRowScratchPad<OCIEProcessorCounter> ocieProcessorRowScratchPad = currentProcessorMetricsRow.getOCIEProcessorScratchPad();
        final long totalDomainBusyDelta = getSumOfDeltas(processorDomainRowScratchPad, targetDomains);

        final long deltaProcessorTime = getProcessorTimeDelta();
        final float percentBusy = (deltaProcessorTime == 0)
                  ? NULL_FLOAT
                  : (deltaProcessorTime > 0 && totalDomainBusyDelta > 0)
                        ? (float) (100d * (double) totalDomainBusyDelta / (double) deltaProcessorTime)
                        : NULL_FLOAT;
        ocieProcessorRowScratchPad.setFinalValueForDebug(ocieCounter, percentBusy);
        return percentBusy;
    }


    /**
     * Get the percentage of time spent in the specified target domains, and store in the given counter
     *
     * @param ocieCounter the counter to hold the percent of time being calculated
     * @param targetDomain the processor domain to include in the calculation
     * @return domain Busy time as a percentage
     */
    private float getDomainBusyPercent(final OCIEProcessorCounter ocieCounter, final BusyDomain targetDomain) {
        final ProcessorMetricsRowScratchPad currentProcessorMetricsRow = (ProcessorMetricsRowScratchPad) currentValues;
        final MetricsRowScratchPad<BusyDomain> processorDomainRowScratchPad = currentProcessorMetricsRow.getProcessorDomainScratchPad();
        final MetricsRowScratchPad<OCIEProcessorCounter> ocieProcessorRowScratchPad = currentProcessorMetricsRow.getOCIEProcessorScratchPad();
        final Long domainDelta = processorDomainRowScratchPad.getCounterDelta(targetDomain);
        if (domainDelta == null) {
            // DE4569: ZAPI to query column names may have failed.  Failure is expected to be transient.
            // For this poll, we can't identify the busy-domain subcolumns, so return null
            return NULL_FLOAT;
        }
        final long deltaProcessorTime = getProcessorTimeDelta();
        final float percentBusy = (deltaProcessorTime == 0)
                  ? NULL_FLOAT
                  : (deltaProcessorTime > 0 && domainDelta > 0)
                        ? (float) (100d * (double) domainDelta / (double) deltaProcessorTime)
                        : NULL_FLOAT;
        ocieProcessorRowScratchPad.setFinalValueForDebug(ocieCounter, percentBusy);
        return percentBusy;
    }

    /**
     * Get the sum of deltas for the set of BusyDomains identified by targetDomains
     * @param rowScratchPad the container of the delta values for all BusyDomains
     * @param targetDomains the set of BusyDomains to sum
     * @return the sum of deltas for the set of BusyDomains identified by targetDomains
     */
    private long getSumOfDeltas(final MetricsRowScratchPad<BusyDomain> rowScratchPad, final EnumSet<BusyDomain> targetDomains) {
        long sumOfDeltas = 0;
        for (final BusyDomain targetDomain : targetDomains) {
            final Long domainDelta = rowScratchPad.getCounterDelta(targetDomain);
            if (domainDelta != null) {
                sumOfDeltas += domainDelta;
            }
        }
        return sumOfDeltas;
    }

    /**
     * Get the difference in total processor time polled from ZAPIthe difference in total processor time polled from ZAPI
     * @return the difference in total processor time polled from ZAPI
     */
    private long getProcessorTimeDelta() {
        return currentValues.getCounterDelta(ProcessorCounter.processor_elapsed_time);
    }


    /**
     * This class is a custom scratchpad for processor metrics.
     *
     * Normally, for each metric reported to OCIE, there is a corresponding source metric from ZAPI,
     * and a simple transformation recipe to normalize the value to OCIE expected units.
     *
     * The ScratchPad's Enum normally defines both the metrics to request from ZAPI, and the columns to report to OCIE.
     *
     * The OCIE Processor metrics columns are orthogonal to the set of metrics requested from ZAPI,
     * so a custom ScratchPad is required.
     *
     * As an extention of MetricsRowScratchPad, ProcessorMetricsRowScratchPad provides support for ProcessorCounter
     * values & calculations.
     *
     * Additionally, we maintain a MetricsRowScratchPad<BusyDomain> to support BusyDomain values & calculations.
     *
     * Finally, we maintain a MetricsRowScratchPad<OCIEProcessorCounter> to support OCIEProcessorCounter values
     * & calculations, which are the only values reported to OCIE.
     */
    public static final class ProcessorMetricsRowScratchPad extends MetricsRowScratchPad<ProcessorCounter> {

        private MetricsRowScratchPad<BusyDomain> domainBusyRowScratchPad = null;
        private MetricsRowScratchPad<OCIEProcessorCounter> ocieProcessorRowScratchPad = null;
        private List<String> discoveredCpuColumnNames;

        public ProcessorMetricsRowScratchPad(final String perfInstanceUUID, final String inventoryUUID,
                                             final long serverPollStartTimeStamp, final long pollStartTimeStamp,
                                             final MetricsRowScratchPad<ProcessorCounter> previousValues,
                                             final List<String> discoveredCpuColumnNames, boolean metricsDebugEnabled) {

            super(ProcessorCounter.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp, pollStartTimeStamp,
                    previousValues, metricsDebugEnabled);
            this.discoveredCpuColumnNames = discoveredCpuColumnNames;
        }

        /**
         * Get the processor domain ScatchPad.  If not initialized, ScatchPad is generated from domain_busy CSV value
         * @return the processor domain ScatchPad
         */
        public MetricsRowScratchPad<BusyDomain> getProcessorDomainScratchPad() {
            if (domainBusyRowScratchPad == null) {
                initDomainBusyScratchPad();
            }
            return domainBusyRowScratchPad;
        }

        /**
         * Get the OCIEProcessor ScratchPad.  If not initialized, ScatchPad is generated from processor domain ScatchPad values.
         * @return the OCIE processor ScatchPad
         */
        private MetricsRowScratchPad<OCIEProcessorCounter> getOCIEProcessorScratchPad() {
            if (ocieProcessorRowScratchPad == null) {
                initOCIEProcessorScratchPad();
            }
            return ocieProcessorRowScratchPad;
        }

        /**
         * Get previous ProcessorMetrics ScratchPad.
         * @return previous ProcessorMetrics ScratchPad
         */
        private ProcessorMetricsRowScratchPad getPreviousProcessorMetrics() {
            return (ProcessorMetricsRowScratchPad) previousValues;
        }

        /**
         * Get previous BusyDomain ScratchPad safely (contained by previous PreviousProcessor ScratchPad).
         * @return previous DomainBusy ScratchPad
         */
        private MetricsRowScratchPad<BusyDomain> getPreviousProcessorDomainScratchPad() {
            final ProcessorMetricsRowScratchPad previousProcessorRowScratchPad = getPreviousProcessorMetrics();
            return (previousProcessorRowScratchPad != null)
                   ? previousProcessorRowScratchPad.getProcessorDomainScratchPad()
                   : null;
        }

        /**
         * Get previous OCIEProcessorMetrics ScratchPad safely (contained by previous PreviousProcessor ScratchPad).
         * @return previous OCIEProcessorMetrics ScratchPad
         */
        private MetricsRowScratchPad<OCIEProcessorCounter> getPreviousOCIECounterScratchPad() {
            final ProcessorMetricsRowScratchPad previousProcessorRowScratchPad = getPreviousProcessorMetrics();
            return (previousProcessorRowScratchPad != null)
                   ? previousProcessorRowScratchPad.getOCIEProcessorScratchPad()
                   : null;
        }

        /**
         * Create a MetricsRowScratchPad<ProcessorBusyDomain> to hold each domain counter individually
         * domain_busy is actually an array of values for each ProcessorDomain.
         *
         * ZAPI effectively returns a Map of ZAPIProcessorCounter to List of ProcessorDomain values.
         */
        private void initDomainBusyScratchPad() {

            // get cpu column names from context
            final int domainBusyColumnCount;
            if (discoveredCpuColumnNames == null || discoveredCpuColumnNames.size() == 0) {
                logger.warn("Failed to discover domain_busy CPU domains");
                domainBusyColumnCount = 0;
            } else {
                domainBusyColumnCount = discoveredCpuColumnNames.size();
            }

            final EnumMap<ProcessorCounter, String> rawValues = getRawValues();
            final String rawDomainBusyValues = rawValues.get(ProcessorCounter.domain_busy);         // Raw value is CSV list
            // raw value is a CSV array of values for each domain
            final String[] domainBusyRawValues = rawDomainBusyValues.split(",");

            // Because ZAPI returns these values as a CSV list, MetricsRowScratchPad doesn't calculate deltas
            // or poll start estimates.   Now that we've split the list into individual values, we can
            // perform the required calculations on the raw ZAPI values.
            final MetricsRowScratchPad<BusyDomain> previousDomainValues = getPreviousProcessorDomainScratchPad();
            this.domainBusyRowScratchPad =
                    new MetricsRowScratchPad<BusyDomain>(BusyDomain.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp,
                                                     pollStartTimeStamp, previousDomainValues, metricsDebugEnabled);
            
            if (domainBusyRawValues.length != domainBusyColumnCount) {
                logger.warn("Discovered domain_busy CPU domains ({}) != domain busy raw values ({}) for processor instance with uuid :{}", 
                        new Object[] {domainBusyColumnCount, domainBusyRawValues.length, inventoryUUID});
                return;
            }

            // Populate the BusyDomain ScratchPad with parsed values from the parent ProcessorCounter ScratchPad
            int i=-1;
            for (final String domainName : discoveredCpuColumnNames) {
                i++;
                final BusyDomain busyDomain;
                try {
                    busyDomain = BusyDomain.valueOf(domainName);
                } catch (IllegalArgumentException iae) {
//                    logger.info("Uninteresting domain_busy column " + domainName);
                    continue;
                }

                // Current Raw Value (String)
                final String rawCounterValue = domainBusyRawValues[i];
                // Current Counter Value (Number)
                final BigInteger currentValue = new BigInteger(rawCounterValue);
                domainBusyRowScratchPad.getCounterValues().put(busyDomain, currentValue);
            }
            
            // Now that DomainBusy ScratchPad has values, we can estimate the values at Poll Start
            //domainBusyRowScratchPad.calculateEstimateValuesAtPollStart();
        }

        /**
         * Create an OCIEProcessorMetrics ScratchPad for final values reported to OCIE
         */
        private void initOCIEProcessorScratchPad() {
            final MetricsRowScratchPad<OCIEProcessorCounter> previousOCIEValues = getPreviousOCIECounterScratchPad();
            this.ocieProcessorRowScratchPad = new MetricsRowScratchPad<OCIEProcessorCounter>(OCIEProcessorCounter.class,
                    perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp, pollStartTimeStamp, previousOCIEValues,
                    metricsDebugEnabled);
        }


        // Debug CSV support

        /**
         * Get the column headers for debug CSV
         * @param targetObjectName the ZAPI target object name for metric values
         * @return the column headers for debug CSV
         */
        public String getDebugColumnHeaders(final String targetObjectName) {
            final StringBuilder debugBuilder = new StringBuilder("# ");
            debugBuilder.append(targetObjectName).append("\n#\n");

            // Record the N Blade and D Blade domains
            debugBuilder.append("# N Blade Domains: ");
            for (final BusyDomain nBladeDomain : NBLADE_DOMAINS) {
                debugBuilder.append(nBladeDomain).append(' ');
            }
            debugBuilder.append('\n');
            debugBuilder.append("# D Blade Domains: ");
            for (final BusyDomain dBladeDomain : DBLADE_DOMAINS) {
                debugBuilder.append(dBladeDomain).append(' ');
            }
            debugBuilder.append("\n#\n");

            // Now the column headers
            debugBuilder.append(getStaticDebugColumnHeaders());
            for (final ProcessorCounter counterName : EnumSet.allOf(ProcessorCounter.class)) {
                if (ProcessorCounter.domain_busy == counterName) {
                    continue;  // defer domain busy because it is an array of values and needs special handling
                }
                debugBuilder.append("curr raw ").append(counterName).append(",prev raw ").append(counterName);
                debugBuilder.append(",delta raw ").append(counterName);
                debugBuilder.append(",final ").append(counterName).append(",,");
            }

            // Now do domain_busy that is an array of values
            for (final BusyDomain domainName : EnumSet.allOf(BusyDomain.class)) {
                final String fullyQualifiedName = ProcessorCounter.domain_busy + " " + domainName;
                debugBuilder.append("curr raw ").append(fullyQualifiedName).append(",prev raw ").append(fullyQualifiedName);
                debugBuilder.append(",delta raw ").append(fullyQualifiedName);
                debugBuilder.append(",,");
            }

            debugBuilder.append('\n');
            return debugBuilder.toString();
        }

        /**
         * Get the debug CSV rows for all counter values, including ZAPI return values, intermediate domains, and final OCIE values
         * @return the debug CSV rows for all counter values
         */
        public String toDebugCSV() {

            final StringBuilder debugBuilder = new StringBuilder(getStaticDebugColumns());

            // First debug the raw values from ZAPI
            for (final ProcessorCounter counterName : EnumSet.allOf(ProcessorCounter.class)) {
                if (ProcessorCounter.domain_busy == counterName) {
                    continue;  // defer domain busy because it is an array of values and needs special handling
                }
                debugBuilder.append(getCounterDebugCSV(counterName));
            }

            // Now do domain_busy that is an array of values for each BusyDomain
            for (final BusyDomain domainName : EnumSet.allOf(BusyDomain.class)) {
                debugBuilder.append(getBusyDomainDebugCSV(domainName));
            }

            debugBuilder.append('\n');
            return debugBuilder.toString();
        }

        /**
         * Get the counter CSV for a busy domain values (partial row)
         * @param busyDomain the busy domain to generate debug CSV for
         * @return the counter CSV for a busy domain (partial row)
         */
        String getBusyDomainDebugCSV(final BusyDomain busyDomain) {
            final MetricsRowScratchPad<BusyDomain> domainBusyRowScratchPad = getProcessorDomainScratchPad();
            final StringBuilder debugBuilder = new StringBuilder();
            debugBuilder.append(getDebugString(domainBusyRowScratchPad.counterValues.get(busyDomain))).append(',');   // current counter
            debugBuilder.append(getDebugString(domainBusyRowScratchPad.getPreviousValue(busyDomain))).append(',');    // prev counter
            debugBuilder.append(previousValues != null ?
                    getDebugString(domainBusyRowScratchPad.getCounterDelta(busyDomain)) : "").append(",,");   // delta counter
            return debugBuilder.toString();
        }
    }
}
