package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.netapp.oci.netapp.performance.qosworkload.interfaces.data.QosVolumeWorkloadMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.apache.log4j.Logger;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappQosVolumeWorkloadMetrics.QosVolumeWorkloadCounter.*;

public class NetappQosVolumeWorkloadMetrics extends MetricsRowSample<QosVolumeWorkloadMeasurements, QosVolumeWorkloadMeasurements, NetappQosVolumeWorkloadMetrics.QosVolumeWorkloadCounter,
                                                                                                    NetappQosVolumeWorkloadMetrics.OCIEWorkloadCounter> {

    // This is 1 of 2 cases where type parameters <Z, O> are NOT the same
    // NetappProcessorMetrics is the other

    private static final Logger logger = Logger.getLogger(NetappQosVolumeWorkloadMetrics.class);

    /**
     * The processor counter fields targeted in ZAPI
     */
    public static enum QosVolumeWorkloadCounter {
        latency,
        ops,
        total_data,
        read_data,
        read_ops,
        read_latency,
        write_data,
        write_latency,
        write_ops,
        read_io_type_base,
        read_io_type,           // Vector of CSV values per component
    }

    /**
     * read_io_type components: intermediate set of column names used for calculations to final OCIE values
     *
     * <counter-info>
     *   <base-counter>read_io_type_base</base-counter>
     *   <desc>The percentage of reads served from various components (for example, buffer cache, ext_cache or disk).</desc>
     *   <labels>
     *     <label-info>cache,ext_cache,disk,bamboo_ssd,hya_hdd,hya_cache,hya_non_cache</label-info>
     *   </labels>
     *   <name>read_io_type</name>
     *   <privilege-level>diag</privilege-level>
     *   <properties>percent</properties>
     *   <type>array</type>
     *   <unit>percent</unit>
     * </counter-info>
     *
     */
    public static enum ReadIOType {
        read_io_type_base,
        cache,
        ext_cache,
        disk,
        bamboo_ssd,
        hya_hdd,
        hya_cache,
        hya_non_cache,
    }

    /**
     * The counters as required by Server (final columns reported to OCIE)
     */
    public static enum OCIEWorkloadCounter {
        latency,
        ops,
        total_data,
        read_data,
        read_ops,
        read_latency,
        write_data,
        write_latency,
        write_ops,
        read_io_cache,
        read_io_ext_cache,
        read_io_disk,
        read_io_bamboo_ssd,
        read_io_hya_hdd,
        read_io_hya_cache,
        read_io_hya_non_cache,
    }
    
    public NetappQosVolumeWorkloadMetrics(final MetricsRowScratchPad<QosVolumeWorkloadCounter> currentValues, final MetricsRowScratchPad<QosVolumeWorkloadCounter> previousValues) {
        super(QosVolumeWorkloadCounter.class, currentValues, previousValues);
    }

    @Override
    public QosVolumeWorkloadMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        return new QosVolumeWorkloadMeasurements()
                   .setLatency(getNormalizedFloatMetric(OCIEWorkloadCounter.latency))
                   .setOps(getNormalizedFloatMetric(OCIEWorkloadCounter.ops))
                   .setReadData(getNormalizedFloatMetric(OCIEWorkloadCounter.read_data))
                   .setReadLatency(getNormalizedFloatMetric(OCIEWorkloadCounter.read_latency))
                   .setReadOps(getNormalizedFloatMetric(OCIEWorkloadCounter.read_ops))
                   .setTotalData(getNormalizedFloatMetric(OCIEWorkloadCounter.total_data))
                   .setWriteData(getNormalizedFloatMetric(OCIEWorkloadCounter.write_data))
                   .setWriteLatency(getNormalizedFloatMetric(OCIEWorkloadCounter.write_latency))
                   .setWriteOps(getNormalizedFloatMetric(OCIEWorkloadCounter.write_ops));
    }

    @Override
    float getNormalizedFloatMetric(final OCIEWorkloadCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            case latency:
                normalizedMetric = currentValues.getLatencyPerOp(latency, ops);
                break;
            case read_latency:
                normalizedMetric = currentValues.getLatencyPerOp(read_latency, read_ops);
                break;
            case write_latency:
                normalizedMetric = currentValues.getLatencyPerOp(write_latency, write_ops);
                break;

            case total_data:
                normalizedMetric = currentValues.getCounterRatePerSec(total_data);
                break;
            case read_data:
                normalizedMetric = currentValues.getCounterRatePerSec(read_data);
                break;
            case write_data:
                normalizedMetric = currentValues.getCounterRatePerSec(write_data);
                break;
            case ops:
                normalizedMetric = currentValues.getCounterRatePerSec(ops);
                break;
            case read_ops:
                normalizedMetric = currentValues.getCounterRatePerSec(read_ops);
                break;
            case write_ops:
                normalizedMetric = currentValues.getCounterRatePerSec(write_ops);
                break;

            // Custom transforms based on intermediate scratchpad
            case read_io_cache:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.cache);
                break;
            case read_io_ext_cache:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.ext_cache);
                break;
            case read_io_disk:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.disk);
                break;
            case read_io_bamboo_ssd:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.bamboo_ssd);
                break;
            case read_io_hya_hdd:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.hya_hdd);
                break;
            case read_io_hya_cache:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.hya_cache);
                break;
            case read_io_hya_non_cache:
                normalizedMetric = getReadIOPercentFromDeltas(counterName, ReadIOType.hya_non_cache);
                break;

            default:
                normalizedMetric = NULL_FLOAT;
                break;
        }
        return normalizedMetric;
    }

    // Delegate to intermediate scratchpad
    private float getReadIOPercentFromDeltas(final OCIEWorkloadCounter ocieCounter, final ReadIOType numerator) {
        final WorkloadRowMetricsScratchPad workloadMetricsScratchPad = (WorkloadRowMetricsScratchPad)currentValues;
        final MetricsRowScratchPad<ReadIOType> readIOScratchPad = workloadMetricsScratchPad.getReadIOScratchPad();
        final Long deltaNumeratorPollStartEstimate = readIOScratchPad.getCounterDelta(numerator);
        final MetricsRowScratchPad<OCIEWorkloadCounter> ocieScratchPad = workloadMetricsScratchPad.getOCIEWorkloadScratchPad();
        final float readIOPercent = readIOScratchPad.getPercentFromDeltas(numerator, ReadIOType.read_io_type_base);
        readIOScratchPad.setFinalValueForDebug(numerator, readIOPercent);
        ocieScratchPad.setFinalValueForDebug(ocieCounter, readIOPercent);
        return readIOPercent;
    }


    /**
     * This class is a custom scratchpad for workload metrics.
     *
     * Normally, for each metric reported to OCIE, there is a corresponding source metric from ZAPI,
     * and a simple transformation recipe to normalize the value to OCIE expected units.
     *
     * The ScratchPad's Enum normally defines both the metrics to request from ZAPI, and the columns to report to OCIE.
     *
     * The OCIE workload metrics columns are orthogonal to the set of metrics requested from ZAPI,
     * so a custom ScratchPad is required.
     *
     * As an extension of MetricsScratchPad, WorkloadMetricsScratchPad provides support for QosVolumeWorkloadCounter
     * values & calculations.
     *
     * Additionally, we maintain a MetricsScratchPad<ReadIOType> to support ReadIOType values & calculations.
     *
     * Finally, we maintain a MetricsScratchPad<OCIEWorkloadCounter> to support OCIEProcessorCounter values
     * & calculations, which are the only values reported to OCIE.
     */
    public static final class WorkloadRowMetricsScratchPad extends MetricsRowScratchPad<NetappQosVolumeWorkloadMetrics.QosVolumeWorkloadCounter> {

        private MetricsRowScratchPad<ReadIOType> readIOScratchPad = null;
        private MetricsRowScratchPad<OCIEWorkloadCounter> ocieWorkloadScratchPad = null;
        private List<String> discoveredReadIOColumnNames;

        public WorkloadRowMetricsScratchPad(final String perfInstanceUUID, final String inventoryUUID,
                                            final long serverPollStartTimeStamp, final long pollStartTimeStamp,
                                            final MetricsRowScratchPad<QosVolumeWorkloadCounter> previousValues,
                                            final List<String> discoveredReadIOColumnNames, boolean metricsDebugEnabled) {

            super(QosVolumeWorkloadCounter.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp, pollStartTimeStamp,
                    previousValues, metricsDebugEnabled);
            this.discoveredReadIOColumnNames = new ArrayList<String>(discoveredReadIOColumnNames.size() + 1);
            this.discoveredReadIOColumnNames.add(read_io_type_base.name());       // Make sure the base counter is added to the scratchpad
            this.discoveredReadIOColumnNames.addAll(discoveredReadIOColumnNames);
        }

        private MetricsRowScratchPad<ReadIOType> getReadIOScratchPad() {
            if (readIOScratchPad == null) {
                initReadIOScratchPad();
            }
            return readIOScratchPad;
        }

        /**
         * Get the OCIE Workload ScratchPad.  If not initialized, ScatchPad is generated from processor domain ScatchPad values.
         * @return the OCIE processor ScatchPad
         */
        private MetricsRowScratchPad<OCIEWorkloadCounter> getOCIEWorkloadScratchPad() {
            if (ocieWorkloadScratchPad == null) {
                initOCIEWorkloadScratchPad();
            }
            return ocieWorkloadScratchPad;
        }

        /**
         * Get previous WorkloadMetrics ScratchPad.
         * @return previous WorkloadMetrics ScratchPad
         */
        private WorkloadRowMetricsScratchPad getPreviousWorkloadMetrics() {
            return (WorkloadRowMetricsScratchPad) previousValues;
        }

        /**
         * Get previous OCIEProcessorMetrics ScratchPad safely (contained by previous PreviousProcessor ScratchPad).
         * @return previous OCIEProcessorMetrics ScratchPad
         */
        private MetricsRowScratchPad<OCIEWorkloadCounter> getPreviousOCIECounterScratchPad() {
            final WorkloadRowMetricsScratchPad previousProcessorScratchPad = getPreviousWorkloadMetrics();
            return (previousProcessorScratchPad != null)
                   ? previousProcessorScratchPad.getOCIEWorkloadScratchPad()
                   : null;
        }


        /**
         * Create a MetricsScratchPad<ReadIOType> to hold each IO type counter individually
         * read_io_type, read_io_type_base are actually an array of values for each ReadIOType.
         *
         */
        private void initReadIOScratchPad() {

            // get read IO column names from context
            final int readIOColumnCount;
            if (discoveredReadIOColumnNames == null || discoveredReadIOColumnNames.size() == 0) {
                logger.warn("Failed to discover read IO subcolumn names");
                readIOColumnCount = 0;
            } else {
                readIOColumnCount = discoveredReadIOColumnNames.size()-1;    // omit read_io_type_base
            }

            final EnumMap<QosVolumeWorkloadCounter, String> rawValues = getRawValues();
            final String rawReadIONumeratorCSV = rawValues.get(read_io_type);
            // raw value is a CSV array of values for each domain
            final String[] rawReadIONumeratorValues = rawReadIONumeratorCSV.split(",");
            if (rawReadIONumeratorValues.length != readIOColumnCount) {
                logger.warn("Discovered read_io types (" + readIOColumnCount
                        + ") != read_io type raw values (" + rawReadIONumeratorValues.length + ")");
            }

            // Because ZAPI returns these values as a CSV list, MetricsScratchPad doesn't calculate deltas
            // or poll start estimates from raw values.   Now that we've split the list into individual values, we can
            // perform the required calculations on the raw ZAPI values.
            final MetricsRowScratchPad<ReadIOType> previousReadIOScratchPad = getPreviousReadIOScratchPad();
            final MetricsRowScratchPad<ReadIOType> currentReadIOScratchPad =
                    new MetricsRowScratchPad<ReadIOType>(ReadIOType.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp,
                                                     pollStartTimeStamp, previousReadIOScratchPad, metricsDebugEnabled);


            // Populate the ReadIOType ScratchPad with parsed values from the parent QosVolumeWorkloadCounter ScratchPad
            // Manually add read_io_type_base
            this.readIOScratchPad = currentReadIOScratchPad;
            final String rawReadIOTypeBaseString = rawValues.get(read_io_type_base);
            // Current Counter Value (Number)
            final BigInteger baseCounterValue = new BigInteger(rawReadIOTypeBaseString);
            currentReadIOScratchPad.getCounterValues().put(ReadIOType.read_io_type_base, baseCounterValue);

            int i=-1;
            // Skip read_io_type_base, it is in the scratch pad because ll read_io values are compared to it,
            // however it is not in the CSV of read_io values
            for (final String domainName : discoveredReadIOColumnNames.subList(1, discoveredReadIOColumnNames.size())) {
                i++;
                final ReadIOType readIOType;
                try {
                    readIOType = ReadIOType.valueOf(domainName);
                } catch (IllegalArgumentException iae) {
//                    logger.info("Uninteresting domain_busy column " + domainName);
                    continue;
                }
                // Current Raw Value (String)
                final String rawCounterValue = rawReadIONumeratorValues[i];
                // Current Counter Value (Number)
                final BigInteger counterValue = new BigInteger(rawCounterValue);
                currentReadIOScratchPad.getCounterValues().put(readIOType, counterValue);
            }

            // Now that DomainBusy ScratchPad has values, we can estimate the values at Poll Start
           // currentReadIOScratchPad.calculateEstimateValuesAtPollStart();
        }

        final MetricsRowScratchPad<ReadIOType> getPreviousReadIOScratchPad() {
            final WorkloadRowMetricsScratchPad previousProcessorScratchPad = getPreviousWorkloadMetrics();
            return (previousProcessorScratchPad != null)
                   ? previousProcessorScratchPad.getReadIOScratchPad()
                   : null;
        }

        /**
         * Create an OCIEProcessorMetrics ScratchPad for final values reported to OCIE
         */
        private void initOCIEWorkloadScratchPad() {
            final MetricsRowScratchPad<OCIEWorkloadCounter> previousOCIEValues = getPreviousOCIECounterScratchPad();
            this.ocieWorkloadScratchPad = new MetricsRowScratchPad<OCIEWorkloadCounter>(OCIEWorkloadCounter.class,
                    perfInstanceUUID, inventoryUUID,serverPollStartTimeStamp, pollStartTimeStamp, previousOCIEValues,
                    metricsDebugEnabled);
        }


        // Debug CSV support

        /**
         * Get the column headers for debug CSV
         * @param targetObjectName the ZAPI target object name for metric values
         * @return the column headers for debug CSV
         */
        public String getDebugColumnHeaders(final String targetObjectName) {
            final StringBuilder debugBuilder = new StringBuilder("# ");
            debugBuilder.append(targetObjectName).append("\n#\n");

            // Now the column headers
            debugBuilder.append(getStaticDebugColumnHeaders());
            for (final QosVolumeWorkloadCounter counterName : EnumSet.allOf(QosVolumeWorkloadCounter.class)) {
                if (QosVolumeWorkloadCounter.read_io_type == counterName) {
                    continue;  // defer read_io_type because it is an array of values and needs special handling
                }
                debugBuilder.append("curr raw ").append(counterName).append(",prev raw ").append(counterName);
                debugBuilder.append(",delta raw ").append(counterName);
                debugBuilder.append(",final ").append(counterName).append(",,");
            }

            // Now do read_io_type that is an array of ReadIOType values
            for (final ReadIOType ioType : EnumSet.allOf(ReadIOType.class)) {
                if (ReadIOType.read_io_type_base == ioType) {
                    continue;  // read_io_type_base debugged with QosVolumeWorkloadCounter values
                }
                final String fullyQualifiedName = QosVolumeWorkloadCounter.read_io_type + " " + ioType;
                debugBuilder.append("curr raw ").append(fullyQualifiedName).append(",prev raw ").append(fullyQualifiedName);
                debugBuilder.append(",delta raw ").append(fullyQualifiedName);
                debugBuilder.append(",final ").append(fullyQualifiedName).append(",,");
            }

            debugBuilder.append('\n');
            return debugBuilder.toString();
        }

        /**
         * Get the debug CSV rows for all counter values, including ZAPI return values, intermediate read IO types, and final OCIE values
         * @return the debug CSV rows for all counter values
         */
        public String toDebugCSV() {
            final StringBuilder debugBuilder = new StringBuilder(getStaticDebugColumns());

            // First debug the raw values from ZAPI
            for (final QosVolumeWorkloadCounter counterName : EnumSet.allOf(QosVolumeWorkloadCounter.class)) {
                if (QosVolumeWorkloadCounter.read_io_type == counterName) {
                    continue;  // defer read_io_type because it is an array of values and needs special handling
                }
                debugBuilder.append(getCounterDebugCSV(counterName));
            }

            final MetricsRowScratchPad<ReadIOType> readIOScratchPad = getReadIOScratchPad();
            // Now do read_io_type that are an arrays of values for each ReadIOType
            for (final ReadIOType ioType : EnumSet.allOf(ReadIOType.class)) {
                if (ReadIOType.read_io_type_base == ioType) {
                    continue;  // read_io_type_base debugged with QosVolumeWorkloadCounter
                }
                debugBuilder.append(readIOScratchPad.getCounterDebugCSV(ioType));
            }

            debugBuilder.append('\n');
            return debugBuilder.toString();
        }
    }
}
