package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics;

import com.netapp.oci.netapp.performance.wafl.interfaces.data.WaflMeasurements;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import org.apache.log4j.Logger;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.NetappWaflMetrics.CpPhaseTimes.*;

import java.math.BigInteger;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.List;

/**
 * Processor Metrics are very different from other metrics.
 *
 * Normally, for each metric reported to OCIE (MetricsRowSample.E), there is a corresponding source metric from ZAPI
 * (MetricsRowSample.Z), and a simple transformation recipe to normalize the value to OCIE expected units
 * (rate or ratio to other counter/column delta)
 *
 * The MetricsRowSample'S Enum <Z> normally defines both the metrics to request from ZAPI, and the columns to report to OCIE
 * (Z == O).
 *
 * The required OCIE Processor metrics columns <O> are orthogonal to the set of metrics requested from ZAPI <Z>,
 * so specific logic is required.
 *
 * This Metrics class uses 3 Enums:
 * 1) ZAPIProcessorCounter <Z> defines the metrics to request from ZAPI
 * 2) ProcessorBusyDomain is used to interpret domain_busy, which is a CSV list of values per domain, rather than a single value
 * 3) OCIEProcessorCounter <O> represents the columns required by OCIE, including aggregations of subsets of domain_busy values
 */
public class NetappWaflMetrics extends MetricsRowSample<WaflMeasurements, WaflMeasurements, NetappWaflMetrics.WaflCounter,
        NetappWaflMetrics.OCIEWaflCounter> {

    // This is 1 of 3 cases where type parameters <Z, E> are NOT the same
    // NetappQosVolumeWorkloadMetrics and NetappProcessorMetrics are the others

    @SuppressWarnings({"UnusedDeclaration"})
    private static final Logger logger = Logger.getLogger(NetappWaflMetrics.class);

    /**
     * The wafl counter fields targeted in ZAPI
     */
    public enum WaflCounter {
        total_cp_msecs,
        cp_phase_times, // Vector of CSV values per component
        wafl_cp_blks_read,
        cp_count
    }

    /**
     * The list of all processor busy domain names (for domain_busy).  SN supports all domain names, while RR supports a subset of domain names
     * See:
     * <perf-object-counter-list-info>
     * <objectname>wafl</objectname>
     * </perf-object-counter-list-info>
     *
     * <labels>
     * <label-info>SETUP,PRE_P0,P0_SNAP_DEL,P1_CLEAN,P1_QUOTA,IPU_DISK_ADD,P2V_INOFILE,P2V_INO_PUB,P2V_INO_PRI,P2V_FSINFO,P2V_DLOG1,P2V_DLOG2,P2V_IPU_BITMAP_GROW,P2V_REFCOUNT,P2V_TOPAA,P2V_DF_SCORES_SUB,P2V_BM,P2V_SNAP,P2V_DF_SCORES,P2V_VOLINFO,P2V_CONT,P2A_INOFILE,P2A_INO,P2A_DLOG1,P2A_HYA,P2A_DLOG2,P2A_FSINFO,P2A_IPU_BITMAP_GROW,P2A_REFCOUNT,P2A_TOPAA,P2A_BM,P2A_SNAP,P2A_VOLINFO,P2_FLUSH,P2_FINISH,P3_WAIT,P3V_VOLINFO,P3A_VOLINFO,P3_FINISH,P4_FINISH,P5_FINISH</label-info>
     * </labels>
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public enum CpPhaseTimes {
        total_cp_msecs,
        SETUP,
        PRE_P0,
        P0_SNAP_DEL,
        P1_CLEAN,
        P1_QUOTA,
        IPU_DISK_ADD,
        P2V_INOFILE,
        P2V_INO_PUB,
        P2V_INO_PRI,
        P2V_FSINFO,
        P2V_DLOG1,
        P2V_DLOG2,
        P2V_IPU_BITMAP_GROW,
        P2V_REFCOUNT,
        P2V_TOPAA,
        P2V_DF_SCORES_SUB,
        P2V_BM,
        P2V_SNAP,
        P2V_DF_SCORES,
        P2V_VOLINFO,
        P2V_CONT,
        P2A_INOFILE,
        P2A_INO,
        P2A_DLOG1,
        P2A_HYA,
        P2A_DLOG2,
        P2A_FSINFO,
        P2A_IPU_BITMAP_GROW,
        P2A_REFCOUNT,
        P2A_TOPAA,
        P2A_BM,
        P2A_SNAP,
        P2A_VOLINFO,
        P2_FLUSH,
        P2_FINISH,
        P3_WAIT,
        P3V_VOLINFO,
        P3A_VOLINFO,
        P3_FINISH,
        P4_FINISH,
        P5_FINISH,
        cp_count,
        wafl_cp_blks_read,
    }

    /**
     * The counters as required by Server
     */
    public enum OCIEWaflCounter {
        total_cp_msecs, // total cp time
        P2_FLUSH, // cp time in p2Flush
        wafl_cp_blks_read,
        cp_count,
    }


    /**
     * Get a ProcessorMetrics MetricsRowScratchPad.  Contains ZAPI counters, intermediate busy values per domain, and final values reported to OCIE.
     *
     * @param currentValues  the current raw values from ZAPI
     * @param previousValues the previous raw values from ZAPI
     */
    public NetappWaflMetrics(final MetricsRowScratchPad<WaflCounter> currentValues, final MetricsRowScratchPad<WaflCounter> previousValues) {
        super(WaflCounter.class, currentValues, previousValues);
    }

    /**
     * Get the normalized float metric for report to OCIE.   This is NOT an override, because set of reported
     * metrics is derived from but not equal to the set of collected metrics
     *
     * @param counterName The enum constant for the desired counter
     *
     * @return The measurement
     */
    float getNormalizedFloatMetric(final OCIEWaflCounter counterName) {
        final float normalizedMetric;
        switch (counterName) {
            default:
                normalizedMetric = MetricsRowSample.NULL_FLOAT;
                break;
        }
        return normalizedMetric;
    }

    private float getDelta(final CpPhaseTimes cpPhaseTime) {
        final WaflMetricsRowScratchPad waflMetricsRowScratchPad = (WaflMetricsRowScratchPad) currentValues;
        final MetricsRowScratchPad<CpPhaseTimes> cpPhaseTimesScratchPad = waflMetricsRowScratchPad.getCpPhaseTimesScratchPad();
        return (float) cpPhaseTimesScratchPad.getCounterDelta(cpPhaseTime);
    }
    
    private float getPercentFromDeltas(final CpPhaseTimes cpPhaseTimeNumerator, final CpPhaseTimes cpPhaseTimeDenominator){
    	final float deltaNumerator = getDelta(cpPhaseTimeNumerator);
    	final float deltaDenominator = getDelta(cpPhaseTimeDenominator);
    	final float percent = (deltaDenominator == 0 || deltaNumerator == 0)
    						  ? 0
    						  : (deltaDenominator > 0 && deltaNumerator > 0)
    						  		? (float) (100d * (double) deltaNumerator / (double) deltaDenominator)
    						  		: MetricsRowSample.NULL_FLOAT;
    	return percent;
    }


    @Override
    public WaflMeasurements normalizeMetrics(final NetAppDefaultStorageContext context) {
        final WaflMeasurements m = new WaflMeasurements();
        m.setTotalCpMsecs(getDelta(CpPhaseTimes.total_cp_msecs));
        m.setP2Flush(getPercentFromDeltas(CpPhaseTimes.P2_FLUSH, CpPhaseTimes.total_cp_msecs));
        m.setCpCount(getDelta(CpPhaseTimes.cp_count));
        m.setWaflCpBlksRead(getDelta(CpPhaseTimes.wafl_cp_blks_read));
        return m;
    }

    /**
     * This class is a custom scratchpad for cpPhaseTime metrics.
     *
     * Normally, for each metric reported to OCIE, there is a corresponding source metric from ZAPI,
     * and a simple transformation recipe to normalize the value to OCIE expected units.
     *
     * The ScratchPad's Enum normally defines both the metrics to request from ZAPI, and the columns to report to OCIE.
     *
     * The OCIE CP Phase Time metrics columns are orthogonal to the set of metrics requested from ZAPI,
     * so a custom ScratchPad is required.
     *
     * As an extension of MetricsRowScratchPad, WaflMetricsRowScratchPad provides support for cp Phase times
     * values & calculations.
     *
     * Additionally, we maintain a MetricsRowScratchPad<CpPhaseTimes> to support cp Phase time values & calculations.
     *
     * Finally, we maintain a MetricsRowScratchPad<OCIEWaflCounter> to support OCIEWaflCounter values
     * & calculations, which are the only values reported to OCIE.
     */
    public static final class WaflMetricsRowScratchPad extends MetricsRowScratchPad<WaflCounter> {

        private MetricsRowScratchPad<CpPhaseTimes> cpPhaseTimesRowScratchPad = null;
        private MetricsRowScratchPad<OCIEWaflCounter> ocieWaflRowScratchPad = null;
        private List<String> discoveredWaflCpPhaseTimesColumnNames = null;

        public WaflMetricsRowScratchPad(final String perfInstanceUUID, final String inventoryUUID,
                                        final long serverPollStartTimeStamp, final long pollStartTimeStamp,
                                         final MetricsRowScratchPad<WaflCounter> previousValues,
                                        final List<String> discoveredWaflCpPhaseTimesColumnNames, boolean metricsDebugEnabled) {

            super(WaflCounter.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp, pollStartTimeStamp,
                  previousValues, metricsDebugEnabled);
            this.discoveredWaflCpPhaseTimesColumnNames = discoveredWaflCpPhaseTimesColumnNames;
        }

        /**
         * Get the processor domain ScatchPad.  If not initialized, ScatchPad is generated from domain_busy CSV value
         *
         * @return the processor domain ScatchPad
         */
        public MetricsRowScratchPad<CpPhaseTimes> getCpPhaseTimesScratchPad() {
            if (cpPhaseTimesRowScratchPad == null) {
                initCpPhaseTimeScratchPad();
            }
            return cpPhaseTimesRowScratchPad;
        }

        /**
         * Get the OCIEProcessor ScratchPad.  If not initialized, ScatchPad is generated from processor domain ScatchPad values.
         *
         * @return the OCIE processor ScatchPad
         */
        private MetricsRowScratchPad<OCIEWaflCounter> getOCIEProcessorScratchPad() {
            if (ocieWaflRowScratchPad == null) {
                initOCIEProcessorScratchPad();
            }
            return ocieWaflRowScratchPad;
        }

        /**
         * Get previous ProcessorMetrics ScratchPad.
         *
         * @return previous ProcessorMetrics ScratchPad
         */
        private WaflMetricsRowScratchPad getPreviousProcessorMetrics() {
            return (WaflMetricsRowScratchPad) previousValues;
        }

        /**
         * Get previous BusyDomain ScratchPad safely (contained by previous PreviousProcessor ScratchPad).
         *
         * @return previous DomainBusy ScratchPad
         */
        private MetricsRowScratchPad<CpPhaseTimes> getPreviousCpPhaseTimesScratchPad() {
            final WaflMetricsRowScratchPad previousProcessorRowScratchPad = getPreviousProcessorMetrics();
            return (previousProcessorRowScratchPad != null)
                   ? previousProcessorRowScratchPad.getCpPhaseTimesScratchPad()
                   : null;
        }

        /**
         * Get previous OCIEProcessorMetrics ScratchPad safely (contained by previous PreviousProcessor ScratchPad).
         *
         * @return previous OCIEProcessorMetrics ScratchPad
         */
        private MetricsRowScratchPad<OCIEWaflCounter> getPreviousOCIECounterScratchPad() {
            final WaflMetricsRowScratchPad previousProcessorRowScratchPad = getPreviousProcessorMetrics();
            return (previousProcessorRowScratchPad != null)
                   ? previousProcessorRowScratchPad.getOCIEProcessorScratchPad()
                   : null;
        }

        /**
         * Create a MetricsRowScratchPad<ProcessorBusyDomain> to hold each domain counter individually
         * domain_busy is actually an array of values for each ProcessorDomain.
         *
         * ZAPI effectively returns a Map of ZAPIProcessorCounter to List of ProcessorDomain values.
         */
        private void initCpPhaseTimeScratchPad() {

            final int cpPhaseTimeColumnCount =  discoveredWaflCpPhaseTimesColumnNames.size();  //CpPhaseTimes.values().length - 3; // -1 for total_cp_msecs which are not returned by ZAPI but returned by the metric

            // Because ZAPI returns these values as a CSV list, MetricsRowScratchPad doesn't calculate deltas
            // or poll start estimates.   Now that we've split the list into individual values, we can
            // perform the required calculations on the raw ZAPI values.
            final MetricsRowScratchPad<CpPhaseTimes> previousCpPhaseTimeValues = getPreviousCpPhaseTimesScratchPad();
            this.cpPhaseTimesRowScratchPad =
            		new MetricsRowScratchPad<CpPhaseTimes>(CpPhaseTimes.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp,
            				pollStartTimeStamp, previousCpPhaseTimeValues, metricsDebugEnabled);

            final EnumMap<WaflCounter, String> rawValues = getRawValues();
            final String rawDomainBusyValues = rawValues.get(WaflCounter.cp_phase_times);         // Raw value is CSV list
            // raw value is a CSV array of values for each domain
            final String[] cpPhaseTimeRawValues = rawDomainBusyValues.split(",");
            if (cpPhaseTimeRawValues.length != cpPhaseTimeColumnCount) {
            	logger.warn("Discovered cp_phase_times Column Names (" + cpPhaseTimeColumnCount
            			+ ") != domain busy raw values (" + cpPhaseTimeRawValues.length + ")");
            	cpPhaseTimesRowScratchPad.getCounterValues().put(P2_FLUSH, MetricsRowSample.NULL_BIG_INT);
            }
            
            else {
            	int i = -1;
            	for (final String cpPhaseTimesColumnName : discoveredWaflCpPhaseTimesColumnNames) {
            		i++;
            		if(cpPhaseTimesColumnName.equals("P2_FLUSH")){
            			// Current Raw Value (String)
            			final String rawCounterValue = cpPhaseTimeRawValues[i];// P2_FLUSH
            			cpPhaseTimesRowScratchPad.getRawValues().put(P2_FLUSH, rawCounterValue);
            			// Current Counter Value (Number)
            			final BigInteger counterValue = new BigInteger(rawCounterValue);
            			cpPhaseTimesRowScratchPad.getCounterValues().put(P2_FLUSH, counterValue);
            		}
            		else
            			continue;

            	}
            }
            {
                // Current Counter Value (Number)
                final BigInteger counterValue = getCounterValues().get(WaflCounter.total_cp_msecs);
                cpPhaseTimesRowScratchPad.getCounterValues().put(total_cp_msecs, counterValue);
            }
            {
            	//Current Raw Value (String)
            	final String rawCounterValue = getRawValues().get(WaflCounter.cp_count);
            	BigInteger counterValue = BigInteger.ZERO;
            	if(rawCounterValue != null){
            		final String[] CpCountRawValues = rawCounterValue.split(",");
            		for(int i=0 ; i < CpCountRawValues.length ; i++){
            			BigInteger cpCountRawBigIntValue = new BigInteger(CpCountRawValues[i]);
            			counterValue = counterValue.add(cpCountRawBigIntValue);
            		}
            	}
            	cpPhaseTimesRowScratchPad.getCounterValues().put(cp_count, counterValue);
            }
            {
                // Current Counter Value (Number)
                final BigInteger counterValue = getCounterValues().get(WaflCounter.wafl_cp_blks_read);
                cpPhaseTimesRowScratchPad.getCounterValues().put(wafl_cp_blks_read, counterValue);
            }

            // Now that DomainBusy ScratchPad has values, we can estimate the values at Poll Start
           // cpPhaseTimesRowScratchPad.calculateEstimateValuesAtPollStart();
        }

        /**
         * Create an OCIEProcessorMetrics ScratchPad for final values reported to OCIE
         */
        private void initOCIEProcessorScratchPad() {
            final MetricsRowScratchPad<OCIEWaflCounter> previousOCIEValues = getPreviousOCIECounterScratchPad();
            this.ocieWaflRowScratchPad = new MetricsRowScratchPad<OCIEWaflCounter>(OCIEWaflCounter.class, perfInstanceUUID, inventoryUUID, serverPollStartTimeStamp,
                                                                                   pollStartTimeStamp, previousOCIEValues, metricsDebugEnabled);
        }

        // Debug CSV support

        /**
         * Get the column headers for debug CSV
         *
         * @param targetObjectName the ZAPI target object name for metric values
         *
         * @return the column headers for debug CSV
         */
        public String getDebugColumnHeaders(final String targetObjectName) {

            // this should just be the final OCIEWaflCounter column names
            final StringBuilder debugBuilder = new StringBuilder("# ");
            debugBuilder.append(targetObjectName).append("\n#\n");
            for (final CpPhaseTimes cpPhaseTime : CpPhaseTimes.values()) {
                debugBuilder.append(cpPhaseTime).append(' ');
            }

            // Now the column headers
            debugBuilder.append(getStaticDebugColumnHeaders());
            for (final WaflCounter counterName : EnumSet.allOf(WaflCounter.class)) {
                if (WaflCounter.cp_phase_times == counterName) {
                    continue;  // defer cp phase times because it is an array of values and needs special handling
                } else {
                	debugBuilder.append("curr raw ").append(counterName).append(",prev raw ").append(counterName);
                	debugBuilder.append(",delta raw ").append(counterName);
                	debugBuilder.append(",,");
                }
            }

            // Now do P2_Flush
            for (final CpPhaseTimes domainName : EnumSet.range(CpPhaseTimes.P2_FLUSH, CpPhaseTimes.P2_FLUSH)) {
                final String fullyQualifiedName = WaflCounter.cp_phase_times + " " + domainName;
                debugBuilder.append("curr raw ").append(fullyQualifiedName).append(",prev raw ").append(fullyQualifiedName);
                debugBuilder.append(",delta raw ").append(fullyQualifiedName);
                debugBuilder.append(",,");
            }

            debugBuilder.append('\n');
            return debugBuilder.toString();
        }

        /**
         * Get the debug CSV rows for all counter values, including ZAPI return values, intermediate domains, and final OCIE values
         *
         * @return the debug CSV rows for all counter values
         */
        public String toDebugCSV() {

            final StringBuilder debugBuilder = new StringBuilder(getStaticDebugColumns());

            // First debug the raw values from ZAPI
            for (final WaflCounter counterName : EnumSet.allOf(WaflCounter.class)) {
                if (WaflCounter.cp_phase_times == counterName) {
                    continue;  // defer cp_phase_times because it is an array of values and needs special handling
                }
                else if(WaflCounter.cp_count == counterName){
                	debugBuilder.append(getCpPhaseTimesDebugCSV(CpPhaseTimes.cp_count));
                }
                else {
                    debugBuilder.append(getDebugString(counterValues.get(counterName))).append(',');             // current counter
                    debugBuilder.append(getDebugString(getPreviousValue(counterName))).append(',');              // prev counter
                    debugBuilder.append(previousValues != null ?
                            getDebugString(getCounterDelta(counterName)) : "").append(",,");                    // delta counter
                }
            }
            
            // Now do domain_busy that is an array of values for each BusyDomain
            for (final CpPhaseTimes cpPhaseTimes : EnumSet.range(CpPhaseTimes.P2_FLUSH, CpPhaseTimes.P2_FLUSH)) {
                debugBuilder.append(getCpPhaseTimesDebugCSV(cpPhaseTimes));
            }

            debugBuilder.append('\n');
            return debugBuilder.toString();
        }

        /**
         * Get the counter CSV for a cpPhaseTimes values (partial row)
         *
         * @param cpPhaseTimes the cpPhaseTimes to generate debug CSV for
         *
         * @return the counter CSV for a cpPhaseTimes (partial row)
         */
        String getCpPhaseTimesDebugCSV(final CpPhaseTimes cpPhaseTimes) {
            final MetricsRowScratchPad<CpPhaseTimes> domainBusyRowScratchPad = getCpPhaseTimesScratchPad();
            final StringBuilder debugBuilder = new StringBuilder();
            debugBuilder.append(getDebugString(domainBusyRowScratchPad.counterValues.get(cpPhaseTimes))).append(',');   // current counter
            debugBuilder.append(getDebugString(domainBusyRowScratchPad.getPreviousValue(cpPhaseTimes))).append(',');    // prev counter
            debugBuilder.append(previousValues != null ?
                    getDebugString(domainBusyRowScratchPad.getCounterDelta(cpPhaseTimes)) : "").append(",,");   // delta counter
            return debugBuilder.toString();
        }
    }
}
