package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderModel;
import com.onaro.sanscreen.acquisition.framework.partialpoll.PartialPollConfigurationRetriever;
import com.onaro.sanscreen.acquisition.framework.partialpoll.generated.ZapiCommandListType;
import com.onaro.sanscreen.acquisition.framework.partialpoll.generated.ZapiCommandType;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.util.VersionComparator;
import com.onaro.util.VersionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * The partial poll context represents the information needed to determine if the results of a zapi should be retrieved
 * from the onTap Cluster or Partial Poll Cache.
 */
public class PartialPollContext implements BuilderModel
{
  private Logger logger = LoggerFactory.getLogger( PartialPollContext.class );

  private static final String fileExtension = "xml";

  private final String dataSourceName;
  private final String baseDirName;
  private final Set<String> supportedZapi = new HashSet<>();
  private final Set<String> readFromDevice = new HashSet<>();
  private final Set<String> readFromCache = new HashSet<>();
  private final Map<String, ChangeResult> changeResultMap = Collections.synchronizedMap( new HashMap<String, ChangeResult>( ) );

  private List<String> supportedZapiNames;
  private boolean globallyEnabled = false;
  private boolean enabled = false;
  private PollType pollType = PollType.Full;
  private Calendar nextPollTime;
  private GregorianCalendar fullPollDailyTime;
  private boolean initialPoll = true;

  public enum PollType{Full, Partial, Daily}

  public PartialPollContext( final String dataSourceName, final String baseDirName )
  {
    this(dataSourceName, null,baseDirName);
  }

  public PartialPollContext(final String dataSourceName, final List<String> zapiNames, final String baseDirName  )
  {
    //Validation
    if( baseDirName == null ) throw new PartialPollException( "Partial Polling Base Directory is not set" );
    if( dataSourceName == null ) throw new PartialPollException( "Partial Polling DataSource name is not set" );

    //Setters
    this.dataSourceName = dataSourceName;
    this.supportedZapiNames = zapiNames == null ? new ArrayList<String>() : zapiNames ;
    this.baseDirName = baseDirName;
    this.supportedZapi.addAll( supportedZapiNames );
    PartialPollConfigurationRetriever ppcRetriever = PartialPollConfigurationRetriever.getInstance();
    fullPollDailyTime = ppcRetriever.getFullPollDailyTime().toGregorianCalendar();

    //create the next daily poll time
    nextPollTime = Calendar.getInstance();
    nextPollTime.set( Calendar.HOUR_OF_DAY, fullPollDailyTime.get( Calendar.HOUR_OF_DAY ) );
    nextPollTime.set( Calendar.MINUTE, fullPollDailyTime.get( Calendar.MINUTE ) );
    nextPollTime.set( Calendar.SECOND, 0 );
    nextPollTime.set( Calendar.MILLISECOND, 0 );
  }

  public void addReadFromCache(String zapiName)
  {
    readFromCache.add( zapiName );
  }

  public void addReadFromDevice(String zapiName)
  {
    readFromDevice.add( zapiName );
  }

  public PollType getPollType()
  {
    return pollType;
  }

  public void setPollType(PollType pollType)
  {
    this.pollType = pollType;
  }

  public boolean isGloballyEnabled()
  {
    return globallyEnabled;
  }

  public void setGloballyEnabled(boolean globallyEnabled)
  {
    this.globallyEnabled = globallyEnabled;
  }

  public String getBaseDirName()
  {
    return baseDirName;
  }

  public List<String> getSupportedZapiNames()
  {
    return supportedZapiNames;
  }

  public void setSupportedZapiNames(List<String> supportedZapiNames)
  {
    this.supportedZapiNames = supportedZapiNames;
    for( String name : supportedZapiNames )
      supportedZapi.add( name );
  }

  public boolean isSupportedZapi(String zapiName)
  {
    return supportedZapi.contains( zapiName );
  }

  /**
   * @return true if the the partial poll cache is being used, else false.
   */
  public boolean isEnabled()
  {
    return globallyEnabled && enabled;
  }

  /**
   * @param enabled Set true if the the partial poll cache is to be used, if not set than to false.
   */
  public void setEnabled(boolean enabled)
  {
    this.enabled = enabled;
  }

  public String getFileExtension()
  {
    return fileExtension;
  }

  /**
   * Returns if the current state for a zapi is different then the previous, ultimately this means that the
   * Partial poll cache for this is zapi is stale.
   *
   * @param zapiName Name of the zapi data will be retrieved for
   *
   * @return true if the the current change data is different from the previous state, else false.
   */
  public boolean needsUpdate(String zapiName)
  {
    //if partial poll disabled then everything needs updating
    if( !enabled ) return true;

    //full poll or daily means reload cache
    if( pollType == PollType.Full || pollType == PollType.Daily ) return true;

    //if zapi not supported then it does not need updating in the cache
    if( !isSupportedZapi( zapiName ) ) return false;

    //pair is the last two change records for this zapi y is null if only 1 poll has been performed
    ChangeResult changeResult = changeResultMap.get( zapiName );

    //if the pair is null then this is the first time after start up or this is not a partial poll supported zapi
    return changeResult == null || changeResult.isChanged();
  }

  /**
   * Updates the current state of the context so the connection can determine if the results of a zapi should be retrieved
   * from the onTap Cluster or Partial Poll Cache.
   *
   * @param currentChange zapi change result
   */
  public void addResult( ChangeResult currentChange )
  {
    //validation
    if( currentChange == null ) throw new PartialPollException( "Current change can't be null" );

    //Ignore if sure partial polling is disabled
    if( enabled )
    {
      //Get the pair of changesResults if one already exists and move the x to the y and put the new one if y
      ChangeResult previousChange = changeResultMap.get( currentChange.getZapiName() );

      //If the previous change is null the value needs to update or if the change time is different from the previous change
      if( previousChange == null || currentChange.getChangeTime() > previousChange.getChangeTime() )
        currentChange.setChanged( true );
      else
        currentChange.setChanged( false );

      //Update the map with the zapi change info
      changeResultMap.put( currentChange.getZapiName(), currentChange );
    }
  }

  /**
   * @param clusterVersion the string version of the cluster
   * @param properties     properties specifying the type of poll
   */
  public void initialize(String clusterVersion, Properties properties)
  {
    //clear old state information
    cleanUp();

    //Disable because the last run may have set it to enabled; Partial poll contexts are long lived objects.
    setEnabled( false );

    //Make sure that partial poll should be enabled
    if( isGloballyEnabled() )
    {
      //set type of poll default to partial
      setPollType( PartialPollContext.PollType.Partial );
      if( initialPoll )
        setPollType( PollType.Full );
      else if( properties != null && properties.get( "type" ) != null && properties.get( "type" ).equals( "f" ) )
        setPollType( PollType.Full );
      else if( shouldPerformFullDailyPoll() )
        setPollType( PollType.Daily );

      //load configuration version
      String version = convertToVersionType( clusterVersion );
      if( version == null )
      {
        setPollType( PollType.Full );
        logger.info( "Disabling partial poll because version " + clusterVersion + " is unsupported." );
      }
      else
      {
        //Check for supported zapis for this version
        PartialPollConfigurationRetriever ppcRetriever = PartialPollConfigurationRetriever.getInstance();
        ZapiCommandListType versionedZapis = ppcRetriever.getZapicalls().get( version );
        if( versionedZapis == null )
          logger.info( "Partial Polling is not configured for version " + clusterVersion );
        else
        {
          //Load supported zapis
          List<String> supportedZapiNames = new ArrayList<>();
          for( ZapiCommandType zapiCommand : versionedZapis.getZAPI() )
            supportedZapiNames.add( zapiCommand.getName() );
          setSupportedZapiNames( supportedZapiNames );

          //Enable partial polling for this data source
          setEnabled( true );
        }
      }
    }

    logger.info( dataSourceName + " " + pollType + " Poll started." );
  }

  /**
   * Clear old state information from last execute
   */
  public void cleanUp()
  {
    //clear old state information
    readFromCache.clear();
    readFromDevice.clear();
  }

  /**
   * Must be called after complete a full poll
   */
  public void completed()
  {
    logger.info( dataSourceName + " " + pollType + " Poll completed." );

    //After the first poll has completed make sure we don't repeat the init poll logic
    initialPoll = false;

    //clean up state info
    cleanUp();

    //if this was a daily poll update the next full daily poll
    if( pollType == PollType.Daily )
      nextPollTime.add( Calendar.DATE, 1 );
  }

  /**
   * Convert Cluster String version to Partial Poll <major>.<minor>.<patch>
   *
   * @param clusterVersion version
   * @return VersionType
   */
  private String convertToVersionType(String clusterVersion)
  {
    final VersionComparator versionComparator = new VersionComparator();
    String version = null; //Earlier versions then FS 1 are unsupported
    if( versionComparator.compare( clusterVersion, Cluster.FULL_STEAM_1 ) >= 0 )
      version = VersionUtils.versionDigits( clusterVersion );
    return version;
  }

  /**
   *
   * @return true if the daily full poll needs to be executed
   */
  public boolean shouldPerformFullDailyPoll()
  {
    //Check if the time for the next poll has passed
    return Calendar.getInstance().after( nextPollTime );
  }

  /**
   * @return true if the runtime values need to be read from the device
   */
  public boolean needsRuntimeUpdate( String zapiName )
  {
    return supportedZapi.contains( zapiName ) && readFromCache.contains( zapiName );
  }
}
