package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.partialpoll;

import com.onaro.sanscreen.acquisition.framework.filecache.DirectoryCacheKey;
import com.onaro.sanscreen.acquisition.framework.filecache.FileCacheEngine;
import com.onaro.sanscreen.acquisition.framework.filecache.StringCacheData;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import org.apache.commons.codec.digest.DigestUtils;

/**
 * ZAPIPartialPollCache is a singleton object using the FileCacheEngine to persist zapi data.
 *
 * Stores zapi data under the $home/partial_polling_cache/<cluster_name>/<zapi_name.iteration.fileExtension>
 */
public class ZAPIPartialPollCache
{
  private static final String PARTIAL_POLL_BASE_DIR = "partial_polling_cache";
  private static final ZAPIPartialPollCache instance = new ZAPIPartialPollCache(); //make singleton

  private FileCacheEngine fileCache = new FileCacheEngine();


  /**
   * To use as singleton make the constructors private
   */
  private ZAPIPartialPollCache()
  {
  }

  /**
   * @return the single instance of this class
   */
  public static ZAPIPartialPollCache getInstance()
  {
    return instance;
  }

  /**
   * Read the string data associated with this cluster input type and index
   *
   * @param context The context
   * @param clusterName name of the cluster this zapi data belongs too
   * @param zapiName The name of the zapi data to be retrieved
   * @param request The request for this zapi
   *
   * @return The zapi data that was stored for this call
   * @throws IOException
   */
    public String read(PartialPollContext context, String clusterName, String zapiName, String request ) throws IOException
    {
        String cachedData = null;
        if( context != null && context.isEnabled()  )
        {
            if( context.isSupportedZapi( zapiName ) && !context.needsUpdate( zapiName ) )
            {
                DirectoryCacheKey key = createDirectoryKey( context, clusterName );
                key.setFileName( constructFileName( context, zapiName, request ) );
                StringCacheData data = null;
                try
                {
                    data = fileCache.read( key );
                }
                catch( FileNotFoundException e )
                {
                    //ignore file not found because this just means that the data has not been collected yet
                }
                cachedData = data != null ? data.getData() : null;
            }
        }

        return cachedData;
    }

  /**
   * Save data for the zapi.
   *
   * @param context The partial poll context
   * @param clusterName The name of the cluster data is to be retrieved for
   * @param zapiName The name of the zapi data to be retrieved
   * @param request The request for this zapi
   * @param data The string data to be saved in the files
   * @throws IOException
   */
    public boolean save(PartialPollContext context, String clusterName, String zapiName, String request, String data) throws IOException
    {
        boolean persisted = false;
        if( context != null && context.isEnabled() )
        {
            //ignore all unsupported zapis
            if( context.isSupportedZapi( zapiName ) )
            {
                //only supported zapis
                DirectoryCacheKey key = createDirectoryKey( context, clusterName );
                key.setFileName( constructFileName( context, zapiName, request ) );

                //Update the saved data only if it has changed or the file is missing
                if( context.needsUpdate( zapiName ) || !fileCache.exists( key ) )
                {
                    fileCache.save( key, new StringCacheData( data ) );
                    persisted = true;
                }
            }
        }

        return persisted;
    }


    /**
   * Standardize the file name construction
   *
   * @param context The partial poll context
   * @param zapiName The name of the zapi data to be retrieved
   * @param request The request for this zapi
   *
   * @return The name file name in the format zapi_name.iteration.fileExtension
   */
    private String constructFileName( PartialPollContext context, String zapiName, String request )
    {
        String hash = DigestUtils.sha256Hex(request);
        return zapiName + "." + hash + "." + context.getFileExtension();
    }


    /**
   * Create the base part of a key that represents a file name or directory name stored in the partial poll cache
   *
   * @param clusterName the base part of a key
   *
   * @return a new DirectoryCacheKey contain the partial poll base plus the clusterName
   *
   * @throws IOException
   */
  public DirectoryCacheKey createDirectoryKey(PartialPollContext context, String clusterName) throws IOException
  {
    //validation
    if( clusterName == null ) throw new PartialPollException( "Cluster Name can't be null" );

    //check for OS paths
    String base = new File( context.getBaseDirName() ).getCanonicalPath() + File.separator + PARTIAL_POLL_BASE_DIR;
    DirectoryCacheKey key = new DirectoryCacheKey( base );
    key.addDir( clusterName );
    return key;
  }
}
