package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.builders;


import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.GlobalScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;
import com.onaro.sanscreen.acquisition.schema.Key;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * Base class for all the synthetic builders which merge counter data from existing
 * scratchPads by implementing {@link #generateMetricsRowScratchPads}. The derived builders
 * report cooked counters and optionally report report raw counters.
 * 
 * @param <S> Measurement class type for cooked counters.
 * @param <R> the raw performance Sample to report to the server
 * @param <K> The foundation UUID, used to determine the container key e.g. Cluster, VServer etc.
 * @param <Z> Enum with the list of counters names
 */
public abstract class SyntheticBuilderBase<S, R, K extends Key, Z extends Enum<Z>> {
    private static final Logger logger = LoggerFactory.getLogger(SyntheticBuilderBase.class);

    protected NetAppOCIEPerformancePackage reportingPackage;
    
    protected SyntheticBuilderBase(final NetAppOCIEPerformancePackage reportingPackage) {
        this.reportingPackage = reportingPackage;
    }

    /**
     * Get and report raw and cooked counters.
     * 
     * @param context
     * @param zapiConnection
     * @return Samples of raw counters
     * @throws DataSourceErrorException
     */    
    public Map<K, Collection<RawSample<S>>> getAndReportPerformance(final NetAppDefaultStorageContext context,
                                                                    final ZAPIConnection zapiConnection)
            throws DataSourceErrorException {
        final Map<K, Collection<RawSample<S>>> allMeasurementsToReport = new HashMap<>();
        final Map<K, Collection<RawSample<R>>> rawMeasurementsToReport = new HashMap<>();
        
        Map<String, MetricsRowScratchPad<Z>> metricsRowScratchPads = 
        		generateMetricsRowScratchPads(context, zapiConnection, reportingPackage.getGlobalScratchPad());
        
        boolean rawPersistenceEnabled = context.isRawPersistenceEnabled();
        
        boolean nullInventoryUUIDReported = false;
        for (Map.Entry<String, MetricsRowScratchPad<Z>> metricsRowScratchPadEntry : metricsRowScratchPads.entrySet()) {
            final MetricsRowScratchPad<Z> metricsRowScratchPad = metricsRowScratchPadEntry.getValue();
            final String inventoryUUID = getInventoryUUID(context, metricsRowScratchPadEntry.getKey(), metricsRowScratchPad);
            if (!metricsRowScratchPad.isPreviousPollValuesAvailable()) {
                continue;// first poll, nothing to report
            }
            if (inventoryUUID == null) {
                nullInventoryUUIDReported = true;
                continue;
            }
            final K parentInventoryKey = getParentInventoryUUIDKey(context, inventoryUUID, metricsRowScratchPad);
            if (parentInventoryKey == null) {
                logger.info("Unable to determine parent inventory UUID for {}", inventoryUUID);
                continue;
            }
            final S measurement = normalizeMetrics(metricsRowScratchPad, context);
            RawSample<S> cookedSample = new RawSample<>(inventoryUUID, measurement);
            if (metricsRowScratchPad.isZeroSuppressDetected()) {
                // Make no report at all if zero suppression has been detected.
                cookedSample = null;
            } else if (metricsRowScratchPad.isWrapResetDetected()) {
                // detected Wrap/Reset, contract with server requires reporting the UUID with a null payload.
                cookedSample = new RawSample<>(inventoryUUID, getWrapResetSample(metricsRowScratchPad, context));
            }

            Collection<RawSample<S>> measurementsToReportForParentKey = allMeasurementsToReport.get(parentInventoryKey);
            if (measurementsToReportForParentKey == null) {
                measurementsToReportForParentKey = new LinkedList<>();
                allMeasurementsToReport.put(parentInventoryKey, measurementsToReportForParentKey);
            }
            if (cookedSample != null) {
                measurementsToReportForParentKey.add(cookedSample);
            }
            
            if(rawPersistenceEnabled){
                if(!rawMeasurementsToReport.containsKey(parentInventoryKey)) {
                	rawMeasurementsToReport.put(parentInventoryKey, new LinkedList<RawSample<R>>());
                }

                final Collection<RawSample<R>> countersForReport = rawMeasurementsToReport.get(parentInventoryKey);
                countersForReport.add(getRawSample(metricsRowScratchPad, inventoryUUID, context));
            }
        }
        
        if(nullInventoryUUIDReported) {
            logger.warn("Null Inventory UUID's were reported from {}, ignoring scratchpad with null inventory UUID", this.getClass().getSimpleName());
        }

        if(rawPersistenceEnabled){
            reportRawPerformance(context.getKey(), rawMeasurementsToReport);
        }

        reportPerformance(context.getKey(), allMeasurementsToReport);
        return allMeasurementsToReport;
    }

    /**
     * Find the inventory UUID for persistence. Default value is the inventoryUUID saved scratchpad. 
     * Derived builders need to override this function if the inventoryUUID is different from the one
     * saved in scratchpad, e.g. inventoryUUID for lif_vserver is vserverUUID.
     * 
     * @param context               the foundation context used to determine the parent foundation UUID key
     * @param counterId             counter ID used to find the corresponding inventoryUUID
     * @param metricsRowScratchPad  scratchpad containing the counter data
     * @return
     */
    protected String getInventoryUUID(NetAppDefaultStorageContext context, String counterId, MetricsRowScratchPad<Z> metricsRowScratchPad) {
    	return metricsRowScratchPad.getInventoryUUID();
    }
    
    /**
     * Get the parent foundation UUID key to use to report the samples under
     *
     * @param context              the foundation context used to determine the parent foundation UUID key
     * @param inventoryUUID        the inventory UUID to find the parent for
     * @param metricsRowScratchPad scratchpad containing the counter data
     * @return the parent inventory/foundation UUID key to use to report the samples under
     */
    protected abstract K getParentInventoryUUIDKey(NetAppDefaultStorageContext context, String inventoryUUID, MetricsRowScratchPad<Z> metricsRowScratchPad);

    /**
     * New scratch pads will be rated using the scratch pads from other builders.
     *
     * @param context          Context containing data from the foundation poll
     * @param zapiConnection   A connection to the cluster to execute ZAPI's
     * @param globalScratchPad A wrapper class containing the scratch pads from other builders.
     * @return A Map of performance uuid to  scratch pads
     */
    protected abstract Map<String, MetricsRowScratchPad<Z>> generateMetricsRowScratchPads(final NetAppDefaultStorageContext context,
                                                                                          final ZAPIConnection zapiConnection,
                                                                                          final GlobalScratchPad globalScratchPad);

    /**
     * Normalized this metrics instance against the previous metrics instance, and wrap in the
     * measurements class expected by the server
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return normalized metric values wrapped in the class expected by the server
     */
    protected abstract S normalizeMetrics(MetricsRowScratchPad<Z> currentScratchPad, NetAppDefaultStorageContext context);

    /**
     * Returns the Measurement object to report if a wrap reset is detected.
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return normalized metric values wrapped in the class expected by the server
     */
    protected S getWrapResetSample(final MetricsRowScratchPad<Z> currentScratchPad, final NetAppDefaultStorageContext context) {
        return null;
    }

    /**
     * Report the given samples to the server
     *
     * @param measurements the samples to report
     * @throws DataSourceErrorException when there is a problem reporting samples to the server
     */
    public abstract void reportPerformance(Cluster.ClusterKey clusterKey, Map<K, Collection<RawSample<S>>> measurements) throws DataSourceErrorException;

    /**
     * Get the raw counters out of the current metric.
     * 
     * @param currentScratchPad Current Scratch pad
     * @param inventoryUUID the inventory id of the sample
     * @param context Context with data from foundation poll
     * @return Raw Sample
     */
    final RawSample<R> getRawSample(final MetricsRowScratchPad<Z> currentScratchPad, final String inventoryUUID,
            final NetAppDefaultStorageContext context) {
        final R measurements = rawMetrics(currentScratchPad, context);
        return new RawSample<>(inventoryUUID, measurements);
    }

    /**
     * Raw metrics
     *
     * @param currentScratchPad Scratch Pad containing the current and the previous values.
     * @param context           in case other data from the context is required to normalize data
     * @return raw metric values wrapped in the class expected by the server
     */
    protected abstract R rawMetrics(MetricsRowScratchPad<Z> currentScratchPad, NetAppDefaultStorageContext context);

    /**
     * Report raw samples to the server
     * 
     * @param clusterKey the cluster key of the sample
     * @param rawMeasurements the samples to report
     * @throws DataSourceErrorException
     */
    public abstract void reportRawPerformance(Cluster.ClusterKey clusterKey, final Map<K, Collection<RawSample<R>>> rawMeasurements) throws DataSourceErrorException;   
    
}
