package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.builders;

import com.netapp.oci.netapp.performance.vserver.interfaces.data.VserverMeasurements;
import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.NetAppOCIEPerformancePackage;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.GlobalScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowSample;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.metrics.MetricsRowScratchPad;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.mergers.VserverMetricsMerger;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.synthetic.mergers.VserverMetricsMergerSN;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIConnection;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster.ClusterKey;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppDefaultStorageContext;

import java.util.Collection;
import java.util.Map;

public class VserverMetricsBuilder extends SyntheticBuilderBase<VserverMeasurements, VserverMeasurements, Cluster.ClusterKey, VserverMetricsBuilder.VserverCounter> {

    public static enum VserverCounter {
        opm_ops,
        opm_read_ops,
        opm_write_ops,
        opm_throughput,
        opm_read_throughput,
        opm_write_throughput,
        opm_read_latency,
        opm_write_latency,
        opm_latency,
    }

    public VserverMetricsBuilder(NetAppOCIEPerformancePackage reportingPackage) {
        super(reportingPackage);
    }

    @Override
    protected Cluster.ClusterKey getParentInventoryUUIDKey(NetAppDefaultStorageContext context,
                                                           String inventoryUUID,
                                                           MetricsRowScratchPad<VserverCounter> metricsRowScratchPad) {
        return context.getKey();
    }

    @Override
    protected Map<String, MetricsRowScratchPad<VserverCounter>> generateMetricsRowScratchPads(
            NetAppDefaultStorageContext context, ZAPIConnection zapiConnection, GlobalScratchPad globalScratchPad) {
        if (context.isLessThanFullSteam()) {
            // In the below case for SN all the counterObjects have performanceUUID as vServerUUID
            return globalScratchPad.merge(new VserverMetricsMergerSN(), "nfsv3", "nfsv4", "nfsv4_1", "cifs:vserver", "iscsi_lif:vserver", "fcp_lif:vserver");
        } else {
            return globalScratchPad.merge(new VserverMetricsMerger(), "nfsv3", "nfsv4", "nfsv4_1", "cifs", "iscsi_lif", "fcp_lif");
        }
    }


    @Override
    protected VserverMeasurements normalizeMetrics(MetricsRowScratchPad<VserverCounter> currentScratchPad,
                                               NetAppDefaultStorageContext context) {
        VserverMeasurements measurements = new VserverMeasurements();
        measurements.setOpmThroughput(getNormalizedFloatMetric(VserverCounter.opm_throughput, currentScratchPad));
        measurements.setOpmReadThroughput(getNormalizedFloatMetric(VserverCounter.opm_read_throughput, currentScratchPad));
        measurements.setOpmWriteThroughput(getNormalizedFloatMetric(VserverCounter.opm_write_throughput, currentScratchPad));
        measurements.setOpmOps(getNormalizedFloatMetric(VserverCounter.opm_ops, currentScratchPad));
        measurements.setOpmReadOps(getNormalizedFloatMetric(VserverCounter.opm_read_ops, currentScratchPad));
        measurements.setOpmWriteOps(getNormalizedFloatMetric(VserverCounter.opm_write_ops, currentScratchPad));
        measurements.setOpmReadLatency(getNormalizedFloatMetric(VserverCounter.opm_read_latency, currentScratchPad));
        measurements.setOpmWriteLatency(getNormalizedFloatMetric(VserverCounter.opm_write_latency, currentScratchPad));
        measurements.setOpmLatency(getNormalizedFloatMetric(VserverCounter.opm_latency, currentScratchPad));
        return measurements;
    }

    private float getNormalizedFloatMetric(VserverCounter counterName, MetricsRowScratchPad<VserverCounter> currentValues) {
        final float normalizedMetric;

        switch (counterName) {
            case opm_ops:
            case opm_read_ops:
            case opm_write_ops:
            case opm_read_throughput:
            case opm_throughput:
            case opm_write_throughput:
                normalizedMetric = currentValues.getCounterRatePerSec(counterName);
                break;
            case opm_read_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName, VserverCounter.opm_read_ops);
                break;
            case opm_write_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName, VserverCounter.opm_write_ops);
                break;
            case opm_latency:
                normalizedMetric = currentValues.getLatencyPerOp(counterName, VserverCounter.opm_ops);
                break;
            default:
                normalizedMetric = MetricsRowSample.NULL_FLOAT;
                break;
        }
        return normalizedMetric;

    }

    @Override
    public void reportPerformance(Cluster.ClusterKey clusterKey,
                                  Map<Cluster.ClusterKey, Collection<RawSample<VserverMeasurements>>> measurements)
            throws DataSourceErrorException {
        for (final Map.Entry<Cluster.ClusterKey, Collection<RawSample<VserverMeasurements>>> measurementEntry : measurements.entrySet()) {
            final Cluster.ClusterKey key = measurementEntry.getKey();
            final Collection<RawSample<VserverMeasurements>> vserverMeasurements = measurementEntry.getValue();
            reportingPackage.reportVserverPerformance(key, vserverMeasurements);
        }
    }

    @Override
    protected VserverMeasurements rawMetrics(MetricsRowScratchPad<VserverCounter> currentScratchPad, NetAppDefaultStorageContext context) {
        return null; //DO nothing, as raw stats are not persisted
    }

    @Override
    public void reportRawPerformance(ClusterKey clusterKey, Map<ClusterKey, Collection<RawSample<VserverMeasurements>>> rawMeasurements) throws DataSourceErrorException {
        // Do nothing as raw stats are not persisted
    }
}