package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transform.ocie;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.UUIDKey;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeAggregate;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeCluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Aggregate;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Cluster;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.VServer;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.Volume;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Common utility methods that can be used in different transform classes.
 */
public class NetAppOCIETransformCommon {
    /**
     * Retrieves and returns a list containing node name and the partner node name associated with the volume.
     *
     * @param cluster cluster with the vServer and the volumes populated
     * @param cModeCluster The complete cMode cluster
     * @param vServerKey vServer associated with the volume
     * @param volumeKey Volume Key
     * @param partnerNodeNames A Map of node name to partner node name.
     * @return A list containing node name and the partner node name associated with the volume or an empty list if the
     * node details are not available.
     */
    public static List<String> getNodeNamesForVolume(final Cluster cluster, final CModeCluster cModeCluster, final VServer.VServerKey vServerKey,
                                               final Volume.VolumeKey volumeKey, final Map<String, String> partnerNodeNames) {
        final CModeAggregate cModeAggregate = getAggregateForVolume(cluster, cModeCluster, vServerKey, volumeKey);
        return getNodeNamesForAggregate(cModeAggregate, partnerNodeNames);
    }

    /**
     * Retrieves and returns a list containing node name and the partner node name associated with the aggregate.
     *
     * @param cModeAggregate aggregate
     * @param partnerNodeNames A Map of node name to partner node name.
     * @return
     */
    public static List<String> getNodeNamesForAggregate(final CModeAggregate cModeAggregate, final Map<String, String> partnerNodeNames) {
        if (cModeAggregate == null || cModeAggregate.getNodeNameStrings().isEmpty()) {
            return new ArrayList<String>(0);
        } else {
            final String nodeName = cModeAggregate.getNodeNameStrings().get(0);
            final List<String> nodeNames = new ArrayList<String>(2);
            nodeNames.add(nodeName);
            if (partnerNodeNames.containsKey(nodeName)) {
                nodeNames.add(partnerNodeNames.get(nodeName));
            }
            return nodeNames;
        }
    }

    /**
     * Retrieve the aggregate (CModeAggregate) associated with the volume.
     *
     * @param cluster cluster with the vServer and the volumes populated
     * @param cModeCluster The complete cMode cluster
     * @param vServerKey  vServer associated with the volume
     * @param volumeKey  Volume Key
     * @return   aggregate (CModeAggregate) associated with the volume or null if not available.
     */
    public static CModeAggregate getAggregateForVolume(final Cluster cluster, final CModeCluster cModeCluster, final VServer.VServerKey vServerKey,
                                                 final Volume.VolumeKey volumeKey) {
        final VServer vServer = cluster.getVServers().get(vServerKey);
        if (vServer == null) {
            return null;
        }
        final Volume volume = vServer.getVolumes().get(volumeKey);
        if (volume == null) {
            return null;
        }
        final Aggregate.AggregateKey aggregateKey = volume.getAggregateKey();
        final String aggregateUUID = aggregateKey.getOcieUUID();
        final UUIDKey cModeAggregateKey = new UUIDKey("aggregate", aggregateUUID);
        return cModeCluster.getAggregates().get(cModeAggregateKey);
    }

}
