package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi;

import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppCommonConfig;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.config.BasePollingPackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSInputStream;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSample;
import com.onaro.sanscreen.acquisition.util.cli.TextSimulator;
import netapp.manage.NaServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Closeable;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPINames.*;

/**
 * Represents a connection to a ZAPI server.
 *
 * This class is thread-safe, i.e. all public methods may be called on different threads concurrently.
 */
public class PackageZAPIConnection extends ZAPIConnection
{
    private static final Logger logger = LoggerFactory.getLogger( PackageZAPIConnection.class );

    private final NetAppCommonConfig config;
    private final TransportType transportType;
    private final String keyPrefix;
    private final Properties properties;

    /**
     * Records all invoked ZAPIs including requested attributes
     */
    private final DSFileWriter zapiAttributesWriter;

    /**
     * Records ZAPI iterator response times
     */
    private final DSFileWriter responseTimesWriter;

    private volatile boolean isSimulated = false;
    private volatile boolean connected = false;

    public PackageZAPIConnection() throws DataSourceErrorException
    {
      this( null, null );
    }

    public PackageZAPIConnection(Package thePackage) throws DataSourceErrorException
    {
      this(thePackage, null );
    }

    public PackageZAPIConnection(Package thePackage, Properties properties) throws DataSourceErrorException {
        this.properties = properties;
        this.config = (NetAppCommonConfig) thePackage.getConfiguration();
        this.dsName = thePackage.getConfiguration().getDataSourceName();
        this.keyPrefix = thePackage.getPackageName();
        this.transportType = TransportType.valueOf( config.getConnectionType() );
        if (logger.isDebugEnabled()) {
            logger.debug(PackageContext.getLogPrefix() + " - initialize NetAppApi with ip=" + config.getIP() + ", user=" + config.getUser());
        }

        // Setup recording
        this.recording = new TextSimulator();
        final RecordingSample recordingSample = thePackage.getRecordingSession().getCurrentSample();
        try {
            this.recording.setRecordingFile(recordingSample.createRecordingTextWithDescription(thePackage.getConfiguration(), keyPrefix, config.getIP()));
        }
        catch (IOException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Failed to setup recording", null, e);
        }

        DSFileWriter responseTimesWriter = null;
        try {
            responseTimesWriter = thePackage.getRecordingSession().getCurrentSample().createTextualOutputFile(thePackage.getConfiguration(), DataSourceOutputFileType.log, "iterator_response_times", "txt");
            responseTimesWriter.writeln("datasource, zapi, timestamp, max-records, iterations, total-clock-time, total-zapi-time, total-invocations, total-records, returned-records-list");
        }
        catch (IOException ioe) {
            logger.error(thePackage.getLogPrefix() + " - Failed to record ZAPI response times: " + ioe, ioe);
        }
        this.responseTimesWriter = responseTimesWriter;

        DSFileWriter zapiAttributesWriter = null;
        try {
            zapiAttributesWriter = thePackage.getRecordingSession().getCurrentSample().createTextualOutputFile(thePackage.getConfiguration(), DataSourceOutputFileType.log, "zapi_attributes", "txt");
        }
        catch (IOException ioe) {
            logger.error(thePackage.getLogPrefix() + " - Failed to record ZAPI attributes: " + ioe, ioe);
        }
        this.zapiAttributesWriter = zapiAttributesWriter;

        // Flag indicating if we had to back-off on the requested max-records due to timeout
        final String backoffPropertyVal = ((BasePollingPackageConfiguration)config).getProperty("iterator.backoff.max.records", "50");
        try {
            configuredBackoffMaxRecords = Integer.parseInt(backoffPropertyVal);
        } catch (NumberFormatException nfe) {
            configuredBackoffMaxRecords = 50;
        }
        setupSimulation();
        connect();
    }

  @Override
  protected boolean isSimulating()
  {
    final Package thePackage = PackageContext.getPackage();
    return thePackage.getConfiguration().isSimulating();
  }

  protected boolean isDataSourceStopped() {
        final Package thePackage = PackageContext.getPackage();
        return thePackage.getDataSource().isStopped();
    }

  @Override
  protected int getMaxRetries()
  {
    return config.getMaxRetries();
  }

  public final void close() throws DataSourceErrorException {
        final NaServer server = getServer();
        if (server != null) {
            server.close();
            setServer( null );
        }
        closeQuietly(responseTimesWriter);
        closeQuietly(zapiAttributesWriter);
        // Close recording last since it might throw an exception
        recording.close();
    }

    private static void closeQuietly(final Closeable c) {
        if (c != null) {
            try {
                c.close();
            }
            catch (IOException ioe) {
                // ignore
            }
        }
    }

    /**
     * If we are running simulated, setup appropriate details. This is a no-op if the datasource package is
     * not being run simulated.
     * <p/>
     * This method is only called once from the constructor, so it doesn't need to be thread-safe.
     *
     * @throws com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException for framework type issues.
     */
    private void setupSimulation() throws DataSourceErrorException {
        final Package thePackage = PackageContext.getPackage();
        if (thePackage.getConfiguration().isSimulating()) {
            this.isSimulated = true;
            final TextSimulator.RequestNormalizer requestNormalizer = new TextSimulator.RequestNormalizer() {

                // The idea here is that desired attributes and query elements in XML requests
                // frequently change, and invalidate old recordings.
                // Instead of using the complete XML request document as the key
                // for the request, filter out the desired-attributes and query elements from the raw request
                // to produce a normalized request
                public String normalizeRequest(final String rawRequest) {
                    String normalizedRequest = rawRequest;

                    // Filter <desired-attributes>...</desired-attributes>
                    // from the request key as this element is changed frequently
                    // and causes simulation failures:
                    //      Failed to simulate request (Not found request: ...
                    // unless we ignore it here
                    final int desiredAttrsIndex = normalizedRequest.indexOf(DESIRED_ATTRIBUTES_START_TAG);
                    if (desiredAttrsIndex >= 0) {
                        final int desiredAttrsEndIndex = normalizedRequest.indexOf(DESIRED_ATTRIBUTES_END_TAG) + DESIRED_ATTRIBUTES_END_TAG.length();
                        final String requestStart = normalizedRequest.substring(0, desiredAttrsIndex);   // Everything before <desired-attributes>
                        final String requestEnd = normalizedRequest.substring(desiredAttrsEndIndex);     // Everything after </desired-attributes>
                        normalizedRequest = requestStart + requestEnd;
                    }

                    // Filter <query>...</query>
                    // from the request key as this element is changed frequently
                    // and causes simulation failures:
                    //      Failed to simulate request (Not found request: ...
                    // unless we ignore it here
                    final int queryIndex = normalizedRequest.indexOf(QUERY_START_TAG);
                    if (queryIndex >= 0) {
                        final int queryEndIndex = normalizedRequest.indexOf(QUERY_END_TAG) + QUERY_END_TAG.length();
                        final String requestStart = normalizedRequest.substring(0, queryIndex);
                        final String requestEnd = normalizedRequest.substring(queryEndIndex);
                        normalizedRequest = requestStart + requestEnd;
                    }

                    // Filter <max-records>...</max-records>
                    // from the request key as this element may be changed dynamically during iteration
                    // and causes simulation failures:
                    //      Failed to simulate request (Not found request: ...
                    // unless we ignore it here
                    final int maxRecordsIndex = normalizedRequest.indexOf(MAX_RECORDS_START_TAG);
                    if (maxRecordsIndex >= 0) {
                        final int maxRecordsEndIndex = normalizedRequest.indexOf(MAX_RECORDS_END_TAG) + MAX_RECORDS_END_TAG.length();
                        final String requestStart = normalizedRequest.substring(0, maxRecordsIndex);
                        final String requestEnd = normalizedRequest.substring(maxRecordsEndIndex);
                        normalizedRequest = requestStart + requestEnd;
                    }
                    return normalizedRequest;
                }
            };
            this.simulator = new TextSimulator(requestNormalizer);
            try {
                final DSInputStream dsInputStream = thePackage.getConfiguration().getSimulationSession().getCurrentSample().getStreamByDescription(DataSourceOutputFileType.recording, keyPrefix, config.getIP());
                final List<DSInputStream> done = new LinkedList<DSInputStream>();
                done.add(dsInputStream);
                this.simulator.initializeSimulation(done);
            }
            catch (IOException e) {
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Failed to setup simulation", e.getMessage(), e);
            }
            catch (FrameworkException e) {
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Failed to setup simulation", e.getMessage(), e);
            }
        }
    }

    /**
     * Are we running simulated?
     *
     * @return true if running simulated.
     */
    public boolean isSimulated() {
        return isSimulated;
    }

    /**
     * Create a default NaServer instance.
     * <p/>
     * This method is only called once from the constructor, so it doesn't need to be thread-safe.
     *
     * @param apiIP      IP on which API is running.
     * @param ontapMajor major version of ontap to specify.
     * @param ontapMinor minor version of ontap to specify.
     * @param timeout    timeout in seconds
     * @return NaServer instance suitably constructed for this API instance.
     * @throws com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException for any connection issues.
     */
    private NaServer createDefaultNaServer(final String apiIP, final int ontapMajor, final int ontapMinor, final int timeout) throws DataSourceErrorException {
        final NaServer defaultServer;
        try {
            defaultServer = new NaServer(apiIP, ontapMajor, ontapMinor);
            defaultServer.setTimeout(timeout);

            //in case we are using https we should set the transport type ( default http ).
            if (transportType == TransportType.HTTPS) {
                defaultServer.setTransportType(transportType.getNaServerType());
                defaultServer.disableSSLv3();
            }
            //port 443 is the default for https
            defaultServer.setPort(config.getPort());
            defaultServer.setAdminUser(config.getUser(), config.getPassword());
            defaultServer.setServerType(NaServer.SERVER_TYPE_FILER);

            // Now we have successfully connected.   This means that if the user provided
            // a fully qualified host name rather than an IP, then that hostname has
            // been successfully translated to an IP.   We can now safely lookup that
            // IP in order to verify the role of the LIF we are communicating with.
            final InetAddress clusterInetAddress = InetAddress.getByName(apiIP);
            this.targetIPAddressString = clusterInetAddress.getHostAddress();
        }
        catch (UnknownHostException e) {
            throw new DataSourceErrorException(DataSourceErrorType.CONNECTING, "Failed to connect to " + config.getIP(), e.getMessage(), e);
        }
        return defaultServer;
    }

    /**
     * If we are running against a live device, create appropriate connection.
     * <p/>
     * This method is only called once from the constructor, so it doesn't need to be thread-safe.
     */
    private void connect() throws DataSourceErrorException {
        if (!isSimulated()) {
            setServer( createDefaultNaServer( config.getIP(), 1, 0, config.getTimeout() ) );
        }
        else {
            this.targetIPAddressString = config.getIP();
        }
        connected = true;
    }

    /**
     * Is this API connected?
     *
     * @return true if API is connected.
     */
    @SuppressWarnings("UnusedDeclaration")
    public boolean isConnected() {
        return connected;
    }

  public String getIP() {
    return config.getIP();
  }

    public DSFileWriter getResponseTimesWriter() {
        return responseTimesWriter;
    }

    @SuppressWarnings({"unused declaration"})
    public TextSimulator getRecording() {
        return recording;
    }

    public static enum TransportType {
        HTTP(NaServer.TRANSPORT_TYPE_HTTP),
        HTTPS(NaServer.TRANSPORT_TYPE_HTTPS);
        private final int naServerType;

        TransportType(final int naServerType) {
            this.naServerType = naServerType;
        }

        public int getNaServerType() {
            return naServerType;
        }
    }

    public void recordRequestedAttribute(final String attribute) {
        attributes.add(attribute);
    }

    /**
     * Record attributes to the underlying attributes writer.
     * <p/>
     * This method is thread-safe, even though it is typically just invoked once after all requests have executed.
     */
    public void recordAttributes() {
        try {
            synchronized (attributes) {
                for (final String attribute : attributes) {
                    zapiAttributesWriter.writeln(attribute);
                }
            }
        }
        catch (IOException ioe) {
            logger.error(PackageContext.getPackage().getLogPrefix() + " - Failed to record ZAPI attributes: " + ioe, ioe);
        }
    }
}
