package com.onaro.sanscreen.acquisition.diskmodelutils;

import com.onaro.commons.util.Zip;
import org.apache.commons.io.FileUtils;

import java.io.*;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

/**
 * Extracts disk models from phonehome via DWH & the Aurora PhoneHome filesystem
 */
public class DiskModelPhoneHomeExtractor {
    // database host and credentials
    private static final String USER = "root";
    private static final String PASSWORD = "ugauga";
    private static final String MYSQL_HOST = "ssphonehomedwh";

    //Constants for phone home file access
    private static final String OUTPUT_FILE = "phonehomediskmodels.txt";
    private static final String PHONEHOME_DIR = "\\\\aurora.nane.netapp.com\\SHARE\\SANscreenPhoneHome\\Sites";
    private static final String destDirPath = ".\\disk_models";

    private static final String JDBC_DRIVER = System.getProperty("JDBC_DRIVER", "com.mysql.jdbc.Driver");

    //A collection of disk models
    private final static Map<String, Set<DiskModelAttributes>> diskModels = new HashMap<String, Set<DiskModelAttributes>>();

    //The destination dir
    private final static File destDir = new File(destDirPath);

    /**
     * Executes this application. This application will parse the phonehome sites from the DWH source
     * as well as the aurora phonehome file system. It will place it within a file called "phonehomediskmodels.txt"
     * which will contain all the diskmodels & any collisions between all the extracted data. It may be useful to run
     * DiskModelConflictResolver to easly go through these things.
     *
     * In addition, if you provide this application with an argument it will be taken as "true disk models", or, disk
     * models that we beleive to be correct in their information. Any disk model that is found during this extraction
     * process which also resides in this "true disk model" file will be dropped since we already have it's information.
     * The result of this is ONLY disk models that do not exist in the given file.
     *
     * @param args Arguments
     *
     * @throws Exception Thrown when bad things happen
     */
    public static void main(final String[] args) throws Exception {
        //Grab from the phone home database
        System.out.println("Getting data from Phone Home DWH");
        getFromSSPhoneHomeDWH();

        //Grab from the phone home file system
        if (!destDir.exists()) {
            destDir.mkdirs();
            System.out.println("Getting data from Phone Home On Aurora");
            getFromNewDiskModelInPhoneHome();
        }
        else {
            System.out.println(destDir.getPath() + " Already exists. Files From Aurora already downloaded.");
        }

        System.out.println("Parsing fiels from Aurora");
        parseFiles();

        //The first argument is "known to be true" disk models
        //Any disk model found within this file is automatically removed
        //from the list of disk models we generated because we beleive that diskmodel to be true
        //and do not need to see it within our mine.
        if (args.length > 0) {
            final File f = new File(args[0]);
            if (f.exists()) {
                System.out.println("Parsing known disk models...");
                int filtered = 0;
                final List<DiskModelAttributes> trueDiskModels = DiskModelFileExtractor.parseDiskModelDescriptionStream(new FileInputStream(f));
                for (final DiskModelAttributes diskModel : trueDiskModels) {
                    if (diskModels.remove(diskModel.key.toUpperCase()) != null) {
                        filtered++;
                    }
                }

                System.out.println(filtered + " disk models filterd.");
            }
            else {
                throw new Exception("Could not find " + args[0]);
            }
        }

        //Output the data
        System.out.println("Generating output...");
        generateOutput();

        System.out.println("Done.");
    }

    /**
     * Given a directory, sort the subdirectories/files in descending order for last modified time
     *
     * @param dir The parent directory
     *
     * @return A list of files/folders sorted by modified time in descending order
     */
    public static File[] lastFileModified(final File dir) {
        final File[] files = dir.listFiles();
        Arrays.sort(files, new Comparator<File>() {
            public int compare(final File f1, final File f2) {
                return Long.valueOf(f2.lastModified()).compareTo(f1.lastModified());
            }
        });
        return files;
    }

    /**
     * Runs through the aurora phonehome directories, pulling the latest extendedlog files
     * from every site and extracting the new_disk_models.txt
     */
    private static void getFromNewDiskModelInPhoneHome() {
        final File phoneHomeDir = new File(PHONEHOME_DIR);

        for (final File site : phoneHomeDir.listFiles()) {
            if (!site.isDirectory()) {
                continue;
            }

            try {
                final File[] dirsByDate = lastFileModified(site);
                if (dirsByDate == null) {
                    continue;
                }

                //We'll typically only grab the first dir, however, if there is an error we
                //want to continue to the next one
                for (final File dateDir : dirsByDate) {
                    //File all the files that start with "extendedlog"
                    final File[] extendedLogFiles = dateDir.listFiles(new FilenameFilter() {
                        public boolean accept(final File parent, final String filename) {
                            return filename.toLowerCase().startsWith("extendedlog") && filename.toLowerCase().endsWith(".gz");
                        }
                    });

                    //There is no extended logs
                    if (extendedLogFiles == null) {
                        continue;
                    }

                    int i = 0;
                    boolean successful = false;
                    for (final File extendedLog : extendedLogFiles) {
                        final File tempExtendedLogFolder = new File("tempExtendedLogFolder");
                        tempExtendedLogFolder.mkdirs();
                        tempExtendedLogFolder.deleteOnExit();
                        i++;

                        try {

                            //Extract the disk model into the temp log folder
                            Zip.fromZip(extendedLog.getCanonicalPath(), tempExtendedLogFolder);

                            System.out.println("Processing " + site.getName() + "/" + dateDir.getName() + "/" + extendedLog.getName());

                            final File diskModelFile = new File(tempExtendedLogFolder, "new_disk_models.txt");
                            if (diskModelFile.exists()) {
                                //Save the file away for later
                                final File dest = new File(destDir, site.getName() + "_" + dateDir.getName() + "_diskmodel_" + i + ".txt");
                                FileUtils.copyFile(diskModelFile, dest);
                                successful = true;
                            }

                            if (getDiskModelsFromFoundation2(tempExtendedLogFolder)) {
                                 successful = true;
                            }
                        }
                        catch (Exception e) {
                            System.out.println("Cannot process " + site.getName() + "/" + dateDir.getName() + "/" + extendedLog.getName() + ": " + e.getMessage());
                        }
                        finally {
                            deleteRecursively(tempExtendedLogFolder);
                        }
                    }

                    //If we got atleast one disk_model file out of this date then consider this successful
                    //and do not attempt to go back further in time to get this information.
                    if (successful) {
                        break;
                    }
                }
            }
            catch (Exception e) {
                System.out.println(e.getMessage());
            }
        }
    }


    private static boolean getDiskModelsFromFoundation2(final File extendedLogs) {
        //File all the files that start with "extendedlog"
        final File[] foundationFiles = extendedLogs.listFiles(new FilenameFilter() {
            public boolean accept(final File parent, final String filename) {
                return filename.toLowerCase().startsWith("foundation") && filename.toLowerCase().endsWith(".zip");
            }
        });

        //There is no extended logs
        if (foundationFiles == null) {
            return false;
        }

        boolean successful = false;
        for (final File extendedLog : foundationFiles) {
            try {
                final ZipFile zipFile = new ZipFile(extendedLog);
                final Enumeration<? extends ZipEntry> entries = zipFile.entries();
                InputStream zipInput = null;

                while (entries.hasMoreElements()) {
                    final ZipEntry zipEntry = entries.nextElement();
                    if (!zipEntry.isDirectory()) {
                        final String fileName = zipEntry.getName();
                        if (fileName.endsWith("diskmodel-descriptions.txt")) {
                            zipInput = zipFile.getInputStream(zipEntry);
                            final byte[] buffer = new byte[1024];
                            final String name = zipEntry.getName();
                            final int index = name.lastIndexOf('\\') + 1;
                            final File target = new File(destDir, name.substring(index));
                            final OutputStream output = new FileOutputStream(target);
                            int i = 0;
                            while ((i = zipInput.read(buffer)) > 0) {
                                output.write(buffer, 0, i);
                            }

                            output.close();
                            zipInput.close();
                            zipFile.close();
                            return true;
                        }
                    }
                }
                zipFile.close();
            }
            catch (final IOException ioe) {
                System.err.println("Unhandled exception:");
                ioe.printStackTrace();
            }
        }

        return false;
    }

    /**
     * Parses the files from phonehome and loads them into our list of disk models
     */
    private static void parseFiles() {
        final File[] files = destDir.listFiles();
        for (final File f : files) {
            final String site = f.getName().split("_")[0];

            try {
                final List<DiskModelAttributes> diskModels = DiskModelFileExtractor.parseDiskModelDescriptionStream(new FileInputStream(f));
                for (final DiskModelAttributes dm : diskModels) {
                    dm.sites.add(site);
                    addDiskModelToList(dm);
                }
            }
            catch (IOException e) {
                System.err.println("Unable to open: " + f.getPath());
            }
        }
    }

    /**
     * Delets files recursively under a directory
     *
     * @param f The target directory
     */
    private static void deleteRecursively(final File f) {
        if (f.isDirectory()) {
            for (final File c : f.listFiles()) {
                deleteRecursively(c);
            }
        }
        f.delete();
    }

    /**
     * Generates the output of all of the disk models
     *
     * @throws IOException Throws an IOException when bad things happen
     */
    private static void generateOutput() throws IOException {
        BufferedWriter writer = null;
        try {
            writer = new BufferedWriter(new FileWriter(OUTPUT_FILE));
            writer.append("## COLLISIONS\n\n");
            for (final Set<DiskModelAttributes> d : diskModels.values()) {
                if (d.size() > 1) {
                    writer.append("## Model = ").append(d.iterator().next().key).append("\n");

                    for (final DiskModelAttributes dm : d) {
                        writer.append("## Sites = ").append(dm.sites.toString()).append("\n");
                        writer.append(dm.to60CSV());
                    }

                    writer.append("\n");
                }
            }

            writer.append("\n");
            writer.append("## OK DATA\n\n");
            for (final Set<DiskModelAttributes> d : diskModels.values()) {
                if (d.size() == 1) {
                    writer.append(d.iterator().next().to60CSV());
                }
            }
        }
        finally {
            if (writer != null) {
                try {
                    writer.close();
                }
                catch (IOException ignore) {
                }
            }
        }
    }

    /**
     * Adds a disk model attribute to the global list
     *
     * @param diskModel The disk model object to add
     */
    private static void addDiskModelToList(final DiskModelAttributes diskModel) {
        Set<DiskModelAttributes> currentList = diskModels.get(diskModel.key.toUpperCase());
        if (currentList == null) {
            currentList = new HashSet<DiskModelAttributes>();
            currentList.add(diskModel);
            diskModels.put(diskModel.key.toUpperCase(), currentList);
        }
        else {
            if (!currentList.contains(diskModel)) {
                currentList.add(diskModel);
            }
        }
    }

    /**
     * Gets a list of disk model attributes from the phone home DWH server
     *
     * @throws Exception Throws an exception when bad news occurs
     */
    private static void getFromSSPhoneHomeDWH() throws Exception {
        Connection connection = null;

        try {

	    Class.forName(JDBC_DRIVER).newInstance();


            final Properties props = new Properties();
            props.put("user", USER);
            props.put("password", PASSWORD);
            final String JDBC_URL = "jdbc:mysql://" + MYSQL_HOST + "/" + "recordings";
            connection = DriverManager.getConnection(JDBC_URL, props);

            final String sql = "SELECT DISTINCT " + "d.disk_model, d.disk_vendor, d.disk_type, d.disk_speed, d.disk_capacity_gb, GROUP_CONCAT(DISTINCT CONCAT_WS(':', s.customer_name, s.site_name)) AS sites " + "FROM " + "skit_report.disk_models d, skit_report.dimension_site_name s " + "WHERE " + "d.site_name_tk = s.id AND d.disk_model != \"\" " + "GROUP BY " + "d.disk_model, d.disk_vendor, d.disk_type, d.disk_speed, d.disk_capacity_gb";

            final ResultSet result = connection.prepareStatement(sql).executeQuery();
            while (result.next()) {
                try {
                    final String model = result.getString(1).trim().toUpperCase();
                    final String vendor = result.getString(2).trim();

                    final int speed = result.getInt(4);
                    final Long capacity = (long) result.getInt(5);
                    final String interfaceType = result.getString(3).trim();
                    final String[] sites = result.getString(6).split(",");

                    final Float seekTime = null;
                    final Float potentialTransferRate = null;
                    final Float interfaceSpeed = null;
                    final String extraInfo = "";

                    final DiskModelAttributes dma = new DiskModelAttributes(model, vendor, model, speed, capacity, interfaceType, seekTime, potentialTransferRate, interfaceSpeed, extraInfo);
                    for (final String site : sites) {
                        dma.sites.add(site);
                    }

                    addDiskModelToList(dma);
                }
                catch (NumberFormatException nfe) {
                    System.err.println("Unable to parse: " + nfe.getMessage() + " for " + result.toString());
                }
            }
        }
        finally {
            if (connection != null) {
                connection.close();
            }
        }
    }
}
