package com.onaro.sanscreen.acquisition.foundation.datasource;

import com.onaro.sanscreen.acquisition.foundation.datasource.config.FoundationConfiguration;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFile;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileOutputStream;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileWriter;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSample;
import com.onaro.sanscreen.acquisition.schema.Root;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.odata2.OData2Root;
import com.onaro.util.CrcCalculator;
import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.*;
import java.util.*;
import java.util.zip.CRC32;

/**
 * Base data source for Foundation based packages.
 * <p/>
 * <p/>The implementor of this class has to override 3 methods: {@link #doStart()}, {@link #doStop()}, {@link #doFoundation}.
 */
public abstract class BaseFoundationPackage<C extends FoundationConfiguration> extends BasePollingPackage<C> implements BaseFoundationPackageMBean, DataSourceSampleListener {
    public final static String PACKAGE = "foundation";

    private final static String crlf = System.getProperty("line.seperator");
    private final static Logger logger = Logger.getLogger(BaseFoundationPackage.class);

    private final static long phoneHomeRecordingFrequency = 1000 * 60 * 60 * 24 * Long.parseLong(ServiceManager.getInstance().getProperty("FOUNDATION_RECORDING_FREQUENCY_DAYS", "1"));

    private OriginatorProvider originatorProvider;

    private final List<FoundationUpdatesListener> updatesListeners = new LinkedList<FoundationUpdatesListener>();

    protected boolean checkIntegrity = true;

    /**
     * Device specific context passed when reporting odata
     */
    private Serializable context = null;

    /**
     * Keeps track of the odata reports made in this sample. This will later be used to identify changes and to prevent
     * the acquisition from sending redundant data to the server
     */
    private Map<String, ODataInfo> odataInfoByOriginatorKey = new HashMap<String, ODataInfo>(50);

    /**
     * Keeps track of the originators reported in the current sample.
     */
    private final Set<Long> currentlyReportedOriginatorIds = new TreeSet<Long>();

    /**
     * Keeps track of the last error reported by this data source to prevent from multiple error recordings to be sent
     * through phone home. When this value is <code>null</code>, no error is recorded.
     */
    private Long lastErrorCRC = null;

    /**
     * Indicates if changes detected during this DS poll. This hints the system to create a recording file in the acq folder
     * and potentially send it to phone home.
     */
    private boolean changesDetected = false;

    /**
     * Indicates that this poll was interesting enough to send a recording file to phone home (regardless of changes). This
     * flag is triggered when an error occurs, crawl issues, etc...
     */
    private boolean sendRecordingFilesToPhoneHomeAnyways = false;

    /**
     * Timestamp of last time recording was sent to PhoneHome
     */
    private long lastRecordingReport = 0;

    /**
     * Constructs a data source instance
     * <p/>
     * <p/>Register for the datasource's life cycle events. This class will forward all event to the appropriate classes
     * (e.g. the {@link RecordingSample} and the {@link FoundationListener})
     *
     * @param dataSource
     * @param configuration
     * @param dataSourceLifeCycleListener A listener to receive DS life cycle events (to write audits, etc...)
     * @param dataSourceSampleListener    A listener to receive sample events (to write audits, etc...)
     * @param foundationUpdatesListener   A listener to receieve foundation updates
     * @param originatorProvider          Originator ID provider
     */
    protected BaseFoundationPackage(final BaseDataSource dataSource,
                                    final C configuration,
                                    final DataSourceLifeCycleListener dataSourceLifeCycleListener,
                                    final DataSourceSampleListener dataSourceSampleListener,
                                    final FoundationUpdatesListener foundationUpdatesListener,
                                    final OriginatorProvider originatorProvider) {
        super(dataSource, configuration, dataSourceSampleListener);

        // Register the parent listener to receive data source life cycle events
        dataSource.addDataSourceLifeCycleListener(dataSourceLifeCycleListener);

        // Register the parent listener to receive data source sample events
        // Make sure this class registers BEFORE the parent data source life cycle listener to be able to process things before they are being packed
        // But this class will not actually send zip to server
        dataSource.addDataSourceSampleListener(getPackageName(), 0, this);

        // Set the originator provider for this package.
        setOriginatorProvider(originatorProvider);

        updatesListeners.add(foundationUpdatesListener);

        this.logPrefix = "[" + getPackageName() + ((configuration.isDrLicensed()) ? " and DR" : "") + "] " + getConfiguration().getDataSourceName();
    }

    /**
     * @return device specific context passed when reporting odata
     */
    public final Serializable getContext() {
        return context;
    }

    /**
     * Sets device specific context passed when reporting odata
     *
     * @param context
     */
    public final void setContext(final Serializable context) {
        this.context = context;
    }

    public Set<Long> getCurrentlyReportedOriginatorIds() {
        return Collections.unmodifiableSet(currentlyReportedOriginatorIds);
    }

    public final RecordingSample getRecordingSample() {
        return getRecordingSession().getCurrentSample();
    }

    public final DSFileOutputStream createBinaryRecordingFile() throws IOException {
        return getRecordingSample().createBinaryRecordingFile(getConfiguration());
    }

    public final DSFileWriter createTextualOutputFile(final DataSourceOutputFileType type, final String custom, final String extension) throws IOException {
        return getRecordingSample().createTextualOutputFile(getConfiguration(), type, custom, extension);
    }

    public final DSFileWriter createTextRecordingFile() throws IOException {
        return getRecordingSample().createTextRecordingFile(getConfiguration());
    }

    public final DSFileWriter createTextRecordingFile(final String custom) throws IOException {
        return getRecordingSample().createTextRecordingFile(getConfiguration(), custom);
    }

    public final void addRecordingFile(final String fileName) throws FileNotFoundException {
        getRecordingSample().addRecordingFile(fileName);
    }

    public final void addRecordingStream(final DSFile stream) {
        getRecordingSample().addRecordingStream(stream);
    }

    public String crawlNow() {
        if (!getConfiguration().isCrawlSupported())
            return "Crawl is not supported by this datasource";
        getConfiguration().turnOnCrawlingOnce();
        this.getDataSource().run(this);
        return "Success";
    }

    /**
     * Performs the poll logic
     */
    public final void pollNow() throws DataSourceErrorException, FrameworkException {
        try {
            try {
                doFoundation(dataSource);
            }
            catch (SchemaException e) {
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Failed to create schema objects", null, e);
            }
        }
        catch (DataSourceErrorException e) {
            // Set the originator ID for this exception (if was not already set)
            if (e.getDeviceKey() == null) {
                e.setDeviceKey("General Device");
            }

            throw e;
        }
    }

    public final void start() throws DataSourceErrorException, FrameworkException {
        context = null;

        doStart();
    }

    public final void stop() throws DataSourceErrorException, FrameworkException {
        doStop();

        context = null;
    }

    /**
     * Initialization code. This method is called everytime the data source is started or when the configuration of the
     * data source changes or when license is changed.
     * <p/>
     * Static information (like configuration parameters) should be read by the configuration class (extending
     * {@link FoundationConfiguration}). Other package state like persistent sessions, etc... can be initialized in
     * this method. However, packages should try to remain as stateless as possible.
     *
     * @throws FrameworkException
     * @throws DataSourceErrorException
     */
    protected abstract void doStart() throws FrameworkException, DataSourceErrorException;

    /**
     * This method is called everytime the data source is stoped, or when the configuration of the data source changes
     * or when license is changed.
     * <p/>
     * Package implementations should perform any neccessary cleanup here (like closing openned persistent sessions, etc...).
     * Packages should try to remain as stateless as possible, minimizing the amount of cleanup required here.
     *
     * @throws FrameworkException
     * @throws DataSourceErrorException
     */
    protected abstract void doStop() throws FrameworkException, DataSourceErrorException;

    /**
     * Query foundation data.
     * <p/>
     * <p/>This method is called by the framework when it is time for this package to poll the device for foundation data.
     * <p/>
     * <p/>The framework allocates a dedicated thread (for this poll) and prevents from other foundation polls to occur
     * while this method is still executing.
     * <p/>
     * <p/>If an error occured during the polling of the device, the implementation should throw a {@link DataSourceErrorException}
     * exception. The framework will also catch any other unhandled exception thrown by the implementation and report about
     * it as an <code>Internal Error</code>.
     * <p/>
     * <p/>If the poll ended successfully and there is nothing to report, the method should just return. If a new
     * schame object hierarchy need to be reported, the implementation should use the {@link #reportOData2} method to send
     * this new report to the server.
     *
     * @param dataSource Datasource initiating this poll
     * @throws SchemaException          when an error occured during the creation of schema objects
     * @throws FrameworkException       when an error occured with framework related mechanisms (e.g. communication problem).
     * @throws DataSourceErrorException when an error occurs during the sample (this error contains information needed by the framework to create an audit update)
     */
    protected abstract void doFoundation(final BaseDataSource dataSource) throws SchemaException, FrameworkException, DataSourceErrorException;

    /**
     * Allocates/returns an originator ID to match the given originaor key
     *
     * @param originatorKey
     * @throws FrameworkException
     */
    public final synchronized long getOriginatorID(final String originatorKey) throws FrameworkException {
        return getOriginatorID(originatorKey, originatorKey);
    }

    /**
     * Allocates/returns originator ID to match the given originaor key
     *
     * @param originatorKey         Key identifying originator ID
     * @param originatorDisplayName User friendly description to display in the originator audit screen
     * @throws FrameworkException
     */
    public final synchronized long getOriginatorID(final String originatorKey, final String originatorDisplayName) throws FrameworkException {
        // Sanity check: A MOID data source should never try to obtain an originator ID with null key. The
        // null key is reserved for non-MOID data sources only!
        if (originatorKey == null && configuration.isMOID())
            throw new FrameworkException("A MOID datasource (" + dataSource.getDataSourceName() + ") is trying to get originator ID with null key");

        return originatorProvider.getOriginatorID(dataSource.getDataSourceID(), originatorKey, originatorDisplayName != null ? originatorDisplayName : dataSource.getDataSourceName());
    }

    /**
     * Allocates/returns originator ID to match the given originaor key
     *
     * @param originatorKey         Key identifying originator ID
     * @param originatorDisplayName User friendly description to display in the originator audit screen
     * @throws FrameworkException
     */
    public final synchronized long getOriginatorID(final String originatorKey, final String originatorDisplayName, final boolean performDuplicateDatasourceCheck) throws FrameworkException {
        // Sanity check: A MOID data source should never try to obtain an originator ID with null key. The
        // null key is reserved for non-MOID data sources only!
        if (originatorKey == null)
            throw new FrameworkException("A MOID datasource (" + dataSource.getDataSourceName() + ") is trying to get originator ID with null key");

        return originatorProvider.getOriginatorID(dataSource.getDataSourceID(),
                                                  originatorKey,
                                                  originatorDisplayName != null ? originatorDisplayName : dataSource.getDataSourceName(),
                                                  performDuplicateDatasourceCheck);
    }

    public final String[] getOriginatorCache() {
        return originatorProvider.getOriginatorCache();
    }

    public final void cleanOriginatorCache() {
        originatorProvider.cleanOriginatorCache();
    }

    public final boolean isRetireUnreportedOriginators() {
        return getConfiguration().isRetireUnreportedOriginators();
    }

    public final void setRetireUnreportedOriginators(boolean retireUnreportedOriginators) {
        getConfiguration().setRetireUnreportedOriginators(retireUnreportedOriginators);
    }

    public final void setManageURL(final String manageURL) {
        for (final FoundationUpdatesListener listener : updatesListeners)
            listener.reportManageURL(dataSource, manageURL);
    }

    /**
     * Used by the data source to report back odata2 (unmarshaled). This method will marshal the XML and then update the
     * listener.
     * <p/>
     * <p/>This method throws no exception because after a data source finished querying a device, it should not worry
     * about errors occurring during the report to the server. This is a task for the framework.
     *
     * @param odataRoot XML to report or null if only files should be reported
     */
    public final void reportOData2(final OData2Root odataRoot) throws DataSourceErrorException, FrameworkException {
        reportOData2(odataRoot, null);
    }


    /**
     * Used by the data source to report back odata2 (unmarshaled). This method will marshal the XML and then update the
     * listener.
     * <p/>
     * <p/>This method throws no exception because after a data source finished querying a device, it should not worry
     * about errors occurring during the report to the server. This is a task for the framework.
     *
     * @param odata2Root XML to report or null if only files should be reported
     */
    public final void reportOData2(final OData2Root odata2Root, final String originatorDisplayName) throws DataSourceErrorException, FrameworkException {
        // If we're stopping right now, ignore this report!
        if (dataSource.isStopped())
            return;

        // Make any necessary final adjustments using odata finalizers.
        // Since finalizeOdata() often needs to create new schema objects, it throws
        // SchemaException, so we rethrow as a FrameworkException to keep the odata2Root
        // interface unchanged.
        try {
            for (final OdataFinalizer finalizer : getOdataFinalizers(odata2Root)) {
                finalizer.finalizeOdata();
            }
        }
        catch (SchemaException e) {
            throw new FrameworkException("Error finalizing the Root structure.", e);
        }

        try {     
            final String odataXML = marshalOData(odata2Root);

            // Check if this data changed from the last odata we reported for this originator
            ODataInfo odataInfo = odataInfoByOriginatorKey.get(odata2Root.getOriginatorKey());
            if (odataInfo == null) {
                odataInfo = new ODataInfo();
                odataInfoByOriginatorKey.put(odata2Root.getOriginatorKey(), odataInfo);
            }

            // add originator to the current list
            currentlyReportedOriginatorIds.add(odata2Root.getOriginator().getOriginatorID());

            // Write the XML to a file (this will also add it to the list of files generated in this sample)
            // Note:  We have to make file's name unique, otherwise we can run in situations that 2 files (odatas)
            // have the same name.  This situation could arise if we collect references of Root objects in order to
            // report them lately on.  For example:
            //
            //  for (final FabricRoot switchRoot : switchRoots)
            //     reportOData(switchRoot);
            //
            // Note: Do not use the "odataRoot.getInfo().getOriginatorKey()" because it can contain chars that are not
            // allowed at the time of creating the file.
            final String custom = Long.toString(odata2Root.getOriginator().getOriginatorID());
            final DSFileWriter os = getRecordingSample().createTextualOutputFile(getConfiguration(), DataSourceOutputFileType.odata, custom, "xml");
            os.write(odataXML);
            os.close();

            // Keep track of which file is generated where. This will later be used for simulating this environment.
            odataInfo.setLastODataFileName(new File(os.getFileName()).getName());

            // Perform integrity checks on the odata generated. We're doing that after files are saved.
            // If checks failed, we will fail the acquisition (reporting an error)
            if (checkIntegrity)
                odata2Root.checkIntegrity();

            // Check for changes. If this originator report contains no changes since the last report, don't report it to
            // the server.
            boolean changed = odataInfo.changed(odataXML);

            // If data contains changes, we want to keep this recording!
            changesDetected |= changed;

            // Report this new data to anyone that is interested
            for (final FoundationUpdatesListener listener : updatesListeners) {
                listener.reportOData2(dataSource,
                        odata2Root.getOriginatorKey(),
                        originatorDisplayName,
                        odata2Root,
                        odataXML,
                        changed,
                        context);
            }


            dataSource.addToStatusHistory("Reported odata for " + originatorDisplayName);
        }
        catch (IntegrityException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INCONSISTENT_DATA, e.getMessage(), "Odata Integrity checks failed (" + e.getMessage() + ")", e);
        }
        catch (SchemaException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, null, e.getMessage(), e);
        }
        catch (IOException e) {
            throw new FrameworkException(e.getMessage(), e);
        }

    }

    /**
     * Return a list of odata finalizers to be executed in sequence before
     * reporting the specified odata.
     *
     * @param odata2Root The odata root to be finalized.
     * @return The list of finalizers for the specified odata root.
     */
    protected List<OdataFinalizer> getOdataFinalizers(
            final OData2Root odata2Root) {
        // No finalizer by default. Let subclasses define finalizers.
        return new LinkedList<OdataFinalizer>();
    }

    /**
     * Save model data as xml in a file. The model is only persisted if isRecording() == true
     *
     * @param modelRoot the device model
     * @throws DataSourceErrorException if an IntegrityException or SchemaException is caught
     * @throws FrameworkException       if an io exception is caught
     */
    public final void reportModelData(final Root modelRoot) throws DataSourceErrorException, FrameworkException {
        // If we're stopping right now, ignore this report!
        if (dataSource.isStopped())
            return;

        try {

            if (isRecording()) {
                final DSFileWriter fw = getRecordingSample().createTextualOutputFile(getConfiguration(), DataSourceOutputFileType.model, "", "xml");
                modelRoot.marshall(fw);
                fw.close();
            }

            // Perform integrity checks on the model generated. We're doing that after files are saved.
            // If checks failed, we will fail the acquisition (reporting an error)
            if (checkIntegrity)
                modelRoot.checkIntegrity();
        }
        catch (IntegrityException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INCONSISTENT_DATA, e.getMessage(), "Device model integrity checks failed (" + e.getMessage() + ")", e);
        }
        catch (SchemaException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, null, e.getMessage(), e);
        }
        catch (IOException e) {
            throw new FrameworkException(e.getMessage(), e);
        }
    }


    /**
     * Updates the management URL for this data source. Data sources usually use this method when the data source is
     * configured to point to a management unit that manages more than one device.
     * <p/>
     * <p/>Note: Each device reported by the data source also has the <code>manaeURL</code> attribute that allows the acquisition
     * to specify specific access to this device.
     *
     * @param manageURL
     */
    public final void reportManageURL(final String manageURL) {
        // Report this new data to anyone that is interested
        for (final FoundationUpdatesListener listener : updatesListeners)
            listener.reportManageURL(dataSource, manageURL);
    }

    private void setOriginatorProvider(final OriginatorProvider foundationListener) {
        this.originatorProvider = foundationListener;
    }

    /**
     * Converts the oData root object to an XML form
     * <p/>
     * <p/>Note: This method also saves the marshalled XML to the output directory and adds the file
     * name to the reported files for this sample.
     *
     * @param odata Root of o-data to marshal
     * @return The XML string
     * @throws SchemaException
     * @throws IOException
     */
    private String marshalOData(final Root odata) throws SchemaException, IOException {
        // Mark the XML with the sample date
        odata.getInfo().setDate(Calendar.getInstance().getTime().toString());

        // Add OriginatorKey
        odata.getInfo().setOriginatorKey(odata.getOriginatorKey());

        // Now marshal the object heirarchy into an XML string
        final StringWriter writer = new StringWriter(256 * 1024);
        odata.marshall(writer);

        return writer.toString();
    }


    public final void startingSample(final BaseDataSource dataSource, final PollingPackage aPackage) {
        sendRecordingFilesToPhoneHomeAnyways = false;

        // Defaults to false, assumes no changes until odata is compared.   When set to true by
        // a data source, we will create a recording zip regardless of odata.
        changesDetected = ((FoundationConfiguration)aPackage.getConfiguration()).isForceCreateZipOfRecording();

        // clear the set of currently reported originators
        currentlyReportedOriginatorIds.clear();

        // Foundation re-runs the same sample over and over !
        if (isSimulating()) {
            try {
                final DSSessionFileReader simulationSession = new DSSessionFileReader(getConfiguration().getSimulationSession());
                getConfiguration().setSimulationSession(simulationSession);
            }
            catch (IOException e) {
                logger.error(getLogPrefix() + " - failed to reload simulation session", e);
            }
        }


        getRecordingSession().reset();
        getRecordingSession().createSample();
    }

    public final void sampleSuccessfull(final BaseDataSource baseDataSource, final PollingPackage aPackage) {
        // Zeroize the error state for this data source. This will make sure that the next time the data source fails, it
        // will keep the recording files for this data source.
        lastErrorCRC = null;
        getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.SUCCESS);
    }


    public final void sampleFailed(final BaseDataSource baseDataSource, final PollingPackage aPackage, final DataSourceErrorException e) {
        // This sample failed. We would like to keep the recording files for this one in order to make it possible
        // to simulate this issue in the lab

        // If we're considering not to keep the recording files for this sample, check out if the error reported here
        // was already reported by this data source last time. This will allow us to minimize the number of identical
        // reports we're keeping in the log file and to minimize the size of the phone home report.
        // Calculate CRC on the new Data

        final long newCRC = e.getExceptionCRC();

        if (lastErrorCRC == null || lastErrorCRC != newCRC)
            sendRecordingFilesToPhoneHomeAnyways = true;

        lastErrorCRC = newCRC;

        //setting the status of the sample
        if (e instanceof AggregatedDataSourceErrorsException) {
            if (e.getType() == DataSourceErrorType.PARTIAL_SUCCESS) {
                getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.PARTIAL_SUCCESS);
                getRecordingSample().setSamplePartialSuccessInformation(((AggregatedDataSourceErrorsException) e).getTotalNumberOfOriginators(), ((AggregatedDataSourceErrorsException) e).getExceptions().size());
            } else
                getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.FAILED);
        } else
            getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.FAILED);

    }

    public final void sampleFailed(final BaseDataSource baseDataSource, final PollingPackage aPackage, final Throwable e) {
        // This sample failed. We would like to keep the recording files for this one in order to make it possible
        // to simulate this issue in the lab

        // If we're considering not to keep the recording files for this sample, check out if the error reported here
        // was already reported by this data source last time. This will allow us to minimize the number of identical
        // reports we're keeping in the log file and to minimize the size of the phone home report.
        // Calculate CRC on the new Data
        final StringBuilder sb = new StringBuilder(1000);
        sb.append(baseDataSource.getDataSourceName()).append(" - ").append(e.getMessage()).append(crlf);

        DataSourceErrorException.printStackStraceToBuffer(sb, e);

        final long newCRC = CrcCalculator.calculateCRC(sb);

        if (lastErrorCRC == null || lastErrorCRC != newCRC)
            sendRecordingFilesToPhoneHomeAnyways = true;

        lastErrorCRC = newCRC;
    }

    public final void finishedSample(final BaseDataSource dataSource, final PollingPackage aPackage) throws FrameworkException {

        // Make sure crawling is done only once!
        if (configuration.isCrawling()) {
            // Crawl support: If the data source was activated in crawl mode, turn it off. Crawl
            // should be activated only once!
            configuration.turnOffCrawling();

            // When crawling, we always want to keep the generated recording file (this is what crawl is all about)
            sendRecordingFilesToPhoneHomeAnyways = true;
        }

        packOdataInfo();

        try {
            final long currentTime = System.currentTimeMillis();
            final boolean enoughTimePassedSinceLastReport = (phoneHomeRecordingFrequency == 0) ? true :
                    ((currentTime - lastRecordingReport) > phoneHomeRecordingFrequency);

            // Is there any reason to create a zip file ?
            if (enoughTimePassedSinceLastReport || changesDetected || sendRecordingFilesToPhoneHomeAnyways) {
                // Create recording zip file
                final File recordingSession = getRecordingSession().buildRecordingFile();

                final boolean shouldReport = enoughTimePassedSinceLastReport || sendRecordingFilesToPhoneHomeAnyways;
                getDataSource().recordingSessionDone(this, recordingSession, shouldReport);

                if (shouldReport) {
                    lastRecordingReport = currentTime;
                }

                // send recording to server if necessary
               /* if (enoughTimePassedSinceLastReport || sendRecordingFilesToPhoneHomeAnyways) {
                    getDataSource().reportRecordingSession(this, recordingSession);
                    lastRecordingReport = currentTime;
                } */
            }
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to send ZIP file: " + e.getMessage(), e);
        }
        finally {
            getRecordingSession().reset();
            if (getConfiguration().getSimulationSession() != null) {
                getConfiguration().getSimulationSession().close();
            }
        }
    }

    public void recordingSessionDone(final BaseDataSource dataSource,
                                       final PollingPackage aPackage,
                                       final File recordingSession,
                                       final boolean shouldReport) throws FrameworkException {
        // Do nothing ! Relevant FoundationListener that implements DataSourceSampleListener will do the work of actually sending the recording to server !
    }

    /**
     * Fill sample description with a odata file name to originator ID mapping
     */
    private void packOdataInfo() {
        for (final Map.Entry<String, ODataInfo> entry : odataInfoByOriginatorKey.entrySet()) {
            // A MOID data source cannot report the empty (null) OID!
            assert entry.getKey() != null && getConfiguration().isMOID();

            final String key = (entry.getKey() != null ? entry.getKey() : "");
            getRecordingSession().getCurrentSample().addFileDescription(entry.getValue().getLastODataFileName(), "system.originator.key", key);
        }
    }

    public final ODataInfo getODataInfo(final String originatorKey) {
        return odataInfoByOriginatorKey.get(originatorKey);
    }

    /**
     * Called by anyone that wishes to listen to foundation updates
     *
     * @param updatesListener Listener interface for reporting back updates
     */
    public final void addUpdatesListener(final FoundationUpdatesListener updatesListener) {
        updatesListeners.add(updatesListener);
    }

    /**
     * This class saved CRC state for a data source foundation sample. The CRC is calculated based on the generated
     * OData XML.
     */
    final class ODataInfo {

        private long lastFullCRC = 0;             // All Attributes
        private long lastStaticCRC = 0;          // Static Attributes only
        private long lastReportTimestamp = 0;

        private String lastODataFileName = null;

        public ODataInfo() {
        }

        public String getLastODataFileName() {
            return lastODataFileName;
        }

        public void setLastODataFileName(final String lastODataFileName) {
            this.lastODataFileName = lastODataFileName;
        }

        /**
         * Using CRC comparison to the last <b>good</b> report, detect if this XML contains any changes in the environment.
         * <p/>
         * <p/>Note that the CRC is being calculated only when a datasource has successfully finished reporting about a device and
         * called the {@link FoundationUpdatesListener#reportOData2} method. This ensures us that errors during the acquisition will
         * not cause us to save redundant reports again and again.
         * <p/>
         * <p/>Note: Since the <Info> element contains dynamic information that changes from sample to sample, we have
         * to remove it before calculating any CRC, otherwise the CRC will always indicate change.
         */
        public boolean changed(final String odataXML) {
            final int ignoreDynamicChangesForHours = configuration.getIgnoreDynamicChangesForHours();
            final long now = System.currentTimeMillis();
            final long maxIgnoreDynamicForMS = ignoreDynamicChangesForHours * 60 * 60 * 1000;    // convert hours to MS

            // Assume report must be sent
            boolean changed = true;

            // Truncate the dynamic part of the oData XML. It ends with the <Info> element
            final String staticOData = odataXML.replaceFirst("<Info .*/>", "");

            // Calculate CRCs on the new odata
            try {
                final SAXParserFactory saxFactory = SAXParserFactory.newInstance();
                final SAXParser saxParser = saxFactory.newSAXParser();
                final InputStream xmlInputStream = new ByteArrayInputStream(staticOData.getBytes("UTF-8"));
                final OdataChangeDetector odataChangeDetector = new OdataChangeDetector();
                saxParser.parse(xmlInputStream, odataChangeDetector);
                final long newFullCRC = odataChangeDetector.fullCRC.getValue();
                final long newStaticCRC = odataChangeDetector.staticCRC.getValue();
                if (logger.isDebugEnabled()) {
                    logger.debug(configuration.getDataSourceName() + " full CRC: " + newFullCRC
                        + ", static CRC: " + newStaticCRC + ", last full CRC: " + lastFullCRC
                        + ", last static CRC: " + lastStaticCRC);
                }

                if (newStaticCRC == lastStaticCRC) {

                    // No changes detected in static attributes.   Do we need to check dynamic attributes?
                    final long timeSinceLastReport = (now - lastReportTimestamp);
                    if (timeSinceLastReport > maxIgnoreDynamicForMS) {

                        // It's been too long, all attributes must be examined
                        if (newFullCRC == lastFullCRC) {

                            // Same full CRC means we can suppress this report (not likely to occur)
                            changed = false;
                            final String message = String.format(getLogPrefix() +
                                    " - Suppressing identical odata report.");
                            logger.info(message);
                        }
                    } else {

                        // Static CRC is the same, and we've reported a full ofdata recently.
                        // We can ignore any dynamic changes and suppress this report.
                        changed = false;
                        if (newFullCRC == lastFullCRC) {
                            final String message = String.format(getLogPrefix() +
                                    " - Suppressing identical odata report.");
                            logger.info(message);
                        } else {
                            final int timeLeftToWaitTotalSeconds = (int)(maxIgnoreDynamicForMS - timeSinceLastReport) / (1000);
                            final int timeLeftToWaitHours = timeLeftToWaitTotalSeconds / (60 * 60);
                            final int timeLeftToWaitMins = (timeLeftToWaitTotalSeconds / 60) % 60;
                            final int timeLeftToWaitSecs = timeLeftToWaitTotalSeconds % 60;
                            final String message = String.format(getLogPrefix() +
                                    " - Suppressing dynamic odata report, will report in approximately %d:%02d:%02d",
                                    timeLeftToWaitHours, timeLeftToWaitMins, timeLeftToWaitSecs);
                            logger.info(message);
                        }
                    }
                }

                if (changed) {
                    lastReportTimestamp = now;
                    lastFullCRC = newFullCRC;
                    lastStaticCRC = newStaticCRC;
                }
            } catch (Exception saxe) {
                // If XML parsing fails, we default to sending the report.   Only log when debugging is enabled
                if (logger.isDebugEnabled()) {
                    logger.error("Failed to parse generated odata during static CRC calculation: " + saxe.getLocalizedMessage(), saxe);
                }
            }
            return changed;
        }
    }

    private static final class OdataChangeDetector extends DefaultHandler {

        private final CRC32 staticCRC = new CRC32();
        private final CRC32 fullCRC = new CRC32();


        // Map of Element QNames to Set of dynamic Attribute names for that Element
        private static final Map<String, Set<String>> elementNameToDynamicAttributes = new HashMap<String, Set<String>>();
        static {

            // StoragePool
            final Set<String> storagePoolDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("StoragePool", storagePoolDynamicAttributes);
            storagePoolDynamicAttributes.add("DedupeSavings");
            storagePoolDynamicAttributes.add("SnapshotUsedCapacityMB");
            storagePoolDynamicAttributes.add("DataUsedCapacityMB");
            storagePoolDynamicAttributes.add("OtherUsedCapacityMB");
            storagePoolDynamicAttributes.add("TotalUsedCapacityMB");

            // InternalVolume
            final Set<String> internalVolumeDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("InternalVolume", internalVolumeDynamicAttributes);
            internalVolumeDynamicAttributes.add("DedupeSavings");
            internalVolumeDynamicAttributes.add("LastSnapshotTime");
            internalVolumeDynamicAttributes.add("SnapshotCount");
            internalVolumeDynamicAttributes.add("LastKnownAccessTime");  
            internalVolumeDynamicAttributes.add("SnapshotUsedCapacityMB");
            internalVolumeDynamicAttributes.add("DataUsedCapacityMB");
            internalVolumeDynamicAttributes.add("DataAllocatedCapacityMB");
            internalVolumeDynamicAttributes.add("OtherUsedCapacityMB");
            internalVolumeDynamicAttributes.add("TotalUsedCapacityMB");
            internalVolumeDynamicAttributes.add("TotalUsedCapacityMB_2");
            internalVolumeDynamicAttributes.add("TotalCloneSavedCapacityMB");

            // Qtree
            final Set<String> qtreeDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("Qtree", qtreeDynamicAttributes);
            qtreeDynamicAttributes.add("QuotaUsedCapacityMB");

            // Volume
            final Set<String> volumeDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("Volume", volumeDynamicAttributes);
            volumeDynamicAttributes.add("DedupeSavings");
            volumeDynamicAttributes.add("LastKnownAccessTime");
            volumeDynamicAttributes.add("UsedCapacityMB");

            // Quota
            final Set<String> quotaDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("Quota", quotaDynamicAttributes);
            quotaDynamicAttributes.add("UsedFiles");
            quotaDynamicAttributes.add("UsedCapacityMB");

            // HV_DataStore
            final Set<String> hvDataStoreDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("HV_DataStore", hvDataStoreDynamicAttributes);
            hvDataStoreDynamicAttributes.add("UsedCapacityMB");

            // HV_VirtualDisk
            final Set<String> hvVirtualDiskDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("HV_VirtualDisk", hvVirtualDiskDynamicAttributes);
            hvVirtualDiskDynamicAttributes.add("UsedCapacityMB");

            //HV_VirtualMachine
            final Set<String> hvVirtualMachineDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("HV_VirtualMachine", hvVirtualMachineDynamicAttributes);
            hvVirtualDiskDynamicAttributes.add("UsedCapacityMB");


            // IscsiSession
            final Set<String> iscsiSessionDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("IscsiSession", iscsiSessionDynamicAttributes);
            iscsiSessionDynamicAttributes.add("InitiatorIps");
            iscsiSessionDynamicAttributes.add("TargetSessionId");
            iscsiSessionDynamicAttributes.add("InitiatorSessionId");

            // StorageEfficiency
            final Set<String> storageEfficiencyDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("StorageEfficiency", storageEfficiencyDynamicAttributes);
            storageEfficiencyDynamicAttributes.add("Technology");
            storageEfficiencyDynamicAttributes.add("GainLossMB");
            storageEfficiencyDynamicAttributes.add("PotentialGainLossMB");

            // FileSystem
            final Set<String> fileSystemEfficiencyDynamicAttributes = new HashSet<String>();
            elementNameToDynamicAttributes.put("FileSystem", fileSystemEfficiencyDynamicAttributes);
            fileSystemEfficiencyDynamicAttributes.add("AllocatedCapacityMB");
            fileSystemEfficiencyDynamicAttributes.add("UsedCapacityMB");
        }


        private OdataChangeDetector() {}

        public void startElement(String uri, String localName, String qName, Attributes attributes)
                    throws SAXException {

            final Set<String> dynamicAttributes = elementNameToDynamicAttributes.containsKey(qName)
                    ? elementNameToDynamicAttributes.get(qName)
                    : new HashSet<String>();
            final int attributeCount = attributes.getLength();
            for (int i=0; i< attributeCount; i++) {
                final String name = attributes.getQName(i);
                final String value = attributes.getValue(i);
                if (name == null || value == null) {
                    continue;
                }

                // Always update the full CRC
                fullCRC.update(value.getBytes());

                // Update static CRC if this attribute is static
                if (!dynamicAttributes.contains(name)) {
                    staticCRC.update(value.getBytes());
                }
            }
        }
    }

}
