package com.onaro.sanscreen.acquisition.foundation.datasource;

import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.odata2.*;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Finalize all connected hubs in a fabric.
 * <p/>
 * Hubs are used to model one-to-one connectivity, many-to-one connectivity on
 * a single VSAN, many-to-one connectivity on multiple VSANs, and connectivity
 * to NPV devices. They are finalized as follows:
 * <ul>
 * <li> Hubs representing an NPV device with known connectivity are replaced
 * with an NPV physical switch, regardless of how many connected ports they
 * have. If the same NPV device is connected to multiple ports on the same
 * switch, the same NPV physical switch is used for all ports.
 * <li> Hubs with no connectivity are disconnected from their switch port.
 * <li> Hubs representing one-to-one connectivity are replaced with a single
 * port.
 * <li> Hubs representing many-to-one connectivity on a single VSAN are
 * replaced with an NPV hub.
 * <li> Hubs representing many-to-one connectivity on multiple VSANs cause an
 * exception since we currently can't represent them.
 * </ul>
 */
public class HubFinalizer implements OdataFinalizer {

    private final static Logger logger =
            Logger.getLogger(HubFinalizer.class);

    /** The root of the fabric where the connected hubs are defined. */
    private final FabricRoot fabricRoot;

    public HubFinalizer(final FabricRoot fabricRoot) {
        this.fabricRoot = fabricRoot;
    }

    /**
     * Finalize all connected hubs in the current fabric.
     */
    @Override
    public void finalizeOdata() throws SchemaException {
        finalizeHubs();
    }

    /**
     * Finalize all connected hubs in the current fabric.
     */
    protected void finalizeHubs() throws SchemaException {
        // First find all hubs connected to fabric switches
        final List<Hub> fabricHubs = new ArrayList<Hub>();
        for (final Fabric fabric : fabricRoot.getFabrics()) {
            // Find hubs on fabric switches
            for (final Switch fabricSwitch : fabric.getSwitches()) {
                findHubs(fabricSwitch, fabricHubs);
            }
            // Find hubs on logical switches
            if (fabric instanceof VSAN) {
                for (final LogicalSwitch logicalSwitch :
                        ((VSAN)fabric).getLogicalSwitches()) {
                    findHubs(logicalSwitch, fabricHubs);
                }
            }
        }

        // Keep track of finalized switch ports, indexed by switch port WWN.
        // key = finalized switch port WWN
        // value = finalized switch port
        final Map<String, SwitchPort> finalizedSwitchPorts =
                new HashMap<String, SwitchPort>();

        // Get all unknown ports indexed by port WWN
        // key = WWN of unknown port
        // value = unknown port
        final Map<String, UnknownPort> unknownPorts = mapUnknownPorts(fabricRoot);

        // For each fabric switch hub, create an appropriate replacement.
        // We start with fabric hubs because the VSAN information is more
        // readily available than on chassis hubs.
        for (final Hub hub : fabricHubs) {
            finalizeFabricHub(hub, finalizedSwitchPorts, unknownPorts);
        }

        // Now find all hubs connected to chassis switches (regular and NPV)
        // Note: NPV chassis hubs are created when NPV switches are acquired
        // directly
        final List<Hub> chassisHubs = new ArrayList<Hub>();
        for (final ChassisSwitch chassisSwitch : fabricRoot.getChassisSwitches()) {
            switch (chassisSwitch.getSwitchType()) {
                case CHASSIS:
                case NPV_CHASSIS:
                    findHubs(chassisSwitch, chassisHubs);
                    break;
                default:
                    // Skip
                    continue;
            }
        }

        // Process all chassis hubs
        // We now have two types of chassis hubs:
        // 1- hubs whose connectivity we've already figured out because they're
        //    also connected to fabric switches
        // 2- hubs with unknown connectivity so far
        for (final Hub hub : chassisHubs) {
            finalizeChassisHub(hub, finalizedSwitchPorts);
        }
    }

    /**
     * Finalize the specified connected fabric hub.
     *
     * @param hub The connected fabric hub to finalize.
     * @param finalizedSwitchPorts Used to keep track of the WWN of the hub
     *        replacements.
     */
    protected void finalizeFabricHub(final Hub hub,
            final Map<String, SwitchPort> finalizedSwitchPorts,
            final Map<String, UnknownPort> unknownPorts)
            throws SchemaException {
        // Determine what type of hub replacement is needed
        final HubPorts nodePorts = hub.getNodePorts();
        final SwitchPort switchPort = hub.getSwitchPort();
        if (hub.isConnectedNpvDevice()) {
            // Always create NPV switch and ports
            createNpvSwitchAndPorts(hub, unknownPorts);

            // Keep track of finalized switch ports
            finalizedSwitchPorts.put(switchPort.getWWN(), switchPort);
        }
        else {
            if (nodePorts.size() == 0) {
                // We have no node ports, so just disconnect this hub from the
                // switch port
                switchPort.disconnect();
                if (logger.isDebugEnabled()) {
                    logger.debug("Fabric hub " + hub.getWWN() + " was disconnected");
                }
            }
            else if (nodePorts.size() == 1) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Fabric hub " + hub.getWWN()
                            + " is connected to a single port");
                }
                // If a hub has one reported NPV port and only one hub port, this
                // should just be a regular connected port
                final HubPort nodePort = nodePorts.iterator().next();
                final ConnectedPort connectedPort = switchPort.createConnectedPort(
                        nodePort.getWWN());
                switchPort.setConnectedPort(connectedPort);
                finalizedSwitchPorts.put(switchPort.getWWN(), switchPort);
            }
            else { // nodePorts.size() >= 2
                if (logger.isDebugEnabled()) {
                    logger.debug("Fabric hub " + hub.getWWN()
                            + " is connected to multiple ports");
                }

                // Create NPV switch and ports
                createNpvSwitchAndPorts(hub, unknownPorts);

                // Keep track of WWN of hub replacement
                finalizedSwitchPorts.put(switchPort.getWWN(), switchPort);
            }
        }
    }

    /**
     * Finalize the specified hub connected to a chassis switch.
     *
     * @param hub The hub to finalize.
     * @param finalizedSwitchPorts Previously established switch port
     * connectivity.
     * @throws SchemaException
     */
    protected void finalizeChassisHub(final Hub hub,
            final Map<String, SwitchPort> finalizedSwitchPorts)
            throws SchemaException {
        // Check if we've already figured out connectivity
        final SwitchPort switchPort = hub.getSwitchPort();
        final SwitchPort finalizedSwitchPort = finalizedSwitchPorts.get(
                switchPort.getWWN());
        if (finalizedSwitchPort != null) {
            switchPort.setConnectedPort(
                    switchPort.createConnectedPort(finalizedSwitchPort.getConnectedPort().getWWN()),
                    finalizedSwitchPort.isUnknownConnectivity());
            return;
        }

        // Now we need to figure out connectivity ourselves
        final HubPorts nodePorts = hub.getNodePorts();
        if (nodePorts.size() == 0) {
            // We have no node ports, so just disconnect this hub from the
            // switch port
            switchPort.disconnect();
            if (logger.isDebugEnabled()) {
                logger.debug("Chassis switch hub " + hub.getWWN() + " is disconnected");
            }
        }
        else {
            throw new SchemaException("Unsupported configuration: Chassis switch port "
                    + hub.getSwitchPort().getWWN() + " is connected to "
                    + nodePorts.size()
                    + " port(s) with no matching logical switch connectivity");
        }
    }

    /**
     * We need to create a separate switch for the specified
     * hub when we recognize the port as an NPV port. We take the NPV port to
     * name the switch, and then add that port to the switch with connectivity
     * to the switch port to which it was connected. All other host ports added
     * to this hub will be given connectivity to a new "fake" port on this NPV
     * switch before marshaling.
     *
     * @throws SchemaException - if there is an error changing the model
     */
    private void createNpvSwitchAndPorts(final Hub hub,
            final Map<String, UnknownPort> unknownPorts)
            throws SchemaException {
        // Make sure the type of the main switch is what we expect
        final SwitchPort switchPort = hub.getSwitchPort();
        final FabricSwitchType mainSwitch = (FabricSwitchType)switchPort.getSwitch();
        final SwitchType.Type mainSwitchType = mainSwitch.getSwitchType();
        if ((mainSwitchType != SwitchType.Type.FABRIC_LOGICAL)
                && (mainSwitchType != SwitchType.Type.FABRIC_PHYSICAL)) {
            throw new SchemaException("Unsupported switch type " + mainSwitchType
                    + " for fabric switch " + mainSwitch.getWWN());
        }

        // Get or create the fabric to contain the generated NPV switch devices
        final Fabric switchFabric = mainSwitch.getFabric();
        final String fabricWwn = switchFabric.getWWN();
        Fabric npvFabric = switchFabric.getParent().getNPVFabrics().find(fabricWwn);
        if (npvFabric == null) {
            // Create a new Fabric element with the same fabric name.
            npvFabric = switchFabric.getParent().createFabric(fabricWwn);
            switchFabric.getParent().getNPVFabrics().add(npvFabric);
            npvFabric.setName(switchFabric.getName());
        }

        // If the connected NPV port/node WWNs are known, use them
        final boolean isConnectedNpvDevice = hub.isConnectedNpvDevice();
        final String npvPortWWN = isConnectedNpvDevice ? hub.getConnectedNpvPortWWN() : "HUB_" + switchPort.getWWN();
        final String npvNodeWWN = isConnectedNpvDevice ? hub.getConnectedNpvNodeWWN() : npvPortWWN;
        // Note: the switch might already have been created, e.g. when multiple
        // ports on an NPV device are connected to multiple ports on the switch
        Switch npvSwitch = null;
        if (isConnectedNpvDevice) {
            for (final Switch aSwitch : npvFabric.getSwitches()) {
                if (npvNodeWWN.equals(aSwitch.getWWN())) {
                    // Found existing switch!
                    npvSwitch = aSwitch;
                    break;
                }
            }
        }
        if (npvSwitch == null) {
            // Create new switch
            npvSwitch = isConnectedNpvDevice
                ? npvFabric.createNpvPhysicalSwitch(npvNodeWWN)
                : npvFabric.createNpvHubSwitch(npvNodeWWN);
            npvSwitch.setGenerated(true);
            npvFabric.getSwitches().add(npvSwitch);
        }

        // Now add the NPV port to this switch
        final SwitchPort npvPort = npvSwitch.createPort(npvPortWWN);
        final boolean unknownConnectivity;
        if (isConnectedNpvDevice) {
            npvPort.setPortType(PortType.Type.NP_PORT);
            unknownConnectivity = false;
        }
        else {
            npvPort.setPortType(PortType.Type.N_PORT);
            npvPort.setGenerated(true);
            unknownConnectivity = true;
        }
        npvSwitch.getPorts().add(npvPort);

        // Now connect this port to the other NPV port on the core switch
        final ConnectedPort switchNpvConnectedPort = npvPort.createConnectedPort(switchPort.getWWN());
        npvPort.setConnectedPort(switchNpvConnectedPort, unknownConnectivity);
        // and the other direction
        final ConnectedPort npvSwitchConnectedPort = switchPort.createConnectedPort(npvPort.getWWN());
        switchPort.setConnectedPort(npvSwitchConnectedPort, unknownConnectivity);

        // Create NPV switch ports
        final Map<String, Set<Zone>> switchPortZonesCache = new HashMap<String, Set<Zone>>();
        for (final HubPort hubPort : hub.getNodePorts()) {
            // Create a generated port and add to the NPV switch.
            // We use the Host port WWN concatenated with the fabric switch port WWN.
            final SwitchPort hostNpvPort = npvSwitch.createPort(generateHubPortWwn(hubPort));
            hostNpvPort.setPortType(PortType.Type.FL_PORT);
            hostNpvPort.setGenerated(true);
            npvSwitch.getPorts().add(hostNpvPort);

            // Connect this generated port to the host port.
            final ConnectedPort connectedPort =
                    hostNpvPort.createConnectedPort(hubPort.getWWN());
            hostNpvPort.setConnectedPort(connectedPort, true);

            // Adjust unknowns
            adjustUnknownPort(hubPort.getWWN(), hostNpvPort.getWWN(), unknownPorts);

            // Adjust zones
            adjustZones(mainSwitch, hostNpvPort.getWWN(), switchPort.getWWN(), switchPortZonesCache);
        }
    }

    /**
     * Find all hubs connected to ports on the specified switch, and add them to
     * the specified list of connected hubs.
     */
    protected void findHubs(final SwitchType<?> aSwitch,
            final List<Hub> connectedHubs) {
        for (final SwitchPort switchPort : aSwitch.getPorts()) {
            final ConnectedPort connectedPort = switchPort.getConnectedPort();
            if (connectedPort instanceof Hub) {
                connectedHubs.add((Hub)connectedPort);
            }
        }
    }

    /**
     * Find all the zones where the switch port is a member. We will need to add
     * members for each of our new fake ports.
     *
     * @return the set of zones containing the core switch port as a member
     */
    protected Set<Zone> findSwitchPortZones(final Fabric fabric,
            final String wwn) {
        final Set<Zone> switchPortZones = new HashSet<Zone>();
        findSwitchPortZones(wwn, fabric.getActiveZoningCapabilities(),
                switchPortZones);
        if (fabric.getInactiveZoningCapabilities() != null) {
            for (final InactiveZoningCapabilities inactiveZoningCapabilities :
                    fabric.getInactiveZoningCapabilities()) {
                findSwitchPortZones(wwn, inactiveZoningCapabilities,
                        switchPortZones);
            }
        }
        return switchPortZones;
    }

    /**
     * Find all zones in the specified zoning capabilities where the switch port
     * is a member, and add them to the specified set of zones.
     */
    protected void findSwitchPortZones(final String wwn,
            final ZoningCapabilities zoningCapabilities,
            final Set<Zone> switchPortZones) {
        if (zoningCapabilities != null) {
            for (final Zone zone : zoningCapabilities.getZones()) {
                for (final ZoneMember zoneMember : zone.getZoneMembers()) {
                    if ((zoneMember.getType() ==
                            ZoneMember.ZoneMemberTypeEnum.SWITCH_PORT)
                            && wwn.equals(zoneMember.getWWN())) {
                        switchPortZones.add(zone);
                    }
                }
            }
        }
    }

    /**
     * Create a map of all the {@link UnknownPort} objects by their WWN.
     * <p/>
     * This is used when we want to replace the connectivity of an Unknown port
     * to point to one of our new fake ports.
     *
     * @param root The odata root.
     * @return a map of unknown ports.
     */
    protected Map<String, UnknownPort> mapUnknownPorts(
            final OData2Root root) {
        final Map<String, UnknownPort> unknownPorts =
                new HashMap<String, UnknownPort>(1024);

        for (final Unknown unknown : root.getUnknowns()) {
            for (final UnknownPort unknownPort : unknown.getPorts()) {
                unknownPorts.put(unknownPort.getWWN(), unknownPort);
            }
        }

        return unknownPorts;
    }

    /**
     * Adjust unknown port connectivity by replacing connectivity to the
     * original port with connectivity to the new port.
     *
     * @param originalPortWwn WWN of the original port.
     * @param generatedPortWwn WWN of the new port.
     * @param unknownPorts Map of unknown ports.
     */
    protected void adjustUnknownPort(final String originalPortWwn,
            final String generatedPortWwn,
            final Map<String, UnknownPort> unknownPorts)
            throws SchemaException {
        final UnknownPort unknownPort = unknownPorts.get(originalPortWwn);
        if (unknownPort != null) {
            // Now replace the connection on the unknown with connection to the
            // new port
            unknownPort.setConnectedPort(
                    unknownPort.createConnectedPort(generatedPortWwn));
            unknownPort.setUnknownConnectivity(true);
        }
    }

    /**
     * Adjust zones by adding the generated port as a member to any
     * zone that contains the switch port
     */
    protected void adjustZones(final FabricSwitchType fabricSwitch,
            final String generatedPortWwn, final String switchPortWwn,
            final Map<String, Set<Zone>> switchPortZonesCache)
            throws SchemaException {
        final String fabricWwn = fabricSwitch.getFabric().getWWN();
        Set<Zone> switchPortZones = switchPortZonesCache.get(fabricWwn);
        if (switchPortZones == null) {
            switchPortZones = findSwitchPortZones(fabricSwitch.getFabric(), switchPortWwn);
            switchPortZonesCache.put(fabricWwn, switchPortZones);
        }
        for (final Zone zone : switchPortZones) {
            zone.getZoneMembers().add(zone.createPortZoneMember(generatedPortWwn));
        }
    }

    /**
     * Generate a WWN for the specified hub port.
     */
    protected String generateHubPortWwn(final HubPort hubPort) {
        return hubPort.getWWN() + "-"
                + hubPort.getHub().getSwitchPort().getWWN();
    }

}
