package com.onaro.sanscreen.acquisition.foundation.datasource;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;

/**
 * Defined the interface for allocating originator IDs for data sources
 */
public interface OriginatorProvider {
    /**
     * Called by the data source to determine which originatorID it should report on based on a datasourceName and a deviceKey
     *
     * @param deviceKey
     * @param originatorDisplayName
     */
    public long getOriginatorID(final Long datasourceID, final String deviceKey, final String originatorDisplayName) throws FrameworkException;

    /**
     * Calls into the database to retrieve an OriginatorID from the server
     * 1) If datasourceID is null, an exception will be thrown in the server
     * 2) If datasourceID is non-null, and devicekey is null, and a datasource row exists in the database, the
     * originatorID will be returned. If none exists, it will be created and returned
     * 4) If datasourceID is non-null and devicekey is non-null, an existing datasourceID will be returned if present,
     * otherwise a new originatorID will be returned, and entered into the database.
     *
     * If performDuplicateDatasourceCheck flag is set to true, then a check is performed to detect duplicate datasources of same type and same device key.
     * @param datasourceID
     * @param deviceKey
     * @param performDuplicateDatasourceCheck
     *
     * @return The ID allocated for this device key
     *
     * @throws com.onaro.sanscreen.acquisition.sessions.AcquisitionUnitException
     */
    public long getOriginatorID(final long datasourceID, final String deviceKey, final String originatorName, final boolean performDuplicateDatasourceCheck) throws FrameworkException;

    /**
     * Removes the specified originator ID. This will also close all the originator lines in the data base reported by
     * this originator.
     */
    public void deleteOriginatorID(final Long originatorID, final String deviceKey) throws FrameworkException;

    /**
     * For monitoring purposes
     *
     * @return A string representation of the cache contents
     */
    public String[] getOriginatorCache();

    /**
     * For monitoring purposes. Cleans the originator cache.
     */
    public void cleanOriginatorCache();
}
