package com.onaro.sanscreen.acquisition.foundation.datasource.config;

import com.onaro.sanscreen.acquisition.foundation.datasource.BaseFoundationPackage;
import com.onaro.sanscreen.acquisition.foundation.datasource.FoundationListener;
import com.onaro.sanscreen.acquisition.foundation.datasource.FoundationUpdatesListener;
import com.onaro.sanscreen.acquisition.foundation.datasource.OriginatorProvider;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceLifeCycleListener;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceSampleListener;
import com.onaro.sanscreen.acquisition.framework.datasource.config.BasePollingPackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;
import com.onaro.util.DeviceFilter;

import java.lang.reflect.Constructor;
import java.util.Properties;

/**
 * Represents the configuration of a data source (types, properties, etc...)
 */
public abstract class FoundationConfiguration<FP extends BaseFoundationPackage> extends BasePollingPackageConfiguration<FP> {

    private boolean enabled = false;
    private String ip = "";
    private String user = "";
    private String password;
    private long trapMinimum = 0;
    private DeviceFilter filter = null;
    private boolean retireUnreportedOriginators;
    private boolean efficiencyEnabled = false;
    private boolean forceCreateZipOfRecording = false;

    /**
     * Time, in hours, that dynamic attribute change reports should be suppressed for.
     * Any value less than or equal to 0 will disable suppression of dynamic attribute
     * reports.  TODO: do we need to enforce a maximum value?   No more than 24 hours?
     */
    private int ignoreDynamicChangesForHours = 0;

    /**
     * Instantiates this configuration object. This object is contructed by the {@link PackageMetaData#instantiateConfiguration} method
     *
     * @param dataSourceConfiguration the data source configuration
     * @param metaData the data source package metadata
     * @param dataSourceName the data source name
     * @param dataSourceID the data source ID
     * @param packageAuditName the data source package audit name
     * @param props the data source properties
     *
     * @throws FrameworkException throws when data source is not valid at framework level
     */
    public FoundationConfiguration(final DataSourceConfiguration dataSourceConfiguration, final PackageMetaData metaData, final String dataSourceName, final long dataSourceID, final String packageAuditName, final Properties props) throws FrameworkException {
        super(dataSourceConfiguration, metaData, dataSourceName, dataSourceID, packageAuditName, props, true);

        // get common properties
        final String ip = properties.getProperty("ip");
        if (ip == null)
            throw new FrameworkException("Configuration for " + toString() + " is missing IP");
        this.ip = ip.trim();
        user = properties.getProperty("user");
        password = properties.getProperty("password");

        trapMinimum = Long.parseLong(properties.getProperty("trap.minimum", "10000"));

        filter = new DeviceFilter(getProperty("filter.exclude", ""));

        retireUnreportedOriginators = Boolean.valueOf(properties.getProperty("retire.unreported.originators", Boolean.FALSE.toString()));

        enabled = Boolean.parseBoolean(props.getProperty("enabled", Boolean.TRUE.toString()));
        props.remove("enabled");

        ignoreDynamicChangesForHours = Integer.parseInt(properties.getProperty("max.ignore.dynamic.for.hours", "6")); // hours

        efficiencyEnabled = Boolean.parseBoolean(props.getProperty("efficiency.enabled", Boolean.FALSE.toString()));

        forceCreateZipOfRecording = Boolean.parseBoolean(props.getProperty("force.recording","false"));
    }

    /**
     * @return true if this is a MOID data source
     */
    public abstract boolean isMOID();

    public final String getIP() {
        return ip;
    }

    public final String getUser() {
        return user;
    }

    public final String getPassword() {
        return password;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public final long getTrapMinimum() {
        return trapMinimum;
    }

    public final boolean isDrLicensed() {
        return true;
    }

    public final boolean isiSCSILicensed() {
        return true;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public final boolean isNASLicensed() {
        return true;
    }

    public final boolean isHostVirtulizationLicensed() {
        return true;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public int getMaxRecordingSamples() {
        return 1;
    }

    public boolean isEnabled() {
        return enabled;
    }

    public boolean isEfficiencyEnabled() {
        return efficiencyEnabled;
    }

    /**
     * @param id the device id
     * @return true if the device identified by <code>id</code> is filtered out
     */
    public final boolean isFiltered(final String id) {
        return filter.contains(id);
    }

    public DeviceFilter getFilter() {
        return filter;
    }

    /**
     * If true, acquisition will delete all originators ever reported
     * by the datasource that are not reported in the current
     * *completely successful* poll.
     * If false, no originators will be deleted.
     * Default value is true.
     *
     * @return true if retire unreported originator
     */
    public boolean isRetireUnreportedOriginators() {
        return retireUnreportedOriginators;
    }

    /**
     * Here to enable control through JConsole.
     * Set to true to tell acquisition to delete all originators ever reported
     * by the datasource that are not reported in the current
     * *completely successful* poll.
     * Set to false to disable this so that no originators will be deleted.
     * Default value is true.
     *
     * @param value Set to true to enable the deletion of unreported originator
     */
    public void setRetireUnreportedOriginators(final boolean value) {
        retireUnreportedOriginators = value;
    }

    /**
     * Get the length of time in hours that dynamic changes should be suppressed for
     *
     * @return the length of time in hours that dynamic changes should be suppressed for
     */
    public int getIgnoreDynamicChangesForHours() {
        return ignoreDynamicChangesForHours;
    }

    /**
     * Hook specifically for DFM annotations datasource that does not report odata and thus
     * no listener creates a recording.   The "force.recording" configuration property can
     * be set (e.g. by DFM Annotations data source) to force a recording even when no odata
     * is reported
     * @return boolean value of "force.recording", indicating whether or not to force the generation
     * of a recording zip.
     */
    public boolean isForceCreateZipOfRecording() {
        return forceCreateZipOfRecording;
    }


    /**
     * Instantiate a new configuration object for this package based on the meta data described in this class.
     *
     * <p/>The configuration object was already created in the right class loader (see {@link PackageMetaData#instantiateConfiguration})
     * so we can simply instantiate a new Foundation package without having to worry about it.
     */
    public FP instantiatePackage(final BaseDataSource dataSource) throws FrameworkException {
        // Create a foundation listener for this package...
        final FoundationListener foundationListener = new FoundationListener(this);

        return instantiatePackage(dataSource, foundationListener, foundationListener, foundationListener, foundationListener);
    }

    public FP instantiatePackage(final BaseDataSource dataSource,
                                 final DataSourceLifeCycleListener dataSourceLifeCycleListener,
                                 final DataSourceSampleListener dataSourceSampleListener,
                                 final FoundationUpdatesListener foundationUpdatesListener,
                                 final OriginatorProvider originatorProvider) throws FrameworkException {
        try {
            final Class packageImplementationClass = getMetaData().getPackageImplementationClass();

            final Class[] constructorSignature = {BaseDataSource.class,
                                                  this.getClass(),
                                                  DataSourceLifeCycleListener.class,
                                                  DataSourceSampleListener.class,
                                                  FoundationUpdatesListener.class,
                                                  OriginatorProvider.class};

            final Constructor constructor = packageImplementationClass.getConstructor(constructorSignature);

            // Construct the data source instance
            final Object[] constructorValues = {dataSource, this, dataSourceLifeCycleListener, dataSourceSampleListener, foundationUpdatesListener, originatorProvider};

            //noinspection unchecked
            return (FP) constructor.newInstance(constructorValues);
        }
        catch (Throwable e) {
            throw new FrameworkException("Failed to instantiate foundation package for data source " + getDataSourceName(), e);
        }
    }
}
