package com.onaro.sanscreen.acquisition.foundation.datasource.dependant.integrity;

import com.netapp.oci.platform.common.interfaces.data.performance.RawSample;
import com.onaro.sanscreen.acquisition.framework.datasource.IntegrityException;
import org.apache.log4j.Logger;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * Abstract class serving to coordinate run-time integrity check activity for foundation dependent packages.   Checks tend
 * to be of the "sanity" level in nature (e.g. no negatives, no absurd values, no obvious inconsistencies, etc.).
 *
 * @param <T> the type of samples being checked by the instance of this class.
 */
public abstract class FoundationDependentIntegrityChecker<T> {

    private final static Logger logger = Logger.getLogger(FoundationDependentIntegrityChecker.class);

    private final Collection<RawSample<T>> samples;
    private final String deviceId;
    private final String logPrefix;
    private final IntegrityCheckType type;

    private final List<String> errors = new LinkedList<String>();

    /**
     * Create a new integrity checker.
     *
     * @param samples samples being checked.
     * @param deviceId ID of device in question (e.g. storage array IP).
     * @param logPrefix the log prefix for the package in question.
     * @param type type of samples being checked (used internally).
     */
    protected FoundationDependentIntegrityChecker(final Collection<RawSample<T>> samples,
                                       final String deviceId,
                                       final String logPrefix,
                                       final IntegrityCheckType type) {
        
        this.samples = Collections.unmodifiableCollection(samples);
        this.deviceId = deviceId;
        this.logPrefix = logPrefix;
        this.type = type;
    }

    /**
     * Get samples being collected.  Collection is unmodifiable.
     *
     * @return unmodifiable collection of samples being checked.
     */
    public final Collection<RawSample<T>> getSamples() {
        return samples;
    }

    /**
     * Current set of errors that will be reported.   Should not be null - may be zero length in the event of no errors.
     *
     * @return the list of errors found during integrity check.
     */
    public final List<String> getErrors() {
        return errors;
    }

    /**
     * Get the type  of integrity check being performed (based on sample type).
     *
     * @return the type of check (based on sample type).
     */
    final IntegrityCheckType getType() {
        return type;
    }

    /**
     * The package log prefix.
     *
     * @return package log prefix.
     */
    public final String getLogPrefix() {
        return logPrefix;
    }

    /**
     * The ID of the device for which samples are being reported.  E.g. the IP identification of a storage array.
     *
     * @return  the ID of the device being reported.
     */
    public final String getDeviceId() {
        return deviceId;
    }

    /**
     * Did this checker find any integrity errors?
     *
     * @return  true if errors were found.
     */
    public final boolean hasErrors() {
        return errors.size() > 0;
    }

    /**
     * Add an error to internal list of integrity failures found during this check.
     *
     * @param errorStr  the formatted error string.
     */
    protected void addError(final String errorStr) {
        errors.add(errorStr);
    }

    /**
     * Return an appropriate exception that may be thrown in response to errors found during the integrity check
     * run.
     *
     * @return appropriate exception to throw if errors were found (with formatted messaging).   Returns NULL if
     * no errors were found.
     */
    protected IntegrityException generateException() {
        if (!hasErrors()) {
            return null;
        }

        return new IntegrityException("Device " + deviceId + " had " + errors.size() + " integrity violation" 
                                      + (errors.size() > 1 ? "s " : " ")
                                      + " in " + type.getDisplayStr() + " performance samples.   See log for more details.");
    }

    /**
     * Dump all errors found in a formatted string to the Acquisition log file.  Will appear annotated as an ERROR.
     */
    protected void dumpErrors() {
        if (!hasErrors()) {
            return;
        }

        final StringBuffer errorStr = new StringBuffer("Device ").append(deviceId).append(" had ")
                                            .append(errors.size()).append(" integrity violations in ")
                                            .append(type.getDisplayStr()).append(" performance samples.  Summary: ");

        int count = 1;
        for (final String error : errors) {
            errorStr.append("\n\t").append(count).append(": ").append(error);
        }

        logger.error(logPrefix + " " + errorStr.toString());

    }

    /**
     * Run integrity check on the samples in this checker instance.   This will throw an exception if there are any
     * errors, as well as dump a full accounting of errors to the log file.
     *
     * @throws IntegrityException  if any integrity failure occured.
     */
    public abstract void checkIntegrity () throws IntegrityException;

}
