package com.onaro.sanscreen.acquisition.framework.datasource;

import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import org.apache.log4j.Logger;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class encapsulates the Implementation of detailed error messages.
 * This class keeps references, through its member variables, any entity that plays
 * a role in the exchange of error messages between: Acquisition Site (producer) and
 * the Server Site (consumer).
 */
public class EnhancedErrorDetail implements Serializable {
    private static final long serialVersionUID = 1L;
    private static Logger logger = Logger.getLogger(EnhancedErrorDetail.class);

    private TemplateConverter templateConverter;
    private String guiMessage;      // Message that will be displayed on SANScreen GUI.
    private String detailedMessage; // Detailed Message built from HTML Template file.
    private String messageTitle;
    private String deviceName;

    /**
     * <u>Called inside Datasource's code only!</u>
     *
     * <p/>This constructor assummes that the input parameter <code>errorFileName</code> is provided by the Data source's code
     * creating this class. If <code>errorFileName</code> is not found, then the code will try to use the error specified
     * in <code>defaultErrorFileName</code>.
     *
     * <p/>Each Data source is responsable to create the HTML Template files.  The location of these files
     * are under <code>data-source-home/conf/error/</code> directory.  When the datasource is packaged in its JAR file
     * the jar file will contains all the HTML template files in it. If you unpack the data source jar file you'll
     * see a directory named "error" and under it you'll find all the HTML Template files.
     *
     * <p/>This class delegated the work of scaning and converting the HTML Template file to an Inner Class
     * named {@link TemplateConverter}.
     *
     * <p/>It is assume that the <code>errorFileName</code> corresponds to a file included in the Datasource's jar file.
     * This file should be inside the jar under "error" directory.
     *
     * @param errorFileName Error to use
     * @param defaultErrorFileName Error to use when <code>errorFileName</code> was not found.
     */
    public EnhancedErrorDetail(final String errorFileName, final String defaultErrorFileName) {
        templateConverter = new TemplateConverter(errorFileName, defaultErrorFileName);
    }


    public EnhancedErrorDetail(final String errorFileName, final String deviceName, final String messageTitle) {
        this.guiMessage = messageTitle;
        this.deviceName = deviceName;
        templateConverter = new TemplateConverter(errorFileName, null, messageTitle);

    }

    /**
     * <u>Called inside Datasource's code only!</u>
     *
     * <p/>This constructor assummes that the input parameter <code>errorFileName</code> is provided by the Data source's code
     * creating this class.
     *
     * <p/>Each Data source is responsable to create the HTML Template files.  The location of these files
     * are under <code>data-source-home/conf/error/</code> directory.  When the datasource is packaged in its JAR file
     * the jar file will contains all the HTML template files in it. If you unpack the data source jar file you'll
     * see a directory named "error" and under it you'll find all the HTML Template files.
     *
     * <p/>This class delegated the work of scaning and converting the HTML Template file to an Inner Class
     * named {@link TemplateConverter}.
     *
     * <p/>It is assume that the <code>errorFileName</code> corresponds to a file included in the Datasource's jar file.
     * This file should be inside the jar under "error" directory.
     *
     * @param errorFileName
     */
    @Deprecated
    public EnhancedErrorDetail(final String errorFileName) {
        templateConverter = new TemplateConverter(errorFileName);
    }

    /**
     * This constructor assumes that you are passing the context of the HTML Template file into a {@link BufferedReader}
     * instance.
     *
     * <p/>This class delegated the work of scaning and converting the HTML Template file to an Inner Class
     * named {@link TemplateConverter}.
     */
    public EnhancedErrorDetail(BufferedReader reader) {
        templateConverter = new TemplateConverter(reader);
    }

    /**
     * Adds key/value pair data used in the construction of the variable-text messages found in the
     * HTML template file being used by this instance.  During construction HTML template file name is set {@link EnhancedErrorDetail}
     *
     * @param paramName token's name as it appears in HTML template file.
     * @param paramValue token's value used to replace with variable text found in HTML template.
     */
    public void addParam(final String paramName, final String paramValue) {
        templateConverter.addParam(paramName, paramValue);
    }


    protected TemplateConverter getTemplateConverter() {
        return templateConverter;
    }

    /**
     * Returns a String representing the enhanced error message built on the fly.
     * The error message is created only once the first time this method is called.
     * Subsequent calls will return the already created message.
     *
     * <p/>Once the user has called the method name {@link #getMessage()} an internal String named {@link #detailedMessage}
     * will have the file's context with the variable tokens replaced.
     *
     * @return converted error message.
     */
    public final String getMessage() {
        if (detailedMessage == null)
            detailedMessage = templateConverter.getMessage();

        return detailedMessage;
    }

    /**
     * Returns a String representing the GUI message built on the fly.
     * If the error message already was created then it is returned.
     *
     * @return converted error message.
     */
    public final String getGuiMessage() {
        if (guiMessage == null) {
            guiMessage = templateConverter.getTitle();
        }
        return guiMessage;
    }

    public String getMessageTitle() {
        return messageTitle;
    }

    public void setMessageTitle(final String messageTitle) {
        this.messageTitle = messageTitle;
    }

    public String getDeviceName() {
        return deviceName;
    }

    public void setDeviceName(final String deviceName) {
        this.deviceName = deviceName;
    }

    public void setGuiMessage(final String guiMessage) {
        this.guiMessage = guiMessage;
    }

    /**
     * Utility class that encapsulate the logic for reading, scanning and replacing the variable tokens from a HTML
     * template file.
     */
    static class TemplateConverter implements Serializable {
        private static final int ONE_KB = 1024;
        private static final String LEFT_MARK_REGEX = ".*\\$\\{";
        private static final String RIGHT_MARK_REGEX = "\\}.*";
        private static final String LEFT_MARK = "${";
        private static final String RIGHT_MARK = "}";

        private Map<String, String> parameters = new HashMap<String, String>();
        private Set<String> tokens = new HashSet<String>(); // list of variable tokens presents in the template file.
        private String dataFile;                            // template file data.
        private String fileName;                            // HTML Template file
        private boolean hasErrors;                          // flag in case that there is an error at creation time.
        private String title;                               // HTML's title.
        private PollingPackage thePackage;
        private String guiMessage;


        /**
         * Constructor.
         *
         * @param errorFile HTML template file name
         */
        public TemplateConverter(final String errorFile) {
            if (errorFile == null) {
                logger.error("Impossible create an Enhanced Error Message.  Template file's name is null");
                return;
            }

            this.fileName = errorFile;

            // The file is read and scaned in order to collect all the variables token available in this file.
            initData(getFileData(errorFile));

        }


        /**
         * Constructor.
         *
         * @param errorFile HTML template file name
         * @param defaultErrorFile HTML template file name to use when <code>errorFile</code> is not found
         */
        public TemplateConverter(final String errorFile, final String defaultErrorFile, final String messageTitle) {
            if (errorFile == null && defaultErrorFile == null) {
                logger.error("Impossible create an Enhanced Error Message.  Template file's name is null");
                return;
            }

            guiMessage = messageTitle;
            this.fileName = errorFile;

            // The file is read and scaned in order to collect all the variables token available in this file.
            String fileData = null;

            if (errorFile != null)
                fileData = getFileData(errorFile);

            // If we couldn't find the error file, let's try the default error file
            if (fileData == null)
                fileData = getFileData(defaultErrorFile);

            initData(fileData);

        }


        /**
         * Constructor.
         *
         * @param errorFile HTML template file name
         * @param defaultErrorFile HTML template file name to use when <code>errorFile</code> is not found
         */
        public TemplateConverter(final String errorFile, final String defaultErrorFile) {
            if (errorFile == null && defaultErrorFile == null) {
                logger.error("Impossible create an Enhanced Error Message.  Template file's name is null");
                return;
            }

            this.fileName = errorFile;

            // The file is read and scaned in order to collect all the variables token available in this file.
            String fileData = null;

            if (errorFile != null)
                fileData = getFileData(errorFile);

            // If we couldn't find the error file, let's try the default error file
            if (fileData == null)
                fileData = getFileData(defaultErrorFile);

            initData(fileData);

        }


        /**
         * Constructor.   BufferedReader with the HTML Template data is requiered as input parameter.
         */
        public TemplateConverter(final BufferedReader reader) {
            if (reader == null) {
                logger.error("Impossible create an Enhanced Error Message.  Reaser obj. is null");
                return;
            }

            // The file is read and scaned in order to collect all the variables token available in this file.
            initData(getFileData(reader));
        }

        /**
         * Enable the client to add parameters that this class will use to replace the variable tokens with.
         *
         * @param paramName
         * @param paramValue
         */
        public void addParam(String paramName, String paramValue) {
            parameters.put(paramName, paramValue);
        }

        /**
         * Parses HTML's title.
         *
         * @return String with HTML Template title.
         */
        public final String getTitle() {
            if (guiMessage != null)
                return guiMessage;
            return parseTitle();
        }

        public PollingPackage getThePackage() {
            return thePackage;
        }

        public void setThePackage(final PollingPackage thePackage) {
            this.thePackage = thePackage;
        }

        /**
         * Parses HTML's context file with variable tokens collected by the invocation of {@link #addParam(String,String)}
         * method.
         *
         * @return String with HTML's context file converted.
         */
        public final String getMessage() {
            return parseTemplate();
        }

        /**
         * Extracts HTML's title tag.  The logic attempts to find the lower case: title.
         * if this is not found then a retry is perform but this time using upper case: TITLE
         *
         * @return title information found in HTML file.
         */
        private String parseTitle() {
            if (hasErrors)
                return "";

            // Regular exp. to finds the title.
            String extractTitle = ".*\\<title\\>" + "(.+)" + "\\</title\\>.*";
            Pattern patternTitle = Pattern.compile(extractTitle);

            // Parse the Template file.  You could change the token's values and re-parse the template.
            String output = parseTemplate();
            Matcher matcherTitle = patternTitle.matcher(output);
            if (matcherTitle.find()) {
                title = matcherTitle.group(1);
            }
            else {
                extractTitle = ".*\\<TITLE\\>" + "(.+)" + "\\</TITLE\\>.*";
                patternTitle = Pattern.compile(extractTitle);
                matcherTitle = patternTitle.matcher(dataFile);
                if (matcherTitle.find()) {
                    title = matcherTitle.group(1);
                }
            }

            return replaceTokens(title, parameters);
        }

        /**
         * Replaces the variable tokens found in the file with the information provided by the user through
         * the internal Map object named <code>parameters</code>, this instance represent a key/value pair used
         * to replace the variable text in the template file.
         *
         * <p/>Before the processing of this parsing method, a non-blocking (no exception thrown) check validation is perfomed
         * to flag all the variable tokens that exist in the message but were not set by the list of parameters given by
         * the user.
         *
         * @return File's context converted.
         */
        private String parseTemplate() {
            if (hasErrors)
                return "";

            // Checks if all the variable tokens where specified by the user.
            // Do not fail is one or more variable token were not set be the user.
            checkParameters();

            // Gets the context of this HTML file.
            String output = dataFile;
            //todo try not to parse twice
            if (thePackage != null)
                createURL();

            output = replaceTokens(output, parameters);


            return output;
        }

        /**
         * Given a text to work on, this method replaces all the tokens in the text with their
         * corresoponding value.
         *
         * @param text text to work on
         * @param parameters values of tokens to replace with
         *
         * @return The fixed version of the text
         */
        private String replaceTokens(final String text, final Map<String, String> parameters) {

            if (text == null) {
                logger.info("Title information not found for file:" + fileName);
                return null;
            }

            String finalText = text;
            // Regular exp. to finds the tokens.
            for (Map.Entry<String, String> entry : parameters.entrySet()) {
                final String tokenToFind = entry.getKey();
                final String valueToReplaceWith = entry.getValue() != null ? entry.getValue() : "";
                finalText = finalText.replace(LEFT_MARK + tokenToFind + RIGHT_MARK, valueToReplaceWith);
            }

            return finalText;
        }

        /**
         * Checks if all the file's tokens are covered by the parameters passed in.
         * If the at least one token on the file is missing from the parameters given to this method
         * then it writes a log error message informing that there are missing tokens not being set.
         */
        private void checkParameters() {

            for (String s : tokens) {
                if (!parameters.containsKey(s)) {
                    if (logger.isDebugEnabled()) {
                        logger.debug("EnhancedErrorDetailsFactory -  Token not set. File named: [" + fileName + "] Token: [" + s + "]");
                    }
                    // In order to avoid the HTML including the ${paramName} being displayed on the scren,
                    // add an space for all the parameters that were not set by the user.
                    parameters.put(s, " ");
                }
            }
        }

        /**
         * Extracts all the variable tokens from the file and builds a list with all of them.
         */
        private Set<String> scanTokens(final String dataFile) {
            Set<String> listOfTokens = new HashSet<String>();

            String extract = LEFT_MARK_REGEX + "(.+)" + RIGHT_MARK_REGEX;
            Pattern pattern = Pattern.compile(extract);

            // Determine if pattern exists in input
            Matcher matcher = pattern.matcher(dataFile);
            while (matcher.find()) {
                // Get matching string
                String match = matcher.group(1);

                // Get indices of matching string
                listOfTokens.add(match);
            }

            return listOfTokens;
        }


        /**
         * Returns a Reader object's context into a String.
         */
        private String getFileData(final BufferedReader reader) {
            final StringBuilder sb = new StringBuilder(ONE_KB * 3);

            try {
                while (true) {
                    final String line = reader.readLine();

                    if (line == null)
                        break;

                    sb.append(line).append("\n");
                }
                return sb.toString();
            }
            catch (IOException e) {
                logger.error("Impossible read BufferedReader obj.", e);
                return null;
            }
        }

        /**
         * Returns a string with the context of the file.
         */
        private String getFileData(final String fileName) {
            // Takes the classloader created when the data source was created.  In this way we'll share the "context"
            // If we don't use the same classloader then we'll never find the HTML Template file.
            // Thread class loader is set in {@link DataSource.pullNow}
            final ClassLoader classLoader = Thread.currentThread().getContextClassLoader();

            // case sensitive lookup name.
            InputStream is = classLoader.getResourceAsStream("errors/" + fileName);
            if (is == null) {
                logger.error("Impossible create an Enhanced Error Message using template file named: [errors/" + fileName + "] not found.");
                return null;
            }

            if (logger.isDebugEnabled())
                logger.debug("Enhanced Error Message using template file named: [errors/" + fileName + "] was loaded.");

            StringBuilder sb = new StringBuilder(ONE_KB * 3);
            BufferedReader in = null;

            try {
                in = new BufferedReader(new InputStreamReader(is, "utf-8"));

                while (true) {
                    String line = in.readLine();
                    if (line == null)
                        break;

                    sb.append(line).append("\n");
                }

                return sb.toString();
            }
            catch (UnsupportedEncodingException e) {
                logger.error("EnhancedErrorDetailsFactory - UnsupportedEncodingException", e);
                hasErrors = true;
            }
            catch (IOException e) {
                logger.error("EnhancedErrorDetailsFactory -  IOException", e);
                hasErrors = true;
            }
            finally {
                if (in != null) {
                    try {
                        in.close();
                    }
                    catch (IOException e) {
                        logger.error("Failed to close error file: " + e.getMessage(), e);
                    }
                }
            }

            return null;
        }

        /**
         * Checks if the HTML Template data is available, if so then extract the variable tokens in the file.
         */
        private void initData(final String data) {
            if (data == null) {
                // Sets the flag that indicates that an error was produced.
                // This flag is used later on to check if we have good data to continue with the extraction logic.
                hasErrors = true;
                return;
            }

            // Original context of the file without any convertion.
            dataFile = data;

            // List of variable tokens found during file scan.
            tokens = scanTokens(dataFile);
        }

        private void createURL() {
            // final StringBuilder nowSiteURL = new StringBuilder(ServiceManager.getInstance().getProperty("now_site_url", "http://google-search.corp.netapp.com"));
            // final StringBuilder nowSiteURL = new StringBuilder("http://now.netapp.com/AskNOW");

            //  Base URL The new Base URL for standard Support Site Search is:
            //      http://support.netapp.com/search?access=a&output=xml_no_dtd&site=gs&client=gs&proxystylesheet=gs&getfields=*&filter=p&num=50
            //  Keywords / Query String The "query" or "keywords" field is the "q" parameter. So if you want to search for "metrocluster", you'd add this parameter:
            //      q=metrocluster
            //  All Together To send a fully-qualified request to Support Site Search to search for the keyword "metrocluster", your final URL would look like this:
            //      http://support.netapp.com/search?access=a&output=xml_no_dtd&site=gs&client=gs&proxystylesheet=gs&getfields=*&filter=p&num=50&q=metrocluster
            
            final String urlBase = "http://support.netapp.com/search?access=a&output=xml_no_dtd&site=gs&client=gs&proxystylesheet=gs&getfields=*&filter=p&num=50";
            final StringBuilder nowSiteURL = new StringBuilder(urlBase);

            final List<String> packageKeyWords = new ArrayList<String>();
            final PackageConfiguration configuration = thePackage.getConfiguration();
            packageKeyWords.addAll(Arrays.asList(configuration.getMetaData().getDataSourceMetaData().getName().split("\\s")));
            packageKeyWords.add(thePackage.getPackageName());
            if (guiMessage != null) {
                packageKeyWords.addAll(Arrays.asList(guiMessage.split("\\s")));
                addParam("err.msg", guiMessage);
            }

            //concate search query params
            nowSiteURL.append("&q=sanscreen");
            for (final String keyword : packageKeyWords) {
                nowSiteURL.append("+").append(keyword);
            }

            parameters.put("kb.link", "<a href=\"" + nowSiteURL.toString() + "\">More information on the NOW Knowledge Base</a>");
        }

        Map<String, String> getParameters() {
            return parameters;
        }
    }
}

