package com.onaro.sanscreen.acquisition.framework.datasource;

import com.onaro.sanscreen.acquisition.framework.datasource.Package;

/**
 * Provides thread local context to the current package.   Datasource package implementations may use static methods
 * directly to get contextual information directly for their package.
 *
 * Only available after it is initialized in framework for the current package.   This happens well before package
 * implementation.
 */
public final class PackageContext {

    /** Thread local variable to store package reference for each thread.*/
    private static final ThreadLocal<Package<?>> threadPackage = new ThreadLocal<Package<?>>();

    /** Thread local variable to store start time (initialize) for each thread.*/
    private static final ThreadLocal<Long> startTime = new ThreadLocal<Long>() {
        @Override
        protected synchronized Long initialValue() {
            return System.currentTimeMillis();
        }
    };

    /**
     * Initialize the package context on the current thread.
     *
     * @param p the current package.
     */
    public static void initialize(Package<?> p) {
        startTime.set(System.currentTimeMillis());
        threadPackage.set(p);
    }

    public static void remove () {
        threadPackage.remove();
        startTime.remove();
    }

    /**
     * Get a handle to the current package implementation.
     *
     * @return the current package.
     */
    public static Package<?> getPackage() {
        return threadPackage.get();
    }

    /**
     * Get the current log prefix.
     *
     * @return  the current log prefix.
     */
    public static String getLogPrefix() {
      Package<?> aPackage = threadPackage.get();
        return aPackage == null ? "unknown" :  aPackage.getLogPrefix();
    }

    /**
     * Get the start time for the current Foundation poll.  Initialized on first call.
     *
     * @return  the start time of the current foundation poll.
     */
    public static long getStartTime() {
        return startTime.get();
    }

    /**
     * How many seconds have progressed since the start of the current foundation poll?
     *
     * @return  seconds since the start of the current foundation poll.
     */
    public static long secondsSinceStart () {
        final Long start = startTime.get();
        return (System.currentTimeMillis() - start)/1000;
    }

    /**
     * Is the current Package in crawling mode?
     *
     * @return is the current package crawling?
     */
    public static boolean isCrawling() {
        return threadPackage.get().getConfiguration().isCrawling();
    }

    /**
     * Is the current Package in being simulated?
     *
     * @return is the current package simulated?
     */
    public static boolean isSimulating() {
        return threadPackage.get().getConfiguration().isSimulating();
    }

}
