package com.onaro.sanscreen.acquisition.framework.datasource.config;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.Attribute;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;

import java.io.File;
import java.util.*;

/**
 * Implementation for the {@link PackageConfiguration} interface.
 */
public abstract class BasePackageConfiguration<P extends Package> implements PackageConfiguration<P> {

    private final DataSourceConfiguration dataSourceConfiguration;
    private final PackageMetaData metaData;

    // Properties identifying the code executed (data source & package)
    private final long dataSourceType;
    private final long dataSourceID;
    private final String dataSourceName;
    private final String packageName;
    private final String auditName;
    private final String vendor;
    private final String model;

    // Recording / Simulation related properties
    private SimulationState simulationState = SimulationState.UNDEFINED;
    private DSSessionFileReader sessionFileReader = null;

    private boolean recording = true;
    private boolean crawling = false;

    protected Properties properties;

    protected BasePackageConfiguration(final DataSourceConfiguration configuration, final PackageMetaData metaData, final String dataSourceName, final long dataSourceID, final String packageAuditName, final Properties props, final boolean defaultForRecording) throws FrameworkException {
        this.dataSourceConfiguration = configuration;
        this.metaData = metaData;

        this.dataSourceName = dataSourceName;
        this.dataSourceType = metaData.getDataSourceMetaData().getDataSourceType();
        this.dataSourceID = dataSourceID;
        this.packageName = metaData.getName();
        this.auditName = packageAuditName;

        properties = props;

        recording = Boolean.parseBoolean(props.getProperty("transport.record", Boolean.toString(defaultForRecording)));
        props.remove("transport.record");

        vendor = props.getProperty("datasource.vendor", "");
        props.remove("datasource.vendor");

        model = props.getProperty("datasource.model", "");
        props.remove("datasource.model");

        crawling = Boolean.parseBoolean(props.getProperty("transport.crawl", "false"));
        props.remove("transport.crawl");

        fillInDefaultValues();
        verifyConfiguration();
    }

    public DataSourceConfiguration getDataSourceConfiguration() {
        return dataSourceConfiguration;
    }

    public final PackageMetaData getMetaData() {
        return metaData;
    }

    public final String getMetaDataName() {
        return metaData.getDataSourceMetaData().getName();
    }

    public final String getDataSourceName() {
        return dataSourceName;
    }

    public final long getDataSourceType() {
        return dataSourceType;
    }

    public final long getDataSourceID() {
        return dataSourceID;
    }

    public final String getPackageName() {
        return packageName;
    }

    public String getAuditName() {
        return auditName;
    }

    /**
     * Retruns <code>true</code> if package is enabled for this data source. It is up to the package implementor to figure
     * out which licenses to query (using {@link com.onaro.sanscreen.acquisition.framework.mgmt.LicenseManager} or use
     * any combinations of flags to figure out if this package should be activated for this data source.
     */
    public abstract boolean isEnabled();

    public final boolean isCrawling() {
        return crawling;
    }

    public final boolean isRecording() {
        return recording;
    }

    public final void setRecording(final boolean recording) {
        this.recording = recording;
    }

    public final boolean isSimulating() {
        if (simulationState == SimulationState.UNDEFINED) {
            throw new IllegalStateException("Using isSimulating() before sessionFileReader is set !");
        }

        return simulationState == SimulationState.YES;
    }

    public final DSSessionFileReader getSimulationSession() {
        if (simulationState == SimulationState.UNDEFINED) {
            throw new IllegalStateException("Using getSimulationSession() before sessionFileReader is set !");
        }

        return sessionFileReader;
    }

    public void setSimulationSession(final DSSessionFileReader sampleFile) {
        simulationState = (sampleFile != null) ? SimulationState.YES : SimulationState.NO;
        sessionFileReader = sampleFile;
    }

    public DSSessionFileReader.Sample getCurrentSimulationSample() {
        return sessionFileReader.getCurrentSample();
    }

    public void turnOnCrawlingOnce() {
        crawling = true;
    }

    public final void turnOffCrawling() {
        crawling = false;
    }

    public final String getVendor() {
        return vendor;
    }

    public final String getModel() {
        return model;
    }

    public final File getBinDirectory() {
        // Get the acq\bin directory (this is where the record utility is saved)
        final String folder = System.getProperty("acq.home").replace('/', File.separatorChar);
        return new File(folder + File.separatorChar + "bin");
    }

    public final Properties getProperties() {
        final Properties copy = new Properties();
        copy.putAll(properties);

        return copy;
    }

    public Properties getPropertiesForRecording() {
        final Properties propertiesToSave = new Properties();

        // Add the properties (only non-empty entries)
        for (final Map.Entry<Object, Object> e : getProperties().entrySet()) {
            final String name = (String) e.getKey();
            String value = (String) e.getValue();

            if (value == null || name == null) {
                continue;
            }

            // Look up the metadata property with this name.
            // By default it is the property name, but it will be changed for list properties
            String metaDataAttributeName = name;

            // a list type attribute is in the format:
            // LIST.foo.x.bar, where x can be any number.
            // It must be looked up in the meta data with number 0
            // By convention, the 3rd item will always be the number
            final String[] parts = name.split("\\.", 4);
            if ((parts != null) && (parts.length == 4) && Attribute.Type.LIST.toString().equalsIgnoreCase(parts[0])) {
                // build the meta data attribute name with o (zero) instead of the number in the property name
                final StringBuilder attributeNameBuffer = new StringBuilder();
                attributeNameBuffer.append(parts[0].toUpperCase()); // must be uppercase "LIST"
                attributeNameBuffer.append(".");
                attributeNameBuffer.append(parts[1]);
                attributeNameBuffer.append(".0.");
                attributeNameBuffer.append(parts[3]);
                
                metaDataAttributeName = attributeNameBuffer.toString();
            }

            final Attribute propertyAttribute = getMetaData().getAttribute(metaDataAttributeName);
            if (propertyAttribute != null) {
                // we do not save encrypted fields in the recording file.

                // OCIE encryption/decryption uses a different encryption mechanism than OCI.
                // For OCIE, encryption/decryption occurs on the server, and the clients (CLI, ACQ..)
                // receive the password in clear text.
                // OCIE password field is isEncrypt() == false in order to avoid OCI style encryption/decryption
                // in the clients.
                // Make sure that the "password" field is always masked.
                if (propertyAttribute.isEncrypt() || "password".equalsIgnoreCase(metaDataAttributeName)){
                    value = "******";
                }
            }

            propertiesToSave.put(name, value);
        }

        return propertiesToSave;
    }

    public final String getProperty(final String propertyName, final String defaultValue) {
        return properties.getProperty(propertyName, defaultValue);
    }

    public final String toString() {
        return dataSourceName + "-" + packageName;
    }

    /**
     * Compares two {@link DataSourceConfiguration} objects. This is used when trying to identify changes to the data
     * source based on it's configuration.
     */
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final BasePackageConfiguration config = (BasePackageConfiguration) o;

        if (dataSourceID != config.dataSourceID) {
            return false;
        }

        if (dataSourceName != null ? !dataSourceName.equals(config.dataSourceName) : config.dataSourceName != null) {
            return false;
        }

        if (packageName != null ? !packageName.equals(config.packageName) : config.packageName != null) {
            return false;
        }

        //noinspection RedundantIfStatement
        if (properties != null ? !properties.equals(config.properties) : config.properties != null) {
            return false;
        }

        return true;
    }

    public int hashCode() {
        int result;
        result = 29 * (dataSourceName != null ? dataSourceName.hashCode() : 0);
        result = 29 * result + (int) (dataSourceID ^ (dataSourceID >>> 32));
        result = 29 * result + (properties != null ? properties.hashCode() : 0);
        return result;
    }

    /**
     * Exposed for MBean management.   Note that returns will have any "Password" property masked (set to
     * '********').
     *
     * @return properties as string array with values 'key=value'
     */
    public String[] getPropertyDescriptions() {
        final String[] descs = new String[properties.size()];

        int i = 0;
        for (final Map.Entry<Object, Object> entry : properties.entrySet()) {
            final String propName = (String) entry.getKey();
            final String propValue = propName.toLowerCase().contains("password") ? "********" : (String) entry.getValue();
            descs[i++] = propName + "=" + propValue;
        }

        return descs;
    }

    private void fillInDefaultValues() throws FrameworkException {
        BasePackageConfiguration.fillInDefaultValues(metaData, properties);
    }

    /**
     * Scans the meta data and adds default values for properties that are defined but missing in the configuration.
     *
     * @throws FrameworkException - exception
     */
    public static void fillInDefaultValues(final PackageMetaData packageMetaData, final Properties packageProperties) throws FrameworkException {
        final Properties defaultProperties = packageMetaData.getDefaultProperties(true);

        final Enumeration e = defaultProperties.propertyNames();
        while (e.hasMoreElements()) {
            final String propertyName = (String) e.nextElement();

            final String propertyDefaultValue = defaultProperties.getProperty(propertyName);

            // LIST properties define the list columns so they are always put in the properties
            if (propertyName.indexOf("LIST") > -1 && !packageProperties.containsKey(propertyName)) {
                packageProperties.setProperty(propertyName, propertyDefaultValue);
                continue;
            }

            if (!packageProperties.containsKey(propertyName) && propertyDefaultValue != null && propertyDefaultValue.length() > 0) {
                packageProperties.setProperty(propertyName, propertyDefaultValue);
            }
        }
    }

    /**
     * Verifies the configuraion against the meta data.
     * Checks missing mandatory fields, property types, redundant properties, etc...
     *
     * @throws FrameworkException -
     */
    private void verifyConfiguration() throws FrameworkException {
        // For each attribute that is defined...
        for (final Attribute attribute : metaData.getAttributes().values()) {
            final String value = properties.getProperty(attribute.getName());

            // Bypass special system attributes...
            if (attribute.isSystemAttribute()) {
                continue;
            }

            if (attribute.getType().equals(Attribute.Type.LIST)) {
                throw new FrameworkException("LIST type attribute should not be part of configuration !");
            }

            // If this is a LIST member attribute, find all instances of it (in all rows) and verify it
            if (attribute.getName().startsWith("LIST.")) {
                final String[] parts = attribute.getName().split("\\.");
                final String attributeNamePrefix = "LIST." + parts[1]; // the second part, after "LIST"
                final String attributeNameSuffix = parts[parts.length - 1]; // the last part

                // Find all properties that belong to this LIST attrbute
                for (final Map.Entry<Object, Object> entry : properties.entrySet()) {
                    final String propName = (String) entry.getKey();
                    final String propValue = (String) entry.getValue();

                    if (propName.startsWith(attributeNamePrefix) && propName.endsWith(attributeNameSuffix)) {
                        verifyAttribute(attribute, propName, propValue);
                    }
                }

                continue;
            }

            verifyAttribute(attribute, attribute.getName(), value);
        }

        // Remove all the null values (otherwise, their default values will not be read)
        for (Iterator<Map.Entry<Object, Object>> it = properties.entrySet().iterator(); it.hasNext();) {
            final Map.Entry<Object, Object> entry = it.next();

            if (entry.getValue() == null) {
                it.remove();

                //noinspection UnnecessaryContinue
                continue;
            }
        }
    }

    private void verifyAttribute(final Attribute attribute, final String name, final String value) throws FrameworkException {
        // Check mandatory value
        if (attribute.isMandatory() && (value == null || value.length() == 0)) {
            throw new FrameworkException("Attribute " + name + " is missing for data source " + dataSourceName + " package " + packageName);
        }

        if (!attribute.checkValueType(value)) {
            throw new FrameworkException("Attribute " + name + " is not of type " + attribute.getType() + " (" + value + ") for data soruce " + dataSourceName + " package " + packageName);
        }

        if (attribute.isEncrypt() && value != null) {
            properties.setProperty(name, Attribute.decrypt(value));
        }
    }

    public static List<Properties> getListFromProperties(final String listName, final Properties properties) throws FrameworkException {
        final List<Properties> result = getListsFromProperties(properties).get(listName);
        // if no list by that name, give an empty list
        // so the datasource config does try not iterate over null;
        if (result != null) {
            return result;
        }
        else {
            return new ArrayList<Properties>();
        }
    }

    /**
     * pkg.LIST.listName.0.column.name
     *
     * @param properties -
     *
     * @return Map
     *
     * @throws FrameworkException -
     */
    public static Map<String, List<Properties>> getListsFromProperties(final Properties properties) throws FrameworkException {
        final List<String> propertyNames = new ArrayList<String>();
        for (final Object propertyName : properties.keySet()) {
            propertyNames.add((String) propertyName);
        }
        Collections.sort(propertyNames);

        // Map<listName, Map<Row, Map<columnName, value>>>
        final Map<String, Map<Integer, Properties>> valuesByRowAndList = new TreeMap<String, Map<Integer, Properties>>();
        for (final String propertyName : propertyNames) {
            final String[] propertyNameParts = propertyName.split("\\.");
            if (propertyNameParts[0].equalsIgnoreCase("LIST")) {
                if (propertyNameParts.length < 4) {
                    // this is the list property LIST.listName, skip this
                    continue;
                }
                final String listName = propertyNameParts[1];
                final Integer rowNumber = Integer.parseInt(propertyNameParts[2]);

                // we skip row 0 since it defines the headings
                if (rowNumber == 0) {
                    continue;
                }
                final String columnName = buildColumnNameFromPropertyName(propertyNameParts);
                final String columnValue = (String) properties.get(propertyName);
                Map<Integer, Properties> rowsForList = valuesByRowAndList.get(listName);
                if (rowsForList == null) {
                    rowsForList = new TreeMap<Integer, Properties>();
                    valuesByRowAndList.put(listName, rowsForList);
                }
                Properties columnsForRow = rowsForList.get(rowNumber);
                if (columnsForRow == null) {
                    columnsForRow = new Properties();
                    rowsForList.put(rowNumber, columnsForRow);
                }
                columnsForRow.put(columnName, columnValue);
            }
        }

        final Map<String, List<Properties>> result = new TreeMap<String, List<Properties>>();
        for (final String listName : valuesByRowAndList.keySet()) {
            result.put(listName, createListOfRows(listName, valuesByRowAndList.get(listName)));
        }

        return result;
    }

    private static List<Properties> createListOfRows(final String listName, final Map<Integer, Properties> listRows) throws FrameworkException {
        final List<Properties> result = new ArrayList<Properties>();
        for (int rowNumber = 1; rowNumber < listRows.size() + 1; rowNumber++) {
            if (listRows.get(rowNumber) == null) {
                throw new FrameworkException("List " + listName + "is missing row " + rowNumber);
            }
            result.add(listRows.get(rowNumber));
        }
        return result;
    }

    private static String buildColumnNameFromPropertyName(final String[] nameParts) {
        final StringBuffer sb = new StringBuffer();
        for (int i = 3; i < nameParts.length; i++) {
            sb.append(nameParts[i]).append(".");
        }
        sb.deleteCharAt(sb.length() - 1);
        return sb.toString();
    }

    private static enum SimulationState {
        YES,
        NO,
        UNDEFINED
    }
}
