package com.onaro.sanscreen.acquisition.framework.datasource.config;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;

import java.io.File;
import java.util.Properties;

/**
 * Interface for all configuration objects. {@link DataSourceConfiguration} object will create a configuration instance
 * for each package defined in the XML.
 */
public interface PackageConfiguration<P extends com.onaro.sanscreen.acquisition.framework.datasource.Package> extends PackageConfigurationMBean {
    public String getDataSourceName();
    public long getDataSourceType();
    public long getDataSourceID();

    public String getPackageName();

    /**
     * @return A string representation of the package in the audit views in the GUI. These names are more customer facing (e.g. "Service Insight", etc...)
     */
    public String getAuditName();

    /**
     * Allow access to the package properties.
     *
     * Note: Using the specialized accessor methods is the preferred way to access the package configuraiton.
     */
    public Properties getProperties();

    /**
     * Retruns the set of properties the package is interested in recording. These properties can include properties from other
     * packages (performance will include foundation properties) to allow the package to independantely simulate from recording.
     */
    public Properties getPropertiesForRecording();

    /**
     * Allow access to the package properties.
     *
     * Note: Using the specialized accessor methods is the preferred way to access the package configuraiton.
     */
    public String getProperty(final String propertyName, final String defaultValue);

    /**
     * Allow access to the meta data of this package
     */
    public PackageMetaData getMetaData();

    /**
     * When simulating, the data source will retrieve data from a recording file instead of the real device.
     *
     * @return true if the data source is in simulation mode.
     */
    public boolean isSimulating();

    /**
     * When recording, the data source will dump the transport level communication to a file in a format that will later
     * allow us to simulate from this file.
     *
     * @return true if data source is in recording mode.
     */
    public boolean isRecording();

    /**
     * Set the recording mode of this package
     *
     * @param recording
     */
    public void setRecording(final boolean recording);

    /**
     * When crawling, the data source will NOT execute any logic related to the converstion of the data to SANscreen. Instead,
     * it will try to 'record' all the possible 'questions and answers' that the device is capable of answering.
     *
     * Note: Crawling is done only once, upon request. After this time the data source automatically resumes to work normally.
     *
     * @return true if the data source is in crawling mode.
     */
    public boolean isCrawling();
    public void turnOnCrawlingOnce();
    public void turnOffCrawling();

    public String getVendor();
    public String getModel();

    /**
     * Crawling is a fundamental part of the recording/simulation framework. However, not all implementation
     * can support this type of recording. This method is used by the framework to identify which of the data sources
     * is able to perform crawling.
     *
     * @return true if this implementation supports crawling
     */
    public boolean isCrawlSupported();

    public boolean isEnabled();

    public DSSessionFileReader getSimulationSession();
    public DSSessionFileReader.Sample getCurrentSimulationSample();
    public void setSimulationSession(final DSSessionFileReader recording);

    /**
     * @return The location of the bin directory (where all executables are located)
     */
    public File getBinDirectory();

    /**
     * @param dataSource
     *
     * @return A new package instance based on this configuration
     *
     * @throws FrameworkException
     */
    public P instantiatePackage(final BaseDataSource dataSource) throws FrameworkException;
}
