package com.onaro.sanscreen.acquisition.framework.datasource.io;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceOutputFileType;

import java.io.*;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class DSSessionFileReader extends DSSessionFileBase {

    private static final Logger LOG = LoggerFactory.getLogger(DSSessionFileReader.class);

    private final File zipFile;
    private ZipFile simulationZip = null;

    private final Sample rootSample = new Sample(rootFolder);

    private long dataSourceType;
    private String dataSourceName;
    private long dataSourceID;
    private String packageName;
    private final Properties properties = new Properties();

    private final Set<Sample> samples = new TreeSet<Sample>();
    private Iterator<Sample> iterator;
    private Sample currentSample;

    public DSSessionFileReader(final DSSessionFileReader o) throws IOException {
        this(o.zipFile);
    }

    public DSSessionFileReader(final File dir, final String fileName) throws IOException {
        this(new File(dir, fileName));
    }

    public DSSessionFileReader(final File zipFile) throws IOException {
        this.zipFile = zipFile;

        if (!zipFile.exists()) {
            throw new IOException("Simulation ZIP file " + zipFile.getAbsoluteFile() + " does not exist");
        }

        if (!zipFile.getName().toLowerCase().endsWith("zip")) {
            throw new IOException("File " + zipFile.getAbsoluteFile() + " is not a ZIP file!");
        }

        retrieveSimulation();
    }

    private void retrieveSimulation() throws IOException {
        if (zipFile.isFile()) {
            retrieveFromZip();
        } else {
            retrieveFromZipFolder();
        }

        iterator = samples.iterator();
        currentSample = iterator.next();
    }

    public Set<Sample> getSamples() {
        return samples;
    }

    public Sample getCurrentSample() {
        return currentSample;
    }

    public Sample nextSample() {
        if (iterator.hasNext()) {
            currentSample = iterator.next();
        } else {
            samples.clear();
            try {
                // Need to re-open the ZIP file because you can't read the same
                // stream from the same zip-file twice
                retrieveSimulation();
            } catch (IOException ioe) {
                LOG.warn("Failed to re-initialize {}", zipFile.getName());
            }
        }
        return currentSample;
    }

    private void retrieveFromZipFolder() throws IOException {
        final File[] samplesFolders = zipFile.listFiles();
        if (samplesFolders == null) {
            return;
        }
        for (final File sampleFolder : samplesFolders) {
            if (!sampleFolder.isDirectory()) {
                continue;
            }

            final String folderName = sampleFolder.getName();
            final Sample sample;
            if (folderName.equals(rootFolder)) {
                sample = rootSample;
            } else {
                samples.add(sample = new Sample(folderName));
            }

            final File[] filesInSample = sampleFolder.listFiles();
            if (filesInSample == null) {
                return;
            }
            for (final File file : filesInSample) {
                final String fileName = file.getName();
                final DSInputStream dsInputStream = new DSInputStream(new BufferedInputStream(new FileInputStream(file)), fileName);

                if (fileName.startsWith(DataSourceOutputFileType.description.toString())) {
                    sample.description.loadFromXML(dsInputStream);
                    continue;
                }

                sample.addStream(dsInputStream);
            }
        }

        final File confFile = new File(zipFile, rootFolder + File.separator + DataSourceOutputFileType.configuration.toString());
        if (!confFile.exists()) {
            throw new IOException("Failed to find root configuration file: " + confFile.getAbsolutePath());
        }

        properties.loadFromXML(new FileInputStream(confFile));
        setupConfigurationFromProperties();
    }

    private void retrieveFromZip() throws IOException {

        simulationZip = new ZipFile(zipFile);
        // if this is an old version zip file, use backward compatinle code to read it.
        if (!isNewVersion(simulationZip)) {
            retrieveOldVersionFromZip(simulationZip);
            return;
        }

        // get InputStreams from zip file
        final Enumeration<?> entries = simulationZip.entries();

        final Map<String, Sample> nameToSample = new HashMap<String, Sample>();

        // enumerate files in the zipfile to find the ones with configured prefix
        while (entries.hasMoreElements()) {
            final ZipEntry ze = (ZipEntry) entries.nextElement();

            if (ze.isDirectory()) {
                continue;
            }

            final String folderName = figureOutFolderName(ze);
            if (folderName == null) {
                continue;
            }

            final String fileName = ze.getName().substring(folderName.length() + 1);
            final DSInputStream dsInputStream = new DSInputStream(simulationZip.getInputStream(ze), fileName);

            Sample sample;

            if (folderName.equals(rootFolder)) {
                sample = rootSample;
            } else {
                sample = nameToSample.get(folderName);
                if (sample == null) {
                    sample = new Sample(folderName);
                    nameToSample.put(folderName, sample);
                }
            }

            if (fileName.startsWith(DataSourceOutputFileType.description.toString())) {
                sample.description.loadFromXML(dsInputStream);
                continue;
            }

            sample.addStream(dsInputStream);
        }

        samples.addAll(nameToSample.values());

        loadConfigurationProperties(rootFolder, DataSourceOutputFileType.configuration.toString());
        setupConfigurationFromProperties();
    }

    private void retrieveOldVersionFromZip(final ZipFile simulationZip) throws IOException {
        final DSSessionFileReader.Sample sample = new Sample("1");
        samples.add(sample);

        final Enumeration<?> entries = simulationZip.entries();
        while (entries.hasMoreElements()) {
            final ZipEntry ze = (ZipEntry) entries.nextElement();
            final String fileName = ze.getName();
            final DSInputStream dsInputStream = new DSInputStream(simulationZip.getInputStream(ze), fileName);

            if (fileName.startsWith("configuration")) {

                final Properties loadedProperties = new Properties();
                loadedProperties.loadFromXML(dsInputStream);

                // Remove the embedded system attributes from the properties
                dataSourceType = Long.parseLong((String) loadedProperties.remove(dataSourceTypeProperty));
                dataSourceName = (String) loadedProperties.remove(dataSourceNameProperty);
                dataSourceID = Long.parseLong((String) loadedProperties.remove(dataSourceIDProperty));
                packageName = "foundation";

                for (final Map.Entry<Object, Object> entry : loadedProperties.entrySet()) {
                    final String key = (String) entry.getKey();
                    if (key.startsWith("system.originator.key.")) {
                        sample.description.put("system.originator.key" + "-" + key.substring("system.originator.key.".length()), entry.getValue());
                    } else if (key.equals("dr.package.installed") && Boolean.valueOf((String) entry.getValue())) {
//                        licenses.add(Feature.DISASTER_RECOVERY);
                    } else {
                        properties.put(key, entry.getValue());
                    }
                }

                continue;
            }

            sample.addStream(dsInputStream);
        }

        // Brocade/McDATA/QLogic snmp
        if (dataSourceType == 39 || dataSourceType == 20 || dataSourceType == 46) {
            for (final DSInputStream stream : sample.getStreams(DataSourceOutputFileType.recording)) {
                final String fileName = stream.getFileName();
                if (fileName.endsWith("_snmpbase.xml")) {
                    sample.description.setProperty("snmpbase-" + properties.getProperty("ip"), fileName);
                    break;
                }
            }
        }
        // Cisco SNMP
        else if (dataSourceType == 57) {
            for (final DSInputStream stream : sample.getStreams(DataSourceOutputFileType.recording)) {
                final String fileName = stream.getFileName();
                if (fileName.contains("_snmpbase") && !fileName.endsWith("_snmpbase.xml")) {
                    sample.description.setProperty("snmpbase-" + properties.getProperty("ip"), fileName);
                    break;
                }
            }
        }
    }

    /**
     * returns true if "root" folder is found in the zip file
     *
     * @param simulationZip the simulation zip file
     * @return true if it is new version, otherwise false
     */
    private boolean isNewVersion(final ZipFile simulationZip) {
        final Enumeration<?> entries = simulationZip.entries();
        while (entries.hasMoreElements()) {
            final ZipEntry ze = (ZipEntry) entries.nextElement();
            final String folderName = figureOutFolderName(ze);
            if (folderName != null && folderName.equals(rootFolder)) {
                return true;
            }
        }
        return false;
    }

    private String figureOutFolderName(final ZipEntry ze) {
        final String[] parts = ze.getName().split("[\\\\/]");

        return (parts.length > 1) ? parts[0] : null;
    }

    public Sample getRootSample() {
        return rootSample;
    }

    public File getFile() {
        return zipFile;
    }

    public long getDataSourceType() {
        return dataSourceType;
    }

    public String getDataSourceName() {
        return dataSourceName;
    }

    public long getDataSourceID() {
        return dataSourceID;
    }

    public String getPackageName() {
        return packageName;
    }

    public Properties getProperties() {
        return properties;
    }

    void loadConfigurationProperties(final String folder, final String configuratioFilePrefix) throws IOException {
        ZipInputStream zip = null;
        boolean found = false;
        try {
            zip = new ZipInputStream(new FileInputStream(zipFile));

            while (true) {
                final ZipEntry entry = zip.getNextEntry();
                if (entry == null) {
                    break;
                }

                if (zipEntryNameMatches(entry, folder, configuratioFilePrefix)) {
                    properties.loadFromXML(zip);
                    found = true;
                    break;
                }
            }

            if (!found) {
                throw new IOException("Could not find configuration file");
            }
        } finally {
            if (zip != null) {
                try {
                    zip.close();
                } catch (IOException e) {
                    // Fall through...
                }
            }
        }
    }

    private void setupConfigurationFromProperties() {
        // Remove the embedded system attributes from the properties
        dataSourceType = Long.parseLong((String) properties.remove(dataSourceTypeProperty));
        dataSourceName = (String) properties.remove(dataSourceNameProperty);
        dataSourceID = Long.parseLong((String) properties.remove(dataSourceIDProperty));
        packageName = (String) properties.remove(packageNameProperty);

        // HACK: for the sake of backward compatibility, default to "foundation"
        if (packageName == null) {
            packageName = "foundation";
        }
    }

    private boolean zipEntryNameMatches(final ZipEntry zipEntry, final String folder, final String prefix) {
        final String[] parts = zipEntry.getName().split("[\\\\/]");

        if (parts.length != 2) {
            return false;
        }

        //noinspection SimplifiableIfStatement
        if (!parts[0].equals(folder)) {
            return false;
        }

        return parts[1].startsWith(prefix);
    }

    public void close() {
        rootSample.close();

        for (final Sample sample : samples) {
            sample.close();
        }

        if (simulationZip != null) {
            try {
                simulationZip.close();
            } catch (final IOException ioe) {
                System.out.println("Failure: " + ioe);
                //ignore.
            }
        }
    }

    public static class Sample implements Comparable<Sample> {
        private final String name;
        private final List<DSInputStream> streams = new LinkedList<DSInputStream>();
        private final Properties description = new Properties();

        Sample(final String name) {
            this.name = name;
        }

        public List<DSInputStream> getStreams() {
            return streams;
        }

        public DSInputStream getStream(final String name) {
            for (final DSInputStream stream : streams) {
                if (stream.getFileName().equals(name)) {
                    return stream;
                }
            }
            return null;
        }

        public List<DSInputStream> getStreams(final DataSourceOutputFileType type) {
            final List<DSInputStream> result = new LinkedList<DSInputStream>();
            for (final DSInputStream stream : streams) {
                if (stream.getFileName().startsWith(type.toString())) {
                    result.add(stream);
                }
            }
            return result;
        }

        public DSInputStream getStream(final DataSourceOutputFileType type) throws FrameworkException {
            final List<DSInputStream> result = getStreams(type);
            if (result.size() != 1) {
                throw new FrameworkException("Expected to find one simulation stream of type " + type.toString() + " but found : " + result.size());
            }
            return result.get(0);
        }

        public DSInputStream getStreamByDescription(final DataSourceOutputFileType type, final String keyPrefix, final String key) throws FrameworkException {
            final String fileName = getDescriptionProperty(keyPrefix, key);
            return getStream(fileName);
        }

        public String getName() {
            return name;
        }

        public Properties getDescription() {
            return description;
        }

        public void addStream(final DSInputStream dsInputStream) {
            streams.add(dsInputStream);
        }

        @Override
        public int compareTo(final Sample o) {
            return name.compareTo(o.name);
        }

        public Object loadObject(final String keyPrefix, final String key) throws IOException {
            final String fileName = getDescriptionProperty(keyPrefix, key);

            final DSInputStream stream = getStream(fileName);
            final ObjectInputStream ois = new ObjectInputStream(stream);
            final Object o;
            try {
                o = ois.readObject();
            } catch (ClassNotFoundException e) {
                throw new IOException("Failed to read an object " + e.getMessage());
            } finally {
                ois.close();
            }
            return o;
        }

        public String getDescriptionProperty(final String keyPrefix, final String key) {
            return getDescription().getProperty(keyPrefix + "-" + key);
        }

        public Properties getDescription(final String keyPrefix) {
            final Properties result = new Properties();
            for (final Map.Entry<Object, Object> entry : description.entrySet()) {
                final String key = (String) entry.getKey();
                if (key.startsWith(keyPrefix + "-")) {
                    result.put(key.substring(keyPrefix.length() + 1), entry.getValue());
                }
            }

            return result;
        }

        public void close() {
            for (final DSInputStream stream : streams) {
                try {
                    stream.close();
                } catch (IOException ignore) {}
            }
        }
    }
}
