package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import java.net.URLClassLoader;
import java.net.URL;
import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;

/**
 * Override the existing URL class loader and change only the access method to get the URLs of this class.
 *
 * <p/>This is done to prevent from the {@link sun.rmi.server.MarshalOutputStream} implemented in to put the list of
 * URLs in the class annotation data. When these URLs are written, the Connectrix server tries to download the
 * eccApi.jar classes from the SANScreen server instead of the local class path. If we don't annotate the class
 * with these URLs, the Connectrix server falls back to the local class path and finds its local copy of the eccApi.jar.
 *
 * <p/>To track down the way the class is annotated, start in {@link sun.rmi.server.MarshalOutputStream#annotateClass}
 * which calls {@link java.rmi.server.RMIClassLoader#getClassAnnotation}. The <code>RMIClassLoader</code> in turn calls
 * the {@link sun.rmi.server.LoaderHandler#getClassAnnotation} method that checks the type of the class. If it is an
 * instance of {@link java.net.URLClassLoader} then it calls {@link java.net.URLClassLoader#getURLs()} to get all the
 * URLs to put in the annotation.
 *
 * <p/>This solved bug #6259
 */
class DataSourceInstanceClassLoader extends URLClassLoader {
    private final String name;
    private final ClassLoader parent;

    /**
     * Constructs the class loader with no parent. The parent provided as an argument we will keep in a member to
     * prevent from the {@link #loadClass} method to try and access the parent class loader without our control.
     *
     * @param name Name of package/instance owning this class loader
     * @param urls list of JAR URLs to include in the class path
     * @param parent Parent class loader
     */
    public DataSourceInstanceClassLoader(final String name, final URL[] urls, final ClassLoader parent) {
        super(urls, null);
        this.name = name;
        this.parent = parent;
    }

    /**
     * Remove URL annotation from the class (see full explanation above)
     */
    public URL[] getURLs() {
        return new URL[] {};
    }

    public final String getName() {
        return name;
    }

    /**
     * Override the load class to first try to look for this class in this class loader URL list and only if not
     * found, delegate it to the parent class loader.
     */
    protected synchronized Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
        try {
            // Try to load this class in this class loader. Remember that no parent is set for this class loader, which
            // means that if this class loader doesn't have it, the search will fail.
            return super.loadClass(name, resolve);
        }
        catch (ClassNotFoundException e) {
            // This class loader could not find it, let's try the parent
            return parent.loadClass(name);
        }
    }

    public InputStream getResourceAsStream(final String name) {
        final InputStream is = super.getResourceAsStream(name);
        if (is != null)
            return is;

        return parent.getResourceAsStream(name);
    }

    public URL getResource(String name) {
        final URL url = super.getResource(name);
        if (url != null)
            return url;

        return parent.getResource(name);
    }

    public Enumeration<URL> getResources(String name) throws IOException {
        final Enumeration<URL> resources = super.getResources(name);
        if (resources != null)
            return resources;

        return parent.getResources(name);
    }

    @Override
    public String toString() {
        return "[" + name + "] " + super.toString();
    }
}
