package com.onaro.sanscreen.acquisition.framework.datasource.metadata;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.net.URL;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Represents the configuration part of a data source, including the configuration, class loading policy, implementation
 * etc..
 *
 * <p/>This class represents the deploy descriptor file for the data source
 */
public class DataSourceMetaData implements Serializable {
    private final static long serialVersionUID = 0L;

    private static final Logger LOG = LoggerFactory.getLogger(DataSourceMetaData.class);

    private final long dataSourceType;
    private final String name;
    private final String description;

    private final VendorModelSection vendorModels;

    // Note that the order is important here (for dependencies)
    private final Map<String, PackageMetaData> packages = new LinkedHashMap<String, PackageMetaData>();

    /**
     * Create a data source meta data based on a URL. This is used by SANscreen to create a data source directly from
     * the deploy directory in the server.
     *
     * @param deployDescriptorURL
     *
     * @return The meta data object
     *
     * @throws IOException
     * @throws FrameworkException
     */
    public static DataSourceMetaData createMetaData(final URL deployDescriptorURL) throws IOException, FrameworkException {
        return new DataSourceMetaData(deployDescriptorURL.getPath(), deployDescriptorURL.openStream(), false);
    }

    /**
     * Create a data source meta data based on a deploy descriptor file.
     *
     * @param deployDescriptor
     *
     * @return The meta data object
     *
     * @throws IOException
     * @throws FrameworkException
     */
    public static DataSourceMetaData createMetaData(final String path, final InputStream deployDescriptor) throws IOException, FrameworkException {
        return new DataSourceMetaData(path, deployDescriptor, false);
    }

    /**
     * Creates a meta data class based on the given deploy descriptor XML. The class loaders in this meta data will be
     * disabled, assuming all required JARs are already loaded into the system class loader.
     *
     * @param deployDescriptor
     *
     * @return The meta data object tailored for testing
     *
     * @throws FrameworkException
     * @throws IOException
     */
    public static DataSourceMetaData createWithClassLoadersDisabled(final String path, final InputStream deployDescriptor) throws FrameworkException, IOException {
        return new DataSourceMetaData(path, deployDescriptor, true);
    }

    /**
     * Constructs the meta data object directly from the deployee file.
     *
     * @param path
     * @param deployeeFile
     * @param disableClassLoaders
     *
     * @throws FrameworkException
     */
    private DataSourceMetaData(String path, InputStream deployeeFile, boolean disableClassLoaders)
            throws FrameworkException
    {
        final Element root = parseDeployDescriptor(deployeeFile, path);

        // Now get the data...
        final Element datasourceType = root.getChild("dataSourceType");
        if (datasourceType == null) {
            throw new FrameworkException("Missing 'dataSourceType' element in deployee xml " + path);
        }

        this.dataSourceType = Long.parseLong(datasourceType.getTextNormalize());

        final Element name = root.getChild("name");
        if (name == null) {
            throw new FrameworkException("Missing 'name' element in deployee xml " + path);
        }

        this.name = name.getTextNormalize();

        final Element description = root.getChild("description");
        if (description == null) {
            throw new FrameworkException("Missing 'description' element in deployee xml " + path);
        }

        this.description = description.getTextNormalize();

        this.vendorModels = new VendorModelSection(this, root);

        //noinspection unchecked
        final List<Element> packageDescriptors = root.getChildren("package");
        if (packageDescriptors.size() == 0) {
            throw new FrameworkException("No packages defined in deploy descriptor " + path);
        }

        // Read XML for all packages
        for (Element packageRootElement : packageDescriptors) {
            PackageMetaData packageMetaData = PackageMetaData.create(this, packageRootElement);
            packages.put(packageMetaData.getName(), packageMetaData);
            packageMetaData.initializeClassLoaders(disableClassLoaders);
        }
    }

    /**
     * Reads the deploy descriptor XML describing the meta data of this data source
     */
    protected Element parseDeployDescriptor(final InputStream deployeeFile, final String path) throws FrameworkException {
        try {
            // Parse the deployee XML
            final Document doc = new SAXBuilder().build(deployeeFile);
            final Element root = doc.getRootElement();

            LOG.debug("Data source type - Parsed deploy file at {}", path);

            return root;
        }
        catch (JDOMException e) {
            throw new FrameworkException("Failed to parse Deployee XML file at " + path, e);
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to read Deployee XML file at " + path, e);
        }
    }

    public VendorModelSection getVendorModels() {
        return vendorModels;
    }

    public long getDataSourceType() {
        return dataSourceType;
    }

    public String getName() {
        return name;
    }

    public String getDescription() {
        return description;
    }

    /**
     * Returns the {@link PackageMetaData}s known to this data source.  The
     * order of the returned array matches the order in which the packages were
     * declared in, and subsequently loaded from, the deployment descriptor.
     *
     * @return
     */
    public PackageMetaData[] getPackages() {
        return packages.values().toArray(new PackageMetaData[] {});
    }

    public PackageMetaData getPackage(final String name) {
        return packages.get(name);
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder('[')
            .append(getDataSourceType()).append("] ")
            .append(getName()).append(" - ").append(getDescription());
        return sb.toString();
    }
}
