package com.onaro.sanscreen.acquisition.framework.filecache;

import com.onaro.commons.util.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;

/**
 * The engine for executing File Caching
 */
public class FileCacheEngine implements FileCache<DirectoryCacheKey, StringCacheData>
{
  private static final Logger log = LoggerFactory.getLogger( FileCacheEngine.class );

  /**
   * Read the data from the file located by the key
   *
   * @param key The key to the file to read
   * @return The data object with the data from the file
   * @throws IOException
   */
  @Override
  public StringCacheData read(DirectoryCacheKey key) throws IOException
  {
    StringCacheData data;
    try( BufferedReader reader = new BufferedReader( new FileReader( key.toString() ) ) )
    {
      String line;
      StringBuilder stringBuilder = new StringBuilder();
      while( (line = reader.readLine()) != null )
        stringBuilder.append( line ).append( System.lineSeparator() );

      data = new StringCacheData( stringBuilder.toString().trim() );
      if( log.isTraceEnabled() ) log.trace( "Read data with key " + key + "\n" + data );
    }
    catch( IOException e )
    {
      if( log.isTraceEnabled() ) log.trace( "Failed to read data with key " + key, e );
      throw e;
    }

    return data;
  }

  /**
   * Save the data in a file with the directory and file name define in the key
   *
   * @param key The key to the file to save the data in
   * @param data The data to store in the file
   * @throws IOException
   */
  @Override
  public void save(DirectoryCacheKey key, StringCacheData data) throws IOException
  {
    //validate the name is set for save
    if( key.getFileName() == null ) throw new RuntimeException( "Can't save key with out file name." );

    //The file to contain the data
    File file = new File( key.toString() );

    //Create the file if it does not exist
    if( !exists( key ) )
    {
      try
      {
        //create parent directories if they don't exists
        if( !file.getParentFile().mkdirs() )
        {
          //try a second time after a small wait
          try
          {
            Thread.sleep( 50 );
          }
          catch( InterruptedException e )
          {
            //ignore
          }

          //if the parent directories don't exist and can't be created
          if( !file.getParentFile().exists() && !file.getParentFile().mkdirs() )
            throw new IOException( "Could not create file " + key );
        }

        //create the file
        if( !file.createNewFile() )
          throw new IOException( "Could not create file " + key );
      }
      catch( IOException e )
      {
        throw new IOException( "Could not create file " + key, e );
      }
    }

    //Write the data to the file
    try( BufferedWriter writer = new BufferedWriter( new FileWriter( file ) ) )
    {
      //write the file
      writer.write( data.getData() );
      if( log.isTraceEnabled() ) log.trace( "Saved data with key " + key + "\n" + data );
    }
    catch( IOException e )
    {
      log.error( "Failed to save data with key " + key, e );
      throw e;
    }
  }

  /**
   * Delete a file, directory or filtered files from a directory
   *
   * @param key The key to the object(s) to delete
   * @throws IOException
   */
  @Override
  public void delete( DirectoryCacheKey key ) throws IOException
  {
    //validate
    key.doValidation();

    //if the key has a filter then delete all files that match the filter in the specified directory
    File fullPath = new File( key.toString() );
    if( key.getFileFilter() != null )
    {
      final File[] files = fullPath.listFiles( key.getFileFilter() );
      for ( final File file : files ) {
        if ( !file.delete() ) {
          log.info( "Can't remove " + file.getAbsolutePath() );
        }
      }
    }

    //delete file if a file key
    else if( key.getFileName() != null )
    {
      if( fullPath.delete() )
        log.debug( "Deleted data with key " + key );
      else
      {
        log.error( "Failed to delete data with key " + key );
        throw new IOException( "Failed to delete data with key " + key );
      }
    }

    //if the key is a directory delete everything below it
    else
    {
      if( FileUtils.deleteDir( fullPath, null ) )
        log.debug( "Deleted data with key " + key );
      else
      {
        if( !fullPath.exists() )
          throw new FileNotFoundException( "Failed to delete data with key " + key );
        else
        {
          log.error( "Failed to delete data with key " + key );
          throw new IOException( "Failed to delete data with key " + key );
        }
      }
    }
  }

  /**
   * Check for file existence
   *
   * @param key The key of the file to test for existence
   *
   * @return true if exists else false
   */
  @Override
  public boolean exists(DirectoryCacheKey key)
  {
    //Create the file if it does not exist
    File file = new File( key.toString() );

    //Validate the name is set for file check
    if( key.getFileName() == null ) throw new RuntimeException( "Missing file name" );

    //check for file existence
    return file.exists();
  }
}
