package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnitAttribute;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.server.dataobject.acquisition.AcquisitionUnitNew;
import org.apache.log4j.Logger;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.Date;
import java.util.Map;
import java.util.jar.Attributes;
import java.util.jar.JarFile;

/**
 * This service handles the life cycle of the Acquisition Unit (the acquisition service) in SANscreen. There can be several
 * acquisition units running at any given time. An acquisition unit can be either local (installed on the same machine as
 * the SANscreen server) or remote (installed on a different machine than the SANscreen server).
 *
 * <p/>Each acquisition unit registers itself in the SANscreen server and reports status of the process periodically (including
 * liveness, errors, etc...). The status of the acquisition unit is displayed in the "Acquisition Unit view" in the SANscreen
 * GUI.
 *
 * <p/>To allow SANscreen to monitor the health of the acquisition units, each AU reports liveness to the server periodically.
 * When the AU does not report for some time, the GUI will show the status of this unit in Red indicating that this service
 * needs attention.
 */
public class AcquisitionUnitManager extends FrameworkService implements AcquisitionUnitManagerMBean {
    private static Logger logger = Logger.getLogger(AcquisitionUnitManager.class);

    private final static String HEART_RATE_TASK = "Acq unit heart rate";
    private final static int LEASE_CONTRACT = 20 * 1000; // 20 secs

    private final CommunicationManager communicationManager;

    private AcquisitionUnit acquisitionUnit = null;

    private long lastSuccessfullLeaseRenewal = 0;

    private String lastAcqStatus;

    private final String acqUnitName;

    AcquisitionUnitManager(final ServiceStatusListener statusListener, final CommunicationManager communicationManager) {
        super("Acquisition Unit Manager", statusListener);
        this.communicationManager = communicationManager;

        acqUnitName = System.getProperty("domain.id", "local");
    }

    /**
     * The acquisition unit is defined in the system property <code>domain.id</code>. This property is being set in the
     * acquisition's service registry when installing.
     */
    protected void doStart() throws FrameworkException {
        register();

        // Schedule a heartbeat task to report acquisition unit status to the server...
        ServiceManager.getInstance().getAsynchronousTaskManager().schedule(HEART_RATE_TASK, new HeartBeat(), LEASE_CONTRACT);
    }

    /**
     * This method should get all system attribute of the acq unit and jar files names and time stamp.
     * if you want to add more attributes to acquisition unit ( key value pairs ) this is the place
     *
     * @param acquisitionUnit The acq unit to add attributes to
     */
    private void populateAcqUnitAttributes(final AcquisitionUnit acquisitionUnit) {
        // TODO: Add acqHome to service manager
        final File acqHomeJars = new File(System.getProperty("acq.home"));
        // Get list of all jars in Program Files/SANscreen/acq folder
        final File[] jars = acqHomeJars.listFiles(new FilenameFilter() {
            public boolean accept(final File dir, final String name) {
                return name.endsWith(".jar");
            }
        });

        // Collect JAR attributes
        for (final File jar : jars) {
            // Add timestamp attribute to the jar
            final Calendar instance = Calendar.getInstance();
            instance.setTimeInMillis(jar.lastModified());
            acquisitionUnit.addAttribute(new AcquisitionUnitAttribute("jar", jar.getName(), "timestamp", instance.getTime().toString()));

            //get jar file size
            acquisitionUnit.addAttribute((new AcquisitionUnitAttribute("jar", jar.getName(), "size", String.valueOf(jar.length()))));

            // Collect jar attributes from manifest
            final JarFile jarFile;
            try {
                jarFile = new JarFile(jar);
                final Attributes attributes = jarFile.getManifest().getMainAttributes();
                for (final Map.Entry<Object, Object> entry : attributes.entrySet()) {
                    final String key = entry.getKey().toString();
                    if (key.startsWith("onaro"))
                        acquisitionUnit.addAttribute(new AcquisitionUnitAttribute("jar", jar.getName(), key, (String) entry.getValue()));
                }
            }
            catch (IOException e) {
                logger.info(getServiceName() + "could not retrieve manifest information from jar file:" + jar.getName());
            }

            // TODO: Collect META-INF attributes (schema version, build time, signature, etc...)
        }

        // add jvm args attribute
        for (final Map.Entry<Object, Object> entry : System.getProperties().entrySet())
            acquisitionUnit.addAttribute(new AcquisitionUnitAttribute("jvm_arg", "", (String) entry.getKey(), (String) entry.getValue()));

        // add env props attribute
        for (final Map.Entry<String, String> entry : System.getenv().entrySet())
            acquisitionUnit.addAttribute(new AcquisitionUnitAttribute("env_variable", "", entry.getKey(), entry.getValue()));
    }

    /**
     * Register this unit in the SANscreen server (called when acquisition is starting).
     *
     *
     * @throws FrameworkException
     */
    public void register() throws FrameworkException {
        try {

            AcquisitionUnit auTemp = communicationManager.getAcquisitionServerSide().getAcquisitionUnit(acqUnitName);
            if (auTemp == null) {
                //this happens on first call of a newly instantiated Acquisition Unit.   In this case "AuTemp" will
                //have no data and no id.
                auTemp = new AcquisitionUnit();
                auTemp.setName(acqUnitName);
            }

            auTemp.setIp(InetAddress.getLocalHost().getHostAddress());
            auTemp.setLeasePeriod(LEASE_CONTRACT * 2);
            populateAcqUnitAttributes(auTemp);

            //set in memory representation of au to the return of register.   This has the in-memory merged
            //data from the database.  This includes an appropriate ID, even on a newly created AU the first time.
            acquisitionUnit = communicationManager.getAcquisitionServerSide().register(auTemp);
        }
        catch (UnknownHostException e) {
            throw new FrameworkException("Failed to register acquisition unit", e);
        }
        catch (AcquisitionServerException e) {
            throw new FrameworkException("Failed to register acquisition unit", e);
        }
    }

    protected void doStop() {
        ServiceManager.getInstance().getAsynchronousTaskManager().unschedule(HEART_RATE_TASK);

        // Now un-register from the acquisition unit
        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Un-registering acquisition unit...");

        unregister();
    }

    /**
     * Un-registers this acquisition unit in the SANscreen server (called when acquisition is stopping)
     */
    public void unregister() {
        try {
            communicationManager.getAcquisitionServerSide().unregister(acquisitionUnit.getId());
        }
        catch (FrameworkException e) {
            reportError("Failed to unregister acquisition unit", e);
        }
        catch (AcquisitionServerException e) {
            reportError("Failed to unregister acquisition unit", e);
        }
    }

    /**
     * @return The {@link AcquisitionUnitNew} object representing this acquisition unit
     *
     * @throws FrameworkException When not initialized
     */
    public AcquisitionUnit getAcquisitionUnit() throws FrameworkException {
        if (!isInitialized())
            throw new FrameworkException("Acquisition Unit Manager not initialized");

        return acquisitionUnit;
    }

    public String getAcquisitionUnitName() {
        return acquisitionUnit.getName();
    }

    public String getAcquisitionUnitIP() {
        return acquisitionUnit.getIp();
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Date getLastSuccessfulRenewalDate() {
        return new Date(lastSuccessfullLeaseRenewal);
    }

    public boolean renewLease() {
        try {
            communicationManager.getAcquisitionServerSide().renewLease(acquisitionUnit.getId());
            lastSuccessfullLeaseRenewal = System.currentTimeMillis();

            return true;
        }
        catch (AcquisitionServerException e) {
            reportError("Failed to renew lease for acquisition unit " + acqUnitName, e);
            return false;
        }
        catch (FrameworkException e) {
            reportError("Failed to renew lease for acquisition unit " + acqUnitName, e);
            return false;
        }
    }

    /**
     * Updates the status of the acquisition unit. This method is used by the {@link ServiceManager} when it
     * needs to report an error detected in one of the services.
     *
     * @param status New status for acq unit
     */
    public void reportAcquisitionUnitAudit(final String status) {
        //we will not report same message to server.
        if (String.valueOf(status).equals(String.valueOf(lastAcqStatus)))
            return;

        lastAcqStatus = status;

        try {
            communicationManager.getAcquisitionServerSide().reportAcquisitionUnitAudit(acquisitionUnit.getId(), status);
        }
        catch (FrameworkException e) {
            reportError("Failed to report audit for acquisition unit " + acqUnitName, e);
        }
        catch (AcquisitionServerException e) {
            reportError("Failed to report audit for acquisition unit " + acqUnitName, e);
        }
    }

    /**
     * This is a thread that is responsible for reporting the health status of this acquisition unit
     */
    private class HeartBeat implements Runnable {
        private int count = 0;

        public HeartBeat() {
        }

        public void run() {
            try {
                count++;

                renewLease();
            }
            catch (Throwable e) {
                logger.error(getServiceName() + " - Failed to report liveness", e);
            }
        }
    }
}
