package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.netapp.acquisition.aquisitionserverside.interfaces.remote.DiskModelRemote;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeForClientSession;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSession;
import com.onaro.sanscreen.server.sessions.SummarySession;
import com.onaro.util.sessions.DefaultClientLogin;
import com.onaro.util.sessions.SessionCache;
import com.onaro.util.sessions.SessionCacheException;

import org.jboss.util.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLHandshakeException;
import javax.net.ssl.SSLSocketFactory;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Handels all EJB communications with the SANscreen server.
 * <p/>
 * <p/>
 * The Communication manager takes care of configuring the right transport layer
 * (RMI or RMI-Over-HTTP), logging in, downloading HTTPS certificates and
 * handling EJB communication problems.
 * <p/>
 * <p/>
 * Data sources should try to use the package's communication services, but they
 * still have direct access (via the {@link ServiceManager}) to any session
 * exposed by the SANscreen server. This was found useful when a data source
 * wants to use the SANscreen server itself as one of the sources of data.
 * <p/>
 * <p/>
 * To avoid having to use JNDI lookup for each call, this CommunicationManager
 * caches home objects and sessions and tries to reuse them for several calls.
 * It handles any error / recovery / cleanup issues related with this cache
 * automatically.
 * <p/>
 * <p>
 * To access an EJB session, one should use the {@link #getSession(String)}
 * method: <br>
 * <code>AcquisitionFacadeSession session = ServiceManager.getInstance().getCommunicationManager().getSession(AcquisitionFacadeSession.JNDI_NAME); </code>
 * </p>
 * s *
 * <p/>
 * The CommunicationManager also provides utility methods for connecting to
 * HTTPs servers like {@link #openHttpConnection(java.net.URL)} which
 * automatically takes care of downloading the right certificates when opening
 * the connection.
 */
public class CommunicationManager extends FrameworkService implements CommunicationManagerMBean {

    private static final Logger LOG = LoggerFactory.getLogger(CommunicationManager.class);

    final String DEFAULT_ACQ_USER = "acquisition";
    final String DEFAULT_ACQ_PASSWORD = "moshekipod";
    private final String codebase;
    private SessionCache sessionCache = null;
    private final HttpBasedUrlConnection httpConnection = new HttpBasedUrlConnection();
    private final boolean standalone;
    private int defaultCertificateDownloadTimeout = 20000;

    // cache these on first successful retrieval from the server
    private String serverVersion = null;
    private String siteName = null;

    private final String userName;
    private final String password;

    /**
     * Constructor
     *
     * @param statusListener status listener to call back
     * @param standalone     a flag to indicate whether the acquisition runs in standalone mode or not
     * @param userName       the user name to communicate with server
     * @param password       the password for the user
     */
    CommunicationManager(ServiceStatusListener statusListener, boolean standalone,
            String userName, String password)
    {
        this(statusListener, standalone, userName, password, System.getProperty("codebase"));
    }

    /**
     * Constructor
     *
     * @param statusListener status listener to call back
     * @param standalone     a flag to indicate whether the acquisition runs in standalone mode or not
     * @param userName       the user name to communicate with server
     * @param password       the password for the user
     * @param codebase       server URL (e.g., "https://hostname:port")
     */
    CommunicationManager(ServiceStatusListener statusListener, boolean standalone,
            String userName, String password, String codebase)
    {
        super("Communication Manager", statusListener);

        this.standalone = standalone;
        if (standalone) {
            // initialize these since there is no server to get them from
            serverVersion = "unknown";
            siteName = "standalone";
        }

        this.userName = userName;
        this.password = password;
        this.codebase = codebase;
    }

    @Override
    protected void doStart() throws FrameworkException {
        if (codebase == null || codebase.length() == 0) {
            throw new FrameworkException("codebase system property is missing");
        }

        // Standalone implementations stop here. We do not want the communication manager to try to communicate with the
        // server.
        if (standalone) {
            // Setup HTTPS and load certificates for the standalone
            httpConnection.initSSL();
            return;
        }

        waitForServerToOpenPorts();

        setUpConnection();

        createSessionCache();

        waitForServerToFullyDeployed();
    }

    /**
     * Used when
     * <p/>
     * TODO: The session cache is not currently implemented to use the socket factory for SSL.
     */
    private void createSessionCache() {
        final int timeout = Integer.parseInt(ServiceManager.getInstance().getProperty("NETWORK_FAILURE_RECOVERY_TIMEOUT")) * 1000;
        sessionCache = new SessionCache(this.getClass().getClassLoader(), new DefaultClientLogin(this.userName, this.password), timeout);
    }

    @Override
    protected void doStop() {
        if (standalone) {
            return;
        }

        // Cleanup communication
        sessionCache.cleanCache();
    }

    @Override
    public int getDefaultCertificateDownloadTimeout() {
        return defaultCertificateDownloadTimeout;
    }

    @Override
    public void setDefaultCertificateDownloadTimeout(final int defaultCertificateDownloadTimeout) {
        this.defaultCertificateDownloadTimeout = defaultCertificateDownloadTimeout;
    }

    /**
     * Returnes an initialized {@link java.net.URLConnection} setup to work.
     * <p/>
     * This includes setting up the correct truststore (for HTTPS certificates) if neccessary.
     *
     * @param url URL to open
     * @return An initialized connection to this URL
     * @throws FrameworkException On any error
     * @throws IOException        On any IO error
     */
    public HttpURLConnection openHttpConnection(final URL url) throws IOException, FrameworkException {
        return httpConnection.openHttpConnection(url, defaultCertificateDownloadTimeout);
    }

    /**
     * Create and return an HTTP connection to the given URL with the default acq username and password
     *
     * @param url the target URL to connect to
     * @return an HTTP connection to the given URL with the provided username and password
     * @throws FrameworkException On any error
     * @throws IOException        On any IO error
     */
    public HttpURLConnection openAcqAuthenticatedHttpConnection(final URL url) throws IOException, FrameworkException {
        return openAuthenticatedHttpConnection(url, DEFAULT_ACQ_USER, DEFAULT_ACQ_PASSWORD);
    }

    /**
     * Create and return an HTTP connection to the given URL with the provided username and password
     *
     * @param url      the target URL to connect to
     * @param user     the username used to authenticate
     * @param password the password used to authenticate
     * @return an HTTP connection to the given URL with the provided username and password
     * @throws FrameworkException On any error
     * @throws IOException        On any IO error
     */
    public HttpURLConnection openAuthenticatedHttpConnection(final URL url, final String user, final String password) throws IOException, FrameworkException {
        final HttpURLConnection httpConnection = openHttpConnection(url);
        final String userPasswordString = user + ":" + password;
        final String base64UserPass = Base64.encodeBytes(userPasswordString.getBytes());
        httpConnection.addRequestProperty("Authorization", "Basic " + base64UserPass);
        return httpConnection;
    }

    public DiskModelRemote getDiskModelRemote() throws FrameworkException {
        return getSession(DiskModelRemote.JNDI_NAME);
    }

    public AcquisitionServerSession getAcquisitionServerSide() throws FrameworkException {
        return getSession(AcquisitionServerSession.JNDI_NAME);
    }

    /**
     * @return Acquisition facade session to communicate with server
     * @throws FrameworkException On any error
     * @deprecated Use {@link #getAcquisitionServerSide} instead
     */
    @Deprecated
    public AcquisitionFacadeSession getAcquisitionFacadeSession() throws FrameworkException {
        return getSession(AcquisitionFacadeSession.JNDI_NAME);
    }

    public AcquisitionFacadeForClientSession getAcquisitionFacadeForClientSession() throws FrameworkException {
        return getSession(AcquisitionFacadeForClientSession.JNDI_NAME);
    }

    public <E> E getSession(final String jndiName) throws FrameworkException {
        try {
            return (E) sessionCache.getSession(jndiName);
        } catch (SessionCacheException e) {
            throw new FrameworkException(e.getMessage(), e);
        }
    }

    @Override
    public String[] getCachedSessions() {
        return sessionCache.getCachedSessions();
    }

    @Override
    public void cleanCache() {
        sessionCache.cleanCache();
    }

    /**
     * Try to get a fresh connection to a session in the server
     */
    @Override
    public boolean testConnectionWithServer() {
        try {
            cleanCache();
            getAcquisitionServerSide();
            return true;
        } catch (FrameworkException e) {
            return false;
        }
    }

    /**
     * Used to allow the acquisition side to synchronize clocks with the server. This is especially important when
     * collecting performance data from multiple acquisition units that might have clock synchronization issues.
     *
     * @return Server time (System.currentTimeMillis() performed on the server side).
     * @throws FrameworkException On any communication issue
     */
    public long getServerTime() throws FrameworkException {
        try {
            return getAcquisitionServerSide().getServerTime();
        } catch (NullPointerException npe) {
            // Temporary fix for standalone, real fix is to use clock synchronizer, currently under review
            return System.currentTimeMillis();
        } catch (AcquisitionServerException e) {
            throw new FrameworkException("Failed to get server time", e);
        }
    }

    private SummarySession getSummarySession() throws FrameworkException {
        try {
            return getSession("SummarySession");
        } catch (FrameworkException ex) {
            throw new FrameworkException("Unable to get Summary Session from Server", ex);
        }

    }

    /**
     * gets the server version
     * <p/>
     * * The value is cached the first time it is successfully retrieved from the server
     *
     * @return The server version as a string. May return null.
     * @throws FrameworkException if communication with the server fails
     */
    public synchronized String getServerVersion() throws FrameworkException {
        if (serverVersion != null) {
            return serverVersion;
        }
        try {
            serverVersion = getSummarySession().getDatabaseVersion(); // this is the server version
            return serverVersion;
        } catch (Exception e) {
            throw new FrameworkException("Unable to get version from Server", e);
        }
    }

    /**
     * gets the site name
     * <p/>
     * The value is cached the first time it is successfully retrieved from the server
     *
     * @return The site name as a string. May return null.
     * @throws FrameworkException if communication with the server fails
     */
    public synchronized String getSiteName() throws FrameworkException {
        if (siteName != null) {
            return siteName;
        }
        try {
            siteName = getSummarySession().getSiteName();
            return siteName;
        } catch (Exception e) {
            throw new FrameworkException("Unable to get site name from server", e);
        }
    }

    /**
     * Prepare the RMI connection to connect either through normal RMI, or RMI-over-HTTP. This also sets up the proper
     * prefix that we need to add to the JNDI name of the session to get the right one.
     * <p/>
     * <p/>Here are the explanations of the system properties that control the RMI transport method:<br>
     * When working with RMI over HTTP/HTTPS: <br>
     * <ul>
     * <code>com.onaro.rmiOverHttp=true</code><br>
     * </ul>
     * URL to connect to -<br>
     * <ul>
     * <li>OVER HTTP: <code>java.naming.provider.url=${protocol}://${host}:${port}/invoker/JNDIFactory</code>
     * <li>RMI: <code>java.naming.provider.url=${host}:1099</code>
     * </ul>
     * Factory for the InitialContext -
     * <ul>
     * <li>OVER HTTP: <code>java.naming.factory.initial=org.jboss.naming.HttpNamingContextFactory</code>
     * <li>RMI: <code>java.naming.factory.initial=org.jnp.interfaces.NamingContextFactory</code>
     * </ul>
     * We are not sure what this is used for -
     * <ul>
     * <li><code>java.naming.factory.url.pkgs=org.jboss.naming:org.jnp.interfaces</code>
     * </ul>
     * Let's the system know where is the URL to use when connecting with RMI over HTTP -
     * <ul>
     * <li><code>com.onaro.http.invoker=${protocol}://${host}:${port}/invoker/JMXInvokerServlet</code>
     * </ul
     * <p/>
     * <p/><b>Note:</b>When setting up a connection using HTTPS, we have to make sure the certificate is setup correctly.
     * To do this, if the certificate does not already exist in the acq\conf\cert directory, we will download the certificate
     * from the server and save it there.
     *
     * @throws FrameworkException On any error
     */
    private void setUpConnection() throws FrameworkException {
        final Transport transportType = Transport.fromConfiguration(ServiceManager.getInstance().getProperty("TRANSPORT", "RMI"));

        LOG.debug("{} - setting up RMI connection [{}]", getServiceName(), transportType);

        // Setup HTTPS and load certificates
        httpConnection.initSSL();

        if (transportType == Transport.RMI) {
            // Tells us which JNDI name to use (NOT to add the prefix Http)
            System.setProperty("com.onaro.rmiOverHttp", Boolean.FALSE.toString());

            // Allow overriding the port by setting the system parameter
            final String namingServicePort = System.getProperty("com.onaro.naming.server.port", "1099");
            System.setProperty("java.naming.provider.url", "localhost:" + namingServicePort);

            System.setProperty("java.naming.factory.initial", "org.jnp.interfaces.NamingContextFactory");
        } else if (transportType == Transport.RMI_OVER_HTTP) {
            try {
                // Tells us which JNDI name to use (to add the prefix Http)
                System.setProperty("com.onaro.rmiOverHttp", Boolean.TRUE.toString());

                final URL codebaseURL = new URL(codebase);

                final String jndiFactory;
                if (codebaseURL.getProtocol().equalsIgnoreCase("https")) {
                    jndiFactory = codebase + "/invoker/JNDIFactorySSL";
                } else {
                    jndiFactory = codebase + "/invoker/JNDIFactory";
                }

                System.setProperty("java.naming.provider.url", jndiFactory);
                System.setProperty("java.naming.factory.initial", "org.jboss.naming.HttpNamingContextFactory");
                System.setProperty("java.naming.factory.url.pkgs", "org.jboss.naming:org.jnp.interfaces");
                System.setProperty("com.onaro.http.invoker", codebase + "/invoker/JMXInvokerServlet");
            } catch (MalformedURLException e) {
                throw new FrameworkException("Failed to read codebase as URL: " + e.getMessage(), e);
            }
        } else {
            throw new FrameworkException("Unrecognized TRANSPORT " + transportType);
        }

        try {
            // Now, if the SANscreen server is configured to use HTTPS, download and install the server's certificate
            final URL url = new URL(codebase);

            if (url.getProtocol().equalsIgnoreCase("HTTPS")) {
                // Make sure we download the certificate from the server if none is installed (and update the SSL factory)
                try {
                    // Set override=true because this is the jboss certificate and we should always download it.
                    // Possible changes include installation of a new certificate and the signing of the existing one.
                    httpConnection.downloadCertificate(url, defaultCertificateDownloadTimeout, true);
                } catch (FrameworkException e) {
                    LOG.error(null, e);
                    // Fall through in hope that this is bootup issue, and server is just not ready yet.
                }
            }
        } catch (MalformedURLException e) {
            throw new FrameworkException("Failed to parse codebase URL - " + codebase, e);
        }
    }

    /**
     * 'Pings' the server for availability. If the server is not responding, this method will wait a while and retry.
     * If the server is still not responding, then after a while, the method will give up and throw an exception.
     *
     * @throws FrameworkException If we wait past the max time as defined in the system properties.
     */
    private void waitForServerToOpenPorts() throws FrameworkException {
        final long maxTimeToWaitMs = Integer.parseInt(ServiceManager.getInstance().getProperty("WAIT_FOR_SERVER_TO_INITIALIZE")) * 1000;
        final long sleepTimeMs = 10000;
        final long startMark = System.currentTimeMillis();

        while (!isServerUp()) {
            try {
                LOG.info("Could not communicate with the server; will retry in {}ms", sleepTimeMs);
                Thread.sleep(sleepTimeMs);
            } catch (InterruptedException ignore) {}

            final long totalWait = System.currentTimeMillis() - startMark;
            if (totalWait >= maxTimeToWaitMs) {
                throw new FrameworkException("Failed to connect to server for " + maxTimeToWaitMs + " milliseconds");
            }
        }

        LOG.info("Server up!");
    }

    /**
     * Is the server up?   Returns true if so, false otherwise.  Checks by opening a simple URL connection to the
     * CODEBASE specified at startup.   Nothing is done with this connection other than confirming it can be
     * opened.
     *
     * @return true if the server is up and receiving network connections.
     */
    private boolean isServerUp() {
        LOG.info("Attempting to verify server at: {}", codebase);
        HttpURLConnection connection = null;
        try {
            final URL url = new URL(codebase);
            connection = (HttpURLConnection) url.openConnection();
            connection.connect();
        } catch (SSLHandshakeException sslException) {
            // server is up; we'll attempt to auto-fix the SSL cert issue later
            LOG.info("SSL handshake failure for {}", codebase, sslException);
        } catch (IOException ioe) {
            LOG.warn("server unreachable for {}", codebase, ioe);
            return false;
        } finally {
            if (connection != null) {
                connection.disconnect();
            }
        }
        return true;
    }

    /**
     * After the login, this method waits until the server application is fully initiailized and the acquisition
     * service side services are fully deployed.
     *
     * @throws FrameworkException On any error
     */
    private void waitForServerToFullyDeployed() throws FrameworkException {
        final long waitSeconds = Integer.parseInt(ServiceManager.getInstance().getProperty("WAIT_FOR_SERVER_TO_INITIALIZE"));
        // abortEpoch heralds when this method should give up
        final long abortEpoch = System.currentTimeMillis() + (waitSeconds * 1000L);

        Exception exceptionToReport;

        while (!isStopping()) {
            // Just try to connect to the acquisition facade session
            try {
                getAcquisitionServerSide();

                // We connected to the session. Login is successfull!
                return;
            } catch (FrameworkException fe) {
                exceptionToReport = fe;
                LOG.info("server not deployed [time remaining={}ms, cause={}]",
                        (abortEpoch - System.currentTimeMillis()), fe.getMessage());
            } catch (RuntimeException re) {
                exceptionToReport = re;
                LOG.info("server not deployed [time remaining={}ms, cause={}]",
                        (abortEpoch - System.currentTimeMillis()), re.getMessage());
            }

            if (System.currentTimeMillis() >= abortEpoch) {
                throw new FrameworkException("server not deployed; gave up after " + waitSeconds + " seconds", exceptionToReport);
            }

            try {
                Thread.sleep(10000);
            } catch (InterruptedException ignore) {}
        }
    }

    public SSLSocketFactory getSSLSocketFactory() {
        synchronized (httpConnection) {
            return httpConnection.getSslSocketFactory();
        }
    }

}
