package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionServerException;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.AcquisitionUnit;
import com.netapp.acquisition.aquisitionserverside.interfaces.data.DataSource;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.datasource.Package;
import com.onaro.sanscreen.acquisition.framework.datasource.PollingPackage;
import com.onaro.sanscreen.acquisition.framework.datasource.config.ConfigurationFactory;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PollingPackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.mgmt.scheduler.Limit;
import com.onaro.sanscreen.acquisition.framework.mgmt.scheduler.Scheduler;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSession;
import org.apache.log4j.Logger;

import javax.management.ObjectName;
import java.lang.ref.WeakReference;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Manages running data sources inside this acquisition unit
 */
public class DataSourceManager extends FrameworkService implements DataSourceManagerMBean {
    private static final Logger logger = Logger.getLogger(DataSourceManager.class);
    private final static int HOUR_MILLIS = 60 * 60 * 1000;
    private final static int MINUTE_MILLIS = 60 * 1000;
    private final static String SYNCHRONIZER_TASK = "Data source manager synchronizer";

    private final WeakReference<CommunicationManager> communicationManager;
    private final WeakReference<AcquisitionUnitManager> acquisitionUnitManager;

    private final Map<String, DataSourceInfo> datasources = new HashMap<String, DataSourceInfo>(100);
    private final Map<String, BaseDataSource> acquiringDatasource = Collections.synchronizedMap(new HashMap<String, BaseDataSource>(100));
    private final Map<String, DataSource> markedDown = new HashMap<String, DataSource>(100);

    private Scheduler<String, BaseDataSource, DataSourceManager> scheduler;

    private final ExecutorService threadPool;

    private boolean slowStart = true;

    DataSourceManager(final ServiceStatusListener statusListener,
                      final CommunicationManager communicationManager,
                      final AcquisitionUnitManager acquisitionUnitManager) {
        super("Data Source Manager", statusListener);

        this.communicationManager = new WeakReference<CommunicationManager>(communicationManager);
        this.acquisitionUnitManager = new WeakReference<AcquisitionUnitManager>(acquisitionUnitManager);

        // Initialize a thread pool that will handle the actual execution of the tasks. The thread pool is not
        // bounded because the scheduler already manages the number of tasks that can run together.
        this.threadPool = Executors.newCachedThreadPool();
    }

    protected void doStart() throws FrameworkException {
        // TODO: Currently the set of limits is hard-coded with the names of the packages. This should be revised!
        final Map<String, Limit> limits = new HashMap<String, Limit>(10);
        limits.put("foundation", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_DATASOURCES", "30"))));
        limits.put("nas", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_DATASOURCES", "30"))));
        limits.put("performance", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_PERFORMANCE_POLLS", "30"))));
        limits.put("storageperformance", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_PERFORMANCE_POLLS", "30"))));
        limits.put("hostvirtualization", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_PERFORMANCE_POLLS", "30"))));
        limits.put("rpo", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_RPO_POLLS", "30"))));
        limits.put("nerfoundation", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_DATASOURCES", "30"))));
        limits.put("netappfoundation", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_DATASOURCES", "30"))));
        limits.put("netappstorageperformance", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_PERFORMANCE_POLLS", "30"))));
        limits.put("cloud", new Limit(Integer.parseInt(ServiceManager.getInstance().getProperty("MAX_NUMBER_OF_CONCURRENT_DATASOURCES", "30"))));


        synchronized (this) {
            scheduler = new Scheduler<String, BaseDataSource, DataSourceManager>(limits, this, threadPool);
        }

        scheduler.start();
        registerMBean(scheduler, formatSchedulerObjectName());

        // When synchronizer starts, it will immediately synchronize status with the server
        ServiceManager.getInstance().getAsynchronousTaskManager().schedule(SYNCHRONIZER_TASK, new Synchronizer(this), 4000);

        slowStart = Boolean.parseBoolean(ServiceManager.getInstance().getProperty("DATASOURCES_SLOW_START", Boolean.TRUE.toString()));
    }

    protected void doStop() {
        ServiceManager.getInstance().getAsynchronousTaskManager().unschedule(SYNCHRONIZER_TASK);

        unregisterMBean(formatSchedulerObjectName());
        scheduler.stop();
        scheduler = null;

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Stopping data sources...");

        // The synchronized guarantees that we will get all data sources that were scheduled to run and tell them
        // to stop!
        final Collection<BaseDataSource> localAcquiring = new HashSet<BaseDataSource>();
        synchronized (acquiringDatasource) {
            localAcquiring.addAll(acquiringDatasource.values());
            acquiringDatasource.clear();
        }

        // Stop all running data sources
        for (final BaseDataSource dataSource : localAcquiring) {
            if (logger.isDebugEnabled())
                logger.debug(getServiceName() + " - Stopping data source " + dataSource.getDataSourceName());

            dataSource.stop();
        }
    }

    DataSourceInfo getDataSourceInfo(final String dataSourceName) {
        return datasources.get(dataSourceName);
    }

    public String[] getAcquiringDatasourceNames() {
        synchronized (acquiringDatasource) {
            final String[] names = new String[acquiringDatasource.size()];
            acquiringDatasource.keySet().toArray(names);
            return names;
        }
    }

    public synchronized String[] getAllDatasourceNames() {
        final String[] names = new String[datasources.size()];
        datasources.keySet().toArray(names);
        return names;
    }

    public synchronized String[] getMarkedDownDatasourceNames() {
        final String[] names = new String[markedDown.size()];
        markedDown.keySet().toArray(names);
        return names;
    }

    /**
     * This method syncronizes the data source manager with the server.
     * It starts all the data sources that need to be started. Stops all the data sources that need to be stopped
     *
     * @param includingUpdates If true, this method will also synchronize updates to data sources. Otherwise (false) it
     *                         would only detect data sources up / down events.
     */
    @SuppressWarnings({"ConstantConditions"})
    public synchronized void synchronizeDataSourcesWithServer(final boolean includingUpdates) {
        try {
            // First get which data sources are supposed to be active...
            final List<DataSource> openDataSourcesFromServer = communicationManager.get().getAcquisitionServerSide().getDataSourcesToRun(getAcquisitionUnit().getId());

            if (isStopping())
                return;

            // Create this result in a map for easier access
            final Map<String, DataSource> openDataSourceNamesFromServerMap = new HashMap<String, DataSource>(100);
            for (final DataSource dataSource : openDataSourcesFromServer)
                openDataSourceNamesFromServerMap.put(dataSource.getName(), dataSource);

            // Get the two groups we're working on...
            final Set<String> openDataSourceNamesFromServer = new HashSet<String>(openDataSourceNamesFromServerMap.keySet());
            final Set<String> currentlyOpenDataSourceNames = datasources.keySet();

            // Figure out which data sources to start (openDataSourceNamesFromServer - currentlyOpenDataSourceNames)
            final Set<String> dataSourceNamesToStart = new HashSet<String>(openDataSourceNamesFromServer);
            dataSourceNamesToStart.removeAll(currentlyOpenDataSourceNames);

            // Figure out which data sources to stop (currentlyOpenDataSourceNames - openDataSourceNamesFromServer)
            final Set<String> dataSourceNamesToStop = new HashSet<String>(currentlyOpenDataSourceNames);

            // Add all the marked down data sources. We need this to be sure we remove and update the data base about the status of data sources that
            // are marked down. Otherwise, the data base will indicate that they are "starting" or not "fully stopped". If a data source is not supposed
            // to be stopped, then its name will be immediately removed by next 'removeAll' line.
            dataSourceNamesToStop.addAll(markedDown.keySet());
            dataSourceNamesToStop.removeAll(openDataSourceNamesFromServer);

            // Figure out which data sources to check for updates (currentlyOpenDataSourceNames && openDataSourceNamesFromServer)
            final Set<String> dataSourceNamesToCheckForUpdate = new HashSet<String>(currentlyOpenDataSourceNames);
            dataSourceNamesToCheckForUpdate.retainAll(openDataSourceNamesFromServer);

            // Update the log (only if there is any change - this method is being called a lot!)
            if (dataSourceNamesToStart.size() > 0 || dataSourceNamesToStop.size() > 0) {
                if (logger.isDebugEnabled())
                    logger.debug(getServiceName() + " - Found " + openDataSourcesFromServer.size() +
                            " active data sources for this acquisition unit. We need to: start " +
                            dataSourceNamesToStart.size() + " data sources, stop " + dataSourceNamesToStop.size() +
                            " data sources.");
            }

            // Now do the job!
            for (final String dataSourceName : dataSourceNamesToStop) {
                if (isStopping())
                    return;

                stopDataSource(dataSourceName);
            }

            for (final String dataSourceName : dataSourceNamesToStart) {
                if (isStopping())
                    return;

                startDataSource(openDataSourceNamesFromServerMap.get(dataSourceName));
            }

            if (includingUpdates) {
                for (final String dataSourceName : dataSourceNamesToCheckForUpdate) {
                    if (isStopping())
                        return;

                    updateDataSource(dataSourceName, openDataSourceNamesFromServerMap.get(dataSourceName));
                }
            }
        }
        catch (FrameworkException e) {
            reportError("Unable to query server for open data sources", e);
        }
        catch (Throwable e) {
            reportFatal("Unexpected error when synchronizing with server - " + e.getMessage(), e);
        }
    }

  public void pollNow(final String dataSourceName, final String packageName ) {
    pollNow( dataSourceName, packageName, null );
  }

    /**
     * "Force acquisition" the given data source. This will affect only the selected package...
     *
     * <p/>The action triggers a new task in the thread pool to be executed. Since the thread pool is unbounded - this
     * action basically opens a new thread and executes this data source. If used incorrectly - this could spawn a number
     * of concurrent polls that is substantially larger than the thread limit indicated in the scheduler.
     *
     * @param dataSourceName name of datasource to run
     * @param packageName    which package to run
     * @param event The event information as Properties object
     */
    public synchronized void pollNow(final String dataSourceName, final String packageName, Properties event ) {
        // Find the data source
        final DataSourceInfo dataSourceInfo = datasources.get(dataSourceName);
        if (dataSourceInfo == null) {
            logger.error(getServiceName() + "Could not find data source to poll (" + dataSourceName + ")");
            return;
        }

        final PollingPackage packageToRun = dataSourceInfo.getDataSourceInstance().getPackage(packageName);
        if (packageToRun == null) {
            logger.error(getServiceName() + " - Could not 'pollNow' on " + dataSourceName + "/" + packageName);
            return;
        }

        // Force this task to execute now. This will not schedule it anywhere. It will allocate a thread in the thread
        // pool and run it as is.
        threadPool.submit(new PollLogic(this, packageToRun, dataSourceInfo.getDataSourceInstance(), event));
    }


  public void pollNow(final String dataSourceName) {
    pollNow( dataSourceName, (Properties)null );
  }
    /**
     * "Force Acquisition" on all packages for a specific data source
     *
     * <p/>The action triggers a new task in the thread pool to be executed. Since the thread pool is unbounded - this
     * action basically opens a new thread and executes this data source. If used incorrectly - this could spawn a number
     * of concurrent polls that is substantially larger than the thread limit indicated in the scheduler.
     *
     * @param dataSourceName name of datasource to run
     * @param event The event information as Properties object
     */
    public synchronized void pollNow(final String dataSourceName, Properties event ) {
        logger.info(dataSourceName + " - Force acquisition");

        // Find the data source
        final DataSourceInfo dataSourceInfo = datasources.get(dataSourceName);
        if (dataSourceInfo == null) {
            logger.error(getServiceName() + "Could not find data source to poll (" + dataSourceName + ")");
            return;
        }

        for (final Package aPackage : dataSourceInfo.getDataSourceInstance().getPackages().values())
            pollNow(dataSourceName, aPackage.getPackageName(), event);
    }


    public void triggerRetireOriginatorScan() throws Exception {
        final AcquisitionFacadeSession session = ServiceManager.getInstance().getCommunicationManager().getSession(AcquisitionFacadeSession.JNDI_NAME);

        session.retireExpiredOriginators();
    }


    /**
     * Starts a data source based on information acquired from the database.
     *
     * @param dataSourceFromServer datasource to start
     */
    protected synchronized void startDataSource(final DataSource dataSourceFromServer) {
        if (isStopping())
            return;

        // make sure not to start marked down datasource (unless configuarion has changed)
        final DataSource previousDataSourceInfo = markedDown.get(dataSourceFromServer.getName());
        if (previousDataSourceInfo != null) {
            // Check if dataSourceFromServer contains different information from the time this data source failed (marked down). If it does,
            // we want to retry the initiailization. Otherwise, we will not try.
            if (!DataSourceInfo.isChanged(previousDataSourceInfo, dataSourceFromServer)) {
                if (logger.isDebugEnabled())
                    logger.debug(getServiceName() + " - Cannot start data source " + dataSourceFromServer.getName() + " (it is marked down)");

                return;
            }

            // Configuration has changed... start the data source...
            markedDown.put(dataSourceFromServer.getName(), dataSourceFromServer);
        }

        // new datasource instance
        final BaseDataSource baseDataSource;

        try {
            // Double check the names of the acquisition units
            if (dataSourceFromServer.getAuId() != getAcquisitionUnit().getId())
                throw new FrameworkException("Acquisition has been asked to start a data source from a different acquisition unit (" + dataSourceFromServer.getAuId() + " vs " + getAcquisitionUnit().getId() + ")");

            if (logger.isDebugEnabled())
                logger.debug(getServiceName() + " - Starting data source " + dataSourceFromServer.getName() + "...");

            baseDataSource = instantiateDataSource(dataSourceFromServer);

            initializeDataSource(baseDataSource, dataSourceFromServer);

            // Initialization of the data source is successfull. Let's make sure the data source is not marked down
            markedDown.remove(dataSourceFromServer.getName());
        }
        catch (FrameworkException e) {
            markedDown.put(dataSourceFromServer.getName(), dataSourceFromServer);

            // Report the error to the service listener and continue...
            reportError(e.getMessage(), e);

            return;
        }

        // At this point DataSource is instantiated, initialized and started. The only things left are:
        // - tell server about it
        // - schedule datasource to run
        // - add datasource to collection ("datasources") to keep track of active datasources.

        if (!updateServerWithDataSourceStatus(baseDataSource.getDataSourceID(), baseDataSource.getDataSourceName(), true)) {
            baseDataSource.stop();
            return;
        }

        try {
            // Schedule the data source for polling interval (it will be scheduled to run immediately)
            scheduleDataSource(baseDataSource);
        }
        catch (FrameworkException e) {
            // Report the error to the service listener and continue...
            reportError("Failed to schedule data source " + baseDataSource.getDataSourceName() + ": " + e.getMessage(), e);

            markedDown.put(dataSourceFromServer.getName(), dataSourceFromServer);

            baseDataSource.stop();
            return;
        }

        // Mark the datasource as started successfully
        datasources.put(baseDataSource.getDataSourceName(), new DataSourceInfo(baseDataSource, dataSourceFromServer));

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Data source " + dataSourceFromServer.getName() + " started!");
    }

    @SuppressWarnings({"CatchGenericClass"})
    private void initializeDataSource(final BaseDataSource dataSource, final DataSource dataSourceInfo) throws FrameworkException {
        try {
            // Initiailize the data source!
            if (!dataSource.start())
                throw new FrameworkException("Data source " + dataSource.getDataSourceName() + " failed during start");

            if (logger.isDebugEnabled())
                logger.debug(dataSourceInfo.getName() + " - Data source implementation initialized!");
        }
        catch (RuntimeException e) {
            // Catch any runtime exception that the data source might throw during initialization...
            throw new FrameworkException("Data source " + dataSource.getDataSourceName() + " failed during start: " + e.getMessage(), e);
        }
    }

    private BaseDataSource instantiateDataSource(final DataSource dataSource) throws FrameworkException {
        // Instantiate a proper configuration object based on the given DS config
        final ConfigurationFactory factory = new ConfigurationFactory(dataSource.getName(), dataSource.getId());
        final DataSourceConfiguration configuration = factory.create(dataSource.getDsTypeId(), dataSource.getAttributesAsProperties());

        // Finally, instantiate the data source instance!
        return configuration.instantiateDataSource();
    }

    /**
     * Stops a running data source and removes it from the pool
     *
     * @param dataSourceName name of datasource to stop
     */
    protected synchronized void stopDataSource(final String dataSourceName) {
        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Stopping data source " + dataSourceName);

        // Remove DataSource from being marked down
        final DataSource markedDownDataSource = markedDown.remove(dataSourceName);
        if (markedDownDataSource != null) {
            // Data source is marked down! All we have to do is to update the server.
            updateServerWithDataSourceStatus(markedDownDataSource.getId(), dataSourceName, false);
            return;
        }

        // Find the data source
        final DataSourceInfo dataSourceInfo = datasources.remove(dataSourceName);
        if (dataSourceInfo == null) {
            logger.error(getServiceName() + " - Could not find data source " + dataSourceName + " to stop");
            return;
        }

        // Update server
        updateServerWithDataSourceStatus(dataSourceInfo.getDataSourceInfoFromServer().getId(), dataSourceName, false);

        // And stop the data source instance
        final BaseDataSource dataSource = dataSourceInfo.getDataSourceInstance();

        // Cancel all scheduled tasks for this data source
        scheduler.unschedule(dataSource);

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Unscheduled all events for data source " + dataSourceName);

        // Finally, stop the data source itself...
        dataSource.stop();

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Implementation stopped for data source " + dataSourceName);

        // We're done!
        logger.info(getServiceName() + " - Stopped data source " + dataSourceName);
    }

    private boolean updateServerWithDataSourceStatus(final long dataSourceId, final String dataSourceName, final boolean status) {
        try {
            // Update the status of the datasource in the GUI from 'starting' to 'running'
            communicationManager.get().getAcquisitionServerSide().updateIsActive(dataSourceId, status);
            return true;
        }
        catch (FrameworkException e) {
            reportError("Failed to update active status of data source " + dataSourceName, e);
            return false;
        }
        catch (AcquisitionServerException e) {
            reportError("Failed to update active status of data source " + dataSourceName, e);
            return false;
        }
    }

    /**
     * Updates an existing data source. This will update the data source's configuration and restart it if it is running.
     *
     * <p/>Note: A check is first made to see if the new data actually contains changes. Only if changes are detected, the
     * data source will be updated!
     *
     * @param dataSourceName    name of datasource to update
     * @param newDataSourceInfo
     * @return true if data source was actually updated, false if no changes were detected
     */
    @SuppressWarnings({"JavaDoc"})
    protected synchronized boolean updateDataSource(final String dataSourceName, final DataSource newDataSourceInfo) {
        try {
            // Find the data source
            final DataSourceInfo currentDataInfo = datasources.get(dataSourceName);
            if (currentDataInfo == null)
                throw new FrameworkException("Could not find data source to update (" + dataSourceName + ")");

            // Now compare the two configurations...
            if (!currentDataInfo.isChanged(newDataSourceInfo))
                return false;
        }
        catch (FrameworkException e) {
            reportError(e.getMessage(), e);
            return false;
        }

        logger.info(getServiceName() + " - Updating data source " + dataSourceName);

        // Stop the existing data source
        stopDataSource(dataSourceName);

        // And start a new one based on the new configuration
        startDataSource(newDataSourceInfo);

        // Immediately force the datasource (all packages) to poll (RFE 1355) after config change
        pollNow(dataSourceName);

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Restarted data source " + dataSourceName);

        return true;
    }

    private AcquisitionUnit getAcquisitionUnit() throws FrameworkException {
        return acquisitionUnitManager.get().getAcquisitionUnit();
    }

    /**
     * Schedule a data source to run. Depending on the capabilities of this data source, we need to schedule the right
     * events (foundation, performance, etc...)
     *
     * @param dataSource
     */
    @SuppressWarnings({"JavaDoc"})
    private void scheduleDataSource(final BaseDataSource dataSource) throws FrameworkException {
        try {
            for (final PackageConfiguration packageConfiguration : dataSource.getConfiguration().getPackageConfigurations().values()) {
                // Don't try to schedule packages that are not enabled!
                if (!packageConfiguration.isEnabled()) {
                    ServiceManager.getInstance().getCommunicationManager().getAcquisitionServerSide().stopReportingAuditForPackage(dataSource.getDataSourceID(), packageConfiguration.getAuditName());

                    continue;
                }

                // Don't try to schedule packages that are not polling!
                if (!(packageConfiguration instanceof PollingPackageConfiguration))
                    continue;

                if (logger.isDebugEnabled())
                    logger.debug(getServiceName() + " - Scheduling data source " + dataSource.getDataSourceName() + " for " + packageConfiguration.getPackageName() + " poll");

                final PollingPackageConfiguration pollingPackageConfiguration = (PollingPackageConfiguration) packageConfiguration;

                final int interval = pollingPackageConfiguration.getPollTimeMillis();
                if (interval < 0)
                    throw new FrameworkException("Invalid time interval (" + interval + ") for data source " + packageConfiguration.getDataSourceName() + " package " + packageConfiguration.getPackageName());

                ServiceManager.getInstance().getCommunicationManager().getAcquisitionServerSide().reportStandbyForPackage(dataSource.getDataSourceID(), packageConfiguration.getAuditName());

                final Calendar now = Calendar.getInstance();

                final Calendar nextMidnight = Calendar.getInstance();
                nextMidnight.set(now.get(Calendar.YEAR), now.get(Calendar.MONTH), now.get(Calendar.DAY_OF_MONTH), 0, 0, 0);
                nextMidnight.add(Calendar.HOUR, 24);

                final int startTimeMins = pollingPackageConfiguration.getPollStartTimeMins();

                final long initialDelay;
                if ("netappstorageperformance".equals(pollingPackageConfiguration.getPackageName())) {
                    // TA38128	Acquisition	US15248: OPM - Fixed tick netappstorageperformance polling
                    // OCIE performance gets special behavior with respect to initialDelay/SlowStart.
                    initialDelay = getOCIEPerformanceInitialDelay(pollingPackageConfiguration, interval, now);
                } else if (startTimeMins < 0) {
                    // Determine the initial delay for this data source. If we're supposed to 'slow start', pick a random time
                    // in the range of the interval to 'spread' the polls as much as possible.
                    // FR1526 - Perform lag too long after acq restart
                    // Some customer may set a very large poll interval that could schedule performance poll with a large delay
                    // Fix it to have a minimum number between the random initial deplay and one hour
                    initialDelay = slowStart ? (long) (Math.random() * Math.min(interval, HOUR_MILLIS)) : 0;
                } else {
                    // If datasource was configured to poll at preconfigured time of the day (NER foundation),
                    // we will calculate initialDelay between current time and next possible start.
                    final Calendar startTime = (Calendar) nextMidnight.clone();
                    startTime.add(Calendar.MINUTE, startTimeMins);

                    initialDelay = (startTime.getTimeInMillis() - now.getTimeInMillis()) % (24 * HOUR_MILLIS);
                }

                // Schedule Foundation with the right interval. The initial poll will happen immediately
                final PollingPackage aPackage = dataSource.getPackage(packageConfiguration.getPackageName());
                final PollLogic logicToRun = new PollLogic(this, aPackage, dataSource);

                scheduler.schedule(aPackage.getPackageName(), dataSource, logicToRun, initialDelay, interval);
            }
        }
        catch (AcquisitionServerException e) {
            // Report this error all the way to the service manager
            reportError("Error while scheduling data source " + dataSource.getDataSourceName(), e);
        }
    }

    /**
     * OCIE performance polls need to be synchronized with a specific wall clock minute.
     * This method gets the target Minute from the configuration, and calculates the
     * initial delay required so that performance polls align with the Minute
     * @param pollingPackageConfiguration the data source configuration indicating the target clock Minute
     * @param interval the performance poll period/interval
     * @param now the current time
     * @return the initial delay required so that performance polls align with the Minute
     */
    private long getOCIEPerformanceInitialDelay(final PollingPackageConfiguration pollingPackageConfiguration,
                                                final int interval, final Calendar now) {
        final Properties packageProperties = pollingPackageConfiguration.getProperties();
        final String clockPollEnabledVal = packageProperties.getProperty("clock.polling.enabled");
        final String clockPollOffsetVal = packageProperties.getProperty("clock.polling.offset");
        final boolean isClockPollEnabled = Boolean.parseBoolean(clockPollEnabledVal);
        final long initialDelay;
        if (isClockPollEnabled) {
            final int intervalInMinutes = interval / MINUTE_MILLIS;
            final int thisMinute = now.get(Calendar.MINUTE);

            Integer clockPollOffset = 0;
            try {
                clockPollOffset = Integer.parseInt(clockPollOffsetVal);
            } catch (NumberFormatException nfe) {
                reportError("Error while scheduling data source " + pollingPackageConfiguration.getDataSourceName()
                        + " bad poll delay value " + clockPollOffsetVal, nfe);
            }

            // If this minute is less than the poll offset, just wait for the poll offset.
            // Otherwise, keep adding the poll interval to the offset until we find a minute
            // in the future
            int nextPollMinute = clockPollOffset;
            while (thisMinute >= nextPollMinute) {
                nextPollMinute += intervalInMinutes;
            }

            final Calendar nextPollCalendar = (Calendar) now.clone();
            // Clear the minute, second & millisecond so we can set the target minute
            nextPollCalendar.set(Calendar.MILLISECOND, 0);
            nextPollCalendar.set(Calendar.SECOND, 0);
            nextPollCalendar.set(Calendar.MINUTE, 0);

            // Now set the target minute for the poll.
            // This should handle any carrying of minute to hour, hour to day, day to month...
            nextPollCalendar.add(Calendar.MINUTE, nextPollMinute);
            initialDelay = nextPollCalendar.getTimeInMillis() - now.getTimeInMillis();
        } else {
            initialDelay = slowStart ? (long) (Math.random() * Math.min(interval, HOUR_MILLIS)) : 0;
        }
        return initialDelay;
    }

    private ObjectName formatSchedulerObjectName() {
        return formatObjectName("Acquisition Framework", "DataSourceManagerScheduler");
    }

    /**
     * The logic executed when polling a data source
     */
    private static class PollLogic implements Runnable {
        private final WeakReference<DataSourceManager> dataSourceManager;
        private final PollingPackage packageToPoll;
        private final BaseDataSource dataSource;
        private final Properties event;

        public PollLogic(final DataSourceManager dataSourceManager, final PollingPackage packageToPoll, final BaseDataSource dataSource) {
          this(dataSourceManager, packageToPoll, dataSource, null );
        }

        public PollLogic(final DataSourceManager dataSourceManager, final PollingPackage packageToPoll, final BaseDataSource dataSource, final Properties event) {
            this.dataSourceManager = new WeakReference<DataSourceManager>(dataSourceManager);
            this.packageToPoll = packageToPoll;
            this.dataSource = dataSource;
          this.event = event;
        }

        public DataSourceManager getDataSourceManager() {
            return dataSourceManager.get();
        }

        /**
         * Runs a data source's single poll
         * <p/>
         * Note: This method HAS to catch ALL exceptions!!!
         */
        @SuppressWarnings({"CatchGenericClass"})
        public void run() {
            if (logger.isDebugEnabled())
                logger.debug(getDataSourceManager().getServiceName() + " - Polling with data source " + dataSource.getDataSourceName());

            synchronized (getDataSourceManager().acquiringDatasource) {
                // Check if the data source manager is stopping...
                if (getDataSourceManager().isStopping())
                    return;

                getDataSourceManager().acquiringDatasource.put(dataSource.getDataSourceName(), dataSource);
            }

            try {
                dataSource.run(packageToPoll, event);
            }
            catch (Throwable e) {
                logger.error(dataSource.getDataSourceName() + " - Uncaught exception while performing acquisition: " + e.getMessage(), e);

                // Report this error all the way to the service manager
                getDataSourceManager().reportError("Uncaught error while running data source " + dataSource.getDataSourceName(), e);
            }
            finally {
                getDataSourceManager().acquiringDatasource.remove(dataSource.getDataSourceName());
            }
        }
    }

    /**
     * An asynchronous scheduled task that runs and periodically sychronizes the state of data sources between server
     * and the acquisition framework.
     * <p/>
     * Note: This implementation should be replaced by better communication protocol between the acquisition and the
     * server. Instead of polling, the server should notify the acuisition about changes!
     */
    private static class Synchronizer implements Runnable {
        private final WeakReference<DataSourceManager> dataSourceManager;

        public Synchronizer(final DataSourceManager dataSourceManager) {
            this.dataSourceManager = new WeakReference<DataSourceManager>(dataSourceManager);
        }

        public DataSourceManager getDataSourceManager() {
            return dataSourceManager.get();
        }

        public void run() {

            // Synchronize without updates (these will be caught by the force acq listener)
            getDataSourceManager().synchronizeDataSourcesWithServer(false);
        }
    }

    public static class DataSourceInfo {
        private final BaseDataSource dataSourceInstance;
        private final DataSource dataSourceInfoFromServer;

        public DataSourceInfo(final BaseDataSource dataSourceInstance, final DataSource dataSourceInfoFromServer) {
            this.dataSourceInstance = dataSourceInstance;
            this.dataSourceInfoFromServer = dataSourceInfoFromServer;
        }

        public BaseDataSource getDataSourceInstance() {
            return dataSourceInstance;
        }

        public DataSource getDataSourceInfoFromServer() {
            return dataSourceInfoFromServer;
        }

        public boolean isChanged(final DataSource that) {
            return isChanged(dataSourceInfoFromServer, that);
        }

        public static boolean isChanged(final DataSource one, final DataSource another) {
            return (one.getId() != another.getId()) ||
                    !one.getName().equals(another.getName()) ||
                    !one.getAttributes().equals(another.getAttributes());
        }
    }
}
