package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.CustomManifestInfo;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.DataSourceMetaData;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.ref.WeakReference;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;

/**
 * This class is responsible for deployment of all the available data sources (a singleton)
 *
 * Deployment process:
 * 1. Figure out the deploy directory (from the codebase attribute in the registry).
 * 2. Read the deployfile.dat file to get a list of deployee.xml files.
 * 3. Download and parse each file and create a {@link com.onaro.sanscreen.acquisition.framework.datasource.metadata.DataSourceMetaData} instance
 * 4. Keep the deployed files in a list.
 */
public class DeployManager extends FrameworkService implements DeployManagerMBean {
    private static final Logger logger = LogManager.getLogger(DeployManager.class);

    private static final String ROOT_LOADER_PROPS = "rootloader.properties";

    private final WeakReference<CommunicationManager> communicationManager;

    private Map<Long, DataSourceMetaData> deployedDataSources = new TreeMap<Long, DataSourceMetaData>();

    private URLClassLoader rootLoader;
    private List<String> rootJars;

    DeployManager(final ServiceStatusListener statusListener, final CommunicationManager communicationManager) {
        super("Deploy Manager", statusListener);

        this.communicationManager = new WeakReference<CommunicationManager>(communicationManager);
    }

    /**
     * @return All the deployed data sources
     */
    public Map<Long, DataSourceMetaData> getDeployedDataSources() {
        return deployedDataSources;
    }

    /**
     * For monitoring purposes, copy the descriptions of all the deployed data sources in the deploy manager
     */
    public String[] getDeployedDataSourceDescriptions() {
        final String[] descriptions = new String[deployedDataSources.size()];

        int i = 0;
        for (DataSourceMetaData metaData : deployedDataSources.values()) {
            descriptions[i] = metaData.toString();
            i++;
        }

        return descriptions;
    }

    public DataSourceMetaData lookup(final long datasourceType) {
        return deployedDataSources.get(datasourceType);
    }

    protected void doStart() throws FrameworkException {

        deployedDataSources.clear();

        // force the SAX parser to be xerces
        System.setProperty("javax.xml.parsers.SAXParserFactory", "org.apache.xerces.jaxp.SAXParserFactoryImpl");

        // Figure out the URL to the deployfile.dat in the datasources.war directory
        final String codebase = System.getProperty("codebase");
        if (codebase == null)
            throw new FrameworkException("'codebase' system attribute is missing from configuration");

        final String datasourceDeployDir = codebase + "/datasources";

        //create the single root loader of the system and packages.
        createRootLoader(datasourceDeployDir);

        BufferedReader deployFileReader = null;
        try {

            final String activeDeployFile = "deployfile.dat";
            deployFileReader = new BufferedReader(new InputStreamReader(openFileInDeployDir(activeDeployFile), "utf-8"));

            final DeployParser dp = new DeployParser();

            // Downloads the server's version of the deployfile.dat file (in the datasources.war directory)
            final List<String> deployFiles = dp.parseDeployFile(deployFileReader);

            // Now deploy each data source!
            for (final String line : deployFiles) {
                // The line is basically a name of a file that is located in the datasources deploy directory (the same
                // directory as the deployfile.dat file we're reading now
                final URL datasourceDeployFile = new URL(datasourceDeployDir + "/" + line);
                deployDataSource(datasourceDeployFile);
            }
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to load and parse deployfile.dat at " + datasourceDeployDir, e);
        }
        finally {
            try {
                if (deployFileReader != null)
                    deployFileReader.close();
            }
            catch (IOException e) {
                //can not do anything here.
            }
        }
    }

    protected void doStop() {
        // Nothing special to do...
    }

    /**
     * Sets up the root class loader for this AU.  This becomes the parent for all package level loaders.  It loads
     * various system level jars such as all base package jars and certain common utility libraries.
     *
     * @param datasourceDeployDir the deployment path on the server.
     *
     * @throws FrameworkException if there are formatting issues with jar URLs or with loading the root loader properties.
     */
    private void createRootLoader(final String datasourceDeployDir) throws FrameworkException {

        this.rootJars = parseRootJarsFromFile();
        final URL[] urls = new URL[rootJars.size()];

        try {
            int j = 0;
            for (final String jarName : rootJars) {
                //TODO- check URL constructor
                urls[j] = new URL(datasourceDeployDir + "/" + jarName);
                final String version = new CustomManifestInfo(urls[j], jarName).getVersionString();
                logger.info("Root jar loaded: " + version);
                j++;
            }

            //todo - use this instead of service manager to get parent loader
            rootLoader = new URLClassLoader(urls, ServiceManager.getInstance().getClass().getClassLoader());


        }
        catch (final Exception e) {
            throw new FrameworkException("Failed to load root JARs!", e);
        }

    }

    /**
     * Parse root jar files from root loader properties file (created during ant deploy of full acquisition).
     *
     * @return the list of jar files to be deployed from the server into the root loader.
     *
     * @throws FrameworkException for problems parsing the root loader file.
     */
    private List<String> parseRootJarsFromFile() throws FrameworkException {

        BufferedReader rootLoaderPropsReader = null;
        try {
            rootLoaderPropsReader = new BufferedReader(new InputStreamReader(openFileInDeployDir("rootloader.properties"), "utf-8"));
            final Properties props = new Properties();
            props.load(rootLoaderPropsReader);

            final String jarFilesStr = props.getProperty("rootloader.jars");
            if (jarFilesStr == null) {
                throw new FrameworkException("Could not load root loader property file! No property defined.");
            }

            final String[] jarFiles = jarFilesStr.split(";");
            return Arrays.asList(jarFiles);
        }
        catch (final IOException ioe) {
            throw new FrameworkException("Could not load root loader property file!", ioe);
        }
        finally {
            if (rootLoaderPropsReader != null) {
                try {
                    rootLoaderPropsReader.close();
                }
                catch (final Exception e) {
                    //fall through;
                }
            }
        }

    }

    /**
     * This is the root loader of the system, obtained from deploy manager.   It becomes the parent loader for all package
     * level loaders.   It holds the base package jars and certain system wide libraries.   There is one root loader
     * per AU.
     *
     * @return the root loader.
     */
    public ClassLoader getRootLoader() {
        return rootLoader;
    }

    /**
     * Is the given jar file loaded in the root class loader?
     *
     * @param jarName name of the jar.
     *
     * @return true if loaded in root loader.
     */
    public boolean isRootLoaded(final String jarName) {
        return rootJars.contains(jarName);
    }

    /**
     * Based on the given URL, read the deploy file and create a {@link DataSourceMetaData} object to represent this
     * data source type
     *
     * @param dataSourceDeployeeFile URL of data source deployee file to deploy
     *
     * @throws FrameworkException On any error
     */

    public void deployDataSource(final URL dataSourceDeployeeFile) throws FrameworkException {
        try {
            final DataSourceMetaData metaData = DataSourceMetaData.createMetaData(dataSourceDeployeeFile);

            // If this type was already deployed, this is BAD!!!!
            if (deployedDataSources.containsKey(metaData.getDataSourceType()))
                throw new FrameworkException("Duplicate deploy file definition for data source type " + metaData.getDataSourceType() + " found when loading deploy file " + dataSourceDeployeeFile + " existing datasource is " + deployedDataSources.get(metaData.getDataSourceType()).getName());

            deployedDataSources.put(metaData.getDataSourceType(), metaData);

            logger.info(getServiceName() + " - Deployed data source type " + metaData.getName());
        }
        catch (IOException e) {
            logger.error(getServiceName() + " - Failed to deploy data source from deploy file " + dataSourceDeployeeFile, e);
            throw new FrameworkException(e.getMessage(), e.getCause());
        }
    }

    /**
     * Opens (remotely) a file located in the datasources.war directory on the SANscreen server. This
     * provides services for accessing files from the central deploy location - such as the 'deployfile.dat' or 'diskmodels.txt'.
     *
     * @param fileName Name of a file in the datasources.war directory on JBoss
     *
     * @return A input stream to the file specified
     *
     * @throws FrameworkException On any error
     */
    public InputStream openFileInDeployDir(final String fileName) throws FrameworkException {
        try {
            // Figure out the URL to the deployfile.dat in the datasources.war directory
            final String codebase = System.getProperty("codebase");
            if (codebase == null)
                throw new FrameworkException("'codebase' system attribute is missing from configuration");

            return communicationManager.get().openHttpConnection(new URL(codebase + "/datasources/" + fileName)).getInputStream();
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to load " + fileName, e);
        }
    }
}
