package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import org.apache.log4j.Logger;

import javax.management.*;
import java.util.List;
import java.util.LinkedList;
import java.util.concurrent.ScheduledFuture;
import java.lang.management.ManagementFactory;

/**
 * A base class defining the basic functionality for a framework service.
 *
 * <p/>This class specifically handles the start / stop process of a service. It's always an issue how to manage a service
 * in a reliable way. Especially when stopping it, we need to signal the service to stop and then wait until it is
 * actually stopped (all threads are stopped, etc...).
 *
 * <p/>The class also defines how the service can report errors back to the service manager ({@link ServiceStatusListener}).
 */
public abstract class FrameworkService implements FrameworkServiceMBean {
    private static final Logger logger = Logger.getLogger(FrameworkService.class);

    private final List<ServiceStatusListener> statusListeners = new LinkedList<ServiceStatusListener>();
    private final String serviceName;

    private volatile boolean initialized = false;
    private volatile boolean stopping = false;
    private volatile boolean stopped = false;

    /**
     * Constructor
     *
     * @param name The name of the service
     * @param statusListener The default status listener for this service
     */
    protected FrameworkService(final String name, final ServiceStatusListener statusListener) {
        this.serviceName = name;

        if (statusListener != null)
            statusListeners.add(statusListener);
    }

    public final String getServiceName() {
        return serviceName;
    }

    /**
     * Register an additional status listener (in addition to the default one) for this service
     *
     * @param statusListener Listener to add
     */
    public final void registerStatusListener(final ServiceStatusListener statusListener) {
        statusListeners.add(statusListener);
    }

    /**
     * Start the service. Overriding classes should use {@link #doStart} to add specific initialization code.
     * After this method returns, the {@link #isInitialized} method will return true.
     *
     * @throws FrameworkException On any problem
     */
    public final void start() throws FrameworkException {
        if (initialized) {
            logger.error(getServiceName() + " - Already started!");
            return;
        }

        stopping = false;
        stopped = false;

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Initializing...");

        doStart();

        registerMBean(this, formatServiceObjectName());

        logger.info(getServiceName() + " - Initialized!");
        initialized = true;
    }

    /**
     * Stop the service. This method will return after the service was verified as completely stopped.
     * Overriding classes should use {@link #doStop} to add specific stopping code.
     *
     * After this method returns, the {@link #isStopped} will return true
     */
    public final void stop() {
        // Prevent unnecessary re-entry
        if (isStopping())
            return;

        if (logger.isDebugEnabled())
            logger.debug(getServiceName() + " - Stopping...");

        unregisterMBean(formatServiceObjectName());

        // Signal the stop flag
        stopping = true;
        initialized = false;

        // Call additional stop actions
        doStop();

        stopped = true;

        logger.info(getServiceName() + " - Stopped!");
    }

    private ObjectName formatServiceObjectName() {
        return formatObjectName("Acquisition Framework", getServiceName());
    }

    /**
     * Report an error to the status listener.
     *
     * @param msg Message explaining the problem
     * @param cause Exception causing this error (null if no exception is available)
     */
    public final void reportError(final String msg, final Throwable cause) {
        for (final ServiceStatusListener serviceStatusListener : statusListeners)
            serviceStatusListener.error(getServiceName(), msg, cause);
    }

    /**
     * Report a fatal error (acquisition will shutdown) to the status listener
     *
     * @param msg Message explaining the problem
     * @param cause Exception causing this error (null if no exception is available)
     */
    public final void reportFatal(final String msg, final Throwable cause) {
        for (final ServiceStatusListener serviceStatusListener : statusListeners)
            serviceStatusListener.fatal(getServiceName(), msg, cause);
    }

    /**
     * @return true if the service started successfully
     */
    public final boolean isInitialized() {
        return initialized;
    }

    /**
     * @return true if the service is currently stopping
     */
    public final boolean isStopping() {
        return stopping;
    }

    /**
     * @return true if the service has stopped completely
     */
    public final boolean isStopped() {
        return stopped;
    }

    public String getStatus() {
        if (stopped)
            return "Stopped";

        if (stopping)
            return "Stopping";

        if (initialized)
            return "Initialized";

        return "Unknown";
    }

    protected ObjectName formatObjectName(final String domain, final String name) {
        try {
            return new ObjectName(domain, "name", name);
        }
        catch (MalformedObjectNameException e) {
            // This should never happen...
            throw new Error(e.getMessage(), e);
        }
    }

    protected void registerMBean(final Object implementation, final ObjectName objectName) {
        // No MBeans in standalone mode
        if (ServiceManager.getInstance().isStandalone())
            return;
        
        // Unregister any left overs for this MBean (this is usefull only when testing)
        if (ManagementFactory.getPlatformMBeanServer().isRegistered(objectName)) {
            try {
                ManagementFactory.getPlatformMBeanServer().unregisterMBean(objectName);
            }
            catch (Exception e) {
                if (logger.isDebugEnabled())
                    logger.debug(getServiceName() + " - Failed to unregister MBean " + objectName);

                // Fall through... ignore this error...
            }
        }

        try {
            ManagementFactory.getPlatformMBeanServer().registerMBean(implementation, objectName);

            if (logger.isDebugEnabled())
                logger.debug(getServiceName() + " - Registered MBean as " + getServiceName());
        }
        catch (Exception e) {
            reportError("Failed to register MBean for " + getServiceName(), e);
        }
    }

    protected void unregisterMBean(final ObjectName objectName) {
        // No MBeans in standalone mode
        if (ServiceManager.getInstance().isStandalone())
            return;        
        try {
            ManagementFactory.getPlatformMBeanServer().unregisterMBean(objectName);

            if (logger.isDebugEnabled())
                logger.debug(getServiceName() + " - Unregistered MBean " + getServiceName());
        }
        catch (Exception e) {
            reportError("Failed to unregister MBean for " + getServiceName(), e);
        }
    }

    protected String getTaskStatus(final ScheduledFuture<?> task) {
        if (task == null)
            return "Not available";

        if (task.isCancelled())
            return "Canceled";

        if (task.isDone())
            return "Done";

        return "ok";
    }

    /**
     * Perform service specific initializations. This method returns only when service is fully initialized!
     * 
     * @throws FrameworkException On any problem
     */
    protected abstract void doStart() throws FrameworkException;

    /**
     * Perform service specific logic required to stop the service.
     * This method must return only when service is completely stopped!
     *
     * NOTE: Do not synchronize these method!
     */
    protected abstract void doStop();
}
