package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import org.apache.log4j.Logger;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;

/**
 * This class should be treated as singleton and only used via CommunicationManager
 * because it sets up and works with the only certificate store file used in acquisition
 */
class HttpBasedUrlConnection {
    private static final Logger logger = Logger.getLogger(HttpBasedUrlConnection.class);

    private SSLSocketFactory sslSocketFactory = null;
    private String trustStorePath;

    /**
     * Iniialize SSL communication including downloading of certificates from the server
     *
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *          When having certificate issues
     */
    public void initSSL() throws FrameworkException {
        if (logger.isDebugEnabled()) {
            logger.debug("HttpBasedUrlConnection - Initializing SSL");
        }

        // Configure truststore location for everybody
        final String acqHome = System.getProperty("acq.home");
        if (acqHome == null) {
            throw new FrameworkException("acq.home system property not set!");
        }

        synchronized (this) {
            trustStorePath = acqHome + "/conf/cert/client.keystore";

            System.setProperty("javax.net.ssl.trustStore", trustStorePath);
        }

        // HEADSUP: This violates HTTPS host verification, since we arnt setting up SSL certs right.
        final HostnameVerifier hv = new HostnameVerifier() {
            public boolean verify(final String urlHostName, final SSLSession session) {
                return true;
            }
        };

        HttpsURLConnection.setDefaultHostnameVerifier(hv);

        // Prepare SSL socket factory with all available certificates
        loadCertificates();
    }

    private void loadCertificates() throws FrameworkException {
        final File trustStore = new File(trustStorePath);
        if (!trustStore.exists() || !trustStore.isFile()) {
            final String defaultKeyStorePath = System.getProperties().getProperty("java.home") + File.separator + "lib" + File.separator + "security" + File.separator + "cacerts";

            if (logger.isDebugEnabled()) {
                logger.debug("HttpBasedUrlConnection - Copying default truststore from " + defaultKeyStorePath);
            }

            try {
                //noinspection ResultOfMethodCallIgnored
                trustStore.getParentFile().mkdirs();

                // Simply copy the file
                final InputStream is = new BufferedInputStream(new FileInputStream(new File(defaultKeyStorePath)));
                final OutputStream os = new BufferedOutputStream(new FileOutputStream(trustStore));
                final byte[] buffer = new byte[100000];

                try {
                    while (true) {
                        final int l = is.read(buffer);
                        if (l == -1) {
                            break;
                        }

                        os.write(buffer, 0, l);
                    }
                } finally {
                    //noinspection EmptyCatchBlock
                    try {
                        os.close();
                    } catch (IOException e) {
                    }
                    //noinspection EmptyCatchBlock
                    try {
                        is.close();
                    } catch (IOException e) {
                    }
                }
            } catch (IOException e) {
                throw new FrameworkException("Failed to copy default key store", e);
            }
        }

        final CertificateDownloader downloader = new CertificateDownloader(trustStorePath);
        synchronized (this) {
            sslSocketFactory = downloader.loadCertificates();
        }

        if (logger.isDebugEnabled()) {
            logger.debug("HttpBasedUrlConnection - Loaded certificates!");
        }
    }

    /**
     * Downloads the SSL certificate from the given URL and updates the local TrustStore.
     * This will also modify the default SSL socket factory to reflect the newly downloaded certificate
     *
     * @param url     URL of server to download certificates from
     * @param timeout In millis
     * @throws FrameworkException When failed to download
     */
    public synchronized void downloadCertificate(final URL url, final int timeout, final boolean override) throws FrameworkException {
        downloadCertificate(url.getHost(), url.getPort(), timeout, override);
    }

    public synchronized void downloadCertificate(final String host, final int port, final int timeout, final boolean override) throws FrameworkException {
        if (logger.isDebugEnabled()) {
            logger.debug("HttpBasedUrlConnection - Downloading certificate from host - " + host + ":" + port);
        }

        final CertificateDownloader downloader = new CertificateDownloader(trustStorePath);
        downloader.downloadIfNeeded(host, port, timeout, override);

        sslSocketFactory = downloader.loadCertificates();

        // Modify the default SSL socket factory to reflect the downloaded certificate!
        HttpsURLConnection.setDefaultSSLSocketFactory(sslSocketFactory);
    }

    /**
     * Returnes an initialized {@link java.net.URLConnection} setup to work.
     * <p/>
     * This includes setting up the correct truststore (for HTTPS certificates) if neccessary.
     *
     * @param url     URL to open
     * @param timeout in millis
     * @return An initialized connection to this URL
     * @throws com.onaro.sanscreen.acquisition.framework.FrameworkException
     *
     */
    public HttpURLConnection openHttpConnection(final URL url, final int timeout) throws IOException, FrameworkException {
        if (logger.isDebugEnabled()) {
            logger.debug("HttpBasedUrlConnection - setting up http connection to url : " + url.toString());
        }

        final URLConnection urlConnection = url.openConnection();
        if (!(urlConnection instanceof HttpURLConnection)) {
            throw new FrameworkException("Not http based url : " + url.toString());
        }

        if (urlConnection instanceof HttpsURLConnection) {
            if (logger.isDebugEnabled()) {
                logger.debug("HttpBasedUrlConnection - enabling SSL for url : " + url.toString());
            }

            // Make sure we download the certificate from the server if none is installed (and update the SSL factory)
            downloadCertificate(url, timeout, false);

            // enable SSL on this connection
            synchronized (this) {
                ((HttpsURLConnection) urlConnection).setSSLSocketFactory(sslSocketFactory);
            }
        }

        return (HttpURLConnection) urlConnection;
    }

    public synchronized SSLSocketFactory getSslSocketFactory() {
        return sslSocketFactory;
    }
}
