package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.mgmt.log.SampleAppenderDistributor;
import org.apache.commons.io.FileUtils;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.LogManager;
import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * This service keeps the acquisition log files directory in control. It
 * periodically runs a scan on the log directory and deletes old files to keep
 * the directory size in control.
 */
public class LogsManager extends FrameworkService implements LogsManagerMBean {
    private final static Logger logger = Logger.getLogger(LogsManager.class);

    private static final String RECORDING_TASK = "Recording Cleaner";

    /**
     * Milliseconds in one minute (millis/sec * sec/min)
     */
    private static final long ONE_MINUTE_MILLIS = 1000L * 60L;
  private final ServiceManager serviceManager;

  private File acqLogDir;

    private File acqRecordingDir;

    private SampleAppenderDistributor sampleAppenderDistributor = new SampleAppenderDistributor();

    LogsManager(final ServiceManager serviceManager) {
        super("Logs Manager", serviceManager);
        this.serviceManager = serviceManager;
    }

    @Override
    protected void doStart() throws FrameworkException {
        final String acqHome = System.getProperty("acq.home");
        final long frequency = Long.parseLong(ServiceManager.getInstance().getProperty("FREQUENCY_RECORDING_DIRECTORY_CLEANUP_JOB", "60").trim());
        
        if (acqHome == null) {
            throw new FrameworkException("Failed to resolve acquisition log directory. 'acq.home' is not set!");
        }

        if (acqHome.startsWith("/opt/netapp"))
        {
            acqLogDir = new File("/var/log/ocie/");
            acqRecordingDir = new File(acqLogDir, "recording");
        } else {
            acqLogDir = new File(acqHome, "log");
            acqRecordingDir = acqLogDir;
        }

        //Configure the recording logs if define in the general.properties
        if( serviceManager.getProperty( "RECORDING_LOGS_HOME" ) != null )
          acqRecordingDir = new File( serviceManager.getProperty( "RECORDING_LOGS_HOME" ) );

        if (!acqLogDir.exists()) {
            throw new FrameworkException("Failed to resolve acquisition log directory. " + acqLogDir.getAbsolutePath() + " does not exist");
        }

        if (!acqLogDir.isDirectory()) {
            throw new FrameworkException("Failed to resolve acquisition log directory. " + acqLogDir.getAbsolutePath() + " is not a directory");
        }

        if (!acqRecordingDir.exists()) {
            throw new FrameworkException("Failed to resolve acquisition log directory. " + acqRecordingDir.getAbsolutePath() + " does not exist");
        }

        if (!acqRecordingDir.isDirectory()) {
            throw new FrameworkException("Failed to resolve acquisition log directory. " + acqRecordingDir.getAbsolutePath() + " is not a directory");
        }

        // Schedule a task that periodically checks the recording directory
        ServiceManager.getInstance().getAsynchronousTaskManager().schedule(RECORDING_TASK, new RecordingCleanerTask(getServiceName(), acqRecordingDir), frequency * ONE_MINUTE_MILLIS);

        LogManager.getRootLogger().addAppender(sampleAppenderDistributor);
    }

    @Override
    protected void doStop() {
        ServiceManager.getInstance().getAsynchronousTaskManager().unschedule(RECORDING_TASK);

        LogManager.getRootLogger().removeAppender(sampleAppenderDistributor);
    }

    @Override
    public void setDebugEnabled(boolean debug) {
        Logger.getRootLogger().setLevel(debug ? Level.DEBUG : Level.INFO);
    }

    @Override
    public boolean isDebugEnabled() {
        return Logger.getRootLogger().isDebugEnabled();
    }

    public File getLogDirectory() {
        return acqLogDir;
    }

    @Override
    public String getLogDirectoryPath() {
        return acqLogDir.getAbsolutePath();
    }

    public File getAcqRecordingDir() {
        return acqRecordingDir;
    }

    @Override
    public boolean cleanLogsNow() {
        return false;
    }

    /**
     * Send a ZIP file to the server to be later send via Phone Home. If this is the local acuqisition unit, the file will
     * not be sent (it will be later collected by the server using a normal file system copy).
     *
     * @param acquisitionUnitReporting Identifies the acq unit sending this ZIP file
     * @param zipFile ZIP file (recording) to send to server
     *
     * @throws IOException on any problem sending file
     */
    public void sendZipToServer(final String acquisitionUnitReporting, final File zipFile) throws IOException {
        if (acquisitionUnitReporting.equalsIgnoreCase("local")) {
            return;
        }

//        final String ip = InetAddress.getLocalHost().getHostAddress().replace('.', '_');
//
//        // TODO: Fix remote appender to work with streams!!!
//        FileInputStream fis = null;
//        try {
//            byte[] data = new byte[(int) zipFile.length()];
//
//            fis = new FileInputStream(zipFile);
//            fis.read(data);
//
//            final RemoteAppender appender = new RemoteAppender();
//            appender.send(ip + "_" + zipFile.getName(), data);
//
//            if (logger.isDebugEnabled())
//                logger.debug(getServiceName() + " - Sent ZIP file " + zipFile.getName());
//        }
//        catch (NamingException e) {
//            throw new IOException(e);
//        } finally {
//            if (fis != null)
//                fis.close();
//        }
    }
    
    public final void addSampleAppender(final String threadName, final Writer out) {
        sampleAppenderDistributor.addSampleAppender(threadName, out);
    }

    public final void removeSampleAppender(final String threadName) {
        sampleAppenderDistributor.removeSampleAppender(threadName);
    }

    /**
     * Enforces an on-disk quota for the recording directory.  If the quota
     * is exceeded, files are deleted in increasing order based upon 1) last
     * modified timestamp, and 2) native OS sorting (e.g., lexicographic).
     */
    static class RecordingCleanerTask implements Runnable {

        private final String serviceName;

        private final File logDir;

        public RecordingCleanerTask(String serviceName, File logDir) {
            this.serviceName = serviceName;
            this.logDir = logDir;
        }

        @Override
        public void run() {
            final long quota = Long.parseLong(ServiceManager.getInstance().getProperty("RECORDING_QUOTA_IN_GB", "8"));
            // bitwise multiplication of 1024^3 to convert gigabytes-->bytes
            final long quotaBytes = quota << 30L;

            long dirSize = FileUtils.sizeOfDirectory(logDir);
            if (dirSize < quotaBytes) {
                // under quota; nothing to do here
                return;
            }

            Collection<File> files = FileUtils.listFiles(logDir, null, true);

            SortedSet<File> sortedFiles = new TreeSet<File>(new FileLastModifiedComparator());
            sortedFiles.addAll(files);

            Iterator<File> itr = sortedFiles.iterator();
            try {
                while (itr.hasNext() && (dirSize >= quotaBytes)) {
                    File f = itr.next();
                    logger.info("[quota=" + quotaBytes + ", used=" + dirSize
                            + "] deleting " + f.getCanonicalPath());
                    dirSize -= f.length();
                    f.delete();
                }
            } catch (Exception e) {
                logger.warn(serviceName, e);
            }
        }
    }

    /**
     * Compares {@link File}s to sort them in 1) ascending order since last
     * modification date, and 2) native OS sort (e.g., lexicographic).  For
     * example:
     * <pre>
     *   unsorted:  foo (2014/02), baz (2014/01), bar (2014/02)
     *     sorted:  baz (2014/01), bar (2014/02), foo (2014/02)
     * </pre>
     */
    private static class FileLastModifiedComparator implements Comparator<File> {

        @Override
        public int compare(File o1, File o2) {
            if (o1.lastModified() == o2.lastModified()) {
                // compare paths to avoid a telescoping effect where files of
                // the same timestamp would collapse to just the first-seen
                return o1.compareTo(o2);
            }
            return (o1.lastModified() < o2.lastModified() ? -1 : 1);
        }
    }

}
