package com.onaro.sanscreen.acquisition.framework.mgmt;

import com.netapp.oci.profiler.ProfilerService;
import com.netapp.oci.profiler.ProfilerServiceFactory;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.Main;
import com.onaro.sanscreen.acquisition.framework.standalone.Standalone;

import java.io.File;
import java.io.IOException;
import java.security.Permission;
import java.util.*;
import java.util.concurrent.ExecutorService;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The ServiceManager is the core of the acquisition framework. This singleton
 * class acts as the single point of access
 * to all the framework services.
 *
 * <p/>
 * When starting acquisition, the framework's main method {@link Main#main}
 * calls the
 * {@link com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager#start()}
 * upon startup to initialize all the services and the ServiceManager makes sure
 * they are initialized in the right order.
 *
 * <p/>
 * When stopping acquisition, the
 * {@link com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager#stop()}
 * method will be called to stop all the services gracefully (in the right
 * order).
 *
 * <p/>
 * Even though each package is designed to provide all the necessary facilities
 * for a data source to communicate with the outside world, the ServiceManager
 * is available (using the {@link #getInstance()} method at any given time. It
 * provides access to services like the {@link CommunicationManager}, etc...
 *
 * <p/>
 * When running unit tests, the ServiceManager should be initialized with a
 * config object with initCommLayerOnly set to indicate to the ServiceManager
 * that it should not try to perform a full startup, including deployment of
 * datasources, but rather initialize the minimum set of services required to
 * run data sources without communicating this information back to the server.
 *
 * <p/>
 * When starting up, the ServiceManager changes the SecurityManager for the JVM
 * to control the permissions of data sources running in the JVM. This prevents
 * data sources from trying to shutdown the JVM, etc... Refer to
 * {@link #shutdownExecutor} when trying to shut down an executor service
 * created by the data source.
 */
public final class ServiceManager extends FrameworkService
        implements ServiceStatusListener, ServiceManagerMBean
{

    private static final Logger LOG = LoggerFactory.getLogger(ServiceManager.class);

    private static final ServiceManager INSTANCE = new ServiceManager();

    private final List<FrameworkService> globalServices = new LinkedList<FrameworkService>();
    private final List<FrameworkService> acqServices = new LinkedList<FrameworkService>();

    // General resource management services
    private AsynchronousTaskManager asynchronousTaskManager;
    private LogsManager logsManager;

    // Acquisition specific services
    private CommunicationManager communicationManager;
    private AcquisitionUnitManager acquisitionUnitManager;
    private DeployManager deployManager;
    private DataSourceManager dataSourceManager;
    private DataSourceEventsManager dataSourceEventsManager;
    private AcqProfilerService acqProfilerService;

    // General properties
    private Properties generalProperties = new Properties();

    private boolean initCommLayerOnly = false;
    private boolean initForUtilOnly = false;        // For Acq CLI
    private boolean standalone = false;

    private String username;
    private String password;

    public static ServiceManager getInstance() {
        return INSTANCE;
    }

    private ServiceManager() {
        super("Service Manager", null);

        // User default credentials
        this.username = "acquisition";
        this.password = "moshekipod";
    }

    public AsynchronousTaskManager getAsynchronousTaskManager() {
        return asynchronousTaskManager;
    }

    public LogsManager getLogsManager() {
        return logsManager;
    }

    public CommunicationManager getCommunicationManager() {
        return communicationManager;
    }

    public AcquisitionUnitManager getAcquisitionUnitManager() {
        return acquisitionUnitManager;
    }

    public DeployManager getDeployManager() {
        return deployManager;
    }

    public DataSourceManager getDataSourceManager() {
        return dataSourceManager;
    }

    public DataSourceEventsManager getDataSourceEventsManager() {
        return dataSourceEventsManager;
    }

    public AcqProfilerService getAcqProfilerService() {
        return acqProfilerService;
    }

    public boolean isStandalone() {
        return standalone;
    }

    /**
     * Restart the service manager. Performed asynchronously to calling thread.    Errors on restart are logged in
     * acquisition log.
     */
    @Override
    public void restart() {

        LOG.info(getServiceName() + " - Restarting acquisition services...");

        final Thread stopper = new Thread() {
            @Override
            public void run() {
                ServiceManager.getInstance().stop();
                try {
                    sleep(2500);
                }
                catch (final InterruptedException ie) {
                    //continue;
                }
                try {
                    ServiceManager.getInstance().start();
                }
                catch (final FrameworkException e) {
                    LOG.error("Failed restarting Acquisition services!", e);
                }
            }
        };
        stopper.start();
    }

    /**
     * Initialize the list of services in the framework. This is done manually in the <b>RIGHT</b> order
     *
     *  <p/>Note: The services currently to not provide an init level. This is why we manually set the order of initialization.
     *
     * {@link #initialize} should be called prior to start() and doStart() to configure the ServiceManager with
     * credentials, or for standalone, etc. Default configuration settings will be used if it is not called.
     *
     * @throws FrameworkException
     */
    @Override
    protected void doStart() throws FrameworkException {
        try {
            if (!isStandalone()) {
                // Install a new security manager for the acquisition...
                System.setSecurityManager(new NoExitSecurityManager());
            }

            // Fix required system properties to the default values (if they are not already set)
            if (System.getProperty("acq.home") == null) {
                System.setProperty("acq.home", "C:/Program Files/SANscreen/acq");
            }

            if (System.getProperty("codebase") == null) {
                System.setProperty("codebase", "https://localhost");
            }

            generalProperties = loadGeneralProperties();

            globalServices.clear();
            acqServices.clear();

            initializeGlobalServices();

            initializeAcquisitionServices();

            // Now start all the services...
            for (final FrameworkService globalService : globalServices) {
                globalService.start();
            }

            for (final FrameworkService acqService : acqServices) {
                acqService.start();
            }
        }
        catch (Throwable e) {
            // Remove the security manager (this is important especially when running in unit tests, where
            // the JUnit framework will try to exit the JVM
            System.setSecurityManager(null);

            // And propagate the error
            throw new FrameworkException("Failed to start acquisition - " + e.getMessage(), e);
        }
    }

    public String getProperty(final String key) {
        return generalProperties.getProperty(key);
    }

    public String getProperty(final String key, final String defaultValue) {
        return generalProperties.getProperty(key, defaultValue);
    }

    public File getAcqHomeDir() {
        return new File(System.getProperty("acq.home"));
    }

    public File getSanscreenHomeDir() {
        return new File(System.getProperty("onaro.home"));
    }

    private void initializeAcquisitionServices() {
        communicationManager = new CommunicationManager(this, standalone, username, password);

        acqServices.add(communicationManager);

        if (standalone) {
            return;
        }

        if (!initForUtilOnly) {
            deployManager = new DeployManager(this, communicationManager);
            acqServices.add(deployManager);

            acquisitionUnitManager = new AcquisitionUnitManager(this, communicationManager);
            acqServices.add(acquisitionUnitManager);

            startProfilerService();
        }

        if (!initCommLayerOnly) {
            dataSourceManager = new DataSourceManager(this, communicationManager, acquisitionUnitManager);
            acqServices.add(dataSourceManager);

            dataSourceEventsManager = new DataSourceEventsManager(this, communicationManager, acquisitionUnitManager, dataSourceManager);
            acqServices.add(dataSourceEventsManager);
        }
    }

    private void startProfilerService() {
        final ProfilerService profilerService = ProfilerServiceFactory.getInstance();
        if (profilerService.isEnabled()) {
            profilerService.start();
            Runtime.getRuntime().addShutdownHook(new Thread() {
                @Override
                public void run() {
                    profilerService.stop();
                }
            });
        }
    }

    private void initializeGlobalServices() {
        asynchronousTaskManager = new AsynchronousTaskManager(this);
        globalServices.add(asynchronousTaskManager);

        logsManager = new LogsManager(this);
        globalServices.add(logsManager);
    }

    /**
     * Signal the ServiceManager to start only the {@link CommunicationManager}. Used for Unit tests!
     *
     * @throws FrameworkException On any error
     */
    public void startOnlyCommunicationLayer() throws FrameworkException {
        if (isInitialized()) {
            throw new FrameworkException("Communication manager already initialized");
        }

        initCommLayerOnly = true;
    }

    /**
     * Signal the ServiceManager to start only the global services ({@link LogsManager} and {@link AsynchronousTaskManager}.
     * Used for standalone !
     *
     * @throws FrameworkException On any error
     */
    public void startForStandalone() throws FrameworkException {
        if (isInitialized()) {
            throw new FrameworkException("Already initialized");
        }

        standalone = true;
    }

    /**
     * Signal the ServiceManager to start only a minimal set of services, for use by Acq CLI
     * @throws FrameworkException On any error
     */
    public void startForUtil() throws FrameworkException {
        if (isInitialized()) {
            throw new FrameworkException("Already initialized");
        }

        initForUtilOnly = true;
        initCommLayerOnly = true;
    }

    /**
     * Signal the ServiceManager to start all services (default behavior). Used for unit tests!
     *
     * @throws FrameworkException On any error
     */
    public void startAllServices() throws FrameworkException {
        if (isInitialized()) {
            throw new FrameworkException("Communication manager already initialized");
        }

        initCommLayerOnly = false;
    }

    /**
     * Since the ServiceManager controls the SecurityManager of the JVM, threads are not allowed to control the shutdown
     * of threads (to prevent data sources from interfering with framework thread handling). If a data source created it's
     * own executor service to run some logic on several threads, it should use this method to shut this executor down.
     * This method will ensure that the right privileges are assigned to the thread when trying to "modifyThread".
     *
     * @param executorService executor to shut down
     */
    public void shutdownExecutor(final ExecutorService executorService) {
        // HEADSUP: This is a workaround. We should try to find a way to have the framework allow these operations
        final SecurityManager sm = System.getSecurityManager();
        System.setSecurityManager(null);
        try {
            executorService.shutdown();
        }
        catch (Throwable t) {
            LOG.error(null, t);
        }
        finally {
            System.setSecurityManager(sm);
        }
    }

    @Override
    protected void doStop() {
        final int sleepTime;
        int maxTimeToWait;
        try {
            // Wait until services are stopped (but we don't want to wait forever)...
            sleepTime = 200;
            maxTimeToWait = 19000;

            LOG.info("{} - Stopping acquisition...", getServiceName());

            // Stop all the services... (go on the reverse order)
            for (ListIterator<FrameworkService> it = acqServices.listIterator(acqServices.size()); it.hasPrevious();) {
                final FrameworkService acqService = it.previous();
                acqService.stop();
            }

            LOG.info("{} - Waiting until acq services are stopped...", getServiceName());

            // Wait until main acq services are stopped...
            final List<FrameworkService> unstoppedAcqServices = new LinkedList<FrameworkService>(acqServices);
            while (unstoppedAcqServices.size() > 0 && maxTimeToWait > 0) {
                for (Iterator<FrameworkService> iterator = unstoppedAcqServices.iterator(); iterator.hasNext();) {
                    final FrameworkService frameworkService = iterator.next();
                    if (frameworkService.isStopped()) {
                        iterator.remove();
                    }
                }

                // Wait a while... and try again.
                try {
                    Thread.sleep(sleepTime);
                }
                catch (InterruptedException e) {
                    // Do nothing...
                }

                maxTimeToWait -= sleepTime;
            }
        }
        finally {
            // Disable the security manager we set when starting... (this is important for unit tests! JUnit is trying to
            // shutdown the JVM and this won't let it!)
            System.setSecurityManager(null);
        }

        LOG.info("{} - Signaling global services to stop...", getServiceName());

        // Now begin shutdown of global services...
        for (ListIterator<FrameworkService> it = globalServices.listIterator(globalServices.size()); it.hasPrevious();) {
            final FrameworkService globalService = it.previous();
            globalService.stop();
        }

        LOG.info("{} - Waiting until global services are stopped...", getServiceName());

        final List<FrameworkService> unstoppedGlobalServices = new LinkedList<FrameworkService>(globalServices);
        while (unstoppedGlobalServices.size() > 0 && maxTimeToWait > 0) {
            for (Iterator<FrameworkService> iterator = unstoppedGlobalServices.iterator(); iterator.hasNext();) {
                final FrameworkService frameworkService = iterator.next();
                if (frameworkService.isStopped()) {
                    iterator.remove();
                }
            }

            // Wait a while... and try again.
            try {
                Thread.sleep(sleepTime);
            }
            catch (InterruptedException e) {
                // Do nothing...
            }

            maxTimeToWait -= sleepTime;
        }

        if (maxTimeToWait <= 0) {
            LOG.info("{} - Waited too long for service to stop...", getServiceName());
            reportError("Timeout while waiting for services to stop", null);
        }
        else {
            LOG.info("{} - All services stopped...", getServiceName());
        }
    }

    /**
     * Receives status notifications from the services...
     * These notifications are sent to the acquisition unit manager to try and update the acquisition unit
     * status view in the GUI
     *
     * @param serviceName Service reporting this error
     * @param status      error message
     */
    @Override
    public void error(final String serviceName, final String status, final Throwable cause) {
        StringBuilder sb = new StringBuilder();
        sb.append(serviceName).append(" reported an error: ").append(status);
        LOG.error(sb.toString(), cause);

        // Allow any listener of the service manager itself to be notified that something has happened
        reportError(status, cause);

        // Report error back to acq unit manager (if it was configured)
        if (!isStandalone()) {
            acquisitionUnitManager.reportAcquisitionUnitAudit(status);
        }
    }

    /**
     * A fatal error occurred. The acquisition needs to shut down!
     *
     * @param serviceName Service reporting this error
     * @param status      error message
     */
    @Override
    public void fatal(final String serviceName, final String status, final Throwable cause) {
        StringBuilder sb = new StringBuilder();
        sb.append(serviceName).append(" reported a fatal error: ").append(status);
        LOG.error(sb.toString(), cause);

        if (!isStopping()) {
            // Allow any listener of the service manager itself to be notified that something has happened
            reportFatal(status, cause);

            // Signal the system (the main thread) to initiate shutdown sequence.
            // HEADSUP: We cannot initiate the shutdown sequence in this thread since this is the worker thread and we
            // might cause a dead lock. The system is designed to be started and stopped from a single thread which is
            // the main thread!
            LOG.error("Acquisition is shutting down!");
            Main.getInstance().stop();
        }
    }

    static private Properties loadGeneralProperties() throws FrameworkException {
        final String generalFileName = "preferences/general.properties";
        final java.net.URL general = ServiceManager.class.getClassLoader().getResource(generalFileName);
        if (general == null) {
            throw new FrameworkException("Couldn't locate " + generalFileName + "  Current dir:" + System.getProperty("user.dir"));
        }

        try {
            LOG.info("Importing general preferences from {}", general);

            final Properties properties = new Properties();
            properties.load(ServiceManager.class.getClassLoader().getResourceAsStream(generalFileName));

            return properties;
        }
        catch (IOException e) {
            throw new FrameworkException("Failed importing general preferences", e);
        }
    }

    /**
     * Override the security manager to prevent from any application running under this framework from trying to
     * perform System.exit.
     *
     * <p/>We have seen this happen when one of the datasources was using an API that tried to perform <code>System.exit</code>.
     */
    private static class NoExitSecurityManager extends SecurityManager {
        private static final Set<Class<?>> classesThatCanExitJVM = new HashSet<Class<?>>(10);

        static {
            // TODO: We should probably prevent from changing the security manager itself as well!
            classesThatCanExitJVM.add(Main.class);
            classesThatCanExitJVM.add(Standalone.class);
        }

        /**
         * Prevent from anyone except the classes of the framework from calling System.exit!
         *
         * @param status Exit code
         */
        @Override
        public void checkExit(final int status) {
            if (!canExitJVM()) {
                LOG.warn("Another class is trying to exit the JVM ({})",
                        getClassContext()[0].getName());
                throw new SecurityException("No class but the framework's is allowed to exit the JVM");
            }

            super.checkExit(status);
        }

        /**
         * Allow everything else...
         *
         * @param perm Permission to check
         */
        @Override
        public void checkPermission(final Permission perm) {
            // Prevent from anyone (but framework) from changing the security manager itself
            if (perm.getName().equals("setSecurityManager") && !canChangeSecurityManager()) {
                LOG.error("Another class is trying to change the security manager ({})",
                        getClassContext()[0].getName());
                throw new SecurityException("Changing the security manager is allowed only to the ");
            }

            // Allow everything else...
        }

        /**
         * Allow everything else...
         *
         * @param perm    Permission to check
         * @param context Context
         */
        @Override
        public void checkPermission(final Permission perm, final Object context) {
            // Prevent from anyone (but framework) from changing the security manager itself
            if (perm.getName().equals("setSecurityManager") && !canChangeSecurityManager()) {
                LOG.error("Another class is trying to change the security manager ({})",
                        getClassContext()[0].getName());
                throw new SecurityException("Changing the security manager is allowed only to the ");
            }

            // Allow everything else...
        }

        private boolean canExitJVM() {
            final Class<?>[] stack = getClassContext();

            for (int i = 0; i < stack.length; i++) {
                final Class<?> clazz = stack[i];

                // The only class allowed to exit the JVM is a permitted class that runs immediately after it the system call.
                if (classesThatCanExitJVM.contains(clazz)) {
                    return ((i > 0) && stack[i - 1].equals(System.class));
                }
            }

            return false;
        }

        private boolean canChangeSecurityManager() {
            final Class<?>[] stack = getClassContext();

            for (int i = 0; i < stack.length; i++) {
                final Class<?> clazz = stack[i];

                // The only class allowed to set the security manager is the ServiceManager. There should be a System call immediately
                // after it.
                if (clazz == ServiceManager.class) {
                    return ((i > 0) && stack[i - 1].equals(System.class));
                }
            }

            return false;
        }
    }

}
