package com.onaro.sanscreen.acquisition.framework.mgmt.log;

import org.apache.log4j.Appender;
import org.apache.log4j.AppenderSkeleton;
import org.apache.log4j.spi.LoggingEvent;

import java.io.Writer;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

/**
 * Implements a log4j appender that redirects data source log entries to the assigned to individual
 * samples.
 *
 * <p/>The entries for a given sample are identified by the framework thread assigned for this sample. Currently, if
 * a data source spawns new threads, the log entries coming from these threads will not be included in the sample
 * log gile.
 */
public class SampleAppenderDistributor extends AppenderSkeleton {
    private final Map<String, SampleAppender> sampleAppenders = Collections.synchronizedMap(new HashMap<String, SampleAppender>());

    public SampleAppenderDistributor() {
        super();
    }

    /**
     * Adds a log4j appender for this sample. The file was created externally by {@link com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource#startingAcquisition}
     *
     * @param threadName thread name to identify this sample
     * @param out file to contain log entires (a recording file?)
     */
    public void addSampleAppender(final String threadName, final Writer out) {
        sampleAppenders.put(threadName, new SampleAppender(out));
    }

    /**
     * Remove the sample appender.
     *
     * <p/>Note: This will <b>not</b> close the file. Since the file was created outside this appender
     * ({@link com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource#finishedAcquisition}) we will manage
     * it outside this class (since this file is created as a recording file, the framework will take care of closing this
     * file before it is creating the ZIP)
     *
     * @param threadName thread name to identify this sample
     */
    public void removeSampleAppender(final String threadName) {
        sampleAppenders.remove(threadName);
    }

    /**
     * Based on the current thread, tries to identify where to redirect the logging event
     */
    protected void append(LoggingEvent event) {
        final SampleAppender sampleAppender = sampleAppenders.get(Thread.currentThread().getName());
        if (sampleAppender == null)
            return;

        sampleAppender.append(event);
    }

    public boolean requiresLayout() {
        return false;
    }

    public void close() {
        synchronized (sampleAppenders) {
            for (Appender appender : sampleAppenders.values()) {
                appender.close();
            }

            sampleAppenders.clear();
        }
    }
}
