package com.onaro.sanscreen.acquisition.framework.mgmt.scheduler;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.mgmt.FrameworkService;
import com.onaro.sanscreen.acquisition.framework.mgmt.FrameworkThread;

import java.util.Map;
import java.util.List;
import java.util.LinkedList;
import java.util.concurrent.ExecutorService;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class handles scheduling of events.
 *
 * <p/>
 * There are 3 parameters for this class:<br>
 * <b><code>Type</code></b>: An object passed as-is by the scheduler from the
 * calling method. This object is used when adding details to the scheduled
 * event like the
 * {@link com.onaro.sanscreen.acquisition.framework.datasource.Package} used
 * when scheduling data sources.<br>
 *
 * <b><code>Target</code></b>: The event identifier. This object is also passed
 * by the calling method. For data sources, the data source instance is used as
 * the event's target object. This allows us to treat all the events from a
 * specific target (like unscheduling all events for a specific data source)<br>
 *
 * <b><code>Service</code></b>: The service this scheduler is running under.
 * This is important to allow the scheduler to report errors back to the parent
 * service. The Service object has to extend {@link FrameworkService}.<br>
 *
 * <p/>
 * When scheduling an evant, the logic to run has to be provided as well. This
 * logic will be executed when possible in a special thread assigned for this
 * poll alone.
 */
public class Scheduler<Type, Target, Service extends FrameworkService> implements SchedulerMBean {

    private static final Logger LOG = LoggerFactory.getLogger(Scheduler.class);

    private final Service parentService;
    private final ExecutorService threadPool;

    private SchedulerTasksTimer<Type, Target, Service> tasksTimer;
    private SchedulerPendingQueue<Type, Target, Service> pendingQueue;
    private SchedulerTasksExecutor<Type, Target, Service> tasksExecutor;

    // Utility threads
    private List<TaskAssigner<Type, Target, Service>> taskAssigners;

    private final Map<Type, Limit> limits;

    public Scheduler(final Map<Type, Limit> limits, final Service parentService, final ExecutorService threadPool) {
        this.limits = limits;
        this.parentService = parentService;
        this.threadPool = threadPool;

        this.taskAssigners = new LinkedList<TaskAssigner<Type, Target, Service>>();
    }

    /**
     * Initializes the scheduler
     */
    public synchronized void start() {
        LOG.debug("initializing scheduler {}", this);

        tasksTimer = new SchedulerTasksTimer<Type, Target, Service>();
        pendingQueue = new SchedulerPendingQueue<Type, Target, Service>(limits);
        tasksExecutor = new SchedulerTasksExecutor<Type, Target, Service>(this, limits, threadPool);

        // Create a task assigner for each poll type
        for (final Type type : limits.keySet()) {
            final TaskAssigner<Type, Target, Service> taskAssigner = new TaskAssigner<Type, Target, Service>(type, this, parentService);
            taskAssigners.add(taskAssigner);

            taskAssigner.start();
        }

        LOG.debug("initialized scheduler {}", this);
    }

    /**
     * Stops the scheduler, removing all scheduled events
     */
    public synchronized void stop() {
        LOG.debug("stopping scheduler {}", this);

        if (!parentService.isStopping()) {
            throw new IllegalStateException("Scheduler must be stopped by stopping parent service");
        }

        for (final TaskAssigner<Type, Target, Service> taskAssigner : taskAssigners) {
            taskAssigner.cancel();
        }

        taskAssigners.clear();

        tasksExecutor.cancel();
        pendingQueue.cancel();
        tasksTimer.cancel();

        LOG.debug("stopped scheduler {}", this);
    }

    public boolean isStopping() {
        return parentService.isStopping();
    }

    /**
     * Schedule an event for the target object.
     *
     * All we need to do here is just add a task to the scheduler's timer. When the timer will tick the events, the tasks
     * will be added to the <code>tasksPendingForExecutions</code> to be later collected by the task assigner.
     *
     * @param type
     * @param target
     * @param logicToRun Logic to run when schduler can execute the task
     * @param delay Time to schedule event for (millis from now)
     * @param interval Time interval between scheduled events (millis)
     */
    public void schedule(final Type type, final Target target, final Runnable logicToRun, final long delay, final long interval) throws FrameworkException {
        if (!limits.containsKey(type)) {
            throw new FrameworkException("Scheduler does not support tasks of type " + type);
        }

        // Create a task for this target
        final SchedulerTask<Type, Target, Service> task = new SchedulerTask<Type, Target, Service>(type, target, this, logicToRun);

        // And add it to the timer...
        tasksTimer.schedule(task, delay, interval);
    }

    /**
     * Unschedules this target from the scheduler...
     *
     * @param target
     *
     * @return true if this task was found, false otherwise
     */
    public synchronized boolean unschedule(final Target target) {
        final boolean foundInPending = pendingQueue.removeAllForTarget(target);
        final boolean foundInTimer = tasksTimer.unschedule(target);

        return foundInPending | foundInTimer;
    }

    /**
     * Runs a task outside the boundaries of the scheduler. This is used when 'forcing acquisition' and designed to
     * give prompt response to a user request.
     *
     * This action will be ignored if task is already running
     *
     * @param type
     * @param target
     *
     * @throws FrameworkException If could not find the task
     */
    public synchronized void runNow(final Type type, final Target target) throws FrameworkException {
        // Check if currently executing and ignore
        if (tasksExecutor.currentlyRunningTask(type, target)) {
            return;
        }

        // Lookup the task in the timer queue...
        final SchedulerTask<Type, Target, Service> theTask = tasksTimer.find(target, type);
        if (theTask == null) {
            throw new FrameworkException("Could not find task [" + type + "," + target + "]");
        }

        // And run it now!
        tasksExecutor.executeNow(theTask);
    }

    @Override
    public String[] getSchedulerTasksInTimer() {
        return tasksTimer.describeStatus();
    }

    @Override
    public String[] getSchedulerTasksInPendingQueue() {
        return pendingQueue.describeStatus();
    }

    @Override
    public String[] getSchedulerTasksInExecution() {
        return tasksExecutor.describeStatus();
    }

    @Override
    public int getNumOfSchedulerTasksInTimer() {
        return tasksTimer.size();
    }

    @Override
    public int getNumOfSchedulerTasksInPendingQueue() {
        return pendingQueue.size();
    }

    @Override
    public int getNumOfSchedulerTasksInExecution() {
        return tasksExecutor.size();
    }

    @Override
    public String[] getRejectHistory() {
        return pendingQueue.getRejectHistory();
    }

    SchedulerTasksExecutor<Type, Target, Service> getTasksExecutor() {
        return tasksExecutor;
    }

    SchedulerPendingQueue<Type, Target, Service> getPendingQueue() {
        return pendingQueue;
    }

    /**
     * The main purpose of this thread is to impose the scheduler limits on the tasks. This thread moves threads that
     * are pending execution to the execution queue (when available)
     */
    private static class TaskAssigner<Type, Target, Service extends FrameworkService> extends FrameworkThread<Service> {
        private Type type;
        private final Scheduler<Type, Target, Service> scheduler;

        public TaskAssigner(final Type type, final Scheduler<Type, Target, Service> scheduler, final Service frameworkService) {
            super("TaskAssigner", frameworkService);
            this.type = type;
            this.scheduler = scheduler;
        }

        @Override
        public boolean doRun() {
            try {
                while (!isStopping()) {
                    // BLOCKING: Wait until a task is pending...
                    final SchedulerTask<Type, Target, Service> task = scheduler.pendingQueue.take(type);

                    // BLOCKING: This task is ready for execution!
                    scheduler.tasksExecutor.execute(task);
                }
            }
            catch (InterruptedException e) {
                // We're stopping. Fall through...
            }

            return isStopping();
        }
    }

    /**
     * Used when running a task in the thread pool. When the task is done, this
     * object makes sure the resources allocated
     * for this tasks are released so the next task will be able to run.
     */
    static class LogicWrapper<Type, Target, Service extends FrameworkService> implements Runnable {
        private final Scheduler<Type, Target, Service> scheduler;
        private final SchedulerTask<Type, Target, Service> task;

        public LogicWrapper(final Scheduler<Type, Target, Service> scheduler, final SchedulerTask<Type, Target, Service> task) {
            this.scheduler = scheduler;
            this.task = task;
        }

        @Override
        public void run() {
            try {
                LOG.debug("executing {}", task);
                task.runLogic();
            }
            finally {
                scheduler.tasksExecutor.clearFinishedTask(task);
            }
        }
    }
}
