package com.onaro.sanscreen.acquisition.framework.mgmt.scheduler;

import com.onaro.sanscreen.acquisition.framework.mgmt.FrameworkService;

import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;

/**
 * Handles the queue of pending tasks for execution
 */
class SchedulerPendingQueue<Type, Target, Service extends FrameworkService> {
    private static Logger logger = Logger.getLogger(SchedulerTasksExecutor.class);

    private final Map<Type, LinkedBlockingQueue<SchedulerTask<Type, Target, Service>>> pendingQueues;
    private final Map<Target, Integer> rejectHistory;

    SchedulerPendingQueue(Map<Type, Limit> limits) {
        pendingQueues = new HashMap<Type, LinkedBlockingQueue<SchedulerTask<Type, Target, Service>>>(10);
        rejectHistory = new HashMap<Target, Integer>(10);

        // Populate the pending queue with all possible poll types
        for (final Type type : limits.keySet()) {
            final LinkedBlockingQueue<SchedulerTask<Type, Target, Service>> pendingTasks = new LinkedBlockingQueue<SchedulerTask<Type, Target, Service>>();
            pendingQueues.put(type, pendingTasks);
        }
    }

    void cancel() {
        pendingQueues.clear();
    }

    boolean removeAllForTarget(final Target target) {
        boolean found = false;

        for (final LinkedBlockingQueue<SchedulerTask<Type, Target, Service>> pendingTasks : pendingQueues.values()) {
            // Find all occurences of this target in the tasks pending for execution...
            for (final Iterator<SchedulerTask<Type, Target, Service>> it = pendingTasks.iterator(); it.hasNext();) {
                final SchedulerTask<Type, Target, Service> pendingTask = it.next();
                if (pendingTask.getTarget().equals(target)) {
                    // Remove it from the queue
                    it.remove();

                    found = true;
                }
            }
        }

        rejectHistory.remove(target);

        return found;
    }

    /**
     * Takes the next task from the pending queue. If there are no tasks pending, this method will block until a task
     * is added to the queue.
     *
     * Note: We do not need to synchronize between this method and {@link #addTask}! If during the <code>contains</code>
     * the pending task will be taken it's ok because it means that the task is currently being executed.
     *
     * @throws InterruptedException
     */
    SchedulerTask<Type, Target, Service> take(final Type type) throws InterruptedException {
        final LinkedBlockingQueue<SchedulerTask<Type, Target, Service>> pendingTasks = pendingQueues.get(type);
        assert pendingTasks != null;

        // Get the first task in the queue. This will be the first task scheduled to run according to the
        // time and interval constraints
        return pendingTasks.take();
    }

    /**
     * Add a task to the pending queue. If the task already exists, this method will be ignored (and return false)
     *
     * @param task
     *
     * @return true if task was added, false otherwise
     */
    boolean addTask(final SchedulerTask<Type, Target, Service> task) {
        // Lookup the right queue for this task
        final LinkedBlockingQueue<SchedulerTask<Type, Target, Service>> pendingTasks = pendingQueues.get(task.getType());
        if (pendingTasks == null) {
            logger.error("Scheduler - Task type " + task.getType() + " has no pending queue assigned!");
            return false;
        }

        // If pending tasks already contain this task, there's no reason to over populate the queue
        if (pendingTasks.contains(task)) {
            Integer r = rejectHistory.get(task.getTarget());
            if (r == null)
                r = 0;

            rejectHistory.put(task.getTarget(), r + 1);
            return false;
        }

        if (logger.isDebugEnabled())
            logger.debug("Scheduler - Task goes to pending : " + task.toString());

        pendingTasks.add(task);
        return true;
    }

    public String[] describeStatus() {
        final String[] status = new String[size()];

        int i = 0;

        for (final LinkedBlockingQueue<SchedulerTask<Type, Target, Service>> pendingTasks : pendingQueues.values()) {
            for (SchedulerTask<Type, Target, Service> task : pendingTasks) {
                status[i] = task.toString();
                i++;
            }
        }

        return status;
    }

    public String[] getRejectHistory() {
        final String[] history = new String[rejectHistory.size()];

        int i = 0;

        for (final Map.Entry<Target, Integer> entry : rejectHistory.entrySet()) {
            history[i] = entry.getKey() + " = " + entry.getValue();
            i++;
        }

        return history;
    }

    public int size() {
        int s = 0;

        for (final LinkedBlockingQueue<SchedulerTask<Type, Target, Service>> pendingTasks : pendingQueues.values())
            s += pendingTasks.size();

        return s;
    }
}
