package com.onaro.sanscreen.acquisition.framework.mgmt.scheduler;

import com.netapp.oci.platform.acq.interfaces.data.PackagePollSkippedStatus;
import com.netapp.oci.platform.acq.interfaces.remote.AcquisitionStatusRemote;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.BaseDataSource;
import com.onaro.sanscreen.acquisition.framework.mgmt.FrameworkService;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.util.Schema;
import com.onaro.util.Average;
import com.onaro.util.CyclicArray;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.TimerTask;

/**
 * This object represents the scheduler task that is being managed inside the scheduler's queues.
 */
class SchedulerTask<Type, Target, Service extends FrameworkService> extends TimerTask {
    private static final Logger LOG = LoggerFactory.getLogger(SchedulerTask.class);

    private final Type type;
    private final Target target;
    private final Scheduler<Type, Target, Service> scheduler;
    private final Runnable logicToRun;
    private final Average sampleTimeAverage = new Average(3);

    private CyclicArray<Long> executionHistory = new CyclicArray<Long>(20);

    public SchedulerTask(final Type type, final Target target, final Scheduler<Type, Target, Service> listener, final Runnable logicToRun) {
        this.type = type;
        this.target = target;
        this.scheduler = listener;
        this.logicToRun = logicToRun;
    }

    /**
     * Called by the timer when the task is scheduled. This will not trigger execution. Instead, it will add the task
     * the the pending tasks for execution to be later polled by the {@link Scheduler.TaskAssigner} thread
     *
     * Note: Since we're using java Timer (that utilizes only one thread), threre is no need to synchronize this code
     * since there will be only one thread adding to the pending queue
     */
    public void run() {
        if (scheduler.isStopping())
            return;

        // If task is already running, log that the poll is skipped and send out a JMS message (US22983), then skip.
        if (scheduler.getTasksExecutor().currentlyRunningTask(this.getType(), this.getTarget())) {
            sendPollSkipped(this.getType(), this.getTarget());
            return;
        }

        // Add the task to the pending queue (if it's not already there)
        scheduler.getPendingQueue().addTask(this);
    }

    public Type getType() {
        return type;
    }

    public Target getTarget() {
        return target;
    }

    /**
     * Execute the underlying logic for this task
     */
    public void runLogic() {
        long startTime = System.currentTimeMillis();

        logicToRun.run();
        executionHistory.add(startTime);

        // Update the time it took this task to execute (we keep an average of the last several polls). This will
        // help us understand which task is considered to be a 'fast' task.
        sampleTimeAverage.add(System.currentTimeMillis() - startTime);
    }

    public Long getSampleTimeAverage() {
        return sampleTimeAverage.calculateAverage();
    }

    /**
     * Used when trying to find the task in a collection. The key of the task is the <code>type</code>
     * and the <code>target</code>.
     */
    public boolean equals(final Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final SchedulerTask that = (SchedulerTask)o;

        if (type != null ? !type.equals(that.type) : that.type != null) return false;
        if (target != null ? !target.equals(that.target) : that.target != null) return false;

        return true;
    }

    public int hashCode() {
        int result;
        result = (type != null ? type.hashCode() : 0);
        result = 29 * result + (target != null ? target.hashCode() : 0);
        return result;
    }

    public String toString() {
        return type + ":" + target;
    }

    public String formatExecutionHistory() {
        final StringBuilder sb = new StringBuilder(200);

        long referenceTime = System.currentTimeMillis();

        sb.append('[');

        for (final Long time : executionHistory)
            sb.append((referenceTime - time) / 1000).append(",");

        sb.deleteCharAt(sb.length() - 1);
        sb.append(']');

        return sb.toString();
    }


    /**
     * This method is to log a message saying "a poll is skipped" and also send out a JMS message
     *
     * @param type   String of packageName
     * @param target  BaseDataSource object
     */
    private void sendPollSkipped(Type type, Target target) {
        final ServiceManager serviceManager = ServiceManager.getInstance();
        if (serviceManager.isStandalone()) {
            return;
        }

        Long datasourceID = null;
        String datasourceName = null;
        String packageName = null;
        try {
            final BaseDataSource thisDataSource = (BaseDataSource)target;
            datasourceID = thisDataSource.getDataSourceID();
            datasourceName = thisDataSource.getDataSourceName();
            packageName = (String)type;
        }
        catch(ClassCastException cce) {
            LOG.error("Failed to send JMS message due to ClassCastException - ",cce.getMessage());
            return;
        }

        LOG.warn(datasourceName + " - Data source is already performing a poll of " + packageName + ". Therefore this poll is skipped. ");

        final PackagePollSkippedStatus pollSkippedStatus = new PackagePollSkippedStatus(datasourceName, datasourceID, packageName, System.currentTimeMillis());
        try {
            final AcquisitionStatusRemote acqStatusSession = serviceManager.getCommunicationManager().getSession(AcquisitionStatusRemote.JNDI_NAME);
            if (acqStatusSession == null) {
                LOG.error("Failed to update AcquisitionStatusRemote packageSkipped {} {}", datasourceName, packageName);
                return;
            }
            acqStatusSession.packageSkipped(Schema.NETAPP, pollSkippedStatus);
        }
        catch (FrameworkException | ServerInterfaceException ex) {
            LOG.error("Failed to update AcquisitionStatusRemote packageStarted {} {}",new Object[]{datasourceName, packageName});
        }

    }
}
