package com.onaro.sanscreen.acquisition.framework.mgmt.scheduler;

import com.onaro.sanscreen.acquisition.framework.mgmt.FrameworkService;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.*;

import org.apache.log4j.Logger;

/**
 * Handles the execution of tasks.
 *
 * <p/>This executor maintains differnt queues for each type of task to allow for example performance polls to be resourced
 * independantly from foundation polls.
 */
class SchedulerTasksExecutor<Type, Target, Service extends FrameworkService> {
    private static Logger logger = Logger.getLogger(SchedulerTasksExecutor.class);

    private final Scheduler<Type, Target, Service> scheduler;
    private final Map<Type, Limit> limits;
    private final Map<Type, TaskQueue> taskQueues;
    private final ExecutorService threadPool;

    SchedulerTasksExecutor(final Scheduler<Type, Target, Service> scheduler, final Map<Type, Limit> limits, final ExecutorService threadPool) {
        this.scheduler = scheduler;
        this.limits = limits;
        this.threadPool = threadPool;

        taskQueues = new HashMap<Type, TaskQueue>(10);
    }

    void cancel() {
        // OPHIR: Since the introduction of the Security Manager, this kind of actions are a little hard to perform.
        // We don't really need this shutdown since every thread will check for stopping condition when running.
        //threadPool.shutdown();

        final List<TaskQueue> copyOfQueues = new LinkedList<TaskQueue>();
        synchronized (taskQueues) {
            final Iterator<Map.Entry<Type, TaskQueue>> it = taskQueues.entrySet().iterator();
            while (it.hasNext()) {
                final Map.Entry<Type, TaskQueue> entry = it.next();
                copyOfQueues.add(entry.getValue());
                it.remove();
            }
        }

        for (final TaskQueue taskQueue : copyOfQueues)
            taskQueue.clear();
    }

    boolean currentlyRunningTask(final Type type, final Target target) {
        final TaskQueue queue;

        synchronized (taskQueues) {
            // Look up the right queue (based on the type)
            queue = taskQueues.get(type);
        }

        // If no queue is defined yet for this type, then this task is definitely not running!
        if (queue == null)
            return false;

        // We have the right queue. Now see if this task is in this queue
        final SchedulerTask<Type, Target, Service> dummyTask = new SchedulerTask<Type, Target, Service>(type, target, null, null);
        return queue.contains(dummyTask);
    }

    /**
     * Executes a task.
     *
     * This method is called by the {@link Scheduler.TaskAssigner} when a task is really supposed to execute.
     * It will put the task in the completion service allocate a thread from the scheduler's thread pool and run the task.
     * When the task is done, it will use the same thread (from the thread pool) to remove itself from the running queue
     * and to make room for another task to run.
     *
     * @param task Task to execute
     */
    void execute(final SchedulerTask<Type, Target, Service> task) throws InterruptedException {
        TaskQueue queue;

        synchronized (taskQueues) {
            // Find the right queue (based on the task type)
            queue = taskQueues.get(task.getType());

            // If we could not find a queue for this type, create it now
            if (queue == null) {
                // Lookup the limit for this queue...
                Limit limit = limits.get(task.getType());
                if (limit == null)
                    limit = Limit.createDefault();

                queue = new TaskQueue(limit, false);
                taskQueues.put(task.getType(), queue);
            }
        }

        if (logger.isDebugEnabled())
            logger.debug("Scheduler - Queuing for execution : " + task.toString());

        // BLOCKING: Wait until there's room to run an additional task (since the running tasks queue is bounded by size, this method will
        // block until there's an available place in the queue)
        queue.put(task);

        executeNow(task);
    }

    /**
     * Executes a task without considering the execution limits. This is used for 'forcing acquisition'
     *
     * @param task Task to execute
     */
    void executeNow(final SchedulerTask<Type, Target, Service> task) {
        // Task can run. Add it to the completion service to run within a thread pool
        threadPool.submit(new Scheduler.LogicWrapper<Type, Target, Service>(scheduler, task));
    }

    /**
     * Called by the {@link Scheduler.LogicWrapper} when a task is finished
     *
     * Note: this method does not need to be synchronized since the queue remove is already protected and the queue
     * is bounded by the number of resources. There is only one thread (TaskAssigner) adding to this queue, and the
     * thread pool's threads only take things out.
     */
    boolean clearFinishedTask(final SchedulerTask<Type, Target, Service> task) {
        final TaskQueue queue;

        synchronized (taskQueues) {
            queue = taskQueues.get(task.getType());
            assert queue != null;
        }

        // If this method was triggered by a 'force acq', then the task will not be in the queue. This is ok, since
        // the remove method will just return false.
        return queue.remove(task);
    }

    public String[] describeStatus() {
        final List<String> tasksDescs = new LinkedList<String>();

        final List<TaskQueue> copyOfQueues = new LinkedList<TaskQueue>();
        synchronized (taskQueues) {
            copyOfQueues.addAll(taskQueues.values());
        }

        // Go over all the queues and let each describe itself
        for (final TaskQueue queue : copyOfQueues)
            queue.describeTaskStatus(tasksDescs);

        final String[] status = new String[tasksDescs.size()];
        return tasksDescs.toArray(status);
    }

    public int size() {
        int size = 0;

        synchronized (taskQueues) {
            for (final TaskQueue queue : taskQueues.values())
                size += queue.size();
        }

        return size;
    }

    private class TaskQueue extends ArrayBlockingQueue<SchedulerTask<Type, Target, Service>> {
        public TaskQueue(final Limit limit, final boolean fair) {
            super(limit.getMaxTasks(), fair);
        }

        void describeTaskStatus(final List<String> tasksDescs) {
            for (SchedulerTask<Type, Target, Service> task : this)
                tasksDescs.add(task.toString());
        }
    }
}
