package com.onaro.sanscreen.acquisition.framework.recordings;

import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileWriter;
import com.onaro.util.CyclicArray;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Encapsulates recording support in framework:
 * - support for variable number of samples in session
 * - support for self-contained, simulatable zip
 * - support for simulation script, i.e. ability to simulate mutiple samples in order
 * - support for different recording mechanisms
 */
public class RecordingSession {
    final private PackageConfiguration configuration;

    private final RecordingSample rootSample = new RecordingSample();
    private final SamplesCollection samples;
    private RecordingSample currentSample;

    public RecordingSession(final PackageConfiguration configuration) {
        this.configuration = configuration;
        samples = new SamplesCollection(1);
    }

    public Collection<RecordingSample> getSamples() {
        return samples;
    }

    public RecordingSample getRootSample() {
        return rootSample;
    }

    public RecordingSample createSample() {
        final RecordingSample sample = new RecordingSample();
        samples.add(sample);
        currentSample = sample;
        return sample;
    }

    public void reset() {
        rootSample.reset();

        for (final RecordingSample sample : samples)
            sample.reset();

        samples.clear();
    }

    /**
     * Closes all the files generated during the sample. it will ZIP up all samples.
     *
     * @return The ZIP file created (when <code>keepRecording</code> is true), null otherwise
     *
     * @throws IOException
     */
    public File buildRecordingFile() throws IOException {
        if (rootSample.getFileHandles().size() == 0 && samples.size() == 0)
            return null; // nothing to do !

        // Close all the files that were created during the sample
        closeFiles();

        // And zip them!
        try {
            return new DSSessionFileWriter(rootSample, samples, configuration).toFile();
        }
        finally {
            reset();
        }
    }

    protected void closeFiles() {
        rootSample.closeFiles();
        for (final RecordingSample sample : samples)
            sample.closeFiles();
    }

    protected void deleteFiles() {
        rootSample.deleteFiles();
        for (final RecordingSample sample : samples)
            sample.deleteFiles();
    }

    public RecordingSample getCurrentSample() {
        return currentSample;
    }

    public int getMaxRecordingSamples() {
        return samples.maxSize();
    }

    /**
     * @return true if session has accumulated enough samples
     */
    public boolean hasMaxRecordingSamples(){
        return samples.size() >= samples.maxSize();
    }

    /**
     * @param maxRecordingSamples The maximum number of recording samples in the recording session 
     */
    public void setMaxRecordingSamples(int maxRecordingSamples) {
        maxRecordingSamples = maxRecordingSamples < 1 ? 1 : maxRecordingSamples;
        final List<RecordingSample> old = new LinkedList<RecordingSample>();
        old.addAll(samples);
        samples.reset(maxRecordingSamples);
        if (old.size() <= samples.size()) {
            samples.addAll(old);
        }
        else {
            samples.addAll(old.subList(old.size() - samples.size(), old.size()));
        }
    }

    static class SamplesCollection extends CyclicArray<RecordingSample> {
        public SamplesCollection(final int maxNumberOfSamples) {
            super(maxNumberOfSamples);
        }

        protected void onOverwrite(final RecordingSample sample) {
            sample.reset();
        }
    }
}
