/*
 * Copyright (c) 2013 NetApp
 * All rights reserved
 */

package com.onaro.sanscreen.acquisition.framework.standalone;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.config.ConfigurationFactory;
import com.onaro.sanscreen.acquisition.framework.datasource.config.DataSourceConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.config.PackageConfiguration;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.Attribute;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.DataSourceMetaData;
import com.onaro.sanscreen.acquisition.framework.datasource.metadata.PackageMetaData;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;

/**
 * This class creates an command line interface for activating a data source without having to need a SANscreen
 * installation. This is usefull when trying out new data sources or new products at customer sites.
 *
 * <p/>This code uses the deploy descriptor XML file to dynamically create the list of required properties for this
 * data source.
 *
 * <p/>This class is abstract and does not have a <code>main</code> method. See specific package implementations
 * for details.
 */
public abstract class Standalone {
    private final static String VERSION = "Standalone v.2.0";
    private final static Logger logger = Logger.getLogger(Standalone.class);

    /**
     * Standalone specific properties, not coming from XML file.
     */
    protected Map<String, PropertyDetails> standaloneProperties;

    protected Standalone() {
        standaloneProperties = new HashMap<String, PropertyDetails>();
        standaloneProperties.put("xml", new PropertyDetails("xml", "Deploy Descriptor file", "", true));
        standaloneProperties.put("datasource", new PropertyDetails("datasource", "Standalone", "Data source name", true));
        standaloneProperties.put("packages", new PropertyDetails("packages", "foundation", "Comma seperated list of packages to run", true));
        standaloneProperties.put("help", new PropertyDetails("help", Boolean.FALSE.toString(), "Help", false));
        standaloneProperties.put("level", new PropertyDetails("level", Level.ERROR.toString(), "Log level: error|warn|info|debug", false));
    }

    /**
     * Translate the given command line attributes to a set of properties that will later be used to instantiate the
     * data source.
     *
     * @param args command line arguments
     *
     * @return A {@link PollDescription} object describing what and how should be run or <code>null</code> if help was displayed.
     *
     * @throws FrameworkException When something is missing or not configured properly
     */
    public PollDescription initConfiguration(final String[] args) throws FrameworkException, IOException {
        return initConfiguration(args, null);
    }

    /**
     * Translate the given command line attributes to a set of properties that will later be used to instantiate the
     * data source.
     *
     * @param args command line arguments
     * @param initialConfiguration Set of properties set initially for the data source
     *
     * @return A {@link PollDescription} object describing what and how should be run or <code>null</code> if help was displayed.
     *
     * @throws FrameworkException When something is missing or not configured properly
     */
    public PollDescription initConfiguration(final String[] args, final Properties initialConfiguration) throws FrameworkException, IOException {
        // Create the properties from command line.
        final Properties properties = createPropertiesFromCommandLineArgs(args, initialConfiguration);

        // Now we can reset the logger to the required log level
        initLogger(getLogLevel(properties));

        // Gets XML file.
        final File deployDescriptor = extractDeployDescriptorPath(properties.getProperty("xml"));

        final Set<String> packages = figureOutPackages(properties);

        // Checks for HELP option. This will also parse the XML file to display a full help message.
        final String helpOption = properties.getProperty("help");
        if (helpOption != null && Boolean.parseBoolean(helpOption)) {
            printUsage(deployDescriptor, packages);
            return null;
        }

        try {
            // Create a configuration object from these properties
            final ConfigurationFactory factory = new ConfigurationFactory(properties.getProperty("datasource", "standalone"), 1L);


            // IBG-651: Foundation only standalone attempts to load performance classes
            // This section of code previously would instantiation a DataSourceConfiguration without specifying
            // a package name. This would try to instantiate a DataSourceConfiguration object for each package type
            // specified in the data source xml config file. This would cause class not found exceptions when the
            // standalone was only created for one package, such as foundation.
            //
            // The solution here is to instantiate 1 DataSourceConfiguration and add additional PackageConfiguration
            // instance to it as specified by the desired packages.
            DataSourceConfiguration configuration = null;

            for (final String packageName : packages) {
                // Create the DataSourceConfiguration if it has not been already
                if(configuration == null){
                    configuration = factory.createForStandalone(packageName, properties, deployDescriptor);

                    final PackageConfiguration packageConfiguration = configuration.getPackageConfiguration(packageName);
                    if (packageConfiguration == null)
                        throw new FrameworkException("Package " + packageName + " not defined for data source!");
                    if ( !packageConfiguration.isEnabled() )
                        throw new FrameworkException("Package " + packageName + " not enabled for data source!");

                }
                // If the DataSourceConfiguration has already been created, create a temp one so it's PackageConfiguration
                // can be registered to the main DataSourceConfiguration object.
                else{
                    // Create a temp DataSourceConfiguration
                    final DataSourceConfiguration anotherConfiguration =  factory.createForStandalone(properties, deployDescriptor);

                    final PackageConfiguration packageConfiguration = anotherConfiguration.getPackageConfiguration(packageName);
                    if (packageConfiguration == null)
                        throw new FrameworkException("Package " + packageName + " not defined for data source!");
                    if ( !packageConfiguration.isEnabled() )
                        throw new FrameworkException("Package " + packageName + " not enabled for data source!");

                    // Register PackageConfiguration with main DataSourceConfiguration
                    configuration.registerPackageConfiguraiton(packageConfiguration);
                }
            }

            return new PollDescription(packages, configuration);
        }
        catch (FrameworkException e) {
            logger.error(e.getMessage(), e);

            // If we have a problem creating the configuration based on the given properties, let's print the usage.
            // There might be some missing mandatory properties...
            printUsage(deployDescriptor, packages);
            return null;
        }
    }

    private Set<String> figureOutPackages(final Properties properties) {
        final String[] packageNames = properties.getProperty("packages", "foundation").split(",");

        if (packageNames.length == 0 || !packageNames[0].contains("foundation")) {
            throw new RuntimeException("Foundation package is not specified first or at all.");
        }

        final Set<String> packages = new LinkedHashSet<String>();
        for (final String packageName : packageNames)
            packages.add(packageName.trim());

        return packages;
    }

    private File extractDeployDescriptorPath(final String deployDesciptorPath) throws FrameworkException {
        if (deployDesciptorPath == null || deployDesciptorPath.trim().length() == 0)
            throw new FrameworkException("xml is a mandatory attribute");

        final File deployDescriptor = new File(deployDesciptorPath);
        if (!deployDescriptor.exists())
            throw new FrameworkException("Could not find deploy descriptor " + deployDescriptor.getAbsolutePath());

        return deployDescriptor;
    }

    /**
     * Init the logger instance.
     *
     * @param logLevel Log level to set the logger to
     */
    protected static void initLogger(final Level logLevel) {
        final ConsoleAppender consoleAppender = new ConsoleAppender(new SimpleLayout());
        consoleAppender.setTarget("System.out");
        Logger.getRootLogger().addAppender(consoleAppender);
        Logger.getRootLogger().setLevel(logLevel);
    }

    /**
     * Gets the "level" option's value if is present in the command line.
     *
     * @param cmdProperties Properties entered at command line
     *
     * @return Level (Log4J) instance with the corresponding value based on input entry.
     */
    protected final Level getLogLevel(final Properties cmdProperties) {
        if (cmdProperties == null || !cmdProperties.containsKey("level"))
            return Level.INFO;

        return Level.toLevel(cmdProperties.getProperty("level"));
    }

    /**
     * Parses the command line to set the properties for this data source
     *
     * <p/>Assuming user's imput follows the format: -&lt;property-name&gt;=&lt;value&gt;
     *
     * <p/>example: standalone -ip=10.5.6.76 -name="ds-8000" -path="c:\program files\cli" ...
     *
     * @param args User's input command line.
     *
     * @param initialConfiguration -
     * @return Properties that has been set with their values.
     */
    protected final Properties createPropertiesFromCommandLineArgs(final String[] args, final Properties initialConfiguration) throws FrameworkException {
        final Properties properties = new Properties();

        if (initialConfiguration != null)
            properties.putAll(initialConfiguration);

        // Loops though all the User's input arguments.
        for (final String token : args) {
            // Option HELP is a special case, because it doesn't have a value.
            if (token.contains("-help")) {
                properties.setProperty("help", Boolean.TRUE.toString());
                continue;
            }

            // Finds the dash sign...
            final int dash = token.indexOf("-");
            if (dash == -1)
                throw new FrameworkException("Option entered: [" + token + "] do not have prefix '-' (dash) char.");

            // Finds the equal sign...
            final int equal = token.indexOf("=");
            if (equal == -1)
                throw new FrameworkException("Equal sign not found in option entered: [" + token + "]");

            // Gets name and value.
            final String optionName = token.substring(dash + 1, equal);
            final String optionValue = token.substring(equal + 1);

            properties.setProperty(optionName, optionValue);
        }

        return properties;
    }

    /**
     * Print usage help. This code gives a description of all the available properties for this data source (as read from the XML).
     */
    protected final void printUsage(final File deployDescriptor, final Set<String> packages) throws FrameworkException, IOException {
        final StringBuilder sb = new StringBuilder(5000);

        // Parse the meta data from the deploy descriptor to add a little more information when printing the usage
        final DataSourceMetaData dataSourceMetaData = DataSourceMetaData.createWithClassLoadersDisabled(deployDescriptor.getPath(), new FileInputStream(deployDescriptor) );

        printCopyRight(sb, dataSourceMetaData);
        printUsageTitle(sb);
        printDetailedUsage(sb, dataSourceMetaData, packages);

        System.out.println(sb.toString());
    }

    private void printCopyRight(final StringBuilder sb, final DataSourceMetaData metaData) {
        final Calendar cal = new GregorianCalendar();
        final int currentYear = cal.get(Calendar.YEAR);

        sb.append("------------------------------------------------------------").append("\n");
        sb.append("COPYRIGHT 2000 to ").append(currentYear).append(" by ONARO, Inc. All rights reserved.").append("\n");
        sb.append("Data source type ").append(metaData.getName()).append(" - ").append(metaData.getDescription()).append("\n");
        sb.append("Standalone version ").append(VERSION).append("\n");
        sb.append("------------------------------------------------------------").append("\n");
    }

    private void printUsageTitle(final StringBuilder sb) {
        sb.append("\n").append("Usage").append("\n").append("=====").append("\n");
        sb.append("standalone.bat [-options=args...]").append("\n");
    }

    private void printDetailedUsage(final StringBuilder sb, final DataSourceMetaData dataSourceMetaData, final Set<String> packages) throws FrameworkException {
        sb.append("\n").append("Where").append("\n").append("=====").append("\n");

        int longestAttributeName = 0;

        for (final String packageName : packages) {
            final PackageMetaData packageMetaData = dataSourceMetaData.getPackage(packageName);
            if (packageMetaData == null)
                throw new FrameworkException("Package " + packageName + " is not defined for data source");

            // Get a list of all the non hidden attributes in the meta data
            final Map<String, Attribute> attributes = packageMetaData.getAttributes();

            // Figure out the length of the longest attribtue name (including the package name prefix)
            final int max = maxAttrName(attributes, standaloneProperties) + packageMetaData.getName().length() + 1;

            longestAttributeName = Math.max(longestAttributeName, max);
        }

        printStandaloneAttributesUsage(sb, longestAttributeName);

        for (final String packageName : packages) {
            final PackageMetaData packageMetaData = dataSourceMetaData.getPackage(packageName);

            printDataSourceAttributesUsage(sb, packageMetaData.getName(), packageMetaData.getAttributes(), longestAttributeName);
        }

        sb.append("\n");
    }

    private void printStandaloneAttributesUsage(final StringBuilder sb, final int max) {
        for (final PropertyDetails propertyDesc : standaloneProperties.values()) {
            sb.append(appendSpaces(propertyDesc.getName(), max)).append(" : ");

            if (propertyDesc.isMandatory())
                sb.append("(mandatory) ");

            sb.append(propertyDesc.getDescription());

            final String defaultValue = propertyDesc.getDefaultValue();
            if (defaultValue != null && defaultValue.length() > 0)
                sb.append(" (default=").append(defaultValue).append(")");

            sb.append("\n");
        }
    }

    private void printDataSourceAttributesUsage(final StringBuilder sb, final String packageName, final Map<String, Attribute> attributes, final int max) {
        for (final Attribute attribute : attributes.values()) {
            sb.append(appendSpaces(packageName + "." + attribute.getName(), max)).append(" : ");

            if (attribute.isMandatory())
                sb.append("(mandatory) ");

            sb.append(attribute.getDescription());

            final String defaultValue = attribute.getValue();
            if (defaultValue != null && defaultValue.length() > 0)
                sb.append(" (default=").append(defaultValue).append(")");

            sb.append("\n");
        }
    }

    /**
     * Returns the longest property's name. This is done to normalize the output on the screen, formatting the
     * properties name and description in a user-friendly manner.
     */
    private int maxAttrName(final Map<String, Attribute> attributes, final Map<String, PropertyDetails> standaloneProperties) {
        int max = 0;

        for (final Attribute attribute : attributes.values())
            max = Math.max(max, attribute.getName().length());

        for (final String propertyName : standaloneProperties.keySet())
            max = Math.max(max, propertyName.length());

        return max;
    }

    /**
     * Format the program's options output on the screen with equal amount of spaces between properties.
     */
    private String appendSpaces(final String attrName, final int max) {
        final StringBuilder sb = new StringBuilder(max);
        sb.append(attrName);
        int spaces = max - attrName.length();
        while (spaces-- > 0)
            sb.append(" ");

        return sb.toString();
    }

    /**
     * Brief Property Desc. used to represent Application specific attributes.
     */
    protected static final class PropertyDetails {
        private String name;
        private String defaultValue;
        private String description;
        private boolean mandatory;

        public PropertyDetails(final String name, final String value, final String description, final boolean mandatory) {
            this.name = name;
            this.defaultValue = value;
            this.description = description;
            this.mandatory = mandatory;
        }

        public String getName() {
            return name;
        }

        public final String getDefaultValue() {
            return defaultValue;
        }

        public final String getDescription() {
            return description;
        }

        public boolean isMandatory() {
            return mandatory;
        }
    }

    protected static final class PollDescription {
        private final Set<String> packages;
        private final DataSourceConfiguration configuration;

        public PollDescription(final Set<String> packageName, final DataSourceConfiguration configuration) {
            this.packages = packageName;
            this.configuration = configuration;
        }

        public final Set<String> getPackages() {
            return packages;
        }

        public final DataSourceConfiguration getConfiguration() {
            return configuration;
        }
    }
}
