package com.onaro.sanscreen.acquisition.netappfoundation.datasource;

import com.netapp.acquisition.aquisitionserverside.interfaces.remote.AcquisitionServerSession;
import com.netapp.oci.platform.acq.interfaces.data.PackagePollCompleteStatus;
import com.netapp.oci.platform.acq.interfaces.remote.AcquisitionStatusRemote;
import com.netapp.sanscreen.server.api.OriginatorUpdateClientUtils;
import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.StorageRoot;
import com.onaro.sanscreen.acquisition.schema.BaseModelRoot;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionUnitException;
import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.util.Schema;
import org.apache.log4j.Logger;

import java.io.*;
import java.net.HttpRetryException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * Main listener for ner foundation events. This object contains the logic needed to link the foundation sample
 * to the SANscreen server / product.
 * <p/>
 * Replacing this set of interfaces allows the framework to run a data source completely detached from SANscreen
 *
 */
public class NetAppFoundationListener implements NetAppFoundationUpdatesListener, DataSourceSampleListener, NetAppOriginatorProvider, DataSourceLifeCycleListener, NetAppClockSynchronizer {
    private final static Logger logger = Logger.getLogger(NetAppFoundationListener.class);

    private final NetAppFoundationConfiguration configuration;
    private final AuditSession auditSession;

    private boolean sendOnlyUpdates = true;

    private final Map<String, Long> originatorCache = new HashMap<String, Long>(10);

    public NetAppFoundationListener(final NetAppFoundationConfiguration configuration) {
        this.configuration = configuration;
        sendOnlyUpdates = ServiceManager.getInstance().getProperty("SEND_ONLY_UPDATES", Boolean.TRUE.toString()).equalsIgnoreCase(Boolean.TRUE.toString());

        auditSession = new AuditSession(configuration);
    }

    public void dataSourceStarting(final BaseDataSource dataSource) throws FrameworkException {
        // Register this data source for receiving trap notifications from it's primary IP
    }

    public void dataSourceStopping(final BaseDataSource dataSource) throws FrameworkException {
        // Unregistering this data source from any traps
    }

    public void dataSourceStatusChanged(final BaseDataSource dataSource, final DataSourceStatus oldStatus, final DataSourceStatus newStatus, final String message) {
    }

    public void startingSample(final BaseDataSource dataSource, final PollingPackage pollType) {
        auditSession.createSession("Acquiring device");
    }

    public void sampleSuccessfull(final BaseDataSource baseDataSource, final PollingPackage aPackage) throws FrameworkException {
        // This acquisition is completely successfull. This is the time to handle upgrades from non MOID to MOID implementations
        auditSession.updateFinalStatus(AcquisitionServerSession.AuditType.SUCCESS, "Success");

        // private final AcquisitionStatusRemote acqStatusSession;
        retireUnreportedOriginators(aPackage);
        updateAcqPollStatus(baseDataSource, aPackage, true);
    }

    /**
     * Retires all originators reported by the datasource in the past that
     * are not reported on the current successful poll.
     *
     * @param aPackage foundation package being run
     */
    private void retireUnreportedOriginators(final PollingPackage aPackage) {
        // convert ID strings to Long values
        final Set<Long> activeOriginatorIDs = ((NetAppFoundationPackage) aPackage).getCurrentlyReportedOriginatorIds();

        // Now, remove all originators for this datasource that we did not successfully report
        NetAppFoundationConfiguration foundationConfig = (NetAppFoundationConfiguration) aPackage.getConfiguration();
        if (foundationConfig.isRetireUnreportedOriginators() && foundationConfig.isMOID()) {
            try {
                ServiceManager.getInstance().
                        getCommunicationManager().
                        getAcquisitionFacadeSession().
                        retireUnreportedOriginators(activeOriginatorIDs, aPackage.getDataSource().getDataSourceID());
            }
            catch (FrameworkException e) {
                logger.error(e.getMessage(), e);
            }
            catch (RemoteException e) {
                logger.error(e.getMessage(), e);
            }
            catch (AcquisitionUnitException e) {
                logger.error(e.getMessage(), e);
            }
        }
    }


    public void finishedSample(final BaseDataSource dataSource, final PollingPackage aPackage) throws FrameworkException {
        auditSession.closeSession(aPackage);
    }

    public void sampleFailed(final BaseDataSource dataSource, final PollingPackage aPackage, final DataSourceErrorException e) {
        auditSession.updateFinalStatus(e);
        updateAcqPollStatus(dataSource, aPackage, false);
    }

    public void sampleFailed(final BaseDataSource dataSource, final PollingPackage aPackage, final Throwable e) {
        logger.error(aPackage.getLogPrefix() + " - Unhandled exception: " + e.getMessage(), e);
        auditSession.updateFinalStatus(new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Package " + aPackage.getPackageName() + " Failed with error:" + e.getMessage(), aPackage.getLogPrefix() + " - Unhandled exception: " + e.getMessage(), e));
        updateAcqPollStatus(dataSource, aPackage, false);
    }

    /**
     * Returns an originator ID for the given key.
     * <p/>
     * This method will try to use the originator cache first to minimize
     * network calls to the server. If an originator ID could not be found in the cache for the given originatorKey,
     * the server will be queried and the cache updated.
     * <p/>
     * Calls into the AcquisitionFacadeRMI to retrieve an OriginatorID from the server
     * <p/>
     * 1) If datasourceID is null, an exception will be thrown in the server<br>
     * 2) If datasourceID is non-null, and devicekey is null, and a datasource row exists in the database, the
     * datasourceID will be returned as the originatorId<br>
     * 3) If datasourceID is non-null, device key is null, and no datasource row exists in the database, null will be
     * returned<br>
     * 4) If datasourceID is non-null and devicekey is non-null, an existing datasourceID will be returned if present,
     * otherwise a new originatorID will be returned, and entered into the database.<br>
     *
     * @param datasourceID
     * @param originatorKey
     * @param originatorDisplayName
     *
     * @throws FrameworkException
     */
    public long getOriginatorID(final Long datasourceID, final String originatorKey, final String originatorDisplayName) throws FrameworkException {
        // we use the server's acquisitionFacadeSession method "getOriginatorID" to also update the originator's last access time
        // When we use the cached originator id, we skip the last access update. This should be updated at the end of
        // of the datasource run, but for now, we will disable the cache and force it to always update here.
        // TODO : impelement the proper update of this at the end of the datasource poll cycle. See bug 15498.
        Long originatorID = null;
//        Long originatorID = originatorCache.get(originatorKey);
//        if (originatorID == null) {
        try {
            originatorID = ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().getOriginatorID(datasourceID, originatorKey, originatorDisplayName, true);
        }
        catch (FrameworkException e) {
            throw new FrameworkException("Failed to get originator ID from server", e);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Failed to get originator ID from server", e);
        }

        originatorCache.put(originatorKey, originatorID);
//        }

        return originatorID;
    }

    public void deleteOriginatorID(final Long originatorID, final String originatorKey) throws FrameworkException {
        if (logger.isDebugEnabled())
            logger.debug("deleteOriginatorID() called. Using originatorID: [" + originatorID + "] Originator Key:[" + originatorKey + "]");

        originatorCache.remove(originatorKey);

        try {
            ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().deleteOriginatorID(originatorID, originatorKey);
        }
        catch (RemoteException e) {
            throw new FrameworkException("Failed to delete originator ID from server", e);
        }
        catch (FrameworkException e) {
            throw new FrameworkException("Failed to delete originator ID from server", e);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Failed to delete originator ID from server", e);
        }
    }

    //TODO: Server Interface required and also for other dependent packages
    // As only one Model is maintained for all storages, we are having a method to report the device model
    public void reportNetAppDeviceModel(final BaseDataSource dataSource,
                              final BaseModelRoot netappModelRoot,
                              final boolean changed,
                              final Serializable context) throws FrameworkException {
        assert auditSession.isOpen();

        if (logger.isDebugEnabled())
            logger.debug(dataSource.getDataSourceName() + " - Closing audit 'Sending data to server'");

        // Check if data source was stopped. If it was, skip the update...
        if (!dataSource.getStatus().equals(DataSourceStatus.Aqcuiring) && !dataSource.getStatus().equals(DataSourceStatus.StandBy)) {
            logger.warn(dataSource.getDataSourceName() + " - Data source is not currently in Acquiring state (" + dataSource.getStatus() + ") skipping server update.");
            return;
        }

        if (changed || !sendOnlyUpdates) {
//            try {
                // Send the new XML back to the server
                //TODO: Need an interface from Server to update Model. Not sure how Server will parse the model file
                //ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().updateNerData(nerdataXML);
//            }
//            catch (AcquisitionUnitException e) {
//                throw new FrameworkException("Cannot update server (" + e.getMessage() + ")", e);
//            }
//
                auditSession.updateProgress("Sent DFM Device Model to server");
        }

    }

    public void reportNetAppData(final BaseDataSource dataSource,
                              final String originaorKey,
                              final String origintorDisplayName,
                              final StorageRoot netAppRoot,
                              final File netappDataFile,
                              final boolean changed,
                              final Serializable context) throws FrameworkException {
        assert auditSession.isOpen();

        if (logger.isDebugEnabled())
            logger.debug(dataSource.getDataSourceName() + " - Closing audit 'Sending data to server'");

        // Check if data source was stopped. If it was, skip the update...
        if (!dataSource.getStatus().equals(DataSourceStatus.Aqcuiring) && !dataSource.getStatus().equals(DataSourceStatus.StandBy)) {
            logger.warn(dataSource.getDataSourceName() + " - Data source is not currently in Acquiring state (" + dataSource.getStatus() + ") skipping server update.");
            return;
        }

        if (changed || !sendOnlyUpdates) {
            try {
                streamReportToServer(dataSource, origintorDisplayName, netAppRoot, netappDataFile);
            } catch (HttpRetryException hre) {
                // If we get an HttpRetryException, retry again 1 time only
                logger.warn(dataSource.getDataSourceName() + " - Failed to report data to server: " + hre.getMessage() + ", attempting 1 retry...");
                try {
                    streamReportToServer(dataSource, origintorDisplayName, netAppRoot, netappDataFile);
                } catch (HttpRetryException hre2) {
                    // Now it is a hard failure
                    throw new FrameworkException("Cannot update server (" + hre2.getMessage() + ")", hre2);
                }
            }
        }

        // Update the originator last access time
        try {
            ServiceManager.getInstance().getCommunicationManager().getAcquisitionFacadeSession().updateOriginatorLastAccessTime(netAppRoot.getOriginator().getOriginatorID());

            if (logger.isDebugEnabled())
                logger.debug(dataSource.getDataSourceName() + " - Update to originator access time");
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Cannot update server (" + e.getMessage() + ")", e);
        }

    }

    private void streamReportToServer(final BaseDataSource dataSource,
                                  final String origintorDisplayName,
                                  final StorageRoot netAppRoot,
                                  final File netappDataFile) throws HttpRetryException, FrameworkException {
        final long startTime = System.currentTimeMillis();
        InputStreamReader netappDataReader = null;
        BufferedWriter streamWriter = null;
        // OutputStreamWriter streamWriter = null;
        try {
            // Send the new XML back to the server
            final String codebase = System.getProperty("codebase");
            if (codebase == null) {
                throw new FrameworkException("'codebase' system attribute is missing from configuration");
            }
            final Long originatorID = netAppRoot.getOriginator().getOriginatorID();
            final String targetURLString = OriginatorUpdateClientUtils.createURL(codebase, Schema.NETAPP, String.valueOf(originatorID));
            final URL targetURL = new URL(targetURLString);
            final HttpURLConnection httpURLConnection = ServiceManager.getInstance().getCommunicationManager().openAcqAuthenticatedHttpConnection(targetURL);
            httpURLConnection.setDoOutput(true);
            httpURLConnection.setChunkedStreamingMode(65536);

            // Stream the data from a file to the URL.
            streamWriter = new BufferedWriter(new OutputStreamWriter(httpURLConnection.getOutputStream(), "UTF-8"));
            if (logger.isDebugEnabled()) {
                logger.info("Sending netappdata to: " + targetURLString);
            }

            InputStream fis = new FileInputStream(netappDataFile);
            netappDataReader = new InputStreamReader(fis, "UTF-8");
            // Specify large buffer size here: 65536
            // Too large a buffer exacerbates memory
            char[] buffer = new char[65536];
            int len = -1;
            while((len = netappDataReader.read(buffer)) > -1){
                streamWriter.write(buffer, 0, len);
                streamWriter.flush();
            }
            streamWriter.close();
            assertResponse(httpURLConnection);
        }
        catch (AcquisitionUnitException e) {
            throw new FrameworkException("Cannot update server (" + e.getMessage() + ")", e);
        }
        catch (MalformedURLException mue) {
            throw new FrameworkException("Cannot update server (" + mue.getMessage() + ")", mue);
        }
        catch (HttpRetryException hre) {
            // We sporadically encounter this failure:
            // java.net.HttpRetryException: cannot retry due to server authentication, in streaming mode
            // Looks like an issue on the server side:
            // Error during authenticate: java.lang.IllegalStateException: Security Context has not been set
            // The current theory is if we retry, we will not encounter this error.
            // Throw this exception to let the caller retry
            throw hre;
        }
        catch (IOException ioe) {
            throw new FrameworkException("Cannot update server (" + ioe.getMessage() + ")", ioe);
        }
        finally {
            if (netappDataReader != null) {
                try {
                    netappDataReader.close();
                } catch (Exception ignore) {}
            }
            if (streamWriter != null) {
                try {
                    streamWriter.close();
                } catch (Exception ignore) {}
            }

        }
        final long finishTime = System.currentTimeMillis();
        final long reportTime = finishTime - startTime;
        logger.info(dataSource.getDataSourceName() + " - Finished sending report to server (" + reportTime + " msecs)");
        if (origintorDisplayName != null) {
            auditSession.updateProgress("Sent " + origintorDisplayName + " data to server");
        }
    }

    @Override
    public void recordingSessionDone(final BaseDataSource dataSource,
                                       final PollingPackage aPackage,
                                       final File recordingSession,
                                       final boolean shouldReport) throws FrameworkException {
        try {
            if (recordingSession != null && shouldReport)  //don't send to server if foundation indicated not to report
                ServiceManager.getInstance().getLogsManager().sendZipToServer(ServiceManager.getInstance().getAcquisitionUnitManager().getAcquisitionUnit().getName(), recordingSession);
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to send ZIP file: " + e.getMessage(), e);
        }
    }


    public synchronized void cleanOriginatorCache() {
        originatorCache.clear();
    }

    public final String[] getOriginatorCache() {
        final String[] cache = new String[originatorCache.size()];

        int i = 0;
        for (final Map.Entry<String, Long> entry : originatorCache.entrySet())
            cache[i++] = entry.getKey() + " - " + entry.getValue();

        return cache;
    }

    private void assertResponse(final HttpURLConnection httpURLConnection) throws FrameworkException, AcquisitionUnitException{
        // Need to define errorReason here so that it may be logged.
        // Calling httpURLConnection.getResponseCode() throws IOException when an error is encountered.
        String failureReason = null;
        try {
            failureReason = httpURLConnection.getHeaderField("OCIE_ORIGINATOR_UPDATE_FAILURE_REASON");
            final InputStream errorStream = httpURLConnection.getErrorStream();
            final String errorString = getStringFromStream(errorStream);
            final String statusMessage = httpURLConnection.getResponseMessage();
            final InputStream inputStream = httpURLConnection.getInputStream();
            final String inputString = getStringFromStream(inputStream);
            final int statusCode = httpURLConnection.getResponseCode();
            if (logger.isDebugEnabled()) {
                logger.info(statusCode + " " + statusMessage);
                logger.info(inputString);
                logger.info(errorString);
            }
            OriginatorUpdateClientUtils.assertResponse(statusCode, statusMessage, inputString, errorString);
        } catch (HttpRetryException hre) {
            // TODO: throw an exception that causes Acq to attempt 1 single resend
        } catch (IOException ioe) {
            final String errorMessage = (failureReason != null)
                    ? failureReason
                    : ioe.getMessage();
            throw new FrameworkException("Cannot update server (" + errorMessage + ")", ioe);
        } catch (Exception e) {
            throw new FrameworkException("Cannot update server (" + e.getMessage() + ")", e);
        }
    }

    private String getStringFromStream(final InputStream stream) {
        if (stream == null) {
            return null;
        }
        BufferedReader streamReader = null;
        StringBuilder stringBuilder = new StringBuilder();
        String line;
        try {
            streamReader = new BufferedReader(new InputStreamReader(stream));
            while ((line = streamReader.readLine()) != null) {
                stringBuilder.append(line).append('\n');
            }
        } catch (Exception ignore) {
        } finally {
            if (streamReader != null) {
                try {
                    streamReader.close();
                } catch (Exception ignore) {}
            }
        }
        return stringBuilder.toString();
    }

    private void updateAcqPollStatus(final BaseDataSource dataSource, final PollingPackage aPackage, final boolean success) {
        if (dataSource.isStopped()) {
            return;
        }
        final NetAppFoundationConfiguration config = (NetAppFoundationConfiguration) aPackage.getConfiguration();
        final NetAppFoundationConfiguration.AcquisitionPollStatus acquisitionPollStatus = config.getAcquisitionPollStatus(aPackage.getPackageName());
        if (acquisitionPollStatus == null) {
            logger.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageCompleted "
                + dataSource.getDataSourceName() + " " + aPackage.getPackageName());
            return;
        }
        acquisitionPollStatus.setDataReported(success);
        final PackagePollCompleteStatus packageCompletePollStatus = acquisitionPollStatus.getPackagePollCompleteStatus(success);
        try {
            final AcquisitionStatusRemote acqStatusSession = ServiceManager.getInstance().
                            getCommunicationManager().getSession(AcquisitionStatusRemote.JNDI_NAME);
            if (acqStatusSession == null) {
                logger.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageCompleted "
                    + dataSource.getDataSourceName() + " " + aPackage.getPackageName());
                return;
            }
            acqStatusSession.packageCompleted(NetAppFoundationConfiguration.AcquisitionPollStatus.schema, packageCompletePollStatus);
        }
        catch (FrameworkException fe) {
            logger.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageCompleted "
                + dataSource.getDataSourceName() + " " + aPackage.getPackageName());
        } catch (ServerInterfaceException sie) {
            logger.error(dataSource.getDataSourceName() + " - Failed to update AcquisitionStatusRemote packageCompleted "
                + dataSource.getDataSourceName() + " " + aPackage.getPackageName());
        }
    }

    public long getServerTime() throws FrameworkException {
        return ServiceManager.getInstance().getCommunicationManager().getServerTime();
    }
}
