package com.onaro.sanscreen.acquisition.netappfoundation.datasource;

import com.onaro.sanscreen.acquisition.framework.FrameworkException;
import com.onaro.sanscreen.acquisition.framework.datasource.*;
import com.onaro.sanscreen.acquisition.framework.datasource.io.*;
import com.onaro.sanscreen.acquisition.framework.recordings.RecordingSample;
import com.onaro.sanscreen.acquisition.netappfoundation.datasource.config.NetAppFoundationConfiguration;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.StorageRoot;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.StreamingCRCRoot;
import com.onaro.util.CrcCalculator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.util.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Base data source for Foundation based packages.
 *
 * <p/>The implementor of this class has to override 3 methods: {@link #doStart()}, {@link #doStop()}, {@link #doFoundation}.
 */
public abstract class NetAppFoundationPackage<C extends NetAppFoundationConfiguration> extends BasePollingPackage<C> implements NetAppFoundationPackageMBean, DataSourceSampleListener {

    private static final Logger LOG = LoggerFactory.getLogger(NetAppFoundationPackage.class);

    public static final String PACKAGE = "netappfoundation";

    private static final String crlf = System.getProperty("line.seperator");

    private NetAppOriginatorProvider originatorProvider;

    private final List<NetAppFoundationUpdatesListener> updatesListeners = new LinkedList<NetAppFoundationUpdatesListener>();

    protected boolean checkIntegrity = true;

    private final NetAppClockSynchronizer clockSynchronizer;

    /**
     * Device specific context passed when reporting odata
     */
    private Serializable context = null;

    /**
     * Keeps track of the netappdata reports made in this sample. This will later be used to identify changes and to prevent
     * the acquisition from sending redundant data to the server
     */
    private Map<String, NetAppDataInfo> netappDataInfoByOriginatorKey = new HashMap<String, NetAppDataInfo>(50);

    /**
     * Keeps track of the originators reported in the current sample.
     */
    private final Set<Long> currentlyReportedOriginatorIds = new TreeSet<Long>();

    /**
     * Keeps track of the last error reported by this data source to prevent from multiple error recordings to be sent
     * through phone home. When this value is <code>null</code>, no error is recorded.
     */
    private Long lastErrorCRC = null;

    private boolean keepRecordingFiles = false;

    /**
     * Keep track of server response time poll publish.
     */
    private long ocieReportDuration = -1;

    /**
     * Cached server time at poll start
     */
    private long cachedServerTime = 0;

    /**
     * Constructs a data source instance
     *
     * <p/>Register for the datasource's life cycle events. This class will forward all event to the appropriate classes
     * (e.g. the {@link RecordingSample} and the {@link NetAppFoundationListener})
     *
     * @param dataSource
     * @param configuration
     * @param dataSourceLifeCycleListener A listener to receive DS life cycle events (to write audits, etc...)
     * @param dataSourceSampleListener A listener to receive sample events (to write audits, etc...)
     * @param foundationUpdatesListener A listener to receieve foundation updates
     * @param originatorProvider Originator ID provider
     */
    protected NetAppFoundationPackage(final BaseDataSource dataSource,
                                   final C configuration,
                                   final DataSourceLifeCycleListener dataSourceLifeCycleListener,
                                   final DataSourceSampleListener dataSourceSampleListener,
                                   final NetAppFoundationUpdatesListener foundationUpdatesListener,
                                   final NetAppOriginatorProvider originatorProvider,
                                   final NetAppClockSynchronizer clockSynchronizer) {
        super(dataSource, configuration, dataSourceSampleListener);

        // Register the parent listener to receive data source life cycle events
        dataSource.addDataSourceLifeCycleListener(dataSourceLifeCycleListener);

        // Register the parent listener to receive data source sample events
        // Make sure this class registers BEFORE the parent data source life cycle listener to be able to process things before they are being packed
        // But this class will not actually send zip to server
        dataSource.addDataSourceSampleListener(getPackageName(), 0, this);

        // Set the originator provider for this package.
        setOriginatorProvider(originatorProvider);

        this.clockSynchronizer = clockSynchronizer;

        updatesListeners.add(foundationUpdatesListener);

        this.logPrefix = "[" + getPackageName() + "] " + getConfiguration().getDataSourceName();
    }

    /**
     * @return device specific context passed when reporting odata
     */
    public final Serializable getContext() {
        return context;
    }

    /**
     * Sets device specific context passed when reporting odata
     *
     * @param context
     */
    public final void setContext(final Serializable context) {
        this.context = context;
    }

    public Set<Long> getCurrentlyReportedOriginatorIds() {
        return Collections.unmodifiableSet(currentlyReportedOriginatorIds);
    }

    public final RecordingSample getRecordingSample() {
        return getRecordingSession().getCurrentSample();
    }

    public final DSFileOutputStream createBinaryRecordingFile() throws IOException {
        return getRecordingSample().createBinaryRecordingFile(getConfiguration());
    }

    public final DSFileWriter createTextualOutputFile(final DataSourceOutputFileType type, final String custom, final String extension) throws IOException {
        return getRecordingSample().createTextualOutputFile(getConfiguration(), type, custom, extension);
    }

    public final DSFileWriter createTextRecordingFile() throws IOException {
        return getRecordingSample().createTextRecordingFile(getConfiguration());
    }

    public final DSFileWriter createTextRecordingFile(final String custom) throws IOException {
        return getRecordingSample().createTextRecordingFile(getConfiguration(), custom);
    }

    public final void addRecordingFile(final String fileName) throws FileNotFoundException {
        getRecordingSample().addRecordingFile(fileName);
    }

    public final void addRecordingStream(final DSFile stream) {
        getRecordingSample().addRecordingStream(stream);
    }

    @Override
    public String crawlNow() {
        if (!getConfiguration().isCrawlSupported()) {
            return "Crawl is not supported by this datasource";
        }
        getConfiguration().turnOnCrawlingOnce();
        this.getDataSource().run(this);
        return "Success";
    }

  /**
   * Performs the poll logic
   */
  @Override
  public final void pollNow( ) throws DataSourceErrorException, FrameworkException {
    pollNow( null );
  }

    /**
     * Performs the poll logic
     */
    @Override
    public final void pollNow( final Properties properties ) throws DataSourceErrorException, FrameworkException {
        cachedServerTime = getCurrentServerTime();
        try {
            try {
                doFoundation(dataSource, properties);
            }
            catch (SchemaException e) {
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, "Failed to create schema objects", null, e);
            }
        }
        catch (DataSourceErrorException e) {
            // Set the originator ID for this exception (if was not already set)
            if (e.getDeviceKey() == null) {
                e.setDeviceKey("General Device");
            }

            throw e;
        }
    }

    @Override
    public final void start() throws DataSourceErrorException, FrameworkException {
        context = null;

        doStart();
    }

    @Override
    public final void stop() throws DataSourceErrorException, FrameworkException {
        doStop();

        context = null;
    }

    /**
     * Initialization code. This method is called everytime the data source is started or when the configuration of the
     * data source changes or when license is changed.
     *
     * Static information (like configuration parameters) should be read by the configuration class (extending
     * {@link NetAppFoundationConfiguration}). Other package state like persistent sessions, etc... can be initialized in
     * this method. However, packages should try to remain as stateless as possible.
     *
     * @throws FrameworkException
     * @throws DataSourceErrorException
     */
    protected abstract void doStart() throws FrameworkException, DataSourceErrorException;

    /**
     * This method is called everytime the data source is stoped, or when the configuration of the data source changes
     * or when license is changed.
     *
     * Package implementations should perform any neccessary cleanup here (like closing openned persistent sessions, etc...).
     * Packages should try to remain as stateless as possible, minimizing the amount of cleanup required here.
     *
     * @throws FrameworkException
     * @throws DataSourceErrorException
     */
    protected abstract void doStop() throws FrameworkException, DataSourceErrorException;

    /**
     * Query foundation data.
     *
     * <p/>This method is called by the framework when it is time for this package to poll the device for foundation data.
     *
     * <p/>The framework allocates a dedicated thread (for this poll) and prevents from other foundation polls to occur
     * while this method is still executing.
     *
     * <p/>If an error occured during the polling of the device, the implementation should throw a {@link DataSourceErrorException}
     * exception. The framework will also catch any other unhandled exception thrown by the implementation and report about
     * it as an <code>Internal Error</code>.
     *
     * <p/>If the poll ended successfully and there is nothing to report, the method should just return. If a new
     * schame object hierarchy need to be reported, the implementation should use the {@link #reportNetAppData} method to send
     * this new report to the server.
     *
     * @param dataSource Datasource initiating this poll
     *
     * @throws SchemaException when an error occured during the creation of schema objects
     * @throws FrameworkException when an error occured with framework related mechanisms (e.g. communication problem).
     * @throws DataSourceErrorException when an error occurs during the sample (this error contains information needed by the framework to create an audit update)
     */
    protected abstract void doFoundation(final BaseDataSource dataSource, final Properties properties ) throws SchemaException, FrameworkException, DataSourceErrorException;

    /**
     * Allocates/returns an originator ID to match the given originaor key
     *
     * @param originatorKey
     *
     * @throws FrameworkException
     */
    public final synchronized long getOriginatorID(final String originatorKey) throws FrameworkException {
        return getOriginatorID(originatorKey, originatorKey);
    }

    /**
     * Allocates/returns originator ID to match the given originaor key
     *
     * @param originatorKey Key identifying originator ID
     * @param originatorDisplayName User friendly description to display in the originator audit screen
     *
     * @throws FrameworkException
     */
    public final synchronized long getOriginatorID(final String originatorKey, final String originatorDisplayName) throws FrameworkException {
        // Sanity check: A MOID data source should never try to obtain an originator ID with null key. The
        // null key is reserved for non-MOID data sources only!
        if (originatorKey == null && configuration.isMOID()) {
            throw new FrameworkException("A MOID datasource (" + dataSource.getDataSourceName() + ") is trying to get originator ID with null key");
        }

        return originatorProvider.getOriginatorID(dataSource.getDataSourceID(), originatorKey, originatorDisplayName != null ? originatorDisplayName : dataSource.getDataSourceName());
    }

    @Override
    public final String[] getOriginatorCache() {
        return originatorProvider.getOriginatorCache();
    }

    @Override
    public final void cleanOriginatorCache() {
        originatorProvider.cleanOriginatorCache();
    }

    @Override
    public final boolean isRetireUnreportedOriginators() {
        return true;
    }

    /**
     * Used  to report back netappdata (unmarshaled). This method will marshal the XML and then update the
     * listener.
     *
     * <p/>This method throws no exception because after a data source finished querying a device, it should not worry
     * about errors occuring during the report to the server. This is a task for the framework.
     */

    public final void reportNetAppData(final StorageRoot netappRoot, final String originatorDisplayName) throws DataSourceErrorException, FrameworkException {

        // If we're stopping right now, ignore this report!
        if (dataSource.isStopped()) {
            return;
        }

        try {
            // Check if this data changed from the last odata we reported for this originator
            NetAppDataInfo netAppDataInfo = netappDataInfoByOriginatorKey.get(netappRoot.getOriginatorKey());
            if (netAppDataInfo == null) {
                netAppDataInfo = new NetAppDataInfo();
                netappDataInfoByOriginatorKey.put(netappRoot.getOriginatorKey(), netAppDataInfo);
            }

            // add originator to the current list
            currentlyReportedOriginatorIds.add(netappRoot.getOriginator().getOriginatorID());

            // Write the XML to a file (this will also add it to the list of files generated in this sample)
            // Note:  We have to make file's name unique, otherwise we can run in situations that 2 files (odatas)
            // have the same name.  This situation could arise if we collect references of Root objects in order to
            // report them lately on.  For example:
            //
            //  for (final NetAppStorageRoot netappRoot : nerRoots)
            //     reportNetAppDeviceModel(netappRoot);
            //
            // Note: Do not use the "netappRoot.getInfo().getOriginatorKey()" because it can contain chars that are not
            // allowed at the time of creating the file.
            LOG.info("Saving netappData file for Originator:{}", netappRoot.getOriginator().getOriginatorID());
            final String custom = Long.toString(netappRoot.getOriginator().getOriginatorID());
            final DSFileCRCWriter dsFileCRCWriter = getRecordingSample().createTextualOutputCRCFile(getConfiguration(), DataSourceOutputFileType.netappdata, custom, "xml");

            marshalAndCRCNetAppData(netappRoot, dsFileCRCWriter);
            dsFileCRCWriter.close();
            LOG.info("Saved netappData file for Originator:{}", netappRoot.getOriginator().getOriginatorID());
            // Keep track of which file is generated where. This will later be used for simulating this environment.
            netAppDataInfo.setLastNetAppDataFileName(new File(dsFileCRCWriter.getFileName()).getName());

            final String netappDataFileName = dsFileCRCWriter.getFileName();
            final File netappDataFile = new File(netappDataFileName);

            // Perform integrity checks on the odata generated. We're doing that after files are saved.
            // If checks failed, we will fail the acquisition (reporting an error)
            if (checkIntegrity) {
                netappRoot.checkIntegrity();
            }

            // Check for changes. If this originator report contains no changes since the last report, don't report it to
            // the server.
            final boolean changed = netAppDataInfo.changed(dsFileCRCWriter.getAllAttributesCRC().getValue(), netappDataFile.length());
            LOG.info("Reporting storage {} for Originator:{} to server",
                    originatorDisplayName, netappRoot.getOriginator().getOriginatorID());
            // If data contains changes, we want to keep this recording!
            keepRecordingFiles |= changed;
            // Report this new data to anyone that is interested

            final long listenerStart = System.currentTimeMillis();
            for (final NetAppFoundationUpdatesListener listener : updatesListeners) {
                listener.reportNetAppData(dataSource,
                                          netappRoot.getOriginatorKey(),
                                          originatorDisplayName,
                                          netappRoot,
                                          netappDataFile,
                                          changed,
                                          context);
            }

            // For OCIE Performance Testing
            ocieReportDuration = System.currentTimeMillis() - listenerStart;
            LOG.info("Reporting storage {} for Originator:{} from server Done",
                    originatorDisplayName, netappRoot.getOriginator().getOriginatorID());
            dataSource.addToStatusHistory("Reported odata for " + originatorDisplayName);
        }
        catch (IntegrityException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INCONSISTENT_DATA, e.getMessage(), "Odata Integrity checks failed (" + e.getMessage() + ")", e);
        }
        catch (SchemaException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, null, e.getMessage(), e);
        }
        catch (IOException e) {
            throw new FrameworkException(e.getMessage(), e);
        }
    }

    /**
     * Save model data as xml in a file. The model is only persisted if isRecording() == true
     *
     * @param modelRoot the device model
     *
     * @throws DataSourceErrorException if an IntegrityException or SchemaException is caught
     * @throws FrameworkException if an io exception is caught
     */
    public final void recordModelData(final StreamingCRCRoot modelRoot) throws DataSourceErrorException, FrameworkException {

        // If we're stopping right now, ignore this report!
        if (dataSource.isStopped()) {
            return;
        }

        try {

            if (isRecording()) {
                // Use DSFileCRCWriter to stream XML to the file and avoid building entire XML in memory.
                final DSFileCRCWriter fw = getRecordingSample().createTextualOutputCRCFile(getConfiguration(), DataSourceOutputFileType.model, "", "xml");
                modelRoot.marshallAndCRC(fw);
                fw.close();
            }

            // Perform integrity checks on the model generated. We're doing that after files are saved.
            // If checks failed, we will fail the acquisition (reporting an error)
            if (checkIntegrity) {
                modelRoot.checkIntegrity();
            }
        }
        catch (IntegrityException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INCONSISTENT_DATA, e.getMessage(), "Device model integrity checks failed (" + e.getMessage() + ")", e);
        }
        catch (SchemaException e) {
            throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, null, e.getMessage(), e);
        }
        catch (IOException e) {
            throw new FrameworkException(e.getMessage(), e);
        }
    }


    private void setOriginatorProvider(final NetAppOriginatorProvider foundationListener) {
        this.originatorProvider = foundationListener;
    }

    /**
     * Converts the oData root object to an XML form
     *
     * <p/>Note: This method also saves the marshalled XML to the output directory and adds the file
     * name to the reported files for this sample.
     *
     * @param netAppData Root of o-data to marshal
     * @param dsFileCRCWriter the datasource file CRC writer to marshall to
     *
     * @throws SchemaException
     * @throws IOException
     */
    protected void marshalAndCRCNetAppData(final StorageRoot netAppData, final DSFileCRCWriter dsFileCRCWriter) throws SchemaException, IOException {
        // Mark the XML with the sample date
        netAppData.getInfo().setDate(Calendar.getInstance().getTime().toString());

        // Add OriginatorKey
        netAppData.getInfo().setOriginatorKey(netAppData.getOriginatorKey());

        // Now marshal the object heirarchy into an XML string
        netAppData.marshallAndCRC(dsFileCRCWriter);
    }

    private void marshalAndCRCNetAppModel(final StorageRoot netAppData, final DSFileCRCWriter dsFileCRCWriter) throws SchemaException, IOException {
        // Now marshal the object heirarchy into an XML string
        netAppData.marshallAndCRC(dsFileCRCWriter);
    }


    @Override
    public final void startingSample(final BaseDataSource dataSource, final PollingPackage aPackage) {

        keepRecordingFiles = false;

        // clear the set of currently reported originators
        currentlyReportedOriginatorIds.clear();

        // Foundation re-runs the same sample over and over !
        if (isSimulating()) {
            try {
                final DSSessionFileReader simulationSession = new DSSessionFileReader(getConfiguration().getSimulationSession());
                getConfiguration().setSimulationSession(simulationSession);
            }
            catch (IOException e) {
                LOG.error("{} - failed to reload simulation session", getLogPrefix(), e);
            }
        }

        getRecordingSession().reset();
        getRecordingSession().createSample();
    }

    @Override
    public final void sampleSuccessfull(final BaseDataSource baseDataSource, final PollingPackage aPackage) {
        // Zeroize the error state for this data source. This will make sure that the next time the data source fails, it
        // will keep the recording files for this data source.
        lastErrorCRC = null;
        getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.SUCCESS);
    }


    @Override
    public final void sampleFailed(final BaseDataSource baseDataSource, final PollingPackage aPackage, final DataSourceErrorException e) {
        // This sample failed. We would like to keep the recording files for this one in order to make it possible
        // to simulate this issue in the lab

        // If we're considering not to keep the recording files for this sample, check out if the error reported here
        // was already reported by this data source last time. This will allow us to minimize the number of identical
        // reports we're keeping in the log file and to minimize the size of the phone home report.
        // Calculate CRC on the new Data

        final long newCRC = e.getExceptionCRC();

        if (lastErrorCRC == null || lastErrorCRC != newCRC) {
            keepRecordingFiles = true;
        }

        lastErrorCRC = newCRC;

        //setting the status of the sample
        if (e instanceof AggregatedDataSourceErrorsException) {
            if (e.getType() == DataSourceErrorType.PARTIAL_SUCCESS) {
                getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.PARTIAL_SUCCESS);
                getRecordingSample().setSamplePartialSuccessInformation(((AggregatedDataSourceErrorsException) e).getTotalNumberOfOriginators(), ((AggregatedDataSourceErrorsException) e).getExceptions().size());
            } else {
                getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.FAILED);
            }
        } else {
            getRecordingSample().setSampleStatus(RecordingSample.SampleStatus.FAILED);
        }

    }

    @Override
    public final void sampleFailed(final BaseDataSource baseDataSource, final PollingPackage aPackage, final Throwable e) {
        // This sample failed. We would like to keep the recording files for this one in order to make it possible
        // to simulate this issue in the lab

        // If we're considering not to keep the recording files for this sample, check out if the error reported here
        // was already reported by this data source last time. This will allow us to minimize the number of identical
        // reports we're keeping in the log file and to minimize the size of the phone home report.
        // Calculate CRC on the new Data
        final StringBuilder sb = new StringBuilder(1000);
        sb.append(baseDataSource.getDataSourceName()).append(" - ").append(e.getMessage()).append(crlf);

        DataSourceErrorException.printStackStraceToBuffer(sb, e);

        final long newCRC = CrcCalculator.calculateCRC(sb);

        if (lastErrorCRC == null || lastErrorCRC != newCRC) {
            keepRecordingFiles = true;
        }

        lastErrorCRC = newCRC;
    }

    @Override
    public final void finishedSample(final BaseDataSource dataSource, final PollingPackage aPackage) throws FrameworkException {

        // Make sure crawling is done only once!
        if (configuration.isCrawling()) {
            // Crawl support: If the data source was activated in crawl mode, turn it off. Crawl
            // should be activated only once!
            configuration.turnOffCrawling();

            // When crawling, we always want to keep the generated recording file (this is what crawl is all about)
            keepRecordingFiles = true;
        }

        packNetAppdataInfo();

        // Close all the recordings generated for this sample (send them to server if they are interesting...)
        try {
            if (keepRecordingFiles) {
                final File recordingSession = getRecordingSession().buildRecordingFile();
                getDataSource().recordingSessionDone(this, recordingSession, true);
            }
        }
        catch (IOException e) {
            throw new FrameworkException("Failed to send ZIP file: " + e.getMessage(), e);
        }
        finally {
            getRecordingSession().reset();
        }
    }


    @Override
    public void recordingSessionDone(BaseDataSource dataSource, PollingPackage aPackage, File recordingSession, boolean shouldReport) throws FrameworkException {
        // Do nothing ! Relevant FoundationListener that implements DataSourceSampleListener will do the work of actually sending the recording to server !
    }

    /**
     * Fill sample description with a odata file name to originator ID mapping
     */
    private void packNetAppdataInfo() {
        for (final Map.Entry<String, NetAppDataInfo> entry : netappDataInfoByOriginatorKey.entrySet()) {
            // A MOID data source cannot report the empty (null) OID!
            assert entry.getKey() == null && !getConfiguration().isMOID();

            final String key = (entry.getKey() != null ? entry.getKey() : "");
            getRecordingSession().getCurrentSample().addFileDescription(entry.getValue().getLastNetAppDataFileName(), "system.originator.key", key);
        }
    }

    public final NetAppDataInfo getNetAppDataInfo(final String originatorKey) {
        return netappDataInfoByOriginatorKey.get(originatorKey);
    }

    /**
     * Called by anyone that wishes to listen to foundation updates
     *
     * @param updatesListener Listener interface for reporting back updates
     */
    public final void addUpdatesListener(final NetAppFoundationUpdatesListener updatesListener) {
        updatesListeners.add(updatesListener);
    }

    /**
     * This class saved CRC state for a data source foundation sample. The CRC is calculated based on the generated
     * OData XML.
     */
    static final class NetAppDataInfo {
        private long lastUpdateCRC = 0;
        private long lastUpdateLength = 0;
        private String lastNetAppDataFileName = null;

        public NetAppDataInfo() {
        }

        public String getLastNetAppDataFileName() {
            return lastNetAppDataFileName;
        }

        public void setLastNetAppDataFileName(final String lastNetAppDataFileName) {
            this.lastNetAppDataFileName = lastNetAppDataFileName;
        }

        /**
         * Using CRC comparison to the last <b>good</b> report, detect if this XML contains any changes in the environment.
         *
         * <p/>Note that the CRC is being calculated only when a datasource has successfully finished reporting about a device and
         * called the {@link NetAppFoundationUpdatesListener#reportNetAppData} method. This ensures us that errors during the acquisition will
         * not cause us to save redundant reports again and again.
         *
         * <p/>Note: Since the <Info> element contains dynamic information that changes from sample to sample, we have
         * to remove it before calculating any CRC, otherwise the CRC will always indicate change.
         */
        public boolean changed(final long netAppDataCRC, final long netAppDataLength) {
            // Truncate the dynamic part of the oData XML. It ends with the <Info> element
            // final String staticOData = odataXML.replaceFirst("<Info .*/>", "");
            final boolean changed = (netAppDataLength != lastUpdateLength) || (netAppDataCRC != lastUpdateCRC);
            lastUpdateCRC = netAppDataCRC;
            lastUpdateLength = netAppDataLength;
            return changed;
        }
    }

    /**
     * Get the current server time
     * @return the current server time
     */
    public final long getCurrentServerTime() {
        try {
            return clockSynchronizer.getServerTime();
        }
        catch (FrameworkException fe) {
            LOG.warn("Could not retrieve server time.  Using local time for sample.");
            return System.currentTimeMillis();
        }
    }

    /**
     * Get the cached server poll start time
     * @return the cached server poll start time
     */
    public long getCachedServerTime(){
        return cachedServerTime;
    }

    public long getOcieReportDuration() {
        return ocieReportDuration;
    }
}
