package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;

@OCIETable("export_rule")
@OCIEResourceType("export_rule")
public class ExportRule extends BaseCRCXMLNode<ExportPolicy> {

    public static final class ExportRuleKey extends KeyBase<ExportRuleKey.Field> {

        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "vserver-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.vserver-info.uuid"
            )
            vserver_uuid,

            @OCIEZapiSource(
                sourceZapi = "export-policy-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.export-policy-info.policy-id"
            )
            export_policy_id,

            @OCIEZapiSource(
                sourceZapi = "export-rule-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.export-rule-info.rule-index"
            )
            rule_id
        }

        public ExportRuleKey(final String clusterUUID, final String vServerUUID, final String exportPolicyID, final String ruleID) {
            super("export_rule", new EnumMap<Field, String>(Field.class){{
                put(Field.cluster_uuid, clusterUUID);
                put(Field.vserver_uuid, vServerUUID);
                put(Field.export_policy_id, exportPolicyID);
                put(Field.rule_id, ruleID);
            }});
        }

        @Override
        @OCIEUUID(uuidDescr = "vserver_uuid.export_policy_id.rule_id")
        public String getOcieUUID() {
            final String clusterUUID = getID(Field.cluster_uuid);
            final String vServerUUID = getID(Field.vserver_uuid);
            final String fqVServerUUID = VServer.VServerKey.getOCIEVServerUUID(clusterUUID, vServerUUID);
            final String exportPolicyID = getID(Field.export_policy_id);
            final String ruleID = getID(Field.rule_id);
            return fqVServerUUID + '.' + exportPolicyID + '.' + ruleID;
        }
    }

    private final ExportRuleKey key;

    @OCIEZapiSource(
        sourceZapi = "export-rule-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.export-rule-info.client-match"
    )
    @OCIEValue(
        ocieColumn = "clientMatch"
    )
    private String clientMatch;

    @OCIEZapiSource(
        sourceZapi = "export-rule-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.export-rule-info.client-match"
    )
    @OCIEValue(
        ocieColumn = "accessProtocol"
    )
    private String accessProtocol;

    @OCIEZapiSource(
        sourceZapi = "export-rule-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.export-rule-info.ro-rule.security-flavor"
    )
    @OCIEValue(
        ocieColumn = "roRule",
        notes = "CSV list"
    )
    private String roRule;

    @OCIEZapiSource(
        sourceZapi = "export-rule-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.export-rule-info.rw-rule.security-flavor"
    )
    @OCIEValue(
        ocieColumn = "rwRule",
        notes = "CSV list"
    )
    private String rwRule;

    @OCIEZapiSource(
        sourceZapi = "export-rule-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.export-rule-info.super-user-security.security-flavor"
    )
    @OCIEValue(
        ocieColumn = "superUserSecurity",
        notes = "CSV list"
    )
    private String superUserSecurity;

    ExportRule(final ExportPolicy parent, final VServer.VServerKey vServerKey, final String ruleIndex) {
        super(parent, ExportRule.class.getSimpleName());
        final String clusterUUID = parent.getKey().getID(ExportPolicy.ExportPolicyKey.Field.cluster_uuid);
        final String shortVServerUUID = vServerKey.getID(VServer.VServerKey.Field.uuid);
        this.key = new ExportRuleKey(clusterUUID, shortVServerUUID, parent.getPolicyID(), ruleIndex);
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public ExportRuleKey getKey() {
        return key;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getRuleIndex() {
        return key.getID(ExportRuleKey.Field.rule_id);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getClientMatch() {
        return clientMatch;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getAccessProtocol() {
        return accessProtocol;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getRoRule() {
        return roRule;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getRwRule() {
        return rwRule;
    }

    public void setClientMatch(final String clientMatch) {
        this.clientMatch = clientMatch;
    }

    public void setAccessProtocol(final String accessProtocol) {
        this.accessProtocol = accessProtocol;
    }

    public void setRoRule(final String roRule) {
        this.roRule = roRule;
    }

    public void setRwRule(final String rwRule) {
        this.rwRule = rwRule;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getSuperUserSecurity() {
        return superUserSecurity;
    }


    public void setSuperUserSecurity(final String superUserSecurity) {
        this.superUserSecurity = superUserSecurity;
    }

    /**
     * @see <a href="http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/DOT/mainN/web_ontap_cluster/exports/index.html#access-protocol">access-protocol</a>
     */
    public static enum AccessProtocol {
        NOT_MAPPED,
        ANY,        // Any
        NFS2,       // NFS Version 2
        NFS3,       // NFS Version 3
        NFS,        // Any Version of NFS
        CIFS,       // CIFS
        NFS4,       // NFS Version 4
        FLEXCACHE,  // Flexcache
    }

    /**
     * @see <a href="http://web.netapp.com/engineering/design-depot/appliance-mgmt/zephyr/zapidoc.cache/DOT/mainN/web_ontap_cluster/exports/index.html#security-flavor">security-flavor</a>
     */
    public static enum SecurityFlavor {
        NOT_MAPPED,
        ANY,        // Any
        NONE,       // Anonymous Access Allowed If Security Type Not Already Listed
        NEVER,      // Never
        KRB5,       // Kerberos 5 Authentication
        NTLM,       // CIFS NTLM
        SYS,        // NFS AUTH_SYS
        SPINAUTH,   // SpinAuth
    }


    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "RuleIndex", key.getID(ExportRuleKey.Field.rule_id));
        addRequiredAttributeAndCRC(writer, "ClientMatch", this.clientMatch);
        addRequiredAttributeAndCRC(writer, "AccessProtocol", this.accessProtocol);
        addRequiredAttributeAndCRC(writer, "RORule", this.roRule);
        addRequiredAttributeAndCRC(writer, "RWRule", this.rwRule);
        addRequiredAttributeAndCRC(writer, "SuperUserSecurity", this.superUserSecurity);
    }
}
