package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.*;

import java.io.IOException;
import java.util.EnumMap;


@OCIETable("iscsi_security_entry")
@OCIEResourceType("iscsi_security_entry")
public class ISCSISecurityEntry extends BaseCRCXMLNode<VServer> {

    public static final class ISCSISecurityEntryKey extends KeyBase<ISCSISecurityEntryKey.Field> {

        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "vserver-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.vserver-info.uuid"
            )
            vserver_uuid,

            @OCIEZapiSource(
                sourceZapi = "iscsi-initiator-auth-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.iscsi-security-entry-info.initiator"
            )
            initiator
        }

        public ISCSISecurityEntryKey(final String clusterUUID, final String vServerUUID, final String initiator) {
            super("iscsi_security_entry",
                  new EnumMap<Field, String>(Field.class) {{
                        put(Field.cluster_uuid, clusterUUID);
                        put(Field.vserver_uuid, vServerUUID);
                        put(Field.initiator, initiator);
                  }});
        }

        @Override
        @OCIEUUID(uuidDescr = "cluster_uuid.vserver_uuid.initiator")
        public String getOcieUUID() {
            return getFQID();       // exposed UUID SHOULD include cluster UUID
        }
    }

    private final ISCSISecurityEntryKey key;

    @OCIEZapiSource(
        sourceZapi = "iscsi-initiator-auth-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-security-entry-info.auth-type"
    )
    @OCIEValue(
        ocieColumn = "authenticationType"
    )
    private ISCSISecurityEntryAuthenticationType authenticationType;

    @OCIEZapiSource(
        sourceZapi = "iscsi-initiator-auth-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-security-entry-info.outbound-user-name"
    )
    @OCIEValue(
        ocieColumn = "outboundUserName"
    )
    private String outboundUsername;

    @OCIEZapiSource(
        sourceZapi = "iscsi-initiator-auth-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-security-entry-info.user-name"
    )
    @OCIEValue(
        ocieColumn = "outboundUserName"
    )
    private String username;

    // JSD TODO get rid of clusterUUID param
    ISCSISecurityEntry(final VServer parentVServer, final String clusterUUID, final String initiator) {
        super(parentVServer, ISCSISecurityEntry.class.getSimpleName());
        final String vServerUUID = parentVServer.getKey().getZapiUUID();
        this.key = new ISCSISecurityEntryKey(clusterUUID, vServerUUID, initiator);
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public ISCSISecurityEntryKey getKey() {
        return key;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public ISCSISecurityEntryAuthenticationType getAuthenticationType() {
        return authenticationType;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getInitiator() {
        return key.getID(ISCSISecurityEntryKey.Field.initiator);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getOutboundUsername() {
        return outboundUsername;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getUsername() {
        return username;
    }

    public void setAuthenticationType(final ISCSISecurityEntryAuthenticationType authenticationType) {
        this.authenticationType = authenticationType;
    }


    public void setOutboundUsername(final String outboundUsername) {
        this.outboundUsername = outboundUsername;
    }

    public void setUsername(final String username) {
        this.username = username;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public enum ISCSISecurityEntryAuthenticationType {
        NOT_MAPPED,
        CHAP,
        NONE,
        DENY
    }


    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "AuthenticationType", authenticationType);
        addRequiredAttributeAndCRC(writer, "Initiator", key.getID(ISCSISecurityEntryKey.Field.initiator));
        addRequiredAttributeAndCRC(writer, "OutboundUserName", outboundUsername);
        addRequiredAttributeAndCRC(writer, "Username", username);
    }
}
