package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;

@OCIETable("license")
@OCIEResourceType("license")
public class License extends BaseCRCXMLNode<Cluster> {

    public static final class LicenseKey extends KeyBase<LicenseKey.Field> {
        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "license-v2-list-info",
                zapiResponsePath = "netapp.results.licenses.license-v2-info.package"
            )
            package_name,

            @OCIEZapiSource(
                sourceZapi = "license-v2-list-info",
                zapiResponsePath = "netapp.results.licenses.license-v2-info.serial-number"
            )
            serial_number
        }

        public LicenseKey(final String clusterUUID, final String packageName, final String serialNumber) {
            super("license", new EnumMap<Field, String>(Field.class) {{
                put(Field.cluster_uuid, clusterUUID);
                put(Field.package_name, packageName);
                put(Field.serial_number, serialNumber);
            }});
        }

        @Override
        @OCIEUUID(uuidDescr = "cluster_uuid.package_name.serial_number")
        public String getOcieUUID() {
            return getFQID();           // OCIE UUID has cluster UUID
        }
    }

    private final LicenseKey key;
    /**
     * Name of the controller or Cluster that owns the serial number.
     */
    @OCIEZapiSource(
        sourceZapi = "license-v2-list-info",
        zapiResponsePath = "netapp.results.licenses.license-v2-info.owner"
    )
    @OCIEValue(
        ocieColumn = "ownerName"
    )
    private String owner;

    /**
     * The entity type of the owner. In RR, this will always be CLUSTER.
     */
    @OCIEZapiSource(
            sourceZapi = "license-v2-list-info",
            zapiResponsePath = "netapp.results.licenses.license-v2-info.type"
    )
    @OCIEValue(
        ocieColumn = "ownerType"
    )
    private OwnerType ownerType;

    /**
     * The license type of the owner.
     */
    @OCIEZapiSource(
            sourceZapi = "license-v2-list-info",
            zapiResponsePath = "netapp.results.licenses.license-v2-info.type"
    )
    @OCIEValue(
            ocieColumn = "licenseType"
    )
    private LicenseType licenseType;


    /**
     * The license type of the owner - raw value.
     */
    @OCIEZapiSource(
            sourceZapi = "license-v2-list-info",
            zapiResponsePath = "netapp.results.licenses.license-v2-info.type"
    )
    @OCIEValue(
            ocieColumn = "licenseType"
    )
    private String licenseTypeRaw;

    /**
     * Owner node when {@link #ownerType} is {@link OwnerType#NODE}.
     */
    @OCIEZapiSource(
        sourceZapi = "license-v2-list-info",
        zapiResponsePath = "netapp.results.licenses.license-v2-info.owner"
    )
    @OCIEValue(
        ocieColumn = "ownerNode"
    )
    private Node ownerNode = null;

    /**
     * License expiration time in number of seconds since January 1, 1970.
     */
    @OCIEZapiSource(
        sourceZapi = "license-v2-list-info",
        zapiResponsePath = "netapp.results.licenses.license-v2-info.expiration-time"
    )
    @OCIEValue(
        ocieColumn = "expirationTime"
    )
    private Integer expirationTime;

    /**
     * Description of the licensed package.
     */
    @OCIEZapiSource(
        sourceZapi = "license-v2-list-info",
        zapiResponsePath = "netapp.results.licenses.license-v2-info.description"
    )
    @OCIEValue(
        ocieColumn = "description"
    )
    private String description;


    License(final Cluster parent, final String packageName, final String serialNumber) {
        super(parent, License.class.getSimpleName());
        this.key = new LicenseKey(parent.getClusterUUID(), packageName, serialNumber);
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public LicenseKey getKey() {
        return key;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getPackageName() {
        return key.getID(LicenseKey.Field.package_name);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getSerialNumber() {
        return key.getID(LicenseKey.Field.serial_number);
    }

    public String getOwner() {
        return owner;
    }

    public void setOwner(final String owner) {
        this.owner = owner;
    }

    public OwnerType getOwnerType() {
        return ownerType;
    }

    public void setOwnerType(final OwnerType ownerType) {
        this.ownerType = ownerType;
    }

    public LicenseType getLicenseType() {
        return licenseType;
    }

    public void setLicenseType(final LicenseType licenseType) {
        this.licenseType = licenseType;
    }

    public String getLicenseTypeRaw() {
        return licenseTypeRaw;
    }

    public void setLicenseTypeRaw(final String licenseTypeRaw) {
        this.licenseTypeRaw = licenseTypeRaw;
    }


    public Node getOwnerNode() {
        return ownerNode;
    }

    public void setOwnerNode(final Node ownerNode) {
        this.ownerNode = ownerNode;
    }

    public Integer getExpirationTime() {
        return expirationTime;
    }

    public void setExpirationTime(final Integer expirationTime) {
        this.expirationTime = expirationTime;
    }

    public String getDescription() {
        return description;
    }

    public void setDescription(final String description) {
        this.description = description;
    }

    public static enum OwnerType {
        NODE,
        CLUSTER,
        UNKNOWN,
    }

    public static enum LicenseType {
        NOT_MAPPED,
        SITE,
        LICENSE,
        DEMO,
    }

    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "Package", getPackageName());
        addRequiredAttributeAndCRC(writer, "SerialNumber", getSerialNumber());
        addRequiredAttributeAndCRC(writer, "OwnerName", getOwner());
        addRequiredAttributeAndCRC(writer, "OwnerType", getOwnerType());
        if (getOwnerNode() != null) {
            addRequiredAttributeAndCRC(writer, "OwnerNodeUUID", getOwnerNode().getKey().getID(Node.NodeKey.Field.uuid));
        }
        addRequiredAttributeAndCRC(writer, "ExpirationTime", getExpirationTime());
        addRequiredAttributeAndCRC(writer, "Description", getDescription());

        addRequiredAttributeAndCRC(writer, "LicenseType", getLicenseType());
        addRequiredAttributeAndCRC(writer, "LicenseTypeRaw", getLicenseTypeRaw());
    }
}
