package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;
import java.util.EnumSet;

@OCIETable("portset")
@OCIEResourceType("portset")
public class Portset extends BaseCRCXMLNode<VServer> {

    public final static class PortSetKey extends KeyBase<PortSetKey.Field> {

        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "vserver-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.vserver-info.uuid"
            )
            vserver_uuid,

            // portset-get-iter
            @OCIEZapiSource(
                sourceZapi = "portset-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.portset-info.portset-name"
            )
            portset_name
        }


        public PortSetKey(final String clusterUUID, final String vServerUUID, final String portSetUUID) {
            super("portset", new EnumMap<Field, String>(Field.class) {{
                put(Field.cluster_uuid, clusterUUID);
                put(Field.vserver_uuid, vServerUUID);
                put(Field.portset_name, portSetUUID);
            }}, EnumSet.of(Field.portset_name));                      // These EnumSet<E> emptyAbleFields fields are allowed to be empty but not null
        }

        @Override
        @OCIEUUID(uuidDescr = "vserver_uuid.portset_name")
        public String getOcieUUID() {
            final String clusterUUID = getID(Field.cluster_uuid);
            final String vServerUUID = getID(Field.vserver_uuid);
            final String fqVServerUUID = VServer.VServerKey.getOCIEVServerUUID(clusterUUID, vServerUUID);
            final String portSetUUID = getID(Field.portset_name);
            return fqVServerUUID + '.' + portSetUUID;
        }
    }

    @OCIEValue(
        ocieColumn = "resourceKey"
    )
    private final PortSetKey key;

    @OCIEZapiSource(
        sourceZapi = "portset-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.portset-info.portset-port-info.portset-port-name"
    )
    private final PortsetPorts portsetPorts;

    @OCIEZapiSource(
        sourceZapi = "portset-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.portset-info.initiator-group-info.initiator-group-name"
    )
    private final PortsetInitiatorGroups initiatorGroups;

    @OCIEZapiSource(
        sourceZapi = "portset-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.portset-info.portset-type"
    )
    @OCIEValue(
        ocieColumn = "type"
    )
    private Type type;

    @OCIEZapiSource(
        sourceZapi = "portset-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.portset-info.portset-type"
    )
    @OCIEValue(
        ocieColumn = "typeRaw"
    )
    private String typeRaw;

    Portset(final VServer parent, final String portSetUUID) {
        super(parent, "Portset");
        final VServer.VServerKey parentKey = parent.getKey();
        final String clusterUUID = parentKey.getID(VServer.VServerKey.Field.cluster_uuid);
        final String vServerUUID = parentKey.getZapiUUID();
        this.key = new PortSetKey(clusterUUID, vServerUUID, portSetUUID);
        this.portsetPorts = new PortsetPorts(this);
        this.initiatorGroups = new PortsetInitiatorGroups(this);
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public PortSetKey getKey() {
        return key;
    }

    public String getName() {
        return key.getID(PortSetKey.Field.portset_name);
    }

    public Type getType() {
        return type;
    }

    public void setType(final Type type) {
        this.type = type;
    }

    @SuppressWarnings({"unused declaration"})
    public String getTypeRaw() {
        return typeRaw;
    }

    public void setTypeRaw(final String typeRaw) {
        this.typeRaw = typeRaw;
    }

    public PortsetPort createPortsetPort(final String portName) {
        return new PortsetPort(this, portName);
    }

    public PortsetPorts getPortsetPorts() {
        return portsetPorts;
    }

    public PortsetInitiatorGroup createPortsetInitiatorGroup(final String iGroupUUID) {
        return new PortsetInitiatorGroup(this, iGroupUUID);
    }

    public PortsetInitiatorGroups getPortsetInitiatorGroups() {
        return initiatorGroups;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public PortsetInitiatorGroups getInitiatorGroups() {
        return initiatorGroups;
    }

    public static enum Type {
        NOT_MAPPED,
        ISCSI,
        FCP,
        MIXED,
    }


    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", key);
        addRequiredAttributeAndCRC(writer, "UUID", key.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "Name", key.getID(PortSetKey.Field.portset_name));
        addRequiredAttributeAndCRC(writer, "Type", type);
        addRequiredAttributeAndCRC(writer, "TypeRaw", typeRaw);
    }

    @Override
    protected boolean hasChildren() {
        return !portsetPorts.isEmpty() || !initiatorGroups.isEmpty();
    }

    @Override
    protected void marshallChildrenToFileAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNodeAndCRC(cache, writer, "PortsetPorts", portsetPorts.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "PortsetIgroupRefs", initiatorGroups.values(), tabs);
    }
}
