package com.onaro.sanscreen.acquisition.netappschema.netappdata;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;
import com.onaro.sanscreen.acquisition.netappschema.annotations.*;
import com.onaro.sanscreen.acquisition.schema.BaseCRCXMLNode;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.util.EnumMap;
import java.util.List;

/**
 * NetApp OCIE Foundation Schema implementation of a VServer and its attributes.
 */
@OCIETable("vserver")
@OCIEResourceType("vserver")
public class VServer extends BaseCRCXMLNode<Cluster> {

    /**
     * Globally unique key based on uuid.  For Legacy CSevers (Cluster Admin VServers),
     * uuid as returned by ZAPI is 00000000-0000-0000-0000-000000000000.
     *
     * We must prefix this CSever UUID with the cluster UUID for global uniqueness.
     * Later ZAPI versions provide a globally unique CSever UUID, in which case no prefix is used.
     *
     * Any UUID other than 00000000-0000-0000-0000-000000000000 is assumed to be globally unique
     */
    public final static class VServerKey extends KeyBase<VServerKey.Field> {

        public static final String CSERVER_UUID = "00000000-0000-0000-0000-000000000000";

        public static enum Field {
            @OCIEZapiSource(
                sourceZapi = "cluster-identity-get",
                zapiResponsePath = "netapp.results.attributes.cluster-identity-info.cluster-uuid"
            )
            cluster_uuid,

            @OCIEZapiSource(
                sourceZapi = "vserver-get-iter",
                zapiResponsePath = "netapp.results.attributes-list.vserver-info.uuid"
            )
            uuid
        }

        public VServerKey(final String clusterUUID, final String vServerUUID) {
            super("vserver",
                new EnumMap<Field, String>(Field.class) {{
                    put(Field.cluster_uuid, clusterUUID);
                    put(Field.uuid, vServerUUID);
                }});
        }

        public String getZapiUUID() {
            return getID(Field.uuid);
        }

        @Override
        @OCIEUUID(uuidDescr = "uuid")
        public String getOcieUUID() {
            final String clusterUUID = getID(Field.cluster_uuid);
            final String uuid = getID(Field.uuid);
            return getOCIEVServerUUID(clusterUUID, uuid);
        }

        public static String getOCIEVServerUUIDOrNull(final VServerKey vServerKey) {
            if (vServerKey == null) {
                return null;
            } else {
                return vServerKey.getOcieUUID();
            }
        }

        /**
         * Gets the OCIE "fully qualified" form for the vServer UUID.   The OCIE fully qualified form of a vServer
         * ID is generally just the vServer UUID.  For Admin/CServer UUIDs that return "00000000-0000-0000-0000-000000000000"
         * as there vServer UUID, then we pre-pend the cluster UUID.
         * This is used by children of VServers (Lif, QTree, Quota....) that only
         * have the short form of the vServer UUID in there own Key, but need to express the fully qualified form
         * when marshalling their own UUID, but NOT when marshalling their own ResourceKey.
         * @param clusterUUID the cluster UUID
         * @param shortOrLongVServerUUID the vServer UUID, which may be one or two tokens long
         * @return the OCIE "fully qualified" form for the vServer UUID
         */
        public static String getOCIEVServerUUID(final String clusterUUID, final String shortOrLongVServerUUID) {
            return (CSERVER_UUID.equals(shortOrLongVServerUUID))
                    ? clusterUUID + '.' + shortOrLongVServerUUID        // short form of 0000... requires cluster prefix
                    : shortOrLongVServerUUID;                           // not a cserver UUID or already fully qualified
        }
    }

    // vserver-info from vserver-get-iter
    private final VServerKey vServerKey;              // uuid

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.vserver-name"
    )
    @OCIEValue(
        ocieColumn = "name"
    )
    private String name;                              // vserver-name

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.vserver-type"
    )
    @OCIEValue(
        ocieColumn = "type",
        notes = "derived enum"
    )
    private Type type;                                // vserver-type

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.vserver-type"
    )
    @OCIEValue(
        ocieColumn = "typeRaw"
    )
    private String typeRaw; 

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.vserver-subtype"
    )
    @OCIEValue(
        ocieColumn = "subtype",
        notes = "derived enum"
    )
    private Subtype subtype;                       

    @OCIEZapiSource(
       sourceZapi = "vserver-get-iter",
       zapiResponsePath = "netapp.results.attributes-list.vserver-info.vserver-subtype"
    )
    @OCIEValue(
        ocieColumn = "subtypeRaw"
    )
    private String subtypeRaw;                       

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.state"
    )
    @OCIEValue(
        ocieColumn = "state",
        notes = "derived enum"
    )
    private State state;                             

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.state"
    )
    @OCIEValue(
        ocieColumn = "stateRaw"
    )
    private String stateRaw;   
    
    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.operational-state"
    )
    @OCIEValue(
        ocieColumn = "operationalState",
        notes = "derived enum"
    )
    private OperationalState operationalState;                             

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.operational-state"
    )
    @OCIEValue(
        ocieColumn = "operationalStateRaw"
    )
    private String operationalStateRaw;

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.allowed-protocols.protocol"
    )
    @OCIEValue(
        ocieColumn = "AllowedProtocols"
    )
    private String allowedProtocolsCSVString;         // allowed-protocols

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.allowed-protocols.protocol"
    )
    private List<Protocol> allowedProtocols;          // allowed-protocols

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.disallowed-protocols.protocol"
    )
    @OCIEValue(
        ocieColumn = "DisallowedProtocols"
    )
    private String disAllowedProtocolsCSVString;      // disallowed-protocols

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.disallowed-protocols.protocol"
    )
    private List<Protocol> disAllowedProtocols;       // disallowed-protocols

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.snapshot-policy"
    )
    @OCIEValue(
        ocieColumn = "snapshotPolicyId"
    )
    private SnapshotPolicy.SnapshotPolicyKey snapshotPolicyKey; // snapshot-policy

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.root-volume"
    )
    @OCIEValue(
        ocieColumn = "rootVolumeName"
    )
    private String rootVolumeName;

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.max-volumes"
    )
    @OCIEValue(
        ocieColumn = "maxVolumes"
    )
    private Integer maxVolumes;


    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.language"
    )
    @OCIEValue(
        ocieColumn = "language",
        notes = "derived enum"
    )
    private Language language;

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.language"
    )
    @OCIEValue(
        ocieColumn = "languageRaw"
    )
    private String languageRaw;

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.is-repository-vserver"
    )
    @OCIEValue(
        ocieColumn = "isRepository"
    )
    private Boolean isRepository;
    
    @OCIEZapiSource(
    		sourceZapi = "vserver-get-iter",
    		zapiResponsePath = "netapp.results.attributes-list.vserver-info.is-config-locked-for-changes"
    )
    @OCIEValue(
    		ocieColumn = "isConfigLockedForChanges"
    )
    private Boolean isConfigLockedForChanges;

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.name-mapping-switch.nmswitch"
    )
    @OCIEValue(
        ocieColumn = "nmSwitch"
    )
    private String nmSwitch;

    // CIFS Service
    @OCIEZapiSource(
        sourceZapi = "cifs-server-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-server-config.administrative-status"
    )
    @OCIEValue(
        ocieColumn = "isCifsEnabled",
        notes = "For RR, CIFS Admin status is not defined, so if we discover a CIFS server we will report it as enabled."
    )
    private Boolean isCifsEnabled;

    @OCIEZapiSource(
        sourceZapi = "cifs-server-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-server-config.cifs-server"
    )
    @OCIEValue(
        ocieColumn = "cifsServer"
    )
    private String cifsServer;

    @OCIEZapiSource(
        sourceZapi = "cifs-server-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-server-config.domain"
    )
    @OCIEValue(
        ocieColumn = "cifsDomain"
    )
    private String cifsDomain;

    @OCIEZapiSource(
        sourceZapi = "cifs-server-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-server-config.domain-workgroup"
    )
    @OCIEValue(
        ocieColumn = "cifsDomainWorkgroup"
    )
    private String cifsDomainWorkgroup;

    @OCIEZapiSource(
        sourceZapi = "cifs-server-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-server-config.auth-style"
    )
    @OCIEValue(
        ocieColumn = "cifsAuthStyle",
        notes = "derived enum"
    )
    private CifsAuthStyle cifsAuthStyle;

    // CIFS Options

    @OCIEZapiSource(
            sourceZapi = "cifs-options-get-iter",
            zapiResponsePath = "netapp.results.attributes-list.cifs-options.default-unix-user"
    )
    @OCIEValue(
            ocieColumn = "defaultUnixUser"
    )
    private String defaultUnixUser;

    @OCIEZapiSource(
            sourceZapi = "cifs-options-get-iter",
            zapiResponsePath = "netapp.results.attributes-list.cifs-options.default-unix-group"
    )
    @OCIEValue(
            ocieColumn = "defaultUnixGroup"
    )
    private String defaultUnixGroup;

    @OCIEZapiSource(
            sourceZapi = "cifs-options-get-iter",
            zapiResponsePath = "netapp.results.attributes-list.cifs-options.is-exportpolicy-enabled"
    )
    @OCIEValue(
            ocieColumn = "exportPolicyEnabled"
    )
    private Boolean exportPolicyEnabled;

    @OCIEZapiSource(
        sourceZapi = "cifs-server-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.cifs-server-config.auth-style"
    )
    @OCIEValue(
        ocieColumn = "cifsAuthStyle"
    )
    private String cifsAuthStyleRaw;

    // NFS Service
    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfs-access-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFSAccessEnabled"
    )
    private Boolean isNFSAccessEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv2-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFSV2Enabled"
    )
    private Boolean isNFSV2Enabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv3-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFSV3Enabled"
    )
    private Boolean isNFSV3Enabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv40-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFSV40Enabled"
    )
    private Boolean isNFSV40Enabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv41-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFSV41Enabled"
    )
    private Boolean isNFSV41Enabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv41-pnfs-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS41PnfsEnabled"
    )
    private Boolean isNFS41PnfsEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv41-acl-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS41AclEnabled"
    )
    private Boolean isNFS41AclEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv40-read-delegation-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS40ReadDelegationEnabled"
    )
    private Boolean isNFS40ReadDelegationEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv40-write-delegation-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS40WriteDelegationEnabled"
    )
    private Boolean isNFS40WriteDelegationEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv40-migration-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS40MigrationEnabled"
    )
    private Boolean isNFS40MigrationEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv40-referrals-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS40ReferralsEnabled"
    )
    private Boolean isNFS40ReferralsEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv41-migration-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS41MigrationEnabled"
    )
    private Boolean isNFS41MigrationEnabled;

    @OCIEZapiSource(
        sourceZapi = "nfs-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nfs-info.is-nfsv41-referrals-enabled"
    )
    @OCIEValue(
        ocieColumn = "isNFS41ReferralsEnabled"
    )
    private Boolean isNFS41ReferralsEnabled;

    // FCP Service     fcp-service-get-iter
    @OCIEZapiSource(
        sourceZapi = "fcp-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.fcp-service-info.node-name"
    )
    @OCIEValue(
        ocieColumn = "fcpNodeName",
        notes = "formated WWN"
    )
    private String fcpNodeName;

    @OCIEZapiSource(
        sourceZapi = "fcp-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.fcp-service-info.is-available"
    )
    @OCIEValue(
        ocieColumn = "isFcpAvailable"
    )
    private Boolean isFcpAvailable;     // Default to False in case no FCP Service exists for this vServer.

    // ISCSI Service
    @OCIEZapiSource(
        sourceZapi = "iscsi-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-service-info.node-name"
    )
    @OCIEValue(
        ocieColumn = "iscsiIQN"
    )
    private String iscsiIQN;

    @OCIEZapiSource(
        sourceZapi = "iscsi-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-service-info.alias-name"
    )
    @OCIEValue(
        ocieColumn = "iscsiAliasName"
    )
    private String iscsiAliasName;

    @OCIEZapiSource(
        sourceZapi = "iscsi-service-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.iscsi-service-info.is-available"
    )
    @OCIEValue(
        ocieColumn = "isISCSIAvailable"
    )
    private Boolean isISCSIAvailable;

    // DNS Service   net-dns-get-iter
    @OCIEZapiSource(
        sourceZapi = "net-dns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-dns-info.dns-state"
    )
    @OCIEValue(
        ocieColumn = "dnsState"
    )
    private DNSState dnsState;

    @OCIEZapiSource(
        sourceZapi = "net-dns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-dns-info.name-servers.ip-address"
    )
    @OCIEValue(
        ocieColumn = "dnsServers",
        notes = "CSV list"
    )
    private String dnsServers;

    @OCIEZapiSource(
        sourceZapi = "net-dns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-dns-info.domains.string"
    )
    @OCIEValue(
        ocieColumn = "dnsDomains",
        notes = "CSV list"
    )
    private String dnsDomains;

    // DDNS Service   net-ddns-get-iter
    @OCIEZapiSource(
        sourceZapi = "net-ddns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-ddns-info.domain-name"
    )
    @OCIEValue(
        ocieColumn = "domainName"
    )
    private String domainName;
    
    @OCIEZapiSource(
        sourceZapi = "net-ddns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-ddns-info.is-enabled"
    )
    @OCIEValue(
        ocieColumn = "isEnabled"
    )
    private Boolean isEnabled;
    
    @OCIEZapiSource(
        sourceZapi = "net-ddns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-ddns-info.use-secure"
    )
    @OCIEValue(
        ocieColumn = "useSecure"
    )
    private Boolean useSecure;
    
    @OCIEZapiSource(
        sourceZapi = "net-ddns-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.net-ddns-info.ttl"
    )
    @OCIEValue(
        ocieColumn = "ttl"
    )
    private Long ttl;
    
    // LDAP Service   ldap-config-get-iter
    @OCIEZapiSource(
        sourceZapi = "ldap-config-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.ldap-config.client-enabled"
    )
    @OCIEValue(
        ocieColumn = "isLDAPClientEnabled"
    )
    private Boolean isLDAPClientEnabled;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "volumeBytesTotal",
        notes = "Sum of TODO"
    )
    private Long volumeBytesTotal = 0L;

    @OCIEZapiSource(
        sourceZapi = "derived"
    )
    @OCIEValue(
        ocieColumn = "volumeBytesTotal",
        notes = "Sum of TODO"
    )
    private Long volumeBytesUsed = 0L;

    // Kerberos Service   kerberos-config-get-iter
    @OCIEZapiSource(
        sourceZapi = "kerberos-config-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.kerberos-config-info.is-kerberos-enabled"
    )
    @OCIEValue(
        ocieColumn = "isKerberosEnabled"
    )
    private Boolean isKerberosEnabled;

    // Qos Policy Group     qos-policy-group
    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.qos-policy-group"
    )
    @OCIEValue(
        ocieColumn = "qosPolicyGroupUUID"
    )
    private String qosPolicyGroupUUID;

    // nis-get-iter if any entry is active
    @OCIEZapiSource(
        sourceZapi = "nis-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nis-domain-config-info.is-active"
    )
    @OCIEValue(
        ocieColumn = "isNISActive"
    )
    private Boolean isNISActive = null;

    @OCIEZapiSource(
        sourceZapi = "nis-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nis-domain-config-info.nis-servers.ip-address"
    )
    @OCIEValue(
        ocieColumn = "activeNISServerIPs",
        notes = "CSV list"
    )
    private String activeNISServerIPs;

    @OCIEZapiSource(
        sourceZapi = "nis-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.nis-domain-config-info.nis-domain"
    )
    @OCIEValue(
        ocieColumn = "activeNISDomainName",
        notes = "Name of active nis domain"
    )
    private String activeNISDomainName;

    @OCIEZapiSource(
        sourceZapi = "vserver-get-iter",
        zapiResponsePath = "netapp.results.attributes-list.vserver-info.ipspace"
    )
    @OCIEValue(
        ocieColumn = "networkIpSpaceId"
    )
    private NetworkIpSpace.NetworkIpSpaceKey ipSpaceKey;

    @OCIEZapiSource(
            sourceZapi = "cifs-security-get-iter",
            zapiResponsePath = "cifs-security.is-smb-encryption-required"
    )
    @OCIEValue(
            ocieColumn = "isSmbEncryptionRequired"
    )
    private Boolean isSmbEncryptionRequired;

    // Children
    private final Volumes volumes;
    private final VolumeSelfRelationships volumeSelfRelationships;
    private final VolumeFlexCaches volumeFlexCaches;
    private final StorageServiceVolumeRelationships storageServiceVolumeRelationships;
    private final Snapshots snapshots;
    private final VolumeMoves volumeMoves;
    private final NetworkLifs networkLifs;
    private final FcpLifs fcpLifs;
    private final IGroups iGroups;
    private final Portsets portSets;
    private final LUNMaps lunMaps;
    private final CIFSShares cifsShares;
    private final ExportPolicies exportPolicies;
    private final ISCSIPortalGroups iscsiPortalGroups;
    private final ISCSISecurityEntries iscsiSecurityEntries;
    private final SisPolicies sisPolicies;
    private final StorageServices storageServices;
    private final RoutingGroups routingGroups;
    private final NISDomains nisDomains;
    private final VServerNameMappings vServerNameMappings;


    VServer(final Cluster parent, final String clusterUUID, final String volumeUUID) {
        super(parent, "Vserver");
        this.vServerKey = new VServerKey(clusterUUID, volumeUUID);
        this.volumes = new Volumes(this);
        this.volumeSelfRelationships = new VolumeSelfRelationships(this);
        this.volumeFlexCaches = new VolumeFlexCaches(this);
        this.storageServiceVolumeRelationships = new StorageServiceVolumeRelationships(this);
        this.snapshots = new Snapshots(this);
        this.volumeMoves = new VolumeMoves(this);
        this.networkLifs = new NetworkLifs(this);
        this.fcpLifs = new FcpLifs(this);
        this.iGroups = new IGroups(this);
        this.portSets = new Portsets(this);
        this.lunMaps = new LUNMaps(this);
        this.cifsShares = new CIFSShares(this);
        this.exportPolicies = new ExportPolicies(this);
        this.iscsiPortalGroups = new ISCSIPortalGroups(this);
        this.iscsiSecurityEntries = new ISCSISecurityEntries(this);
        this.sisPolicies = new SisPolicies(this);
        this.storageServices = new StorageServices(this);
        this.routingGroups = new RoutingGroups(this);
        this.nisDomains = new NISDomains(this);
        this.vServerNameMappings = new VServerNameMappings(this);
    }

    @Override
    @SuppressWarnings({"unchecked"})
    public VServerKey getKey() {
        return vServerKey;
    }

    public String getUUID() {
        return vServerKey.getOcieUUID();
    }

    public String getName() {
        return name;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public Type getType() {
        return type;
    }

    public void setType(final Type type) {
        this.type = type;
    }
    
    public void setTypeRaw(final String typeRaw) {
        this.typeRaw = typeRaw;
    }
    
    public Subtype getSubtype() {
        return subtype;
    }

    public void setSubtype(final Subtype subtype) {
        this.subtype = subtype;
    }

    public void setSubtypeRaw(final String subtypeRaw) {
        this.subtypeRaw = subtypeRaw;
    }

    public String getFcpNodeName() {
        return fcpNodeName;
    }

    public void setFcpNodeName(final String fcpNodeName) {
        this.fcpNodeName = fcpNodeName;
    }

    public Boolean isFcpAvailable() {
        return isFcpAvailable;
    }

    public void setFcpAvailable(final Boolean fcpAvailable) {
        isFcpAvailable = fcpAvailable;
    }

    public State getState() {
        return state;
    }

    public void setState(final State state) {
        this.state = state;
    }
    
    public void setStateRaw(final String stateRaw) {
        this.stateRaw = stateRaw;
    }
    
    public String getStateRaw() {
        return stateRaw;
    }
    
    public OperationalState getOperationalState() {
        return operationalState;
    }

    public void setOperationalState(final OperationalState operationalState) {
        this.operationalState = operationalState;
    }
    
    public String getOperationalStateRaw() {
        return operationalStateRaw;
    }
    
    public void setOperationalStateRaw(final String operationalStateRaw) {
        this.operationalStateRaw = operationalStateRaw;
    }

    public void setRawAllowedProtocols(final String allowedProtocols) {
        this.allowedProtocolsCSVString = allowedProtocols;
    }

    public void setRawDisAllowedProtocols(final String disAllowedProtocols) {
        this.disAllowedProtocolsCSVString = disAllowedProtocols;
    }

    public SnapshotPolicy.SnapshotPolicyKey getSnapshotPolicyKey() {
        return snapshotPolicyKey;
    }

    public void setSnapshotPolicyKey(final SnapshotPolicy.SnapshotPolicyKey snapshotPolicyKey) {
        this.snapshotPolicyKey = snapshotPolicyKey;
    }

    public String getRootVolumeName() {
        return rootVolumeName;
    }

    public void setRootVolumeName(final String rootVolumeName) {
        this.rootVolumeName = rootVolumeName;
    }

    public Integer getMaxVolumes() {
        return maxVolumes;
    }

    public void setMaxVolumes(final Integer maxVolumes) {
        this.maxVolumes = maxVolumes;
    }

    public Language getLanguage() {
        return language;
    }

    public void setLanguage(final Language language) {
        this.language = language;
    }

    public void setLanguageRaw(final String languageRaw) {
        this.languageRaw = languageRaw;
    }

    public Boolean isRepository() {
        return isRepository;
    }

    public void setRepository(final Boolean repository) {
        isRepository = repository;
    }

    public String getNmSwitch() {
        return nmSwitch;
    }

    public void setNmSwitch(final String nmSwitch) {
        this.nmSwitch = nmSwitch;
    }

    public Boolean isNFSAccessEnabled() {
        return isNFSAccessEnabled;
    }

    public void setNFSAccessEnabled(final Boolean NFSAccessEnabled) {
        isNFSAccessEnabled = NFSAccessEnabled;
    }

    public Boolean isNFSV2Enabled() {
        return isNFSV2Enabled;
    }

    public void setNFSV2Enabled(final Boolean NFSV2Enabled) {
        isNFSV2Enabled = NFSV2Enabled;
    }

    public Boolean isNFSV3Enabled() {
        return isNFSV3Enabled;
    }

    public void setNFSV3Enabled(final Boolean NFSV3Enabled) {
        isNFSV3Enabled = NFSV3Enabled;
    }

    public Boolean isNFSV40Enabled() {
        return isNFSV40Enabled;
    }

    public void setNFSV40Enabled(final Boolean NFSV40Enabled) {
        isNFSV40Enabled = NFSV40Enabled;
    }

    public Boolean isNFSV41Enabled() {
        return isNFSV41Enabled;
    }

    public void setNFSV41Enabled(final Boolean NFSV41Enabled) {
        isNFSV41Enabled = NFSV41Enabled;
    }

    public String getIscsiIQN() {
        return iscsiIQN;
    }

    public void setISCSIIQN(final String iscsiIQN) {
        this.iscsiIQN = iscsiIQN;
    }

    public String getIscsiAliasName() {
        return iscsiAliasName;
    }

    public void setISCSIAliasName(final String iscsiAliasName) {
        this.iscsiAliasName = iscsiAliasName;
    }

    public Boolean isISCSIAvailable() {
        return isISCSIAvailable;
    }

    public void setISCSIAvailable(final Boolean ISCSIAvailable) {
        isISCSIAvailable = ISCSIAvailable;
    }

    public Boolean isCifsEnabled() {
        return isCifsEnabled;
    }

    public void setCifsEnabled(final Boolean cifsEnabled) {
        isCifsEnabled = cifsEnabled;
    }

    public Boolean isExportPolicyEnabled() {
        return exportPolicyEnabled;
    }

    public void setExportPolicyEnabled(final Boolean exportPolicyEnabled) {
        this.exportPolicyEnabled = exportPolicyEnabled;
    }

    public String getDefaultUnixUser() {
        return defaultUnixUser;
    }

    public void setDefaultUnixUser(final String defaultUnixUser) {
        this.defaultUnixUser = defaultUnixUser;
    }

    public String getDefaultUnixGroup() {
        return defaultUnixGroup;
    }

    public void setDefaultUnixGroup(final String defaultUnixGroup) {
        this.defaultUnixGroup = defaultUnixGroup;
    }

    public String getCifsServer() {
        return cifsServer;
    }

    public void setCifsServer(final String cifsServer) {
        this.cifsServer = cifsServer;
    }

    public String getCifsDomain() {
        return cifsDomain;
    }

    public void setCifsDomain(final String cifsDomain) {
        this.cifsDomain = cifsDomain;
    }

    public String getCifsDomainWorkgroup() {
        return cifsDomainWorkgroup;
    }

    public void setCifsDomainWorkgroup(final String cifsDomainWorkgroup) {
        this.cifsDomainWorkgroup = cifsDomainWorkgroup;
    }

    public CifsAuthStyle getCifsAuthStyle() {
        return cifsAuthStyle;
    }

    public void setCifsAuthStyle(final CifsAuthStyle cifsAuthStyle) {
        this.cifsAuthStyle = cifsAuthStyle;
    }

    public void setCifsAuthStyleRaw(final String cifsAuthStyleRaw) {
        this.cifsAuthStyleRaw = cifsAuthStyleRaw;
    }

    public Boolean isNFS41PnfsEnabled() {
        return isNFS41PnfsEnabled;
    }

    public void setNFS41PnfsEnabled(final Boolean NFS41PnfsEnabled) {
        isNFS41PnfsEnabled = NFS41PnfsEnabled;
    }

    public Boolean isNFS41AclEnabled() {
        return isNFS41AclEnabled;
    }

    public void setNFS41AclEnabled(final Boolean NFS41AclEnabled) {
        isNFS41AclEnabled = NFS41AclEnabled;
    }

    public Boolean isNFS40ReadDelegationEnabled() {
        return isNFS40ReadDelegationEnabled;
    }

    public void setNFS40ReadDelegationEnabled(final Boolean NFS40ReadDelegationEnabled) {
        isNFS40ReadDelegationEnabled = NFS40ReadDelegationEnabled;
    }

    public Boolean isNFS40WriteDelegationEnabled() {
        return isNFS40WriteDelegationEnabled;
    }

    public void setNFS40WriteDelegationEnabled(final Boolean NFS40WriteDelegationEnabled) {
        isNFS40WriteDelegationEnabled = NFS40WriteDelegationEnabled;
    }

    public Boolean isNFS40MigrationEnabled() {
        return isNFS40MigrationEnabled;
    }

    public void setNFS40MigrationEnabled(final Boolean NFS40MigrationEnabled) {
        isNFS40MigrationEnabled = NFS40MigrationEnabled;
    }

    public Boolean isNFS40ReferralsEnabled() {
        return isNFS40ReferralsEnabled;
    }

    public void setNFS40ReferralsEnabled(final Boolean NFS40ReferralsEnabled) {
        isNFS40ReferralsEnabled = NFS40ReferralsEnabled;
    }

    public Boolean isNFS41MigrationEnabled() {
        return isNFS41MigrationEnabled;
    }

    public void setNFS41MigrationEnabled(final Boolean NFS41MigrationEnabled) {
        isNFS41MigrationEnabled = NFS41MigrationEnabled;
    }

    public Boolean isNFS41ReferralsEnabled() {
        return isNFS41ReferralsEnabled;
    }

    public void setNFS41ReferralsEnabled(final Boolean NFS41ReferralsEnabled) {
        isNFS41ReferralsEnabled = NFS41ReferralsEnabled;
    }

    public DNSState getDnsState() {
        return dnsState;
    }

    public void setDnsState(final DNSState dnsState) {
        this.dnsState = dnsState;
    }

    public String getDnsServers() {
        return dnsServers;
    }

    public void setDnsServers(final String dnsServers) {
        this.dnsServers = dnsServers;
    }

    public String getDnsDomains() {
        return dnsDomains;
    }

    public void setDnsDomains(final String dnsDomains) {
        this.dnsDomains = dnsDomains;
    }
    
    public String getDomainName() {
        return domainName;
    }

    public void setDomainName(final String domainName) {
        this.domainName = domainName;
    }
    
    public Boolean getIsEnabled() {
        return isEnabled;
    }

    public void setIsEnabled(final Boolean isEnabled) {
        this.isEnabled = isEnabled;
    }
    
    public Boolean getUseSecure() {
        return useSecure;
    }

    public void setUseSecure(final Boolean useSecure) {
        this.useSecure = useSecure;
    }
    
    public Long getTtl() {
        return ttl;
    }

    public void setTtl(final Long ttl) {
        this.ttl = ttl;
    }

    public Boolean getLDAPClientEnabled() {
        return isLDAPClientEnabled;
    }

    public void setLDAPClientEnabled(final Boolean LDAPClientEnabled) {
        isLDAPClientEnabled = LDAPClientEnabled;
    }

    public Long getVolumeBytesTotal() {
        return volumeBytesTotal;
    }

    public void addToVolumeBytesTotal(final Long volumeBytes) {
        this.volumeBytesTotal += volumeBytes;
    }

    public Long getVolumeBytesUsed() {
        return volumeBytesUsed;
    }

    public void addToVolumeBytesUsed(final Long volumeBytes) {
        this.volumeBytesUsed += volumeBytes;
    }

    public Long getVolumeBytesAvail() {
        return (volumeBytesTotal - volumeBytesUsed);
    }

    public Boolean isKerberosEnabled() {
        return isKerberosEnabled;
    }

    public void setKerberosEnabled(final Boolean kerberosEnabled) {
        isKerberosEnabled = kerberosEnabled;
    }

    public String getQosPolicyGroupUUID() {
        return qosPolicyGroupUUID;
    }

    public void setQosPolicyGroupUUID(final String qosPolicyGroupUUID) {
        this.qosPolicyGroupUUID = qosPolicyGroupUUID;
    }

    public void setNISActive(final boolean NISActive) {
        this.isNISActive = NISActive;
    }

    public Boolean getNISActive() {
        return isNISActive;
    }

    public void setActiveNISServerIPs(final String activeNISServerIPs) {
        this.activeNISServerIPs = activeNISServerIPs;
    }

    public void setActiveNISDomainName(final String activeNISDomainName) {
        this.activeNISDomainName = activeNISDomainName;
    }

    public Volume createVolume(final String uuid, final String instanceUUID, final Volume.VolumeStyle volumeStyle, final Boolean isConstituent) {
        return new Volume(this, uuid, instanceUUID, volumeStyle, isConstituent);
    }

    public Volumes getVolumes() {
        return volumes;
    }

    public VolumeSelfRelationship createVolumeSelfRelationship(final String volumeUUID) {
        return new VolumeSelfRelationship(this, volumeUUID);
    }

    public VolumeSelfRelationships getVolumeSelfRelationships() {
        return volumeSelfRelationships;
    }

    public VolumeFlexCache createVolumeFlexCache(final String volumeFlexCacheInstanceUUID) {
        return new VolumeFlexCache(this, volumeFlexCacheInstanceUUID);
    }

    public VolumeFlexCaches getVolumeFlexCaches() {
        return volumeFlexCaches;
    }

    public StorageServiceVolumeRelationship createStorageServiceVolumeRelationship(final StorageService.StorageServiceKey storageServiceKey,
                                                                                   final Volume.VolumeKey volumeKey) {
        return new StorageServiceVolumeRelationship(this, storageServiceKey, volumeKey);
    }

    public StorageServiceVolumeRelationships getStorageServiceVolumeRelationships() {
        return storageServiceVolumeRelationships;
    }

    public Snapshot createSnapshot(final String volumeInstanceUUID, final String snapshotInstanceUUID) {
        return new Snapshot(this, volumeInstanceUUID, snapshotInstanceUUID);
    }

    public Snapshots getSnapshots() {
        return snapshots;
    }

    public VolumeMove createVolumeMove(final String jobUUID) {
        return new VolumeMove(this, jobUUID);
    }

    public VolumeMoves getVolumeMoves() {
        return volumeMoves;
    }

    public SisPolicy createSisPolicy(final String vServerUUID, final String sisPolicyName) {
        return new SisPolicy(this, vServerUUID, sisPolicyName);
    }

    public SisPolicies getSisPolicies() {
        return sisPolicies;
    }

    public StorageService createStorageService(final Long serviceID, final Volume.VolumeKey volumeKey) {
        return new StorageService(this, serviceID, getKey(), volumeKey);
    }

    public StorageServices getStorageServices() {
        return storageServices;
    }

    public RoutingGroup createRoutingGroup(final String groupUUID) {
        return new RoutingGroup(this, groupUUID);
    }

    public RoutingGroups getRoutingGroups() {
        return routingGroups;
    }

    public NISDomains getNisDomains() {
        return nisDomains;
    }

    public NISDomain createNISDomain(final String domainName) {
        return new NISDomain(this, vServerKey.getID(VServerKey.Field.cluster_uuid), vServerKey.getOcieUUID(), domainName);
    }

    public NetworkLif createLif(final String lifUUID) {
        return new NetworkLif(this, lifUUID);
    }

    public NetworkLifs getNetworkLifs() {
        return networkLifs;
    }

    public FcpLif createFcpLif(final String lifName) {
        return new FcpLif(this, lifName);
    }

    public FcpLifs getFcpLifs() {
        return fcpLifs;
    }

    public IGroup createIGroup(final String iGroupUUID) {
        return new IGroup(this, iGroupUUID);
    }

    public IGroups getIGroups() {
        return iGroups;
    }

    public Portset createPortSet(final String portSetUUID) {
        return new Portset(this, portSetUUID);
    }

    public Portsets getPortSets() {
        return portSets;
    }

    public LUNMap createLUNMap(final String lunUUID, final String iGroupUUID) {
        return new LUNMap(this, lunUUID, iGroupUUID);
    }

    public LUNMaps getLUNMaps() {
        return lunMaps;
    }

    public CIFSShare createCIFSShare(final String shareName) {
        return new CIFSShare(this, shareName);
    }

    public CIFSShares getCIFSShares() {
        return cifsShares;
    }

    public ExportPolicies getExportPolicies() {
        return exportPolicies;
    }

    public ExportPolicy createExportPolicy(final Long policyID) {
        return new ExportPolicy(this, policyID);
    }

    public ISCSIPortalGroups getISCSIPortalGroups() {
        return iscsiPortalGroups;
    }

    public ISCSIPortalGroup createISCSIPortalGroup(final String portalGroupName) {
        return new ISCSIPortalGroup(this, portalGroupName);
    }

    public ISCSISecurityEntries getISCSISecurityEntries() {
        return iscsiSecurityEntries;
    }

    public ISCSISecurityEntry createISCSISecurityEntry(final String clusterUUID, final String initiator) {
        return new ISCSISecurityEntry(this, clusterUUID, initiator);
    }

    public VServerNameMappings getVServerNameMappings() {
        return vServerNameMappings;
    }

    public VServerNameMapping createVServerNameMapping(final String direction, final Integer position) {
        return new VServerNameMapping(this, direction, position);
    }


    public List<Protocol> getAllowedProtocols() {
        return allowedProtocols;
    }

    public void setAllowedProtocols(final List<Protocol> allowedProtocols) {
        this.allowedProtocols = allowedProtocols;
    }


    public List<Protocol> getDisAllowedProtocols() {
        return disAllowedProtocols;
    }

    public void setDisAllowedProtocols(final List<Protocol> disAllowedProtocols) {
        this.disAllowedProtocols = disAllowedProtocols;
    }

    public NetworkIpSpace.NetworkIpSpaceKey getIpSpaceKey() {
        return ipSpaceKey;
    }

    public void setIpSpaceKey(NetworkIpSpace.NetworkIpSpaceKey ipSpaceKey) {
        this.ipSpaceKey = ipSpaceKey;
    }
    
    public Boolean getIsConfigLockedForChanges(){
    	return isConfigLockedForChanges;
    }
    
    public void setIsConfigLockedForChanges(Boolean isConfigLockedForChanges){
    	this.isConfigLockedForChanges = isConfigLockedForChanges;
    }

    public Boolean getIsSmbEncryptionRequired() { return isSmbEncryptionRequired; }

    public void setIsSmbEncryptionRequired(Boolean isSmbEncryptionRequired) {
        this.isSmbEncryptionRequired = isSmbEncryptionRequired;
    }


    /**
     * VServer State
     */

    public static enum State {
        NOT_MAPPED,
        DELETING,
        INITIALIZING,
        RUNNING,
        STARTING,
        STOPPED,
        STOPPING, }

    /**
     * VServer Operational-State
     */

    public static enum OperationalState {
        NOT_MAPPED,
        RUNNING,
        STOPPED,}

    /**
     * VServer Type
     */

    public static enum Type {
        NOT_MAPPED,
        CLUSTER,
        ADMIN,
        NODE,
        DATA, }
    
    /**
     * VServer Subtype
     */     

    public static enum Subtype {
        NOT_MAPPED,
        DEFAULT,
        DP_DESTINATION,
        SYNC_SOURCE,
        SYNC_DESTINATION, }

    /**
     * VServer Protocols
     */

    public static enum Protocol {
        NOT_MAPPED,
        NFS,
        CIFS,
        ISCSI,
        FCP, }


    public static enum Language {
        NOT_MAPPED,
        C,                            // 'C' .............. POSIX
        ar,                           // 'ar' ............. Arabic
        cs,                           // 'cs' ............. Czech
        da,                           // 'da' ............. Danish
        de,                           // 'de' ............. German
        en,                           // 'en' ............. English
        en_US,                        // 'en_US' .......... English (US)
        es,                           // 'es' ............. Spanish
        fi,                           // 'fi' ............. Finnish
        fr,                           // 'fr' ............. French
        he,                           // 'he' ............. Hebrew
        hr,                           // 'hr' ............. Croatian
        hu,                           // 'hu' ............. Hungarian
        it,                           // 'it' ............. Italian
        ja,                           // 'ja' ............. Japanese euc-j
        ja_v1,                        // 'ja_v1' .......... Japanese euc-j
        ja_JP_PCK("ja_JP.PCK"),       // 'ja_JP.PCK' ...... Japanese PCK (sjis)
        ja_JP_932("ja_JP.932"),       // 'ja_JP.932' ...... Japanese cp932
        ja_JP_PCK_v2("ja_JP.PCK_v2"), // 'ja_JP.PCK_v2' .... Japanese PCK (sjis)
        ko,                           // 'ko' ............. Korean
        no,                           // 'no' ............. Norwegian
        nl,                           // 'nl' ............. Dutch
        pl,                           // 'pl' ............. Polish
        pt,                           // 'pt' ............. Portuguese
        ro,                           // 'ro' ............. Romanian
        ru,                           // 'ru' ............. Russian
        sk,                           // 'sk' ............. Slovak
        sl,                           // 'sl' ............. Slovenian
        sv,                           // 'sv' ............. Swedish
        tr,                           // 'tr' ............. Turkish
        zh,                           // 'zh' ............. Simplified Chinese
        zh_ZBK("zh.GBK"),             // 'zh.GBK' ......... Simplified Chinese (GBK)
        zh_TW,                        // 'zh_TW' .......... Traditional Chinese euc-tw
        zh_TW_BIG5("zh_TW.BIG5");     // 'zh_TW.BIG5' ..... Traditional Chinese Big 5

        private final String xmlValue;

        private Language() {
            this(null);
        }

        private Language(final String xmlValue) {
            this.xmlValue = xmlValue;
        }

        public String toXML() {
            return (xmlValue == null ? name() : xmlValue);
        }

        public static Language fromString(final String stringVal) {
            if (null == stringVal) {
                return null;
            }
            for (final Language candidateLanguage : Language.values()) {
                if (candidateLanguage.toXML().equalsIgnoreCase(stringVal)) {
                    return candidateLanguage;
                }
            }
            return NOT_MAPPED;
        }
    }

    public static enum CifsAuthStyle {
        NOT_MAPPED,
        DOMAIN,
        WORKGROUP,
    }

    public static enum CifsAdminStatus {
        NOT_MAPPED,
        UP,
        DOWN,
    }

    public static enum DNSState {
        NOT_MAPPED,
        ENABLED,
        DISABLED,
    }

    @Override
    protected void marshallAttributesToFileAndCRC(final MarshallingCache marshallingCache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        super.marshallAttributesToFileAndCRC(marshallingCache, writer);
        addRequiredAttributeAndCRC(writer, "ResourceKey", vServerKey);
        addRequiredAttributeAndCRC(writer, "UUID", vServerKey.getOcieUUID());
        addRequiredAttributeAndCRC(writer, "Name", name);
        addRequiredAttributeAndCRC(writer, "State", state);
        addRequiredAttributeAndCRC(writer, "StateRaw", stateRaw);
        addRequiredAttributeAndCRC(writer, "OperationalState", operationalState);
        addRequiredAttributeAndCRC(writer, "OperationalStateRaw", operationalStateRaw);
        addRequiredAttributeAndCRC(writer, "Type", type);
        addRequiredAttributeAndCRC(writer, "TypeRaw", typeRaw);
        addRequiredAttributeAndCRC(writer, "Subtype", subtype);
        addRequiredAttributeAndCRC(writer, "SubtypeRaw", subtypeRaw);
        addRequiredAttributeAndCRC(writer, "AllowedProtocols", allowedProtocolsCSVString);
        addRequiredAttributeAndCRC(writer, "DisallowedProtocols", disAllowedProtocolsCSVString);

        addRequiredAttributeAndCRC(writer, "RootVolumeName", rootVolumeName);
        addRequiredAttributeAndCRC(writer, "MaximumVolumes", maxVolumes);
        addRequiredAttributeAndCRC(writer, "Language", languageRaw);

        addRequiredAttributeAndCRC(writer, "IsRepository", isRepository);
        addRequiredAttributeAndCRC(writer, "IsConfigLockedForChanges", isConfigLockedForChanges);
        addRequiredAttributeAndCRC(writer, "NameServerSwitch", nmSwitch);

        addRequiredAttributeAndCRC(writer, "FCPEnabled", isFcpAvailable);
        addRequiredAttributeAndCRC(writer, "FCPNodeName", fcpNodeName);

        addRequiredAttributeAndCRC(writer, "ISCSIEnabled", isISCSIAvailable);
        addRequiredAttributeAndCRC(writer, "ISCSINodeName", iscsiIQN);
        addRequiredAttributeAndCRC(writer, "ISCSIAliasName", iscsiAliasName);

        addRequiredAttributeAndCRC(writer, "CIFSEnabled", isCifsEnabled);
        addRequiredAttributeAndCRC(writer, "CIFSServer", cifsServer);
        addRequiredAttributeAndCRC(writer, "CIFSDomain", cifsDomain);
        addRequiredAttributeAndCRC(writer, "ExportPolicyEnabled", exportPolicyEnabled);
        addRequiredAttributeAndCRC(writer, "DefaultUnixUser", defaultUnixUser);
        addRequiredAttributeAndCRC(writer, "DefaultUnixGroup", defaultUnixGroup);
        //        addRequiredAttributeAndCRC(writer, "CIFSDomainWorkgroup", cifsDomainWorkgroup);
        addRequiredAttributeAndCRC(writer, "CIFSAuthenticationStyle", cifsAuthStyle);
        addRequiredAttributeAndCRC(writer, "CIFSAuthenticationStyleRaw", cifsAuthStyleRaw);

        addRequiredAttributeAndCRC(writer, "NFSEnabled", isNFSAccessEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV2Supported", isNFSV2Enabled);
        addRequiredAttributeAndCRC(writer, "NFSV3Supported", isNFSV3Enabled);
        addRequiredAttributeAndCRC(writer, "NFSV4Supported", isNFSV40Enabled);

        addRequiredAttributeAndCRC(writer, "NFSV4ReadDelegationEnabled", isNFS40ReadDelegationEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV4WriteDelegationEnabled", isNFS40WriteDelegationEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV4MigrationEnabled", isNFS40MigrationEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV4ReferralsEnabled", isNFS40ReferralsEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV41Supported", isNFSV41Enabled);
        addRequiredAttributeAndCRC(writer, "NFSV41PnfsEnabled", isNFS41PnfsEnabled);
        addRequiredAttributeAndCRC(writer, "NFS41AclEnabled", isNFS41AclEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV41MigrationEnabled", isNFS41MigrationEnabled);
        addRequiredAttributeAndCRC(writer, "NFSV41ReferralsEnabled", isNFS41ReferralsEnabled);

        addRequiredAttributeAndCRC(writer, "NISEnabled", isNISActive);
        addRequiredAttributeAndCRC(writer, "ActiveNISDomainName", activeNISDomainName);
        addRequiredAttributeAndCRC(writer, "ActiveNISServers", activeNISServerIPs);

        if (dnsState != null) {
            addRequiredAttributeAndCRC(writer, "DNSEnabled", (DNSState.ENABLED == dnsState));
        }
        addRequiredAttributeAndCRC(writer, "DNSServers", dnsServers);
        addRequiredAttributeAndCRC(writer, "DNSDomainNames", dnsDomains);
        
        addRequiredAttributeAndCRC(writer, "DomainName", domainName);
        addRequiredAttributeAndCRC(writer, "IsEnabled", isEnabled);
        addRequiredAttributeAndCRC(writer, "UseSecure", useSecure);
        addRequiredAttributeAndCRC(writer, "Ttl", ttl);

        addRequiredAttributeAndCRC(writer, "LdapClientEnabled", isLDAPClientEnabled);

        addRequiredAttributeAndCRC(writer, "VolumeBytesTotal", volumeBytesTotal);
        addRequiredAttributeAndCRC(writer, "VolumeBytesUsed", volumeBytesUsed);
        addRequiredAttributeAndCRC(writer, "VolumeBytesAvail", getVolumeBytesAvail());

        addRequiredAttributeAndCRC(writer, "SnapshotPolicyUUID", KeyBase.getOcieUUIDOrNull(snapshotPolicyKey));
        addRequiredAttributeAndCRC(writer, "QosPolicyGroupUUID", qosPolicyGroupUUID);
        addRequiredAttributeAndCRC(writer, "IsKerberosEnabled", isKerberosEnabled);
        if (ipSpaceKey != null) {
            addRequiredAttributeAndCRC(writer, "NetworkIpSpaceUUID", ipSpaceKey.getOcieUUID());
        }

        addRequiredAttributeAndCRC(writer, "IsSmbEncryptionRequired", isSmbEncryptionRequired);
    }

    @Override
    protected boolean hasChildren() {
        return !exportPolicies.isEmpty() || !volumes.isEmpty() || !cifsShares.isEmpty() || !networkLifs.isEmpty() || !fcpLifs.isEmpty()
               || !iGroups.isEmpty() || !portSets.isEmpty() || !lunMaps.isEmpty() || !iscsiPortalGroups.isEmpty() || !iscsiSecurityEntries.isEmpty()
               || !storageServices.isEmpty() || !routingGroups.isEmpty() || !storageServiceVolumeRelationships.isEmpty() || !volumeSelfRelationships.isEmpty();

    }

    @Override
    protected void marshallChildrenToFileAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNodeAndCRC(cache, writer, "ExportPolicies", exportPolicies.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "SisPolicies", sisPolicies.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "StorageServices", storageServices.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "Volumes", volumes.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "VolumeSelfRelationships", volumeSelfRelationships.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "VolumeFlexCaches", volumeFlexCaches.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "StorageServiceVolumeRelationships", storageServiceVolumeRelationships.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "CIFSShares", cifsShares.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "RoutingGroups", routingGroups.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "NetworkLIFs", networkLifs.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "FCPLIFs", fcpLifs.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "Igroups", iGroups.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "LUNMaps", lunMaps.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "ISCSIPortalGroups", iscsiPortalGroups.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "ISCSISecurityEntries", iscsiSecurityEntries.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "Portsets", portSets.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "Snapshots", snapshots.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "VolumeMoveJobs", volumeMoves.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "NisDomains", nisDomains.values(), tabs);
        addAllChildrenUnderNodeAndCRC(cache, writer, "NameMappings", vServerNameMappings.values(), tabs);
    }
}
