package com.onaro.sanscreen.acquisition.netappstorageperformance.datasource;

import java.util.Collection;

public class NetAppMetricCalculator {

    private NetAppMetricCalculator() {
    }

    /**
     * Calculates average amount of time spent on each IO.
     * Calculated based on disk specifications (rather than read from device)
     *
     * Assumptions:
     * 1. Average latency : we only need to rotate 1/2 of revolution to fulfill the request
     * 2. Seek time is always the max seek time, but the disk only performs seek once per request
     * 3. Formula for RAID is RAID-1 (reading from raid-0 or raid-1) or Writing to RAID-1.
     * 4. Transfer rate is the same for reading and writing. Calculated as READ operation
     *
     * @param numIO
     * @param byteIO
     * @param diskSpecifications
     *
     * @return access time in millis
     */
    public static float figureOutAccessTime(final long numIO,
                                            final long byteIO,
                                            final Collection<NetAppDiskSpecification> diskSpecifications) {
        final float accessTimeMillis;
        if (diskSpecifications == null || diskSpecifications.size() == 0) {
            accessTimeMillis = -1;
        }
        else {
            if (numIO == 0) {
                accessTimeMillis = 0;
            }
            else {
                final float volumeServiceTimeMS = NetAppMetricCalculator.getVolumeServiceTimeMillis(byteIO, numIO, diskSpecifications);
                accessTimeMillis = volumeServiceTimeMS / numIO;
            }

        }
        return accessTimeMillis;
    }

    /**
     * @param deltaBytes        - delta bytes (can be read write or both)
     * @param deltaNumRequest   - the delta number IO requests (can be read write or both)
     * @param deltaTime         the delta time in milliseconds
     * @param diskSpecification - for extracting the speed, latency or access time
     *
     * @return the utilization
     */
    public static float getDiskUtilization(final long deltaBytes,
                                           final long deltaNumRequest,
                                           final long deltaTime,
                                           final NetAppDiskSpecification diskSpecification) {
        final float serviceTimeMillis = getDiskServiceTimeMillis(deltaBytes, deltaNumRequest, diskSpecification);
        if (serviceTimeMillis == 0)
            return 0;

        // Service time may accidentally (calculated based on averages, not read from device)
        // represent greater time to read data from disk than actually happened
        // Thats why we are capping utilization so not to report greater than 100%

        final float diskUtilization = (100 * serviceTimeMillis) / deltaTime;
        return Math.min(100, diskUtilization);
    }

    /**
     * @param deltaBytes      - delta bytes (can be read write or both)
     * @param deltaTimeMillis the delta time in milliseconds
     *
     * @return throughput in MB/s
     */
    public static float getThoughputInMBPerSec(final long deltaBytes, final long deltaTimeMillis) {
        if (deltaBytes == 0)
            return 0;

        // from Millies to Sec and from byte to MB
        final float deltaBytesMB = (deltaBytes / 1024F / 1024F);
        final float deltaTimeSecs = (deltaTimeMillis / 1000F);

        return (deltaBytesMB / deltaTimeSecs);
    }

    /**
     * Calc and return IOPS
     *
     * @param deltaNumRequest amount of requests happened in this interval
     * @param deltaTimeMills  interval betwen samples
     *
     * @return IOPS
     */
    public static float getIOPS(final long deltaNumRequest, final long deltaTimeMills) {
        if (deltaNumRequest == 0)
            return 0;

        final float delaTimeSeconds = deltaTimeMills / 1000F;
        return (deltaNumRequest / delaTimeSeconds);
    }

    /**
     * return transfer time in milliseconds
     * This method calculates avarage transfer time per IO request using the potential transfer rate of the disk
     *
     * @param deltaBytes        - delta bytes (can be read write or both)
     * @param deltaNumRequest   - the delta number IO requests (can be read write or both)
     * @param diskSpecification - for extracting the speed, latency or access time
     *
     * @return get average time in milliscond
     */
    private static float getDiskTransferTimeMillis(long deltaBytes, long deltaNumRequest, NetAppDiskSpecification diskSpecification) {
        //we need to convert transfer rate from Byte per second to byte per millisecond
        final float averageBytesPerRequest = getAverageBytesPerRequest(deltaBytes, deltaNumRequest);
        final float diskPotentialTransferRate = getDiskPotentialTransferRateBytesPerSecond(diskSpecification) / 1000F;
        return averageBytesPerRequest / diskPotentialTransferRate;
    }

    /**
     * disk potential transfer rate in bytes per second
     *
     * @param diskSpecification
     *
     * @return
     */
    private static float getDiskPotentialTransferRateBytesPerSecond(NetAppDiskSpecification diskSpecification) {
        // disk spec transfer rate is MB per second, for example 80 MB/s
        // converting MB to byte
        return diskSpecification.getPotentialTransferRate() * 1024 * 1024;
    }

    /**
     * @param deltaBytes      - delta bytes (can be read write or both)
     * @param deltaNumRequest - the delta number IO requests (can be read write or both)
     *
     * @return the Average Bytes Per Request
     */
    static public float getAverageBytesPerRequest(long deltaBytes, long deltaNumRequest) {
        if (deltaNumRequest == 0)
            return 0;

        return ((float) deltaBytes) / (float) deltaNumRequest;
    }

    /**
     * @param deltaBytes        - delta bytes (can be read write or both)
     * @param deltaNumRequest   - the delta number IO requests (can be read write or both)
     * @param diskSpecification - for extracting the speed, latency or access time
     *
     * @return the Service Time in ms
     */
    public static float getDiskServiceTimeMillis(long deltaBytes, long deltaNumRequest, NetAppDiskSpecification diskSpecification) {
        final float diskAccessTimeMS = getDiskAccessTimeMillis(diskSpecification);
        final float diskTransferTimeMS = getDiskTransferTimeMillis(deltaBytes, deltaNumRequest, diskSpecification);
        return deltaNumRequest * (diskAccessTimeMS + diskTransferTimeMS);
    }

    /**
     * Although we calculated correctly the write for each raid I believe that we made another mistake there, that fixing
     * it will make our life easier at the long run. Our calculation are based on a bare bone interaction with RAID
     * controller, but this is not the case with arrays. Modern arrays try to mitigate the penalty of writing by using a
     * cache. Let me try to explain it in steps.\
     * 1. Client submit a write request to the array
     * 2. Assume that the cache is not propagate with the data this is a cache miss, than a read from all the data stripes in performed (as we calculated for the first part of the RAID 4,5 and 6)
     * 3. One step 2 is completed the cache represent the data therefore the client can complete the write to the cache
     * 4. The array send ACK to the client
     * 5. In offline the array will calculate the XOR and write the data to the relevant place. This mean that when
     * calculating response time we need to calculate it between step 1-4 and exclude step 5 So basically for every X
     * IO it should take X/(disks-1) and not X+2X/(disks-1)
     */
    public static float getVolumeWriteServiceTimeMillis(long deltaBytes, long deltaNumRequest, Collection<NetAppDiskSpecification> diskSpecifications) {
        float res = 0;
        for (NetAppDiskSpecification diskSpecification : diskSpecifications) {
            res = Math.max(res, getDiskServiceTimeMillis(deltaBytes, deltaNumRequest, diskSpecification));
        }

        // Write formula is X/(disks-1) where X - IOs, disks- number of disks
        // RAID5 specific calculation is taken as an average for writes
        int disks = diskSpecifications.size() > 1 ? diskSpecifications.size() : 1;
        return res / disks;
    }

    /**
     * Calculates amount of time spent to service all of the IO.
     * Calculated based on disk specifications (rather than read from device)
     * will provide the slowest disk service time
     *
     * @param deltaBytes
     * @param deltaNumRequest
     * @param diskSpecifications
     *
     * @return
     */
    public static float getVolumeServiceTimeMillis(long deltaBytes, long deltaNumRequest, Collection<NetAppDiskSpecification> diskSpecifications) {
        float res = 0;
        for (final NetAppDiskSpecification diskSpecification : diskSpecifications) {
            final float diskServiceTimeMS = getDiskServiceTimeMillis(deltaBytes, deltaNumRequest, diskSpecification);
            res = Math.max(res, diskServiceTimeMS);
        }

        // Read formula is X/disks where X = IOs, disks= number of disks
        final int numDisks = diskSpecifications.size();
        return (res / numDisks);
    }

    /**
     * retrive disk access time in ms (AccessTime(ms)=(SeekTime(ms)+Latency(ms))
     *
     * @param diskSpecification - disk specification
     *
     * @return the disk access time in ms AccessTime(ms)=(SeekTime(ms)+Latency(ms))
     */
    public static float getDiskAccessTimeMillis(NetAppDiskSpecification diskSpecification) {
        final float diskAverageLatancyMS = getDiskAverageLatencyMillis(diskSpecification);
        return (diskAverageLatancyMS + diskSpecification.getSeekTimeMs());
    }

    /**
     * Latency is time in ms it takes disk to make full spin, this method is
     * returning average latency in ms
     *
     * @param diskSpecification - disk specification
     *
     * @return the disk latnecy(in ms)
     */

    private static float getDiskAverageLatencyMillis(NetAppDiskSpecification diskSpecification) {
        //Convert the revolution per minute to revolution per millisecond
        float speed = ((float) diskSpecification.getSpeed()) / (60000F);
        if (speed == 0)
            return 0;

        // 1 revolution  / by revolution per millisecond  = time for 1 revolution
        // divid by 2 will provide the time of half a revolution.
        return ((1 / speed) / 2F);
    }

}
