package com.onaro.sanscreen.acquisition.schema;

import com.onaro.sanscreen.acquisition.framework.datasource.io.DSFileCRCWriter;

import java.io.IOException;
import java.util.Collection;


public abstract class BaseCRCXMLNode<P> extends XMLNode<P> implements Node<P> {

    protected BaseCRCXMLNode(final P parent, final String name) {
        super(parent, name);
    }

    // CRC infrastructure
    /**
     * Marshal the object into the appropriate XML schema.
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshallAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer) throws SchemaException, IOException {
        marshallAndCRC(cache, writer, 0);
    }

    /**
     * Marshal the object into the appropriate XML schema.
     * <p/>
     * Usually this method should not be overriden. Insead, override {@link #marshallAttributes}, {@link #marshallChildren} and {@link #hasChildren}
     * to handle the object marshalling.
     *
     * @param cache Object passed during marshalling that can be used to pass information between marshalled objects
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshallAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final int tabs) throws IOException, SchemaException {
        addTabulation(writer, tabs);
        writer.append("<").append(getNodeName()).append(' ');

        marshallAttributesToFileAndCRC(cache, writer);

        if (!hasChildren()) {
            writer.append("/>").append(lineSeperator);
        }
        else {
            writer.append(">").append(lineSeperator);
            marshallChildrenToFileAndCRC(cache, writer, tabs + 1);
            addTabulation(writer, tabs);
            writer.append("</").append(getNodeName()).append(">").append(lineSeperator);
        }
    }

    protected void marshallAttributesToFileAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer) throws IOException, SchemaException {
        // Default implementation marshals with no attributes
    }

    protected void marshallChildrenToFileAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final int tabs) throws IOException, SchemaException {
        // Default implementation marshals with no children
    }

    /**
     * Marshall a required schema object attribute to the given writer, and update both CRCs
     * @param writer the writer to marshall to
     * @param attributeName the name of the attribute to marshall
     * @param attributeValue the value of the attribute to marshall
     * @throws IOException on IO error
     */
    protected final void addOptionalAttributeAndCRC(final DSFileCRCWriter writer, final String attributeName, final Object attributeValue) throws IOException {
        if (attributeValue == null)
            return;
        final String escapedValue = escape(String.valueOf(attributeValue));
        writer.addOptionalAttributeAndCRC(attributeName, escapedValue);
    }

    /**
     * Marshall an optional schema object attribute to the given writer, and update CRC representing all attributes
     * @param writer the writer to marshall to
     * @param attributeName the name of the attribute to marshall
     * @param attributeValue the value of the attribute to marshall
     * @throws IOException on IO error
     */
    protected final void addRequiredAttributeAndCRC(final DSFileCRCWriter writer, final String attributeName, final Object attributeValue) throws IOException {
        if (attributeValue == null)
            return;
        final String escapedValue = escape(String.valueOf(attributeValue));
        writer.addRequiredAttributeAndCRC(attributeName, escapedValue);
    }

    protected final void addAttributeNoCRC(final DSFileCRCWriter writer, final String attributeName, final Object attributeValue) throws IOException {
        if (attributeValue == null)
            return;
        final String escapedValue = escape(String.valueOf(attributeValue));
        writer.addAttributeNoCRC(attributeName, escapedValue);
    }

    protected void addAllChildrenUnderNodeAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final String nodeName, final Collection<? extends BaseCRCXMLNode> childList, final int tabs) throws IOException, SchemaException {
        //Dont print container element if the child list is empty
        //this will save us the check for every container element
        if (childList == null || childList.size() == 0)
            return;

        addTabulation(writer, tabs);
        writer.append("<").append(nodeName).append(">").append(lineSeperator);

        for (final BaseCRCXMLNode<P> node : childList)
            node.marshallAndCRC(cache, writer, tabs + 1);

        addTabulation(writer, tabs);
        writer.append("</").append(nodeName).append(">").append(lineSeperator);
    }

    protected final void addChildAndCRC(final MarshallingCache cache, final DSFileCRCWriter writer, final BaseCRCXMLNode childNode, final int tabs) throws SchemaException, IOException {
        if (childNode == null)
            return;

        childNode.marshallAndCRC(cache, writer, tabs);
    }
}
