package com.onaro.sanscreen.acquisition.schema;

import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Map;

public abstract class KeyBase<E extends Enum<E>> implements Key<KeyBase<E>> {
    private final String keyType;
    private final KeyIDs<E> idValues;

    public KeyBase(final String keyType, final EnumMap<E, String> idValues) {
        this(keyType, idValues, null);
    }

    public KeyBase(final String keyType, final EnumMap<E, String> idValues, final EnumSet<E> emptyAbleFields) {
        this.keyType = keyType;
        this.idValues = new KeyIDs<E>(idValues, emptyAbleFields);
    }

    /**
     * Get the OCIE UUID for this key.
     * In many cases, the OCIE ID is a dot separated concatenation of multiple, somewhat arbitrary fields,
     * that may or may not be part of the public ResourceKey fields.   Each schema object must override
     * this method to build the correct UUID as defined by the model view UUID descriptions.
     * Relationships need a unique key for identification in Maps, but usually don't martial their
     * own OCIE UUID, instead relationships marshall the OCIE UUIDs of the related objects.
     * Relationships generall implement getOcieUUID by calling getFQID. getOcieUUID is not generally invoked
     * on relationship objects.
     *
     * Eventually OCIE UUIDs will go away in favor of the ResourceKey values.   In the mean time, both
     * most co-exist.
     *
     * The previous implementation of returning all ordered key fields except the fisrt (cluster UUID) field
     * was a super bad idea because the Resource Key based field sets and ordering often had nothing to
     * do with the actual OCIE UUID expected by the server.
     *
     * @return the OCIE UUID for this key
     */
    public abstract String getOcieUUID();

    /**
     * Get any field of the key by field name
     * @param key the field name
     * @return the field value
     */
    public String getID(final E key) {
        return idValues.idValueMap.get(key);
    }

    /**
     * Get the ResourceKey form of the key.   This value is defined at:
     *  http://wikid.netapp.com/w/MPO/Cross_Teams_Alignment/Unified_Keys
     * @return the ResourceKey form of the key.
     */
    public String getResourceKey() {
        StringBuilder resourceKey = new StringBuilder(128);
        for (final Map.Entry<E, String> fieldEntry : idValues.idValueMap.entrySet()) {
            final String value = fieldEntry.getValue();

            // Assume first field is the already rendered root (cluster)
            if (resourceKey.length() == 0) {
                // All resource keys start with: <cluster_uuid>:type=<type>
                resourceKey.append(value).append(":type=").append(keyType);
            } else {
                // All resource keys continue with CSV list of <key>=<val> for all subsequent key field values
                final String cleanValue = cleanKeyFieldValue(value);
                final E key = fieldEntry.getKey();
                resourceKey.append(',').append(key).append('=').append(cleanValue);
            }
        }
        return resourceKey.toString();
    }

    protected String cleanKeyFieldValue(final String keyFieldValue) {
        String cleanKeyFieldValue = new String(keyFieldValue);
        boolean needsQuotes = false;

        // Do we need quotes?
        if (cleanKeyFieldValue.isEmpty()) {
            needsQuotes = true;
        } else if (cleanKeyFieldValue.contains(" ")
                || cleanKeyFieldValue.contains(",")
                || cleanKeyFieldValue.contains("=")
                || cleanKeyFieldValue.contains(":")
                || cleanKeyFieldValue.contains("*")
                || cleanKeyFieldValue.contains("?")) {
            needsQuotes = true;
        }
        // Escape back-slashes
        if (cleanKeyFieldValue.contains("\\")) {
            cleanKeyFieldValue = cleanKeyFieldValue.replaceAll("\\\\", "\\\\\\\\");
        }
        // Escape quotes
        if (cleanKeyFieldValue.contains("\"")) {
            needsQuotes = true;
            cleanKeyFieldValue = cleanKeyFieldValue.replaceAll("\"", "\\\\\\\"");
        }
        // Surround with quotes if needed
        if (needsQuotes) {
            cleanKeyFieldValue = "\"" + cleanKeyFieldValue + "\"";
        }
        return cleanKeyFieldValue;
    }

    /**
     * Get the fully qualified ID as a String, consisting of all ID field values
     * from Root to Leaf separated by '.'
     *
     * @return the fully qualified ID as a String, consisting of all ID field
     *         values from Root to Leaf separated by '.'
     */
    protected String getFQID() {
        StringBuilder sb = new StringBuilder(128);
        for (String value : idValues.idValueMap.values()) {
            sb.append(value).append('.');
        }
        // trim trailing '.'
        if (sb.length() > 0) {
            sb.setLength(sb.length() - 1);
        }
        return sb.toString();
    }

    /**
     * Get the OCIE UUID value from the key if the key is not null, otherwise return null
     * @param k the key to get the UUID from, may be null
     * @return the OCIE UUID value from the key if the key is not null, otherwise return null
     */
    public static String getOcieUUIDOrNull(KeyBase<?> k) {
        return (k == null) ? null : k.getOcieUUID();
    }

    @Override
    public int compareTo(final KeyBase<E> o) {
        return idValues.compareTo(o.idValues);
    }

    @Override
    public boolean equals(final Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        final KeyBase<?> keyBase = (KeyBase<?>) o;
        if (idValues != null ? !idValues.equals(keyBase.idValues) : keyBase.idValues != null) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        return idValues != null ? idValues.hashCode() : 0;
    }

    @Override
    public String toString() {
        return getResourceKey();
    }

    public final class KeyIDs<E extends Enum<E>> implements Comparable<KeyIDs<E>> {
        private final EnumMap<E, String> idValueMap;
        final EnumSet<E> emptyAbleFields;

        private KeyIDs(final EnumMap<E, String> idValueMap, final EnumSet<E> emptyAbleFields) {
            if (idValueMap == null || idValueMap.size() < 1) {
                throw new IllegalArgumentException(keyType + " key fields may not be null or empty.");
            }
            if (emptyAbleFields == null) {
                // Stupid trick to create an empty EnumSet without having a reference to the Enum's class
                this.emptyAbleFields = EnumSet.copyOf(idValueMap.keySet());     // Create typed EnumSet
                this.emptyAbleFields.clear();                                   // Make sure it's empty
            } else {
                this.emptyAbleFields = EnumSet.copyOf(emptyAbleFields);         // Create a defensive copy of passed parameter
            }
            this.idValueMap = new EnumMap<E, String>(idValueMap);               // Create a defensive copy of passed parameter
            for (final Map.Entry<E, String> fieldEntry : idValueMap.entrySet()) {
                final E key = fieldEntry.getKey();
                final String value = fieldEntry.getValue();
                if (value == null) {
                    throw new NullPointerException(keyType + " key field '" + key + "' may not be null.");
                }

                // Only some key fields are allowed to be empty
                if (value.isEmpty()) {
                    if (!this.emptyAbleFields.contains(key)) {
                        throw new IllegalArgumentException(keyType + " key field '" + key + "' may not be empty.");
                    }
                }
            }
        }

        @Override
        public boolean equals(final Object o) {
            if (this == o) {
                return true;
            }
            if (o == null || getClass() != o.getClass()) {
                return false;
            }

            final KeyIDs<?> keyUtil = (KeyIDs<?>) o;
            if (!idValueMap.equals(keyUtil.idValueMap)) {
                return false;
            }
            return true;
        }

        @Override
        public int hashCode() {
            return idValueMap.hashCode();
        }

        @Override
        public int compareTo(final KeyIDs<E> o) {
            final int sizeCompare = idValueMap.size() - o.idValueMap.size();
            if (sizeCompare != 0) {
                return sizeCompare;
            }
            for (final Map.Entry<E, String> fieldEntry : idValueMap.entrySet()) {
                final E key = fieldEntry.getKey();
                final String thisVal = fieldEntry.getValue();
                final String otherVal = o.idValueMap.get(key);
                final int valCompare = thisVal.compareTo(otherVal);
                if (valCompare != 0) {
                    return valCompare;
                }
            }
            return 0;
        }
    }
}
