package com.onaro.sanscreen.acquisition.schema;

import java.util.*;
import java.io.Serializable;

/**
 * Utility class for representing list of nodes. This list implements the {@link Set} interface and behaves in terms of
 * performance like {@link TreeSet}. It performs parental checks to verify that list refers to elements from the same
 * parent.
 */
public class Nodes<Parent, N extends Node> implements Set<N>, Serializable {
    private final static long serialVersionUID = 1;
    private Parent parent;
    private final TreeSet<N> set = new TreeSet<N>(new NodeComparator());

    public Nodes(final Parent parent) {
        this.parent = parent;
    }

    public Parent getParent() {
        return parent;
    }

    public boolean add(N o) {
        if (!o.getParent().equals(parent))
            throw new IllegalArgumentException("Added child object to wrong parent object in Schema");

        return set.add(o);
    }

    public boolean addAll(Collection<? extends N> c) {
        for (N node : c) {
            if (!node.getParent().equals(parent))
                throw new IllegalArgumentException("Added child object to wrong parent object in Schema");
        }

        return (set.addAll(c));
    }

    public boolean isEmpty() {
        return set.isEmpty();
    }

    public boolean containsAll(final Collection<?> c) {
        return set.containsAll(c);
    }

    public boolean removeAll(final Collection<?> c) {
        return set.removeAll(c);
    }

    public boolean retainAll(final Collection<?> c) {
        return set.retainAll(c);
    }

    public boolean equals(final Object o) {
        return set.equals(o);
    }

    public int hashCode() {
        return set.hashCode();
    }

    public Iterator<N> iterator() {
        return set.iterator();
    }

    public boolean contains(final Object o) {
        //noinspection SuspiciousMethodCalls
        return set.contains(o);
    }

    public int size() {
        return set.size();
    }

    public boolean remove(final Object o) {
        //noinspection SuspiciousMethodCalls
        return set.remove(o);
    }

    public void clear() {
        set.clear();
    }

    public Object[] toArray() {
        return set.toArray();
    }

    public <T>T[] toArray(final T[] a) {
        return set.toArray(a);
    }

    public static class NodeComparator implements Comparator<Node>, Serializable {
        private final static long serialVersionUID = 1;
        public int compare(final Node o1, final Node o2) {
            return o1.getKey().compareTo(o2.getKey());
        }
    }
}
