package com.onaro.sanscreen.acquisition.schema;

import com.onaro.sanscreen.acquisition.framework.datasource.IntegrityException;

import java.io.IOException;
import java.io.Writer;

/**
 * Defines the basic functionality of a odata root object.
 *
 * <p/>Each acquisition report is sent to the SANscreen server in XML format. On the data source package implementation
 * side, building this report is done by utilizing the odata schema library (<code>com.onaro.sanscreen.acquisition.schema.odata2.*</code>).
 *
 * <p/>To start building an <code>odata</code> object tree, a <code>Root</code> object has to be created. Using the {@link BaseRootFactory}
 * factory class, root of different kinds can be created, depending on the type of report the data source will generate.
 * There is a specific root object for every type of report that can be generated by the package <code>BaseStorageRoot</code> for
 * example is used to report about a storage array to the server).
 *
 * <p/>A package implementation can choose to report more than one roots in a single sample (this is achieved by construction
 * a "Multiple Originator ID (MOID)" package).
 */
public interface Root extends Node<BaseRootFactory> {
    /**
     * Contains general information about the sample like timestamp, ID of data source, etc...
     *
     * @return Info object describing this XML
     */
    public Info getInfo();

    /**
     * Originator types are strings describing the type of the reporter. For example, a netapp package would say ""NetAppApi".
     *
     * <p/>The only importance of this value is to visually identify the data source reporting when looking at the XML file
     * generated by the sample.
     *
     * @return Originator type
     */
    public OriginatorType getOriginatorType();

    /**
     * The originator ID is a unique ID assigned by the server for this reporter. This ID is obtained by calling the
     * {@link com.onaro.sanscreen.acquisition.foundation.datasource.BaseFoundationPackage#getOriginatorID(String)} and
     * providing an Originator key.
     *
     * <p/>It is essential that the <u>same originator key</u> is used between samples (this will ensure the same Originator ID
     * is assigned to the package every time!).
     *
     * <p/>The server uses this ID to identify changes made by the same reporter and is using this information to merge reports
     * from different "Originators".
     *
     * @return originator ID
     */
    public Originator getOriginator();

    /**
     * The originator key is a device specific string that uniquely identifies this device. The package implementation uses
     * this string to obtain the same "Originator ID" for every sample.
     *
     * <p/>A typical originator key for a switch would be
     * the IP of this switch (this is a value that identifies the switch and that does not change between samples). Another
     * example would be the WWN of this switch.
     *
     * @return String representation of originator (device specific)
     */
    public String getOriginatorKey();

    public void setOriginatorKey(String originatorKey);

    /**
     * Used by the framework to convert the schema object hierarchy into XML. This is done when sending the report to
     * the server.
     *
     * @param writer Writer to dump XML to
     *
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    public void marshall(final Writer writer) throws IOException, SchemaException;

    /**
     * Perform integrity checks on device data
     *
     * @throws IntegrityException If there is a problem with the integrity of the XML report
     */
    public void checkIntegrity() throws IntegrityException ;
}
