package com.onaro.sanscreen.acquisition.schema;

import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.lang.ref.WeakReference;

/**
 * Base class for all o-data nodes. This object introduces the notion of a parent object since the o-data model is
 * tree based.
 */
public abstract class XMLNode<P> extends BaseXMLNode<MarshallingCache> implements Node<P> {
    private SerializableWeakReference<P> parent;

    protected XMLNode(final P parent, final String name) {
        super(name);

        this.parent = new SerializableWeakReference<P>(new WeakReference<P>(parent));
    }

    /**
     * Returns a reference (WEAK!!!) to the parent node. This is maintained as a weak reference to allow garbage collection
     * of the parent section of the tree if it is not necessary. To avoid this (only when needed), you can keep a strong
     * reference to the parent node.
     *
     * <p/>For example, keeping a variable pointing to the Root of the odata - will make sure
     * the entire tree is kept in memory. This is wanted in foundation implementaions but unwanted when keeping some of
     * the foundation objects in performance.
     *
     * @return The parent of this node (might be <code>null</code>) if there is no strong reference to the parent and
     * parent was garbage collected
     */
    public P getParent() {
        return parent.getWeakReference().get();
    }

    /**
     * Utiltiy method that compares between two object taking <code>null</code> values into account
     * @param o1 the first Object to be compared
     * @param o2 the second Object to be compared.
     *
     * @return  a negative integer, zero, or a positive integer as this object is less than, equal to, or greater than the specified object.
     */
    public static int compareObjects(Comparable o1, Comparable o2) {
        if (o1 == null && o2 == null)
            return 0;

        if (o1 != null && o2 == null)
            return 1;

        if (o1 == null /*&& o2 != null*/)
            return -1;

        //noinspection unchecked
        return o1.compareTo(o2);
    }

     /* Weak references (as used on the parent class) are not serializable.   As a work around, provide
     * a wrapper class that will attempt to serialize the reference of the weak reference instead.
     *
     * Use the serializable proxy pattern (Bloch, Effective Java 2nd Edition, p. 312 [Item 78]).  This approach
     * is non-standard, but seems to be the safest way to attempt serialization of a weak reference.
     */
    private static class SerializableWeakReference<T> implements Serializable {

        private WeakReference<T> reference;
        private static final long serialVersionUID = 1;

        SerializableWeakReference (final WeakReference<T> reference) {
            this.reference = reference;
        }

        public WeakReference<T> getWeakReference() {
            return reference;
        }


        // Nominate our proxy as a replacement serialization object for this class.
        private Object writeReplace() {
            return new SerializationProxy<T>(this);
        }

        //Read object is invalid given we have a writeReplace.  Make sure it doesn't work under any condition.
        private void readObject(ObjectInputStream stream) throws InvalidObjectException {
            throw new InvalidObjectException ("Serialization proxy required");
        }

        /**
         * Serialized proxy class for SerializableWeakReference.
         */
        private static class SerializationProxy<T> implements Serializable {

            private T weak;
            private static final long serialVersionUID = 1;

            SerializationProxy(SerializableWeakReference<T> swf) {
                this.weak = swf.getWeakReference().get();
            }

            //Override the de-serialized proxy class with the appropriate
            //SerializableWeakReference that we are proxy-ing instead.
            private Object readResolve() {
                  return new SerializableWeakReference<T>(new WeakReference<T>(weak));
            }

        }


    }
}
