package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

/**
 * EMC Symmetrix only.
 *
 * FAST-VP supports volumes that reside on multiple storage pools, moving "hot" and "cold" data to and
 * from faster/slower storage groups (groups of storage pools) based on a policy. A FAST policy defines the upper usage
 * limits for up to three tiers for any storage groups associated with the policy.
 *
 * A volume that participates in an auto tier will also have a reference one auto tier policy.
 */
public class AutoTierPolicy extends XMLNode<Storage> {
    private final AutoTierPolicyKey key;
    private String name = null;
    private final AutoTierPolicyConstraints policyConstraints;

    protected AutoTierPolicy(final Storage parent, final String id) throws SchemaException {
        super(parent, "AutoTierPolicy");

        if (id == null)
            throw new SchemaException(getNodeName() + " Type requires non-null id");

        this.key = new AutoTierPolicyKey(id, parent.getKey());
        policyConstraints = new AutoTierPolicyConstraints(this);
    }

    @SuppressWarnings({"unchecked"})
    public AutoTierPolicyKey getKey() {
        return key;
    }

    public String getId() {
        return key.getId();
    }

    public String getName() {
        return name;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public AutoTierPolicyConstraint createAutoTierPolicyConstraint(final String policyName) throws SchemaException {
        return new AutoTierPolicyConstraint(this, policyName);
    }

    public AutoTierPolicyConstraints getPolicyConstraints() {
        return policyConstraints;
    }

    @Override
    protected boolean hasChildren() {
        return super.hasChildren() || !getPolicyConstraints().isEmpty();
    }

    @Override
    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        super.marshallChildren(cache, writer, tabs);
        addAllChildrenUnderNode(cache, writer, "AutoTierPolicyConstraints", getPolicyConstraints().values(), tabs);
    }

    @Override
    protected void marshallAttributes(final MarshallingCache marshallingCache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(marshallingCache, writer);
        addAttribute(writer, "Identifier", getId());
        addAttribute(writer, "Name", getName());
    }

    public final static class AutoTierPolicyKey implements Key<AutoTierPolicyKey> {
        private final String id;
        private final Storage.StorageKey storageKey;

        public AutoTierPolicyKey(final String id, final Storage.StorageKey storageKey) {
            this.id = id;
            this.storageKey = storageKey;
        }

        public int compareTo(final AutoTierPolicyKey o) {
            int i = storageKey.compareTo(o.storageKey);
            if (i != 0)
                return i;

            i = id.compareTo(o.id);
            if (i != 0)
                return i;

            return 0;
        }

        public String getId() {
            return id;
        }

        public String toString() {
            return storageKey + ":" + id;
        }

        public Storage.StorageKey getStorageKey() {
            return storageKey;
        }

        @SuppressWarnings({"RedundantIfStatement"})
        public boolean equals(final Object o) {
            if (this == o) return true;
            if (!(o instanceof AutoTierPolicyKey)) return false;

            final AutoTierPolicyKey that = (AutoTierPolicyKey) o;

            if (id != null ? !id.equals(that.id) : that.id != null) return false;
            if (storageKey != null ? !storageKey.equals(that.storageKey) : that.storageKey != null) return false;
            return true;
        }

        public int hashCode() {
            int result;
            result = (id != null ? id.hashCode() : 0);
            result = 31 * result + (storageKey != null ? storageKey.hashCode() : 0);
            return result;
        }
    }
}
