package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

/**
 * Back end volume class represents a volume visible to the storage virtualizer on the "real"  storage array
 */
public class BackEndLun extends XMLNode<Storage> {
    private final String name;
    private String group;
    private final BackEndPaths backendPaths;
    private final BackEndLunKey key;
    private String uuid;

    public BackEndLun(final Storage parent, final String name) throws SchemaException {
        super(parent, "BackendLun");
        if (name == null || name.length() == 0)
            throw new SchemaException("Back end volume name cannot be null or empty");

        this.name = name;
        backendPaths = new BackEndPaths(this);

        key = new BackEndLunKey(parent.getKey(), name);
    }

    @SuppressWarnings({"unchecked"})
    public final BackEndLunKey getKey() {
        return key;
    }

    public final String getName() {
        return name;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public final String getGroup() {
        return group;
    }

    public final void setGroup(final String backEndVolumeGroup) {
        this.group = backEndVolumeGroup;
    }

    public String getUuid() {
        return uuid;
    }

    public void setUuid(final String uuid) {
        this.uuid = uuid;
    }

    /**
     * Creates a connection object that maps the virutalizer's BackEnd volume object to the backend storage.
     *
     * @param localPortWWN Virtualizer port (<code>null</code> if all ports are mapped)
     * @param remotePortWWN The backend storage port
     * @param lun LUN assigned during mapping
     *
     * @return BackEndConnection
     *
     * @throws com.onaro.sanscreen.acquisition.schema.SchemaException will be thrown when remote port is empty or null
     */
    public final BackEndPath createBackEndPath(final String localPortWWN, final String remotePortWWN, final long lun) throws SchemaException {
        return new BackEndPath(this, localPortWWN, remotePortWWN, lun);
    }                  

    public final BackEndPaths getBackendPaths() {
        return backendPaths;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        addAttribute(writer, "Name", name);
        addAttribute(writer, "UUID", uuid);
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNode(cache, writer, "BackendPaths", backendPaths, tabs);
    }

    protected boolean hasChildren() {
        return backendPaths.size() > 0;
    }

    public static class BackEndLunKey implements Key<BackEndLunKey> {
        private final String name;
        private final Storage.StorageKey storageKey;

        public BackEndLunKey(final Storage.StorageKey storageKey, final String name) {
            this.name = name;
            this.storageKey = storageKey;
        }

        public int compareTo(final BackEndLunKey o) {
            final int i = storageKey.compareTo(o.getStorageKey());
            if (i != 0) return i;

            return getName().compareTo(o.getName());
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final BackEndLunKey that = (BackEndLunKey) o;

            if (name != null ? !name.equals(that.name) : that.name != null) return false;
            if (storageKey != null ? !storageKey.equals(that.storageKey) : that.storageKey != null) return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (name != null ? name.hashCode() : 0);
            result = 31 * result + (storageKey != null ? storageKey.hashCode() : 0);
            return result;
        }

        public String toString() {
            return storageKey.toString() + ":" + name;
        }

        public String getName() {
            return name;
        }

        public Storage.StorageKey getStorageKey() {
            return storageKey;
        }
    }
}
