package com.onaro.sanscreen.acquisition.schema.odata2;
/**
 *   this table document what typs of volume we have and there attributes.
 *
 *          Storage Pool | Internal Volume | Map/Mask | Volume Members | ratio
 *  ------|---------------------------------------------------------------------
 *  Meta        no               no            yes          yes           members
 *  ------|--------------|-----------------|----------|----------------|--------
 * member |    yes       |       no        |   no     |     no         |  storage pool
 * -------|----------------------------------------------------------------------------
 *internal|    yes       |       no        |   no     |     no         |  storage pool
 * -------|----------------------------------------------------------------------------
 * volume |    yes       |       yes       |   ys     |     no         |  storage pool
 *
 *
 *
 */

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;
import java.util.SortedSet;
import java.util.TreeSet;

public abstract class BaseVolume<P extends XMLNode> extends XMLNode<P> {
    protected final String name;
    protected String label = null;
    protected String type = null;
    protected Boolean replicaSource = null;
    protected Boolean replicaTarget = null;
    protected Boolean snapshot = null;

    protected String diskType = null;
    protected String diskGroup = null;


    protected Boolean thinProvisioned = false;
    /**
     * Set to true if volume is "Virtual Volume" on the storage virtualizer
     */
    protected Boolean virtual = null;

    // Auto Tiering attributes
    /**
     * true indicates that the volume is participating in auto tiering (EMC specific)
     */
    private Boolean autoTiering = null;

    /**
     * The id of the tier policy applied to the volume (EMC specific)
     * Vendor Tiers are not yet modeled, just the reference to the tier name
     */
    private AutoTierPolicy.AutoTierPolicyKey autoTierPolicyKey = null;


    protected final SortedSet<Integer> diskSizes;
    protected final SortedSet<Integer> diskSpeeds;

    protected final boolean drPackageInstalled;

    protected StoragePool storagePool;
    protected final VolumeKey key;
    protected Qtree qtree;


    protected Long lastKnownAccessTime = null;
    protected String virtualStorage = null;
    protected String head;
    protected String protectionType;
    private String uuid;

    protected BaseVolume(final P parent, final Storage storage, final String nodeName, final String name, final StoragePool storagePool, final Qtree qtree, final boolean drPackageInstalled) throws SchemaException {
        super(parent, nodeName);
        this.drPackageInstalled = drPackageInstalled;

        if (name == null || name.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null name");

        this.name = name;

        diskSizes = new TreeSet<Integer>();
        diskSpeeds = new TreeSet<Integer>();

        key = new VolumeKey(storage.getKey(), name);

        this.storagePool = storagePool;
        this.qtree = qtree;
    }

    @SuppressWarnings({"unchecked"})
    public final VolumeKey getKey() {
        return key;
    }

    /**
     * @return the storage associated with this volume / volume member
     */
    public abstract Storage getStorage();

    @SuppressWarnings({"UnusedDeclaration"})
    public final String getDiskGroup() {
        return diskGroup;
    }

    public Boolean isThinProvisioned() {
        return thinProvisioned;
    }

    public void setThinProvisioned(final Boolean thinProvisioned) {
        this.thinProvisioned = thinProvisioned;
    }

    public final void setDiskGroup(final String diskGroup) {
        this.diskGroup = diskGroup;
    }

    public final boolean isDrPackageInstalled() {
        return drPackageInstalled;
    }

    public String getLabel() {
        return label;
    }

    public final void setLabel(final String label) {
        this.label = label;
    }

    public final String getName() {
        return name;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public final boolean isReplicaSource() {
        return (replicaSource != null) && replicaSource;
    }

    public final void setReplicaSource(final boolean replicaSource) {
        this.replicaSource = replicaSource;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public final boolean isReplicaTarget() {
        return (replicaTarget != null) && replicaTarget;
    }

    public final void setReplicaTarget(final boolean replicaTarget) {
        this.replicaTarget = replicaTarget;
    }

    public final Boolean isSnapshot() {
        return (snapshot != null) && snapshot;
    }

    public final void setSnapshot(final Boolean snapshot) {
        this.snapshot = snapshot;
    }

    public final String getType() {
        return type;
    }

    public final void setType(final String type) {
        this.type = type;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public Long getLastKnownAccessTime() {
        return lastKnownAccessTime;
    }

    public void setLastKnownAccessTime(final Long lastKnownAccessTime) {
        this.lastKnownAccessTime = lastKnownAccessTime;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getVirtualStorage() {
        return virtualStorage;
    }

    public void setVirtualStorage(final String virtualStorage) {
        this.virtualStorage = virtualStorage;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getHead() {
        return head;
    }

    public void setHead(final String head) {
        this.head = head;
    }

    public final String getDiskType() {
        return diskType;
    }

    public final void setDiskType(final String diskType) {
        this.diskType = diskType;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getProtectionType() {
        return protectionType;
    }

    public void setProtectionType(final String protectionType) {
        this.protectionType = protectionType;
    }

    /**
     * @return All disk sizes (GB) being used for this volume
     */
    public final SortedSet<Integer> getDiskSizes() {
        return diskSizes;
    }

    /**
     * @return All disk speeds (rpm) being used for this volume.
     */
    public final SortedSet<Integer> getDiskSpeeds() {
        return diskSpeeds;
    }

    /**
     * Returns set to true if volume is "Virtual Volume" on the storage virtualizer
     *
     * @return set to true if volume is "Virtual Volume" on the storage virtualizer
     */
    public final boolean isVirtual() {
        return (virtual != null) && virtual;
    }

    /**
     * EMC Symmetrix only
     * returns true if the volume is participating in auto tiering
     * @return may return null
     */
    public Boolean getAutoTiering() {
        return autoTiering;
    }

    /**
     * EMC Symmetrix only.
     *
     * If setting to true, the storage pool must also have autoTiering set to true.
     *
     * @param autoTiering set to true to indicate that the volume is participating in auto tiering
     */
    public void setAutoTiering(final Boolean autoTiering) {
        this.autoTiering = autoTiering;
    }

    /**
     * EMC Symmetrix only.
     *
     * gets the id of the auto tier policy assigned to the volume
     * @return The auto tier policy key or null
     */
    @SuppressWarnings({"UnusedDeclaration"})
    public AutoTierPolicy.AutoTierPolicyKey getAutoTierPolicyKey() {
        return autoTierPolicyKey;
    }

    /**
     * EMC Symmetrix only.
     * 
     * sets the id of the auto tier policy assigned to the volume
     *
     * @param autoTierPolicyKey the policy key.
     */
    public void setAutoTierPolicyKey(final AutoTierPolicy.AutoTierPolicyKey autoTierPolicyKey) {
        this.autoTierPolicyKey = autoTierPolicyKey;
    }

    public Qtree getQtree() {
        return qtree;
    }

    /**
     * Sets set to true if volume is "Virtual Volume" on the storage virtualizer
     *
     * @param virtual set to true if volume is "Virtual Volume" on the storage virtualizer
     */
    public final void setVirtual(final Boolean virtual) {
        this.virtual = virtual;
    }

    /**
     * @return Storage Pool directly under the Volume.
     *         This might be null if the volume sits on an internal volume.
     */
    public StoragePool getStoragePool() {
        return storagePool;
    }

    /**
     * @return Underlying Storage Pool for the Volume.
     *         Can be StoragePool or the Storage Pool of the InternalVolume.
     */
    public StoragePool getUnderlyingStoragePool() {
        if (storagePool != null) {
            return storagePool;
        }
        else {
            return qtree.getStoragePool();
        }
    }


    public void setStoragePool(final StoragePool storagePool) {
        this.storagePool = storagePool;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getUuid() {
        return uuid;
    }

    public void setUuid(final String uuid) {
        this.uuid = uuid;
    }

    protected void marshallAttributes(final MarshallingCache marshallingCache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(marshallingCache, writer);

        addAttribute(writer, "Name", getName());
        addAttribute(writer, "Label", getLabel());
        addAttribute(writer, "VirtualStorage", virtualStorage);
        addAttribute(writer, "Head", head);
        addAttribute(writer, "ProtectionType", protectionType);

        //this is a workaround
        //todo MetaVolume should be refactored 
        if (!(this instanceof MetaVolume))
            addAttribute(writer, "StoragePoolId", getUnderlyingStoragePool().getStoragePoolId());

        if (qtree != null)
            addAttribute(writer, "QtreeId", qtree.getKey());

        // Auto Tiering attributes
        addAttribute(writer, "AutoTiering", autoTiering);
        if (autoTierPolicyKey != null){
            addAttribute(writer, "AutoTierPolicyIdentifier", autoTierPolicyKey.getId());
        }

        addAttribute(writer, "UUID", uuid);
    }

    protected final String convertToCSV(final SortedSet<Integer> listOfDiskSize) {
        if (listOfDiskSize.size() == 0)
            return null;

        final StringBuilder sb = new StringBuilder();
        for (final Integer memberDiskSize : listOfDiskSize)
            sb.append(memberDiskSize.toString()).append(",");

        return sb.deleteCharAt(sb.length() - 1).toString();
    }

    public final static class VolumeKey implements Key<VolumeKey> {
        private final Storage.StorageKey storageKey;
        private final String name;

        public VolumeKey(final Storage.StorageKey storageKey, final String name) {
            this.storageKey = storageKey;
            this.name = name;
        }

        public final int compareTo(final BaseVolume.VolumeKey o) {
            final int i = getStorageKey().compareTo(o.getStorageKey());
            if (i != 0)
                return i;

            return getName().compareTo(o.getName());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final VolumeKey volumeKey = (VolumeKey) o;

            if (name != null ? !name.equals(volumeKey.name) : volumeKey.name != null) return false;
            //noinspection RedundantIfStatement
            if (storageKey != null ? !storageKey.equals(volumeKey.storageKey) : volumeKey.storageKey != null)
                return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (storageKey != null ? storageKey.hashCode() : 0);
            result = 29 * result + (name != null ? name.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return storageKey.toString() + ":" + getName();
        }

        public final String getName() {
            return name;
        }

        public final Storage.StorageKey getStorageKey() {
            return storageKey;
        }
    }
}
