package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;

@SuppressWarnings({"JavaDoc", "UnusedDeclaration"})
public class Disk extends XMLNode<Storage> {
    private final String id;
    private final String name;

    private Integer speed;
    /**
     * Capacity of disk - must be positive!
     */
    private Capacity capacity = Capacity.create();
    private String location;
    private String role;
    private DiskRole diskRoleEnum;
    private String model;
    private String vendor;
    private String vendorID;
    private String type;
    private String group;
    private Float seekTime;
    private Float potentialTransferRate;

    private final DiskKey key;
    private DiskType diskTypeEnum;
    private String status;
    private DiskStatus statusEnum;
    private String serialNumber;

    /**
     * Use {@link Storage#createDisk} to create this object
     */
    Disk(final Storage parent, final String id, final String name) throws SchemaException {
        super(parent, "Disk");

        if (id == null || id.length() == 0 || name == null || name.length() == 0)
            throw new SchemaException(getNodeName() + " Type require non-null name, id");

        this.id = id;
        this.name = name;

        key = new DiskKey(id, name);
    }

    public final Storage getStorage() {
        return getParent();
    }

    @SuppressWarnings({"unchecked"})
    public final DiskKey getKey() {
        return key;
    }

    public DiskType getDiskTypeEnum() {
        return diskTypeEnum;
    }

    public void setDiskTypeEnum(final DiskType diskTypeEnum) {
        this.diskTypeEnum = diskTypeEnum;
    }

    public final String getID() {
        return id;
    }

    /**
     * @return disk capacity (GB)
     */
    public final Capacity getCapacity() {
        return capacity;
    }

    /**
     * Set disk capacity (GB)
     *
     * @param capacity
     */
    public final void setCapacity(final Capacity capacity) {
        this.capacity = Capacity.create(capacity);
    }

    /**
     * @return Disk speed (RPM)
     */
    public final Integer getSpeed() {
        return speed;
    }

    public DiskRole getDiskRoleEnum() {
        return diskRoleEnum;
    }

    public void setDiskRoleEnum(final DiskRole diskRoleEnum) {
        this.diskRoleEnum = diskRoleEnum;
    }

    /**
     * Set disk speed (RPM)
     *
     * @param speed
     */
    public final void setSpeed(final Integer speed) {
        this.speed = speed;
    }

    public final String getLocation() {
        return location;
    }

    public final void setLocation(final String location) {
        this.location = location;
    }

    /**
     * Disk Role: Data / Parity / Hot Spare, etc...
     */
    public final String getRole() {
        return role;
    }

    /**
     * Disk Role: Data / Parity / Hot Spare, etc...
     * we are setting to known role an enum all the unknown disk Roles will use OTHER enum
     */
    public final void setRole(final String role) {
        this.role = role;
    }

    public final String getModel() {
        return model;
    }

    public final void setModel(final String model) {
        this.model = model;
    }

    public final String getVendor() {
        return vendor;
    }

    public final void setVendor(final String vendor) {
        this.vendor = vendor;
    }

    /**
     * Vendor ID is a string identifying this disk exact model (should be used when trying to query disk's technical spec)
     */
    public final String getVendorID() {
        return vendorID;
    }

    /**
     * Vendor ID is a string identifying this disk exact model (should be used when trying to query disk's technical spec)
     */
    public final void setVendorID(final String vendorID) {
        this.vendorID = vendorID;
    }

    /**
     * Disk type describes the type technology used in this disk like: SCSI, SATA, FATA, Fibre channel, etc..
     */
    public final String getType() {
        return type;
    }

    /**
     * Disk type describes the type technology used in this disk like: SCSI, SATA, FATA, Fibre channel, etc..
     */
    public final void setType(final String type) {
        this.type = type;
    }

    /**
     * @return The group associated with this disk
     */
    public final String getGroup() {
        return group;
    }

    /**
     * Sets the group associated with this disk. Groups can be indicate which raid group this disk belong to, or indicate
     * a set of disks allocated to certain storage pools, etc...
     *
     * @param group
     */
    public final void setGroup(final String group) {
        this.group = group;
    }

    public void setSeekTime(final float seekTime) {
        this.seekTime = seekTime;
    }

    public void setPotentialTransferRate(final float potentialTransferRate) {
        this.potentialTransferRate = potentialTransferRate;
    }

    public float getSeekTime() {
        return seekTime;
    }

    public float getPotentialTransferRate() {
        return potentialTransferRate;
    }

    public String getStatus() {
        return status;
    }

    public void setStatus(final String status) {
        this.status = status;
    }

    public DiskStatus getStatusEnum() {
        return statusEnum;
    }

    public void setStatusEnum(final DiskStatus statusEnum) {
        this.statusEnum = statusEnum;
    }

    public String getSerialNumber() {
        return serialNumber;
    }

    public void setSerialNumber(final String serialNumber) {
        this.serialNumber = serialNumber;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "DiskId", id);
        addAttribute(writer, "Name", name);
        addAttribute(writer, "Speed", speed);
        addAttribute(writer, "CapacityGB", Capacity.valueForXML(capacity, Capacity.Unit.GB_1024));
        addAttribute(writer, "Location", location);
        addAttribute(writer, "RoleEnum", diskRoleEnum);
        addAttribute(writer, "Role", (role == null && diskRoleEnum != null) ? diskRoleEnum : role);
        addAttribute(writer, "Status", status);
        addAttribute(writer, "StatusEnum", statusEnum);
        addAttribute(writer, "SerialNumber", serialNumber);
        addAttribute(writer, "Vendor", vendor);
        addAttribute(writer, "VendorId", vendorID);
        addAttribute(writer, "Model", model);


        // Type is taken as set by the datasource. The TypeEnum should be taken from the list of valid enums.
        // It is validated by the server. The convertDiskType
        addAttribute(writer, "Type", type);
        addAttribute(writer, "TypeEnum", (diskTypeEnum != null) ?
                                         convertDiskType(diskTypeEnum.name()) :
                                         convertDiskType(type));
        addAttribute(writer, "Group", group);
        addAttribute(writer, "SeekTime", seekTime);
        addAttribute(writer, "PotentialTransferRate", potentialTransferRate);
    }

    public final static class DiskKey implements Key<DiskKey> {
        private final String id;
        private final String name;

        public DiskKey(final String id, final String name) {
            this.id = id;
            this.name = name;
        }

        public int compareTo(final DiskKey o) {
            int i;

            i = getId().compareTo(o.getId());
            if (i != 0)
                return i;

            return getName().compareTo(o.getName());
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final DiskKey diskKey = (DiskKey) o;

            if (id != null ? !id.equals(diskKey.id) : diskKey.id != null) return false;
            if (name != null ? !name.equals(diskKey.name) : diskKey.name != null) return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (id != null ? id.hashCode() : 0);
            result = 29 * result + (name != null ? name.hashCode() : 0);
            return result;
        }

        public final String getId() {
            return id;
        }

        public final String getName() {
            return name;
        }

        public final String toString() {
            return getId() + ":" + getName();
        }
    }

    public static enum DiskRole {
        DATA,
        PARITY,
        SPARE,
        UNALLOCATED,
        FAULTY,
        OTHER
    }

    //please do not modify without communicating to server/dwh team.
    // NOTE: order of enums is in increasing performance
    public static enum DiskType {
        OTHER,
        ATA,
        FATA,
        SATA,
        SATA2,
        SATA3,
        FC,
        SAS,
        SSD,
        // SCSI,   The server does not accept SCSI as a disk type.
        //todo we will add "PATA" Type after understanding it
        //   PATA
    }

    /*
    An optional attribute for disk
    */

    public static enum DiskStatus {
        NORMAL,
        FAILED,
        QUESTIONABLE,
        ZEROING,
        COPYING,
        PENDING,
        PARTNER,
        RECONSTRUCTING,
        OTHER
    }

    /**
     * Converts a string to its valid enum value. Returns FC for FCAL and SATA for NL.
     * Otherwise, it returns OTHER.
     *
     * @param diskType
     *
     * @return {@link Disk.DiskType} enum for the given string.
     */
    private Disk.DiskType convertDiskType(final String diskType) {
        if (diskType == null)
            return null;


        Disk.DiskType diskTypeEnum = null;
        try {
            diskTypeEnum = Disk.DiskType.valueOf(diskType.toUpperCase());
            return diskTypeEnum;
        }
        catch (IllegalArgumentException e) {
            //fall thru

        }

        // we could not resolve the disk type  to enum
        assert (diskTypeEnum == null);

        // NL is "nearline ready SATA" -> SATA
        if (diskType.equals("NL")) {
            return Disk.DiskType.SATA;
        }

        // FCAL is "FC over Arbitrated Loop" -> FC
        if (diskType.equals("FCAL")) {
            return Disk.DiskType.FC;
        }

        // We don't recognize the interface
        return Disk.DiskType.OTHER;
    }


}
