package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.io.Writer;

/**
 * Object representing a many node to one switch port connection, This can be a Brocade Access Gateway,
 * a Cisco NPV mode switch, an NPIV connection, or an arbitrated loop.
 *
 * <p/> This object extends {@link ConnectedPort}, so using it it like any other connected port - calling {@link
 * Port#setConnectedPort}. Creating this object is done using {@link SwitchPort#createHub}.
 *
 * <p/> Note: If we have more than one node port in the hub, the hub will be represented as a separate switch.
 * The ports on this switch that connect to the node ports will be named as switchPortWWN-nodePortWWN.
 * If this is a separate device that can be acquired by sanscreen, and if it is acquired, these ports will be
 * replaced by those reported by the direct acquisition of the device.
 * In case we do not know the physically connected port to the switch port (connectedNpvPortWWN), we will generate
 * the port name as Hub_switchPortWWN.
 */
public class Hub extends ConnectedPort {
    private final HubPorts nodePorts;
    private String connectedNpvPortWWN;   // port of the NPV device that is connected to the switch port of this Hub
    private String connectedNpvNodeWWN;

    /**
     * Use {@link com.onaro.sanscreen.acquisition.schema.odata2.SwitchPort#createHub} to create this object
     * @param port the switch port
     * @throws com.onaro.sanscreen.acquisition.schema.SchemaException throws when the fake hub port is invalid
     */
    Hub(final SwitchPort port) throws SchemaException {
        super(port, "HUB_" + port.getWWN());
        nodePorts = new HubPorts(this);
    }

    /**
     * The port on the switch to which this hub is connected.
     * @return The port on the switch connected to this hub.
     */
    public SwitchPort getSwitchPort() {
        //noinspection unchecked
        return (SwitchPort) getParent();
    }

    /**
     * Node ports that are reported in the name server as connected to the switch port
     * and are not Physical NPV or Access Gateway links.
     * @return
     */
    public HubPorts getNodePorts() {
        return nodePorts;
    }

    public HubPort createPort(final String wwn) throws SchemaException {
        return new HubPort(this, wwn);
    }

    /**
     * Indicates that this connected port (Hub) object should represent
     * an NPV device and that the port on that device connected to the
     * core switch has wwn npvPortWWN.
     * If this is set, then a new switch will be added to the fabric containing
     * a port with the given wwn connected to the core switch port of the Hub,
     * a port for each hub/host port of this Hub connected to the associated
     * host port unknows. 
     *
     * @param connectedNpvPortWWN - wwn of the port on the NPV device connected to the core switch port of this Hub
     * @throws SchemaException - if our model construction is in error
     */
    public void setConnectedNpvPortWWN(final String connectedNpvPortWWN) throws SchemaException {

        if (connectedNpvPortWWN == null) {
            throw new SchemaException("Cannot set a null NPV port");
        }
        assert this.connectedNpvPortWWN == null;
        this.connectedNpvPortWWN = connectedNpvPortWWN;

    }

    public String getConnectedNpvPortWWN() {
        return connectedNpvPortWWN;
    }

    /**
     * Indicates that this Hub object should represent
     * an NPV device with wwn connectedNpvNodeWWN.
     *
     * @param connectedNpvNodeWWN - wwn of the NPV device connected to the core switch port of this Hub
     * @throws SchemaException - if our model construction is in error
     */
    public void setConnectedNpvNodeWWN(final String connectedNpvNodeWWN) throws SchemaException {
        if (connectedNpvNodeWWN == null) {
            throw new SchemaException("Cannot set a null NPV node WWN");
        }
        assert this.connectedNpvNodeWWN == null;
        this.connectedNpvNodeWWN = connectedNpvNodeWWN;
    }

    public String getConnectedNpvNodeWWN() {
        return connectedNpvNodeWWN;
    }

    /**
     * @return <tt>true</tt> if this hub represents a known NPV device,
     * <tt>false</tt> otherwise.
     */
    public boolean isConnectedNpvDevice() {
        return StringUtils.isNotEmpty(connectedNpvPortWWN)
                && StringUtils.isNotEmpty(connectedNpvNodeWWN);
    }

    /**
     * All hubs should have been removed from the model by
     * {@link com.onaro.sanscreen.acquisition.foundation.datasource.HubFinalizer}.
     * So fail if this method is ever called.
     */
    public void marshall(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        throw new SchemaException("Hub " + getWWN()
                + " should no longer be in odata after finalization!");
    }

}
