package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

public class ISCSINetworkPortalGroup extends XMLNode<Storage> {
    private final String portalGroupName;
    private final String portalGroupTag;

    private final ISCSIPortalGroupKey key;
    private final ISCSINetworkPortals iscsiNetworkPortals;

    ISCSINetworkPortalGroup(final Storage parent, final String portalGroupName, final String portalGroupTag) throws SchemaException {
        super(parent, "ISCSINetworkPortalGroup");

        if (portalGroupName == null || portalGroupTag == null)
            throw new SchemaException("portalGroupName and portalGroupTag must not be null");

        this.portalGroupName = portalGroupName;
        this.portalGroupTag = portalGroupTag;

        this.iscsiNetworkPortals = new ISCSINetworkPortals(this);

        key = new ISCSIPortalGroupKey(portalGroupName, portalGroupTag, parent.getKey());
    }

    public final String getPortalGroupName() {
        return portalGroupName;
    }

    public final String getPortalGroupTag() {
        return portalGroupTag;
    }

    public ISCSIPortalGroupKey getKey() {
        return key;
    }

    public final ISCSINetworkPortals getIscsiPortals() {
        return iscsiNetworkPortals;
    }

    /**
     * Create target portal with specific port (use {@link ISCSITargetNetworkPortal#defaultPort} for default)
     *
     * @param ip
     * @pararm port Specifies the listening port for this portal
     */
    public ISCSITargetNetworkPortal createTargetPortal(final String ip, final Integer port) throws SchemaException {
        return new ISCSITargetNetworkPortal(this, ip, port);
    }

    public ISCSINetworkPortal createInitiatorPortal(final String ip) throws SchemaException {
        return new ISCSINetworkPortal(this, ip);
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "OID", getKey().toString());
        addAttribute(writer, "PortalGroupName", portalGroupName);
        addAttribute(writer, "PortalGroupTag", portalGroupTag);
    }

    protected boolean hasChildren() {
        return iscsiNetworkPortals.size() > 0;
    }

    protected void marshallChildren(final MarshallingCache marshallingCache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(marshallingCache, writer, iscsiNetworkPortals, tabs);
    }

    public static final class ISCSIPortalGroupKey implements Key<ISCSIPortalGroupKey> {
        private final String portalGroupName;
        private final String portalGroupTag;
        private final Key parentKey;

        public ISCSIPortalGroupKey(final String portalGroupName, final String portalGroupTag, final Key key) throws SchemaException {
            if (portalGroupName == null || portalGroupTag == null || key == null)
                throw new SchemaException("portalGroupName, portalGroupTag and key must not be null");

            this.portalGroupName = portalGroupName;
            this.portalGroupTag = portalGroupTag;
            this.parentKey = key;
        }

        public int compareTo(final ISCSIPortalGroupKey o) {
            int i = getPortalGroupName().compareTo(o.getPortalGroupName());
            if (i != 0)
                return i;

            i = getPortalGroupTag().compareTo(o.getPortalGroupTag());
            if (i != 0)
                return i;

            return getParentKey().compareTo(o.getParentKey());
        }

        public Key getParentKey() {
            return parentKey;
        }

        public String getPortalGroupName() {
            return portalGroupName;
        }

        public String getPortalGroupTag() {
            return portalGroupTag;
        }

        public String toString() {
            return portalGroupName + ":" + portalGroupTag + getParentKey();
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final ISCSIPortalGroupKey that = (ISCSIPortalGroupKey) o;

            if (portalGroupName != null ? !portalGroupName.equals(that.portalGroupName) : that.portalGroupName != null)
                return false;

            if (portalGroupTag != null ? !portalGroupTag.equals(that.portalGroupTag) : that.portalGroupTag != null)
                return false;

            if (parentKey != null ? !parentKey.equals(that.parentKey) : that.parentKey != null) return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (portalGroupName != null ? portalGroupName.hashCode() : 0);
            result = 31 * result + (portalGroupTag != null ? portalGroupTag.hashCode() : 0);
            result = 31 * result + (parentKey != null ? parentKey.hashCode() : 0);
            return result;
        }
    }
}
