package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

public abstract class ISCSINode<Parent> extends XMLNode<Parent> {
    private final String iscsiNodeName;

    private String iscsiNodeAlias;
    private ISCSINodeType type;

    private final ISCSISessions sessions;
    private final ISCSIPortalGroupMaps portalGroupsMaps;

    private final ISCSINodeKey key;

    ISCSINode(final Parent parent, final String iscsiNodeName) throws SchemaException {
        super(parent, "ISCSINode");

        if (iscsiNodeName == null || iscsiNodeName.length() == 0)
            throw new SchemaException("node name and type must not be null");

        this.iscsiNodeName = iscsiNodeName;
        this.sessions = new ISCSISessions(this);
        this.portalGroupsMaps = new ISCSIPortalGroupMaps(this);

        this.key = new ISCSINodeKey(iscsiNodeName);
    }

    public String getIscsiNodeAlias() {
        return iscsiNodeAlias;
    }

    public void setIscsiNodeAlias(final String iscsiNodeAlias) {
        this.iscsiNodeAlias = iscsiNodeAlias;
    }

    public String getIscsiNodeName() {
        return iscsiNodeName;
    }

    public ISCSINodeType getType() {
        return type;
    }

    public void setType(final ISCSINodeType type) {
        this.type = type;
    }

    public ISCSISessions getISCSISessions() {
        return sessions;
    }

    public ISCSIPortalGroupMaps getISCSIPortalGroupMaps() {
        return portalGroupsMaps;
    }

    public ISCSISession createISCSISession(final String initiatorNodeName, final String targetNodeName, ISCSINetworkPortalGroup.ISCSIPortalGroupKey targetPortalGroupKey) throws SchemaException {
        return new ISCSISession(this, initiatorNodeName, targetNodeName, targetPortalGroupKey);
    }

    public ISCSIPortalGroupMap createISCSIPortalGroupMap(final ISCSINetworkPortalGroup.ISCSIPortalGroupKey portalGroupKey, final String iscsiTargetNodeName) throws SchemaException {
        return new ISCSIPortalGroupMap(this, portalGroupKey, iscsiTargetNodeName);
    }

    protected void marshallChildren(final MarshallingCache marshallingCache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildrenUnderNode(marshallingCache, writer, "ISCSINodeMaps", portalGroupsMaps, tabs);
        addAllChildrenUnderNode(marshallingCache, writer, "ISCSISessions", sessions, tabs);
    }

//    protected void marshall(final MarshallingCache marshallingCache, final Writer writer) throws SchemaException, IOException {
//        super.marshall(marshallingCache, writer);
//        addAttribute(writer, "OID", iscsiNodeName);
//        addAttribute(writer, "nodeName", iscsiNodeName);
//        addAttribute(writer, "alias", iscsiNodeAlias);
//        addAttribute(writer, "type", type);
//    }

    protected void marshallAttributes(final MarshallingCache marshallingCache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(marshallingCache, writer);
        addAttribute(writer, "OID", iscsiNodeName);
        addAttribute(writer, "NodeName", iscsiNodeName);
        addAttribute(writer, "Alias", iscsiNodeAlias);
        addAttribute(writer, "Type", type);
    }

    protected boolean hasChildren() {
        return sessions.size() > 0 || portalGroupsMaps.size() > 0;
    }

    public ISCSINodeKey getKey() {
        return key;
    }

    public static enum ISCSINodeType {
        Target,
        Initiator,
        Both
    }

    public static class ISCSINodeKey implements Key<ISCSINodeKey> {
        private final String nodeName;

        public ISCSINodeKey(final String nodeName) throws SchemaException {
            if (nodeName == null)
                throw new SchemaException("node name must not be null");

            this.nodeName = nodeName;
        }

        public String getNodeName() {
            return nodeName;
        }

        public String toString() {
            return nodeName;
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final ISCSINodeKey that = (ISCSINodeKey) o;

            if (nodeName != null ? !nodeName.equals(that.nodeName) : that.nodeName != null) return false;

            return true;
        }

        public int hashCode() {
            return (nodeName != null ? nodeName.hashCode() : 0);
        }

        public int compareTo(final ISCSINodeKey o) {
            return nodeName.compareTo(o.getNodeName());
        }
    }
}
