package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.*;
import com.onaro.util.WWNComparator;

import java.io.IOException;
import java.io.Writer;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

/**
 * Represents an IVR fabric discovered in this Chassis
 */
public class IVRPhysicalFabric extends XMLNode<Root> {
    private final Set<String> ivrChassisWWNs = new TreeSet<String>(new WWNComparator());
    private final IVRZoneSets inactiveIVRZoneSets;

    private IVRZoneSet activeIVRZoneSet = null;

    private final IVRPhysicalFabricKey key;

    IVRPhysicalFabric(final Root parent, final Collection<String> ivrChassisWWNs) throws SchemaException {
        super(parent, "IVRPhysicalFabric");

        if (ivrChassisWWNs == null || ivrChassisWWNs.size() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null and non-empty ivrChassisWWNs");

        this.ivrChassisWWNs.addAll(ivrChassisWWNs);
        this.inactiveIVRZoneSets = new IVRZoneSets(this);

        key = new IVRPhysicalFabricKey(ivrChassisWWNs);
    }

    public final IVRPhysicalFabricKey getKey() {
        return key;
    }

    public final Root getRoot() {
        return getParent();
    }

    public final Set<String> getIvrChassisWWNs() {
        return ivrChassisWWNs;
    }

    public final IVRZoneSet createActiveIVRZoneSet(final String name) {
        return new IVRZoneSet(this, name, true);
    }

    public final IVRZoneSet createInactiveIVRZoneSet(final String name) {
        return new IVRZoneSet(this, name, true);
    }

    public final IVRZoneSet getActiveIVRZoneSet() {
        return activeIVRZoneSet;
    }

    public final void setActiveIVRZoneSet(final IVRZoneSet activeIVRZoneSets) {
        this.activeIVRZoneSet = activeIVRZoneSets;
    }

    public final IVRZoneSets getInactiveIVRZoneSets() {
        return inactiveIVRZoneSets;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        // The first ivr Chassis WWN (the smallest) is the key for the
        addAttribute(writer, "LowestIVRChassisWWN", ivrChassisWWNs.iterator().next());

        addAttribute(writer, "IVRChassisWWNs", formatChassisWWNsList(ivrChassisWWNs));
    }

    /**
     * Now format a comma-seperated list of the Chassis WWNs
     */
    private static String formatChassisWWNsList(final Collection<String> ivrChassisWWNs) {
        final StringBuilder sb = new StringBuilder(1000);
        for (final String wwn : ivrChassisWWNs)
            sb.append(wwn).append(",");

        sb.deleteCharAt(sb.length() - 1);

        return sb.toString();
    }

    protected boolean hasChildren() {
        return true;
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        if (activeIVRZoneSet == null)
            throw new SchemaException("Active IVR zoneset was not set for IVR fabric");

        addChild(cache, writer, activeIVRZoneSet, tabs);
        addAllChildren(cache, writer, inactiveIVRZoneSets, tabs);
    }

    public static final class IVRPhysicalFabricKey implements Key<IVRPhysicalFabricKey> {
        private final Collection<String> ivrChassisWWNs;

        public IVRPhysicalFabricKey(final Collection<String> ivrChassisWWNs) {
            this.ivrChassisWWNs = ivrChassisWWNs;
        }

        public final int compareTo(final IVRPhysicalFabricKey o) {
            if (ivrChassisWWNs.size() != o.getIvrChassisWWNs().size())
                return ivrChassisWWNs.size() - o.getIvrChassisWWNs().size();

            int i;

            final Iterator<String> it1 = ivrChassisWWNs.iterator();
            final Iterator<String> it2 = o.getIvrChassisWWNs().iterator();

            while (it1.hasNext()) {
                i = it1.next().compareTo(it2.next());
                if (i != 0)
                    return i;
            }

            return 0;
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final IVRPhysicalFabric that = (IVRPhysicalFabric) o;

            if (ivrChassisWWNs != null ? !ivrChassisWWNs.equals(that.ivrChassisWWNs) : that.ivrChassisWWNs != null)
                return false;

            return true;
        }

        public final int hashCode() {
            return (ivrChassisWWNs != null ? ivrChassisWWNs.hashCode() : 0);
        }

        public final Collection<String> getIvrChassisWWNs() {
            return ivrChassisWWNs;
        }

        public final String toString() {
            return formatChassisWWNsList(ivrChassisWWNs);
        }
    }
}
