package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.io.Writer;

public abstract class Port<P> extends PortType<P> {
    private String physicalSwitchID = null;
    private Boolean unknownConnectivity = Boolean.FALSE;
    private ConnectedPort connectedPort = null;
    private Boolean generated = Boolean.FALSE;

    protected Port(final P parent, final String portWwn) throws SchemaException {
        super(parent, new PortTypeKey(portWwn));
    }

    /**
     * Create a connected port by port wwn.
     *
     * @param portWwn the port wwn of the connected port to create
     * @return the connected port object
     * @throws SchemaException throws when the port wwn is invalid
     */
    public ConnectedPort createConnectedPort(final String portWwn) throws SchemaException {
        return new ConnectedPort(this, portWwn);
    }

    /**
     * Create a connected port by switch wwn and port ID.
     *
     * @param switchWwn the switch wwn of the connected port to create
     * @param portId the port ID of the connected port to create
     * @return the connected port object
     * @throws SchemaException throws when either switch wwn or port ID is invalid
     */
    public ConnectedPort createConnectedPort(final String switchWwn, final String portId) throws SchemaException {
        return new ConnectedPort(this, switchWwn, portId);
    }

    public ConnectedPort getConnectedPort() {
        return connectedPort;
    }

    /**
     * @return The port WWN (all upper case) formatted as AA:BB:CC:DD:EE:FF:GG:HH.
     */
    public final String getWWN() {
        return getKey().toString();
    }

    public void setConnectedPort(final ConnectedPort connectedPort) {
        setConnectedPort(connectedPort, false);
    }

    /**
     * Use this method to set the connected port and force the value of
     * <tt>UnknownConnectivity</tt>, which by default should be set to
     * <tt>false</tt>.
     * <p/>
     * This is useful in cases where we're dealing with generated NPV hubs and
     * switches, because the server requires <tt>UnknownConnectivity</tt> to be
     * set to <tt>true</tt> on both ends of connections with generated ports.
     */
    public void setConnectedPort(final ConnectedPort connectedPort, final boolean unknownConnectivity) {
        if (!connectedPort.getPort().equals(this))
            throw new RuntimeException("Added child object to wrong parent object in Schema");

        this.connectedPort = connectedPort;
        this.unknownConnectivity = unknownConnectivity;
    }

    public void disconnect() {
        this.connectedPort = null;
        // We know there is no connectivity, so connectivity is known
        unknownConnectivity = Boolean.FALSE;
    }

    public String getPhysicalSwitchID() {
        return physicalSwitchID;
    }

    public void setPhysicalSwitchID(final String physicalSwitchID) throws SchemaException {
        this.physicalSwitchID = physicalSwitchID;
    }

    public boolean isUnknownConnectivity() {
        return unknownConnectivity;
    }

    public void setUnknownConnectivity(final boolean unknownConnectivity) {
        this.unknownConnectivity = unknownConnectivity;
    }

    public Boolean isGenerated() {
        return generated;
    }

    public void setGenerated(final boolean generated) {
        this.generated = generated;
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addChild(cache, writer, connectedPort, tabs);
    }

    protected boolean hasChildren() {
        return (connectedPort != null);
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "UnknownConnectivity", unknownConnectivity ? 1 : null);
        addAttribute(writer, "Generated", generated ? 1 : null);
        addAttribute(writer, "physicalSwitchId", physicalSwitchID);
    }

}
