package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.sanscreen.acquisition.schema.utils.NasShareUtils;

import java.io.IOException;
import java.io.Writer;
import java.util.SortedSet;
import java.util.TreeSet;

public class Share extends XMLNode<FileShare> {
    private final String shareName;
    private final ShareType protocol;

    private String description;

    private final ShareInitiators initiators;
    private final ShareKey shareKey;

    //Sorted for server delivery.
    private SortedSet<String> hosts = null;

    public Share(final FileShare parent, final String shareName, final ShareType protocol) throws SchemaException {
        super(parent, "Share");

        if (shareName == null || protocol == null) {
            throw new SchemaException(getNodeName() + " Type requires non-null name and protocol");
        }

        this.protocol = protocol;
        this.shareName = shareName;
        shareKey = new ShareKey(shareName, protocol);
        initiators = new ShareInitiators(this);
        hosts = new TreeSet<String>();
    }

    public ShareInitiators getInitiators() {
        return initiators;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    public String getShareName() {
        return shareName;
    }

    public ShareType getProtocol() {
        return protocol;
    }

    /**
     * Get list of host names / IP addresses on which this share is exposed.   This list will be sorted according
     * to natural String sort of the addresses.
     *
     * @return SortedSet of the hosts.
     */
    public SortedSet<String> getHosts() {
        return hosts;
    }

    @SuppressWarnings({"unchecked"})
    public Key getKey() {
        return shareKey;
    }


    public String getDescription() {
        return description;
    }

    public void setDescription(final String description) {
        this.description = description;
    }

    public ShareInitiator createShareInitiator(final String initiator, final String permission) throws SchemaException {
        return new ShareInitiator(this, initiator, permission);
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "Name", NasShareUtils.normalizeNasShareName(shareName, protocol == ShareType.CIFS));
        addAttribute(writer, "Protocol", protocol);
        addAttribute(writer, "Description", description);
        addAttribute(writer, "IPInterfaces", getHostsAsFormattedString());
    }

    protected void marshallChildren(final MarshallingCache marshallingCache, final Writer writer, final int tabs) throws IOException, SchemaException {
        super.marshallChildren(marshallingCache, writer, tabs);
        addAllChildrenUnderNode(marshallingCache, writer, "ShareInitiators", initiators, tabs);
    }

    protected boolean hasChildren() {
        return initiators.size() > 0;
    }

    /**
     * Get the lsit of host names / IP addresses on which this share is exposed as a formatted single string list.
     * This return is comma separated and sorted according to natural String sort.      Note that this is the
     * String currently inserted into the "IPInterfaces" of the ODATA marshalling.
     *
     * @return Comma separated single string list of IP addresses / host names on which this share is exposed.
     *         Sorted.
     */
    private String getHostsAsFormattedString() {
        final StringBuilder builder = new StringBuilder();
        int count = 0;
        final int total = hosts.size();
        for (final String host : hosts) {
            builder.append(host);
            if (count != total - 1) {
                builder.append(",");
            }
            count++;
        }
        return builder.toString();
    }

    public static final class ShareKey implements Key<ShareKey> {
        final String shareName;
        final ShareType protocol;

        public ShareKey(final String shareName, final ShareType protocol) throws SchemaException {
            if (shareName == null || protocol == null)
                throw new SchemaException(" Type requires non-null name and protocol");
            this.shareName = shareName;
            this.protocol = protocol;
        }

        public ShareType getProtocol() {
            return protocol;
        }


        public int compareTo(final ShareKey o) {
            int i = getShareName().compareTo(o.getShareName());
            if (i != 0)
                return i;
            i = getProtocol().compareTo(o.getProtocol());
            if (i != 0)
                return i;

            return getShareName().compareTo(o.getShareName());
        }

        public String toString() {
            return shareName + ":" + protocol;
        }


        public String getShareName() {
            return shareName;
        }
    }

    public static enum ShareType {
        NFS("NFS"),
        CIFS("CIFS");

        private final String s;

        ShareType(final String s) {
            this.s = s;
        }

        @SuppressWarnings({"UnusedDeclaration"})
        public String toEnum() {
            return s;
        }

        public String toString() {
            return s;
        }
    }
}
