package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;

/**
 * Storage Capacity Efficiency. <br>
 * <p>StorageEfficiency class represent  vendor spacific feature that influance storage efficiency.<br>
 * for example: Redundancy will decrease storage capacity efficiency due too extra sapce consumed with the same data. <br>
 * on the other hand dedup will incrase efficiency by removing duplicate data.<br></p>
 * <p>
 * Attributes:<br>
 * <b>technology</b> = vendor specific feature name.<br>
 * <b>gainLossMB</b> = impact of active feature  negative or positive.<br>
 * <b>potentialGainLossMB</b> = impact of enabling a feater  negative or positive.<br>
 * </p>
 * example:   <br>
 * redundancy RAID1 on volume x with capacity of 20mb : gainLossMB=-20mb <br>
 * dedup enabled on volume x with capacity of 20mb and dedup saving in 50% : gainLossMB=10mb <br>
 * dedup disabled on volume y: with capacity of 10mb                       : PotentialGainLossMB = 5mb.<br>
 */
@SuppressWarnings({"UnusedDeclaration"})
public class StorageEfficiency extends XMLNode<Storage> {

    private final String technology;
    /**
     * Gain/Loss is positive for gain and negative for loss
     */
    private final Capacity gainLossMB = Capacity.createAllowNegative();
    /**
     * Potential gain CANNOT be negative - we are not supposed to add this technology as 'potential' if it has negative numbers.
     */
    private final Capacity potentialGainLossMB = Capacity.create();

    private final StorageEfficiencyKey storageEfficiencyKey;

    protected StorageEfficiency(final Storage parent, final String technology) throws SchemaException {
        super(parent, "StorageEfficiency");
        if (technology == null || technology.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null technology");

        this.technology = technology;
        storageEfficiencyKey = new StorageEfficiencyKey(parent.getKey(), technology);
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "Technology", technology);
        addAttribute(writer, "GainLossMB", gainLossMB, Capacity.Unit.MB_1024);

        if (potentialGainLossMB.value(Capacity.Unit.BYTE) > 0)
            addAttribute(writer, "PotentialGainLossMB", potentialGainLossMB, Capacity.Unit.MB_1024);
    }

    public final String getTechnology() {
        return technology;
    }

    public final Capacity getGainLossMB() {
        return gainLossMB;
    }

    public final Capacity getPotentialGainLossMB() {
        return potentialGainLossMB;
    }

    @SuppressWarnings({"unchecked"})
    public final StorageEfficiencyKey getKey() {
        return storageEfficiencyKey;
    }

    public boolean equals(final Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final StorageEfficiency that = (StorageEfficiency) o;

        if (gainLossMB != null ? !gainLossMB.equals(that.gainLossMB) : that.gainLossMB != null) return false;
        if (potentialGainLossMB != null ? !potentialGainLossMB.equals(that.potentialGainLossMB) : that.potentialGainLossMB != null)
            return false;
        if (storageEfficiencyKey != null ? !storageEfficiencyKey.equals(that.storageEfficiencyKey) : that.storageEfficiencyKey != null)
            return false;
        if (technology != null ? !technology.equals(that.technology) : that.technology != null) return false;

        return true;
    }

    public int hashCode() {
        int result;
        result = (technology != null ? technology.hashCode() : 0);
        result = 31 * result + (gainLossMB != null ? gainLossMB.hashCode() : 0);
        result = 31 * result + (potentialGainLossMB != null ? potentialGainLossMB.hashCode() : 0);
        result = 31 * result + (storageEfficiencyKey != null ? storageEfficiencyKey.hashCode() : 0);
        return result;
    }

    public static class StorageEfficiencyKey implements Key<StorageEfficiencyKey> {
        private final Storage.StorageKey storageKey;
        private final String technology;

        public StorageEfficiencyKey(final Storage.StorageKey storageKey, final String technology) throws SchemaException {
            if (technology == null || technology.length() == 0)
                throw new SchemaException("StorageEfficiency Type requires non-null technology");

            this.storageKey = storageKey;
            this.technology = technology;
        }

        public int compareTo(final StorageEfficiencyKey storageEfficiencyKey) {
            int i = getTechnology().compareTo(storageEfficiencyKey.getTechnology());
            if (i != 0)
                return i;

            return getStorageKey().compareTo(storageEfficiencyKey.getStorageKey());
        }

        public Storage.StorageKey getStorageKey() {
            return storageKey;
        }

        public String getTechnology() {
            return technology;
        }
    }
}
