package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;
import java.math.RoundingMode;
import java.text.NumberFormat;
import java.util.Locale;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

public class StoragePool extends XMLNode<Storage> {
    private final String name;
    private final StoragePoolType type;
    private Boolean thinProvisioningSupported;

    /**
     * true indicates that the storage pool (NetApp aggregate) contains both SSD and non-SSD RAID groups(NetApp specific)
     */
    private Boolean usesFlashPools;

    /**
     * true indicates that the storage pool is participating in an auto tier with other pools (EMC specific)
     */
    private Boolean autoTiering;

    /**
     * The name of the tier that the pool belongs to (EMC specific)
     * Vendor Tiers are not yet modeled, just the reference to the tier name
     */
    private String vendorTier;

    private Boolean dedupeEnabled;
    private Float dedupeSavings;
    private String status;

    private Boolean includeInDwhCapacity;

    private final StoragePoolCapacity capacity;

    private final SortedSet<Disk.DiskKey> diskRefs;
    private final SortedSet<BackEndLun.BackEndLunKey> backendLunRefs;

    private final StoragePoolKey key;

    /*
     * Create US locale format
     */
    final static NumberFormat usaLocalNumberFormat = NumberFormat.getInstance(Locale.US);


    /* Set to 5 digits
     * Set format to always round down.
     * By rounding down we workaround the condition where higher rawToUsableRatio
     * will cause the calculation of rawUnusedCapacityMB to evaluate to a negative.
     * rawUnusedCapacityMB = physicalDiskCapacity - (totalUsedCapacity * rawToUsableRatio )
     */
    static {
        usaLocalNumberFormat.setMaximumFractionDigits(5);
        usaLocalNumberFormat.setRoundingMode(RoundingMode.DOWN);
    }

    protected StoragePool(final Storage parent, final String name, final StoragePoolType type) throws SchemaException {
        super(parent, "StoragePool");

        this.name = name;
        if (name == null || type == null)
            throw new SchemaException(getNodeName() + " Type requires non-null name and StoragePoolType");

        this.type = type;
        this.key = new StoragePoolKey(name, type, parent.getKey());
        this.capacity = new StoragePoolCapacity();

        this.diskRefs = new TreeSet<Disk.DiskKey>();
        this.backendLunRefs = new TreeSet<BackEndLun.BackEndLunKey>();
    }

    @SuppressWarnings({"unchecked"})
    public StoragePoolKey getKey() {
        return key;
    }

    public StoragePoolCapacity getCapacity() {
        return capacity;
    }

    public String getName() {
        return name;
    }

    public StoragePoolType getType() {
        return type;
    }

    public Boolean getDedupeEnabled() {
        return dedupeEnabled;
    }

    public void setDedupeEnabled(final Boolean dedupeEnabled) {
        this.dedupeEnabled = dedupeEnabled;
    }

    public Float getDedupeSavings() {
        return dedupeSavings;
    }

    public void setDedupeSavings(final Float dedupeSavings) {
        this.dedupeSavings = dedupeSavings;
    }

    /**
     * @return Status of the Storage Pool (e.g offline, online ...)
     */
    public String getStatus() {
        return status;
    }

    public void setStatus(final String status) {
        this.status = status;
    }

    public Boolean isThinProvisioningSupported() {
        return thinProvisioningSupported;
    }

    public void setThinProvisioningSupported(final Boolean thinProvisioningSupported) {
        this.thinProvisioningSupported = thinProvisioningSupported;
    }

    /**
     * NetApp only
     *
     * @return true if the aggregate contains both SSD and non-SSD RAID groups
     */
    public Boolean getUsesFlashPools() {
        return usesFlashPools;
    }

    /**
     * NetApp only. The storage pool type should be HYBRID
     *
     * @param usesFlashPools true to indicate the aggregate contains both SSD and non-SSD RAID groups
     */
    public void setUsesFlashPools(final Boolean usesFlashPools) {
        this.usesFlashPools = usesFlashPools;
    }

    /**
     * EMC Symmetrix only
     * returns true if the pool is part of auto tiering
     *
     * @return may return null
     */
    public Boolean getAutoTiering() {
        return autoTiering;
    }

    /**
     * EMC Symmetrix only
     *
     * @param autoTiering set to true to indicate that the pool contributes to auto tiering
     */
    public void setAutoTiering(final Boolean autoTiering) {
        this.autoTiering = autoTiering;
    }

    /**
     * EMC Symmetrix only
     *
     * @return The vendor Tier name. May return null
     */
    public String getVendorTier() {
        return vendorTier;
    }

    /**
     * EMC Symmetrix only
     *
     * sets the name of the vendor Tier that the pool is part of. The tier itself is not modeled.
     *
     * @param vendorTier the raw vendor tier name
     */
    public void setVendorTier(final String vendorTier) {
        this.vendorTier = vendorTier;
    }

    public Boolean getIncludeInDwhCapacity() {
        return includeInDwhCapacity;
    }

    public void setIncludeInDwhCapacity(final Boolean includeInDwhCapacity) {
        this.includeInDwhCapacity = includeInDwhCapacity;
    }

    public String getStoragePoolId() {
        return getKey().toString();
    }

    public boolean isVirtual() {
        return backendLunRefs.size() > 0;
    }

    public Set<Disk.DiskKey> getDiskRefs() {
        return diskRefs;
    }

    public Set<BackEndLun.BackEndLunKey> getBackendLunRefs() {
        return backendLunRefs;
    }

    protected boolean hasChildren() {
        return !diskRefs.isEmpty() || !backendLunRefs.isEmpty();
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        super.marshallChildren(cache, writer, tabs);

        if (!diskRefs.isEmpty()) {
            // Marshal references for the disk refs
            addTabulation(writer, tabs);
            writer.append("<DiskRefs >");
            writer.append(lineSeperator);

            for (final Disk.DiskKey key : diskRefs) {
                addTabulation(writer, tabs + 1);
                writer.append("<").append("DiskRef").append(' ');
                writer.append("DiskId=\"").append(key.getId()).append("\" ");
                writer.append("/>").append(lineSeperator);
            }

            addTabulation(writer, tabs);
            writer.append("</DiskRefs >");
            writer.append(lineSeperator);
        }

        if (!backendLunRefs.isEmpty()) {
            // Marshal references for the LUN refs
            addTabulation(writer, tabs);
            writer.append("<BackendLunRefs >");
            writer.append(lineSeperator);

            for (final BackEndLun.BackEndLunKey key : backendLunRefs) {
                addTabulation(writer, tabs + 1);
                writer.append("<").append("BackendLunRef").append(' ');
                writer.append("BackendLunName=\"").append(key.getName()).append("\" ");
                writer.append("/>").append(lineSeperator);
            }

            addTabulation(writer, tabs);
            writer.append("</BackendLunRefs >");
            writer.append(lineSeperator);
        }
    }

    protected void marshallAttributes(final MarshallingCache marshallingCache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(marshallingCache, writer);

        addAttribute(writer, "StoragePoolId", getStoragePoolId());
        addAttribute(writer, "Name", name);
        addAttribute(writer, "Type", type);
        addAttribute(writer, "Status", status);
        addAttribute(writer, "ThinProvisioningSupported", thinProvisioningSupported);

        /*don't include the dwh flag if false in order to avoid breaking unit tests.     Server will interpret absence as false
        anyways. Also, always include it if thin provisioning is false, regardless of it's actual value (dwh rule).

         As a matter of policy, it seems odd that we have this flag at all.   Why is Acq telling the data
         warehouse what it should include or not?   This seems to violate separation of concerns, and sets up a
         path leading to buggy interdependence.   I would vote to revisit this post 6.0, or whenever we have appropriate time.
         -jjs
        */
        if ((thinProvisioningSupported != null && thinProvisioningSupported) || includeInDwhCapacity) {
            addAttribute(writer, "IncludeInDwhCapacity", true);
        }

        // Auto Tier attributes
        addAttribute(writer, "AutoTiering", autoTiering);
        addAttribute(writer, "UsesFlashPools", usesFlashPools);        
        addAttribute(writer, "VendorTier", vendorTier);

        addAttribute(writer, "DedupeEnabled", dedupeEnabled);
        addAttribute(writer, "DedupeSavings", dedupeSavings);
        addAttribute(writer, "Virtual", isVirtual());
        addAttribute(writer, "RaidGroup", Boolean.toString(type == StoragePoolType.RAID_GROUP));

        // marshall only attributes of the capacity
        addAttribute(writer, "SnapshotUsedCapacityMB", getCapacity().getSnapshotUsed(), Capacity.Unit.MB_1024);
        addAttribute(writer, "SnapshotAllocatedCapacityMB", getCapacity().getSnapshotAllocated(), Capacity.Unit.MB_1024);
        addAttribute(writer, "DataUsedCapacityMB", getCapacity().getDataUsed(), Capacity.Unit.MB_1024);
        addAttribute(writer, "DataAllocatedCapacityMB", getCapacity().getDataAllocated(), Capacity.Unit.MB_1024);
        addAttribute(writer, "TotalUsedCapacityMB", getCapacity().getTotalUsed(), Capacity.Unit.MB_1024);
        addAttribute(writer, "TotalAllocatedCapacityMB", getCapacity().getTotalAllocated(), Capacity.Unit.MB_1024);
        addAttribute(writer, "OtherUsedCapacityMB", getCapacity().getOtherUsedCapacity(), Capacity.Unit.MB_1024);
        addAttribute(writer, "OtherAllocatedCapacityMB", getCapacity().getOtherAllocatedCapacity(), Capacity.Unit.MB_1024);
        addAttribute(writer, "PhysicalDiskCapacityMB", figureOutPhysicalDiskCapacity(), Capacity.Unit.MB_1024);
        addAttribute(writer, "ReservedCapacityMB", getCapacity().getReservedCapacity(), Capacity.Unit.MB_1024);
        addAttribute(writer, "RawToUsableRatio", usaLocalNumberFormat.format(getCapacity().getRawToUsableRatio()));
        addAttribute(writer, "Redundancy", getCapacity().getRedundancy());

    }

    /**
     * disk based storage pool and virtual storage pool with local disks we need
     * to populate PhysicalDiskCapacityMB attribute with disk capaicty value
     * bug:19277
     *
     * @return PhysicalDiskCapacity
     *
     * @throws SchemaException
     */
    private Capacity figureOutPhysicalDiskCapacity() throws SchemaException {

        //for backend lun we do not report PhysicalDiskCapacityMB
        if (diskRefs.size() == 0 && isVirtual())
            return null;

        return getCapacity().getPhysicalDiskCapacity() != null
               ? getCapacity().getPhysicalDiskCapacity()
               : Capacity.create((long) (getCapacity().getTotalAllocated().value(Capacity.Unit.BYTE) * getCapacity().getRawToUsableRatio()), Capacity.Unit.BYTE);
    }

    public final static class StoragePoolKey implements Key<StoragePoolKey> {
        private final String name;
        private final StoragePoolType type;
        private final Storage.StorageKey storageKey;

        public StoragePoolKey(final String name, final StoragePoolType type, final Storage.StorageKey storageKey) {
            this.name = name;
            this.type = type;
            this.storageKey = storageKey;
        }

        public int compareTo(final StoragePoolKey o) {
            int i = storageKey.compareTo(o.storageKey);
            if (i != 0)
                return i;

            i = name.compareTo(o.name);
            if (i != 0)
                return i;

            i = type.compareTo(o.type);
            if (i != 0)
                return i;

            return 0;
        }

        public String getName() {
            return name;
        }

        public String toString() {
            return storageKey + ":" + name + ":" + type;
        }

        public StoragePoolType getType() {
            return type;
        }

        public Storage.StorageKey getStorageKey() {
            return storageKey;
        }

        @SuppressWarnings({"RedundantIfStatement"})
        public boolean equals(final Object o) {
            if (this == o) return true;
            if (!(o instanceof StoragePoolKey)) return false;

            final StoragePoolKey that = (StoragePoolKey) o;

            if (name != null ? !name.equals(that.name) : that.name != null) return false;
            if (storageKey != null ? !storageKey.equals(that.storageKey) : that.storageKey != null) return false;
            if (type != null ? !type.equals(that.type) : that.type != null) return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (name != null ? name.hashCode() : 0);
            result = 31 * result + (type != null ? type.hashCode() : 0);
            result = 31 * result + (storageKey != null ? storageKey.hashCode() : 0);
            return result;
        }
    }

    /**
     * Reported as string to server.
     */
    public enum StoragePoolType {
        AGGREGATE,
        RAID_GROUP,
        THIN_PROVISIONING,
        BACKEND_GROUP,
        RESERVED,
        FLASH_POOL; // NetApp specific - the aggregate contains SSD and non-SSD RAID groups

        public boolean isPhysicalDiskType() {
            return this == AGGREGATE || this == RAID_GROUP || this == FLASH_POOL;
        }
    }
}
