package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.framework.datasource.IntegrityException;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;
import java.util.Set;

@SuppressWarnings("unchecked")
class StorageRootImpl extends BaseRootImpl implements StorageRoot {
    private static final Logger logger = Logger.getLogger(StorageRootImpl.class);

    private final Storages storages;
    private final DeviceGroups deviceGroups;
    private final ISCSINodes nodes;
    private boolean drPackageInstalled = false;

    StorageRootImpl(final RootFactory parent, final long originatorID, final String originatorType, final String dataSourceName) throws SchemaException {
        super(parent, originatorID, originatorType, dataSourceName);

        storages = new Storages(this);
        deviceGroups = new DeviceGroups(this);
        nodes = new ISCSINodes(this);
    }

    public DeviceGroups getDeviceGroups() {
        return deviceGroups;
    }

    public Storages getStorages() {
        return storages;
    }

    public Storage createStorage(final String ip) throws SchemaException {
        return new Storage(this, ip);
    }

    public WWNAlias createStorageHostAlias(final String alias, final String wwn) throws SchemaException {
        return new WWNAlias(this, WWNAlias.Source.Storage, alias, wwn, WWNAlias.ObjectType.Host);
    }

    public DeviceGroup createDeviceGroup(final String name, final String storageManagementID) throws SchemaException {
        return new DeviceGroup(this, name, storageManagementID);
    }

    public ISCSIUnKnownNode createISCSIInitiatorNode(final String iscsiNodeName) throws SchemaException {
        return new ISCSIUnKnownNode(this, iscsiNodeName);
    }

    public boolean isDrPackageInstalled() {
        return drPackageInstalled;
    }

    public void setDrPackageInstalled(boolean drPackageInstalled) {
        this.drPackageInstalled = drPackageInstalled;
    }


    public ISCSINodes getNodes() {
        return nodes;
    }

    protected void doMarshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, storages, tabs);

        // Now let the base marshall the unknowns and aliases
        super.doMarshallChildren(cache, writer, tabs);

        // Add the device groups!
        if (deviceGroups.size() > 0)
            addAllChildrenUnderNode(cache, writer, "DeviceGroups", deviceGroups.values(), tabs);

        if (nodes.size() > 0)
            addAllChildrenUnderNode(cache, writer, "ISCSINodes", nodes, tabs);
    }

    /**
     * Perform integrity checks on storage device data
     *
     * @throws IntegrityException
     */
    public void checkIntegrity() throws IntegrityException {

        if (getStorages().size() == 0)
            throw new IntegrityException("No storage devices found");

        for (final Storage storage : getStorages()) {
            final Set<String> storagePortWWNs = new HashSet<String>();

            //check that volume capacity is not null
            for (final Volume volume : storage.getVolumes().values()) {
                final VolumeCapacity volumeCapacity = volume.getCapacity();
                if (volumeCapacity.getCapacity() == null)
                    throw new IntegrityException("Volume:" + volume.getName() + " in storage:" + storage.getIP() + " has no capacity value");
            }

            // Note: Do NOT check for existence of mapping or masking since an unconfigured array
            // with only volumes but without mapping & masking can exist and make sense. In AOL they
            // had a Clariion for DR that had this situation. They were mapping and masking the host
            // only when a disaster happens. (they have empty storage groups - with no hosts).

            if (isFcpActive(storage)) {
                // Each controller should have at least one port!
                for (final Controller controller : storage.getControllers()) {
                    checkWWN(controller.getWWN());

                    if (controller.getPorts().size() == 0)
                        throw new IntegrityException("Storage (" + storage.getIP() + ") controller has no ports");

                    for (final StoragePort storagePort : controller.getPorts()) {
                        storagePortWWNs.add(storagePort.getWWN());
                        checkWWN(storagePort.getWWN());
                    }
                }

                // First check if storage has controllers...
                if (storagePortWWNs.size() == 0 && storage.getISCSINodes().size() == 0 && !storageHasShares(storage.getFileShares()))
                    logger.warn("Storage (" + storage.getIP() + ") has no ports nor iSCSI nodes nor shares/exports");

                // Check for virtual storage validity.
                // TODO: Currently, only checked for FCP since we don't support virtualization for anything else
                if (storage.isVirtual()) {
                    // Check storage virtualization issues: Every Connection element must map the local port attribute to an
                    // existing storage (virtualizer) port
                    for (final BackEndLun backEndLun : storage.getBackEndLuns()) {
//                        if (backEndLun.getGroup() == null)
//                            throw new IntegrityException("Storage (" + storage.getIP() + ") : backEndLun " + backEndLun.getKey().toString() + " does not report group attribute");
                        for (final BackEndPath backendPath : backEndLun.getBackendPaths()) {
                            if (backendPath.getLocalPortWWN() != null && !storagePortWWNs.contains(backendPath.getLocalPortWWN()))
                                throw new IntegrityException("Storage (" + storage.getIP() + ") : Connection " + backendPath.getKey().toString() + " points to a non existing storage port " + backendPath.getLocalPortWWN());
                        }
                    }
                }
            }
        }
    }

    private boolean storageHasShares(final FileShares fileShares) {
        for (final FileShare fileShare : fileShares) {
            if (fileShare.getShares().size() > 0)
                return true;
        }
        return false;
    }

    @SuppressWarnings({"SimplifiableIfStatement"})
    private boolean isFcpActive(final Storage storage) {
        // For backward-compatibility - since Capabilities were introduced with iSCSI, we assume that if the recording
        // did not contain any information about capabilities, then FCP was the only protocol supported and enabled
        if (storage.getCapabilities() == null || storage.getCapabilities().size() == 0)
            return true;

        return storage.getCapabilities().get(Storage.SupportedTechnology.FC).isActive();
    }

    private void checkWWN(final String wwn) throws IntegrityException {
        if (wwn.equals("00:00:00:00:00:00:00:00"))
            throw new IntegrityException("Bad WWN generated (00:00:00:00:00:00:00:00)");
    }
}
