package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;
import com.onaro.util.Capacity;

import java.io.IOException;
import java.io.Writer;

/**
 * Represents the properties of a virtual disk and its relashionship with the parent datastore
 */
public class VirtualDisk extends XMLNode<VirtualizationRoot> {
    private final String moid;
    private final String virtualCenterIP;

    private DataStore.DataStoreKey dataStoreKey;
    private String name;
    private Capacity capacity;
    private Capacity usedCapacity;
    private Capacity usedCapacityForReporting;
    private String lunOID;
    private final VirtualDiskKey key;

    VirtualDisk(final VirtualizationRoot parent, final String moid, final String virtualCenterIP) throws SchemaException {
        super(parent, "HV_VirtualDisk");

        if (moid == null || moid.length() == 0)
            throw new SchemaException("moid is mandatory!");

        if (virtualCenterIP == null || virtualCenterIP.length() == 0)
            throw new SchemaException("ip is mandatory!");

        this.moid = moid;
        this.virtualCenterIP = virtualCenterIP;
        this.key = new VirtualDiskKey(moid, virtualCenterIP);
    }

    @SuppressWarnings("unchecked")
    public VirtualDiskKey getKey() {
        return key;
    }

    @SuppressWarnings("UnusedDeclaration")
    public String getMOID() {
        return moid;
    }

    @SuppressWarnings("UnusedDeclaration")
    public String getVirtualCenterIP() {
        return virtualCenterIP;
    }

    public String getName() {
        return name;
    }

    public void setName(final String name) {
        this.name = name;
    }

    public DataStore.DataStoreKey getDataStore() {
        return dataStoreKey;
    }

    public void setDataStoreID(final DataStore.DataStoreKey dataStoreKey) {
        this.dataStoreKey = dataStoreKey;
    }

    /**
     * Capacity of the virtual disk.
     *
     * @return the capacity of the virtual disk.
     */
    public Capacity getCapacity() {
        return capacity;
    }

    public void setCapacity(final Capacity capacity) {
        this.capacity = capacity != null ? Capacity.create(capacity) : null;
    }

    /**
     * Used capacity of virtual disk, if available.
     *
     * @return  the used capacity.
     */
    public Capacity getUsedCapacity() {
        return usedCapacity;
    }

    public void setUsedCapacity(final Capacity usedCapacity) {
         this.usedCapacity = usedCapacity != null ? Capacity.create(usedCapacity) : null;
    }

    public Capacity getUsedCapacityForReporting() {
        return usedCapacityForReporting;
    }

    public void setUsedCapacityForReporting(final Capacity usedCapacityForReporting) {
        this.usedCapacityForReporting = usedCapacityForReporting != null ? Capacity.create(usedCapacityForReporting) : null;
    }

    @SuppressWarnings("UnusedDeclaration")
    public String getLunOID() {
        return lunOID;
    }

    public void setLunOID(final String lunOID) {
        this.lunOID = lunOID;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "OID", getKey());

        addAttribute(writer, "HV_DataStore_OID", dataStoreKey);
        addAttribute(writer, "Name", name);
        if (capacity != null) {
            addAttribute(writer, "CapacityMB", capacity.value(Capacity.Unit.MB_1024));
        }
        if (usedCapacity != null) {
            addAttribute(writer, "UsedCapacityMB", usedCapacity.value(Capacity.Unit.MB_1024));
        }
        if (usedCapacityForReporting != null) {
            addAttribute(writer, "UsedCapacityForReportingMB", usedCapacityForReporting.value(Capacity.Unit.MB_1024));
        }
        addAttribute(writer, "HV_Lun_OID", lunOID);
    }

    public final static class VirtualDiskKey implements Key<VirtualDiskKey> {
        final String moid;
        final String virtualCenterIP;

        public VirtualDiskKey(final String moid, final String virtualCenterIP) {
            this.moid = moid;
            this.virtualCenterIP = virtualCenterIP;
        }

        public int compareTo(final VirtualDiskKey o) {
            final int i = moid.compareTo(o.getMOID());
            if (i != 0)
                return i;

            return virtualCenterIP.compareTo(o.getVirtualCenterIP());
        }

        private String getVirtualCenterIP() {
            return virtualCenterIP;
        }

        public String getMOID() {
            return moid;
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final VirtualDiskKey that = (VirtualDiskKey) o;

            if (moid != null ? !moid.equals(that.moid) : that.moid != null) return false;
            if (virtualCenterIP != null ? !virtualCenterIP.equals(that.virtualCenterIP) : that.virtualCenterIP != null)
                return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (moid != null ? moid.hashCode() : 0);
            result = 29 * result + (virtualCenterIP != null ? virtualCenterIP.hashCode() : 0);
            return result;
        }

        public String toString() {
            return getMOID() + ":" + getVirtualCenterIP();
        }
    }
}
