package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

public class VirtualPath extends XMLNode<VirtualizationRoot> {
    private final String moid;
    private final String virtualCenterIP;

    private final VmLun.VMLunKey lunKey;
    private String hostPortWWN;
    private String storagePortWWN;
    private boolean active;
    private VirtualPathType type;

    private final VirtualPathKey key;

    VirtualPath(final VirtualizationRoot parent, final String moid, final VmLun.VMLunKey lunKey, final String virtualCenterIP) throws SchemaException {
        super(parent, "HV_Path");

        if (moid == null || moid.length() == 0)
            throw new SchemaException("moid is mandatory!");

        if (virtualCenterIP == null || virtualCenterIP.length() == 0)
            throw new SchemaException("virtualCenterIP is mandatory!");

        if (lunKey == null)
            throw new SchemaException("Lun key is mandatory!");

        this.moid = moid;
        this.virtualCenterIP = virtualCenterIP;
        this.lunKey = lunKey;
        key = new VirtualPathKey(moid, virtualCenterIP, lunKey);
    }

    public Key getKey() {
        return key;
    }

    public String getMOID() {
        return moid;
    }

    public String getVirtualCenterIP() {
        return virtualCenterIP;
    }

    public VmLun.VMLunKey getLun() {
        return lunKey;
    }

    public String getHostPortWWN() {
        return hostPortWWN;
    }

    public void setHostPortWWN(final String hostPortWWN) {
        this.hostPortWWN = hostPortWWN;
    }

    public String getStoragePortWWN() {
        return storagePortWWN;
    }

    public void setStoragePortWWN(final String storagePortWWN) {
        this.storagePortWWN = storagePortWWN;
    }

    public boolean isActive() {
        return active;
    }

    public void setActive(final boolean active) {
        this.active = active;
    }

    /**
     * Get the virtual path type.  Optional.   Should discriminate between FC and iSCSI paths.     Server treats
     * missing value as FC.
     *
     * @return  the virtual path type.
     */
    public VirtualPathType getType() {
        return type;
    }

    public void setType(final VirtualPathType type) {
        this.type = type;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "OID", getKey().toString());

        addAttribute(writer, "HV_Lun_OID", lunKey);
        addAttribute(writer, "HostPortWWPN", hostPortWWN);
        addAttribute(writer, "StoragePortWWPN", storagePortWWN);
        addAttribute(writer, "Active", active);
        addAttribute(writer, "Type", type);
    }

    public enum VirtualPathType {
        ISCSI, FC;
    }

    public final static class VirtualPathKey implements Key<VirtualPathKey> {
        private final String moid;
        private final String virtualCenterIP;
        private final VmLun.VMLunKey lunKey;

        public VirtualPathKey(final String moid, final String virtualCenterIP, final VmLun.VMLunKey lunKey) throws SchemaException {

            if (moid == null || moid.length() == 0)
                throw new SchemaException("moid is mandatory!");

            if (virtualCenterIP == null || virtualCenterIP.length() == 0)
                throw new SchemaException("virtualCenterIP is mandatory!");

            if (lunKey == null)
                throw new SchemaException("Lun key is mandatory!");

            this.moid = moid;
            this.virtualCenterIP = virtualCenterIP;
            this.lunKey = lunKey;
        }

        public VmLun.VMLunKey getLunKey() {
            return lunKey;
        }

        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final VirtualPathKey that = (VirtualPathKey) o;

            if (lunKey != null ? !lunKey.equals(that.lunKey) : that.lunKey != null) return false;
            if (moid != null ? !moid.equals(that.moid) : that.moid != null) return false;
            if (virtualCenterIP != null ? !virtualCenterIP.equals(that.virtualCenterIP) : that.virtualCenterIP != null)
                return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (moid != null ? moid.hashCode() : 0);
            result = 31 * result + (virtualCenterIP != null ? virtualCenterIP.hashCode() : 0);
            result = 31 * result + (lunKey != null ? lunKey.hashCode() : 0);
            return result;
        }

        public int compareTo(final VirtualPathKey o) {
            int i = moid.compareTo(o.getMOID());
            if (i != 0)
                return i;
            i = lunKey.compareTo(o.getLunKey());
            if (i != 0)
                return i;
            return virtualCenterIP.compareTo(o.getVirtualCenterIP());
        }

        private String getVirtualCenterIP() {
            return virtualCenterIP;
        }

        public String getMOID() {
            return moid;
        }

        public String toString() {
            return getMOID() + ":" + getLunKey() + ":" + getVirtualCenterIP();
        }
    }
}
