package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

/**
 * Root-level element representing the association of a {virtual center, datacenter, virtual-cluster, host}
 */
public class VirtualServer extends XMLNode<VirtualizationRoot> {
    private final String virtualCenterIP;
    private final Host.HostKey host;
    private String clusterName;
    private String datacenterName;
    private String moid;

    private final VirtualServerKey key;

    /**
     * @param parent
     * @param host:           Host key
     * @param virtualCenterIP IP address of the Virtual Center managing this host
     *
     * @throws SchemaException: schema exeption
     */
    VirtualServer(final VirtualizationRoot parent, final Host.HostKey host, final String virtualCenterIP) throws SchemaException {
        super(parent, "HV_Server");

        if (virtualCenterIP == null || virtualCenterIP.length() == 0)
            throw new SchemaException("virtualCenterIP is mandatory!");

        if (host == null)
            throw new SchemaException("host id is mandatory!");

        this.virtualCenterIP = virtualCenterIP;
        this.host = host;
        key = new VirtualServerKey(virtualCenterIP, host);
    }

    public Key getKey() {
        return key;
    }

    public Host.HostKey getHost() {
        return host;
    }

    public String getVirtualCenterIP() {
        return virtualCenterIP;
    }

    public String getClusterName() {
        return clusterName;
    }

    public void setClusterName(final String clusterName) {
        this.clusterName = clusterName;
    }

    public String getDatacenterName() {
        return datacenterName;
    }

    public void setDatacenterName(final String datacenterName) {
        this.datacenterName = datacenterName;
    }

    public String getMoid() {
        return moid;
    }

    public void setMoid(final String moid) {
        this.moid = moid;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException, SchemaException {
        super.marshallAttributes(cache, writer);

        addAttribute(writer, "OID", "HV_Server:" + getKey());

        addAttribute(writer, "VirtualCenterIp", virtualCenterIP);
        addAttribute(writer, "ClusterName", clusterName);
        addAttribute(writer, "DataCenterName", datacenterName);
        addAttribute(writer, "Host_OID", host);
        addAttribute(writer, "MOID", moid);
    }

    public final static class VirtualServerKey implements Key<VirtualServerKey> {
        private String virtualCenterIP;
        private Host.HostKey host;

        public VirtualServerKey(final String virtualCenterIP, final Host.HostKey host) {
            this.virtualCenterIP = virtualCenterIP;
            this.host = host;
        }

        public int compareTo(final VirtualServerKey o) {
            int i = host.compareTo(o.getHostKey());
            if (i != 0)
                return i;

            return virtualCenterIP.compareTo(o.getVirtualCenterIP());
        }

        private String getVirtualCenterIP() {
            return virtualCenterIP;
        }

        public Host.HostKey getHostKey() {
            return host;
        }

        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final VirtualServerKey that = (VirtualServerKey) o;

            if (host != null ? !host.equals(that.host) : that.host != null) return false;
            if (virtualCenterIP != null ? !virtualCenterIP.equals(that.virtualCenterIP) : that.virtualCenterIP != null)
                return false;

            return true;
        }

        public int hashCode() {
            int result;
            result = (virtualCenterIP != null ? virtualCenterIP.hashCode() : 0);
            result = 29 * result + (host != null ? host.hashCode() : 0);
            return result;
        }

        public String toString() {
            return host.toString();
        }
    }
}
