package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.framework.datasource.IntegrityException;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

public class VirtualizationRootImpl extends BaseRootImpl implements VirtualizationRoot {
    final private VirtualServers servers = new VirtualServers(this);
    final private VirtualMachines machines = new VirtualMachines(this);
    final private DataStores dataStores = new DataStores(this);
    final private VmLuns luns = new VmLuns(this);
    final private Hosts hosts = new Hosts(this);
    final private VirtualDisks disks = new VirtualDisks(this);
    final private VirtualMachineToDisks machinesToDisks = new VirtualMachineToDisks(this);
    final private VirtualPaths paths = new VirtualPaths(this);
    final private NasShares nasShares = new NasShares(this);
    final private HostToNasShares hostToNasShares = new HostToNasShares(this);
    final private DataStoreToNasShares dataStoreToNasShares = new DataStoreToNasShares(this);

    protected VirtualizationRootImpl(final RootFactory parent, final long originatorID, final String originatorType, final String dataSourceName) throws SchemaException {
        super(parent, originatorID, originatorType, dataSourceName);
    }

    public VirtualServers getVirtualServers() {
        return servers;
    }

    public VirtualServer createVirtualServer(final String virtualCenterIP, final Host.HostKey host) throws SchemaException {
        return new VirtualServer(this, host, virtualCenterIP);
    }

    public VirtualMachines getVirtualMachines() {
        return machines;
    }

    public VirtualMachine createVirtualMachine(final String moid, final String managementIp) throws SchemaException {
        return new VirtualMachine(this, moid, managementIp);
    }

    public DataStores getDataStores() {
        return dataStores;
    }

    public DataStore createVirtualDataStore(final String oid, final String virtualCenterIP) throws SchemaException {
        return new DataStore(this, oid, virtualCenterIP);
    }

    public VmLuns getVirtualLuns() {
        return luns;
    }

    public VmLun createVmLun(final String oid, final String virtualCenterIP) throws SchemaException {
        return new VmLun(this, oid, virtualCenterIP);
    }

    public NasShares getNasShares() {
        return nasShares;
    }

    public Hosts getHosts() {
        return hosts;
    }

    public Host createHost(final String ips) throws SchemaException {
        return new Host(this, ips);
    }

    public NasShare createNasShare(final String virutalCenterIp, final String filerIp, final String sharePath, final NasShare.Type type) throws SchemaException {
        return new NasShare(this, virutalCenterIp, filerIp, sharePath, type);
    }

    public VirtualDisks getVirtualDisks() {
        return disks;
    }

    public VirtualDisk createVirtualDisk(final String moid, final String virtualCenterIP) throws SchemaException {
        return new VirtualDisk(this, moid, virtualCenterIP);
    }

    public VirtualMachineToDisks getVirtualMachineToDisks() {
        return machinesToDisks;
    }

    public VirtualMachineToDisk createVirtualMachineToDisk(final VirtualMachine.VirtualMachineKey vmkey, final VirtualDisk.VirtualDiskKey vdiskKey) throws SchemaException {
        return new VirtualMachineToDisk(this, vmkey, vdiskKey);
    }

    public VirtualPaths getVirtualPaths() {
        return paths;
    }

    public VirtualPath createVirtualPath(final String oid, final VmLun.VMLunKey lunOid, final String virtualCenterIP) throws SchemaException {
        return new VirtualPath(this, oid, lunOid, virtualCenterIP);
    }

    public HostToNasShare createHostToNasShare(final Host.HostKey hostKey, final NasShare.NasShareKey nasShareKey) throws SchemaException {
        return new HostToNasShare(this, hostKey, nasShareKey);
    }

    public HostToNasShares getHostToNasShares() {
        return hostToNasShares;
    }

    @Override
    public DataStoreToNasShare createDataStoreToNasShare(final DataStore.DataStoreKey dataStoreKey, final NasShare.NasShareKey nasShareKey) throws SchemaException {
        return new DataStoreToNasShare(this, dataStoreKey, nasShareKey);
    }

    @Override
    public DataStoreToNasShares getDataStoreToNasShares() {
        return dataStoreToNasShares;
    }

    protected void doMarshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, hosts.values(), tabs);
        addAllChildren(cache, writer, servers, tabs);
        addAllChildren(cache, writer, machines.values(), tabs);
        addAllChildren(cache, writer, dataStores.values(), tabs);
        addAllChildren(cache, writer, luns.values(), tabs);
        addAllChildren(cache, writer, disks.values(), tabs);
        addAllChildren(cache, writer, machinesToDisks.values(), tabs);
        addAllChildren(cache, writer, paths, tabs);
        addAllChildren(cache, writer, nasShares, tabs);
        addAllChildren(cache, writer, hostToNasShares, tabs);
        addAllChildren(cache, writer, dataStoreToNasShares, tabs);
    }

    public void checkIntegrity() throws IntegrityException {
        //check that all luns have path
        final Set<VmLun.VMLunKey> pathLansKeys = new HashSet<VmLun.VMLunKey>(paths.size());
        final Set<VmLun.VMLunKey> lunsWithOutPath = new HashSet<VmLun.VMLunKey>();
        final Set<DataStore.DataStoreKey> dataStoreKeys = new HashSet<DataStore.DataStoreKey>();
        final Set<DataStore.DataStoreKey> nonExistDataStore = new HashSet<DataStore.DataStoreKey>();

        //create a set path luns keyes
        for (final VirtualPath path : paths) {
            pathLansKeys.add(path.getLun());
        }

        for (final DataStore ds : dataStores.values()) {
            dataStoreKeys.add(ds.getKey());
        }

        //check that each lun has path
        for (final VmLun lun : luns.values()) {
            if (!pathLansKeys.contains(lun.getKey()))
                lunsWithOutPath.add(lun.getKey());

            //check that the lun point to valid datastore
            final DataStore.DataStoreKey dataStoreKeyFromLun = lun.getDataStore();
            if (dataStoreKeyFromLun != null && !dataStoreKeys.contains(lun.getDataStore())) {
                nonExistDataStore.add(lun.getDataStore());
            }

        }
        if (lunsWithOutPath.size() > 0 || nonExistDataStore.size() > 0)
            throw new IntegrityException((lunsWithOutPath.size() != 0 ? "The luns:" + Arrays.toString(lunsWithOutPath.toArray()) + " does not have any virtual path \n" : "") + (nonExistDataStore.size() != 0 ? "Luns that point to non existing datastor" + Arrays.toString(nonExistDataStore.toArray()) : ""));
    }

}
