package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;
import java.util.Set;
import java.util.TreeSet;

public class VolumeMap extends XMLNode<Volume> {
    private final String storagePortWWN;
    private final String protocolController;
    private Set<Integer> luns = new TreeSet<Integer>();
    private Integer tid = null;
    private Boolean wwnSecurityValidity = null;
    private Volume.MapMaskType type = null;

    private final VolumeMapKey key;

    /**
     * Use {@link Volume#createVolumeMap} to create this object
     *
     * @param storagePortWWN The storage port WWN through which this volume is mapped (if null, it means "map through all ports!")
     * @param protocolController The protocol controller associated with this mapping (if null, no protocol controllers are supported in this storage)
     */
    VolumeMap(final Volume parent, final String storagePortWWN, final String protocolController) {
        super(parent, "VolumeMap");

        this.storagePortWWN = storagePortWWN;
        this.protocolController = protocolController;

        key = new VolumeMapKey(parent.getKey(), storagePortWWN, protocolController);
    }

    public final VolumeMapKey getKey() {
        return key;
    }

    public final Volume getVolume() {
        return getParent();
    }

    public final Set<Integer> getLUNs() {
        return luns;
    }

    public final String getProtocolController() {
        return protocolController;
    }

    public final String getStoragePortWWN() {
        return storagePortWWN;
    }

    public final Integer getTID() {
        return tid;
    }

    public final void setTID(final Integer tid) {
        this.tid = tid;
    }

    public final Boolean getWwnSecurityValidity() {
        return wwnSecurityValidity;
    }

    public final void setWwnSecurityValidity(Boolean wwnSecurityValidity) {
        this.wwnSecurityValidity = wwnSecurityValidity;
    }


    public Volume.MapMaskType getType() {
        return type;
    }

    public void setType(final Volume.MapMaskType type) {
        this.type = type;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        String lun = setToCommaSeparetaedString(luns);
        addAttribute(writer, "LUN", lun);
        addAttribute(writer, "StoragePort", storagePortWWN);
        addAttribute(writer, "TID", tid);
        addAttribute(writer, "MaskingRequired", wwnSecurityValidity);
        addAttribute(writer, "ProtocolController", getProtocolController());
        addAttribute(writer, "Type", type);

    }

    private String setToCommaSeparetaedString(Set<Integer> set) {
        StringBuilder sb = new StringBuilder();
        for (Integer i : set) {
            if (sb.length() != 0) sb.append(",");
            sb.append(i.toString());
        }
        return sb.toString();
    }

    public static final class VolumeMapKey implements Key<VolumeMapKey> {
        private final BaseVolume.VolumeKey volumeKey;
        private final String storagePortWWN;
        private final String protocolController;

        public VolumeMapKey(final BaseVolume.VolumeKey key, final String storagePortWWN, final String protocolController) {
            volumeKey = key;
            this.storagePortWWN = storagePortWWN;
            this.protocolController = protocolController;
        }

        public final int compareTo(final VolumeMapKey o) {
            int i;

            // Must be the same volume
            i = getVolumeKey().compareTo(o.getVolumeKey());
            if (i != 0)
                return i;

            i = compareObjects(getStoragePortWWN(), o.getStoragePortWWN());
            if (i != 0)
                return i;

            return compareObjects(getProtocolController(), o.getProtocolController());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final VolumeMapKey that = (VolumeMapKey) o;

            if (protocolController != null ? !protocolController.equals(that.protocolController) : that.protocolController != null)
                return false;
            if (storagePortWWN != null ? !storagePortWWN.equals(that.storagePortWWN) : that.storagePortWWN != null)
                return false;
            if (volumeKey != null ? !volumeKey.equals(that.volumeKey) : that.volumeKey != null) return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (volumeKey != null ? volumeKey.hashCode() : 0);
            result = 29 * result + (storagePortWWN != null ? storagePortWWN.hashCode() : 0);
            result = 29 * result + (protocolController != null ? protocolController.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return volumeKey.toString() + ":" + storagePortWWN + ":" + protocolController;
        }

        public final Volume.VolumeKey getVolumeKey() {
            return volumeKey;
        }

        public final String getProtocolController() {
            return protocolController;
        }

        public final String getStoragePortWWN() {
            return storagePortWWN;
        }
    }


}
