package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.*;

import java.io.IOException;
import java.io.Writer;

/**
 * A WWN alias is a "hint" object sent from acquisition to the server to help with host resolution.
 *
 * <p/>Many management interfaces of different devices (switches, arrays, etc) contain a repository of identified WWNs.
 * This information is used when managing the device itself (it helps the administrator work with human readable names
 * instead of WWNs).
 *
 * <p/>Arrays typically have an alias mechanism to give names to host port to help with masking. It is common that the
 * array management interfaces require the user to create aliases before being able to perform any masking. When a storage
 * package wants to create an object to represent this kind of alias, it should use the {@link StorageRoot#createStorageHostAlias}
 * method.
 *
 * <p/>For switch management interfaces, aliases are usually used to help with zoning. These types of aliases is typically
 * called "Zone Aliases" and it is a common convention to encode the name of the host (and the storage) in the zone alias
 * name. When a switch package wants to create an object to represent this kind of alias, it should use the
 * {@link com.onaro.sanscreen.acquisition.schema.odata2.FabricRoot#createSwitchZoneAlias} method.
 *
 * <p/>A third type of alias is retrieved when integrating into a SRM system (like the EMC StorageScope application).
 * These systems usually have a repository of identified hosts with their port WWNs. When an SRM package wants to create
 * an object to represent this kind of alias, it should use the {@link com.onaro.sanscreen.acquisition.schema.odata2.SRMRoot#createSRMAlias} method.
 */
public class WWNAlias extends XMLNode<OData2Root> {
    private final Source source;
    private final String alias;
    private final String wwn;
    private final ObjectType objectType;

    private String ip = null;
    private String description = null;
    private String hostOs = null;
    private Integer hostFsFree = null;
    private Integer hostFsTotal = null;
    private Integer hostCpuCount = null;
    private Integer hostDbCount = null;
    private String hostDomain = null;
    private Integer hostInstalledMemory = null;
    private String hostModel = null;
    private String hostOsClass = null;
    private String hostOsLevel = null;
    private String hostOsRelease = null;
    private String hostVendorName = null;
    private String firmware = null;

    private final WWNAliasKey key;

    /**
     * Use {@link OData2Root#getWWNAliases} to access this object
     */
    WWNAlias(final OData2Root parent, final Source source, final String alias, final String wwn, final ObjectType objectType) throws SchemaException {
        super(parent, "WWNAlias");

        if (source == null || alias == null || alias.length() == 0 || wwn == null || wwn.length() == 0 || objectType == null)
            throw new SchemaException(getNodeName() + " Type requires non-null source, alias, wwn, objectType");

        this.source = source;
        this.alias = alias;
        this.wwn = wwn.toUpperCase();
        this.objectType = objectType;

        key = new WWNAliasKey(source, alias, wwn, objectType);
    }

    public final WWNAliasKey getKey() {
        return key;
    }

    public final WWNAlias cloneAlias(final OData2Root parent, final Source source, final String name, final String wwn, final ObjectType objectType) throws SchemaException {
        WWNAlias alias = new WWNAlias(parent, source, name, wwn, objectType);

        alias.ip = this.ip;
        alias.description = this.description;
        alias.hostOs = this.hostOs;
        alias.hostFsFree = this.hostFsFree;
        alias.hostFsTotal = this.hostFsTotal;
        alias.hostCpuCount = this.hostCpuCount;
        alias.hostDbCount = this.hostDbCount;
        alias.hostDomain = this.hostDomain;
        alias.hostInstalledMemory = this.hostInstalledMemory;
        alias.hostModel = this.hostModel;
        alias.hostOsClass = this.hostOsClass;
        alias.hostOsLevel = this.hostOsLevel;
        alias.hostOsRelease = this.hostOsRelease;
        alias.hostVendorName = this.hostVendorName;

        return alias;
    }

    public final Root getRoot() {
        return getParent();
    }

    public final Source getSource() {
        return source;
    }

    public final String getAlias() {
        return alias;
    }

    public final String getWWN() {
        return wwn;
    }

    public final ObjectType getObjectType() {
        return objectType;
    }

    public final String getDescription() {
        return description;
    }

    public final void setDescription(final String description) {
        this.description = description;
    }

    public final String getIP() {
        return ip;
    }

    public final void setIP(final String ip) {
        this.ip = ip;
    }

    public final String getHostVendorName() {
        return hostVendorName;
    }

    public final void setHostVendorName(final String hostVendorName) {
        this.hostVendorName = hostVendorName;
    }

    public final String getHostOsRelease() {
        return hostOsRelease;
    }

    public final void setHostOsRelease(final String hostOsRelease) {
        this.hostOsRelease = hostOsRelease;
    }

    public final String getHostOsLevel() {
        return hostOsLevel;
    }

    public final void setHostOsLevel(final String hostOsLevel) {
        this.hostOsLevel = hostOsLevel;
    }

    public final String getHostOsClass() {
        return hostOsClass;
    }

    public final void setHostOsClass(final String hostOsClass) {
        this.hostOsClass = hostOsClass;
    }

    public final String getHostModel() {
        return hostModel;
    }

    public final void setHostModel(final String hostModel) {
        this.hostModel = hostModel;
    }

    public final Integer getHostInstalledMemory() {
        return hostInstalledMemory;
    }

    public final void setHostInstalledMemory(final Integer hostInstalledMemory) {
        this.hostInstalledMemory = hostInstalledMemory;
    }

    public final String getHostDomain() {
        return hostDomain;
    }

    public final void setHostDomain(final String hostDomain) {
        this.hostDomain = hostDomain;
    }

    public final Integer getHostDbCount() {
        return hostDbCount;
    }

    public final void setHostDbCount(final Integer hostDbCount) {
        this.hostDbCount = hostDbCount;
    }

    public final Integer getHostCpuCount() {
        return hostCpuCount;
    }

    public final void setHostCpuCount(final Integer hostCpuCount) {
        this.hostCpuCount = hostCpuCount;
    }

    public final Integer getHostFsTotal() {
        return hostFsTotal;
    }

    public final void setHostFsTotal(final Integer hostFsTotal) {
        this.hostFsTotal = hostFsTotal;
    }

    public final Integer getHostFsFree() {
        return hostFsFree;
    }

    public final void setHostFsFree(final Integer hostFsFree) {
        this.hostFsFree = hostFsFree;
    }

    public final String getHostOs() {
        return hostOs;
    }

    public final void setHostOs(final String hostOs) {
        this.hostOs = hostOs;
    }

    public final String getFirmware() {
        return firmware;
    }

    public final void setFirmware(final String firmware) {
        this.firmware = firmware;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "Source", source.toString());
        addAttribute(writer, "Alias", alias);
        addAttribute(writer, "SourceDescription", description);
        addAttribute(writer, "WWN", wwn);
        addAttribute(writer, "ObjectType", objectType);
        addAttribute(writer, "IP", ip);

        addAttribute(writer, "HostOs", hostOs);
        addAttribute(writer, "HostFsFree", hostFsFree);
        addAttribute(writer, "HostFsTotal", hostFsTotal);
        addAttribute(writer, "HostCpuCount", hostCpuCount);
        addAttribute(writer, "HostDbCount", hostDbCount);
        addAttribute(writer, "HostDomain", hostDomain);
        addAttribute(writer, "HostInstalledMemory", hostInstalledMemory);
        addAttribute(writer, "HostModel", hostModel);
        addAttribute(writer, "HostOsClass", hostOsClass);
        addAttribute(writer, "HostOsLevel", hostOsLevel);
        addAttribute(writer, "HostOsRelease", hostOsRelease);
        addAttribute(writer, "HostVendorName", hostVendorName);
        addAttribute(writer, "Firmware", firmware);
    }

    public static enum Source {
        /**
         * Alias originated from a switch zone alias
         */
        Switch,
        /**
         * Alias originated from a storage management application host alias
         */
        Storage,
        /**
         * Alias originated from an SRM application (like ECC)
         */
        SRM
    }

    /**
     * The object type is specified to let the server know what kind of object is being reported.
     *
     * <p/>When the package implementation does not know if this object is a <code>host</code>, <code>storage</code> or
     * <code>tape</code>, it should use the <code>unknown</code> value.
     */
    public static enum ObjectType {
        Host,
        Tape,
        Storage,
        Unknown
    }

    public static final class WWNAliasKey implements Key<WWNAliasKey> {
        private final Source source;
        private final String alias;
        private final String wwn;
        private final ObjectType objectType;

        public WWNAliasKey(final Source source, final String alias, final String wwn, final ObjectType objectType) {
            this.source = source;
            this.alias = alias;
            this.wwn = wwn;
            this.objectType = objectType;
        }

        public final int compareTo(final WWNAliasKey o) {
            int i;

            i = source.ordinal() - o.getSource().ordinal();
            if (i != 0)
                return i;

            i = objectType.ordinal() - o.getObjectType().ordinal();
            if (i != 0)
                return i;

            i = alias.compareTo(o.getAlias());
            if (i != 0)
                return i;

            return wwn.compareTo(o.getWWN());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final WWNAliasKey that = (WWNAliasKey) o;

            if (alias != null ? !alias.equals(that.alias) : that.alias != null) return false;
            if (objectType != that.objectType) return false;
            if (source != that.source) return false;
            if (wwn != null ? !wwn.equals(that.wwn) : that.wwn != null) return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (source != null ? source.hashCode() : 0);
            result = 29 * result + (alias != null ? alias.hashCode() : 0);
            result = 29 * result + (wwn != null ? wwn.hashCode() : 0);
            result = 29 * result + (objectType != null ? objectType.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return objectType + ":" + wwn + ":" + alias + ":" + source;
        }

        public final Source getSource() {
            return source;
        }

        public final String getAlias() {
            return alias;
        }

        public final String getWWN() {
            return wwn;
        }

        public final ObjectType getObjectType() {
            return objectType;
        }
    }
}
