package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Nodes;
import com.onaro.sanscreen.acquisition.schema.SchemaException;

import java.util.*;

/**
 * A collection representing a set of HostAliases ({@link WWNAlias}).
 */
public class WWNAliases extends Nodes<OData2Root, WWNAlias> {
    WWNAliases(final OData2Root parent) {
        super(parent);
    }

    /**
     * If two or more aliases referencing the same port - it creates problems on the server side.
     * This method makes sure to return collection of aliases that point to unique members.
     * In case of a conflict, alias names will be concatenated with comma to seperate the names.
     *
     * @return collection of normalized Aliases
     */
    Collection<WWNAlias> prepareForMarshalling() throws SchemaException {
        // Create a map of WWNs to Aliases that reference them (invert the list of aliases)
        // Since each wwn can be referenced by more than one alias - we store them in ordered set
        final Map<String, TreeSet<WWNAlias>> memberToSetOfAliases = new TreeMap<String, TreeSet<WWNAlias>>();
        for (WWNAlias wwnAlias : this) {
            TreeSet<WWNAlias> foundAliases = memberToSetOfAliases.get(wwnAlias.getWWN());
            if (foundAliases == null)
                foundAliases = new TreeSet<WWNAlias>(new Nodes.NodeComparator());

            foundAliases.add(wwnAlias);
            memberToSetOfAliases.put(wwnAlias.getWWN(), foundAliases);
        }

        // Now recombine aliases
        List<WWNAlias> result = new LinkedList<WWNAlias>();
        for (Map.Entry<String, TreeSet<WWNAlias>> entry : memberToSetOfAliases.entrySet()) {
            String wwn = entry.getKey();
            if (entry.getValue().size() == 1) {
                // there is going to be only one alias
                for (WWNAlias alias : entry.getValue())
                    result.add(alias);
            }
            else {
                // multiple aliases should be combined
                StringBuilder sbNames = new StringBuilder();
                StringBuilder sbIPs = new StringBuilder();

                WWNAlias aliasForExample = null;

                for (WWNAlias alias : entry.getValue()) {
                    sbNames.append(alias.getAlias()).append(";");
                    if (alias.getIP() != null && alias.getIP().length() != 0)
                        sbIPs.append(alias.getIP()).append(";");
                    aliasForExample = alias;
                }

                sbNames.deleteCharAt(sbNames.length() - 1);
                if (sbIPs.length() != 0)
                    sbIPs.deleteCharAt(sbIPs.length() - 1);

                WWNAlias newAlias = aliasForExample.cloneAlias(this.getParent(),
                                                               aliasForExample.getSource(),
                                                               sbNames.toString(),
                                                               wwn,
                                                               aliasForExample.getObjectType());


                //IBG-1357: Make sure there is some sort of content to the IP string builder
                //or else do not set this. Otherwise, we get an empty ip field in the WWN alias
                if (sbIPs.length() >  0) {
                    newAlias.setIP(sbIPs.toString());
                }

                result.add(newAlias);
            }
        }

        return result;
    }
}
