package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

/**
 * Represents a zone member
 */
public abstract class ZoneMember extends XMLNode<Zone> {
    private final ZoneMemberTypeEnum type;
    private final String wwn;

    private final ZoneMemberKey key;

    /**
     * Use either {@link Zone#createPortZoneMember} or {@link Zone#createWWNZoneMember} to construct an instance of this object
     */
    ZoneMember(final Zone parent, final ZoneMemberTypeEnum type, final String wwn) throws SchemaException {
        super(parent, "ZoneMember");

        if (type == null || wwn == null || wwn.length() == 0)
            throw new SchemaException(getNodeName() + " Type requires non-null type, oid and wwn");

        this.type = type;
        this.wwn = wwn.toUpperCase();

        key = new ZoneMemberKey(parent.getKey(), type, wwn);
    }

    public final ZoneMemberKey getKey() {
        return key;
    }

    public final Zone getZone() {
        return getParent();
    }

    public final ZoneMemberTypeEnum getType() {
        return type;
    }

    public final String getWWN() {
        return wwn;
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "WWN", wwn);
        addAttribute(writer, "Type", type);
    }

    /**
     * Represents the different zone member types supported by the server
     */
    public enum ZoneMemberTypeEnum {
        SWITCH_PORT,
        WWN;
    }

    public static final class ZoneMemberKey implements Key<ZoneMemberKey> {
        private final Zone.ZoneKey zoneKey;
        private final ZoneMemberTypeEnum type;
        private final String wwn;

        public ZoneMemberKey(final Zone.ZoneKey zoneKey, final ZoneMemberTypeEnum type, final String wwn) {
            this.zoneKey = zoneKey;
            this.type = type;
            this.wwn = wwn;
        }

        public final int compareTo(final ZoneMemberKey o) {
            int i;

            i = wwn.compareTo(o.getWWN());
            if (i != 0)
                return i;

            i = type.compareTo(o.getType());
            if (i != 0)
                return i;

            return zoneKey.compareTo(o.getZoneKey());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final ZoneMemberKey that = (ZoneMemberKey) o;

            if (type != that.type) return false;
            if (wwn != null ? !wwn.equals(that.wwn) : that.wwn != null) return false;
            if (zoneKey != null ? !zoneKey.equals(that.zoneKey) : that.zoneKey != null) return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (zoneKey != null ? zoneKey.hashCode() : 0);
            result = 29 * result + (type != null ? type.hashCode() : 0);
            result = 29 * result + (wwn != null ? wwn.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return zoneKey.toString() + ":" + type + ":" + wwn;
        }

        public final String getWWN() {
            return wwn;
        }

        public final ZoneMemberTypeEnum getType() {
            return type;
        }

        public final Zone.ZoneKey getZoneKey() {
            return zoneKey;
        }
    }
}
