package com.onaro.sanscreen.acquisition.schema.odata2;

import com.onaro.sanscreen.acquisition.schema.Key;
import com.onaro.sanscreen.acquisition.schema.MarshallingCache;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.sanscreen.acquisition.schema.XMLNode;

import java.io.IOException;
import java.io.Writer;

public abstract class ZoningCapabilities extends XMLNode<Fabric> {
    private final String configurationName;
    private final boolean activeConfiguration;

    private ZoningBehavior defaultZoningBehavior = null;
    private MergeControl mergeControl = null;

    private final Zones zones;

    private final ZoningCapabilitiesKey key;

    /**
     * This object is abstract.
     * Use {@link ActiveZoningCapabilities} or {@link InactiveZoningCapabilities}
     */
    ZoningCapabilities(final Fabric parent, final String name, boolean activeConfiguration) {
        super(parent, "ZoningCapabilities");

        this.configurationName = name;
        this.activeConfiguration = activeConfiguration;

        zones = new Zones(this);

        key = new ZoningCapabilitiesKey(parent.getKey(), name, activeConfiguration);
    }

    public final ZoningCapabilitiesKey getKey() {
        return key;
    }

    public final <F extends Fabric> F getFabric() {
        //noinspection unchecked
        return (F) getParent();
    }

    public final Zones getZones() {
        return zones;
    }

    public final Zone createZone(final String zoneName) throws SchemaException {
        return new Zone(this, zoneName);
    }

    public final boolean isActiveConfiguration() {
        return activeConfiguration;
    }

    public final String getConfigurationName() {
        return configurationName;
    }

    public final ZoningBehavior getDefaultZoningBehavior() {
        return defaultZoningBehavior;
    }

    public final void setDefaultZoningBehavior(ZoningBehavior defaultzoningbehavior) {
        this.defaultZoningBehavior = defaultzoningbehavior;
    }

    public final MergeControl getMergeControl() {
        return mergeControl;
    }

    public final void setMergeControl(MergeControl mergeControl) {
        this.mergeControl = mergeControl;
    }

    public final String getWWN() {
        return getParent().getWWN();
    }

    protected void marshallAttributes(final MarshallingCache cache, final Writer writer) throws IOException {
        addAttribute(writer, "WWN", getWWN());
        addAttribute(writer, "DefaultZoningBehavior", defaultZoningBehavior);
        addAttribute(writer, "ActiveConfiguration", (activeConfiguration ? "1" : "0"));
        addAttribute(writer, "ConfigurationName", configurationName);
        addAttribute(writer, "MergeControl", mergeControl);
    }

    protected void marshallChildren(final MarshallingCache cache, final Writer writer, final int tabs) throws IOException, SchemaException {
        addAllChildren(cache, writer, zones, tabs);
    }

    protected boolean hasChildren() {
        return !zones.isEmpty();
    }

    public static final class ZoningCapabilitiesKey implements Key<ZoningCapabilitiesKey> {
        private final Fabric.FabricKey fabricKey;
        private final String name;
        private final Boolean activeConfiguration;

        public ZoningCapabilitiesKey(final Fabric.FabricKey fabricKey, final String name, final boolean activeConfiguration) {
            this.fabricKey = fabricKey;
            this.name = name;
            this.activeConfiguration = activeConfiguration;
        }

        public final int compareTo(final ZoningCapabilitiesKey o) {
            int i;

            i = fabricKey.compareTo(o.getFabricKey());
            if (i != 0)
                return i;

            i = compareObjects(name, o.getName());
            if (i != 0)
                return i;

            return activeConfiguration.compareTo(o.isActiveConfiguration());
        }

        public final boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final ZoningCapabilitiesKey that = (ZoningCapabilitiesKey) o;

            if (activeConfiguration != null ? !activeConfiguration.equals(that.activeConfiguration) : that.activeConfiguration != null)
                return false;
            if (fabricKey != null ? !fabricKey.equals(that.fabricKey) : that.fabricKey != null) return false;
            if (name != null ? !name.equals(that.name) : that.name != null) return false;

            return true;
        }

        public final int hashCode() {
            int result;
            result = (fabricKey != null ? fabricKey.hashCode() : 0);
            result = 29 * result + (name != null ? name.hashCode() : 0);
            result = 29 * result + (activeConfiguration != null ? activeConfiguration.hashCode() : 0);
            return result;
        }

        public final String toString() {
            return fabricKey.toString() + ":" + name + ":" + activeConfiguration;
        }

        public final Fabric.FabricKey getFabricKey() {
            return fabricKey;
        }

        public final boolean isActiveConfiguration() {
            return activeConfiguration;
        }

        public final String getName() {
            return name;
        }
    }

    /**
     * Specifies how the switch behaves when no zoning is defined.
     */
    public static enum ZoningBehavior {
        /**
         * If zoning disabled, all accesses is allowed (permit all)
         */
        PERMIT(0),
        /**
         * If zoning disabled, no accesses is allowed (deny all)
         */
        RESTRICT(1);

        private int e;

        ZoningBehavior(int e) {
            this.e = e;
        }

        public String toString() {
            return Integer.toString(e);
        }

        public int toEnum() {
            return e;
        }
    }

    /**
     * Specifies the behavior of zone merging when trying to merge fabrics
     */
    public static enum MergeControl {
        /**
         * If zones are not identical, fabrics will not merge
         */
        RESTRICT(0),
        /**
         * If the common zones in two fabrics are identical, will merge including uncommon zones
         */
        PERMIT(1);

        private int e;

        MergeControl(int e) {
            this.e = e;
        }

        public String toString() {
            return Integer.toString(e);
        }

        public int toEnum() {
            return e;
        }
    }
}
