package com.onaro.sanscreen.acquisition.schema2;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Collection of utility methods related to marshalling.
 */
public class MarshallingUtil {

    private final static Logger logger = Logger.getLogger(MarshallingUtil.class);

    /**
     * Lists of annotated fields indexed by (class, annotation) pair.
     * Speeds up object marshalling by over an order of magnitude.
     */
    private final static Map<CacheKey, List<Field>> cache =
            new HashMap<CacheKey, List<Field>>(512);

    /**
     * Extract the value of fields annotated with the specified annotation.
     * Excludes fields with <tt>null</tt> values, as well as empty collections
     * and maps.
     *
     * @return A map of field name/value pairs.
     */
    public static Map<Field, Object> extractFieldValues(final Object obj,
            final Class<? extends Annotation> annotationClass) {
        // Look up list of fields for this object. Create if needed.
        List<Field> fields;
        final CacheKey key = new CacheKey(obj.getClass(), annotationClass);
        synchronized (cache) {
            fields = cache.get(key);
            if (fields == null) {
                fields = extractAnnotatedFields(obj.getClass(), annotationClass,
                        true);
                cache.put(key, fields);
            }
        }
        // For consistency, use LinkedHashMap to preserve field order. Even
        // though javadoc says returned fields "are not in any particular
        // order", they do seem to be returned in declared order.
        final Map<Field, Object> attributes =
                new LinkedHashMap<Field, Object>(fields.size());
        for (final Field field : fields) {
            try {
                field.setAccessible(true); // suppress Java language access checking
                final Object value = field.get(obj);
                if (value == null) {
                    // Skip null values
                    continue;
                }
                if ((value instanceof Map<?, ?>)
                        && ((Map<?, ?>)value).isEmpty()) {
                    // Skip empty maps
                    continue;
                }
                if ((value instanceof Collection<?>)
                        && ((Collection<?>)value).isEmpty()) {
                    // Skip empty collections
                    continue;
                }
                attributes.put(field, value);
            }
            catch (Exception e) {
                logger.warn("Error extracting value from field: " + field, e);
            }
        }
        return attributes;
    }

    /**
     * Extract all fields marked as <tt>annotationClass</tt> from the specified
     * class. Includes public, protected, default (package) access, private
     * fields, and inherited fields.
     *
     * @param clazz The target class.
     * @param annotationClass The target field annotation.
     * @param superclassFirst If <tt>true</tt>, start with fields declared in
     *        the top level superclass and work downwards.
     * @return The list of all attribute fields.
     */
    public static List<Field> extractAnnotatedFields(final Class<?> clazz,
            final Class<? extends Annotation> annotationClass,
            final boolean superclassFirst) {
        final List<Field> fields = new LinkedList<Field>();
        final List<Class<?>> classHierarchy = new LinkedList<Class<?>>();
        for (Class<?> c = clazz; c != null; c = c.getSuperclass()) {
            classHierarchy.add(c);
        }
        if (superclassFirst) {
            Collections.reverse(classHierarchy);
        }
        for (final Class<?> c : classHierarchy) {
            for (final Field field : c.getDeclaredFields()) {
                // Only return fields marked as NodeAttribute
                if (field.getAnnotation(annotationClass) != null) {
                    fields.add(field);
                }
            }
        }
        return fields;
    }

    private static class CacheKey {
        private final Class<?> c;
        private final Class<? extends Annotation> a;
        public CacheKey(final Class<?> c, final Class<? extends Annotation> a) {
            this.c = c;
            this.a = a;
        }
        @Override
        public int hashCode() {
            return c.hashCode() ^ a.hashCode();
        }
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof CacheKey) {
                final CacheKey that = (CacheKey)obj;
                return ObjectUtils.equals(this.c, that.c)
                        && ObjectUtils.equals(this.a, that.a);
            }
            return false;
        }
    }

}
