package com.onaro.sanscreen.acquisition.schema2;

import com.onaro.sanscreen.acquisition.schema2.collect.NodeSet;
import com.onaro.sanscreen.acquisition.framework.datasource.IntegrityException;
import com.onaro.sanscreen.acquisition.schema.SchemaException;
import com.onaro.util.Capacity;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Base class for all o-data nodes. This object introduces the notion of a
 * parent object since the o-data model is tree based.
 */
abstract public class XmlNode implements Node {

    protected static final String lineSeperator = System.getProperty(
            "line.separator");

    private final String nodeName;

    protected XmlNode(final String name) {
        this.nodeName = name;
    }

    public final String getNodeName() {
        return nodeName;
    }

    /**
     * Marshal the object into the appropriate XML schema.
     * <p/>
     * Usually this method should not be overridden. Instead, override {@link
     * #marshallAttributes}, {@link #marshallChildren} and {@link #hasChildren}
     * to handle the object marshalling.
     *
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshallNode(final Writer writer, final int tabs)
            throws IOException, SchemaException {
        addTabulation(writer, tabs);
        writer.append("<");
        writer.append(getNodeName());

        marshallAttributes(writer);

        if (hasChildren()) {
            writer.append(">");
            writer.append(lineSeperator);

            marshallChildren(writer, tabs + 1);

            addTabulation(writer, tabs);
            writer.append("</");
            writer.append(getNodeName());
            writer.append(">");
            writer.append(lineSeperator);
        }
        else {
            // No children
            writer.append(" />");
            writer.append(lineSeperator);
        }
    }

    /**
     * Indicates if this object has any XML child elements. When true, {@link
     * #marshallChildren} should also be overriden.
     *
     * @return true if node has children
     */
    protected boolean hasChildren() {
        return hasNodeChild();
    }

    /**
     * @return <tt>true</tt> if this node has at least one non-null child node.
     */
    protected boolean hasNodeChild() {
        return !MarshallingUtil.extractFieldValues(this, NodeChild.class)
                .isEmpty();
    }

    /**
     * Adds all the object's XML attribute during marshalling
     *
     * @param writer Writer to dump XML to
     * @throws com.onaro.sanscreen.acquisition.schema.SchemaException If an XML
     * integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshallAttributes(final Writer writer)
            throws IOException, SchemaException {
        final Map<Field, Object> attributes =
                MarshallingUtil.extractFieldValues(this, NodeAttribute.class);
        final Set<String> attributeNames = new HashSet<String>(
                attributes.size());
        for (final Map.Entry<Field, Object> e : attributes.entrySet()) {
            // Don't allow duplicate attributes
            final Field field = e.getKey();
            final String attributeName = createAttributeName(field);
            if (attributeNames.contains(attributeName)) {
                throw new SchemaException(
                        "Found duplicate attributes: " + attributeName);
            }
            attributeNames.add(attributeName);
            // Special case for fields of type Capacity
            final Object value = e.getValue();
            if (value instanceof Capacity) {
                addAttribute(writer, attributeName, (Capacity)value,
                        field.getAnnotation(NodeAttribute.class).unit());
            }
            else {
                addAttribute(writer, attributeName, value);
            }
        }
    }

    /**
     * Create an XML attribute name for the field with the specified name. By
     * default, converts camel case to lower case separated by dashes. For
     * example, &quot;fooBar&quot; -&gt; &quot;foo-bar&quot;.
     */
    protected String createAttributeName(final Field field) {
        // Check for attribute name override in annotation
        final NodeAttribute nodeAttribute = field.getAnnotation(
                NodeAttribute.class);
        if ((nodeAttribute != null) && StringUtils.isNotEmpty(
                nodeAttribute.name())) {
            return nodeAttribute.name();
        }
        // No override, so generate our own attribute name
        final String fieldName = field.getName();
        final StringBuilder attributeName = new StringBuilder(
                fieldName.length() + 10);
        for (int i = 0, len = fieldName.length(); i < len; i++) {
            char c = fieldName.charAt(i);
            if (Character.isUpperCase(c)) {
                attributeName.append('-');
                c = Character.toLowerCase(c);
            }
            attributeName.append(c);
        }
        return attributeName.toString();
    }

    /**
     * Write an XML attribute to the object during marshalling. This method
     * takes care of <tt>null</tt> values (the attribute will not be added) and
     * XML escaping.
     *
     * @param writer Writer to dump XML to
     * @param attributeName XML attribute name.
     * @param attributeValue Attribute value. If null, the attribute will not be
     * set. The {@link String#valueOf(Object)} method will be used to convert
     * the object's value into a string.
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAttribute(final Writer writer,
            final String attributeName, final Object attributeValue)
            throws IOException {
        // Ignore null values
        if (attributeValue == null) {
            return;
        }
        writer.append(' ');
        writer.append(attributeName);
        writer.append("=\"");
        writer.append(escape(attributeValue.toString()));
        writer.append('"');
    }

    /**
     * Write a capacity XML attribute to the object during marshalling. This
     * method takes care of null values (the attribute will not be added) and
     * XML escaping. This also will translate the capacity to the desired unit.
     *
     * @param writer Writer to dump XML to
     * @param attributeName XML attribute name.
     * @param capacity value. If null, the attribute will not be set.
     * @throws java.io.IOException If an IO error occurs
     */
    protected final void addAttribute(final Writer writer,
            final String attributeName, final Capacity capacity,
            final Capacity.Unit unit) throws IOException {
        if (capacity == null) {
            return;
        }
        writer.append(' ');
        writer.append(attributeName);
        writer.append("=\"");
        writer.append(String.valueOf(capacity.value(unit)));
        writer.append('"');
    }

    /**
     * Adds the objects XML child elements during marshalling
     *
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     * @throws SchemaException If an XML integrity error occurs
     * @throws java.io.IOException If an IO error occurs
     */
    protected void marshallChildren(final Writer writer, final int tabs)
            throws IOException, SchemaException {
        final Map<Field, Object> children = MarshallingUtil.extractFieldValues(
                this, NodeChild.class);
        for (final Map.Entry<Field, Object> e : children.entrySet()) {
            final Field field = e.getKey();
            final Object value = e.getValue();
            if (value instanceof XmlChildNode) {
                ((XmlChildNode)value).marshallNode(writer, tabs);
            }
            else if (value instanceof NodeSet) {
                marshallNodeSet(writer, tabs, (NodeSet<?, ?>)value, field);
            }
            else if (value instanceof Collection) {
                marshallCollection(writer, tabs, (Collection<?>)value, field);
            }
            else {
                throw new SchemaException("Unable to marshall child: "
                        + field.getName() + " of type: "
                        + value.getClass().getName()
                        + " on node of type: " + getClass().getName());
            }
        }
    }

    /**
     * Marshall the specified node set.
     */
    protected void marshallNodeSet(final Writer writer,
            final int tabs, final NodeSet<?, ?> nodeSet, final Field field)
            throws IOException, SchemaException {
        final NodeChild annotation = field.getAnnotation(NodeChild.class);
        String elementName = annotation.elementName();
        if (StringUtils.isEmpty(elementName)) {
            // Use NodeSet class name if element name not explicitly defined
            elementName = nodeSet.getClass().getSimpleName();
        }
        addTabulation(writer, tabs);
        writer.append("<");
        writer.append(elementName);
        writer.append(">");
        writer.append(lineSeperator);

        for (final XmlChildNode node : nodeSet) {
            node.marshallNode(writer, tabs + 1);
        }

        addTabulation(writer, tabs);
        writer.append("</");
        writer.append(elementName);
        writer.append(">");
        writer.append(lineSeperator);
    }

    /**
     * Marshall the specified collection.
     *
     * @param field The field holding the specified collection.
     */
    protected void marshallCollection(final Writer writer,
            final int tabs, final Collection<?> collection, final Field field)
            throws IOException, SchemaException {
        if ((collection == null) || (collection.size() == 0)) {
            // Nothing to marshall
            return;
        }
        final NodeChild annotation = field.getAnnotation(NodeChild.class);
        final String elementName = annotation.elementName();
        final String attributeName = annotation.attributeName();
        if (StringUtils.isEmpty(elementName)) {
            throw new SchemaException("Missing elementName for child "
                    + field.getName());
        }
        if (StringUtils.isEmpty(attributeName)) {
            throw new SchemaException("Missing attributeName for child "
                    + field.getName());
        }
        for (final Object obj : collection) {
            addTabulation(writer, tabs);
            writer.append("<");
            writer.append(elementName);
            writer.append(" ");
            writer.append(attributeName);
            writer.append("=\"");
            writer.append(obj.toString());
            writer.append("\" />");
            writer.append(lineSeperator);
        }
    }

    /**
     * Sets up the tabulation to present the XML is a better format
     *
     * @param writer Writer to dump XML to
     * @param tabs The indentation of this node (to make XML more readable)
     * @throws java.io.IOException If an IO error occurs
     */
    protected static void addTabulation(final Writer writer, final int tabs)
            throws IOException {
        for (int i = 0; i < tabs; i++) {
            writer.append('\t');
        }
    }

    /**
     * Converts the specified string to an XML appropriate version by escaping
     * reserved ascii characters
     *
     * @param unescapedString String to encode
     * @return XML legal string representation of suspect string
     */
    @SuppressWarnings({"ConstantConditions"})
    protected static String escape(final String unescapedString) {
        final StringBuilder escapedString = new StringBuilder(
                unescapedString.length());

        for (int i = 0; i < unescapedString.length(); i++) {
            final char unescapeCharacter = unescapedString.charAt(i);

            switch (unescapeCharacter) {
                case '"':
                    escapedString.append("&quot;");
                    break;
                case '\'':
                    escapedString.append("&apos;");
                    break;
                case '&':
                    escapedString.append("&amp;");
                    break;
                case '<':
                    escapedString.append("&lt;");
                    break;
                case '>':
                    escapedString.append("&gt;");
                    break;
                default:
                    escapedString.append(unescapeCharacter);
            }
        }

        return escapedString.toString();
    }

    /**
     * Recursively call {@link #checkIntegrity()} on all children.
     * <p/>
     * Should never be called directly by subclasses.
     */
    void checkIntegrityOfChildren() throws IntegrityException {
        final Map<Field, Object> children = MarshallingUtil.extractFieldValues(
                this, NodeChild.class);
        for (final Map.Entry<Field, Object> e : children.entrySet()) {
            final Object value = e.getValue();
            if (value instanceof XmlChildNode) {
                final XmlChildNode node = (XmlChildNode)value;
                node.checkIntegrity();
                node.checkIntegrityOfChildren();
            }
            else if (value instanceof NodeSet) {
                final NodeSet<?, ?> nodeSet = (NodeSet<?, ?>)value;
                for (final XmlChildNode node : nodeSet) {
                    node.checkIntegrity();
                    node.checkIntegrityOfChildren();
                }
            }
        }
    }

}
