package com.onaro.sanscreen.acquisition.schema2.collect;

import org.apache.commons.lang3.ObjectUtils;
import org.apache.commons.lang3.StringUtils;

import java.io.Serializable;

/**
 * Collection of static methods related to keyable objects.
 */
public class NkeyUtils {

    private NkeyUtils() {
    }

    /**
     * @param keyable1 The input keyable object.
     * @return A string representation of the specified {@link Keyable1} object.
     */
    public static String toString(final Keyable1<?> keyable1) {
        return ObjectUtils.toString(keyable1.getKey1(), "null");
    }

    /**
     * @param keyable2 The input keyable object.
     * @return A string representation of the specified {@link Keyable1}
     *         object.
     */
    public static String toString(final Keyable2<?, ?> keyable2) {
        return toString(keyable2, ":");
    }

    /**
     * Concatenate the string representations of the keyable object's two keys,
     * using <tt>separator</tt> in between.
     *
     * @param keyable2 The input keyable object.
     * @param separator The separator between the string representations of the
     *        two keys.
     * @return A string representation of the specified {@link Keyable2}
     *         object.
     */
    public static String toString(final Keyable2<?, ?> keyable2,
            final String separator) {
        final StringBuilder buf = new StringBuilder();
        buf.append(ObjectUtils.toString(keyable2.getKey1(), "null"));
        if (StringUtils.isNotEmpty(separator)) {
            buf.append(separator);
        }
        buf.append(ObjectUtils.toString(keyable2.getKey2(), "null"));
        return buf.toString();
    }

    /**
     * @param keyable3 The input keyable object.
     * @return A string representation of the specified {@link Keyable1}
     *         object.
     */
    public static String toString(final Keyable3<?, ?, ?> keyable3) {
        return toString(keyable3, ":", ":");
    }

    /**
     * Concatenate the string representations of the keyable object's two keys,
     * using <tt>separator</tt> in between.
     *
     * @param keyable3 The input keyable object.
     * @param separator1 The separator between the string representations of the
     * first and second keys.
     * @param separator2 The separator between the string representations of the
     * second and third keys.
     * @return A string representation of the specified {@link Keyable3}
     *         object.
     */
    public static String toString(final Keyable3<?, ?, ?> keyable3,
            final String separator1, final String separator2) {
        final StringBuilder buf = new StringBuilder();
        buf.append(ObjectUtils.toString(keyable3.getKey1(), "null"));
        if (StringUtils.isNotEmpty(separator1)) {
            buf.append(separator1);
        }
        buf.append(ObjectUtils.toString(keyable3.getKey2(), "null"));
        if (StringUtils.isNotEmpty(separator2)) {
            buf.append(separator2);
        }
        buf.append(ObjectUtils.toString(keyable3.getKey3(), "null"));
        return buf.toString();
    }

    /**
     * Compares the two specified keyable objects.
     *
     * @return <tt>true</tt> if both objects are equals, <tt>false</tt>
     *         otherwise.
     */
    public static <T1 extends Comparable<? super T1> & Serializable> boolean equals(
            final Keyable1<T1> k1, final Keyable1<T1> k2) {
        if (k1 == null) {
            return (k2 == null);
        }
        if (k2 == null) {
            return false;
        }
        return ObjectUtils.compare(k1.getKey1(), k2.getKey1()) == 0;
    }

    /**
     * Generate a hash code for the specified keyable object.
     *
     * @return A hash code for <tt>k1</tt>.
     */
    public static <T1 extends Comparable<? super T1> & Serializable> int hashCode(
            final Keyable1<T1> k1) {
        return (k1 != null) && (k1.getKey1() != null) ?
                k1.getKey1().hashCode() : 0;
    }

}
