package com.onaro.sanscreen.acquisition.util.common;

import org.apache.log4j.Logger;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * Logger that automatically prepends any number of externally provided
 * prefixes to all log messages.
 */
public class PrefixLogger {

    private final Logger logger;

    private static final ThreadLocal<Deque<String>> prefixes =
            new ThreadLocal<Deque<String>>() {
        @Override
        protected Deque<String> initialValue() {
            return new LinkedList<String>();
        }
    };

    private PrefixLogger(final Logger logger) {
        this.logger = logger;
    }

    public static PrefixLogger getLogger(final Class<?> clazz) {
        return new PrefixLogger(Logger.getLogger(clazz));
    }

    protected static Deque<String> getPrefixes() {
        return prefixes.get();
    }

    /** Push the specified prefix to the top of the prefix stack. */
    public static void pushPrefix(final String prefix) {
        getPrefixes().push(prefix);
    }

    /** Pop the most recent prefix from the prefix stack. */
    public static String popPrefix() {
        return getPrefixes().pop();
    }

    public boolean isDebugEnabled() {
        return logger.isDebugEnabled();
    }

    public boolean isInfoEnabled() {
        return logger.isInfoEnabled();
    }

    public void debug(final String message) {
        logger.debug(createLogMessage(message));
    }

    public void debug(final String message, final Throwable t) {
        logger.debug(createLogMessage(message), t);
    }

    public void info(final String message) {
        logger.info(createLogMessage(message));
    }

    public void warn(final String message) {
        logger.warn(createLogMessage(message));
    }

    public void warn(final String message, final Throwable t) {
        logger.warn(createLogMessage(message), t);
    }

    public void error(final String message) {
        logger.error(createLogMessage(message));
    }

    public void error(final String message, final Throwable t) {
        logger.error(createLogMessage(message), t);
    }

    /** Create a log message with the appropriate prefix. */
    private String createLogMessage(final String message) {
        final StringBuilder buffer = new StringBuilder();
        String separator = "";
        for (Iterator<String> i = getPrefixes().descendingIterator(); i.hasNext(); ) {
            buffer.append(separator);
            buffer.append(i.next());
            separator = " - ";
        }
        buffer.append(separator);
        buffer.append(message);
        return buffer.toString();
    }
}
