package com.onaro.sanscreen.acquisition.util.nameserver;

/**
 * Class to parse HBA data from a Nameserver entry.   When an HBA logs into the
 * fabric, HBA details are registered with a Fabric Name Server.   The details
 * are represented by proprietary strings.   The logic to extract HBA details
 * from name server entries for various manufacturer's HBAs is centralized here. 
 */
public class NameserverHBAParser {
    private final static String EMULEX_VENDOR = "Emulex";
    private final static String BROCADE_VENDOR = "Brocade";
    private final static String ATTO_VENDOR = "ATTO";
    private final static String CISCO_VENDOR = "Cisco";
    private final static String LSI_VENDOR = "LSI";
    private final static String XSIGO_VENDOR = "Xsigo";
    
    /**
     * Extract HBA details from a name server HBA entry
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return an HBAData instance containing fields extracted from the name server entry.
     * Any and all fields may be null.
     */
    public static HBAData parseHBAData(final String nameserverHBAEntry) {
        HBAData hbaData = parseQLogicHBAData(nameserverHBAEntry);
        if (hbaData == null) {
            hbaData = parseEmulexHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseIbmHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseNetAppHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parse3ParHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseBrocadeHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseAttoHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseCiscoHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseLSIHBAData(nameserverHBAEntry);
        }
        if (hbaData == null) {
            hbaData = parseXsigoHBAData(nameserverHBAEntry);
        }
        return hbaData;
    }

    private static HBAData parseXsigoHBAData(final String nameserverHBAEntry) {
        if (!nameserverHBAEntry.startsWith(XSIGO_VENDOR)) {
            return null;
        }
        return new HBAData(XSIGO_VENDOR, null, null, null);
    }

    private static HBAData parseAttoHBAData(final String nameserverHBAEntry) {
        if (!nameserverHBAEntry.startsWith(ATTO_VENDOR)) {
            return null;
        }
        final String[] fields = nameserverHBAEntry.split(",");
        if (fields.length < 2) {
            return null;
        }
        final String[] models = fields[0].trim().split("\\s+");
        if (models.length < 2) {
            return null;
        }
        final String[] drivers = fields[1].trim().split("\\s+");
        if (models.length < 2) {
            return null;
        }
        return new HBAData(ATTO_VENDOR, models[1], null, drivers[1]);
    }

    /**
     * Parse Brocade-1741 HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry. Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseBrocadeHBAData(final String nameserverHBAEntry) {
        if (!nameserverHBAEntry.startsWith(BROCADE_VENDOR)) {
            return null;
        }
        
        final String[] fields = nameserverHBAEntry.split("\\|");
        if (fields.length < 2) {
            return null;
        }
        final String[] models = fields[0].trim().split("-");
        if (models.length < 2) {
            return null;
        }
        return new HBAData(BROCADE_VENDOR, models[1], null, fields[1].trim());
    }


    /**
     * Parse QLogic HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry.  Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseQLogicHBAData(final String nameserverHBAEntry) {

        // Figure out the vendor.  Note we have found a re-branded HPXXX device, that is
        // clearly a Qlogic chipset (same parsing format).   In this case, we put HP as the vendor.
        // See 20210 in devtrack for details
        final HBAData hbaData;
        if (nameserverHBAEntry.startsWith("QLA") || nameserverHBAEntry.startsWith("QLE") ||
            nameserverHBAEntry.startsWith("QMH") || nameserverHBAEntry.startsWith("HP") ||
            nameserverHBAEntry.startsWith("QMI") || nameserverHBAEntry.startsWith("QMC") ||
            nameserverHBAEntry.startsWith("QCP") || nameserverHBAEntry.startsWith("QME") ||
            nameserverHBAEntry.startsWith("QEM") ||
            nameserverHBAEntry.startsWith("FCI/O-CARD2Gb/s") || nameserverHBAEntry.startsWith("DELL2362M")) {

            // The format is like: "QLA FW:xxxx DVR:yyyy "
            // DOES NOT HANDLE: "HP StorageWorks MSA1000"
            final String[] parts = nameserverHBAEntry.split("\\s+");
            final String vendor = nameserverHBAEntry.startsWith("HP")
                    ? "HP"
                    : "QLogic";
            String model = null;
            String firmware = null;
            String driver = null; 

            for (final String part : parts) {
                if (part.startsWith("QLA") || part.startsWith("QLE")
                    || part.startsWith("QMH") || part.startsWith("HP")
                    || part.startsWith("QMI") || part.startsWith("QMC")
                    || part.startsWith("QCP") || part.startsWith("QME")
                    || part.startsWith("QEM")) {
                    final int prefix = part.startsWith("HP") ? 2 : 3;
                    model = part.substring(prefix).trim();
                } else if (part.startsWith("FCI/O-CARD2Gb/s")) {
                    model = "2312";
                } else if (part.startsWith("DELL2362M")) {
                    model = "2362M";
                } else if (part.startsWith("FW:")) {
                    firmware = part.substring(3).trim();
                } else if (part.startsWith("DVR:")) {
                    driver = part.substring(4).trim();
                }
            }
            hbaData = new HBAData(vendor, model, firmware, driver);

        } else {
            hbaData = null;
        }
        return hbaData;
    }


    /**
     * Parse Emulex HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry.  Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseEmulexHBAData(final String nameserverHBAEntry) {
        if (!nameserverHBAEntry.startsWith(EMULEX_VENDOR)) {
            return null;
        }

        // The format is like: "Emulex LP850 FV3.82A1  DV5-4.82A4 "
        // The format is like: "Emulex FV3.82A1  DV5-4.82A4 "
        final String[] parts = nameserverHBAEntry.split("[ ]+");
        String model = null;
        String firmware = null;
        String driver = null;
        String host = null;

        for (int index = 1; index < parts.length; index++) {
            final String part = parts[index];
            if (model == null && (part.startsWith("L") || part.startsWith("A"))) {
                model = part.trim();
            } else if (firmware == null && part.startsWith("FV")) {
                firmware = part.trim();
            } else if (part.startsWith("DV") || part.startsWith("Dv")) {
                // many driver formats:
                // Emulex LP9002 FV3.92A2  DV FERDINAND v5-2.23a6
                // Emulex LP9002 FV3.90A7 Dv5-2.41a1 GRINGOTS
                // Emulex LP9802 FV1.00A2 DV6.00g
                // Emulex 111-00308 FV2.72A2  DV5-2.10A7 IBMX336-RTP02
                // Emulex LP8000 FV3.81A1 Dv5-2.30a2 GVWAPSANAP02

                // 24165 - misreporting driver version for emulex hba / add brocade-1741 hba support
                // Either 2 or 3
                final int numberOfRestParts = parts.length - index;
                if (numberOfRestParts == 2) {
                    driver = part.trim();
                    host = parts[index + 1].trim();
                } else {
                    driver = parts[parts.length - 1];
                    if (index < parts.length - 1) {
                        host = parts[index + 1].trim();
                    }
                }
                break;
            }
        }

        final HBAData hbaData = new HBAData(EMULEX_VENDOR, model, firmware, driver);
        if (host != null && host.length() > 0) {
            hbaData.setHost(host);
        }
        return hbaData;
    }


    /**
     * Parse IBM HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry.  Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseIbmHBAData(final String nameserverHBAEntry) {

        final HBAData hbaData;
        if (nameserverHBAEntry.startsWith("IBM")) {
            // "IBM FW:v3.03.16 DVR:v9.1.0.15 (w32)"
            // The format is like: "IBM FW:v3.03.16 DVR:v9.1.0.15 (w32)"
            final String[] parts = nameserverHBAEntry.split("[ ]+");
            final String vendor = "IBM";
            final String model = null;
            String firmware = null;
            String driver = null;

            for (final String part : parts) {
                if (part.startsWith("IBM")) {
                    // No model
                }
                else if (part.startsWith("FW:")) {
                    firmware = part.substring(3).trim();
                }
                else if (part.startsWith("DVR:")) {
                    driver = part.substring(4).trim();
                }
            }
            hbaData = new HBAData(vendor, model, firmware, driver);
        } else {
            hbaData = null;
        }
        return hbaData;
    }


    /**
     * Parse NetApp HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry.  Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseNetAppHBAData(final String nameserverHBAEntry) {

        final HBAData hbaData;
        if (nameserverHBAEntry.contains("NetApp")) {
            // [NetApp FAS3070 (fas3070-rtp02/fas3070-rtp01)]
            final String[] parts = nameserverHBAEntry.split("[ ]+");
            final String vendor = "NetApp";
            String model = null;
            final String firmware = null;
            final String driver = null;

            if (parts.length > 3 && parts[1].startsWith("FAS"))
                model = parts[1].trim();
            
            hbaData = new HBAData(vendor, model, firmware, driver);
        } else {
            hbaData = null;
        }
        return hbaData;
    }

    private static HBAData parse3ParHBAData(final String nameserverHBAEntry) {
        final HBAData hbaData;
        if (nameserverHBAEntry.startsWith("3PAR")) {
            hbaData = new HBAData("3PAR", null, null, null);   
        } else {
            hbaData = null;
        }
        return hbaData;
    }

    /**
     * Parse Cisco HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry.  Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseCiscoHBAData(final String nameserverHBAEntry) {
        //Oddly, in a few cases, the Cisco name entry is identified not by its vendor name
        //Sometimes only the model name is the only means of identification (fnic for example)
        if (!nameserverHBAEntry.startsWith("fnic"))
            return null;

        //Example: fnic v1.5.0.3-1vmw over fnic2
        //        (Model)  (Driver)
        final String[] parts = nameserverHBAEntry.split("\\s+");
        String model = null;
        String driver = null;

        //The first part is the model
        if (parts.length > 0) {
            model = parts[0].trim();

            //The second part is the driver version
            if (parts.length > 1) {
                driver = parts[1].trim();
            }
        }

        //Return the HBAData object with no firmware.
        return new HBAData(CISCO_VENDOR, model, null, driver);
    }

    /**
     * Parse LSI HBA info
     * @param nameserverHBAEntry the String acquired from the name server describing the HBA
     * @return null if the given nameserverHBAEntry does not contain a known pattern, otherwise
     * an HBAData instance containing fields extracted from the name server entry.  Only
     * the vendor field is guaranteed not to be null, any other field may be null.
     */
    private static HBAData parseLSIHBAData(final String nameserverHBAEntry) {
        if (!nameserverHBAEntry.startsWith(LSI_VENDOR))
            return null;

        //Example: LSI7204EP-LC A.2 L3-25065-00B FW:01.03.24 Port 1
        //        (Model)                           (FW)
        final String[] parts = nameserverHBAEntry.split("\\s+");
        String model = null;
        String firmware = null;

        for (final String part : parts) {
            if (part.startsWith(LSI_VENDOR)) {
                model = part.trim();
            }
            else if (part.startsWith("FW:")) {
                firmware = part.substring(3).trim();
            }
        }

        //Return the HBAData object with no driver.
        return new HBAData(LSI_VENDOR, model, firmware, null);
    }


    /**
     * Returns null if the parameter is null or empty, otherwise returns the unmodified string 
     * @param s the string to reduce
     * @return null if the parameter is null or empty, otherwise returns the unmodified string
     */
    private static String reduceString(final String s) {
        return (s != null && s.isEmpty())
                ? null
                : s;
    }



    /**
     * Simple read-only container class for HBA details
     */
    public static final class HBAData {

        // the HBA vendor
        public final String vendor;

        // the HBA model
        public final String model;

        // the HBA firmware version
        public final String firmware;

        // the HBA driver version
        public final String driver;

        // the host name if available
        private String host;

        /**
         * Create a new HBA Data instance containing known details of the HBA
         * @param vendor the HBA vendor
         * @param model the HBA model
         * @param firmware the HBA firmware version
         * @param driver the HBA driver version
         */
        private HBAData(final String vendor, final String model, final String firmware, final String driver) {
            this.vendor = reduceString(vendor);
            this.model = reduceString(model);
            this.firmware = reduceString(firmware);
            this.driver = reduceString(driver);
        }

        public String getHost() {
            return host;
        }

        public void setHost(final String host) {
            this.host = host;
        }
    }
    
}
