package com.onaro.sanscreen.acquisition.util.resolver;

import java.util.LinkedList;

/**
 * A utility class that helps to generate port WWNs based on the switch WWN and port number.
 *
 * <p/>The implementation mimics the device specific algorithms to come up with the same port WWNs for switch ports.
 * <p/>This utility is especially important when trying to convert switch port zoning (which is specified using FCID) to
 * port WWNs (which is what SANscreen requires). It also helps to describe the remote port WWNs when trying to describe
 * ISLs between seperately managed switches or switches in interop.
 */
public class PortResolver {
    private static LinkedList<VendorPortGenerator> portGenerators = new LinkedList<VendorPortGenerator>();

    private VendorPortGenerator vpg = null;
    protected String chassisWWN;

    static {
        portGenerators.add(new BrocadePortGenerator());
        portGenerators.add(new McDataPortGenerator());
        portGenerators.add(new CiscoPortGenerator());
        portGenerators.add(new QLogicPortGenerator()) ;
        // TODO: Write generator for QLogic!!!
    }

    /**
     * Construct a port resolver object with a specific chassis WWN
     *
     * @param chassisWWN
     */
    public PortResolver(final String chassisWWN) {
        this.chassisWWN = chassisWWN;

        for (final VendorPortGenerator vendorPortGenerator : portGenerators) {
            if (vendorPortGenerator.checkVendor(chassisWWN)) {
                this.vpg = vendorPortGenerator;
                break;
            }
        }
        // Unknown vendors will not be resolved
    }

    public String getVendorName() {
        if ( vpg == null )
            return "";

        return vpg.getVendorName();
    }

    /**
     * Generate a port WWN based on the port number (and the chassis WWN).
     *
     * @param portNumber
     *
     * @return The generated port WWN or null if could not generate.
     *
     * @throws PortResolverException
     */
    public String resolve(final int portNumber) throws PortResolverException {
        if (vpg == null)
            return null;

        return vpg.getPortWWN(chassisWWN, 0, portNumber);
    }

    public String resolve(final int bladeNumber, final int portNumber) throws PortResolverException {
        if (vpg == null)
            return null;

        return vpg.getPortWWN(chassisWWN, bladeNumber, portNumber);
    }
}
