package com.onaro.sanscreen.aquisition.recordingutils;


import com.onaro.sanscreen.acquisition.framework.datasource.io.DSInputStream;
import com.onaro.sanscreen.acquisition.framework.datasource.io.DSSessionFileReader;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.onaro.sanscreen.acquisition.netappfoundation.datasource.NetAppFoundationPackage;
import com.onaro.sanscreen.acquisition.netappstorageperformance.datasource.NetAppStoragePerformancePackage;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

/**
 * Analyze MEGA netappfoundation recordings for iterator response times,
 * generate a WIKI table of results
 */
public class ResponseTimesAnalyzer {

    // ./1378823091917/recording_netappfoundation_marketing_cluster.zip_91_2013_09_10_10_24_52_015_netappfoundation.dat
    private static final Pattern FOUNDATION_RECORDING_PATTERN = Pattern.compile("recording_netappfoundation_.*_netappfoundation.dat");

    // ./1378823091917/recording_netappfoundation_marketing_cluster.zip_91_2013_09_10_10_24_52_015_netappfoundation.dat
    private static final Pattern PERFORMANCE_RECORDING_PATTERN = Pattern.compile("recording_netappstorageperformance_.*_netappstorageperformance.dat");

    private static final Pattern EMBEDDED_FOUNDATION_PATTERN = Pattern.compile("netappfoundation_.*_copy.zip");

    private static enum RecordingParserState {
        NONE,
        REQUEST_START,
        REQUEST_BODY,
        REQUEST_END,
        RESPONSE_START,
        RESPONSE_BODY,
        RESPONSE_END,
        TIME_START,
        TIME_BODY,
        TIME_END,
    }

    /**
     * Map of Zapi name to top-level repeating record Element name,
     * used to count records for Zapis that don't return num-records
     */
    private static final Map<String, String> zapiToRecordEltMap = new HashMap<String, String>() {{
        put("storage-adapter-get-adapter-info", "adapter-detail-info");
        put("cf-status", "backup-io-times");
        put("perf-object-counter-list-info", "counter-info");
        put("system-get-version", "version");      // Node versions + cluster version?
        put("diagnosis-status-get", "diagnosis-status");
        put("storage-shelf-environment-list-info", "shelf-environ-channel-info");
        put("storage-shelf-list-info", "shelf-info");
        put("snapmirror-policy-get-iter", "snapmirror-policy-info");
        put("qos-policy-group-get-iter", "qos-policy-group-info");
        put("storage-adapter-get-adapter-list", "adapter-name-elem");
        put("service-processor-get", "service-processor-info");
        put("cluster-identity-get", "cluster-identity-info");
        put("vserver-peer-get-iter", "vserver-peer-info");
        put("volume-footprint-get-iter", "vserver-peer-info");
        put("sis-get-iter", "sis-status-info");
        put("interim-license-list-get", "interim-license-cluster-info");
        put("clock-get-timezone", "timezone");
        put("perf-object-get-instances", "instance-data");
        put("cluster-ha-get", "cluster-ha-info");
        put("license-v2-list-info", "license-v2-info");
    }};

    private static final String REQUEST =       "<<< REQUEST >>>";
    private static final String RESPONSE =      "<<< RESPONSE >>>";


    public static void main(final String[] args) {
        if (args.length != 1) {
            usage();
            System.exit(1);
        }
        final String targetFileName = args[0];
        final File targetFile = new File(targetFileName);
        if (!targetFile.exists()) {
            System.err.println(targetFileName + " does not exist");
            System.exit(1);
        }
        try {
            if (targetFile.isDirectory()) {
                final FilenameFilter zipFilter = new FilenameFilter() {
                    public boolean accept(final File dir, final String name) {
                        if (!dir.equals(targetFile)) {
                            return false;
                        }
                        if (!name.endsWith(".zip")) {
                            return false;
                        }
                        return true;
                    }

                };
                final String[] recordingFiles = targetFile.list(zipFilter);
                for (final String recordingFileName : recordingFiles) {
                    final File recordingFile = new File(targetFile, recordingFileName);
                    analyzeRecording(recordingFile);
                }
            } else {
                analyzeRecording(targetFile);
            }
        } catch (Exception e) {
            System.out.println("JSD caught: " + e);
        }
    }

    private static void usage() {
        System.out.println("Usage:");
        System.out.println("\tResponseTimesAnalyzer <netappfoundation_recording.zip>");
        System.out.println("\tResponseTimesAnalyzer <directory_of_recordings>");
    }

    private static void analyzeRecording(final File recordingFile) throws IOException {
        final File outputFile = new File("mega_zapi.wiki");
        final FileWriter outputFileWriter = new FileWriter(outputFile);
        final BufferedWriter bufferedOutputWriter = new BufferedWriter(outputFileWriter);
        bufferedOutputWriter.write("= NetApp MEGA ZAPI Use =");
        bufferedOutputWriter.newLine();

        final File parentDir = recordingFile.getParentFile();
        final String fileName = recordingFile.getName();
        System.out.println("Analyzing " + recordingFile.getName());
        DSSessionFileReader fileReader = null;
        try {
            Cluster cluster = null;
            fileReader = new DSSessionFileReader(parentDir, fileName);
            final DSInputStream configStream = fileReader.getRootSample().getStream("configuration");
            if (configStream != null) {
                cluster = processConfiguration(configStream);
                System.out.println("Cluster: " + cluster.clusterName + " " + cluster.clusterIP);
            }

            final Set<DSSessionFileReader.Sample> samples = fileReader.getSamples();
            for (final DSSessionFileReader.Sample sample : samples) {
                final String timestampAsString = sample.getName();
                long timestampAsLong = -1;
                try {
                    timestampAsLong = Long.parseLong(timestampAsString);
                } catch (NumberFormatException nfe) {
                    // Ignore
                }
                final Date sampleTimestamp = new Date(timestampAsLong);

                final List<DSInputStream> inputStreams = sample.getStreams();
                for (final DSInputStream inputStream : inputStreams) {
                    final String streamName = inputStream.getFileName();

                    final Matcher performanceRecordingMatcher = PERFORMANCE_RECORDING_PATTERN.matcher(streamName);
                    if (performanceRecordingMatcher.lookingAt()) {
                        processZAPIRecordingFile(NetAppStoragePerformancePackage.PACKAGE, inputStream, cluster, sampleTimestamp, bufferedOutputWriter);
                        continue;
                    }

                    final Matcher foundationRecordingMatcher = FOUNDATION_RECORDING_PATTERN.matcher(streamName);
                    if (foundationRecordingMatcher.lookingAt()) {
                        processZAPIRecordingFile(NetAppFoundationPackage.PACKAGE, inputStream, cluster, sampleTimestamp, bufferedOutputWriter);
                        continue;
                    }

                    final Matcher embeddedFoundationMatcher = EMBEDDED_FOUNDATION_PATTERN.matcher(streamName);
                    if (embeddedFoundationMatcher.lookingAt()) {
                        processEmbeddedFoundation(inputStream, sample, cluster, sampleTimestamp, bufferedOutputWriter);
                        continue;
                    }
                }
            }
        } catch (Exception e) {
            System.out.println("JSD caught:  " + e);
            e.printStackTrace();
        } finally {
            if (fileReader != null) {
                try {
                    fileReader.close();
                } catch (Exception ignore) {
                    // Ignore
                }
            }
            if (bufferedOutputWriter != null) {
                try {
                    bufferedOutputWriter.close();
                } catch (Exception ignore) {
                    // Ignore
                }
            }
        }
    }

    private static Cluster processConfiguration(final DSInputStream configStream) {
        final Properties configProperties = new Properties();
        try {
            configProperties.loadFromXML(configStream);
            final String clusterName = configProperties.getProperty("system.datasource.name");
            final String clusterMgmtIP = configProperties.getProperty("ip");
            return new Cluster(clusterName, clusterMgmtIP);
        } catch (Exception e) {
            System.out.println("JSD caught " + e);
            return null;
        }
    }


    private static void processEmbeddedFoundation(final DSInputStream foundationInputStream, final DSSessionFileReader.Sample performanceSample,
                                                  final Cluster cluster, final Date sampleTimestamp, final BufferedWriter bufferedOutputWriter) {
        final File outputDir = new File(".");
        final String foundationRecordingFile = performanceSample.getDescriptionProperty("foundation", "recording");
        if (foundationRecordingFile == null) {
            return;
        }
        try {
            //make a copy of this outside the zip file for use.
            final File outputZip = new File(outputDir, foundationRecordingFile);
            if (!outputZip.exists()) {
                BufferedOutputStream out = null;
                try {
                    out = new BufferedOutputStream(new FileOutputStream(outputZip));

                    final int bufferSize = 8 * 1024;
                    final byte[] buffer = new byte[bufferSize];
                    int count = foundationInputStream.read(buffer);

                    while (count > 0) {
                        out.write(buffer, 0, count);
                        count = foundationInputStream.read(buffer);
                    }
                    out.flush();

                } catch (IOException ioe) {
                    // ignore
                } finally {
                    if (out != null) {
                        try {
                            out.close();
                        } catch (IOException e) {

                            //ignore.
                        }
                    }
                }
            }

            final DSSessionFileReader foundationReader = new DSSessionFileReader(outputZip);
            for (final DSSessionFileReader.Sample sample : foundationReader.getSamples()) {
                final List<DSInputStream> inputStreams = sample.getStreams();
                for (final DSInputStream inputStream : inputStreams) {
                    final String streamName = inputStream.getFileName();
                    final Matcher foundationRecordingMatcher = FOUNDATION_RECORDING_PATTERN.matcher(streamName);
                    if (foundationRecordingMatcher.lookingAt()) {
                        processZAPIRecordingFile(NetAppFoundationPackage.PACKAGE, inputStream, cluster, sampleTimestamp, bufferedOutputWriter);
                        continue;
                    }
                }
            }
        } catch (IOException ioe) {
            System.out.println("JSD caught " + ioe);
        }
    }

    private static void processZAPIRecordingFile(final String pollType, final DSInputStream inputStream, final Cluster cluster,
                                                 final Date sampleTimestamp, final BufferedWriter bufferedOutputWriter) {
        System.out.println("\n\nAnalyzing " + inputStream.getFileName());
        System.out.println("Cluster: " + cluster.clusterName + " " + cluster.clusterIP);
        System.out.println(pollType + " poll: " + sampleTimestamp + "\n");
        final Map<String, Zapi> zapis = new HashMap<String, Zapi>();
        BufferedReader dataFileReader = null;
        try {
            dataFileReader = new BufferedReader(new InputStreamReader(inputStream, "utf-8"));
            RecordingParserState parserState = RecordingParserState.NONE;
            Zapi currentZapi = null;
            while (true) {
                final String line = dataFileReader.readLine();
                if (line == null) {
                    break;
                }
                if (line.isEmpty()) {
                    continue;
                }
                String compoundLine = line;
                while (compoundLine.endsWith("&gt;")) {
                    compoundLine += dataFileReader.readLine();
                }
                switch (parserState) {
                    case NONE:
                    case REQUEST_START:
                        if (line.startsWith(REQUEST)) {
                            parserState = RecordingParserState.REQUEST_BODY;
                        }
                        break;

                    case REQUEST_BODY:

                        currentZapi = processZAPIRequest(compoundLine, zapis);
                        parserState = RecordingParserState.REQUEST_END;
                        break;

                    case REQUEST_END:
                        parserState = RecordingParserState.RESPONSE_START;
                        break;

                    case RESPONSE_START:
                        if (line.startsWith(RESPONSE)) {
                            parserState = RecordingParserState.RESPONSE_BODY;
                        }
                        break;

                    case RESPONSE_BODY:
                        // Count top-level records received

                        final long recordCount = getRecordCount(compoundLine, currentZapi);
                        currentZapi.addReceivedRecords(recordCount);
                        parserState = RecordingParserState.RESPONSE_END;
                        break;

                    case RESPONSE_END:
                        parserState = RecordingParserState.TIME_START;
                        break;

                    case TIME_START:
                        parserState = RecordingParserState.TIME_BODY;
                        break;

                    case TIME_BODY:
                        final long invocationTime = parseLong(line);
                        currentZapi.addZapiTime(invocationTime);
                        parserState = RecordingParserState.TIME_END;
                        break;

                    case TIME_END:
                        parserState = RecordingParserState.REQUEST_START;
                        break;

                    default:
                        // Just consume the empty line and look for the next state
                        break;
                }
            }

            final SortedSet<Zapi> sortedZapis = new TreeSet<Zapi>(zapis.values());
            for (final Zapi zapi : sortedZapis) {
                System.out.println(zapi);
            }

            // JSD XXX
            writePollToTable(cluster, sampleTimestamp, pollType, bufferedOutputWriter, zapis);
        }
        catch (IOException ioe) {
            System.out.println("JSD caught " + ioe);
        }
        finally {
            if (dataFileReader != null) {
                try {
                    dataFileReader.close();
                } catch(IOException ignore) {}
            }
        }
    }


    private static Zapi processZAPIRequest(final String zapiRequest, final Map<String, Zapi> zapis) {
        final SAXBuilder builder = new SAXBuilder();
        builder.setExpandEntities(false);

        try {
            final Document responseDoc = builder.build(new StringReader(zapiRequest));
            final Element zapiRequestElt = responseDoc.getRootElement();
            final String zapiName = zapiRequestElt.getName();
            String fqZapiName = zapiName;
            final List<String> arguments = new ArrayList<String>();

            String objectName = null;
            long instanceCount = 0;
            long desiredAttributes = 0;
            boolean hasQuery = false;
            int maxRecords = 0;
            if ("perf-object-get-instances".equals(zapiName)) {
                maxRecords = 500; // or 100 for < 8.2
            }
            @SuppressWarnings("unchecked")
            final List<Element> childElts = zapiRequestElt.getChildren();
            for (final Element childElt : childElts) {
                final String childEltName = childElt.getName();
                if ("desired-attributes".equals(childEltName)) {
                    desiredAttributes = countLeafNodes(childElt);
                    continue;
                }
                if ("query".equals(childEltName)) {
                    hasQuery = true;
                    continue;
                }
                if ("objectname".equals(childEltName)) {
                    arguments.add(childEltName);
                    objectName = childElt.getValue();
                    fqZapiName += ' ' + objectName;
                    // System.out.println("\t" + childEltName + ": " + objectName);

                } else if ("instance-uuids".equals(childEltName)) {
                    instanceCount = countChildElements(childElt, "instance-uuid");

                } else if ("counters".equals(childEltName)) {
                    desiredAttributes = countChildElements(childElt, "counter");
                }  else if ("max-records".equals(childEltName)) {
                    final String maxRecordsAsString = childElt.getValue();
                    try {
                        maxRecords = Integer.parseInt(maxRecordsAsString);
                    } catch (NumberFormatException nfe) {
                        // Ignore
                    }
                } else {
                    arguments.add(childEltName);
                }
            }
            if (objectName == null) {
                fqZapiName = zapiName;
            } else {
                fqZapiName = zapiName + " " + objectName;
            }
            if (!zapis.containsKey(fqZapiName)) {
                zapis.put(fqZapiName, new Zapi(fqZapiName, zapiName));
            }
            final Zapi zapi = zapis.get(fqZapiName);
            zapi.setObjectName(objectName);
            zapi.incrInvocations();
            zapi.setQuery(hasQuery);
            zapi.setDesiredAttributes(desiredAttributes);
            zapi.addRequestedRecords(instanceCount);
            zapi.setMaxRecords(maxRecords);
            for (final String argument : arguments) {
                zapi.addArgument(argument);
            }
            return zapi;
        }
        catch (IOException ioe) {
            System.err.println(zapiRequest);
            System.err.println(ioe);
        }
        catch (JDOMException jdome) {
            System.err.println(zapiRequest);
            System.err.println(jdome);
        }
        return null;
    }

    private static long countLeafNodes(final Element rootElt) {
        long leafNodeCount = 0;
        @SuppressWarnings("unchecked")
        final List<Element> childElts = rootElt.getChildren();
        if (childElts.size() == 0) {
            // This is a leaf element
            leafNodeCount++;
        } else {
            for (final Element childElt : childElts) {
                leafNodeCount += countLeafNodes(childElt);
            }
        }
        return leafNodeCount;
    }

    private static long getRecordCount(final String responseLine, final Zapi zapi) {
        final SAXBuilder builder = new SAXBuilder();
        builder.setExpandEntities(false);
        try {
            final Document responseDoc = builder.build(new StringReader(responseLine));
            final Element zapiResponseElt = responseDoc.getRootElement();

            final Element numRecordsElt = zapiResponseElt.getChild("num-records");
            if (numRecordsElt != null) {
                final String numRecordsAsString = numRecordsElt.getValue();
                try {
                    return Long.parseLong(numRecordsAsString);
                } catch (NumberFormatException nfe) {
                    return -1;
                }

            } else if (zapiToRecordEltMap.containsKey(zapi.shortZapiName)) {
                final String repeatingEltName = zapiToRecordEltMap.get(zapi.shortZapiName);
                return countChildElements(zapiResponseElt, repeatingEltName);
            } else {
                // Unable to find API:
                @SuppressWarnings("unchecked")
                final List<Element> childElts = zapiResponseElt.getChildren();

                if (childElts.size() == 0) {
                    return 0;
                } else {
                    System.out.println(zapi.fqZapiName + " can't find num-records under " + zapiResponseElt.getName()
                        + ", have: " + childElts.size());
                    for (final Element childElt : childElts) {
                        System.out.println("\t\t" + childElt.getName());
                    }
                    System.out.println(responseLine + "\n\n");
                    return -1;
                }
            }
        }
        catch (IOException ioe) {
            System.err.println(ioe);
        }
        catch (JDOMException jdome) {
            System.err.println(jdome);
        }
        return -1;
    }


    private static long countChildElements(final Element rootElement, final String repeatingEltName) {
        long childCount = 0;
        @SuppressWarnings("unchecked")
        final List<Element> childElts = rootElement.getChildren();
        for (final Element childElt : childElts) {
            if (repeatingEltName.equals(childElt.getName())) {
                childCount++;
            } else {
                childCount += countChildElements(childElt, repeatingEltName);
            }
        }
        return childCount;
    }


    private static long parseLong(final String longAsString) {
        try {
            return Long.parseLong(longAsString);
        } catch(NumberFormatException nfe) {
            System.out.println("can't parse " + longAsString);
            return -1L;
        }
    }

    private static void writePollToTable(final Cluster cluster, final Date sampleTimestamp, final String pollType, final BufferedWriter bufferedOutputWriter,
                                final Map<String, Zapi> zapis) throws IOException {
        writeln(bufferedOutputWriter, String.format("== NetApp MEGA %s poll: %s %s %s ==",
                pollType, cluster.clusterName, cluster.clusterIP, sampleTimestamp));
        writeln(bufferedOutputWriter, "{| class=\"prettytable sortable \" border=\"1\"");
        writeln(bufferedOutputWriter, "|-");
        writeln(bufferedOutputWriter, "! ZAPI Name");
        writeln(bufferedOutputWriter, "! Total Time (ms)");
        writeln(bufferedOutputWriter, "! Invocations");
        writeln(bufferedOutputWriter, "! Requested Records");
        writeln(bufferedOutputWriter, "! Total Recv Records");
        writeln(bufferedOutputWriter, "! Filter Query");
        writeln(bufferedOutputWriter, "! Max Records");
        writeln(bufferedOutputWriter, "! Desired Attributes");
        writeln(bufferedOutputWriter, "! Other Arguments");
        for (final Zapi zapi : zapis.values()) {
            writeln(bufferedOutputWriter, "|-");
            writeln(bufferedOutputWriter, "| " + zapi.fqZapiName);
            writeln(bufferedOutputWriter, "| " + zapi.totalZapiTime);
            writeln(bufferedOutputWriter, "| " + zapi.totalInvocations);
            writeln(bufferedOutputWriter, "| " + (zapi.totalRequestedRecords <= 0 ? "All" : String.valueOf(zapi.totalRequestedRecords)));
            writeln(bufferedOutputWriter, "| " + (zapi.totalReceivedRecords < 0 ? "Unknown" : String.valueOf(zapi.totalReceivedRecords)));
            writeln(bufferedOutputWriter, "| " + zapi.query);
            writeln(bufferedOutputWriter, "| " + (zapi.maxRecords <= 0 ? "All" : String.valueOf(zapi.maxRecords)));
            writeln(bufferedOutputWriter, "| " + (zapi.desiredAttributes <= 0 ? "All" : String.valueOf(zapi.desiredAttributes)));
            writeln(bufferedOutputWriter, "| " + zapi.getArgumentsAsCSV());
        }
        writeln(bufferedOutputWriter, "|}");
    }

    private static void writeln(final BufferedWriter writer, final String val) throws IOException {
        writer.write(val);
        writer.newLine();
    }


    private static final class Zapi implements Comparable<Zapi> {

        private final String fqZapiName;
        private final String shortZapiName;
        private String objectName;
        private long totalZapiTime = 0;
        private long totalInvocations = 0;
        private long totalRequestedRecords = 0;
        private long totalReceivedRecords = 0;
        private int maxRecords = -1;
        private long desiredAttributes = -1L;
        private boolean query = false;
        private List<String> arguments = new ArrayList<String>();

        private Zapi(final String fqZapiName, final String shortZapiName) {
            this.fqZapiName = fqZapiName;
            this.shortZapiName = shortZapiName;
        }

        private void setObjectName(final String objectName) {
            this.objectName = objectName;
        }

        private void addZapiTime(final long zapiTime) {
            totalZapiTime += zapiTime;
        }

        private void incrInvocations() {
            totalInvocations++;
        }

        private void addRequestedRecords(final long records) {
            totalRequestedRecords += records;
        }

        private void addReceivedRecords(final long records) {
            this.totalReceivedRecords += records;
        }

        private void setMaxRecords(final int maxRecords) {
            this.maxRecords = maxRecords;
        }

        private void setDesiredAttributes(final long desiredAttributes) {
            this.desiredAttributes = desiredAttributes;
        }

        private void setQuery(final boolean query) {
            this.query = query;
        }

        private void addArgument(final String argument) {
            if (this.arguments.contains(argument)) {
                return;
            }
            this.arguments.add(argument);
        }

        private String getArgumentsAsCSV() {
            if (this.arguments.isEmpty()) {
                return "None";
            } else {
                String argumentsCSV = "";
                for (final String argument : this.arguments) {
                    if (argumentsCSV.isEmpty()) {
                        argumentsCSV = argument;
                    } else {
                        argumentsCSV += ", " + argument;
                    }
                }
                return argumentsCSV;
            }
        }

        @Override
        public boolean equals(final Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            final Zapi zapi = (Zapi) o;

            if (!fqZapiName.equals(zapi.fqZapiName)) return false;

            return true;
        }

        @Override
        public int hashCode() {
            return fqZapiName.hashCode();
        }

        @Override
        public int compareTo(final Zapi o) {
            final long diff = o.totalZapiTime - this.totalZapiTime;
            return Long.signum(diff);
        }

        @Override
        public String toString() {
            if ("perf-object-instance-list-info-iter".equals(shortZapiName)) {
                return shortZapiName + " " + objectName
                        + "\n\t\t   total time (ms): " + totalZapiTime
                        + "\n\t\t       invocations: " + totalInvocations
                        + "\n\t\ttotal recv records: " + (totalReceivedRecords < 0 ? "Unknown" : String.valueOf(totalReceivedRecords))
                        + "\n\t\t      filter query: " + query
                        + "\n\t\t       max records: " + maxRecords
                        + "\n\t\tdesired attributes: " + (-1L == desiredAttributes ? "All" : String.valueOf(desiredAttributes))
                        + "\n\t\t   other arguments: " + getArgumentsAsCSV()
                        + "\n";
            } else if ("perf-object-get-instances".equals(shortZapiName)) {
                return fqZapiName
                        + "\n\t\t   total time (ms): " + totalZapiTime
                        + "\n\t\t       invocations: " + totalInvocations
                        + "\n\t\t requested records: " + (totalRequestedRecords <= 0 ? "All" : String.valueOf(totalRequestedRecords))
                        + "\n\t\ttotal recv records: " + (totalReceivedRecords < 0 ? "Unknown" : String.valueOf(totalReceivedRecords))
                        + "\n\t\t      filter query: " + query
                        + "\n\t\t       max records: " + maxRecords
                        + "\n\t\tdesired attributes: " + (-1L == desiredAttributes ? "All" : String.valueOf(desiredAttributes))
                        + "\n\t\t   other arguments: " + getArgumentsAsCSV()
                        + "\n";
            } else {
                return fqZapiName
                        + "\n\t\t   total time (ms): " + totalZapiTime
                        + "\n\t\t       invocations: " + totalInvocations
                        + "\n\t\ttotal recv records: " + (totalReceivedRecords < 0 ? "Unknown" : String.valueOf(totalReceivedRecords))
                        + "\n\t\t      filter query: " + query
                        + "\n\t\t       max records: " + (-1 == maxRecords ? "All" : String.valueOf(maxRecords))
                        + "\n\t\tdesired attributes: " + (-1L == desiredAttributes ? "All" : String.valueOf(desiredAttributes))
                        + "\n\t\t   other arguments: " + getArgumentsAsCSV()
                        + "\n";
            }
        }
    }

    private static final class Cluster {
        private final String clusterName;
        private final String clusterIP;

        private Cluster(final String clusterName, final String clusterIP) {
            this.clusterName = clusterName;
            this.clusterIP = clusterIP;
        }
    }
}
