package com.onaro.sanscreen.server.startup;

import com.netapp.oci.netapp.startup.NetAppPlugin;
import com.netapp.oci.netapp.upgrade.NetAppUpgradeRegistry;
import com.netapp.oci.platform.common.InternalServerException;
import com.netapp.oci.platform.common.ServerSystem;
import com.netapp.oci.platform.common.param.ParamException;
import com.netapp.oci.platform.common.table.ConnectionHelper;
import com.netapp.oci.platform.common.table.DBPrefixHelper;
import com.netapp.oci.platform.common.table.TableException;
import com.netapp.oci.platform.plugin.remote.PlatformPluginRemote;
import com.netapp.oci.platform.upgrade.UpgradeRunner;
import com.netapp.oci.profiler.ProfilerService;
import com.netapp.oci.profiler.ProfilerServiceFactory;
import com.onaro.commons.util.Unit;
import com.onaro.sanscreen.acquisition.sessions.AcquisitionFacadeSessionLocal;
import com.onaro.sanscreen.server.tables.idgenerator.sessions.IdGeneratorSession;
import com.onaro.sanscreen.server.tables.param.InfoTable;
import org.jboss.ejb3.annotation.Depends;
import org.jboss.ejb3.annotation.Management;
import org.jboss.ejb3.annotation.SecurityDomain;
import org.jboss.ejb3.annotation.Service;
import org.jboss.ejb3.common.proxy.plugins.async.AsyncUtils;
import org.jboss.logging.Logger;
import org.jboss.security.RunAsIdentity;
import org.jboss.security.SecurityContextAssociation;
import javax.annotation.Resource;
import javax.ejb.EJB;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.management.MBeanServerConnection;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.sql.DataSource;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.lang.management.ManagementFactory;
import java.lang.management.RuntimeMXBean;
import java.sql.Connection;
import java.text.NumberFormat;
import java.util.*;
import java.util.regex.Pattern;

/**
 * @author Nancy Walia
 *         Date: Dec 16, 2009
 *         This MBean is used to run methods after jboss starts.
 */
@Service(objectName = "sampleJMX:service=ociServer")
@Management(StartUpService.class)
@TransactionManagement(TransactionManagementType.BEAN)
@SecurityDomain("onaro-login")
@SuppressWarnings({"UnusedDeclaration"})
@Depends({ // enforce dependency on required beans
        "jboss.j2ee:ear=acqServer.ear,jar=acqServer.jar,name=AcquisitionServerSessionBean,service=EJB3",
        "org.hornetq:module=JMS,name=\"NettyConnectionFactory\",type=ConnectionFactory"
})
public class StartUpServiceMBean implements StartUpService {

    private static final Logger logger = Logger.getLogger("com.onaro.sanscreen.server.startup");
    private static String LINE_SEP = System.getProperty("line.separator");
    private static final String DATABASE_SCHEMA_UPGRADED_KEY = "DATABASE_SCHEMA_UPGRADED";

    @Resource(name = "jdbc/OnaroDataSource", mappedName = "java:/onaro_pool", type = javax.sql.DataSource.class, authenticationType = Resource.AuthenticationType.CONTAINER)
    private DataSource dataSource;

    @EJB(name = AcquisitionFacadeSessionLocal.JNDI_NAME)
    private AcquisitionFacadeSessionLocal acquisitionLocal;

    @EJB(mappedName = PlatformPluginRemote.JNDI_NAME)
    private PlatformPluginRemote platformPluginRemote;

    @EJB
    private IdGeneratorSession idGeneratorSession; // hack to preload id generator session, cause policy local will use it behind scenes

    /**
     * This method should not throw any exceptions.
     */
    public void start() {
        try {
            // annotation RunAs did not work for some reason
            SecurityContextAssociation.getSecurityContext().setOutgoingRunAs(new RunAsIdentity("admin", "admin"));

            logger.info("Deploying Acquisition Datasource meta-data");
            acquisitionLocal.deploy(false);

            Connection connection = null;
            try {
                connection = ConnectionHelper.newConnection(dataSource);

                new DBPrefixHelper().handleDbPrefix(connection);

                logger.info("Checking OCIE database");
                try {
                    new UpgradeRunner(NetAppUpgradeRegistry.INSTANCE, connection).run();
                } catch (InternalServerException e) {
                    logger.error("Failed to check database", e);
                }

                StartUpService startUpService = AsyncUtils.mixinAsync(this);
                startUpService.collect();

                startProfilerService();

                logger.info("Aquisition initializing");
                platformPluginRemote.register(new NetAppPlugin());
            } finally {
                ConnectionHelper.close(connection);
            }

        } catch (Throwable ex) {
            logger.error("Exception during server startup", ex);
        }
    }

    @Override
    public void collect() {
        Connection connection = null;
        try {
            connection = ConnectionHelper.newConnection(dataSource);

            boolean isCollecting = Boolean.valueOf(System.getProperty("phonehome.systemInfo", "true"));
            if (isCollecting) {
                collectIpConfig(connection);
                collectSystemInfo(connection);
                collectJVMInformation(connection);
            }

            collectSANscreenMysqlInfo(connection);
        } finally {
            ConnectionHelper.close(connection);
        }
    }

    private void startProfilerService() {
        final ProfilerService profilerService = ProfilerServiceFactory.getInstance();
        if (profilerService.isEnabled()) {
            profilerService.start();
            Runtime.getRuntime().addShutdownHook(new Thread() {
                @Override
                public void run() {
                    profilerService.stop();
                }
            });
        }
    }

    private void collectSANscreenMysqlInfo(Connection connection) {
        try {
            Map<String, String> info = new HashMap<String, String>();
            String sanscreenHome = System.getProperty("onaro.home");
            if (sanscreenHome == null) {
                logger.error("onaro.home cannot be found in system properties");
                return;
            }

            // null these out in case they exist from a backup on a previous OS
            if (ServerSystem.isLinux()) {
                info.put("SANscreen Drive Space(GB) ", null);
                info.put("SANscreen Drive Free Space(GB) ", null);
                info.put("MySQL Drive Space(GB) ", null);
                info.put("MySQL Drive Free Space(GB) ", null);
            } else {
                info.put("Disk Info", null);
            }

            if (ServerSystem.isLinux()) {
                collectLinuxDriveSpaceInfo(connection);
            } else {
                getWindowsDriveSpaceInfo(info, "SANscreen", sanscreenHome);
            }

            Properties p = new Properties();
            p.load(new FileInputStream(ServerSystem.getMyCnf(sanscreenHome)));
            String baseDir = p.getProperty("basedir");
            String innodb = p.getProperty("innodb_data_file_path");
            info.put("MySQL basedir", baseDir);
            info.put("MySQL innodb_data_file_path", innodb);
            if (baseDir != null && innodb != null) {
                if (!ServerSystem.isLinux()) {
                    getWindowsDriveSpaceInfo(info, "MySQL", baseDir);
                }
                int index = innodb.indexOf(File.separatorChar);
                if (index >= 0) {
                    File dataDir = new File(baseDir + File.separator + "data" + File.separator + innodb.substring(0, index));
                    if (dataDir.exists() && dataDir.isDirectory()) {
                        StringBuilder sb = new StringBuilder();
                        File[] files = dataDir.listFiles();
                        if (files != null) {
                            for (File file : files) {
                                sb.append(file.getName()).append(":").append(convertBytesToGB(file.length())).append("(GB)  ");
                            }
                            info.put("MySQL data file size", sb.toString());
                        }
                    }
                }
            }

            saveToInfoTable(connection, info);
        } catch (Exception e) {
            logger.error("Failed to collect OnCommand Insight MySQL info", e);
        }
    }

    private void collectLinuxDriveSpaceInfo(Connection connection) {
        collectSystemInformation(connection, DEFAULT_FILTER, "Disk Info", ServerSystem.getDiskInfoCommand());
    }

    private void getWindowsDriveSpaceInfo(Map<String, String> info, String prefix, String path) {
        int index = path.indexOf(File.separatorChar);
        if (index < 0)
            return;

        String driveName = path.substring(0, index);
        File f = new File(driveName);
        info.put(prefix + " Drive Space(GB) ", convertBytesToGB(f.getTotalSpace()) + " on drive " + driveName);
        info.put(prefix + " Drive Free Space(GB) ", convertBytesToGB(f.getFreeSpace()) + " on drive " + driveName);
    }

    private String convertBytesToGB(double bytes) {
        NumberFormat nf = NumberFormat.getNumberInstance();
        nf.setMaximumFractionDigits(2);

        return nf.format(Unit.convert(bytes, Unit.BYTE, Unit.GIGABYTE));
    }

    private void collectIpConfig(Connection connection) {
        Filter filter = DEFAULT_FILTER;

        collectSystemInformation(connection, filter, "IPCONFIG", ServerSystem.getIpConfigCommand());
    }

    private void collectSystemInfo(Connection connection) {
        Filter filter = DEFAULT_FILTER;
        if (!ServerSystem.isLinux()) {
            filter = new Filter() {
                private boolean filtering = false;
                private Pattern startPattern = Pattern.compile("^Hotfix\\(s\\):.*");
                private Pattern endPattern = Pattern.compile("^NetWork Card\\(s\\):.*");

                public boolean filter(String line) {
                    if (!filtering) {
                        if (startPattern.matcher(line).matches()) {
                            filtering = true;
                        }
                    }

                    if (filtering) {
                        if (endPattern.matcher(line).matches()) {
                            filtering = false;
                        }
                    }

                    return filtering;
                }
            };
        }

        collectSystemInformation(connection, filter, "SYSTEM_INFORMATION", ServerSystem.getSystemInformationCommand());
    }

    private void collectSystemInformation(Connection connection, Filter filter, String key, String... command) {
        logger.info("Collecting information for " + key + ": " + Arrays.asList(command));
        try {
            Process proc = new ProcessBuilder(command).start();
            BufferedReader br = new BufferedReader(new InputStreamReader(proc.getInputStream()));
            StringBuilder sb = new StringBuilder();
            String line = br.readLine();
            while (line != null) {
                line = line.trim();
                if (!filter.filter(line)) {
                    if (line.length() > 0) {
                        sb.append(line).append(LINE_SEP);
                    }
                }
                line = br.readLine();
            }
            saveToInfoTable(connection, key, sb.toString());
        } catch (Throwable t) {
            logger.error("Failed to get " + key + " information", t);
        }
    }

    /**
     * this method using JMX client access JBoss JVM
     */
    private void collectJVMInformation(Connection connection) {

        try {
            String jmxPort = System.getProperty("com.sun.management.jmxremote.port", "33333");
            final JMXServiceURL target = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://localhost:" + jmxPort + "/jmxrmi");

            // Connect to target
            final JMXConnector connector = JMXConnectorFactory.connect(target);

            // Get an MBeanServerConnection on the remote VM.
            final MBeanServerConnection remote =
                    connector.getMBeanServerConnection();

            final RuntimeMXBean remoteRuntime =
                    ManagementFactory.newPlatformMXBeanProxy(
                            remote,
                            ManagementFactory.RUNTIME_MXBEAN_NAME,
                            RuntimeMXBean.class);

            List<String> params = remoteRuntime.getInputArguments();
            connector.close();
            for (String param : params) {
                if (param.startsWith("-Xmx")) {
                    saveToInfoTable(connection, "-Xmx", param.substring(4));
                }
            }


        } catch (Exception exc) {
            logger.error("Faield to get JBoss JVM information", exc);
        }
    }

    private void saveToInfoTable(Connection connection, String param, String value) throws TableException {
        InfoTable infoTable = new InfoTable(connection);
        infoTable.set(param, value);
    }

    private void saveToInfoTable(Connection connection, Map<String, String> info) throws TableException, ParamException {
        InfoTable infoTable = new InfoTable(connection);
        for (Map.Entry<String, String> entry : info.entrySet()) {
            if (entry.getValue() != null) {
                infoTable.set(entry.getKey(), entry.getValue());
            } else {
                infoTable.deletePrefix(entry.getKey());
            }
        }
    }

    private static interface Filter {
        boolean filter(String line);
    }

    private static final Filter DEFAULT_FILTER = new Filter() {
        public boolean filter(String line) {
            return false;
        }
    };
}
