package com.onaro.sanscreen.server.tables.idgenerator;

import com.onaro.sanscreen.server.interfaces.data.ServerInterfaceException;
import com.onaro.sanscreen.server.tables.idgenerator.sessions.IdGeneratorSession;
import com.onaro.sanscreen.util.JndiHelper;

/**
 * Generates object ids for object keys.
 */
public class IdGenerator {

    /**
     * Next seed, provides next portion of ids
     */
    private long idSeedNumber;

    /**
     * The size of the ids portion
     */
    public static final int QUANTUM = 1000;

    /**
     * Current value of index of the ids' portion
     */
    private long index = QUANTUM;

    /**
     * Singleton instance
     */
    private static IdGenerator instance = new IdGenerator();

    /**
     * Ids manager immplementation
     */
    private IdManager idManager;

    private IdGeneratorSession idGeneratorSession;

    /**
     * Returns skeleton instance.
     *
     * @return singleton
     */
    public static IdGenerator getInstance() {
        return instance;
    }

    /**
     * Default private constructor
     */
    private IdGenerator() {
        idManager = new IdManagerDefaultImpl();
    }

    /**
     * Returns new id, freshly baked
     *
     * @return new unqiue id
     * @throws IdGeneratorException if failed to generate new id
     */
    synchronized public Long getNextId() throws IdGeneratorException {
        if (index == QUANTUM) {
            idSeedNumber = getNextSeedNumber();
            index = 0;
        }
        long val = idSeedNumber + index;
        index++;
        return new Long(val);
    }

    /**
     * Retrieves next portion of ids from database
     *
     * @return portion of ids
     * @throws IdGeneratorException if failed to access database
     */
    private long getNextSeedNumber() throws IdGeneratorException {
        try {
            return getIdGeneratorSession().getNextSeedNumber();
        } catch (ServerInterfaceException e) {
            throw new IdGeneratorException("Failed to get next seed number", e);
        }
    }

    synchronized private IdGeneratorSession getIdGeneratorSession() throws IdGeneratorException {
        if (idGeneratorSession == null) {
            try {
                idGeneratorSession = (IdGeneratorSession) JndiHelper.getEJB(IdGeneratorSession.JNDI_NAME);
            } catch (Exception e) {
                throw new IdGeneratorException("Failed to connect to " + IdGeneratorSession.JNDI_NAME, e);
            }
        }
        return idGeneratorSession;
    }

    synchronized public void reset() throws IdGeneratorException {
        try {
            getIdGeneratorSession().reset();
            idManager.reset();
            index = QUANTUM;
        } catch (ServerInterfaceException e) {
            throw new IdGeneratorException("Failed to reset id", e);
        }
    }

    public static void main(String[] args) throws Exception {
        JndiHelper.adminLogin();
        IdGenerator idg = new IdGenerator();
        Long curId = idg.getNextId();
        while (curId.longValue() < 10000) {
            Long nextId = idg.getNextId();
            if (nextId.longValue() - curId.longValue() != 1)
                System.out.println("Error");
        }
    }
}


