package com.onaro.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;

/**
 * A cyclic container of objects keeping a limited amount of resources.
 *
 * This is often used when keeping track of history records and trying to limit the number of history records being kept.
 */
public class CyclicArray<Element> implements Collection<Element> {
    private Element[] elements;
    private int end = 0;
    private int size = 0;

    public CyclicArray(int length) {
        //noinspection unchecked
        elements = (Element[]) new Object[length];
    }

    public boolean add(final Element element) {
        if (element == null)
            throw new IllegalArgumentException("Element is null");

        // call to notify that previous element will be overwritten
        if (size >= elements.length)
            onOverwrite(elements[end]);

        // Put the element at the end of the list
        elements[end] = element;

        // Advance the element to the next free space
        end = (end + 1) % elements.length;

        if (size < elements.length)
            size++;

        return true;
    }

    /**
     * A callback method that is called everytime the array overwrites an element
     *
     * @param e Element being overwritten
     */
    protected void onOverwrite(final Element e) {
        // nothing to do. this is placeholder
    }

    /**
     * @return Current number of elements in the array
     */
    public int size() {
        return size;
    }

    /**
     * @return Maximum size the array can hold
     */
    public int maxSize() {
        return elements.length;
    }

    public boolean isEmpty() {
        return size == 0;
    }

    public boolean contains(final Object o) {
        Iterator<Element> it = iterator();
        while (it.hasNext()) {
            final Element element = it.next();
            if (element.equals(o))
                return true;
        }

        return false;
    }

    public void clear() {
        end = 0;
        size = 0;
    }

    /**
     * Empties and reallocates the array with a new size
     *
     * @param length
     */
    public void reset(int length) {
        end = 0;
        size = 0;
        //noinspection unchecked
        elements = (Element[]) new Object[length];
    }

    public Iterator<Element> iterator() {
        return new CyclicIterator();
    }

    public Object[] toArray() {
        final Object[] a = new Object[size()];
        return toArray(a);
    }

    public <T>T[] toArray(final T[] a) {
        Iterator it = iterator();

        int i = 0;
        while (it.hasNext()) {
            //noinspection unchecked
            a[i++] = (T) it.next();
        }

        return a;
    }

    public boolean containsAll(final Collection<?> c) {
        for (Object o : c) {
            //noinspection SuspiciousMethodCalls
            if (!contains(o))
                return false;
        }

        return true;
    }

    public boolean addAll(final Collection<? extends Element> c) {
        for (Element element : c)
            add(element);

        return true;
    }

    public boolean remove(final Object o) {
        throw new UnsupportedOperationException("Cannot remove elements from cyclic array");
    }

    public boolean removeAll(final Collection<?> c) {
        throw new UnsupportedOperationException("Cannot remove elements from cyclic array");
    }

    public boolean retainAll(final Collection<?> c) {
        throw new UnsupportedOperationException("Cannot remove elements from cyclic array");
    }

    public boolean equals(final Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final CyclicArray that = (CyclicArray) o;

        if (end != that.end) return false;
        if (!Arrays.equals(elements, that.elements)) return false;

        return true;
    }

    public int hashCode() {
        int result = end;
        result = 37 * result + Arrays.hashCode(elements);

        return result;
    }

    private class CyclicIterator implements java.util.Iterator<Element> {
        int currLoc;
        int leftToRead;

        public CyclicIterator() {
            leftToRead = size();

            currLoc = (end - leftToRead + elements.length) % elements.length;
        }

        public boolean hasNext() {
            return leftToRead > 0;
        }

        public Element next() {
            if (leftToRead <= 0)
                throw new IndexOutOfBoundsException();

            leftToRead--;
            return elements[currLoc++ % elements.length];
        }

        public void remove() {
            throw new UnsupportedOperationException("Cannot remove elements from cyclic array");
        }
    }
}
