package com.onaro.util;

import java.util.*;
import java.util.regex.*;

/**
 * Comparator that knows how to compare versions. For example - 5.5.3 vs.3.7 or 10.2.  Assumes a dot based
 * version patterning, version strings should be normalized before passing to this class.  
 *
 * <p>
 * This may be used statically {@link #compareVersions(String, String)}} or as a standard comparator for sorting
 * and the like.
 *
 */
public class VersionComparator implements Comparator<String> {



    @Override
    public int compare(final String o1, final String o2) {
        List<Integer> l1 = extractDigitList(o1) ;
        List<Integer> l2 = extractDigitList(o2) ;
        return compareIntList(l1,l2) ;
    }

    /**
     * Use this class to statically compare two version strings.  Returns an int < 0 if version1
     * is "less" than version2, == 0 if they are the same, and > 0 if version1 is "greater" than
     * version2.
     *
     * @param version1  1st version string.
     * @param version2  2nd version string.
     * @return int indicating one of three states:  < 0 if version1 < version2, ==0 if version1 == version2,
     *   and > 0 if version1 > version2.
     */
    public static int compareVersions (final String version1, final String version2) {
        List<Integer> l1 = extractDigitList(version1) ;
        List<Integer> l2 = extractDigitList(version2) ;
        return compareIntList(l1,l2) ;
    }


     private static int compareIntList(List<Integer> a, List<Integer> b) {
        int minLen = Math.min(a.size(), b.size());
        for (int i = 0; i < minLen; i++) {
            if (a.get(i) < b.get(i)) {
                return -1;
            }
            else if (a.get(i) > b.get(i)) {
                return 1;
            }
        }
        if (a.size() == b.size()) {
            return 0;
        }
        else {
            return a.size() < b.size() ? -1 : 1;
        }
    }

    private static List<Integer> extractDigitList(String version){
        Matcher m = VersionUtils.versionPattern.matcher(version);
        if (!m.find()) {
             throw new IllegalArgumentException("Version string does not contain numeric sequence: "+ version)  ;
        }

        String numbers = m.group(0) ;
        List<Integer> res = new ArrayList<Integer>() ;
        StringTokenizer tok = new StringTokenizer(numbers, ".") ;
        while (tok.hasMoreTokens()){
            res.add(Integer.parseInt(tok.nextToken())) ;
        }
        return res ;
    }
}
