package com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transform.ocie;

import com.netapp.oci.netapp.foundation.qosvolumeworkloaduuidconvertion.interfaces.remote.UuidConversionRemote;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.BuilderUtil;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.builders.zapi.cmode.netappfoundation.ClusterBuilder;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.config.NetAppOCIEConfiguration;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.EnumWrapper;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.UUIDKey;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.*;
import com.onaro.sanscreen.acquisition.datasource.netapp_ocie.transport.zapi.ZAPIPropertiesManager;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorException;
import com.onaro.sanscreen.acquisition.framework.datasource.DataSourceErrorType;
import com.onaro.sanscreen.acquisition.framework.datasource.PackageContext;
import com.onaro.sanscreen.acquisition.framework.mgmt.ServiceManager;
import com.onaro.sanscreen.acquisition.netappschema.netappdata.*;
import com.onaro.sanscreen.acquisition.schema.KeyBase;
import com.onaro.util.VersionComparator;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeAggregateSpare.AggregateSpareKey;
import static com.onaro.sanscreen.acquisition.datasource.netapp_ocie.model.cmode.CModeVolume.VolumeNameKey;

public class NetAppOCIETransform {
    private static final Logger logger = LoggerFactory.getLogger(NetAppOCIETransform.class);
    private static final Integer BYTES_PER_BLOCK = 4096;
    private static final Pattern NODE_VERSION_PATTERN = Pattern.compile("\\D*(\\d+)\\.(\\d+)(\\.(\\d+))?.*");
    private final CModeClusterRoot clusterModelRoot;
    private final StorageRoot storageRoot;
    private static final String SYSTEM_STORAGE_SERVICE = "SYSTEM";
    private static final String DEFAULT_STORAGE_SERVICE = "DEFAULT";
    private static final String CAMPUS = "campus";

    public NetAppOCIETransform(final CModeClusterRoot clusterModelRoot, final StorageRoot storageRoot) {
        this.clusterModelRoot = clusterModelRoot;
        this.storageRoot = storageRoot;
    }

    public void transform(NetAppOCIEConfiguration config) throws DataSourceErrorException {
        for (final CModeCluster cModeCluster : clusterModelRoot.getCModeClusters().values()) {
            if (config.isQosVolumeWorkloadUUIDConversionEnabled()) {
                qosVolumeWorkloadfUUIDConversion(cModeCluster);
            }
            final Cluster cluster = transformCluster(cModeCluster);
            final Map<String, Node> nodesByName = transformNodes(cModeCluster, cluster);
            final Map<String, String> partnerNodeNames = transformHAPairs(cModeCluster, cluster, nodesByName);
            final Set<RaidGroup.RaidGroupKey> raidGroupKeys = new HashSet<RaidGroup.RaidGroupKey>();
            final Map<String, Aggregate.AggregateKey> aggrKeysByName = transformAggregates(cModeCluster, nodesByName, raidGroupKeys, cluster);
            final Map<String, StoragePool> storagePoolByName = transformStoragePools(cModeCluster, cluster);
            transformStoragePoolAvailableSizes(cModeCluster, nodesByName, storagePoolByName);
            transformStoragePoolAggregate(cModeCluster, aggrKeysByName, storagePoolByName);
            final List<Map<String, Disk.DiskKey>> diskMapList = transformDisks(cModeCluster, cluster, nodesByName, aggrKeysByName, raidGroupKeys, storagePoolByName);
            Map<String, Disk.DiskKey> diskByName = diskMapList.get(0);
            Map<String, Disk.DiskKey> diskBySerialNumber = diskMapList.get(1);
            transformDiskAggregateRelationship(cModeCluster, cluster, diskByName, aggrKeysByName);
            final Map<String, NetworkIpSpace.NetworkIpSpaceKey> ipSpaceNameToKeyMap = transformNetworkIpSpace(cModeCluster, cluster);
            final Map<String, NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey> broadcastDomainNameToKeyMap = transformNetworkPortBroadcastDomain(cModeCluster, cluster, ipSpaceNameToKeyMap);
            final Map<Node.NodeKey, Map<String, NetworkPort.PortKey>> portKeysByNameByNodeKey = transformNetworkPorts(cModeCluster, nodesByName, ipSpaceNameToKeyMap, broadcastDomainNameToKeyMap);
            transformNetworkPortRelationships(cluster, portKeysByNameByNodeKey);
            transformFcpPorts(cModeCluster, nodesByName);
            transformInitiatorPorts(cModeCluster, nodesByName);
            final Map<String, VServer.VServerKey> vServerKeysByName = transformVServers(cModeCluster, cluster, aggrKeysByName, ipSpaceNameToKeyMap);
            final Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> failoverGroupKeyByName = transformNetworkFailoverGroup(cModeCluster, cluster, broadcastDomainNameToKeyMap, vServerKeysByName);
            transformKerberos(cModeCluster, cluster, vServerKeysByName);
            transformFcpServices(cModeCluster, cluster, vServerKeysByName);
            transformISCSIServices(cModeCluster, cluster, vServerKeysByName);
            transformCIFSServices(cModeCluster, cluster, vServerKeysByName);
            transformCIFSOptions(cModeCluster, cluster, vServerKeysByName);
            transformNFSServices(cModeCluster, cluster, vServerKeysByName);
            transformNetworkFailoverGroupNetworkPortRelationship(cModeCluster, cluster, nodesByName, failoverGroupKeyByName, portKeysByNameByNodeKey );
            final Map<String, JobSchedule> jobSchedulesByName = transformJobSchedules(cModeCluster, cluster);
            final Map<String, SnapshotPolicy.SnapshotPolicyKey> snapshotPolicyKeysByName = transformSnapshotPolicies(cModeCluster, cluster, vServerKeysByName, jobSchedulesByName);
            final Map<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>> exportPolicyKeysByNameByVServerKey = transformExportPoliciesAndRules(cModeCluster, cluster, vServerKeysByName);
            final Map<Volume, String> volumeToCloneParentName = new HashMap<Volume, String>();
            final Map<VServer.VServerKey, Map<String, SisPolicy.SisPolicyKey>> sisPolicyKeysByNameByVServer = transformSisPolicies(cModeCluster, cluster, vServerKeysByName, jobSchedulesByName);
            final Map<String, QosPolicyGroup.QosPolicyGroupKey> qosPolicyGroupKeysByName = transformQosPolicyGroups(cModeCluster, cluster, vServerKeysByName);
            final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey = transformVolumes(cModeCluster, cluster, exportPolicyKeysByNameByVServerKey, volumeToCloneParentName, snapshotPolicyKeysByName, qosPolicyGroupKeysByName, nodesByName);
            transformVolumeRelationships(cModeCluster, cluster, volumeKeysByNameByVServerKey);
            transformVolumeClones(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey, volumeToCloneParentName);
            transformVolumeFlexCaches(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey);
            transformVolumeSisStatus(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey, sisPolicyKeysByNameByVServer);
            transformVolumeMoves(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey, aggrKeysByName);
            transformQuotaStatuses(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey);
            final Map<VServer.VServerKey, Map<String, QTree.QTreeKey>> qTreeKeysByAliasByVServerKey = transformQTrees(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey, exportPolicyKeysByNameByVServerKey);
            final Map<Volume.VolumeKey, Map<String, Snapshot.SnapshotKey>> snapshotKeysByNameByVolumeKey = transformSnapshots(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey);
            final Map<LUN.LUNKey, LUN> lunsByKey = transformLUNs(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey, qTreeKeysByAliasByVServerKey, qosPolicyGroupKeysByName);
            transformStorageServices(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey);
            final Map<VServer.VServerKey, Map<String, RoutingGroup.RoutingGroupKey>> routingGroupKeysByNameByVServer = transformRouteGroups(cModeCluster, cluster, vServerKeysByName);
            final Map<VServer.VServerKey, Map<String, NetworkLif>> networkLIFsByNameByVServerMap = transformLifs(cModeCluster, cluster, vServerKeysByName, nodesByName, routingGroupKeysByNameByVServer, failoverGroupKeyByName);
            transformIGroups(cModeCluster, cluster, vServerKeysByName);
            transformLunMaps(cModeCluster, cluster, vServerKeysByName, lunsByKey, nodesByName);
            transformDiskPaths(cModeCluster, cluster, nodesByName, diskByName.values());
            final Map<VServer.VServerKey, Map<String, CIFSShare>> cifsSharesByNameByVServerKey = transformCIFSShares(cModeCluster, cluster, vServerKeysByName);
            transformCIFSShareACLs(cModeCluster, cluster, vServerKeysByName, cifsSharesByNameByVServerKey);
            final Map<VServer.VServerKey, Map<String, ISCSIPortalGroup>> iscsiPortalGroupByNameByVServer = transformISCSIPortalGroups(cModeCluster, cluster, vServerKeysByName);
            transformPortSets(cModeCluster, cluster, vServerKeysByName);
            transformISCSIPortalGroupLIFs(cModeCluster, cluster, vServerKeysByName, iscsiPortalGroupByNameByVServer, networkLIFsByNameByVServerMap);
            transformISCSISessions(cModeCluster, cluster, iscsiPortalGroupByNameByVServer, vServerKeysByName, networkLIFsByNameByVServerMap);
            transformISCSISecurity(cModeCluster, cluster, vServerKeysByName);
            transformLicenses(cModeCluster, cluster, nodesByName);
            transformQuotas(cModeCluster, cluster, vServerKeysByName);
            final Map<VServer.VServerKey, Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey>> policyKeysByVServer = transformSnapMirrorPolicies(cModeCluster, cluster, vServerKeysByName);
            transformSnapMirrors(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey, snapshotKeysByNameByVolumeKey,
                    policyKeysByVServer, jobSchedulesByName, nodesByName);

            LunImportTransformer.transform(cModeCluster, cluster, vServerKeysByName, nodesByName, diskBySerialNumber);
            MetroClusterOperationTransformer.transform(cModeCluster, cluster, nodesByName);
            ClusterPeerTransform.transform(cModeCluster,cluster, ipSpaceNameToKeyMap);
            VServerPeerTransformer.transform(cModeCluster, cluster, vServerKeysByName);
            NetworkSubnetTransform.transform(cModeCluster, cluster,ipSpaceNameToKeyMap,broadcastDomainNameToKeyMap );
            SnapMirrorHistoryTransformer.transform(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey);
            transformQosVolumeWorkloads(cModeCluster, cluster, vServerKeysByName, volumeKeysByNameByVServerKey);
            generateQosWorkloadNodeRelationship(cluster);
            final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey = transformQosServiceCenters(cModeCluster, nodesByName, aggrKeysByName);

            // Generating QosWorkloadDetails for both SN and FS for OPM's per aggregate stats story - US21522
            generateQosWorkloadDetails(cModeCluster, cluster, qosServiceCenterUUIDToKey, partnerNodeNames);

            transformServiceProcessors(cModeCluster, nodesByName);
            transformStorageShelves(cModeCluster, cluster, nodesByName);
            /*
                fix of burt 804059. May use this code later.
            */
            //transformStorageSwitches(cModeCluster, cluster, nodesByName);
            transformVServerNameMappings(cModeCluster, cluster, vServerKeysByName);
            transformApplicationRecords(cModeCluster, cluster);
            transformNodePerfUUIDs(cModeCluster, cluster);
            transformVserverPerfUUIDs(cModeCluster, cluster);
            transformSystemHealthAlerts(cModeCluster, cluster, nodesByName);
            // Some capacities are summations from child elements.
            // Now that all parent/child relationships have been established,
            // loop again to sum capacities from children to parents.
            updateAggregateCapacities(cluster);
        }
    }

  private void transformNetworkFailoverGroupNetworkPortRelationship(final CModeCluster cModeCluster, final Cluster cluster, Map<String, Node> nodesByName, final Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> failoverGroupsByName, final Map<Node.NodeKey, Map<String, NetworkPort.PortKey>> networkPortsKeysByName)
  {
    if (!cModeCluster.getNetworkFailoverGroupNetworkPortRelationships().isEmpty())
    {
      for (final CModeNetworkFailoverGroupNetworkPortRelationship relationship : cModeCluster.getNetworkFailoverGroupNetworkPortRelationships().values())
      {
        final NetworkFailoverGroup.NetworkFailoverGroupKey failoverGroupKey = failoverGroupsByName.get(relationship.getNetworkFailoverGroupName());
        if (failoverGroupKey != null)
        {
          final Node node = nodesByName.get(relationship.getServerNodeName());
          if (node == null)
            logger.debug(PackageContext.getLogPrefix() + " - Failed to find the node " + relationship.getNodeName() + " in the cluster  " + cModeCluster.getClusterUUID());
          else
          {
            final Map<String, NetworkPort.PortKey> portKeyMap = networkPortsKeysByName.get(node.getKey());
            if (portKeyMap == null)
              logger.debug(PackageContext.getLogPrefix() + " - Failed to find the portMap for node " + relationship.getNodeName() + " in the cluster  " + cModeCluster.getClusterUUID());
            else
            {
              NetworkPort.PortKey portKey = portKeyMap.get(relationship.getNetworkPortName());
              NetworkFailoverGroupNetworkPortRelationship networkFailoverGroupNetworkPortRelationship = cluster.createNetworkFailoverGroupNetworkPortRelationship(node.getKey(), failoverGroupKey, portKey);
              cluster.getNetworkFailoverGroupNetworkPortRelationships().add( networkFailoverGroupNetworkPortRelationship );
            }
          }
        }
      }
    }
  }
    
    private void qosVolumeWorkloadfUUIDConversion(CModeCluster cModeCluster) throws DataSourceErrorException {
        Map<String, String> nameUUIDMap = new HashMap<String, String>();
        for (final CModeQosVolumeWorkload cModeQosWorkload : cModeCluster.getQosVolumeWorkloads().values()) {
            if (cModeQosWorkload.getWorkloadTypeWrapper().getEnumValue() != QosVolumeWorkload.QosWorkloadClassType.USER_DEFINED 
                && cModeQosWorkload.getName().matches("^_.*$")) {
                nameUUIDMap.put(cModeQosWorkload.getName(), cModeQosWorkload.getKey().getUUID());
            }
        }
        
        if (!nameUUIDMap.isEmpty()) {
            try {
                UuidConversionRemote uuidConverter = ServiceManager.getInstance().getCommunicationManager().getSession(UuidConversionRemote.JNDI_NAME);
                uuidConverter.convert(cModeCluster.getClusterUUID(), nameUUIDMap);
            }
            catch (Exception e) {
                final String msg = PackageContext.getLogPrefix() + " - Dynamic UUID conversion EJB failed, cluster uuid: "
                        +  cModeCluster.getClusterUUID() + ", name UUID map: " +  nameUUIDMap;
                throw new DataSourceErrorException(DataSourceErrorType.INTERNAL_ERROR, msg, msg);
            }
        }
    }

  private Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> transformNetworkFailoverGroup(CModeCluster cModeCluster, Cluster cluster,
                                                                                                  final Map<String, NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey> broadcastDomainNameToKeyMap,
                                                                                                  final Map<String, VServer.VServerKey> vServerKeysByName ) {
    Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> map = new HashMap<String, NetworkFailoverGroup.NetworkFailoverGroupKey>();
    for (CModeNetworkFailoverGroup cModeNetFailoverGroup : cModeCluster.getNetworkFailoverGroups().values()) {
      final NetworkFailoverGroup networkFailoverGroup = cluster.createNetworkFailoverGroup(cModeNetFailoverGroup.getNetworkFailoverGroupName());
      if (cModeNetFailoverGroup.getNetworkPortBroadcastDomainName() != null) {
        NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey broadcastDomainKey = broadcastDomainNameToKeyMap.get(cModeNetFailoverGroup.getNetworkPortBroadcastDomainName());
        if (broadcastDomainKey != null) {
          networkFailoverGroup.setBroadcastDomainKey(broadcastDomainKey);
        } else {
          logger.info("Unable to find the Broadcast domain return with the failover group {}",
                  cModeNetFailoverGroup.getNetworkFailoverGroupName());
        }
      }

      final VServer.VServerKey vserver = vServerKeysByName.get(cModeNetFailoverGroup.getVServer());
      if (vserver != null) {
        networkFailoverGroup.setVserverId(vserver);
      }

      //fill in containers
      cluster.getNetworkFailoverGroups().add(networkFailoverGroup);
      map.put( networkFailoverGroup.getKey().getID(NetworkFailoverGroup.NetworkFailoverGroupKey.Field.failover_group_name), networkFailoverGroup.getKey() );
    }

    return map;
  }

    private Map<String, NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey> transformNetworkPortBroadcastDomain(
            CModeCluster cModeCluster, Cluster cluster, Map<String, NetworkIpSpace.NetworkIpSpaceKey> ipSpaceNameToKeyMap) {
        Map<String, NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey>
                broadcastDomainKeyMap = new HashMap<String, NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey>();
        for (CModeNetworkPortBroadcastDomain cModeBroadcastDomain : cModeCluster.getNetworkPortBroadcastDomains().values()) {
            if (cModeBroadcastDomain.getIpSpaceName() != null) {
                NetworkIpSpace.NetworkIpSpaceKey ipSpaceKey = ipSpaceNameToKeyMap.get(cModeBroadcastDomain.getIpSpaceName());
                if (ipSpaceKey != null) {
                    final NetworkPortBroadcastDomain broadcastDomain =
                            cluster.createNetworkPortBroadcastDomain(cModeBroadcastDomain.getBroadcastDomainName(),
                                    ipSpaceKey);
                    broadcastDomain.setMtu(cModeBroadcastDomain.getMtu());
                    broadcastDomainKeyMap.put(cModeBroadcastDomain.getBroadcastDomainName(), broadcastDomain.getKey());
                    cluster.getNetworkPortBroadcastDomains().add(broadcastDomain);
                } else {
                    logger.info("Unable to find the IPSpace return with the broadcast domain {}",
                            cModeBroadcastDomain.getBroadcastDomainName());
                }
            }
        }

        return broadcastDomainKeyMap;
    }

    private Map<String, NetworkIpSpace.NetworkIpSpaceKey> transformNetworkIpSpace(CModeCluster cModeCluster, Cluster cluster) {
        Map<String, NetworkIpSpace.NetworkIpSpaceKey> ipSpaceKeyMap = new HashMap<String, NetworkIpSpace.NetworkIpSpaceKey>();
        for (CModeNetworkIPSpace cModeNetIPSpace : cModeCluster.getNetworkIPSpaces().values()) {
          String uuid = cModeNetIPSpace.getKey().getUUID();
          if (StringUtils.startsWith(uuid, CModeNetworkIPSpace.CREATED_NETWORKIPSPACE_UUID_PREFIX)) {
            DoNotDelete doNotDelete =  cluster.createDoNotDelete(UUID.randomUUID().toString());
            doNotDelete.setObjectTagName("NetworkIpSpace");
            AlternateKey alternateKey = doNotDelete.createAlternateKey(UUID.randomUUID().toString());
            alternateKey.setName("name");
            alternateKey.setValue(cModeNetIPSpace.getIPSpaceName());
            doNotDelete.getAlternateKeys().add(alternateKey);
            cluster.getDoNotDeleteList().add(doNotDelete);
          }
          else {
            final NetworkIpSpace ipSpace = cluster.createNetworkIPSpace(uuid);
            ipSpace.setId(cModeNetIPSpace.getId()).setIPSpaceName(cModeNetIPSpace.getIPSpaceName());
            ipSpaceKeyMap.put(cModeNetIPSpace.getIPSpaceName(), ipSpace.getKey());
            cluster.getNetworkIpSpaces().add(ipSpace);
          }
        }
        return ipSpaceKeyMap;
    }

    private void transformDiskAggregateRelationship(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Disk.DiskKey> diskByName, final Map<String, Aggregate.AggregateKey> aggrKeysByName) {
        if (!cModeCluster.getDiskAggregateRelationships().isEmpty()) {
            // ONTAP 8.3 and higher
            for (final CModeDiskAggregateRelationship cModeDiskAggregateRelationship : cModeCluster.getDiskAggregateRelationships().values()) {
                final Disk.DiskKey disk = diskByName.get(cModeDiskAggregateRelationship.getDiskName());
                if (disk != null) {
                    final Aggregate.AggregateKey aggregateKey = new Aggregate.AggregateKey(cModeCluster.getClusterUUID(), cModeDiskAggregateRelationship.getAggregateUUID());
                    if (aggrKeysByName.containsValue(aggregateKey)) {
                        final DiskAggregateRelationship diskAggregateRelationship = cluster.createDiskAggregateRelationship(disk, aggregateKey);
                        diskAggregateRelationship.setIsCache(cModeDiskAggregateRelationship.getIsCache());
                        cluster.getDiskAggregateRelationships().add(diskAggregateRelationship);
                    } else {
                        if (logger.isDebugEnabled()) {
                            logger.debug(PackageContext.getLogPrefix() + " - Failed to find the aggregate " + cModeDiskAggregateRelationship.getAggregateUUID() + " in the cluster  " + cModeCluster.getClusterUUID());
                        }
                    }
                } else {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Failed to find the disk " + cModeDiskAggregateRelationship.getDiskName() + " in the cluster  " + cModeCluster.getClusterUUID());
                    }
                }
            }
        } else {
            // ONTAB 8.2 and lower
            for (final Disk disk : cluster.getDisks().values()) {
                final Disk.DiskKey diskKey = disk.getKey();
                final Aggregate.AggregateKey aggregateKey = disk.getAggregateKey();
                if (aggregateKey != null) {
                    final DiskAggregateRelationship diskAggregateRelationship = cluster.createDiskAggregateRelationship(diskKey, aggregateKey);
                    cluster.getDiskAggregateRelationships().add(diskAggregateRelationship);
                }
            }
        }
    }

    private Map<String, StoragePool> transformStoragePools(final CModeCluster cModeCluster, final Cluster cluster) {
        final Map<String, StoragePool> storagePoolByName = new HashMap<String, StoragePool>();
        final StoragePools storagePools = cluster.getStoragePools();
        for (final CModeStoragePool cModeStoragePool : cModeCluster.getStoragePools().values()) {
            final StoragePool storagePool = transformStoragePool(cluster, cModeStoragePool);
            storagePools.add(storagePool);
            storagePoolByName.put(storagePool.getName(), storagePool);
        }
        return storagePoolByName;
    }

    private StoragePool transformStoragePool(final Cluster cluster, final CModeStoragePool cModeStoragePool) {
        final StoragePool storagePool = cluster.createStoragePool(cModeStoragePool.getStoragePoolName());
        storagePool.setAllocationUnitSize(cModeStoragePool.getAllocationUnitSize());
        storagePool.setDiskCount(cModeStoragePool.getDiskCount());
        storagePool.setHealthy(cModeStoragePool.getHealthy());
        storagePool.setPoolTotalSize(cModeStoragePool.getPoolTotalSize());
        storagePool.setPoolUsableSize(cModeStoragePool.getPoolUsableSize());
        storagePool.setStorageType(cModeStoragePool.getStorageType().getEnumValue());
        storagePool.setStorageTypeRaw(cModeStoragePool.getStorageType().getRawValue());
        return storagePool;
    }

    private void transformStoragePoolAvailableSizes(final CModeCluster cModeCluster, final Map<String, Node> nodesByName, final Map<String, StoragePool> storagePoolByName) {
        for (final CModeStoragePoolAvailableSize cModeStoragePoolAvailableSize : cModeCluster.getStoragePoolAvailableSizes().values()) {
            final StoragePool storagePool = storagePoolByName.get(cModeStoragePoolAvailableSize.getStoragePoolName());
            if (storagePool != null) {
                final Node spareNode = nodesByName.get(cModeStoragePoolAvailableSize.getSpareNodeName());
                if (spareNode != null) {
                    final StoragePoolAvailableSize storagePoolAvailableSize = storagePool.createStoragePoolAvailableSize(spareNode.getKey());
                    storagePoolAvailableSize.setAllocationUnitCount(cModeStoragePoolAvailableSize.getAllocationUnitCount());
                    storagePoolAvailableSize.setAllocationUnitSize(cModeStoragePoolAvailableSize.getAllocationUnitSize());
                    storagePoolAvailableSize.setAvailableSize(cModeStoragePoolAvailableSize.getAvailableSize());
                    storagePoolAvailableSize.setStorageType(cModeStoragePoolAvailableSize.getStorageType().getEnumValue());
                    storagePoolAvailableSize.setStorageTypeRaw(cModeStoragePoolAvailableSize.getStorageType().getRawValue());
                    storagePoolAvailableSize.setSyncMirrorPool(cModeStoragePoolAvailableSize.getSyncMirrorPool());
                    storagePool.getStoragePoolAvailableSizes().add(storagePoolAvailableSize);
                }
            }
        }
    }

    private void transformStoragePoolAggregate(final CModeCluster cModeCluster, final Map<String, Aggregate.AggregateKey> aggregateKeysByName, final Map<String, StoragePool> storagePoolByName) {
        for (final CModeStoragePoolAggregate cModeStoragePoolAggregate : cModeCluster.getStoragePoolAggregates().values()) {
            final StoragePool storagePool = storagePoolByName.get(cModeStoragePoolAggregate.getStoragePoolName());
            final Aggregate.AggregateKey aggregateKey = aggregateKeysByName.get(cModeStoragePoolAggregate.getAggregateName());
            if (storagePool != null && aggregateKey != null) {
                final StoragePoolAggregate storagePoolAggregate = storagePool.createStoragePoolAggregate(aggregateKey);
                storagePoolAggregate.setAllocatedUnitCount(cModeStoragePoolAggregate.getAllocatedUnitCount());
                storagePoolAggregate.setUsedCapacity(cModeStoragePoolAggregate.getUsedCapacity());
                storagePool.getStoragePoolAggregates().add(storagePoolAggregate);
            }
        }
    }
    private void transformApplicationRecords(final CModeCluster cModeCluster, final Cluster cluster) {
        for (final CModeApplicationRecord cModeApplicationRecord : cModeCluster.getApplicationRecords().values()) {
            final ApplicationRecord applicationRecord = cluster.createAndAddApplicationRecord(cModeApplicationRecord.getKey().getUUID());
            applicationRecord.setRecordName(cModeApplicationRecord.getRecordName());
            applicationRecord.setRecordValue(cModeApplicationRecord.getRecordValue());
            applicationRecord.setModifiedFrom(cModeApplicationRecord.getModifiedFrom());
            applicationRecord.setLastModifiedTimestamp(cModeApplicationRecord.getLastModifiedTimestamp());
        }
    }

    private Cluster transformCluster(final CModeCluster cModeCluster) {
        final Cluster cluster = storageRoot.createCluster(cModeCluster.getClusterUUID());
        cluster.setName(cModeCluster.getClusterName());
        cluster.setMangementIP(cModeCluster.getManagementIP());
        cluster.setSerialNum(cModeCluster.getSerialNum());
        cluster.setVersionString(cModeCluster.getVersionString());
        cluster.setGeneration(cModeCluster.getGeneration());
        cluster.setMajor(cModeCluster.getMajor());
        cluster.setMinor(cModeCluster.getMinor());
        cluster.setLocation(cModeCluster.getLocation());
        cluster.setContact(cModeCluster.getContact());
        cluster.setTimezone(cModeCluster.getTimezone());
        cluster.setTimezoneUTC(cModeCluster.getTimezoneUTC());
        cluster.setTimezoneVersion(cModeCluster.getTimezoneVersion());
        cluster.setCpuColumns(cModeCluster.getCpuColumns());
        cluster.setWaflPhaseTimesColumns(cModeCluster.getWaflCpPhaseTimesColumns());
        cluster.setVolumeWorkloadReadIOColumns(cModeCluster.getVolumeWorkloadReadIOColumns());
        cluster.setHaConfigured(cModeCluster.getHaConfigured());
        cluster.setSslFipsEnabled(cModeCluster.getSslFipsEnabled());

        // Diag status is returned by a 2nd ZAPI that is not supported by all cluster
        // simulators, so StatusWrapper may be null.
        final EnumWrapper<Cluster.HmStatus> statusWrapper = cModeCluster.getDiagnosisStatus();
        if (statusWrapper != null) {
            cluster.setDiagnosisStatus(statusWrapper.getEnumValue());
            cluster.setDiagnosisStatusRaw(statusWrapper.getRawValue());
        }
        //MCC attributes
        cluster.setMetrocluster(cModeCluster.getIsMetrocluster());
        final EnumWrapper<Cluster.MtConfigurationState> mtConfigurationStateEnumWrapper = cModeCluster.getMtConfigurationState();
        if(mtConfigurationStateEnumWrapper != null) {
            cluster.setMtConfigurationState(mtConfigurationStateEnumWrapper.getEnumValue());
            cluster.setMtConfigurationStateRaw(mtConfigurationStateEnumWrapper.getRawValue());
        }
        final EnumWrapper<Cluster.MtMode> mtModeEnumWrapper = cModeCluster.getMtMode();
        if(mtModeEnumWrapper != null) {
            cluster.setMtMode(mtModeEnumWrapper.getEnumValue());
            cluster.setMtModeRaw(mtModeEnumWrapper.getRawValue());
        }
        final EnumWrapper<Cluster.MtConfigurationState> mtRemoteConfigurationStateEnumWrapper = cModeCluster.getMtRemoteConfigurationState();
        if(mtRemoteConfigurationStateEnumWrapper != null) {
            cluster.setMtRemoteConfigurationState(mtRemoteConfigurationStateEnumWrapper.getEnumValue());
            cluster.setMtRemoteConfigurationStateRaw(mtRemoteConfigurationStateEnumWrapper.getRawValue());
        }
        final EnumWrapper<Cluster.MtMode> mtRemoteModeEnumWrapper = cModeCluster.getMtRemoteMode();
        if(mtRemoteModeEnumWrapper != null) {
            cluster.setMtRemoteMode(mtRemoteModeEnumWrapper.getEnumValue());
            cluster.setMtRemoteModeRaw(mtRemoteModeEnumWrapper.getRawValue());
        }
        cluster.setMtConfigurationType(cModeCluster.getMtConfigurationType());

        storageRoot.getClusters().add(cluster);

        final VersionComparator versionComparator = new VersionComparator();
        final String clusterVersion = cModeCluster.getVersionString();
        final boolean isLessThanSierraNevada = (versionComparator.compare(clusterVersion, Cluster.SIERRA_NEVADA) < 0);     // Rolling Rock  is less than Sierra Nevada
        cModeCluster.setLessThanSierraNevada(isLessThanSierraNevada);
        return cluster;
    }

    private Map<String, Node> transformNodes(final CModeCluster cModeCluster, final Cluster cluster) {
        final Map<String, Node> nodesByName = new HashMap<String, Node>();
        Long maxNodeUptime = 0L;
        for (final CModeNodeDetail cModeNodeDetail : cModeCluster.getNodeDetails().values()) {
            final Node node = transformNode(cluster, cModeCluster, cModeNodeDetail);
            cluster.getNodes().add(node);
            nodesByName.put(cModeNodeDetail.getNodeNodeName(), node);

            final Long nodeUptime = cModeNodeDetail.getNodeUptime();
            if (nodeUptime != null && nodeUptime > maxNodeUptime) {
                maxNodeUptime = nodeUptime;
            }
        }
        // cluster.setUpTime(maxNodeUptime);
        return nodesByName;
    }

    private Map<String, String> transformHAPairs(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Node> nodesByName) {
        // HA Pairs
        final Map<String, String> partnerNodeNames = new HashMap<String, String>();
        final SortedMap<Node.NodeKey, Node.NodeKey> pairedNodes = new TreeMap<Node.NodeKey, Node.NodeKey>();
        for (final CModeNodeFailover failover : cModeCluster.getNodeFailovers().values()) {
            final String partnerAName = failover.getKey().getName();
            if (!nodesByName.containsKey(partnerAName)) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find node " + partnerAName + " for aggregate HA Pair");
                continue;
            }
            final Node nodeA = nodesByName.get(partnerAName);
            final String partnerBName = failover.getPartnerName();
            if (!nodesByName.containsKey(partnerBName)) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find node " + partnerBName + " for aggregate HA Pair");
                }
                continue;
            }
            partnerNodeNames.put(partnerAName, partnerBName);
            partnerNodeNames.put(partnerBName, partnerAName);
            final Node nodeB = nodesByName.get(partnerBName);
            // Pairs are ordered and each node represented only once
            pairedNodes.put(nodeA.getKey(), nodeB.getKey());
            pairedNodes.put(nodeB.getKey(), nodeA.getKey());
        }
        final Set<Node.NodeKey> reportedPairedNodes = new HashSet<Node.NodeKey>();
        for (final Map.Entry<Node.NodeKey, Node.NodeKey> pairEntry : pairedNodes.entrySet()) {
            final Node.NodeKey nodeAKey = pairEntry.getKey();
            final Node.NodeKey nodeBKey = pairEntry.getValue();
            if (reportedPairedNodes.contains(nodeAKey) || reportedPairedNodes.contains(nodeBKey)) {
                continue;
            }
            reportedPairedNodes.add(nodeAKey);
            reportedPairedNodes.add(nodeBKey);
            final HaPair haPair = cluster.createHaPair(nodeAKey);
            haPair.setNodeBUUID(nodeBKey);
            cluster.getHaPairs().add(haPair);
        }
        return partnerNodeNames;
    }

    private Map<String, Aggregate.AggregateKey> transformAggregates(final CModeCluster cModeCluster, final Map<String, Node> nodesByName, final Set<RaidGroup.RaidGroupKey> raidGroupKeys, Cluster cluster) {
        final Map<String, Aggregate.AggregateKey> aggrsKeysByName = new HashMap<String, Aggregate.AggregateKey>();
//        long aggregateBytesTotal = 0;
//        long aggregateBytesUsed = 0;
        for (final CModeAggregate cModeAggregate : cModeCluster.getAggregates().values()) {
            final List<String> aggregateNodeNames = cModeAggregate.getNodeNameStrings();
            final String aggregateID = cModeAggregate.getKey().getUUID();
            if (aggregateNodeNames == null || aggregateNodeNames.size() == 0) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find any node for aggregate " + aggregateID);
                addAggregateToDoNotDeleteList(cluster, aggregateID);
                continue;
            }
            final String firstNodeName = aggregateNodeNames.get(0);
            if (firstNodeName == null || firstNodeName.isEmpty()) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find node for aggregate " + aggregateID);
                addAggregateToDoNotDeleteList(cluster, aggregateID);
                continue;
            }
            if (!nodesByName.containsKey(firstNodeName)) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find node " + firstNodeName + " for aggregate " + aggregateID);
                addAggregateToDoNotDeleteList(cluster, aggregateID);
                continue;
            }
            final Node node = nodesByName.get(firstNodeName);
            final Aggregate aggregate = transformAggregate(node, cModeAggregate, raidGroupKeys);
//            aggregateBytesTotal += getLongForSum(aggregate.getSizeTotal()) + getLongForSum(aggregate.getSnapshotSizeTotal());
//            aggregateBytesUsed += getLongForSum(aggregate.getSizeUsed()) + getLongForSum(aggregate.getSnapshotSizeUsed());
            node.getAggregates().add(aggregate);
            aggrsKeysByName.put(aggregate.getName(), aggregate.getKey());
        }
        // cluster.setAggregateBytesTotal(aggregateBytesTotal);
        // cluster.setAggregateBytesUsed(aggregateBytesUsed);
        return aggrsKeysByName;
    }

    private List<Map<String, Disk.DiskKey>> transformDisks(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Node> nodesByName, final Map<String, Aggregate.AggregateKey> aggrsUUIDsByName, final Set<RaidGroup.RaidGroupKey> raidGroupKeys, final Map<String, StoragePool> storagePoolByName) {
        // Disks
        final Map<String, Disk.DiskKey> diskByName = new HashMap<String, Disk.DiskKey>();
        final Map<String, Disk.DiskKey> diskBySerialNumber = new HashMap<String, Disk.DiskKey>();
        final List<Map<String, Disk.DiskKey>> diskKeyMapList = Arrays.asList(diskByName, diskBySerialNumber);
                
        long rawBytesTotal = 0L;     // Aggregated from disks regardless of state or role
        long rawBytesUsed = 0L;      // Aggregated from disks regardless of state or role

        Set<String> localShelfIds = getLocalShelfIds(cModeCluster);

        for (final CModeDisk cModeDisk : cModeCluster.getDisks().values()) {
            final String diskUUID = cModeDisk.getKey().getUUID();
            final String diskHomeNodeName = cModeDisk.getHomeNodeName();
            final Node homeNode = nodesByName.get(diskHomeNodeName);
            final String ownerNodeName = cModeDisk.getOwnerNodeName();
            final Node ownerNode = nodesByName.get(ownerNodeName);
            if (ownerNode == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find the node UUID for disk owner name " + ownerNodeName
                            + " for disk " + cModeDisk.getDiskName());
                }
            }
            final String activeNodeName = cModeDisk.getActiveNodeName();
            final Node activeNode = nodesByName.get(activeNodeName);
            if (activeNode == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find the node UUID for disk active node name " + activeNodeName
                            + " for disk " + cModeDisk.getDiskName());
                }
            }

            AggregateSpareKey aggrSpareKey = new AggregateSpareKey(cModeDisk.getDiskName());
            CModeAggregateSpare aggrSpare = cModeCluster.getAggregateSpares().get(aggrSpareKey);
            final Disk disk = transformDisk(cluster, cModeCluster, homeNode, ownerNode, activeNode,
                    aggrsUUIDsByName, aggrSpare, cModeDisk, raidGroupKeys, storagePoolByName, localShelfIds);
            disk.setHomeNodeName(cModeDisk.getHomeNodeName());
            disk.setActiveNodeName(cModeDisk.getActiveNodeName());
            disk.setOwnerNodeName(cModeDisk.getOwnerNodeName());
            diskByName.put(disk.getName(), disk.getKey());
            diskBySerialNumber.put(disk.getSerialNumber(), disk.getKey());
            cluster.getDisks().add(disk);

            final Long diskRawBlocks = cModeDisk.getPhysicalBlocks();
            if (diskRawBlocks == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find total blocks for disk " + diskUUID);
                }
            } else {
                rawBytesTotal += diskRawBlocks * BYTES_PER_BLOCK;
            }
            final Long diskUsedBlocks = cModeDisk.getUsedBlocks();
            if (diskUsedBlocks == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find used blocks for disk " + diskUUID);
                }
            } else {
                rawBytesUsed += diskUsedBlocks * BYTES_PER_BLOCK;
            }
        }
        cluster.setRawBytesTotal(rawBytesTotal);
        cluster.setRawBytesUsed(rawBytesUsed);
        return diskKeyMapList;
    }

    private Map<Node.NodeKey, Map<String, NetworkPort.PortKey>> transformNetworkPorts(final CModeCluster cModeCluster, final Map<String, Node> nodesByName, Map<String, NetworkIpSpace.NetworkIpSpaceKey> ipSpaceNameToKeyMap, Map<String, NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey> broadcastDomainNameToKeyMap) {
        final Map<Node.NodeKey, Map<String, NetworkPort.PortKey>> portKeysByNameByNodeKey = new HashMap<Node.NodeKey, Map<String, NetworkPort.PortKey>>();
        for (final CModeNetInterface cModeNetInterface : cModeCluster.getNetInterfaces().values()) {
            final String nodeName = cModeNetInterface.getKey().getID(CModeNetInterface.IFKey.Field.node_name);
            final String interfaceName = cModeNetInterface.getKey().getID(CModeNetInterface.IFKey.Field.interface_name);
            if (!nodesByName.containsKey(nodeName)) {
                logger.info("{} - Failed to find node {} for port {}" ,new Object[]{ PackageContext.getLogPrefix(), nodeName, interfaceName});
                continue;
            }
            final Node node = nodesByName.get(nodeName);
            final Node.NodeKey nodeKey = node.getKey();
            if (!portKeysByNameByNodeKey.containsKey(nodeKey)) {
                final Map<String, NetworkPort.PortKey> nodePortKeysByName = new HashMap<String, NetworkPort.PortKey>();
                portKeysByNameByNodeKey.put(nodeKey, nodePortKeysByName);
            }

            final Map<String, NetworkPort.PortKey> nodePortKeysByName = portKeysByNameByNodeKey.get(nodeKey);
            final NetworkPort networkPort = transformNetInterface(node, cModeNetInterface);
            final NetworkPort.PortKey portKey = networkPort.getKey();
            nodePortKeysByName.put(portKey.getPortName(), portKey);          // Port Names are not unique, need to be qualified by Node
            final String ifgroupNodeName = cModeNetInterface.getIfGroupNode();
            if (!nodesByName.containsKey(ifgroupNodeName)) {
                if (logger.isDebugEnabled() && ifgroupNodeName != null) {
                    logger.debug("{} - Failed to find node {} for ifgroup port {}" , new Object[]{PackageContext.getLogPrefix(), ifgroupNodeName, interfaceName});
                }
            } else {
                final Node ifgroupNode = nodesByName.get(ifgroupNodeName);
                networkPort.setIfGroupNode(ifgroupNode.getKey());
            }
            final String vlanNodeName = cModeNetInterface.getVlanNode();
            if (!nodesByName.containsKey(vlanNodeName)) {
                if (logger.isDebugEnabled() && vlanNodeName != null) {
                    logger.debug("{} - Failed to find node {} for vlan port {}", new Object[]{PackageContext.getLogPrefix(), vlanNodeName, interfaceName});
                }
            } else {
                final Node vlanNode = nodesByName.get(vlanNodeName);
                networkPort.setVlanNode(vlanNode.getKey());
            }
            if (cModeNetInterface.getNetworkIpSpaceName() != null) {
                final NetworkIpSpace.NetworkIpSpaceKey ipSpaceKey = ipSpaceNameToKeyMap.get(cModeNetInterface.getNetworkIpSpaceName());
                if (ipSpaceKey != null) {
                    networkPort.setNetworkIpSpaceKey(ipSpaceKey);
                } else {
                    logger.info("Unable to find the IPSpace return with the net interface {}", interfaceName);
                }
            }
            if (cModeNetInterface.getNetworkPortBroadcastDomainName() != null) {
                final NetworkPortBroadcastDomain.NetworkPortBroadcastDomainKey broadcastDomainKey = broadcastDomainNameToKeyMap.get(cModeNetInterface.getNetworkPortBroadcastDomainName());
                if (broadcastDomainKey != null) {
                    networkPort.setNetPortBroadcastDomainKey(broadcastDomainKey);
                } else {
                    logger.info("Unable to find the broadcast domain return with the net interface {}", interfaceName);
                }
            }
            node.getNetworkPorts().add(networkPort);
        }
        return portKeysByNameByNodeKey;
    }

    private void transformNetworkPortRelationships(final Cluster cluster, final Map<Node.NodeKey, Map<String, NetworkPort.PortKey>> portKeysByNameByNodeKey) {
        for (final Node node : cluster.getNodes().values()) {
            for (final NetworkPort networkPort : node.getNetworkPorts().values()) {
                NetworkPort.PortKey ifGroupPortKey = null;
                final Node.NodeKey ifGroupNodeKey = networkPort.getIfGroupNode();
                final String ifgroupPortName = networkPort.getIfGroupPort();
                if (ifGroupNodeKey != null && ifgroupPortName != null && !ifgroupPortName.isEmpty()) {
                    final String nodeUUID = ifGroupNodeKey.getID(Node.NodeKey.Field.uuid);
                    final Map<String, NetworkPort.PortKey> portKeysByName = portKeysByNameByNodeKey.get(ifGroupNodeKey);
                    if (portKeysByName == null) {
                        if (logger.isDebugEnabled()) {
                            logger.debug(PackageContext.getLogPrefix() + " - Unknown node " + nodeUUID + " for network port group "
                                    + ifgroupPortName);
                        }
                    } else {
                        ifGroupPortKey = portKeysByName.get(ifgroupPortName);
                        if (ifGroupPortKey == null) {
                            if (logger.isDebugEnabled()) {
                                logger.debug(PackageContext.getLogPrefix() + " - Unknown ifgroup port " + ifgroupPortName + " for network port "
                                        + networkPort.getKey().getPortName());
                            }
                        }
                    }
                }

                NetworkPort.PortKey vlanPortKey = null;
                final Node.NodeKey vlanNodeKey = networkPort.getVlanNode();
                final String vlanPortName = networkPort.getVlanPort();
                if (vlanNodeKey != null && vlanPortName != null && !vlanPortName.isEmpty()) {
                    final String vlanNodeUUID = vlanNodeKey.getID(Node.NodeKey.Field.uuid);
                    final Map<String, NetworkPort.PortKey> portKeysByName = portKeysByNameByNodeKey.get(vlanNodeKey);
                    if (portKeysByName == null) {
                        if (logger.isDebugEnabled()) {
                            logger.debug(PackageContext.getLogPrefix() + " - Unknown node " + vlanNodeUUID + " for network port vlan "
                                    + vlanPortName);
                        }
                    } else {
                        vlanPortKey = portKeysByName.get(vlanPortName);
                        if (vlanPortKey == null) {
                            if (logger.isDebugEnabled()) {
                                logger.debug(PackageContext.getLogPrefix() + " - Unknown vlan port " + vlanPortName + " for network port "
                                        + networkPort.getKey().getPortName());
                            }
                        }
                    }
                }
                if (ifGroupPortKey == null && vlanPortKey == null) {
                    continue;
                }

                final NetworkPortRelationship networkPortRelationship = cluster.createNetworkPortRelationship(networkPort.getKey());
                networkPortRelationship.setIfGroupPortKey(ifGroupPortKey);
                networkPortRelationship.setVlanPortKey(vlanPortKey);
                cluster.getPortRelationships().add(networkPortRelationship);
            }
        }
    }

    private void transformFcpPorts(final CModeCluster cModeCluster, final Map<String, Node> nodesByName) {
        for (final CModeFcpAdapter cModeFcpAdapter : cModeCluster.getFcpAdapters().values()) {
            final String nodeName = cModeFcpAdapter.getKey().getID(CModeFcpAdapter.FCPKey.Field.node_name);
            if (!nodesByName.containsKey(nodeName)) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find node " + nodeName + " for fcp adapter " + cModeFcpAdapter.getKey());
                continue;
            }
            final Node node = nodesByName.get(nodeName);
            final FcpPort fcpPort = transformFcpAdapter(cModeCluster, node, cModeFcpAdapter);
            node.getFcpPorts().add(fcpPort);
        }
    }

    private void transformInitiatorPorts(final CModeCluster cModeCluster, final Map<String, Node> nodesByName) {
        for (final CModeAdapter cModeAdapter : cModeCluster.getCModeStorageAdapters().values()) {
            final String nodeName = cModeAdapter.getKey().getID(CModeAdapter.AdapterKey.Field.node_name);
            if (!nodesByName.containsKey(nodeName)) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find node " + nodeName + " for fcp adapter " + cModeAdapter.getKey());
                continue;
            }
            final Node node = nodesByName.get(nodeName);
            final StorageAdapter storageAdapter = transformStorageAdapter(cModeCluster, node, cModeAdapter);
            node.getStorageAdapters().add(storageAdapter);
        }
    }

    private Map<String, VServer.VServerKey> transformVServers(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Aggregate.AggregateKey> aggrKeysByName, Map<String, NetworkIpSpace.NetworkIpSpaceKey> ipSpaceNameToKeyMap) {
        // First transform the vServers
        final Map<String, VServer.VServerKey> vServerKeysByName = new HashMap<String, VServer.VServerKey>();
        for (final CModeVServer cModeVServer : cModeCluster.getVServers().values()) {
            final VServer vServer = transformVServer(cluster, cModeCluster, cModeVServer, aggrKeysByName, ipSpaceNameToKeyMap);
            vServerKeysByName.put(cModeVServer.getName(), vServer.getKey());
            cluster.getVServers().add(vServer);
        }

        // Next correlate NISInfo to VServers, sort the NIS infos by VServer
        for (final CModeVServerNISInfo cModeVServerNISInfo : cModeCluster.getVServerNISInfos().values()) {
            final String nisDomainName = cModeVServerNISInfo.getKey().getID(CModeVServerNISInfo.NISInfoKey.Field.nis_domain_name);
            final String vServerName = cModeVServerNISInfo.getKey().getID(CModeVServerNISInfo.NISInfoKey.Field.vserver_name);
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.info("{} - Failed to find vServer {} for NIS Service.", PackageContext.getLogPrefix(), vServerName);
                }
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            final Boolean isVServerNISActive = vServer.getNISActive();
            final NISDomain nisDomain = vServer.createNISDomain(nisDomainName);
            final Boolean nisDomainActive = cModeVServerNISInfo.isActive();

            // VServer NIS Domain active should be:
            // Null if no NIS servers configured
            // False if 1 or more NIS servers configured, but none active
            // True if 1 active NIS server configured
            if (nisDomainActive != null) {
                if (nisDomainActive == Boolean.TRUE) {
                    vServer.setNISActive(Boolean.TRUE);
                    vServer.setActiveNISDomainName(nisDomainName);
                    vServer.setActiveNISServerIPs(cModeVServerNISInfo.getNisServerIPs());
                } else {
                    // Found disabled NIS server.
                    // Set to NIS enabled to false only if not already set to true.
                    if (isVServerNISActive != Boolean.TRUE) {
                        vServer.setNISActive(Boolean.FALSE);
                    }
                }
            }
            nisDomain.setActive(cModeVServerNISInfo.isActive());
            nisDomain.setNisServerIPs(cModeVServerNISInfo.getNisServerIPs());
            vServer.getNisDomains().add(nisDomain);
        }
        return vServerKeysByName;
    }

    private void transformKerberos(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        final Map<VServer, Boolean> vServerKerberosStatusMap = new HashMap<VServer, Boolean>();
        for (final CModeKerbero cModeKerbero : cModeCluster.getKerberos().values()) {
            final String vServerName = cModeKerbero.getKey().getName();
            final VServer vServer = getVServerForChild(cluster, "kerberos", vServerName, vServerName, vServerKeysByName);
            if (vServer == null) {
                // already logged reason
                continue;
            }
            if (!vServerKerberosStatusMap.containsKey(vServer)) {
                vServerKerberosStatusMap.put(vServer, cModeKerbero.isKerberosEnabled());
            } else {
                // If any lif on the vserver has kerberos on, we want to set the flag.
                // Don't overwrite "true" with "false", only overwrite "false" with "true".
                if (cModeKerbero.isKerberosEnabled()) {
                    vServerKerberosStatusMap.put(vServer, Boolean.TRUE);
                }
            }
        }

        // If there is no kerberos entry for a vserver, we want to leave the vserver's kerberos flag null.
        // Only set the flag for vservers with a known kerberos enabled value.
        for (final Map.Entry<VServer, Boolean> vServerKerberosEntry : vServerKerberosStatusMap.entrySet()) {
            final VServer vServer = vServerKerberosEntry.getKey();
            final Boolean isKerberosEnabled = vServerKerberosEntry.getValue();
            vServer.setKerberosEnabled(isKerberosEnabled);
        }
    }

    private void transformFcpServices(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeFcpServiceInfo cModeFcpServiceInfo : cModeCluster.getFcpServiceInfos().values()) {
            final String vServerName = cModeFcpServiceInfo.getVServerName();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for FCP Service.");
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerKey.getResourceKey() + " for FCP Service.");
                continue;
            }
            vServer.setFcpAvailable(cModeFcpServiceInfo.isAvailable());
            vServer.setFcpNodeName(cModeFcpServiceInfo.getFcpNodeName());
        }
    }

    private void transformISCSIServices(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeISCSIService cModeISCSIService : cModeCluster.getIscsiServices().values()) {
            final String vServerName = cModeISCSIService.getVServerName();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for ISCSI Service.");
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerKey.getResourceKey() + " for ISCSI Service.");
                continue;
            }
            vServer.setISCSIAvailable(cModeISCSIService.isAvailable());
            vServer.setISCSIAliasName(cModeISCSIService.getAliasName());
            vServer.setISCSIIQN(cModeISCSIService.getKey().getUUID());
        }
    }

    private void transformCIFSServices(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeCIFSServer cModeCIFsServer : cModeCluster.getCifsServers().values()) {
            final String vServerName = cModeCIFsServer.getVServerName();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for CIFS Service.");
                }
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerKey.getResourceKey() + " for CIFS Service.");
                continue;
            }
            vServer.setCifsServer(cModeCIFsServer.getName());
            vServer.setCifsAuthStyle(cModeCIFsServer.getAuthSytle().getEnumValue());
            vServer.setCifsAuthStyleRaw(cModeCIFsServer.getAuthSytle().getRawValue());
            vServer.setCifsDomain(cModeCIFsServer.getDomain());
            vServer.setCifsDomainWorkgroup(cModeCIFsServer.getDomainWorkgroup());
            final EnumWrapper<VServer.CifsAdminStatus> cifsServerStatusWrapper = cModeCIFsServer.getAdminStatus();
            // For RR, CIFS Admin status is not defined, so if we discover a CIFS server we will report it as enabled.
            // For SN, there is an explicit attribute.
            final boolean isCifsEnabled = (cifsServerStatusWrapper.getRawValue() == null)
                    || VServer.CifsAdminStatus.UP == cifsServerStatusWrapper.getEnumValue();
            vServer.setCifsEnabled(isCifsEnabled);
        }
    }

    private void transformCIFSOptions(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeCIFSOption cModeCIFSOption : cModeCluster.getCifsOptions().values()) {
            final String vServerName = cModeCIFSOption.getVserverName();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for CIFS Service.");
                }
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerKey.getResourceKey() + " for CIFS Service.");
                continue;
            }
            vServer.setExportPolicyEnabled(cModeCIFSOption.getExportPolicyEnabled());
            vServer.setDefaultUnixUser(cModeCIFSOption.getDefaultUnixUser());
            vServer.setDefaultUnixGroup(cModeCIFSOption.getDefaultUnixGroup());
        }
    }

    private void transformNFSServices(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeNFSService nfsService : cModeCluster.getNFSServices().values()) {
            final String vServerName = nfsService.getKey().getUUID();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for NFS Service.");
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerKey.getResourceKey() + " for NFS Service.");
                continue;
            }
            vServer.setNFSAccessEnabled(nfsService.isNFSAccessEnabled());
            vServer.setNFSV2Enabled(nfsService.isNFSV2Enabled());
            vServer.setNFSV3Enabled(nfsService.isNFSV3Enabled());
            vServer.setNFSV40Enabled(nfsService.isNFSV40Enabled());
            vServer.setNFS40ReadDelegationEnabled(nfsService.isNFS40ReadDelegationEnabled());
            vServer.setNFS40WriteDelegationEnabled(nfsService.isNFS40WriteDelegationEnabled());
            vServer.setNFS40MigrationEnabled(nfsService.isNFS40MigrationEnabled());
            vServer.setNFS40ReferralsEnabled(nfsService.isNFS40ReferralsEnabled());
            vServer.setNFS41AclEnabled(nfsService.isNFS41AclEnabled());
            vServer.setNFS41MigrationEnabled(nfsService.isNFS41MigrationEnabled());
            vServer.setNFS41PnfsEnabled(nfsService.isNFS41PnfsEnabled());
            vServer.setNFS41ReferralsEnabled(nfsService.isNFS41ReferralsEnabled());
            vServer.setNFSV41Enabled(nfsService.isNFSV41Enabled());
        }
    }

    private Map<String, JobSchedule> transformJobSchedules(final CModeCluster cModeCluster, final Cluster cluster) {
        final Map<String, JobSchedule> jobSchedulesByName = new HashMap<String, JobSchedule>();
        for (final CModeJobSchedule cModeJobSchedule : cModeCluster.getJobSchedules().values()) {
            final String jobScheduleName = cModeJobSchedule.getKey().getName();
            final JobSchedule jobSchedule = cluster.createJobSchedule(jobScheduleName);
            final EnumWrapper<JobSchedule.ScheduleType> scheduleType = cModeJobSchedule.getJobScheduleType();
            jobSchedule.setScheduleType(scheduleType.getEnumValue());
            jobSchedule.setScheduleTypeRaw(scheduleType.getRawValue());
            jobSchedule.setDescription(cModeJobSchedule.getDescription());
            jobSchedulesByName.put(jobScheduleName, jobSchedule);
            cluster.getJobSchedules().add(jobSchedule);
        }
        for (final CModeCronJobSchedule cModeCronJobSchedule : cModeCluster.getCronJobSchedules().values()) {
            final String jobScheduleName = cModeCronJobSchedule.getKey().getName();
            if (!jobSchedulesByName.containsKey(jobScheduleName)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find the job schedule for cron job schedule " + jobScheduleName);
                continue;
            }
            final JobSchedule jobSchedule = jobSchedulesByName.get(jobScheduleName);
            jobSchedule.setCronDayOfMonth(cModeCronJobSchedule.getCronDayOfMonthCSV());
            jobSchedule.setCronDayOfWeek(cModeCronJobSchedule.getCronDayOfWeekCSV());
            jobSchedule.setCronHour(cModeCronJobSchedule.getCronHourCSV());
            jobSchedule.setCronMinute(cModeCronJobSchedule.getCronMinuteCSV());
            jobSchedule.setCronMonth(cModeCronJobSchedule.getCronMonthCSV());
        }
        for (final CModeJobScheduleInterval cModeJobScheduleInterval : cModeCluster.getJobScheduleIntervals().values()) {
            final String jobScheduleName = cModeJobScheduleInterval.getKey().getName();
            if (!jobSchedulesByName.containsKey(jobScheduleName)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find the job schedule for job schedule interval" + jobScheduleName);
                continue;
            }
            final JobSchedule jobSchedule = jobSchedulesByName.get(jobScheduleName);
            jobSchedule.setIntervalDays(cModeJobScheduleInterval.getDays());
            jobSchedule.setIntervalHours(cModeJobScheduleInterval.getHours());
            jobSchedule.setIntervalMinutes(cModeJobScheduleInterval.getMinutes());
            jobSchedule.setIntervalSeconds(cModeJobScheduleInterval.getSeconds());
        }
        return jobSchedulesByName;
    }

    private Map<String, SnapshotPolicy.SnapshotPolicyKey> transformSnapshotPolicies(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<String, JobSchedule> jobSchedulesByName) {
        final Map<String, SnapshotPolicy.SnapshotPolicyKey> snapshotPolicyKeysByName = new HashMap<String, SnapshotPolicy.SnapshotPolicyKey>();
        final String clusterUUID = cluster.getClusterUUID();
        for (final CModeSnapshotPolicy cModeSnapshotPolicy : cModeCluster.getSnapshotPolicies().values()) {
            final String policyName = cModeSnapshotPolicy.getKey().getID(CModeSnapshotPolicy.SnapshotPolicyKey.Field.policy_name);
            final String vServerName = cModeSnapshotPolicy.getVServerName();
            // Snapshot Policy may be Cluster or VServer scope.
            //
            final VServer.VServerKey vServerKey;
            if (vServerName != null) {
                if (!vServerKeysByName.containsKey(vServerName)) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for Snapshot Policy " + policyName);
                    continue;
                }
                vServerKey = vServerKeysByName.get(vServerName);
                if (!cluster.getVServers().containsKey(vServerKey)) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for Snapshot Policy " + policyName);
                    continue;
                }
                final String vServerUUID = vServerKey.getZapiUUID();
                final SnapshotPolicy.SnapshotPolicyKey policyKey = new SnapshotPolicy.SnapshotPolicyKey(clusterUUID, vServerUUID, policyName);
                final SnapshotPolicyVserverRelationship relationship = cluster.createSnapshotPolicyVserverRelationship(policyKey, vServerKey);
                cluster.getSnapshotPolicyVserverRelationships().add(relationship);
            } else {
                vServerKey = null;
            }

            final SnapshotPolicy snapshotPolicy = cluster.createSnapshotPolicy(vServerKey, policyName);
            snapshotPolicy.setComment(cModeSnapshotPolicy.getComment());
            snapshotPolicy.setEnabled(cModeSnapshotPolicy.isEnabled());
            cluster.getSnapshotPolicies().add(snapshotPolicy);
            snapshotPolicyKeysByName.put(policyName, snapshotPolicy.getKey());

            for (final CModeSnapshotSchedule cModeSnapshotSchedule : cModeSnapshotPolicy.getCModeSnapshotSchedules().values()) {
                final String scheduleName = cModeSnapshotSchedule.getKey().getName();
                final SnapshotPolicySchedule snapshotPolicySchedule = snapshotPolicy.createSnapshotPolicySchedule(scheduleName);
                if (jobSchedulesByName.containsKey(scheduleName)) {
                    final JobSchedule jobSchedule = jobSchedulesByName.get(scheduleName);
                    snapshotPolicySchedule.setJobScheduleKey(jobSchedule.getKey());
                }
                snapshotPolicySchedule.setVServerKey(vServerKey);
                snapshotPolicySchedule.setCount(cModeSnapshotSchedule.getCount());
                snapshotPolicySchedule.setPrefix(cModeSnapshotSchedule.getPrefix());
                snapshotPolicySchedule.setSnapmirrorLabel(cModeSnapshotSchedule.getSnapmirrorLabel());
                snapshotPolicy.getSnapshotPolicySchedules().add(snapshotPolicySchedule);
            }
        }

        // Now update VServers with SnapshotPolicy references
        for (final CModeVServer cModeVServer : cModeCluster.getVServers().values()) {
            final String snapshotPolicyName = cModeVServer.getSnapshotPolicy();
            if (snapshotPolicyName == null) {
                continue;
            }
            if (!snapshotPolicyKeysByName.containsKey(snapshotPolicyName)) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find snapshot policy " + snapshotPolicyName +
                            " for vServer " + cModeVServer.getName());
                }
                continue;
            }
            final SnapshotPolicy.SnapshotPolicyKey policyKey = snapshotPolicyKeysByName.get(snapshotPolicyName);
            final UUIDKey cModeVServerKey = cModeVServer.getKey();
            final VServer.VServerKey vServerKey = new VServer.VServerKey(cluster.getClusterUUID(), cModeVServerKey.getUUID());
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to vServer " + cModeVServer.getName() +
                            " for snapshot policy " + snapshotPolicyName);
                }
                continue;
            }
            vServer.setSnapshotPolicyKey(policyKey);
        }
        return snapshotPolicyKeysByName;
    }

    private Map<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>> transformExportPoliciesAndRules(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        // Export Policies & Rules
        // Iterate over the discovered Export Policies
        final Map<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>> exportPolicyKeysByNameByVServerKey = new HashMap<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>>();
        for (final CModeExportPolicy cModeExportPolicy : cModeCluster.getExportPolicies().values()) {

            // Find the VServer that this policy belongs to
            final String vServerName = cModeExportPolicy.getVServerName();
            if (!vServerKeysByName.containsKey(vServerName)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find the vserver UUID for name " + vServerName);
                continue;
            }

            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (!cluster.getVServers().containsKey(vServerKey)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find the vserver with UUID " + vServerKey.getResourceKey());
                continue;
            }

            // Get the Vserver
            final VServer parentVserver = cluster.getVServers().get(vServerKey);

            // Create the Export policy
            final ExportPolicy exportPolicy = transformExportPolicy(parentVserver, cModeExportPolicy);

            // Iterate over the discovered Export Rules and find ones that belong to this policy/vserver combination
            for (final CModeExportRule cModeExportRule : cModeCluster.getExportRules().values()) {

                // If the rule's policy ID and Vserver Name match the current, create and add
                if (cModeExportRule.getPolicyName().equals(cModeExportPolicy.getName()) && cModeExportRule.getVServerName().equals(vServerName)) {

                    final ExportRule exportRule = transformExportRule(exportPolicy, parentVserver, cModeExportRule);
                    exportPolicy.getExportRules().add(exportRule);
                }
            }

            // Add the export policy to the vserver
            parentVserver.getExportPolicies().add(exportPolicy);

            if (exportPolicyKeysByNameByVServerKey.containsKey(parentVserver.getKey())) {
                exportPolicyKeysByNameByVServerKey.get(parentVserver.getKey()).put(cModeExportPolicy.getName(), exportPolicy.getKey());
            } else {
                exportPolicyKeysByNameByVServerKey.put(parentVserver.getKey(), new HashMap<String, ExportPolicy.ExportPolicyKey>() {{
                    put(cModeExportPolicy.getName(), exportPolicy.getKey());
                }});
            }
        }
        return exportPolicyKeysByNameByVServerKey;
    }

    private Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> transformVolumes(final CModeCluster cModeCluster, final Cluster cluster,
                                                                                    final Map<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>> exportPolicyKeysByNameByVServerKey,
                                                                                    final Map<Volume, String> volumeToCloneParentName,
                                                                                    final Map<String, SnapshotPolicy.SnapshotPolicyKey> snapshotPolicyKeysByName,
                                                                                    final Map<String, QosPolicyGroup.QosPolicyGroupKey> qosPolicyGroupKeysByName,
                                                                                    final Map<String, Node> nodesByName) {
        // Volumes
        final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey = new HashMap<VServer.VServerKey, Map<String, Volume.VolumeKey>>();
        for (final CModeVolume cModeVolume : cModeCluster.getVolumes().values()) {
            final String volumeNameForDebug = cModeVolume.getVolumeName();
            final Volume.VolumeType type = cModeVolume.getVolumeType().getEnumValue();
            if (Volume.VolumeType.TMP == type) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Skipping temp volume " + volumeNameForDebug);
                }
                continue;
            }
            final String volumeInstanceUUID = cModeVolume.getInstanceUUID();
            final String owningVServerUUID = cModeVolume.getVServerUUID();
            if (owningVServerUUID == null) {
                logger.info(PackageContext.getLogPrefix() + " - Undefined to owning vServer for volume " + volumeNameForDebug);
                continue;
            }
            final VServer.VServerKey parentVServerKey = new VServer.VServerKey(cModeCluster.getClusterUUID(), owningVServerUUID);

            final VServer parentVServer = cluster.getVServers().get(parentVServerKey);
            if (parentVServer == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + parentVServerKey + " for volume " + volumeNameForDebug);
                }
                continue;
            }
            final String aggregateUUID = cModeVolume.getAggregateUUID();
            if (aggregateUUID == null) {
                // Infinivols don't have a specific aggregate.
                // If the aggregate is undefined, this better be an infinivol
                final Volume.VolumeStyle volumeStyle = cModeVolume.getVolumeStyle().getEnumValue();
                if (Volume.VolumeStyle.INFINITEVOL != volumeStyle) {
                    if (logger.isDebugEnabled()) {
                        // This is unexpected, log the error when debug is enabled
                        logger.error(PackageContext.getLogPrefix() + " - Undefined aggregate for non-infinivol volume " + volumeNameForDebug);
                    }
                    continue;   // Skip the volume because it is not infinivol and it has no aggregate
                } // else Infinivols are allowed to have no aggregate
            } else {
                // Aggregate UUID is defined, make sure it corresponds to a valid aggregate
                final UUIDKey aggregateKey = new UUIDKey("aggregate", aggregateUUID);
                final CModeAggregate cModeAggregate = cModeCluster.getAggregates().get(aggregateKey);
                if (cModeAggregate == null) {
                    if (logger.isDebugEnabled()) {
                        logger.error(PackageContext.getLogPrefix() + " - Undefined aggregate " + aggregateUUID + " for volume " + volumeNameForDebug);
                    }
                    continue;       // Skip the volume because its aggregate is unknown
                }
            }

            ExportPolicy.ExportPolicyKey exportPolicyKey = null;
            if (exportPolicyKeysByNameByVServerKey.containsKey(parentVServer.getKey())) {
                exportPolicyKey = exportPolicyKeysByNameByVServerKey.get(parentVServer.getKey()).get(cModeVolume.getExportPolicy());
            }

            if (exportPolicyKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Undefined export policy " + cModeVolume.getExportPolicy() + " for volume " + volumeNameForDebug);
                }
                // Not a required attribute so continue building this instance
            }

            UUIDKey key = cModeVolume.getKey();
            final String uuid = key.getID( UUIDKey.Field.uuid);
            if (StringUtils.startsWith(uuid, CModeVolume.CREATED_VOLUME_UUID_PREFIX)) {
                DoNotDelete doNotDelete =  cluster.createDoNotDelete(UUID.randomUUID().toString());
                doNotDelete.setObjectTagName("Volume");
                AlternateKey alternateKey = doNotDelete.createAlternateKey(UUID.randomUUID().toString());
                alternateKey.setName("name");
                alternateKey.setValue(cModeVolume.getVolumeName());
                doNotDelete.getAlternateKeys().add(alternateKey);
                alternateKey = doNotDelete.createAlternateKey(UUID.randomUUID().toString());
                alternateKey.setName("vServerUUID");
                alternateKey.setValue(cModeVolume.getVServerUUID());
                doNotDelete.getAlternateKeys().add(alternateKey);
                cluster.getDoNotDeleteList().add(doNotDelete);
                continue;
            }

            if (volumeInstanceUUID == null) {
                // We only allow this condition for infinivols
                // If style is NOT infinivol, then don't report the volume.
                if (!Volume.VolumeStyle.INFINITEVOL.equals(cModeVolume.getVolumeStyle().getEnumValue())) {
                    // Don't report this volume
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Undefined instance UUID for non-Infinivol volume " + volumeNameForDebug + ", ignoring.");

                    }
                    continue;
                } else {
                    // This is infinivol, set instance UUID to UUID
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Undefined instance UUID for Infinivol volume " + volumeNameForDebug + ", using uuid.");
                    }
                    cModeVolume.setInstanceUUID(cModeVolume.getKey().getUUID());
                }
            }
            final Volume volume = transformVolume(cluster, parentVServer, cModeVolume, exportPolicyKey, snapshotPolicyKeysByName, qosPolicyGroupKeysByName);
            final Volume.VolumeKey volumeKey = volume.getKey();
            if (volumeKeysByNameByVServerKey.containsKey(parentVServer.getKey())) {
                volumeKeysByNameByVServerKey.get(parentVServer.getKey()).put(cModeVolume.getVolumeName(), volumeKey);
            } else {
                volumeKeysByNameByVServerKey.put(parentVServer.getKey(), new HashMap<String, Volume.VolumeKey>() {{
                    put(cModeVolume.getVolumeName(), volumeKey);
                }});
            }


            if (cModeVolume.isClone()) {
                volumeToCloneParentName.put(volume, cModeVolume.getCloneParentName());
            }

            parentVServer.getVolumes().add(volume);
        }

        return volumeKeysByNameByVServerKey;
    }


    private void transformVolumeRelationships(final CModeCluster cModeCluster, final Cluster cluster, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {
        for (final CModeVolume cModeVolume : cModeCluster.getVolumes().values()) {
            final String junctionParentName = cModeVolume.getJunctionParentName();
            if (junctionParentName == null) {
                continue;
            }
            final String volumeName = cModeVolume.getVolumeName();
            final String vServerUUID = cModeVolume.getVServerUUID();
            final VServer.VServerKey vServerKey = new VServer.VServerKey(cModeCluster.getClusterUUID(), vServerUUID);
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerUUID + " for volume " + volumeName);
                }
                continue;
            }
            final Map<String, Volume.VolumeKey> volumeKeysByName = volumeKeysByNameByVServerKey.get(vServerKey);
            if (volumeKeysByName == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerUUID + " for volume " + volumeName);
                }
                continue;
            }

            final String volumeUUID = cModeVolume.getInstanceUUID();
            final Volume volume = getVolumeForChild(vServer, "volume", volumeName, volumeName, volumeKeysByNameByVServerKey);
            if (volume == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find Volume " + volumeName + " " + volumeUUID);
                }
                continue;   // already logged reason
            }

            final Volume.VolumeKey junctionParentVolumeKey = volumeKeysByName.get(junctionParentName);
            if (junctionParentVolumeKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find junction parent volume " + junctionParentName + " for volume " + volumeName);
                }
                continue;
            }
            final Volume junctionParentVolume = vServer.getVolumes().get(junctionParentVolumeKey);
            if (junctionParentVolume == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find junction parent volume " + junctionParentName + " for volume " + volumeName);
                }
                continue;
            }

            final VolumeSelfRelationship volumeRelationship = getVolumeSelfRelationship(volumeUUID, vServer);
            final String junctionParentUUID = junctionParentVolume.getInstanceUUID();
            volumeRelationship.setVolumeJunctionParentUUID(junctionParentUUID);
        }
    }

    private void transformVolumeClones(final CModeCluster cModeCluster, final Cluster cluster,
                                       final Map<String, VServer.VServerKey> vServerKeysByName,
                                       final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey,
                                       final Map<Volume, String> volumeToCloneParentName) {
        // process volume clone entries for clone split estimate
        for (final CModeVolumeClone cModeVolumeClone : cModeCluster.getVolumeClones().values()) {
            final String vServerName = cModeVolumeClone.getKey().getID(VolumeNameKey.Field.vserver_name);
            final String volumeName = cModeVolumeClone.getKey().getID(VolumeNameKey.Field.volume_name);

            final VServer vServer = getVServerForChild(cluster, "volume", volumeName, vServerName, vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason
            }

            final Volume volume = getVolumeForChild(vServer, "volume", volumeName, volumeName, volumeKeysByNameByVServerKey);
            if (volume == null) {
                continue;   // already logged reason
            }
            final Long cloneSplitEstimateBlocks = cModeVolumeClone.getCloneSplitEstimate();
            if (cloneSplitEstimateBlocks != null) {
                volume.setCloneSplitEstimate(cloneSplitEstimateBlocks * BYTES_PER_BLOCK);
            }
        }

        // process volume clones parent/child relationships
        for (final Map.Entry<Volume, String> entry : volumeToCloneParentName.entrySet()) {
            final Volume clone = entry.getKey();
            final String cloneParentName = entry.getValue();

            final VServer vserver = clone.getParent();

            final Volume.VolumeKey cloneParentKey = volumeKeysByNameByVServerKey.get(vserver.getKey()).get(cloneParentName);
            if (cloneParentKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Undefined clone parent " + cloneParentName + " for volume " + clone.getVolumeName());
                }
                continue;
            }

            final String cloneUUID = clone.getInstanceUUID();
            final VolumeSelfRelationship volumeRelationship = getVolumeSelfRelationship(cloneUUID, vserver);

            final Volume cloneParent = vserver.getVolumes().get(cloneParentKey);
            final String volumeCloneParentUUID = cloneParent.getInstanceUUID();
            volumeRelationship.setVolumeCloneParentUUID(volumeCloneParentUUID);
        }
    }

    /**
     * Get or create the volume self relationship for the target volume
     *
     * @param volumeUUID the UUID of the volume
     * @param vServer    the vServer hosting the target UUID
     * @return The volume self relationship object for the specified volume UUID on the specified vServer.
     *         Created as needed so it never returns <tt>null</tt>.
     */
    private VolumeSelfRelationship getVolumeSelfRelationship(final String volumeUUID, final VServer vServer) {
        final String clusterUUID = vServer.getParent().getClusterUUID();
        final VolumeSelfRelationship.RelationshipKey volumeSelfRelationshipKey =
                new VolumeSelfRelationship.RelationshipKey(clusterUUID, volumeUUID);
        VolumeSelfRelationship volumeRelationship = vServer.getVolumeSelfRelationships().get(volumeSelfRelationshipKey);
        if (volumeRelationship == null) {
            volumeRelationship = vServer.createVolumeSelfRelationship(volumeUUID);
            vServer.getVolumeSelfRelationships().add(volumeRelationship);
        }
        return volumeRelationship;
    }

    private void transformVolumeFlexCaches(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {

        for (final CModeVolumeFlexCache cModeVolumeFlexCache : cModeCluster.getVolumeFlexCaches().values()) {
            final String vServerName = cModeVolumeFlexCache.getKey().getID(VolumeNameKey.Field.vserver_name);
            final String flexCacheVolumeName = cModeVolumeFlexCache.getKey().getID(VolumeNameKey.Field.volume_name);
            final VServer vServer = getVServerForChild(cluster, "flex cache volume", flexCacheVolumeName, vServerName, vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason
            }

            final Volume cacheVolume = getVolumeForChild(vServer, "flex cache volume", flexCacheVolumeName, flexCacheVolumeName, volumeKeysByNameByVServerKey);
            if (cacheVolume == null) {
                continue;   // already logged reason
            }

            final String originVolumeName = cModeVolumeFlexCache.getOriginVolumeName();
            final Volume originVolume = getVolumeForChild(vServer, "flex cache volume origin", flexCacheVolumeName, originVolumeName, volumeKeysByNameByVServerKey);
            if (originVolume == null) {
                continue;   // already logged reason
            }

            final VolumeFlexCache volumeFlexCache = vServer.createVolumeFlexCache(cacheVolume.getInstanceUUID());
            volumeFlexCache.setFlexCacheOriginKey(originVolume.getKey());
            vServer.getVolumeFlexCaches().add(volumeFlexCache);
        }
    }

    private void transformVolumeSisStatus(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey, final Map<VServer.VServerKey, Map<String, SisPolicy.SisPolicyKey>> sisPolicyKeysByNameByVServerKey) {
        for (final CModeSisStatus sisStatus : cModeCluster.getSisStatuses().values()) {
            final String vServerName = sisStatus.getKey().getVServerName();
            final String volumePath = sisStatus.getKey().getVolumePath();
            final VServer vServer = getVServerForChild(cluster, "volume", volumePath, vServerName, vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason
            }
            final String volumeName = volumePath.substring("/vol/".length());
            final Volume volume = getVolumeForChild(vServer, "volume", volumeName, volumeName, volumeKeysByNameByVServerKey);
            if (volume == null) {
                continue;   // already logged reason
            }

            final String sisPolicyName = sisStatus.getSisPolicyName();
            if (sisPolicyName != null) {
                final Map<String, SisPolicy.SisPolicyKey> sisPolicyKeysByName = sisPolicyKeysByNameByVServerKey.get(vServer.getKey());
                if (sisPolicyKeysByName != null) {
                    final SisPolicy.SisPolicyKey sisPolicyKey = sisPolicyKeysByName.get(sisPolicyName);
                    volume.setSisPolicyKey(sisPolicyKey);
                }
            }

            volume.setSisStatus(sisStatus.getStatus().getEnumValue());
            volume.setSisStatusRaw(sisStatus.getStatus().getRawValue());
            volume.setSisState(sisStatus.getState().getEnumValue());
            volume.setSisStateRaw(sisStatus.getState().getRawValue());
            volume.setSisType(sisStatus.getType().getEnumValue());
            volume.setSisTypeRaw(sisStatus.getType().getRawValue());
            volume.setProgress(sisStatus.getProgress());
            volume.setSchedule(sisStatus.getSchedule());
            volume.setLastSisOpState(sisStatus.getLastOpState().getEnumValue());
            volume.setLastSisOpStateRaw(sisStatus.getLastOpState().getRawValue());
            volume.setLastOpBeginTS(sisStatus.getLastOpBeginTS());
            volume.setLastOpEndTS(sisStatus.getLastOpEndTS());
            volume.setLastOpError(sisStatus.getLastOpError());
            volume.setLastSisOpSize(sisStatus.getLastOpSize());
            volume.setCompressionEnabled(sisStatus.isCompressionEnabled());
            volume.setInlineCompressionEnabled(sisStatus.isInlineCompressionEnabled());
        }
    }

    private void transformVolumeMoves(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey, final Map<String, Aggregate.AggregateKey> aggrKeysByName) {

        for (final CModeVolumeMove cModeVolumeMove : cModeCluster.getVolumeMoves().values()) {
            final String volumeName = cModeVolumeMove.getVolume();
            final String vServerName = cModeVolumeMove.getVServer();
            final VServer vServer = getVServerForChild(cluster, "volume move", volumeName, vServerName, vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason
            }
            final Volume volume = getVolumeForChild(vServer, "volume move", volumeName, volumeName, volumeKeysByNameByVServerKey);
            if (volume == null) {
                continue;   // already logged reason
            }

            // Sometimes this is a Name and sometimes a UUID
            final String sourceAggrUUIDOrName = cModeVolumeMove.getSourceAggregate();
            final Aggregate.AggregateKey sourceAggregateKey = getAggrUUIDByNameOrUUID(sourceAggrUUIDOrName, aggrKeysByName);
            if (sourceAggregateKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find sourceAggregate UUID for aggregate " + sourceAggrUUIDOrName + " while transforming volume move object for volume " + volumeName);
                }
                continue;
            }

            final String destAggrUUIDOrName = cModeVolumeMove.getDestinationAggregate();
            final Aggregate.AggregateKey destAggregateKey = getAggrUUIDByNameOrUUID(destAggrUUIDOrName, aggrKeysByName);
            if (destAggregateKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find destinationAggregate UUID for aggregate name " + destAggrUUIDOrName + " while transforming volume move object for volume " + volumeName);
                }
                continue;
            }

            final VolumeMove volumeMove = vServer.createVolumeMove(cModeVolumeMove.getKey().getUUID());

            volumeMove.setCompletionTime(cModeVolumeMove.getActualCompletionTimestamp());
            volumeMove.setDestinationAggregateKey(destAggregateKey);
            volumeMove.setDetails(cModeVolumeMove.getDetails());
            volumeMove.setJobIdentifier(cModeVolumeMove.getJobID());
            volumeMove.setPercentComplete(cModeVolumeMove.getPercentComplete());
            volumeMove.setPhaseRaw(cModeVolumeMove.getPhaseRaw());
            volumeMove.setPhaseEnum(cModeVolumeMove.getPhaseEnum());
            volumeMove.setSourceAggregateKey(sourceAggregateKey);
            volumeMove.setStartTime(cModeVolumeMove.getStartTimestamp());
            volumeMove.setEstimatedCompletionTime(cModeVolumeMove.getEstimatedCompletionTime());
            volumeMove.setState(cModeVolumeMove.getStateEnum());
            volumeMove.setStateRaw(cModeVolumeMove.getStateRaw());
            volumeMove.setVolumeKey(volume.getKey());

            vServer.getVolumeMoves().add(volumeMove);
        }
    }

    @SuppressWarnings("UnusedDeclaration")
    private void transformVolumeFootprints(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {
        // Volume Block Footprints disabled with DE5038, leaving infrastructure in place in case
        // Zapi performance is improved and need for volume blocks footprint returns.
        for (final CModeVolumeFootprint cModeVolumeFootprint : cModeCluster.getVolumeFootprints().values()) {
            final VolumeNameKey volumeNameKey = cModeVolumeFootprint.getKey();
            final String vServerName = volumeNameKey.getID(VolumeNameKey.Field.vserver_name);
            final String volumeName = volumeNameKey.getID(VolumeNameKey.Field.volume_name);

            final VServer vServer = getVServerForChild(cluster, "volume footprints", volumeName, vServerName, vServerKeysByName);
            if (vServer != null) {
                final Volume volume = getVolumeForChild(vServer, "volume footprints", volumeName, volumeName, volumeKeysByNameByVServerKey);
                if (volume != null) {
                    volume.setVolumeBlocksFootPrint(cModeVolumeFootprint.getVolumeBlocksFootPrint());
                }
            }
        }
    }


    private void transformQuotaStatuses(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {
        for (final CModeQuotaStatus cModeQuotaStatus : cModeCluster.getQuotaStatuses().values()) {
            final String vServerName = cModeQuotaStatus.getvServerName();
            final String volumeName = cModeQuotaStatus.getVolumeName();
            final VServer vServer = getVServerForChild(cluster, "quota status", volumeName, vServerName, vServerKeysByName);
            if (vServer != null) {
                final Volume volume = getVolumeForChild(vServer, "quota status", volumeName, volumeName, volumeKeysByNameByVServerKey);
                if (volume != null) {
                    volume.setQuotaStatus(cModeQuotaStatus.getStatus().getEnumValue());
                    volume.setQuotaStatusRaw(cModeQuotaStatus.getStatus().getRawValue());
                }
            }
        }
    }


    private Aggregate.AggregateKey getAggrUUIDByNameOrUUID(final String aggrNameOrUUID, final Map<String, Aggregate.AggregateKey> aggrKeysByName) {

        if (aggrKeysByName.containsKey(aggrNameOrUUID)) {
            return aggrKeysByName.get(aggrNameOrUUID);
        }


        // Doesn't look like a name, so it must be a UUID.
        // Make sure the UUID corresponds to a known aggregate
        for (final Aggregate.AggregateKey candidateKey : aggrKeysByName.values()) {
            if (candidateKey.getOcieUUID().equals(aggrNameOrUUID)) {
                return candidateKey;
            }
        }

        // not found
        return null;
    }

    private Map<VServer.VServerKey, Map<String, SisPolicy.SisPolicyKey>> transformSisPolicies(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<String, JobSchedule> jobSchedulesByName) {
        final Map<VServer.VServerKey, Map<String, SisPolicy.SisPolicyKey>> sisPolicyKeysByNameByVServer = new HashMap<VServer.VServerKey, Map<String, SisPolicy.SisPolicyKey>>();
        for (final CModeSisPolicy cModeSisPolicy : cModeCluster.getSisPolicies().values()) {
            final String sisPolicyName = cModeSisPolicy.getKey().getID(CModeSisPolicy.SisPolicyKey.Field.sis_policy_name);
            final String vServerName = cModeSisPolicy.getKey().getID(CModeSisPolicy.SisPolicyKey.Field.vserver_name);
            if (!vServerKeysByName.containsKey(vServerName)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for SIS Policy " + sisPolicyName);
                continue;
            }
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (!cluster.getVServers().containsKey(vServerKey)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find vServer " + vServerName + " for SIS Policy " + sisPolicyName);
                continue;
            }

            final VServer vServer = cluster.getVServers().get(vServerKey);
            final SisPolicy sisPolicy = vServer.createSisPolicy(vServerKey.getZapiUUID(), sisPolicyName);
            sisPolicy.setComment(cModeSisPolicy.getComment());
            sisPolicy.setDuration(cModeSisPolicy.getDuration());
            sisPolicy.setEnabled(cModeSisPolicy.isEnabled());
            sisPolicy.setQosPolicy(cModeSisPolicy.getQosPolicy().getEnumValue());
            sisPolicy.setQosPolicyRaw(cModeSisPolicy.getQosPolicy().getRawValue());
            if (!sisPolicyKeysByNameByVServer.containsKey(vServerKey)) {
                final Map<String, SisPolicy.SisPolicyKey> sisPoliciesByName = new HashMap<String, SisPolicy.SisPolicyKey>();
                sisPolicyKeysByNameByVServer.put(vServerKey, sisPoliciesByName);
            }
            final Map<String, SisPolicy.SisPolicyKey> sisPoliciesByName = sisPolicyKeysByNameByVServer.get(vServerKey);
            sisPoliciesByName.put(sisPolicyName, sisPolicy.getKey());
            final String scheduleName = cModeSisPolicy.getSchedule();
            if (scheduleName != null) {
                if (jobSchedulesByName.containsKey(scheduleName)) {
                    final JobSchedule jobSchedule = jobSchedulesByName.get(scheduleName);
                    sisPolicy.setJobScheduleKey(jobSchedule.getKey());
                }
            }
            vServer.getSisPolicies().add(sisPolicy);
        }
        return sisPolicyKeysByNameByVServer;
    }

    private Map<VServer.VServerKey, Map<String, QTree.QTreeKey>> transformQTrees(final CModeCluster cModeCluster, final Cluster cluster,
                                                                                 final Map<String, VServer.VServerKey> vServerKeysByName,
                                                                                 final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey,
                                                                                 final Map<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>> exportPolicyKeysByNameByVServerKey) {
        // QTrees
        final Map<VServer.VServerKey, Map<String, QTree.QTreeKey>> qTreeKeysByAliasByVServerKey = new HashMap<VServer.VServerKey, Map<String, QTree.QTreeKey>>();
        for (final CModeQTree cModeQTree : cModeCluster.getQTrees().values()) {
            final String qtreeID = cModeQTree.getQTreeID();
            final VServer vServer = getVServerForChild(cluster, "qtree", qtreeID, cModeQTree.getVServerName(), vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason
            }
            final Volume volume = getVolumeForChild(vServer, "qtree", qtreeID, cModeQTree.getVolumeName(), volumeKeysByNameByVServerKey);
            if (volume == null) {
                continue;   // already logged reason
            }
            final String qTreeAlias = getQTreeAlias(volume.getVolumeName(), cModeQTree.getQTreeName());

            final QTree qTree = transformQTree(volume, cModeQTree, exportPolicyKeysByNameByVServerKey);
            if (qTreeKeysByAliasByVServerKey.containsKey(vServer.getKey())) {
                qTreeKeysByAliasByVServerKey.get(vServer.getKey()).put(qTreeAlias, qTree.getKey());
            } else {
                qTreeKeysByAliasByVServerKey.put(vServer.getKey(), new HashMap<String, QTree.QTreeKey>() {{
                    put(qTreeAlias, qTree.getKey());
                }});
            }
            volume.getQTrees().add(qTree);
        }
        return qTreeKeysByAliasByVServerKey;
    }

    private Map<Volume.VolumeKey, Map<String, Snapshot.SnapshotKey>> transformSnapshots(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {
        final Map<Volume.VolumeKey, Map<String, Snapshot.SnapshotKey>> snapshotKeyByNameByVolumeKey = new HashMap<Volume.VolumeKey, Map<String, Snapshot.SnapshotKey>>();
        for (final CModeSnapshot cModeSnapshot : cModeCluster.getSnapshots().values()) {
            final String snapshotInstanceUUID = cModeSnapshot.getInstanceUUID();
            final VServer vServer = getVServerForChild(cluster, "snapshot", snapshotInstanceUUID, cModeSnapshot.getVServerName(), vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason

            }
            final Volume volume = getVolumeForChild(vServer, "snapshot", snapshotInstanceUUID, cModeSnapshot.getVolumeName(), volumeKeysByNameByVServerKey);
            if (volume == null) {
                continue;   // already logged reason
            }

            final Volume.VolumeKey volumeKey = volume.getKey();
            if (!snapshotKeyByNameByVolumeKey.containsKey(volumeKey)) {
                final Map<String, Snapshot.SnapshotKey> snapshotKeys = new HashMap<String, Snapshot.SnapshotKey>();
                snapshotKeyByNameByVolumeKey.put(volumeKey, snapshotKeys);
            }
            final Map<String, Snapshot.SnapshotKey> snapshotKeys = snapshotKeyByNameByVolumeKey.get(volumeKey);
            final Snapshot snapshot = vServer.createSnapshot(volume.getInstanceUUID(), cModeSnapshot.getInstanceUUID());
            snapshot.setAccessTime(cModeSnapshot.getAccessTime());
            snapshot.setBusy(cModeSnapshot.isBusy());
            snapshot.setCumulativePercentageOfTotalBlocks(cModeSnapshot.getCumulativePercentageOfTotalBlocks());
            snapshot.setDependency(cModeSnapshot.getDependency());
            snapshot.setName(cModeSnapshot.getName());
            snapshot.setPercentageOfTotalBlocks(cModeSnapshot.getPercentageOfTotalBlocks());
            snapshot.setVersionUUID(cModeSnapshot.getVersionUUID());
            vServer.getSnapshots().add(snapshot);
            snapshotKeys.put(snapshot.getName(), snapshot.getKey());

            // update Volume OldestSnapshotTimestamp - make it the smallest among all snapshots
            final Long prevValue = volume.getOldestSnapshotTimestamp();
            if (prevValue == null || prevValue >= snapshot.getAccessTime()) {
                volume.setOldestSnapshotTimestamp(snapshot.getAccessTime());
            }
        }
        return snapshotKeyByNameByVolumeKey;
    }

    private Map<LUN.LUNKey, LUN> transformLUNs(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName,
                                               final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey,
                                               final Map<VServer.VServerKey, Map<String, QTree.QTreeKey>> qTreeKeysByAliasByVServerKey,
                                               final Map<String, QosPolicyGroup.QosPolicyGroupKey> qosPolicyGroupKeysByName) {
        // LUNs
        final Map<LUN.LUNKey, LUN> lunsByUUID = new HashMap<LUN.LUNKey, LUN>();
        for (final CModeLUN cModeLUN : cModeCluster.getLUNs().values()) {
            final String lunUUID = cModeLUN.getKey().getUUID();
            final String vServerName = cModeLUN.getVServerName();
            final VServer vServer = getVServerForChild(cluster, "lun", lunUUID, vServerName, vServerKeysByName);
            if (vServer == null) {
                continue;   // already logged reason
            }
            final Volume volume = getVolumeForChild(vServer, "lun", lunUUID, cModeLUN.getVolumeName(), volumeKeysByNameByVServerKey);
            if (volume == null) {
                continue;   // already logged reason
            }
            final String qTreeAlias = getQTreeAlias(volume.getVolumeName(), cModeLUN.getQTreeName());
            final Map<String, QTree.QTreeKey> qTreeKeysByAlias = qTreeKeysByAliasByVServerKey.get(vServer.getKey());
            QTree.QTreeKey qTreeKey = null;
            if (qTreeKeysByAlias != null) {
                qTreeKey = qTreeKeysByAlias.get(qTreeAlias);
            }
            if (null == qTreeKey && !ZAPIPropertiesManager.getDisableZapisSet().contains("qtree-list-iter")) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find qtreeID for qtree alias " + qTreeAlias + " for Lun " + lunUUID);
                continue;
            }
            final String qosPolicyName = cModeLUN.getQosPolicyGroupName();
            final QosPolicyGroup.QosPolicyGroupKey qosPolicyGroupKey = (qosPolicyName != null) ? qosPolicyGroupKeysByName.get(qosPolicyName) : null;
            final LUN lun = transformLUN(volume, cModeLUN, qTreeKey, qosPolicyGroupKey);
            volume.getLUNs().add(lun);
            lunsByUUID.put(lun.getKey(), lun);
        }
        return lunsByUUID;
    }

    private void transformStorageServices(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {

        final Map<VServer.VServerKey, Map<String, StorageService>> storageServicesByNameByVServer = new HashMap<VServer.VServerKey, Map<String, StorageService>>();
        for (final CModeStorageService cModeStorageService : cModeCluster.getStorageServices().values()) {

            // Don't report Storage Service "-", it will be mapped to the DEFAULT storage service
            final String storageServiceName = cModeStorageService.getStorageService();
            if (storageServiceName == null || "-".equals(storageServiceName)) {
                continue;
            }
            final Long storageServiceID = cModeStorageService.getStorageServiceID();
            final Long constituentGroupID = cModeStorageService.getConstituentGroupID();

            final String vServerName = cModeStorageService.getKey().getID(CModeStorageService.ServiceKey.Field.vserver_name);
            final VServer vServer = getVServerForChild(cluster, "StorageService", String.valueOf(constituentGroupID), vServerName, vServerKeysByName);
            if (vServer == null) {
                // already logged reason
                continue;
            }
            final VServer.VServerKey vServerKey = vServer.getKey();
            final String volumeName = cModeStorageService.getKey().getID(CModeStorageService.ServiceKey.Field.volume_name);
            final Map<String, Volume.VolumeKey> volumeKeysByName = volumeKeysByNameByVServerKey.get(vServerKey);
            final Volume.VolumeKey infinivolKey = volumeKeysByName.get(volumeName);
            if (infinivolKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find volume " + volumeName + " for storage service " + vServerName
                        + " " + constituentGroupID);
                continue;
            }

            // This is the infinivol
            final Volume infiniVol = vServer.getVolumes().get(infinivolKey);
            if (infiniVol == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find volume " + volumeName + " for storage service " + vServerName
                        + " " + constituentGroupID);
                continue;
            }

            // Keep track of storage service by name
            if (!storageServicesByNameByVServer.containsKey(vServerKey)) {
                storageServicesByNameByVServer.put(vServerKey, new HashMap<String, StorageService>());
            }
            final Map<String, StorageService> storageServicesByName = storageServicesByNameByVServer.get(vServerKey);
            final StorageService storageService = vServer.createStorageService(storageServiceID, infinivolKey);
            storageService.setStorageServiceName(storageServiceName);
            vServer.getStorageServices().add(storageService);
            storageServicesByName.put(storageServiceName, storageService);

            final StorageServiceVolumeRelationship storageServiceVolumeRelationship =
                    vServer.createStorageServiceVolumeRelationship(storageService.getKey(), infiniVol.getKey());
            vServer.getStorageServiceVolumeRelationships().add(storageServiceVolumeRelationship);
        }

        // We are to add a DEFAULT and SYSTEM storage service for each Infinite Volume
        for (final VServer vServer : cluster.getVServers().values()) {
            final VServer.VServerKey vServerKey = vServer.getKey();
            if (!storageServicesByNameByVServer.containsKey(vServerKey)) {
                // No discovered storage services, initialize the map for this vServer for
                // implicit storage services
                storageServicesByNameByVServer.put(vServerKey, new HashMap<String, StorageService>());
            }
            final Map<String, StorageService> storageServicesByName = storageServicesByNameByVServer.get(vServer.getKey());

            // We assume only 1 per Infinite Volume per vServer.
            // If there are more, we can't support DEFAULT & SYSTEM storage services
            int infinivolCount = 0;
            for (final Volume volume : vServer.getVolumes().values()) {
                if (volume.getVolumeStyle() != Volume.VolumeStyle.INFINITEVOL) {
                    continue;
                }
                infinivolCount++;
                final Volume.VolumeKey infiniteVolKey = volume.getKey();

                // System
                final StorageService systemStorageService = createImplicitStorageService(vServer, infiniteVolKey, SYSTEM_STORAGE_SERVICE, -1L);
                storageServicesByName.put(SYSTEM_STORAGE_SERVICE, systemStorageService);

                // Default
                final StorageService defaultStorageService = createImplicitStorageService(vServer, infiniteVolKey, DEFAULT_STORAGE_SERVICE, -2L);
                storageServicesByName.put(DEFAULT_STORAGE_SERVICE, defaultStorageService);
            }

            if (infinivolCount > 1) {
                // Unexpected and unsupported, don't add the implicit storage services
                storageServicesByName.remove(SYSTEM_STORAGE_SERVICE);
                storageServicesByName.remove(DEFAULT_STORAGE_SERVICE);
                logger.warn(PackageContext.getLogPrefix() + " - vServer " + vServer.getName()
                        + " has multiple infinite volumes, SYSTEM and DEFAULT storage services are not supported.");
            } else if (infinivolCount > 0) {
                // Add the System storage service, but wait to add the Default storage service.
                // The Default storage service should only be reported if there are DATA volumes that
                // need to reference it (their storage service is null or "" or "-").
                final StorageService systemStorageService = storageServicesByName.get(SYSTEM_STORAGE_SERVICE);
                addImplicitStorageService(vServer, systemStorageService);
            }
        }

        // Aggregate constituent volume capacities up into the storage service
        // For each constituent volume in the system:
        //      If constituentRole is in the set of {NAMESPACE, NS_MIRROR, OLS} AND the storage-service returned from ZAPI is null:
        //          Link volume.storageServiceId to the SYSTEM storage_service created for this vServer's infinite volume.
        // If constituentRole is {DATA} and the storage-service returned from ZAPI is "-" or "" or null:
        //      Link volume.storageServiceId to the DEFAULT storage_service created for this vServer's infinite volume.
        // If constituentRole is in the set of {UNKNOWN, NOT_MAPPED}, do nothing
        for (final VServer vServer : cluster.getVServers().values()) {
            boolean needsDefaultStorageService = false;
            final Map<String, StorageService> storageServicesByName = storageServicesByNameByVServer.get(vServer.getKey());
            for (final Volume volume : vServer.getVolumes().values()) {
                final UUIDKey cModeVolumeKey = new UUIDKey("volume", volume.getKey().getID(Volume.VolumeKey.Field.uuid));
                final CModeVolume cModeVolume = cModeCluster.getVolumes().get(cModeVolumeKey);
                final Volume.ConstituentRole constituentRole = volume.getConstituentRole();
                String storageServiceName = cModeVolume.getStorageServiceName();

                if (constituentRole == Volume.ConstituentRole.NAMESPACE ||
                        constituentRole == Volume.ConstituentRole.NS_MIRROR ||
                        constituentRole == Volume.ConstituentRole.OLS) {

                    if (storageServiceName == null) {
                        final StorageService systemStorageService = storageServicesByName.get(SYSTEM_STORAGE_SERVICE);
                        if (systemStorageService != null) {
                            storageServiceName = SYSTEM_STORAGE_SERVICE;
                            volume.setStorageServiceUUID(systemStorageService.getKey().getOcieUUID());
                        }
                    }
                } else if (constituentRole == Volume.ConstituentRole.DATA) {
                    if (storageServiceName == null || (storageServiceName.equals("") || storageServiceName.equals("-"))) {
                        final StorageService defaultStorageService = storageServicesByName.get(DEFAULT_STORAGE_SERVICE);
                        if (defaultStorageService != null) {
                            needsDefaultStorageService = true;
                            storageServiceName = DEFAULT_STORAGE_SERVICE;
                            volume.setStorageServiceUUID(defaultStorageService.getKey().getOcieUUID());
                        }
                    }
                }

                if (storageServiceName == null || storageServiceName.isEmpty()) {
                    continue;
                }

                if (!storageServicesByName.containsKey(storageServiceName)) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find storage service " + storageServiceName
                            + " for volume " + volume.getVolumeName());
                    continue;
                }
                final StorageService storageService = storageServicesByName.get(storageServiceName);
                volume.setStorageServiceUUID(storageService.getKey().getOcieUUID());
                if (!volume.isConstituent()) {
                    continue;
                }
                if (volume.getVolumeStyle() != Volume.VolumeStyle.FLEX) {
                    continue;
                }
                storageService.addToSizeUsed(getLongForSum(volume.getSizeUsed()));
                storageService.addToSizeTotal(getLongForSum(volume.getSizeTotal()));
                storageService.addToSizeUsedBySnapshots(getLongForSum(volume.getSizeUsedBySnapshots()));
                storageService.addToSnapshotReservedSize(getLongForSum(volume.getSnapshotReserveSize()));
                storageService.addToDeduplicationSpaceSaved(getLongForSum(volume.getDeduplicationSpaceSaved()));
                storageService.addToCompressionSpaceSaved(getLongForSum(volume.getCompressionSpaceSaved()));
            }
            if (needsDefaultStorageService) {
                final StorageService defaultStorageService = storageServicesByName.get(DEFAULT_STORAGE_SERVICE);
                addImplicitStorageService(vServer, defaultStorageService);
            } else {
                storageServicesByName.remove(DEFAULT_STORAGE_SERVICE);
            }
        }
    }


    private StorageService createImplicitStorageService(final VServer vServer, final Volume.VolumeKey infiniteVolKey,
                                                        final String storageServiceName, final Long storageServiceID) {
        final StorageService implicitStorageService = vServer.createStorageService(storageServiceID, infiniteVolKey);
        implicitStorageService.setStorageServiceName(storageServiceName);
        return implicitStorageService;
    }

    private void addImplicitStorageService(final VServer vServer, final StorageService implicitStorageService) {
        final Volume.VolumeKey infiniteVolKey = implicitStorageService.getVolumeKey();
        vServer.getStorageServices().add(implicitStorageService);
        final StorageServiceVolumeRelationship defaultServiceVolumeRelationship =
                vServer.createStorageServiceVolumeRelationship(implicitStorageService.getKey(), infiniteVolKey);
        vServer.getStorageServiceVolumeRelationships().add(defaultServiceVolumeRelationship);
    }

    private Map<VServer.VServerKey, Map<String, RoutingGroup.RoutingGroupKey>> transformRouteGroups(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        final Map<VServer.VServerKey, Map<String, RoutingGroup.RoutingGroupKey>> routingGroupKeysByNameByVServerKey = new HashMap<VServer.VServerKey, Map<String, RoutingGroup.RoutingGroupKey>>();
        for (final CModeNetRoute cModeRouteEntry : cModeCluster.getNetRoutes().values()) {
            final String groupName = cModeRouteEntry.getKey().getID(CModeNetRoute.NetRouteKey.Field.routing_group_name);
            final String vServerName = cModeRouteEntry.getKey().getID(CModeNetRoute.NetRouteKey.Field.vserver_name);
            final VServer vServer = getVServerForChild(cluster, "RouteGroup", groupName, vServerName, vServerKeysByName);
            if (vServer == null) {
                // already logged reason
                continue;
            }
            final VServer.VServerKey vServerKey = vServer.getKey();
            if (!routingGroupKeysByNameByVServerKey.containsKey(vServerKey)) {
                routingGroupKeysByNameByVServerKey.put(vServerKey, new HashMap<String, RoutingGroup.RoutingGroupKey>());
            }
            final Map<String, RoutingGroup.RoutingGroupKey> routingGroupKeysByName = routingGroupKeysByNameByVServerKey.get(vServerKey);
            final RoutingGroups routingGroups = vServer.getRoutingGroups();
            final RoutingGroup.RoutingGroupKey routingGroupKey = new RoutingGroup.RoutingGroupKey(cluster.getClusterUUID(), vServer.getUUID(), groupName);
            if (!routingGroups.containsKey(routingGroupKey)) {
                final RoutingGroup routingGroup = vServer.createRoutingGroup(groupName);
                final EnumWrapper<RoutingGroup.AddressFamily> addressFamilyWrapper = cModeRouteEntry.getAddressFamily();
                routingGroup.setAddressFamily(addressFamilyWrapper.getEnumValue());
                routingGroup.setAddressFamilyRaw(addressFamilyWrapper.getRawValue());
                routingGroups.add(routingGroup);
                routingGroupKeysByName.put(groupName, routingGroup.getKey());
            }
            final RoutingGroup routingGroup = routingGroups.get(routingGroupKey);
            final RoutingGroup.RoutingGroupDestination destination = routingGroup.createRoutingGroupDestination(cModeRouteEntry.getDestinationAddress());
            destination.setGatewayAddress(cModeRouteEntry.getGatewayAddress());
            routingGroup.getRoutingGroupDestinations().add(destination);
        }
        return routingGroupKeysByNameByVServerKey;
    }


    private Map<VServer.VServerKey, Map<String, NetworkLif>> transformLifs(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName,
                                                                           final Map<String, Node> nodesByName,
                                                                           final Map<VServer.VServerKey, Map<String, RoutingGroup.RoutingGroupKey>> routingGroupKeysByNameByVServer,
                                                                           final Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> failoverGroupKeyByName) {
        // Lifs
        // final Map<String, NetworkLif> networkLIFsByNameMap = new HashMap<String, NetworkLif>();
        final Map<VServer.VServerKey, Map<String, NetworkLif>> networkLifsByNameByVServer = new HashMap<VServer.VServerKey, Map<String, NetworkLif>>();
        for (final CModeLif cModeLif : cModeCluster.getLifs().values()) {
            final CModeLif.LifKey lifKey = cModeLif.getKey();
            final String lifName = cModeLif.getInterfaceName();                         // always interface name
            final String vServerName = cModeLif.getKey().getID(CModeLif.LifKey.Field.vserver_name);
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for lif " + lifName);
                continue;
            }
            final VServer parentVServer = cluster.getVServers().get(vServerKey);
            if (parentVServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for lif " + lifName);
                continue;
            }
            if (!networkLifsByNameByVServer.containsKey(vServerKey)) {
                final Map<String, NetworkLif> networkLifsByName = new HashMap<String, NetworkLif>();
                networkLifsByNameByVServer.put(vServerKey, networkLifsByName);
            }

            final Map<String, NetworkLif> networkLifsByName = networkLifsByNameByVServer.get(vServerKey);
            final String currentNodeName = cModeLif.getCurrentNode();
            final Node currentNode = nodesByName.get(currentNodeName);
            final Node.NodeKey currentNodeKey;
            final String currentNodeUUID;
            final String verifiedCurrentNodeName;
            if (currentNode == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find current node " + currentNodeName + " for lif " + lifName);
                }
                currentNodeKey = null;
                currentNodeUUID = null;
                verifiedCurrentNodeName = null;
            } else {
                currentNodeKey = currentNode.getKey();
                currentNodeUUID = currentNodeKey.getOcieUUID();
                verifiedCurrentNodeName = currentNodeName;
            }

            final String homeNodeName = cModeLif.getHomeNode();
            final Node homeNode = nodesByName.get(homeNodeName);
            final Node.NodeKey homeNodeKey;
            final String verifiedHomeNodeName;
            final String homeNodeUUID;
            if (homeNode == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find home node " + homeNodeName + " for lif " + lifName);
                }
                homeNodeKey = null;
                verifiedHomeNodeName = null;
                homeNodeUUID = null;
            } else {
                homeNodeKey = homeNode.getKey();
                verifiedHomeNodeName = homeNodeName;
                homeNodeUUID = homeNodeKey.getOcieUUID();
            }

            final String homePortName = cModeLif.getHomePort();
            final NetworkPort.PortKey homePortKey = getNetworkPortKey(cModeCluster, verifiedHomeNodeName, homeNodeUUID, homePortName);
            final String currentPortName = cModeLif.getCurrentPort();
            final NetworkPort.PortKey currentPortKey = getNetworkPortKey(cModeCluster, verifiedCurrentNodeName, currentNodeUUID, currentPortName);
            if (currentPortKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find current port " + currentNodeName + " " + currentPortName + " for lif " + lifName);
                }
            }

            // If this LIF is an FCP LIF we need to grab more info
            if (cModeLif.getDataProtocolList().contains("fcp")) {
                final FcpPort.FcpPortKey homeFcpPortKey = getFcpPortKey(cModeCluster, verifiedHomeNodeName, homePortName, nodesByName);
                final FcpPort.FcpPortKey currentFcpPortKey = getFcpPortKey(cModeCluster, verifiedCurrentNodeName, currentPortName, nodesByName);
                final CModeFcpLif cModeFcpLif = getCModeFcpLif(cModeCluster, lifKey);
                final FcpLif fcpLif = transformFcpLif(parentVServer, cModeLif, cModeFcpLif, currentNodeKey, currentFcpPortKey, homeNodeKey, homeFcpPortKey, failoverGroupKeyByName);
                parentVServer.getFcpLifs().add(fcpLif);
            } else {
                final Map<String, RoutingGroup.RoutingGroupKey> routingGroupKeysByName = routingGroupKeysByNameByVServer.get(vServerKey);
                final String routingGroupName = cModeLif.getRoutingGroupName();
                final RoutingGroup.RoutingGroupKey routingGroupKey;
                if (routingGroupName != null && routingGroupKeysByName != null && routingGroupKeysByName.containsKey(routingGroupName)) {
                    routingGroupKey = routingGroupKeysByName.get(routingGroupName);
                } else {
                    routingGroupKey = null;
                }
                final NetworkLif networkLif = transformNetworkLif(parentVServer, cModeLif, currentNodeKey, currentPortKey, homeNodeKey, homePortKey, routingGroupKey, failoverGroupKeyByName);
                parentVServer.getNetworkLifs().add(networkLif);
                networkLifsByName.put(lifName, networkLif);
            }
        }
        return networkLifsByNameByVServer;
    }


    private void transformIGroups(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        // IGroups
        for (final CModeIGroup cModeIGroup : cModeCluster.getIGroups().values()) {
            final String vServerName = cModeIGroup.getVServerName();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for iGroup " + cModeIGroup.getIGroupName());
                continue;
            }
            final VServer parentVServer = cluster.getVServers().get(vServerKey);
            if (parentVServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for iGroup " + cModeIGroup.getIGroupName());
                continue;
            }
            final IGroup iGroup = parentVServer.createIGroup(cModeIGroup.getKey().getUUID());
            iGroup.setName(cModeIGroup.getIGroupName());
            iGroup.setOsType(cModeIGroup.getOSType());
            iGroup.setType(cModeIGroup.getType().getEnumValue());
            iGroup.setTypeRaw(cModeIGroup.getType().getRawValue());
            iGroup.setAluaEnabled(cModeIGroup.getAluaEnabled());
            iGroup.setVsaEnabled(cModeIGroup.getVsaEnabled());
            iGroup.setUsePartner(cModeIGroup.getUsePartner());
            parentVServer.getIGroups().add(iGroup);

            // IGroup Initiators

            for (final String initiatorName : cModeIGroup.getInitiatorNames()) {
                final IGroupInitiator initiator = iGroup.createIGroupInitiator(initiatorName);
                iGroup.getIGroupInitiators().add(initiator);
            }
        }
    }

    private void transformPortSets(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModePortSet cModePortSet : cModeCluster.getPortSets().values()) {
            final String vServerName = cModePortSet.getKey().getID(CModePortSet.PortSetKey.Field.vserver_name);
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            final String portSetName = cModePortSet.getKey().getID(CModePortSet.PortSetKey.Field.portset_name);
            if (vServerKey == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for PortSet " + portSetName);
                continue;
            }
            final VServer parentVServer = cluster.getVServers().get(vServerKey);
            if (parentVServer == null) {
                logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for PortSet " + portSetName);
                continue;
            }

            final Portset portSet = parentVServer.createPortSet(portSetName);
            portSet.setType(cModePortSet.getType().getEnumValue());
            portSet.setTypeRaw(cModePortSet.getType().getRawValue());
            parentVServer.getPortSets().add(portSet);

            FcpLifs fcpLifs = parentVServer.getFcpLifs();
            ISCSIPortalGroups iSCSIPortalGroups = parentVServer.getISCSIPortalGroups();

            //build map for searching by name (key is portname, value is uuid)
            Map<String, String>  lifMap = new HashMap<String, String>();
            if(cModePortSet.getType().getEnumValue() == Portset.Type.FCP) {
                if (fcpLifs!=null && !fcpLifs.isEmpty()){
                    for(Map.Entry<FcpLif.LifKey, FcpLif> entry : fcpLifs.entrySet()) {
                        if (logger.isDebugEnabled()) {
                            logger.info("portname =  " + entry.getValue().getInterfaceName() + " uuid = " + entry.getKey().getOcieUUID());
                        }
                        lifMap.put(entry.getValue().getInterfaceName(), entry.getKey().getOcieUUID());
                    }
                 }
            }
            else if(cModePortSet.getType().getEnumValue() == Portset.Type.ISCSI){
                if(iSCSIPortalGroups!=null && !iSCSIPortalGroups.isEmpty()) {
                    for(Map.Entry<ISCSIPortalGroup.ISCSIPortalGroupKey, ISCSIPortalGroup> entry : iSCSIPortalGroups.entrySet()) {
                        if (logger.isDebugEnabled()) {
                            logger.info("portname =  " + entry.getKey().getID(ISCSIPortalGroup.ISCSIPortalGroupKey.Field.portal_group_name) + " uuid = " + entry.getKey().getOcieUUID());
                        }
                        lifMap.put(entry.getKey().getID(ISCSIPortalGroup.ISCSIPortalGroupKey.Field.portal_group_name), entry.getKey().getOcieUUID());
                    }
                }
            }

            // PortsetPorts
            final List<String> portNames = cModePortSet.getPortsetPortNames();
            for (final String portName : portNames) {
                final PortsetPort portsetPort = portSet.createPortsetPort(portName);
                //set fcpportuuid or iscsiportalgroupuuid
                if(!lifMap.isEmpty()) {
                    if(cModePortSet.getType().getEnumValue() == Portset.Type.FCP)  {
                        portsetPort.setFcpLifId(lifMap.get(portName));
                     }
                    else if (cModePortSet.getType().getEnumValue() == Portset.Type.ISCSI)   {
                        portsetPort.setIScsiPortalGroupId(lifMap.get(portName));
                    }
                }

                portSet.getPortsetPorts().add(portsetPort);
            }

            final List<String> igroupNames = cModePortSet.getInitiatorGroupNames();
            for (final String iGroupName : igroupNames) {
                final String iGroupUUID = getIGroupUUIDByName(parentVServer, iGroupName);
                if (iGroupUUID == null) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find IGroup " + iGroupName + " for PortSet " + portSetName);
                    continue;
                }
                final PortsetInitiatorGroup iGroup = portSet.createPortsetInitiatorGroup(iGroupUUID);
                portSet.getPortsetInitiatorGroups().add(iGroup);
            }
        }
    }

    private void transformLunMaps(final CModeCluster cModeCluster,
            final Cluster cluster,
            final Map<String, VServer.VServerKey> vServerKeysByName,
            final Map<LUN.LUNKey, LUN> lunsByKey, Map<String, Node> nodesByName) {
        // LUNMaps
        for (final CModeLUNMap cModeLUNMap : cModeCluster.getLUNMaps().values()) {
            final String vServerName = cModeLUNMap.getVServerName();
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            final CModeLUNMap.LUNMapKey cModeLUNMapKey = cModeLUNMap.getKey();
            if (vServerKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for LUNMap " + cModeLUNMapKey);
                }
                continue;
            }
            final VServer parentVServer = cluster.getVServers().get(vServerKey);
            if (parentVServer == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for LUNMap " + cModeLUNMapKey);
                }
                continue;
            }

            // Assert LUN and iGroup are valid and discovered
            final String lunUUID = cModeLUNMapKey.getID(CModeLUNMap.LUNMapKey.Field.lun_uuid);
            final LUN.LUNKey lunKey = new LUN.LUNKey(cluster.getClusterUUID(), lunUUID);
            final LUN lun = lunsByKey.get(lunKey);
            if (lun == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find lun " + lunUUID + " for LUNMap " + cModeLUNMapKey);
                }
                continue;
            }

            final String iGroupUUID = cModeLUNMapKey.getID(CModeLUNMap.LUNMapKey.Field.igroup_uuid);
            final IGroup.IGroupKey iGroupKey = new IGroup.IGroupKey(cluster.getClusterUUID(), iGroupUUID);
            if (!parentVServer.getIGroups().containsKey(iGroupKey)) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to igroup " + iGroupUUID + " for LUNMap " + cModeLUNMapKey);
                }
                continue;
            }

            final LUNMap lunMap = parentVServer.createLUNMap(lunUUID, iGroupUUID);
            lunMap.setLunID(cModeLUNMap.getLunID());
            lunMap.setPath(cModeLUNMap.getPath());
            String owningNodeName = cModeLUNMap.getOwningNodeName();
            if (nodesByName.containsKey(owningNodeName)) {
                Node owningNode = nodesByName.get(owningNodeName);
                lunMap.setOwningNodeKey(owningNode.getKey());
            }
            else if (owningNodeName != null) {
                logger.warn("{} - Failed to find owning node: {}, for LUNMap: {}",
                        new Object[] {PackageContext.getLogPrefix(), owningNodeName, cModeLUNMapKey });
            }

            LUNMapReportingNodes lunMapReportingNodes = lunMap.getLunMapReportingNodes();
            for (String reportingNodeName : cModeLUNMap.getReportingNodeNames()) {
                if (nodesByName.containsKey(reportingNodeName)) {
                    Node reportingNode = nodesByName.get(reportingNodeName);
                    lunMapReportingNodes.add(lunMap.createLUNMapReportingNode(lunMap.getKey(),
                            reportingNode.getKey()));
                }
                else if (reportingNodeName != null) {
                    logger.warn("{} - Failed to find reporting node: {}, for LUNMap: {}",
                            new Object[]{PackageContext.getLogPrefix(), reportingNodeName, cModeLUNMapKey});
                }
            }
            
            parentVServer.getLUNMaps().add(lunMap);
        }
    }
    
    private void transformDiskPaths(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Node> nodesByName, final Collection<Disk.DiskKey> diskKeys) {
        // Disk Paths
        // Iterating through the disks again here because the Paths needs all disks
        // to have been discovered first.
        final String clusterUUID = cluster.getClusterUUID();
        for (final CModeDisk cModeDisk : cModeCluster.getDisks().values()) {
            for (final CModeDiskPath cModeDiskPath : cModeDisk.getDiskPaths().values()) {

                // Find the parent Node
                final String nodeName = cModeDiskPath.getNode();
                if (!nodesByName.containsKey(nodeName)) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Failed to find node " + nodeName + " for disk path " + cModeDiskPath.getKey());
                    }
                    continue;
                }
                final Node parentNode = nodesByName.get(nodeName);

                // Find the reference initiator
                // Using the schema objects since the key is different than the those generated by the model
                final String initiatorPortName = cModeDiskPath.getInitiatorPortName();
                final String nodeUUID = parentNode.getKey().getID(Node.NodeKey.Field.uuid);
                final StorageAdapter.AdapterKey storageAdapterKey = new StorageAdapter.AdapterKey(clusterUUID, nodeUUID, initiatorPortName);
                if (!parentNode.getStorageAdapters().containsKey(storageAdapterKey)) {
                    // Probably due to unsupported adapter type (e.g. ADT_IF_SAS)
                    // We still want to report the path even if the referenced adapter is unknown.
                    // Set the key to unknown so we don't report a reference to an adapter that does not exist
                    storageAdapterKey.setKnown(false);
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Failed to find storage adapter " + initiatorPortName + " for disk path " + cModeDiskPath.getKey());
                    }
                } else {
                    storageAdapterKey.setKnown(true);
                }

                //Make sure disk had been found in the Cluster
                final String diskUUID = cModeDiskPath.getDiskUUID();
                final Disk.DiskKey diskKey = new Disk.DiskKey(clusterUUID, diskUUID);
                if (!diskKeys.contains(diskKey)) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Failed to find disk " + diskUUID + " for disk path " + cModeDiskPath.getKey());
                    }
                    continue;
                }

                final DiskPath diskPath = transformDiskPath(cluster, cModeDiskPath, parentNode.getKey(), diskKey, storageAdapterKey);
                cluster.getDiskPaths().add(diskPath);
            }
        }
    }

    private Map<VServer.VServerKey, Map<String, CIFSShare>> transformCIFSShares(final CModeCluster cModeCluster, final Cluster cluster,
                                                                                final Map<String, VServer.VServerKey> vServerKeysByName) {

        // Return map of transformed shares
        final Map<VServer.VServerKey, Map<String, CIFSShare>> cifsSharesByNameByVServerKey = new HashMap<VServer.VServerKey, Map<String, CIFSShare>>();

        // Sort the raw set of CIFS Shares by vServer, then correlate Shares to volumes & qtrees within the
        // context of a single vServer
        final Map<VServer.VServerKey, Set<CModeCIFSShare>> cModeSharesByVServer = new HashMap<VServer.VServerKey, Set<CModeCIFSShare>>();
        for (final CModeCIFSShare cModeCIFSShare : cModeCluster.getCifsShares().values()) {
            // Get the CIFS Server
            final String cifsServerName = cModeCIFSShare.getCIFSServerName();
            final UUIDKey cModeCIFSServerKey = new UUIDKey("cifs_service", cifsServerName);
            if (cModeCluster.getCifsServers().containsKey(cModeCIFSServerKey)) {
                final CModeCIFSServer cModeCIFSServer = cModeCluster.getCifsServers().get(cModeCIFSServerKey);
                final VServer.VServerKey vServerKey = vServerKeysByName.get(cModeCIFSServer.getVServerName());
                if (null != vServerKey) {
                    if (!cModeSharesByVServer.containsKey(vServerKey)) {
                        cModeSharesByVServer.put(vServerKey, new HashSet<CModeCIFSShare>());
                    }
                    final Set<CModeCIFSShare> cifsShares = cModeSharesByVServer.get(vServerKey);
                    cifsShares.add(cModeCIFSShare);
                }
            }
        }

        // Now transform cifs shares for each vServer
        for (final VServer vServer : cluster.getVServers().values()) {
            final VServer.VServerKey vServerKey = vServer.getKey();
            if (!cModeSharesByVServer.containsKey(vServerKey)) {
                // No shares to transform for this vServer
                continue;
            }

            // Build maps of volume by name and qtree by path for volume/qtree lookup
            final Map<String, QTree> qTreesByPath = new HashMap<String, QTree>();
            for (final Volume volume : vServer.getVolumes().values()) {
                for (final QTree qTree : volume.getQTrees().values()) {
                    // Create a share path string to look for.
                    final String qTreePath = getQTreePath(qTree);
                    qTreesByPath.put(qTreePath, qTree);
                }
            }

            // Now transform CIFS Shares for this vServer
            final Set<CModeCIFSShare> cModeCIFSShares = cModeSharesByVServer.get(vServerKey);
            for (final CModeCIFSShare cModeCIFSShare : cModeCIFSShares) {
                Volume.VolumeKey shareVolumeKey = null;
                QTree.QTreeKey shareQTreeKey = null;
                final String cifsSharePath = cModeCIFSShare.getPath();
                // ZAPI did not provide volume name, lookup QTree by exact path match
                if (qTreesByPath.containsKey(cifsSharePath)) {
                    final QTree shareQTree = qTreesByPath.get(cifsSharePath);
                    shareQTreeKey = shareQTree.getKey();
                    final Volume shareVolume = shareQTree.getParent();
                    if (shareVolume != null && Volume.State.OFFLINE != shareVolume.getState()) {
                        shareVolumeKey = shareVolume.getKey();
                    }
                }

                final CIFSShare cifsShare = transformCIFSShare(vServer, cModeCIFSShare, shareVolumeKey, shareQTreeKey);
                vServer.getCIFSShares().add(cifsShare);
                // Add new share to return map
                if (!cifsSharesByNameByVServerKey.containsKey(vServerKey)) {
                    cifsSharesByNameByVServerKey.put(vServerKey, new HashMap<String, CIFSShare>());
                }
                final Map<String, CIFSShare> vServerSharesByName = cifsSharesByNameByVServerKey.get(vServerKey);
                vServerSharesByName.put(cifsShare.getName(), cifsShare);
            }
        }
        return cifsSharesByNameByVServerKey;
    }

    private String getQTreePath(final QTree qTree) {
        final Volume volume = qTree.getParent();
        final String junctionPath = volume.getJunctionPath();
        final String sharePath;
        if (qTree.getName().equals("")) {
            sharePath = junctionPath;
        } else {
            sharePath = ("/".equals(junctionPath)) ? "/" + qTree.getName() : junctionPath + "/" + qTree.getName();
        }
        return sharePath;
    }

    private void transformCIFSShareACLs(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName,
                                        final Map<VServer.VServerKey, Map<String, CIFSShare>> cifsSharesByNameByVServerKey) {

        for (final CModeCIFSACL cModeCIFSACL : cModeCluster.getCIFSShareACLs().values()) {
            final String vServerName = cModeCIFSACL.getKey().getID(CModeCIFSACL.ACLKey.Field.vserver_name);
            final String shareName = cModeCIFSACL.getKey().getID(CModeCIFSACL.ACLKey.Field.share_name);
            final String userOrGroup = cModeCIFSACL.getKey().getID(CModeCIFSACL.ACLKey.Field.user_or_group);
            if (!vServerKeysByName.containsKey(vServerName)) {
                if (logger.isDebugEnabled()) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find parent vServer " + vServerName
                            + " for CIFS share " + shareName + ", skipping ACL for owner " + userOrGroup);
                }
                continue;
            }
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                if (logger.isDebugEnabled()) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find parent vServer " + vServerName
                            + " for CIFS share " + shareName + ", skipping ACL for owner " + userOrGroup);
                }
                continue;
            }
            final Map<String, CIFSShare> cifsSharesByName = cifsSharesByNameByVServerKey.get(vServerKey);
            if (cifsSharesByName == null || !cifsSharesByName.containsKey(shareName)) {
                if (logger.isDebugEnabled()) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find CIFS Share " + vServerName
                            + " " + shareName + ", skipping ACL for owner " + userOrGroup);
                }
                continue;
            }
            final CIFSShare cifsShare = cifsSharesByName.get(shareName);
            if (cifsShare == null) {
                if (logger.isDebugEnabled()) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find CIFS Share " + vServerName
                            + " " + shareName + ", skipping ACL for owner " + userOrGroup);
                }
                continue;
            }
            final CIFSShareACL cifsShareACL = cifsShare.createCIFSShareACL(userOrGroup);
            final EnumWrapper<CIFSShareACL.Permission> permissionWrapper = cModeCIFSACL.getPermissionWrapper();
            cifsShareACL.setPermission(permissionWrapper.getEnumValue());
            cifsShareACL.setPermissionRaw(permissionWrapper.getRawValue());
            cifsShare.getCIFSShareACLs().add(cifsShareACL);
        }
    }

    private Map<VServer.VServerKey, Map<String, ISCSIPortalGroup>> transformISCSIPortalGroups(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        // ISCSI Portal Groups
        final Map<VServer.VServerKey, Map<String, ISCSIPortalGroup>> iscsiPortalGroupByNameByVServer = new HashMap<VServer.VServerKey, Map<String, ISCSIPortalGroup>>();
        for (final CModeISCSIPortalGroup cModeISCSIPortalGroup : cModeCluster.getISCSIPortalGroups().values()) {

            final String vServerName = cModeISCSIPortalGroup.getVServerName();
            final VServer vServer = getVServerForChild(cluster, "ISCSIPortalGroup", cModeISCSIPortalGroup.getName(), vServerName, vServerKeysByName);
            if (null == vServer) {
                continue; //Already logged
            }
            final VServer.VServerKey vServerKey = vServer.getKey();
            if (!iscsiPortalGroupByNameByVServer.containsKey(vServerKey)) {
                final Map<String, ISCSIPortalGroup> portalGroupsByName = new HashMap<String, ISCSIPortalGroup>();
                iscsiPortalGroupByNameByVServer.put(vServerKey, portalGroupsByName);
            }
            final Map<String, ISCSIPortalGroup> iscsiPortalGroupByName = iscsiPortalGroupByNameByVServer.get(vServerKey);
            final ISCSIPortalGroup iscsiPortalGroup = transformISCSIPortalGroup(vServer, cModeISCSIPortalGroup);
            vServer.getISCSIPortalGroups().add(iscsiPortalGroup);

            iscsiPortalGroupByName.put(cModeISCSIPortalGroup.getName(), iscsiPortalGroup);
        }
        return iscsiPortalGroupByNameByVServer;
    }

    private void transformISCSIPortalGroupLIFs(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName,
                                               final Map<VServer.VServerKey, Map<String, ISCSIPortalGroup>> iscsiPortalGroupsByNameByVServer,
                                               final Map<VServer.VServerKey, Map<String, NetworkLif>> networkLIFsByNameByVServerMap) {
        // ISCSI Portal Group to LIFs
        for (final CModeISCSIPortalGroupToLIF cModeISCSIPortalGroupToLIF : cModeCluster.getISCSIPortalGroupToLIFs().values()) {

            final String vServerName = cModeISCSIPortalGroupToLIF.getVServerName();
            final VServer vServer = getVServerForChild(cluster, "ISCSIPortalGroup", cModeISCSIPortalGroupToLIF.getKey().getOcieUUID(), vServerName, vServerKeysByName);
            if (null == vServer) {
                continue; //Already logged
            }
            final VServer.VServerKey vServerKey = vServer.getKey();
            final Map<String, ISCSIPortalGroup> iscsiPortalGroupByName = (iscsiPortalGroupsByNameByVServer.containsKey(vServerKey))
                    ? iscsiPortalGroupsByNameByVServer.get(vServerKey)
                    : new HashMap<String, ISCSIPortalGroup>();

            final String portalGroupName = cModeISCSIPortalGroupToLIF.getPortalGroupName();
            final ISCSIPortalGroup parentPortalGroup = iscsiPortalGroupByName.get(portalGroupName);
            if (null == parentPortalGroup) {
                if (logger.isDebugEnabled()) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find parent Portal Group " + portalGroupName);
                }
                continue;
            }

            final Map<String, NetworkLif> networkLifsByNameMap = networkLIFsByNameByVServerMap.containsKey(vServerKey)
                    ? networkLIFsByNameByVServerMap.get(vServerKey)
                    : new HashMap<String, NetworkLif>();

            final String lifName = cModeISCSIPortalGroupToLIF.getInterfaceName();
            final NetworkLif referencedLIF = networkLifsByNameMap.get(lifName);
            if (null == referencedLIF) {
                if (logger.isDebugEnabled()) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find referenced LIF " + lifName);
                }
                continue;
            }

            final String referencedLifUUID = referencedLIF.getKey().getID(Lif.LifKey.Field.lif_uuid);
            final ISCSIPortalGroupToLIF iscsiPortalGroupToLIF = transformISCSIPortalGroupToLIF(parentPortalGroup, referencedLifUUID, cModeISCSIPortalGroupToLIF);

            parentPortalGroup.getISCSIPortalGroupToLIFs().add(iscsiPortalGroupToLIF);
        }
    }

    private void transformISCSISessions(final CModeCluster cModeCluster, final Cluster cluster, final Map<VServer.VServerKey, Map<String, ISCSIPortalGroup>> iscsiPortalGroupsByNameByVServer,
                                        final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, NetworkLif>> networkLIFsByNameByVServerMap) {
        // ISCSI Sessions
        for (final CModeISCSISession cModeISCSISession : cModeCluster.getISCSISessions().values()) {
            final String vServerName = cModeISCSISession.getVServerName();
            final VServer vServer = getVServerForChild(cluster, "ISCSIPortalGroup", cModeISCSISession.getKey().getOcieUUID(), vServerName, vServerKeysByName);
            if (null == vServer) {
                continue; //Already logged
            }
            final VServer.VServerKey vServerKey = vServer.getKey();
            final Map<String, ISCSIPortalGroup> iscsiPortalGroupByName = (iscsiPortalGroupsByNameByVServer.containsKey(vServerKey))
                    ? iscsiPortalGroupsByNameByVServer.get(vServerKey)
                    : new HashMap<String, ISCSIPortalGroup>();

            final String portalGroupName = cModeISCSISession.getPortalGroupName();
            final ISCSIPortalGroup parentPortalGroup = iscsiPortalGroupByName.get(portalGroupName);
            if (null == parentPortalGroup) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find parent Portal Group " + portalGroupName);
                continue;
            }

            final Map<String, NetworkLif> networkLifsByNameMap = networkLIFsByNameByVServerMap.containsKey(vServerKey)
                    ? networkLIFsByNameByVServerMap.get(vServerKey)
                    : new HashMap<String, NetworkLif>();

            final ISCSISession iscsiSession = transformISCSISession(parentPortalGroup, cluster.getClusterUUID(), vServer.getKey(), cModeISCSISession);

            // ISCSI Session Connections
            for (final CModeISCSISessionConnection cModeISCSISessionConnection : cModeISCSISession.getCModeISCSISessionConnections().values()) {

                final String lifName = cModeISCSISessionConnection.getInterfaceName();
                final NetworkLif referencedLIF = networkLifsByNameMap.get(lifName);
                if (null == referencedLIF) {
                    if (logger.isDebugEnabled()) {
                        logger.warn(PackageContext.getLogPrefix() + " - Failed to find referenced LIF " + lifName);
                    }
                    continue;
                }
                final ISCSISessionConnection iscsiSessionConnection = transformISCSISessionConnection(iscsiSession, referencedLIF.getKey(), cModeISCSISessionConnection);
                iscsiSession.getISCSISessionConnections().add(iscsiSessionConnection);
            }

            parentPortalGroup.getISCSISessions().add(iscsiSession);
        }
    }

    private void transformISCSISecurity(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeISCSISecurityEntry cModeISCSISecurityEntry : cModeCluster.getISCSISecurityEntries().values()) {
            final String vServerName = cModeISCSISecurityEntry.getVServerName();
            final VServer parent = getVServerForChild(cluster, "ISCSISession", cModeISCSISecurityEntry.getKey().getOcieUUID(), vServerName, vServerKeysByName);
            if (null == parent) {
                continue; //Already logged
            }
            final ISCSISecurityEntry iscsiSecurityEntry = transformISCSISecurityEntry(parent, cluster.getClusterUUID(), cModeISCSISecurityEntry);
            parent.getISCSISecurityEntries().add(iscsiSecurityEntry);
        }
    }

    private void transformLicenses(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Node> nodesByName) {
        // Licenses
        for (final CModeLicense cModeLicense : cModeCluster.getLicenses().values()) {
            final License license = transformLicense(cluster, cModeLicense, nodesByName);
            cluster.getLicenses().add(license);
        }
    }

    private void transformQuotas(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        // Quotas
        for (final CModeQuota cModeQuota : cModeCluster.getQuotas().values()) {


            final VServer vserver = getVServerForChild(cluster, "Quota", cModeQuota.getKey().getOcieUUID(), cModeQuota.getVServerName(), vServerKeysByName);
            if (null == vserver) {
                continue; //Already logged
            }

            boolean qtreeFound = false;
            for (final Volume volume : vserver.getVolumes().values()) {
                if (volume.getVolumeName().equals(cModeQuota.getVolumeName())) {
                    for (final QTree qtree : volume.getQTrees().values()) {
                        if (qtree.getName().equals(cModeQuota.getQTreeID())) {
                            qtreeFound = true;
                            if (TreeQuota.QuotaType.TREE != cModeQuota.getQuotaType()) {
                                final UserQuota userQuota = transformUserQuota(volume, qtree.getKey(), cModeQuota);
                                final List<String> userNameList = new ArrayList<String>(cModeQuota.getQuotaUsers().size());
                                for (final CModeQuotaUser cModeQuotaUser : cModeQuota.getQuotaUsers().values()) {
                                    final QuotaUser quotaUser = transformQuotaUser(userQuota, cModeQuotaUser);
                                    userQuota.getQuotaUsers().add(quotaUser);
                                    userNameList.add(cModeQuotaUser.getQuotaUsername());
                                }

                                userQuota.setUsers(BuilderUtil.getListAsCSVString(userNameList));
                                volume.getUserQuotas().add(userQuota);
                            } else {
                                final TreeQuota treeQuota = transformTreeQuota(volume, qtree.getKey(), cModeQuota);
                                volume.getTreeQuotas().add(treeQuota);
                            }
                            break;
                        }
                    }
                }
            }

            if (!qtreeFound) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find qtree " + cModeQuota.getQTreeID() + " for Quota "
                        + cModeQuota.getVServerName() + " " + cModeQuota.getVolumeName() + " " + cModeQuota.getQuotaTarget()
                        + " (" + cModeQuota.getKey().getOcieUUID() + ")");
            }
        }
    }

    @SuppressWarnings("ConstantConditions")
    private void transformSnapMirrors(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName,
                                      final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey,
                                      final Map<Volume.VolumeKey, Map<String, Snapshot.SnapshotKey>> snapshotKeysByNameByVolumeKey,
                                      final Map<VServer.VServerKey, Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey>> policyKeysByVServer,
                                      final Map<String, JobSchedule> jobSchedulesByName,
                                      final Map<String, Node> nodesByName) {

        // Snap Mirrors
        for (final CModeSnapMirror cModeSnapMirror : cModeCluster.getSnapMirrors().values()) {
            boolean isVServerSnapMirror = SnapMirror.RelationshipGroupType.VSERVER.equals(cModeSnapMirror.getRelationshipGroupType().getEnumValue()) &&
                    cModeSnapMirror.getDestinationVolume() == null && cModeSnapMirror.getSourceVolume() == null;
            // Mirrors are reported only when the destination is local to the target cluster.
            final String targetClusterName = cluster.getName();
            final String sourceVolumeName = cModeSnapMirror.getSourceVolume();
            final String destinationVolumeName = cModeSnapMirror.getDestinationVolume();
            String destinationClusterName = cModeSnapMirror.getDestinationCluster();
            // 8.1 source and dest mirrors are returned
            // 8.2 only dest mirrors are returned, but the destination cluster name is empty
            if (destinationClusterName == null) {
                // can assume that all mirrors are destinations in this cluster (and can assume 8.2)
                destinationClusterName = cluster.getName();

                final String sourceClusterName = getSourceClusterName(cModeCluster, cModeSnapMirror);

                if (sourceClusterName == null) {
                    logger.error(PackageContext.getLogPrefix() + " - skipping SnapMirror with undefined source cluster " + sourceVolumeName + " " + destinationVolumeName);
                    continue;
                }

                // update the snapMirror model with cluster source and destinations
                cModeSnapMirror.setSourceCluster(sourceClusterName);
                cModeSnapMirror.setDestinationCluster(destinationClusterName);

            }
            if (!destinationClusterName.equals(targetClusterName)) {
                continue;       // only report mirrors where the destination is the local target cluster
            }

            final String destinationVServerName = cModeSnapMirror.getDestinationVServer();
            final String destinationVServerUuid = cModeSnapMirror.getDestinationVServerUuid();
            VServer tmp_destinationVServer = null;
            if(destinationVServerUuid!=null)  {   //if having uuid, by uuid
                tmp_destinationVServer = getVServerByUUIDForChild(cluster, "snapmirror", destinationVolumeName, destinationVServerUuid);
            }
            else {  //not having uuid, then by name.
                tmp_destinationVServer = getVServerForChild(cluster, "snapmirror", destinationVolumeName, destinationVServerName, vServerKeysByName);
            }
            if (tmp_destinationVServer == null) {
                // already logged reason
                continue;
            }
            final VServer destinationVServer = tmp_destinationVServer;

            final VServer.VServerKey destinationVServerKey = destinationVServer.getKey();
            Snapshot.SnapshotKey exportedSnapshotKey = null;
            Snapshot.SnapshotKey newestSnapshotKey = null;
            Volume.VolumeKey destinationVolumeKey = null;
            Volume destinationVolume = null;
            if(!isVServerSnapMirror) {
            if (!volumeKeysByNameByVServerKey.containsKey(destinationVServerKey)) {
                logger.warn(PackageContext.getLogPrefix() + " - Failed to find Volumes for destination VServer " + destinationVServerName);
                continue;
            }
            final Map<String, Volume.VolumeKey> destinationVServerVolumeKeysByName = volumeKeysByNameByVServerKey.get(destinationVServerKey);

            if (!destinationVServerVolumeKeysByName.containsKey(destinationVolumeName)) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find SnapMirror destination Volume by name " + destinationVolumeName);
                }
                continue;
            }

            // Correlate snapmirror snapshot references to discovered snapshots only if snapshot discovery is enabled
                destinationVolumeKey = destinationVServerVolumeKeysByName.get(destinationVolumeName);
                destinationVolume = destinationVServer.getVolumes().get(destinationVolumeKey);
            final boolean isSnapshotDiscoveryEnabled = ClusterBuilder.isSnapshotDiscoveryEnabled();
            if (isSnapshotDiscoveryEnabled) {
                final Map<String, Snapshot.SnapshotKey> snapshotKeys = snapshotKeysByNameByVolumeKey.get(destinationVolumeKey);
                if (snapshotKeys != null) {
                    final String exportedSnapshotName = cModeSnapMirror.getExportedSnapshot();
                    exportedSnapshotKey = snapshotKeys.get(exportedSnapshotName);
                    if (exportedSnapshotKey == null) {
                        logger.warn(PackageContext.getLogPrefix() + " - Failed to find snapshot " + exportedSnapshotName
                                + " referenced by snapmirror: " + cModeSnapMirror.getKey().getID());
                    }

                    final String newestSnapshotName = cModeSnapMirror.getNewestSnapshot();
                    newestSnapshotKey = snapshotKeys.get(newestSnapshotName);
                    if (newestSnapshotKey == null) {
                        logger.warn(PackageContext.getLogPrefix() + " - Failed to find snapshot " + newestSnapshotName
                                + " referenced by snapmirror: " + cModeSnapMirror.getKey().getID());
                    }
                }   // else
                    // The SnapMirror appears to be referencing a destination volume for which no snapshots
                    // were reported
                    // logger.warn(PackageContext.getLogPrefix() + " - Failed to find snapshot keys for volume: " + destinationVolumeName + " " + destinationVolumeKey);
            }
            }

            // If source is local, resolve source UUIDs
            final String sourceClusterName = cModeSnapMirror.getSourceCluster();
            final String sourceVServerName = cModeSnapMirror.getSourceVServer();
            final String sourceVServerUuid = cModeSnapMirror.getSourceVServerUuid();

            VServer.VServerKey sourceVServerKey = null;                     // only known if local
            Volume.VolumeKey sourceVolumeKey = null;                        // only known if local
            if (sourceClusterName.equals(cluster.getName())) {
                if(sourceVServerUuid!=null) {  //for 8.3 or later
                    sourceVServerKey = new VServer.VServerKey(cluster.getClusterUUID(), sourceVServerUuid);
                }
                else {  //for ONTAP 8.2 or earlier version
                if (!vServerKeysByName.containsKey(sourceVServerName)) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Failed to find SnapMirror source VServer UUID by name " + sourceVServerName);
                    }

                } else {
                    sourceVServerKey = vServerKeysByName.get(sourceVServerName);
                    }
                }

                if(sourceVServerKey!=null)  {
                    if (!volumeKeysByNameByVServerKey.containsKey(sourceVServerKey)) {
                        if (logger.isDebugEnabled()) {
                            logger.debug(PackageContext.getLogPrefix() + " - Failed to find volumes for SnapMirror source VServer " + sourceVServerName);
                        }
                        sourceVServerKey = null;
                    } else {
                        final Map<String, Volume.VolumeKey> sourceVServerVolumeKeysByName = volumeKeysByNameByVServerKey.get(sourceVServerKey);
                        if (!sourceVServerVolumeKeysByName.containsKey(sourceVolumeName)) {
                            if (logger.isDebugEnabled()) {
                                logger.debug(PackageContext.getLogPrefix() + " - Failed to find SnapMirror source Volume by name " + sourceVolumeName);
                            }

                            // If source volume is unknown, set the vserver ID to null.
                            // Because if the server sees the vServer ID, server assumes volume ID is also defined.
                            sourceVServerKey = null;
                        } else {
                            final VServer sourceVServer = cluster.getVServers().get(sourceVServerKey);
                            final Volume sourceVolume = getVolumeForChild(sourceVServer, "SnapMirror", sourceVolumeName, sourceVolumeName, volumeKeysByNameByVServerKey);
                            sourceVolumeKey = sourceVolume.getKey();
                        }
                    }
                }
            }

            final SnapMirrorPolicy.SnapMirrorPolicyKey policyKey;
            final String policyName = cModeSnapMirror.getPolicyName();
            if (policyName != null) {
                // Currently, name of the snapmirror policy is unique across all the vservers in the same cluster. I will confirm with Akhil that it will stay that way.
                // Correction: two vservers can have policies with the same name. The following link describes the rules:
                // http://wikid.netapp.com/w/C-Mode_SnapMirror_Functional_Specification#SnapMirror_Policy_2
                // Policy is unique within  a set of {cluster-scoped policies, vserver-scoped polices} for any given vserver.

                final Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey> vServerPolicyKeys;
                if (policyKeysByVServer.containsKey(destinationVServerKey)) {
                    vServerPolicyKeys = policyKeysByVServer.get(destinationVServerKey);
                } else {
                    // logger.error("unknown vserver " + destinationVServerKey);
                    vServerPolicyKeys = new HashMap<String, SnapMirrorPolicy.SnapMirrorPolicyKey>();
                }
                final Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey> allKnownPolicyKeysByName = new HashMap<String, SnapMirrorPolicy.SnapMirrorPolicyKey>();
                for (final Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey> perVServerPolicyKeys : policyKeysByVServer.values()) {
                    allKnownPolicyKeysByName.putAll(perVServerPolicyKeys);
                }
                if (vServerPolicyKeys.containsKey(policyName)) {
                    policyKey = vServerPolicyKeys.get(policyName);
                } else if (allKnownPolicyKeysByName.containsKey(policyName)) {
                    policyKey = allKnownPolicyKeysByName.get(policyName);
                } else {
                    policyKey = null;
                    // TODO: turn down debugging after we play with it
                    logger.warn(PackageContext.getLogPrefix() + " - Undefined snapmirror policy: " + policyName);
                }
            } else {
                policyKey = null;
            }

            final String destinationVolumeNodeName = cModeSnapMirror.getDestinationVolumeNode();
            final Node.NodeKey destinationVolumeNodeKey;
            if (destinationVolumeNodeName == null) {
                destinationVolumeNodeKey = null;
            } else {
                final Node destinationVolumeNode = nodesByName.get(destinationVolumeNodeName);
                if (destinationVolumeNode == null) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Undefined destination volume node: " + destinationVolumeNodeName);
                    }
                    destinationVolumeNodeKey = null;
                } else {
                    destinationVolumeNodeKey = destinationVolumeNode.getKey();
                }
            }


            final EnumWrapper<SnapMirror.MirrorState> mirrorStateWrapper = cModeSnapMirror.getMirrorState();
            final EnumWrapper<SnapMirror.RelationshipType> relationshipTypeWrapper = cModeSnapMirror.getRelationshipType();
            final EnumWrapper<SnapMirror.RelationshipStatusType> relationshipStatusWrapper = cModeSnapMirror.getRelationshipStatus();
            final EnumWrapper<SnapMirror.TransferPriorityType> currentTransferPriority = cModeSnapMirror.getCurrentTransferPriority();
            final EnumWrapper<SnapMirror.TransferType> currentTransferType = cModeSnapMirror.getCurrentTransferType();
            final EnumWrapper<SnapMirror.TransferType> lastTransferType = cModeSnapMirror.getLastTransferType();
            final SnapMirror snapMirror = (isVServerSnapMirror) ?
                    cluster.createSnapMirror(destinationVServerKey, cModeSnapMirror.getDestinationLocation()) :
                    cluster.createSnapMirror(destinationVServerKey, destinationVolumeKey, cModeSnapMirror.getDestinationLocation());
            snapMirror.setSourceLocation(cModeSnapMirror.getSourceLocation());
            snapMirror.setDestinationLocation(cModeSnapMirror.getDestinationLocation());
            snapMirror.setSourceCluster(sourceClusterName);
            snapMirror.setSourceVServer(sourceVServerName);
            snapMirror.setSourceVServerKey(sourceVServerKey);
            snapMirror.setSourceVolume(sourceVolumeName);
            snapMirror.setSourceVolumeKey(sourceVolumeKey);
            snapMirror.setExportedSnapshotKey(exportedSnapshotKey);
            snapMirror.setNewestSnapshotKey(newestSnapshotKey);
            snapMirror.setTransferSnapshot(cModeSnapMirror.getTransferSnapshot());
            snapMirror.setMaxTransferRate(cModeSnapMirror.getMaxTransferRate());
            snapMirror.setMirrorState(mirrorStateWrapper.getEnumValue());
            snapMirror.setMirrorStateRaw(mirrorStateWrapper.getRawValue());
            snapMirror.setConstituent(cModeSnapMirror.isConsituent());
            snapMirror.setHealthy(cModeSnapMirror.isHealthy());
            snapMirror.setRelationshipStatusRaw(relationshipStatusWrapper.getRawValue());
            snapMirror.setRelationshipStatus(relationshipStatusWrapper.getEnumValue());
            snapMirror.setRelationshipTypeRaw(relationshipTypeWrapper.getRawValue());
            snapMirror.setRelationshipType(relationshipTypeWrapper.getEnumValue());
            snapMirror.setRelationshipProgress(cModeSnapMirror.getRelationshipProgress());
            snapMirror.setSnapshotCheckpoint(cModeSnapMirror.getSnapshotCheckpoint());
            snapMirror.setSnapshotProgress(cModeSnapMirror.getSnapshotProgress());
            snapMirror.setTries(cModeSnapMirror.getTries());
            snapMirror.setCurrentMaxTransferRate(cModeSnapMirror.getCurrentMaxTransferRate());
            snapMirror.setCurrentTransferError(cModeSnapMirror.getCurrentTransferError());
            snapMirror.setCurrentTransferPriority(currentTransferPriority.getEnumValue());
            snapMirror.setCurrentTransferPriorityRaw(currentTransferPriority.getRawValue());
            snapMirror.setCurrentTransferType(currentTransferType.getEnumValue());
            snapMirror.setCurrentTransferTypeRaw(currentTransferType.getRawValue());
            snapMirror.setLagTime(cModeSnapMirror.getLagTime());
            snapMirror.setLastTransferError(cModeSnapMirror.getLastTransferError());
            snapMirror.setLastTransferType(lastTransferType.getEnumValue());
            snapMirror.setLastTransferTypeRaw(lastTransferType.getRawValue());
            snapMirror.setLastTransferDuration(cModeSnapMirror.getLastTransferDuration());
            snapMirror.setProgressLastUpdated(cModeSnapMirror.getProgressLastUpdated());
            snapMirror.setLastTransferSize(cModeSnapMirror.getLastTransferSize());
            snapMirror.setRelationshipControlPlane(cModeSnapMirror.getRelationshipControlPlane());
            snapMirror.setRelationshipID(cModeSnapMirror.getRelationshipID());
            snapMirror.setPolicyKey(policyKey);
            snapMirror.setLastTransferDuration(cModeSnapMirror.getLastTransferDuration());
            snapMirror.setLastTransferEndErrorCodes(cModeSnapMirror.getLastTransferErrorCodes());
            snapMirror.setLastTransferEndTimestamp(cModeSnapMirror.getLastTransferEndTimestamp());
            snapMirror.setUnhealthyReason(cModeSnapMirror.getUnhealthyReason());
            if (!isVServerSnapMirror) {
            snapMirror.setDestinationVolumeNodeKey(destinationVolumeNodeKey);
            snapMirror.setDestVolumeDerivedStyle(destinationVolume.getDerivedVolumeStyle());
            }
            snapMirror.setOperationId(cModeSnapMirror.getOperationId());
            snapMirror.setNetworkCompressionRatio(cModeSnapMirror.getNetworkCompressionRatio());
            snapMirror.setLastTransferNetworkCompressionRatio(cModeSnapMirror.getLastTransferNetworkCompressionRatio());
            snapMirror.setRelationshipGroupTypeRaw(cModeSnapMirror.getRelationshipGroupTypeRaw());
            snapMirror.setRelationshipGroupType(cModeSnapMirror.getRelationshipGroupType().getEnumValue());
            //If the volume SnapMirror is a child of a vServer SnapMirror then create SnapMirrorSelfRelationship
            if(!isVServerSnapMirror && SnapMirror.RelationshipGroupType.VSERVER.equals(cModeSnapMirror.getRelationshipGroupType().getEnumValue())) {
                cluster.getSnapMirrorSelfRelationships().add(cluster.createSnapMirrorSelfRelationship(snapMirror.getKey().getOcieUUID(), destinationVServerKey.getZapiUUID()));
            }
            final String scheduleName = cModeSnapMirror.getSchedule();
            if (scheduleName != null && !scheduleName.isEmpty()) {
                if (jobSchedulesByName.containsKey(scheduleName)) {
                    final JobSchedule jobSchedule = jobSchedulesByName.get(scheduleName);
                    snapMirror.setScheduleKey(jobSchedule.getKey());
                }
            }
            cluster.getSnapMirrors().add(snapMirror);
        }
    }

    private Map<VServer.VServerKey, Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey>> transformSnapMirrorPolicies(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        final Map<VServer.VServerKey, Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey>> policyKeysByVServer = new HashMap<VServer.VServerKey, Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey>>();
        for (final CModeSnapMirrorPolicy cModeSnapMirrorPolicy : cModeCluster.getSnapMirrorPolicies().values()) {
            final String policyName = cModeSnapMirrorPolicy.getKey().getID(CModeSnapMirrorPolicy.PolicyKey.Field.snapmirror_policy_name);
            final String vServerName = cModeSnapMirrorPolicy.getKey().getID(CModeSnapMirrorPolicy.PolicyKey.Field.vserver_name);
            if (vServerName == null) {
                logger.warn(PackageContext.getLogPrefix() + " - Undefined vServer for Snapshot Policy " + policyName);
                continue;
            }
            final VServer vServer = this.getVServerForChild(cluster, "snapmirror policy", policyName, vServerName, vServerKeysByName);
            if (vServer == null) {
                // already logged reason
                continue;
            }

            final VServer.VServerKey vServerKey = vServer.getKey();
            if (!policyKeysByVServer.containsKey(vServerKey)) {
                policyKeysByVServer.put(vServerKey, new HashMap<String, SnapMirrorPolicy.SnapMirrorPolicyKey>());
            }
            final Map<String, SnapMirrorPolicy.SnapMirrorPolicyKey> policyKeysByName = policyKeysByVServer.get(vServerKey);
            final SnapMirrorPolicy snapMirrorPolicy = cluster.createSnapMirrorPolicy(vServerKey, policyName);
            snapMirrorPolicy.setCreateSnapshot(cModeSnapMirrorPolicy.getCreateSnapshot());
            snapMirrorPolicy.setIgnoreAccessTime(cModeSnapMirrorPolicy.getIgnoreAccessTime());
            snapMirrorPolicy.setNetworkCompressionEnabled(cModeSnapMirrorPolicy.getNetworkCompressionEnabled());
            snapMirrorPolicy.setRestart(cModeSnapMirrorPolicy.getRestart().getEnumValue());
            snapMirrorPolicy.setRestartRaw(cModeSnapMirrorPolicy.getRestart().getRawValue());
            snapMirrorPolicy.setTries(cModeSnapMirrorPolicy.getTries());
            snapMirrorPolicy.setType(cModeSnapMirrorPolicy.getType().getEnumValue());
            snapMirrorPolicy.setTypeRaw(cModeSnapMirrorPolicy.getType().getRawValue());
            snapMirrorPolicy.setComment(cModeSnapMirrorPolicy.getComment());
            snapMirrorPolicy.setTotalKeep(cModeSnapMirrorPolicy.getTotalKeep());
            snapMirrorPolicy.setTransferPriority(cModeSnapMirrorPolicy.getTransferPriority().getEnumValue());
            snapMirrorPolicy.setTransferPriorityRaw(cModeSnapMirrorPolicy.getTransferPriority().getRawValue());
            cluster.getSnapMirrorPolicies().add(snapMirrorPolicy);
            policyKeysByName.put(policyName, snapMirrorPolicy.getKey());

            for (final CModeSnapMirrorRule cModeSnapMirrorRule : cModeSnapMirrorPolicy.getCmodeSnapMirrorRules().values()) {
                final SnapMirrorPolicyRule snapMirrorPolicyRule = snapMirrorPolicy.createSnapMirrorPolicyRule(cModeSnapMirrorRule.getRuleIndex());
                snapMirrorPolicyRule.setKeep(cModeSnapMirrorRule.getKeep());
                snapMirrorPolicyRule.setPreserve(cModeSnapMirrorRule.getPreserve());
                snapMirrorPolicyRule.setLabel(cModeSnapMirrorRule.getLabel());
                snapMirrorPolicyRule.setWarn(cModeSnapMirrorRule.getWarn());
                snapMirrorPolicy.getSnapMirrorPolicyRules().add(snapMirrorPolicyRule);
            }
        }
        return policyKeysByVServer;
    }

    private String getSourceClusterName(final CModeCluster cModeCluster, final CModeSnapMirror cModeSnapMirror) {
        final String sourceVServer = cModeSnapMirror.getSourceVServer();
        final String destVServer = cModeSnapMirror.getDestinationVServer();
        final String sourceClusterName;
        if (sourceVServer.equals(destVServer)) {
            // if the vServer names are the same, then the source cluster is the current cluster (the destination cluster is also the current cluster)
            sourceClusterName = cModeCluster.getClusterName();
        } else {
            // find the source cluster name by looking it up in the vServer peer model
            sourceClusterName = cModeCluster.getVServerPeers().getClusterName(cModeSnapMirror.getSourceVServer(), cModeSnapMirror.getDestinationVServer()/*local */);
        }
        return sourceClusterName;
    }

    private void updateAggregateCapacities(final Cluster cluster) {
        final Map<Aggregate.AggregateKey, Aggregate> aggregatesByKey = new HashMap<Aggregate.AggregateKey, Aggregate>();

        for (final Node node : cluster.getNodes().values()) {
            for (final Aggregate aggregate : node.getAggregates().values()) {
                aggregatesByKey.put(aggregate.getKey(), aggregate);
                // total includes "total + snapshot"
                node.addToAggregateBytesTotal(aggregate.getSizeTotal());
                node.addToAggregateBytesTotal(aggregate.getSnapshotSizeTotal());

                // used includes "used + snapshot"
                node.addToAggregateBytesUsed(aggregate.getSizeUsed());
                node.addToAggregateBytesUsed(aggregate.getSnapshotSizeUsed());
            }
        }
        for (final VServer vServer : cluster.getVServers().values()) {
            for (final Volume volume : vServer.getVolumes().values()) {
                final Aggregate.AggregateKey aggregateKey = volume.getAggregateKey();
                if (aggregateKey == null) {
                    // infinivol don't have aggregate
                    continue;
                }
                final Aggregate aggregate = aggregatesByKey.get(aggregateKey);
                if (aggregate == null) {
                    logger.warn("Failed to update aggregate capacities for unknown aggregate " + aggregateKey.getOcieUUID());
                    continue;
                }

                final Boolean isSISVolume = volume.isSISVolume();
                if (isSISVolume != null && isSISVolume) {
                    final Long dedupSpaceSavings = volume.getDeduplicationSpaceSaved();
                    if (dedupSpaceSavings != null) {
                        aggregate.addToVolumeDedupSpaceSavings(dedupSpaceSavings);
                    }
                    final Long compressionSpaceSavings = volume.getCompressionSpaceSaved();
                    if (compressionSpaceSavings != null) {
                        aggregate.addToVolumeCompressionSpaceSavings(compressionSpaceSavings);
                    }
                }

                if (volume.getState() == Volume.State.ONLINE) {
                    final Long volumeCommittedSize;
                    if (volume.isFlexCache()) {
                        final long sizeUsedBySnapsForCalc = (volume.getSizeUsedBySnapshots() != null) ? volume.getSizeUsedBySnapshots() : 0;
                        final long volUsed = volume.getSizeUsed() + sizeUsedBySnapsForCalc;
                        volumeCommittedSize = Math.max(volUsed, volume.getFlexCacheMinReserve());
                    } else {
                        final long volSize = volume.getSizeTotal() + volume.getSnapshotReserveSize();
                        long maxSize = 0L;
                        if(volume.getAutoSizeMode() != Volume.VolumeAutoSizeMode.OFF && volume.getVolumeType() != Volume.VolumeType.DP) {
                            maxSize = volume.getMaxAutosizeSize();
                        }
                        volumeCommittedSize = Math.max(volSize, maxSize);
                    }
                    aggregate.addToTotalCommitted(volumeCommittedSize);
                }
            }
        }
    }

    private final String getIGroupUUIDByName(final VServer parentVServer, final String iGroupName) {
        String iGroupUUID = null;
        for (final IGroup candidateIGroup : parentVServer.getIGroups().values()) {
            if (iGroupName.equals(candidateIGroup.getName())) {
                iGroupUUID = candidateIGroup.getKey().getID(IGroup.IGroupKey.Field.uuid);
                break;
            }
        }
        return iGroupUUID;
    }

    private final String getQTreeAlias(final String volumeName, final String qtreeName) {
        final String finalQTreeName = (qtreeName == null) ? "" : qtreeName;
        return volumeName + ':' + finalQTreeName;
    }

    private VServer transformVServer(final Cluster cluster, final CModeCluster cModeCluster, final CModeVServer cModeVServer, final Map<String, Aggregate.AggregateKey> aggrKeysByName, Map<String, NetworkIpSpace.NetworkIpSpaceKey> ipSpaceNameToKeyMap) {
        final String clusterUUID = cluster.getClusterUUID();
        final String vServerUUID = cModeVServer.getKey().getUUID();
        final VServer vServer = cluster.createVServer(clusterUUID, vServerUUID);
        final String vServerName = cModeVServer.getName();
        vServer.setName(vServerName);
        vServer.setState(cModeVServer.getState().getEnumValue());
        vServer.setStateRaw(cModeVServer.getState().getRawValue());
        vServer.setOperationalState(cModeVServer.getOperationalState().getEnumValue());
        vServer.setOperationalStateRaw(cModeVServer.getOperationalState().getRawValue());
        vServer.setType(cModeVServer.getType().getEnumValue());
        vServer.setTypeRaw(cModeVServer.getType().getRawValue());
        vServer.setSubtype(cModeVServer.getSubtype().getEnumValue());
        vServer.setSubtypeRaw(cModeVServer.getSubtype().getRawValue());
        vServer.setRootVolumeName(cModeVServer.getRootVolumeName());
        vServer.setRawAllowedProtocols(CModeVServer.protoListToRawCSVString(cModeVServer.getWrappedAllowedProtos()));
        vServer.setAllowedProtocols(cModeVServer.getAllowedProtos());
        vServer.setRawDisAllowedProtocols(CModeVServer.protoListToRawCSVString(cModeVServer.getWrappedDisAllowedProtos()));
        vServer.setDisAllowedProtocols(cModeVServer.getDisAllowedProtos());
        vServer.setMaxVolumes(cModeVServer.getMaxVolumes());
        vServer.setLanguage(cModeVServer.getLanguage().getEnumValue());
        vServer.setLanguageRaw(cModeVServer.getLanguage().getRawValue());
        vServer.setNmSwitch(cModeVServer.getNameServerSwitches());
        vServer.setRepository(cModeVServer.isRepository());
        
        
        final UUIDKey dnsInfoKey = new UUIDKey("dns_info", vServerName);
        final CModeVServerDNSInfo cModeVServerDNSInfo = cModeCluster.getVServerDNSInfos().get(dnsInfoKey);
        if (cModeVServerDNSInfo != null) {
            vServer.setDnsState(cModeVServerDNSInfo.getDnsState().getEnumValue());
            vServer.setDnsServers(cModeVServerDNSInfo.getNameServers());
            vServer.setDnsDomains(cModeVServerDNSInfo.getDomains());
        }
        
        final UUIDKey ddnsInfoKey = new UUIDKey("ddns_info", vServerName);
        final CModeVServerDDNSInfo cModeVServerDDNSInfo = cModeCluster.getVServerDDNSInfos().get(ddnsInfoKey);
        if (cModeVServerDDNSInfo != null) {
            vServer.setDomainName(cModeVServerDDNSInfo.getDomainName());
            vServer.setIsEnabled(cModeVServerDDNSInfo.getIsEnabled());
            vServer.setUseSecure(cModeVServerDDNSInfo.getUseSecure());
            vServer.setTtl(cModeVServerDDNSInfo.getTtl());
        }

        final UUIDKey ldapInfoKey = new UUIDKey("ldap_info", vServerName);
        final CModeVServerLDAPInfo cModeVServerLDAPInfo = cModeCluster.getVServerLDAPInfos().get(ldapInfoKey);
        if (cModeVServerLDAPInfo != null) {
            vServer.setLDAPClientEnabled(cModeVServerLDAPInfo.isClientEnabled());
        }

        final UUIDKey cifsSecInfoKey = new UUIDKey("cifs_sec_info", vServerName);
        final CModeVServerCIFSSecurityInfo cifsSecInfo = cModeCluster.getVServerCIFSSecInfos().get(cifsSecInfoKey);
        if (cifsSecInfo != null) {
            vServer.setIsSmbEncryptionRequired(cifsSecInfo.isSmbEncryptionRequired());
        }

        final String vServerAggregatesCSV = cModeVServer.getAggregatesCSV();
        if (vServerAggregatesCSV != null) {
            final String[] vServerAggregates = vServerAggregatesCSV.split(",");
            for (final String vServerAggregateName : vServerAggregates) {
                if (!aggrKeysByName.containsKey(vServerAggregateName)) {
                    logger.warn(PackageContext.getLogPrefix() + " - Failed to find aggregate policy " + vServerAggregateName + " for vServer " + vServerName);
                } else {
                    final Aggregate.AggregateKey aggregateKey = aggrKeysByName.get(vServerAggregateName);
                    final VServerAggregate vServerAggregate = cluster.createVServerAggregate(vServer.getKey(), aggregateKey);
                    cluster.getVServerAggregates().add(vServerAggregate);
                }
            }
        }

        if (cModeVServer.getNetworkIpSpaceName() != null) {
            final NetworkIpSpace.NetworkIpSpaceKey ipSpaceKey = ipSpaceNameToKeyMap.get(cModeVServer.getNetworkIpSpaceName());
            if (ipSpaceKey != null) {
                vServer.setIpSpaceKey(ipSpaceKey);
            } else {
                logger.info("Unable to find the IPSpace return with the vServer {}", vServer.getName());
            }
        }
        
        if (cModeVServer.getIsConfigLockedForChanges() != null) {
            vServer.setIsConfigLockedForChanges(cModeVServer.getIsConfigLockedForChanges());
            } else {
                logger.info("is-config-locked-for-changes is not returned by the zapi vserver-get-iter for vServer {}", vServer.getName());
            }
        
        return vServer;
    }


    private Volume transformVolume(final Cluster cluster, final VServer vServer, final CModeVolume cModeVolume, final ExportPolicy.ExportPolicyKey exportPolicyKey,
                                   final Map<String, SnapshotPolicy.SnapshotPolicyKey> snapshotPolicyKeysByName,
                                   final Map<String, QosPolicyGroup.QosPolicyGroupKey> qosPolicyGroupKeysByName) {
        final Volume.VolumeStyle volumeStyle = cModeVolume.getVolumeStyle().getEnumValue();
        final Boolean isConstituent = cModeVolume.getConstituent();
        final Volume volume = vServer.createVolume(cModeVolume.getKey().getUUID(), cModeVolume.getInstanceUUID(), volumeStyle, isConstituent);
        volume.setInstanceUUID(cModeVolume.getInstanceUUID());
        volume.setVolumeName(cModeVolume.getVolumeName());
        final String aggregateUUID = cModeVolume.getAggregateUUID();
        final Aggregate.AggregateKey aggregateKey;
        if (aggregateUUID != null) {
            aggregateKey = new Aggregate.AggregateKey(vServer.getKey().getID(VServer.VServerKey.Field.cluster_uuid), cModeVolume.getAggregateUUID());
        } else {
            aggregateKey = null;
        }
        volume.setAggregateKey(aggregateKey);
        volume.setSize(cModeVolume.getSize());
        volume.setVolumeType(cModeVolume.getVolumeType().getEnumValue());
        volume.setVolumeTypeRaw(cModeVolume.getVolumeType().getRawValue());
        final Long volumeBytesTotal = cModeVolume.getSizeTotal();
        if (volumeBytesTotal != null) {
            volume.setSizeTotal(volumeBytesTotal);
            vServer.addToVolumeBytesTotal(volumeBytesTotal);
        }
        final Long volumeBytesUsed = cModeVolume.getSizeUsed();
        if (volumeBytesUsed != null) {
            volume.setSizeUsed(volumeBytesUsed);
            vServer.addToVolumeBytesUsed(volumeBytesUsed);
        }
        final String snapshotPolicyName = cModeVolume.getSnapshotPolicy();
        final SnapshotPolicy.SnapshotPolicyKey snapshotPolicyKey;
        if (snapshotPolicyName == null) {
            snapshotPolicyKey = null;
        } else {
            if (snapshotPolicyKeysByName.containsKey(snapshotPolicyName)) {
                snapshotPolicyKey = snapshotPolicyKeysByName.get(snapshotPolicyName);
            } else {
                snapshotPolicyKey = null;
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find snapshot policy " + snapshotPolicyName + " for volume " + volume.getVolumeName());
                }
            }
        }

        final String qosPolicyGroupName = cModeVolume.getQosPolicyGroupName();
        final QosPolicyGroup.QosPolicyGroupKey qosPolicyGroupKey = qosPolicyGroupKeysByName.get(qosPolicyGroupName);

        // When a Volume is OFFLINE or RESTRICTED, ZAPI reports 0 for
        // snapshot count
        // In this case, do NOT report 0, instead allow the server to maintain
        // the previous reported snapshot count value.
        // Report null, not 0, for snapshot count when state is OFFLINE or RESTRICTED.
        final EnumWrapper<Volume.State> volumeStateWrapper = cModeVolume.getState();
        final Volume.State volumeState = volumeStateWrapper.getEnumValue();
        Long snapshotCount = cModeVolume.getSnapshotCount();
        if (Volume.State.OFFLINE == volumeState ||
                Volume.State.RESTRICTED == volumeState) {
            snapshotCount = null;
        }
        volume.setSnapshotCount(snapshotCount);
        volume.setState(volumeState);
        volume.setStateRaw(volumeStateWrapper.getRawValue());

        volume.setSizeAvailable(cModeVolume.getSizeAvail());
        volume.setVServerRoot(cModeVolume.isVServerRoot());
        volume.setJunctionPath(cModeVolume.getJunctionPath());
        volume.setJunctionActive(cModeVolume.isJunctionActive());
        volume.setVolumeStyleRaw(cModeVolume.getVolumeStyle().getRawValue());
        volume.setSISVolume(cModeVolume.isSISVolume());
        volume.setDataProtectionMirror(cModeVolume.isDataProtectionMirror());
        volume.setLoadSharingMirror(cModeVolume.isLoadSharingMirror());
        volume.setMoveMirror(cModeVolume.isMoveMirror());
        volume.setReplicaVolume(cModeVolume.isReplicaVolume());
        volume.setSpaceGuarantee(cModeVolume.isSpaceGuarantee());
        volume.setVmAlignSector(cModeVolume.getVmAlignSector());
        volume.setVmAlignSuffix(cModeVolume.getVmAlignSuffix());
        volume.setOverwriteReserve(cModeVolume.getOverwriteReserve());
        volume.setOverwriteReserveRequired(cModeVolume.getOverwriteReserveRequired());
        volume.setOverwriteReserveUsed(cModeVolume.getOverwriteReserveUsed());
        volume.setOverwriteReserveActualUsed(cModeVolume.getOverwriteReserveActualUsed());
        volume.setSnapshotReserveSize(cModeVolume.getSnapshotReserveSize());

        volume.setSnapshotCloneDependencyEnabled(cModeVolume.isSnapshotCloneDependencyEnabled());
        volume.setPercentSnapshotReserved(cModeVolume.getPercentSnapshotReserve());
        volume.setPercentSnapshotReserveUsed(cModeVolume.getPercentSnapshotReserveUsed());
        volume.setSizeUsedPercent(cModeVolume.getPercentSizeUsed());
        volume.setSizeAvailableForSnapshots(cModeVolume.getSizeAvailableForSnapshots());
        volume.setPercentFractionalReserve(cModeVolume.getPercentFractionalReserve());
        volume.setExportPolicyKey(exportPolicyKey);
        volume.setDeduplicationSpaceSaved(cModeVolume.getDeduplicationSpaceSaved());
        volume.setCompressionSpaceSaved(cModeVolume.getCompressionSpaceSaved());
        volume.setSecurityStyle(cModeVolume.getSecurityStyle().getEnumValue());
        volume.setSecurityStyleRaw(cModeVolume.getSecurityStyle().getRawValue());
        volume.setHybridCacheEligibility(cModeVolume.getHybridCacheEligibility().getEnumValue());
        volume.setHybridCacheEligibilityRaw(cModeVolume.getHybridCacheEligibility().getRawValue());
        volume.setSpaceMgmtTryFirst(cModeVolume.getSpaceMgmtTryFirst().getEnumValue());
        volume.setSpaceMgmtTryFirstRaw(cModeVolume.getSpaceMgmtTryFirst().getRawValue());
        volume.setSizeUsedBySnapshots(cModeVolume.getSizeUsedBySnapshots());
        volume.setSpaceGuarantee(cModeVolume.getSpaceGuarantee().getEnumValue());
        volume.setSpaceGuaranteeRaw(cModeVolume.getSpaceGuarantee().getRawValue());
        volume.setI2pEnabled(cModeVolume.isI2pEnabled());
        volume.setAutoSnapshotsEnabled(cModeVolume.isAutoSnapshotsEnabled());
        volume.setSnapdirAccessEnabled(cModeVolume.isSnapdirAccessEnabled());

        // "mode" is supported by SN.0 (main), but not RR.1.
        // for RR.1, we use "is-enabled", and map true to GROW and false to OFF.
        // try "mode" first, if not supported then use "is-enabled"
        EnumWrapper<Volume.VolumeAutoSizeMode> autoSizeMode = cModeVolume.getAutoSizeMode();
        String rawValue = autoSizeMode.getRawValue();
        if (cluster.isLessThanSierraNevada() && rawValue == null) {
            // "mode" not supported, use "is-enabled" to drive mode
            final Boolean isEnabled = cModeVolume.getEnabled();
            if (isEnabled != null) {
                final Volume.VolumeAutoSizeMode autoSizeModeEnum = (isEnabled) ? Volume.VolumeAutoSizeMode.GROW : Volume.VolumeAutoSizeMode.OFF;
                rawValue = autoSizeModeEnum.name().toLowerCase();
                autoSizeMode = new EnumWrapper<Volume.VolumeAutoSizeMode>(rawValue, autoSizeModeEnum);
            } else {
                autoSizeMode = new EnumWrapper<Volume.VolumeAutoSizeMode>(null, null);
            }
        }

        volume.setAutoSizeMode(autoSizeMode.getEnumValue());
        volume.setAutoSizeModeRaw(autoSizeMode.getRawValue());
        volume.setMaxAutosizeSize(cModeVolume.getMaxAutosizeSize());
        volume.setAutosizeIncrementSize(cModeVolume.getAutosizeIncrementSize());
        volume.setAtimeUpdateEnabled(cModeVolume.isAtimeUpdateEnabled());
        volume.setBlockType(cModeVolume.getBlockType().getEnumValue());
        volume.setBlockTypeRaw(cModeVolume.getBlockType().getRawValue());
        volume.setInodeFilesTotal(cModeVolume.getInodeFilesTotal());
        volume.setInodeFilesUsed(cModeVolume.getInodeFilesUsed());
        volume.setLanguageCode(cModeVolume.getLanguageCode());
        volume.setCreateUcodeEnabled(cModeVolume.isCreateUcodeEnabled());
        volume.setConvertUcodeEnabled(cModeVolume.isConvertUcodeEnabled());
        volume.setCloneChildCount(cModeVolume.getCloneChildCount());
        volume.setAutoDeleteEnabled(cModeVolume.isAutoDeleteEnabled());
        volume.setCommitment(cModeVolume.getCommitment().getEnumValue());
        volume.setCommitmentRaw(cModeVolume.getCommitment().getRawValue());
        volume.setDeleteOrder(cModeVolume.getDeleteOrder().getEnumValue());
        volume.setDeleteOrderRaw(cModeVolume.getDeleteOrder().getRawValue());
        volume.setDeferDelete(cModeVolume.getDeferDelete().getEnumValue());
        volume.setDeferDeleteRaw(cModeVolume.getDeferDelete().getRawValue());
        volume.setTargetFreeSpace(cModeVolume.getTargetFreeSpace());
        volume.setTrigger(cModeVolume.getTrigger().getEnumValue());
        volume.setTriggerRaw(cModeVolume.getTrigger().getRawValue());
        volume.setPrefix(cModeVolume.getPrefix());
        volume.setDestroyList(cModeVolume.getDestroyList());
        volume.setDestroyListRaw(cModeVolume.getDestroyListRaw());
        volume.setFlexCacheMinReserve(cModeVolume.getFlexCacheMinReserve());
        volume.setOriginVolumeName(cModeVolume.getFlexCacheOriginVolumeName());
        volume.setPercentageCompressionSpaceSaved(cModeVolume.getPercentageCompressionSpaceSaved());
        volume.setPercentageDeduplicationSpaceSaved(cModeVolume.getPercentageDeduplicationSpaceSaved());
        volume.setSnapshotPolicyKey(snapshotPolicyKey);
        volume.setEnableSnapDiff(cModeVolume.isEnableSnapDiff());
        volume.setManagedByService(cModeVolume.isManagedByService());
        volume.setMaxDataConstituentSize(cModeVolume.getMaxDataConstituentSize());
        volume.setMaxNamespaceConstituentSize(cModeVolume.getMaxNamespaceConstituentSize());
        volume.setConstituentRole(cModeVolume.getConstituentRole().getEnumValue());
        volume.setConstituentRoleRaw(cModeVolume.getConstituentRole().getRawValue());
        volume.setQosPolicyGroupKey(qosPolicyGroupKey);
        volume.setUnixSecurityGroupID(cModeVolume.getUnixSecurityGroupID());
        volume.setUnixSecurityPermissions(cModeVolume.getUnixSecurityPermissions());
        volume.setUnixSecurityUserID(cModeVolume.getUnixSecurityUserID());
        volume.setCachingPolicy(cModeVolume.getCachingPolicy());
        volume.setCachingPolicyRaw(cModeVolume.getCachingPolicyRaw());

        return volume;
    }


    private Node transformNode(final Cluster cluster, final CModeCluster cModeCluster, final CModeNodeDetail cModeNodeDetail) {
        // Node information is acquired from multiple ZAPIs:
        //      system-node-get-iter
        //      system-get-node-info-iter
        //      system-get-version
        //      cf-status

        // node-details-info
        final String nodeUUID = cModeNodeDetail.getKey().getUUID();
        final String nodeName = cModeNodeDetail.getNodeNodeName();
        final String model = cModeNodeDetail.getNodeModel();
        final String serialNum = cModeNodeDetail.getNodeSerialNum();
        final Long upTime = cModeNodeDetail.getNodeUptime();
        final String vendor = cModeNodeDetail.getNodeVendor();
        final String location = cModeNodeDetail.getNodeLocation();

        final Boolean isPre8_1_2 = (cModeCluster.getGeneration() <= 8 && cModeCluster.getMajor() <= 1 && (cModeCluster.getMinor() == null || cModeCluster.getMinor() < 2));

        final Boolean envOverTemperature = (isPre8_1_2) ? null : cModeNodeDetail.isEnvOverTemperature();
        final String envFailedFanMessage = cModeNodeDetail.getEnvFailedFanMessage();
        final String envFailedPowerSupplyMessage = cModeNodeDetail.getEnvFailedPowerSupplyMessage();
        final EnumWrapper<Node.BatteryState> nvramBatteryStatus = cModeNodeDetail.getNvramBatteryStatus();
        final Integer failedFanCount = cModeNodeDetail.getFailedFanCount();
        final Integer failedPowerSupplyCount = cModeNodeDetail.getFailedPowerSupplyCount();
        final String cpuFirmwareRelease = cModeNodeDetail.getCpuFirmwareRelease();
        final String owner = cModeNodeDetail.getNodeOwner();
        final String nvramID = cModeNodeDetail.getNvramID();
        final Boolean isHealthy = cModeNodeDetail.isHealthy();
        final Boolean isEpsilonNode = cModeNodeDetail.getIsEpsilonNode();
        final Boolean allFlashOptimized = cModeNodeDetail.getAllFlashOptimized();

        // system-get-version
        Integer generation = null;
        Integer major = null;
        Integer minor = null;

        // node-version-detail-info
        final String version;

        // Version may come from 1 of 3 places.
        // Ideally, it comes from system-node-get-iter.
        // Older ONTAP releases don't provide node version from system-node-get-iter,
        // so we look to system-get-version, which sometimes provides a version for
        // the cluster and each node, and sometimes just a version for the cluster.


        // Make sure the corresponding node-version-detail-info exists
        final Map<String, String> nodeVersions = cModeCluster.getCModeNodeVersions();
        if (cModeNodeDetail.getProductVersion() != null) {
            // Check system-node-get-iter for node version first
            version = cModeNodeDetail.getProductVersion();
        } else if (nodeVersions.containsKey(nodeUUID)) {
            // If node version unavailble from system-node-get-iter, then look for node version from system-get-version
            version = nodeVersions.get(nodeUUID);
        } else {
            // if node version breakdown not available from system-get-version, then use the
            // cluster version.
            // (Per Node version breakdown only exists when different versions exist in the cluster.)
            version = cModeCluster.getVersionString();
        }

        final Matcher versionMatcher = NODE_VERSION_PATTERN.matcher(version);
        if (versionMatcher.matches()) {
            final String generationString = versionMatcher.group(1);
            try {
                generation = Integer.parseInt(generationString);
            } catch (NumberFormatException ignore) {
            }

            final String majorString = versionMatcher.group(2);
            try {
                major = Integer.parseInt(majorString);
            } catch (NumberFormatException ignore) {
            }

            final String minorString = versionMatcher.group(4);
            if (minorString != null && !minorString.isEmpty()) {
                try {
                    minor = Integer.parseInt(minorString);
                } catch (NumberFormatException ignore) {
                }
            } else {
                minor = cModeCluster.getMinor();
            }
        } else {
            generation = cModeCluster.getGeneration();
            major = cModeCluster.getMajor();
            minor = cModeCluster.getMinor();
        }

        // storage-failover-info
        final UUIDKey nodeNameKey = new UUIDKey("node", nodeName);
        final CModeNodeFailover nodeFailover = cModeCluster.getNodeFailovers().get(nodeNameKey);
        final String nodeFailoverState = (nodeFailover != null) ? nodeFailover.getNodeStateString() : null;

        // system-info
        String cpuType = null;
        String cpuID = null;
        Integer cpuCount = null;
        Integer memSize = null;
        String productType = null;

        // cf-status
        Boolean isFailoverEnabled = null;
        String failoverState = null;
        Boolean isTakeoverPossible = null;
        String partnerFirmwareState = null;
        String localFirmwareState = null;
        Boolean isInterconnectUp = null;
        String interconnectLinks = null;
        String givebackState = null;
        String interconnectType = null;
        String currentMode = null;
        String takeoverFailureReason = null;
        String takeoverByPartnerNotPossibleReasonsCSV = null;
        String takeoverOfPartnerNotPossibleReasonsCSV = null;

        // Make sure the corresponding system-info exists
        final CModeNode cModeNode = cModeCluster.getCModeNodes().get(nodeNameKey);
        if (cModeNode != null) {
            cpuType = cModeNode.getCpuTypeString();
            cpuID = cModeNode.getCpuIDString();
            cpuCount = cModeNode.getNumCPUs();
            memSize = cModeNode.getMemSize();
            productType = cModeNode.getProductType();
            isFailoverEnabled = cModeNode.isFailoverEnabled();
            failoverState = cModeNode.getFailoverState();
            isTakeoverPossible = cModeNode.isTakeoverPossible();
            partnerFirmwareState = cModeNode.getPartnerFirmwareState();
            localFirmwareState = cModeNode.getLocalFirmwareState();
            isInterconnectUp = cModeNode.isInterconnectUp();
            interconnectLinks = cModeNode.getInterconnectLinks();
            givebackState = cModeNode.getGivebackState();
            interconnectType = cModeNode.getInterconnectType();
            currentMode = cModeNode.getCurrentMode();
            takeoverFailureReason = cModeNode.getTakeoverFailureReason();

            // Sort the reasons so they are in a deterministic order
            final List<String> takeoverByPartnerNotPossibleReasonRaw = cModeNode.getTakeoverByPartnerNotPossibleReasons();
            final int byPartnerReasounsCount = takeoverByPartnerNotPossibleReasonRaw.size();
            if (takeoverByPartnerNotPossibleReasonRaw != null && byPartnerReasounsCount > 0) {
                final String[] reasons = takeoverByPartnerNotPossibleReasonRaw.toArray(new String[byPartnerReasounsCount]);
                Arrays.sort(reasons);
                final List<String> sortedReasons = Arrays.asList(reasons);
                takeoverByPartnerNotPossibleReasonsCSV = BuilderUtil.getListAsCSVString(sortedReasons);
            }

            final List<String> takeoverOfPartnerNotPossibleReasonRaw = cModeNode.getTakeoverOfPartnerNotPossibleReasons();
            final int ofPartnerReasonsCount = takeoverOfPartnerNotPossibleReasonRaw.size();
            if (takeoverOfPartnerNotPossibleReasonRaw != null && ofPartnerReasonsCount > 0) {
                final String[] reasons = takeoverOfPartnerNotPossibleReasonRaw.toArray(new String[ofPartnerReasonsCount]);
                Arrays.sort(reasons);
                final List<String> sortedReasons = Arrays.asList(reasons);
                takeoverOfPartnerNotPossibleReasonsCSV = BuilderUtil.getListAsCSVString(sortedReasons);
            }
        } else {
            logger.info(PackageContext.getLogPrefix() + " - No system-info for node " + nodeName + " " + nodeUUID);
        }

        //metrocluster node info
        final UUIDKey nodeUUIDKey = new UUIDKey("node", nodeUUID);
        final CModeNodeMetrocluster nodeMetrocluster = cModeCluster.getNodeMetroclusters().get(nodeUUIDKey);
        String mtDrPartnerClusterUUID = null;
        String mtDrPartnerNodeUUID = null;
        EnumWrapper<Node.MtDrOperationState>  mtDrOperationState = null;
        Boolean mtAutomaticUso = null;

        if(nodeMetrocluster!=null)   {
            mtDrOperationState = nodeMetrocluster.getMtDrOperationState();
            mtDrPartnerNodeUUID = nodeMetrocluster.getMtDrPartnerNodeUUID();
            mtDrPartnerClusterUUID = nodeMetrocluster.getMtDrPartnerClusterUUID();
            mtAutomaticUso = nodeMetrocluster.getMtAutomaticUso();
        }

        final Node node = cluster.createNode(nodeUUID);
        node.setName(nodeName);
        node.setModel(model);
        node.setSerialNum(serialNum);
        node.setUptime(upTime);
        node.setVendor(vendor);
        node.setLocation(location);
        node.setEnvOverTemperature(envOverTemperature);
        node.setEnvFailedFanMessage(envFailedFanMessage);
        node.setEnvFailedPowerSupplyMessage(envFailedPowerSupplyMessage);
        node.setNvramBatteryStatus(nvramBatteryStatus.getEnumValue());
        node.setNvramBatteryStatusRaw(nvramBatteryStatus.getRawValue());
        node.setVersion(version);
        node.setFailoverState(nodeFailoverState);
        node.setCpuType(cpuType);
        node.setCpuID(cpuID);
        node.setCpuCount(cpuCount);
        node.setMemSize(memSize);
        node.setProductType(productType);
        node.setVersionGeneration(generation);
        node.setVersionMajor(major);
        node.setVersionMinor(minor);
        // node.setSystemID(systemID);
        // node.setPartnerID(partnerID);
        // node.setPartnerName(partnerName);
        node.setFailoverEnabled(isFailoverEnabled);
        node.setFailoverState(failoverState);
        node.setTakeoverPossible(isTakeoverPossible);
        node.setPartnerFirmwareState(partnerFirmwareState);
        node.setLocalFirmwareState(localFirmwareState);
        node.setInterconnectUp(isInterconnectUp);
        node.setInterconnectLinks(interconnectLinks);
        node.setInterconnectType(interconnectType);
        // node.setInterconnectName(interconnectName);
        node.setGiveBackState(givebackState);
        node.setCurrentMode(currentMode);
        node.setTakeoverFailureReason(takeoverFailureReason);
        node.setTakeoverByPartnerNotPossibleReasons(takeoverByPartnerNotPossibleReasonsCSV);
        node.setTakeoverOfPartnerNotPossibleReasons(takeoverOfPartnerNotPossibleReasonsCSV);
        node.setFailedFanCount(failedFanCount);
        node.setFailedPowerSupplyCount(failedPowerSupplyCount);
        node.setCpuFirmwareRelease(cpuFirmwareRelease);
        node.setOwner(owner);
        node.setNvramID(nvramID);
        node.setHealthy(isHealthy);
        node.setIsEpsilonNode(isEpsilonNode);
        node.setAllFlashOptimized(allFlashOptimized);

        if(cModeCluster.getIsMetrocluster())  {
            node.setMtDrPartnerNodeUUID(mtDrPartnerNodeUUID);
            node.setMtDrOperationState(mtDrOperationState.getEnumValue());
            node.setMtDrOperationStateRaw(mtDrOperationState.getRawValue());
            node.setMtAutomaticUso(mtAutomaticUso);
            cluster.setMtDrPartnerClusterUUID(mtDrPartnerClusterUUID);
        }

        // Add processor objects
        if (null != cModeNode) {
            final Processors processors = node.getProcessors();
            for (final CModeProcessor processor : cModeNode.getProcessors().values()) {
                final Integer processorID = processor.getProcessorNumber();
                processors.add(node.createProcessor(processorID));
            }
        }

        // Add flash devices
        if (null != cModeNode) {
            final FlashDevices flashDevices = node.getFlashDevices();
            for (final CModeFlashDevice cModeFlashDevice : cModeNode.getFlashDevices().values()) {
                final FlashDevice flashDevice = node.createFlashDevice(cModeFlashDevice.getSerialNumber());
                flashDevice.setSlotNumber(cModeFlashDevice.getSlotNumber());
                flashDevice.setModelName(cModeFlashDevice.getModelName());
                flashDevice.setFirmwareRevision(cModeFlashDevice.getFirmwareRevision());
                flashDevice.setCapacityGB(cModeFlashDevice.getCapacity());
                flashDevice.setPercentOnline(cModeFlashDevice.getPercentOnline());
                flashDevice.setStatus(cModeFlashDevice.getStatus().getEnumValue());
                flashDevice.setStatusRaw(cModeFlashDevice.getStatus().getRawValue());
                flashDevice.setServiceTime(cModeFlashDevice.getServiceTime());
                flashDevice.setThresholdProfile(cModeFlashDevice.getThresholdProfile());
                flashDevices.add(flashDevice);
            }
        }

        return node;
    }

    /**
     *
     * @param cluster
     * @param homeNode
     * @param ownerNode
     * @param activeNode
     * @param aggrKeysByName
     * @param aggrSpare Identifies the spare space on {@code cModeDisk}.  If
     *         {@code null}, no information was available.
     * @param cModeDisk
     * @param knownRaidGroupKeys
     * @param storagePoolByName
     * @return
     */
    @SuppressWarnings("ConstantConditions")
    private Disk transformDisk(final Cluster cluster, final CModeCluster cModeCluster, final Node homeNode, final Node ownerNode,
            final Node activeNode, final Map<String, Aggregate.AggregateKey> aggrKeysByName,
            final CModeAggregateSpare aggrSpare, final CModeDisk cModeDisk,
            final Set<RaidGroup.RaidGroupKey> knownRaidGroupKeys,
            final Map<String, StoragePool> storagePoolByName,
            final Set<String> localShelfIds)
    {
        final String clusterUUID = cluster.getClusterUUID();
        final String aggrName = cModeDisk.getAggregateName();
        // Lookup Aggregate & Raid Group refs
        final Aggregate.AggregateKey aggregateKey;
        if (aggrName == null) {
            aggregateKey = null;
        } else if (!aggrKeysByName.containsKey(aggrName)) {
            aggregateKey = null;
        } else {
            aggregateKey = aggrKeysByName.get(aggrName);
        }
        final String aggregateUUID = (aggregateKey == null) ? null : aggregateKey.getID(Aggregate.AggregateKey.Field.uuid);
        // Check if RaidGroup is defined by disk and known
        final String raidGroupName = cModeDisk.getRaidGroupName();
        RaidGroup.RaidGroupKey raidGroupKey = null;

        if (raidGroupName != null && aggregateKey != null) {
            final String plexName = cModeDisk.getPlexName();
            final String plexPath = "/" + aggrName + '/' + plexName;
            final Plex.PlexKey plexKey = new Plex.PlexKey(clusterUUID, aggregateKey.getID(Aggregate.AggregateKey.Field.uuid), plexPath);
            final String raidGroupPath = plexPath + '/' + raidGroupName;
            final RaidGroup.RaidGroupKey candidateRaidGroupKey = new RaidGroup.RaidGroupKey(clusterUUID, aggregateUUID, plexKey.getID(Plex.PlexKey.Field.plex_name), raidGroupPath);
            if (knownRaidGroupKeys.contains(candidateRaidGroupKey)) {
                raidGroupKey = candidateRaidGroupKey;
            }
        }

        final Disk disk = cluster.createDisk(cModeDisk.getKey().getUUID());
        disk.setAggregateKey(aggregateKey);
        if (homeNode != null) {
            disk.setHomeNodeKey(homeNode.getKey());
            final Long diskBlocksTotal = cModeDisk.getPhysicalBlocks();
            if (diskBlocksTotal != null) {
                homeNode.addToDiskSizeTotal(diskBlocksTotal * BYTES_PER_BLOCK);
            }
            final Long diskBlocksUsed = cModeDisk.getUsedBlocks();
            if (diskBlocksUsed != null) {
                homeNode.addToDiskSizeUsed(diskBlocksUsed * BYTES_PER_BLOCK);
            }
        }
        if (ownerNode != null) {
            disk.setOwnerNodeKey(ownerNode.getKey());
        }
        if (activeNode != null) {
            disk.setActiveNodeKey(activeNode.getKey());
        }
        disk.setBay(cModeDisk.getBay());
        disk.setBytesPerSector(cModeDisk.getBytesPerSector());
        disk.setContainerType(cModeDisk.getContainerType().getEnumValue());
        disk.setContainerTypeRaw(cModeDisk.getContainerType().getRawValue());
        disk.setFailed(cModeDisk.isFailed());
        disk.setFirmwareRevision(cModeDisk.getFirmwareRev());
        disk.setDiskType(cModeDisk.getDiskType().getEnumValue());
        disk.setDiskTypeRaw(cModeDisk.getDiskType().getRawValue());
        disk.setInterfaceType(cModeDisk.getEffectiveDiskType().getEnumValue());
        disk.setInterfaceTypeRaw(cModeDisk.getEffectiveDiskType().getRawValue());
        disk.setModel(cModeDisk.getModel());
        disk.setName(cModeDisk.getDiskName());
        final Long physicalBlocks = cModeDisk.getPhysicalBlocks();
        disk.setPhysicalBlocks(physicalBlocks);
        if (physicalBlocks != null) {
            disk.setPhysicalBytes(physicalBlocks * BYTES_PER_BLOCK);
        }
        disk.setRaidGroupKey(raidGroupKey);
        disk.setRaidPosition(cModeDisk.getPosition().getEnumValue());
        disk.setRaidPositionRaw(cModeDisk.getPosition().getRawValue());
        final StoragePool storagePool = storagePoolByName.get(cModeDisk.getStoragePoolName());
        if(storagePool != null) {
            disk.setStoragePoolKey(storagePool.getKey());
        }
        disk.setRpm(cModeDisk.getRPM());
        disk.setSerialNumber(cModeDisk.getSerialNum());
        disk.setShelf(cModeDisk.getShelf());
        final Long usedBlocks = cModeDisk.getUsedBlocks();
        disk.setUsedBlocks(usedBlocks);
        if (usedBlocks != null) {
            disk.setUsedBytes(usedBlocks * BYTES_PER_BLOCK);
        }
        disk.setVendor(cModeDisk.getVendor());
        disk.setVirtual(cModeDisk.isVirtual());
        disk.setZeroed(cModeDisk.isZeroed());
        disk.setZeroing(cModeDisk.isZeroing());
        disk.setPoolID(cModeDisk.getPoolID());
        disk.setZeroingPercentComplete(cModeDisk.getZeroingPercentComplete());
        disk.setChecksumCompatibility(cModeDisk.getChecksumCompatibility());
        disk.setMediaScrubbing(cModeDisk.isMediaScrubbing());
        disk.setOffline(cModeDisk.isOffline());
        disk.setPreFailed(cModeDisk.isPreFailed());
        disk.setReconstructing(cModeDisk.isReconstructing());
        disk.setReplacing(cModeDisk.isReplacing());
        disk.setShared(cModeDisk.getShared());
        disk.setInFDR(cModeDisk.isInFDR());
        disk.setFailedReason(cModeDisk.getFailedReason());
        disk.setRightSizeSectors(cModeDisk.getRightSizeSectors());
        disk.setCapacitySectors(cModeDisk.getCapacitySectors());
        if(CAMPUS.equalsIgnoreCase(cluster.getMtConfigurationType()))  {
            setLocalAttachForDisk(disk, localShelfIds);
        }
        else {
            disk.setIsLocalAttach(cModeDisk.getIsLocalAttach());
        }
        disk.setPercentSparesConsumed(cModeDisk.getPercentSparesConsumed());
        disk.setPartitioningType(cModeDisk.getPartitioningType().getEnumValue());
        disk.setPartitioningTypeRaw(cModeDisk.getPartitioningType().getRawValue());
        disk.setPercentRatedLifeUsed(cModeDisk.getPercentRatedLifeUsed());

        // entry efficiency; the following augmentation identifies disks that
        // have space on their data/root partitions for additional provisioning
        if (aggrSpare != null) {
            disk.setUsableDataSize(aggrSpare.getUsableDataSize());
            disk.setUsableRootSize(aggrSpare.getUsableRootSize());
        }

        return disk;
    }

    private Aggregate transformAggregate(final Node node, final CModeAggregate cModeAggr, final Set<RaidGroup.RaidGroupKey> raidGroupKeys) {
        final Aggregate aggregate = node.createAggregate(cModeAggr.getKey().getUUID());
        aggregate.setName(cModeAggr.getAggrName());
        aggregate.setType(cModeAggr.getAggrType());
        aggregate.setHybrid(cModeAggr.isHybrid());
        aggregate.setUsesSharedDisks(cModeAggr.getUsesSharedDisks());
        aggregate.setHybridEnabled(cModeAggr.isHybridEnabled());
        final EnumWrapper<Aggregate.AggregateState> aggrStateWrapper = cModeAggr.getAggrState();
        final Aggregate.AggregateState aggrState = aggrStateWrapper.getEnumValue();
        aggregate.setState(aggrState);
        aggregate.setStateRaw(aggrStateWrapper.getRawValue());
        final EnumWrapper<Aggregate.AggregateType> aggregateType = cModeAggr.getAggregateType();
        aggregate.setAggregateType(aggregateType.getEnumValue());
        aggregate.setAggregateTypeRaw(aggregateType.getRawValue());

        final Integer sizeUsedPercent = filterAggregatePercent(aggrState, cModeAggr.getSizeUsedPercent());
        aggregate.setSnapshotSizeTotal(filterAggregateCapacity(aggrState, cModeAggr.getSnapshotSizeTotal()));
        aggregate.setSnapshotSizeUsed(filterAggregateCapacity(aggrState, cModeAggr.getSnapshotSizeUsed()));
        aggregate.setSnapshotSizeAvailable(filterAggregateCapacity(aggrState, cModeAggr.getSnapShotSizeAvailable()));
        aggregate.setSizeTotal(filterAggregateCapacity(aggrState, cModeAggr.getSizeTotal()));
        aggregate.setSizeAvailable(filterAggregateCapacity(aggrState, cModeAggr.getSizeAvailable()));
        aggregate.setSizeUsed(filterAggregateCapacity(aggrState, cModeAggr.getSizeUsed()));
        aggregate.setSizeUsedPercent(sizeUsedPercent);
        aggregate.setTotalReservedSpace(filterAggregateCapacity(aggrState, cModeAggr.getTotalReservedSpace()));
        if (!Aggregate.shouldFreezeValues(aggrState)) {
            if (sizeUsedPercent != null && sizeUsedPercent != 0) {
                // Size Available is derived from size used percent, only report if the volume is not offline
                aggregate.setSizeAvailPercent(cModeAggr.getSizeAvailPercent());
            }
        }

        final String raidStatusRaw = cModeAggr.getRaidStatus();
        aggregate.setRaidStatus(raidStatusRaw);
        String raidTypeRaw = cModeAggr.getRaidType();
        Aggregate.RaidType raidType = null;
        if(raidTypeRaw == null) {
            //For OnTap 8.1 & lover use raid-status attribute to determine the RAID type
            raidType = determineRaidType(raidStatusRaw);
            // Raid Type Raw is not the raw value of raid status, but rather the recognized
            // raid type from raid status, i.e. one of raid0, raid4, raid_dp, mixed_raid_type,
            // or null if raid status does not define a raid type (e.g. "reconstruct,growing").
            switch (raidType) {
                case NOT_MAPPED:
                    raidTypeRaw = null;
                    break;
                default:
                    raidTypeRaw = raidType.name().toLowerCase();
            }
        } else {
            //For OnTap 8.2 & higher use raid-type attribute
            raidType = BuilderUtil.getEnumWrapperByValue(Aggregate.RaidType.class, raidTypeRaw, Aggregate.RaidType.NOT_MAPPED).getEnumValue();
        }
        aggregate.setRaidType(raidType);
        aggregate.setRaidTypeRaw(raidTypeRaw);
        aggregate.setRaidSize(cModeAggr.getRaidSize());
        aggregate.setBlockType(cModeAggr.getBlockType().getEnumValue());
        aggregate.setBlockTypeRaw(cModeAggr.getBlockType().getRawValue());
        aggregate.setIs64BitUpgradeInProgress(cModeAggr.isIs64UpgradeInProgress());
        aggregate.setMirrorStatus(cModeAggr.getMirrorStatus().getEnumValue());
        aggregate.setMirrorStatusRaw(cModeAggr.getMirrorStatus().getRawValue());
        aggregate.setHybridCacheSizeTotal(cModeAggr.getHybridCacheSizeTotal());
        aggregate.setHasLocalRoot(cModeAggr.hasLocalRoot());
        aggregate.setHasPartnerRoot(cModeAggr.hasPartnerRoot());
        transformPlexes(cModeAggr, aggregate, raidGroupKeys);
        return aggregate;
    }

    /**
     * Don't report 0 capacity for offline aggregates.   When an aggregate goes offline, we want to
     * preserve the last known capacity values.   When the aggregate is offline, ZAPI reports 0 for
     * capacity values.   If the aggregate is offline and the capacity is 0, don't report the capacity
     * at all so that the server can retain the last known capacity values.
     *
     * @param aggrState current state of the aggregate
     * @param capacity  an aggregate capacity value
     */
    private Long filterAggregateCapacity(final Aggregate.AggregateState aggrState, final Long capacity) {
        Long returnVal = capacity;
        if (capacity == null) {
            return returnVal;
        }
        if (Aggregate.shouldFreezeValues(aggrState)) {
            if (capacity.equals(0L)) {
                returnVal = null;
            }
        }
        return returnVal;
    }

    /**
     * Don't report 0 capacity for offline aggregates.   When an aggregate goes offline, we want to
     * preserve the last known capacity values.   When the aggregate is offline, ZAPI reports 0 for
     * capacity values.   If the aggregate is offline and the capacity is 0, don't report the capacity
     * at all so that the server can retain the last known capacity values.
     *
     * @param aggrState       current state of the aggregate
     * @param capacityPercent an aggregate capacity percent value
     */
    private Integer filterAggregatePercent(final Aggregate.AggregateState aggrState, final Integer capacityPercent) {
        Integer returnVal = capacityPercent;
        if (capacityPercent == null) {
            return returnVal;
        }
        if (Aggregate.shouldFreezeValues(aggrState)) {
            if (capacityPercent.equals(0)) {
                returnVal = null;
            }
        }
        return returnVal;
    }

    private void transformPlexes(final CModeAggregate cModeAggregate, final Aggregate aggregate, final Set<RaidGroup.RaidGroupKey> raidGroupKeys) {
        for (final CModePlex cModePlex : cModeAggregate.getPlexes().values()) {
            final Plex plex = aggregate.createPlex(cModePlex.getKey().getID(CModePlex.PlexKey.Field.plex_name), cModePlex.isOnline());
            plex.setStatus(cModePlex.getStatus().getEnumValue());
            plex.setStatusRaw(cModePlex.getStatus().getRawValue());
            plex.setResyncPercent(cModePlex.getResyncPercentage());
            aggregate.getPlexes().add(plex);
            transformRaidGroups(cModePlex, plex, raidGroupKeys);
        }
    }

    private void transformRaidGroups(final CModePlex cModePlex, final Plex plex, final Set<RaidGroup.RaidGroupKey> raidGroupKeys) {
        for (final CModeRaidGroup cModeRaidGroup : cModePlex.getRaidGroups().values()) {
            final String raidGroupName = cModeRaidGroup.getKey().getID(CModeRaidGroup.RaidGroupKey.Field.raidgroup_name);
            final RaidGroup raidGroup = plex.createRaidGroup(raidGroupName);
            raidGroup.setReconstructing(cModeRaidGroup.isReconstructing());
            if(cModeRaidGroup.getRaidType() != null) {
                raidGroup.setRaidType(cModeRaidGroup.getRaidType().getEnumValue());
                raidGroup.setRaidTypeRaw(cModeRaidGroup.getRaidType().getRawValue());
            }
            plex.getRaidGroups().add(raidGroup);
            raidGroupKeys.add(raidGroup.getKey());    // to test if raid group is known when transforming disks
        }
    }

    private NetworkPort transformNetInterface(final Node node, final CModeNetInterface cModePort) {
        final NetworkPort networkPort = node.createPort(cModePort.getKey().getID(CModeNetInterface.IFKey.Field.interface_name), cModePort.getMacAddress(), cModePort.getPortType().getEnumValue());
        networkPort.setPortTypeRaw(cModePort.getPortType().getRawValue());
        networkPort.setPortRole(cModePort.getPortRole().getEnumValue());
        networkPort.setRoleRaw(cModePort.getPortRole().getRawValue());
        networkPort.setLinkStatus(cModePort.getLinkStatus().getEnumValue());
        networkPort.setLinkStatusRaw(cModePort.getLinkStatus().getRawValue());
        networkPort.setOperSpeed(cModePort.getOperSpeed().getEnumValue());
        networkPort.setOperSpeedRaw(cModePort.getOperSpeed().getRawValue());
        networkPort.setAdminSpeed(cModePort.getAdminSpeed().getEnumValue());
        networkPort.setAdminSpeedRaw(cModePort.getAdminSpeed().getRawValue());
        networkPort.setMtu(cModePort.getMTU());
        networkPort.setVlanID(cModePort.getVlanID());
        networkPort.setVlanPort(cModePort.getVlanPort());
        networkPort.setIfGroupPort(cModePort.getIfGroupPort());
        networkPort.setOperationalDuplex(cModePort.getOperationalDuplex().getEnumValue());
        networkPort.setOperationalDuplexRaw(cModePort.getOperationalDuplex().getRawValue());
        return networkPort;
    }

    private FcpPort transformFcpAdapter(final CModeCluster cModeCluster, final Node node, final CModeFcpAdapter cModeFcpAdapter) {
        final CModeFcpAdapter.FCPKey fcpKey = cModeFcpAdapter.getKey();
        final CModeAdapter.AdapterKey statusKey = new CModeAdapter.AdapterKey(node.getName(), fcpKey.getID(CModeFcpAdapter.FCPKey.Field.adapter_name));
        final CModeAdapterStatus cModeAdapterStatus = cModeCluster.getAdapterStatuses().get(statusKey);
        final EnumWrapper<FcpPort.Status> status;
        if (cModeAdapterStatus != null) {
            status = cModeAdapterStatus.getStatus();
        } else {
            // some vsims return 0 records for fc-config-list-info while storage-adapter-get-adapter-info returns records
            status = new EnumWrapper<FcpPort.Status>(null, null);
        }

        final String adapterName = cModeFcpAdapter.getKey().getID(CModeFcpAdapter.FCPKey.Field.adapter_name);
        final FcpPort fcpPort = node.createFcpPort(cModeFcpAdapter.getNodeWWN(), cModeFcpAdapter.getPortWWN(), adapterName);
        fcpPort.setFabricEstablished(cModeFcpAdapter.getFabricEstablished());
        fcpPort.setFirmwareRev(cModeFcpAdapter.getFirmwareRev());
        fcpPort.setInfoName(cModeFcpAdapter.getInfoName());
        fcpPort.setDataLinkRate(cModeFcpAdapter.getDataLinkRate());
        fcpPort.setMaxSpeed(cModeFcpAdapter.getMaxSpeed());
        fcpPort.setMediaType(cModeFcpAdapter.getMediaType().getEnumValue());
        fcpPort.setMediaTypeRaw(cModeFcpAdapter.getMediaType().getRawValue());
        fcpPort.setNodeWWN(cModeFcpAdapter.getNodeWWN());
        fcpPort.setPhysicalProtocol(cModeFcpAdapter.getPhysicalProtocol().getEnumValue());
        fcpPort.setPhysicalProtocolRaw(cModeFcpAdapter.getPhysicalProtocol().getRawValue());
        fcpPort.setSpeed(cModeFcpAdapter.getSpeed().getEnumValue());
        fcpPort.setSpeedRaw(cModeFcpAdapter.getSpeed().getRawValue());
        fcpPort.setState(cModeFcpAdapter.getState().getEnumValue());
        fcpPort.setStateRaw(cModeFcpAdapter.getState().getRawValue());
        fcpPort.setStatus(status.getEnumValue());
        fcpPort.setStatusRaw(status.getRawValue());
        fcpPort.setSwitchPort(cModeFcpAdapter.getSwitchPort());
        return fcpPort;
    }

    private StorageAdapter transformStorageAdapter(final CModeCluster cModeCluster, final Node node, final CModeAdapter cModeAdapter) {
        final CModeAdapter.AdapterKey adapterKey = cModeAdapter.getKey();
        final CModeAdapterStatus cModeAdapterStatus = cModeCluster.getAdapterStatuses().get(adapterKey);
        final StorageAdapter storageAdapter = node.createStorageAdapter(adapterKey.getID(CModeAdapter.AdapterKey.Field.adapter_name));
        storageAdapter.setInUse(cModeAdapter.getInUse());
        storageAdapter.setModel(cModeAdapter.getModel());
        storageAdapter.setNodeWWN(cModeAdapter.getNodeWWN());
        storageAdapter.setPortWWN(cModeAdapter.getPortWWN());
        if (cModeAdapterStatus != null) {
            storageAdapter.setStatus(StorageAdapter.Status.valueOf(cModeAdapterStatus.getStatus().getEnumValue().toString()));
        }
        storageAdapter.setType(cModeAdapter.getType().getEnumValue());

        storageAdapter.setCableEndIdentifier(cModeAdapter.getCableEndIdentifier());
        storageAdapter.setCableLength(cModeAdapter.getCableLength());
        storageAdapter.setCableManufacturer(cModeAdapter.getCableManufacturer());
        storageAdapter.setCablePartNumber(cModeAdapter.getCablePartNumber());
        storageAdapter.setCableSerialNumber(cModeAdapter.getCableSerialNumber());
        storageAdapter.setAttachedSerialNumber(cModeAdapter.getAttachedSerialNumber());
        storageAdapter.setCableTechnology(cModeAdapter.getCableTechnology());
        final EnumWrapper<StorageAdapter.State> adapterState = cModeAdapter.getAdapterState();
        if (adapterState != null) {
            storageAdapter.setState(adapterState.getEnumValue());
            storageAdapter.setStateRaw(adapterState.getRawValue());
        }

        return storageAdapter;
    }

    private QTree transformQTree(final Volume volume, final CModeQTree cModeQTree, final Map<VServer.VServerKey, Map<String, ExportPolicy.ExportPolicyKey>> exportPolicyKeysByNameByVServerKey) {
        final String volumeUUID = volume.getInstanceUUID();
        final QTree qTree = volume.createQTree(cModeQTree.getQTreeID(), volumeUUID);
        qTree.setOpLocks(cModeQTree.getOpLocksEnabled());
        qTree.setqTreeName(cModeQTree.getQTreeName());
        qTree.setSecurityStyle(cModeQTree.getSecurityStyle().getEnumValue());
        qTree.setSecurityStyleRaw(cModeQTree.getSecurityStyle().getRawValue());
        qTree.setStatus(cModeQTree.getStatus().getEnumValue());
        qTree.setStatusRaw(cModeQTree.getStatus().getRawValue());

        final Boolean isExportPolicyInherited = cModeQTree.getExportPolicyInherited();
        final ExportPolicy.ExportPolicyKey policyKey;
        if (isExportPolicyInherited != null && isExportPolicyInherited) {
            policyKey = volume.getExportPolicyKey();
        } else {
            final VServer parentVServer = volume.getParent();
            final String exportPolicyName = cModeQTree.getExportPolicyName();
            if (exportPolicyKeysByNameByVServerKey.containsKey(parentVServer.getKey())) {
                policyKey = exportPolicyKeysByNameByVServerKey.get(parentVServer.getKey()).get(exportPolicyName);
            } else {
                policyKey = null;
            }

            if (policyKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Undefined export policy " + exportPolicyName + " for qtree " + cModeQTree.getQTreeName());
                }
            }
        }
        qTree.setExportPolicyKey(policyKey);
        qTree.setExportPolicyInherited(cModeQTree.getExportPolicyInherited());

        return qTree;
    }

    private LUN transformLUN(final Volume volume, final CModeLUN cModeLUN, final QTree.QTreeKey qTreeKey, final QosPolicyGroup.QosPolicyGroupKey qosPolicyGroupKey) {
        final LUN lun = volume.createLUN(cModeLUN.getKey().getUUID());
        lun.setAlignment(cModeLUN.getAlignment().getEnumValue());
        lun.setAlignmentRaw(cModeLUN.getAlignment().getRawValue());
        lun.setComment(cModeLUN.getComment());
        lun.setMapped(cModeLUN.isMapped());
        lun.setMultiprotocolType(cModeLUN.getMultiprotocolTypeString());
        lun.setOnline(cModeLUN.isOnline());
        lun.setPath(cModeLUN.getPath());
        lun.setPrefixSize(cModeLUN.getPrefixSize());
        lun.setQTreeKey(qTreeKey);
        lun.setSerialNumber(cModeLUN.getSerialNumber());
        lun.setSize(cModeLUN.getSize());
        lun.setSizeUsed(cModeLUN.getSizeUsed());
        lun.setSpaceAllocEnabled(cModeLUN.isSpaceAllocEnabled());
        lun.setSpaceReservationEnabled(cModeLUN.isSpaceReservationEnabled());
        lun.setSuffixSize(cModeLUN.getSuffixSize());
        lun.setQosPolicyGroupKey(qosPolicyGroupKey);
        lun.setLunClass(cModeLUN.getLunClass().getEnumValue());
        lun.setLunClassRaw(cModeLUN.getLunClass().getRawValue());
        return lun;
    }

    private NetworkLif transformNetworkLif(final VServer parentVServer, final CModeLif cModeLif, final Node.NodeKey currentNodeKey, final NetworkPort.PortKey currentPortKey,
                                           final Node.NodeKey homeNodeKey, final NetworkPort.PortKey homePortKey, final RoutingGroup.RoutingGroupKey routingGroupKey,
                                           final Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> failoverGroupKeyByName ) {

        final String vServerUUID = parentVServer.getKey().getOcieUUID();
        final String interfaceName = cModeLif.getInterfaceName();
        final String legacyUUID = vServerUUID + '.' + interfaceName;
        final String lifUUID = cModeLif.getLifUUID();
        final String lifID = (lifUUID == null || lifUUID.isEmpty()) ? legacyUUID : lifUUID;
        final NetworkLif lif = parentVServer.createLif(lifID);
        lif.setInterfaceName(interfaceName);
        lif.setCurrentNodeKey(currentNodeKey);
        lif.setCurrentPortKey(currentPortKey);
        lif.setHomeNodeKey(homeNodeKey);
        lif.setHomePortKey(homePortKey);
        lif.setAutoRevert(cModeLif.isAutoRevert());
        lif.setFailoverPolicy(cModeLif.getFailoverPolicy().getEnumValue());
        lif.setFailoverPolicyRaw(cModeLif.getFailoverPolicy().getRawValue());
        lif.setHome(cModeLif.isHome());
        lif.setIpAddress(cModeLif.getIpAddress());
        lif.setNetworkMask(cModeLif.getNetMask());
        lif.setNetmaskLen(cModeLif.getNetmaskLen());
        lif.setDnsDomainName(cModeLif.getDnsDomainName());
        lif.setRoutingGroupKey(routingGroupKey);
        lif.setOperStatus(cModeLif.getOperStatus().getEnumValue());
        lif.setOperStatusRaw(cModeLif.getOperStatus().getRawValue());
        lif.setAdminStatus(cModeLif.getAdminStatus().getEnumValue());
        lif.setAdminStatusRaw(cModeLif.getAdminStatus().getRawValue());
        lif.setRole(cModeLif.getRole().getEnumValue());
        lif.setRoleRaw(cModeLif.getRole().getRawValue());
        lif.setDataProtocols(cModeLif.getDataProtocolList());
        lif.setFailoverGroupName(cModeLif.getFailoverGroupName());
        lif.setIsDNSUpdateEnabled(cModeLif.getIsDNSUpdateEnabled());
        lif.setListenForDNSQuery(cModeLif.getListenForDNSQuery());
        
        return lif;
    }

    private CModeFcpLif getCModeFcpLif(final CModeCluster cModeCluster, final CModeLif.LifKey lifKey) {
        for (final CModeFcpLif cModeFcpLif : cModeCluster.getFcpLifs().values()) {
            if (lifKey.equals(cModeFcpLif.getKey())) {
                return cModeFcpLif;
            }
        }
        return null;
    }

    private FcpLif transformFcpLif(final VServer parentVServer, final CModeLif cModeLif, final CModeFcpLif cModeFcpLif, final Node.NodeKey currentNodeKey,
                                   final FcpPort.FcpPortKey currentPortKey, final Node.NodeKey homeNodeKey, final FcpPort.FcpPortKey homePortKey,
                                   final Map<String, NetworkFailoverGroup.NetworkFailoverGroupKey> failoverGroupKeyByName ) {

        final String vServerUUID = parentVServer.getKey().getOcieUUID();
        final String interfaceName = cModeLif.getInterfaceName();
        final String legacyUUID = vServerUUID + '.' + interfaceName;
        final String lifUUID = cModeLif.getLifUUID();
        final String lifID = (lifUUID == null || lifUUID.isEmpty()) ? legacyUUID : lifUUID;

        final FcpLif fcpLif = parentVServer.createFcpLif(lifID);
        fcpLif.setInterfaceName(cModeLif.getInterfaceName());
        fcpLif.setAdminStatus(cModeLif.getAdminStatus().getEnumValue());
        fcpLif.setAdminStatusRaw(cModeLif.getAdminStatus().getRawValue());
        fcpLif.setAutoRevert(cModeLif.isAutoRevert());
        fcpLif.setCurrentNodeKey(currentNodeKey);
        fcpLif.setCurrentPortKey(currentPortKey);
        fcpLif.setDataProtocols(cModeLif.getDataProtocolList());
        fcpLif.setHome(cModeLif.isHome());
        fcpLif.setHomeNodeKey(homeNodeKey);
        fcpLif.setHomePortKey(homePortKey);
        fcpLif.setNodeName(cModeFcpLif.getNodeWWN());
        fcpLif.setOperStatus(cModeLif.getOperStatus().getEnumValue());
        fcpLif.setOperStatusRaw(cModeLif.getOperStatus().getRawValue());
        fcpLif.setPortName(cModeFcpLif.getPortWWN());
        fcpLif.setRole(cModeLif.getRole().getEnumValue());
        fcpLif.setRoleRaw(cModeLif.getRole().getRawValue());
        fcpLif.setFailoverGroupName(cModeLif.getFailoverGroupName());
        return fcpLif;
    }

    private CIFSShare transformCIFSShare(final VServer parentVServer, final CModeCIFSShare cModeCIFSShare, final Volume.VolumeKey volumeKey, final QTree.QTreeKey qTreeKey) {
        final CIFSShare cifsShare = parentVServer.createCIFSShare(cModeCIFSShare.getCIFSShareName());
        cifsShare.setComment(cModeCIFSShare.getComment());
        cifsShare.setPath(cModeCIFSShare.getPath());
        cifsShare.setQTreeKey(qTreeKey);
        cifsShare.setVolumeKey(volumeKey);
        cifsShare.setShareProperties(cModeCIFSShare.getShareProperties().rawValuesToString());
        return cifsShare;
    }

    private ExportPolicy transformExportPolicy(final VServer parentVServer, final CModeExportPolicy cModeExportPolicy) {
        final ExportPolicy exportPolicy = parentVServer.createExportPolicy(cModeExportPolicy.getPolicyID());
        exportPolicy.setName(cModeExportPolicy.getName());
        return exportPolicy;
    }

    private ExportRule transformExportRule(final ExportPolicy parentPolicy, final VServer encapsulatingVserver, final CModeExportRule cModeExportRule) {
        final ExportRule exportRule = parentPolicy.createExportRule(encapsulatingVserver.getKey(), cModeExportRule.getRuleIndex());
        exportRule.setClientMatch(cModeExportRule.getClientMatch());
        exportRule.setAccessProtocol(cModeExportRule.getProtocol().rawValuesToString());
        exportRule.setRoRule(cModeExportRule.getRoRule().rawValuesToString());
        exportRule.setRwRule(cModeExportRule.getRwRule().rawValuesToString());
        exportRule.setSuperUserSecurity(cModeExportRule.getSuperUserSecurity().rawValuesToString());
        return exportRule;
    }

    private DiskPath transformDiskPath(final Cluster parentCluster, final CModeDiskPath cModeDiskPath, final Node.NodeKey nodeKey,
                                       final Disk.DiskKey diskKey, final StorageAdapter.AdapterKey storageAdapterKey) {
        final DiskPath diskPath = parentCluster.createDiskPath(cModeDiskPath.getTargetWWPN(), diskKey, storageAdapterKey);
        diskPath.setArrayName(cModeDiskPath.getArrayName());
        diskPath.setDiskPort(cModeDiskPath.getDiskPort());
        diskPath.setDiskPortName(cModeDiskPath.getDiskPortName());
        diskPath.setInitiatorSideSwitchPort(cModeDiskPath.getInitiatorSideSwitchPort());
        diskPath.setLun(cModeDiskPath.getLun());
        diskPath.setLunPathUseState(cModeDiskPath.getLunPathUseState().getEnumValue());
        diskPath.setLunPathUseStateRaw(cModeDiskPath.getLunPathUseState().getRawValue());
        diskPath.setNodeKey(nodeKey);
        diskPath.setTargetPortGroupNum(cModeDiskPath.getTargetPortGroupNum());
        diskPath.setTargetSideSwitchPort(cModeDiskPath.getTargetSideSwitchPort());
        diskPath.setTargetWWNN(cModeDiskPath.getTargetWWNN());
        return diskPath;
    }

    private ISCSIPortalGroup transformISCSIPortalGroup(final VServer parent, final CModeISCSIPortalGroup cModeISCSIPortalGroup) {
        final ISCSIPortalGroup portalGroup = parent.createISCSIPortalGroup(cModeISCSIPortalGroup.getName());
        portalGroup.setTag(cModeISCSIPortalGroup.getTag());
        return portalGroup;
    }

    private ISCSIPortalGroupToLIF transformISCSIPortalGroupToLIF(final ISCSIPortalGroup portalGroup, final String lifUUID, final CModeISCSIPortalGroupToLIF cModeISCSIPortalGroupToLIF) {
        final ISCSIPortalGroupToLIF iscsiPortalGroupToLIF = portalGroup.createISCSIPortalGroupToLIF(lifUUID);
        final String clusterUUID = portalGroup.getKey().getID(ISCSIPortalGroup.ISCSIPortalGroupKey.Field.cluster_uuid);
        final String vServerUUID = portalGroup.getKey().getID(ISCSIPortalGroup.ISCSIPortalGroupKey.Field.vserver_uuid);
        final Lif.LifKey lifKey = new Lif.LifKey(NetworkLif.KEY_TYPE, clusterUUID, vServerUUID, lifUUID);
        iscsiPortalGroupToLIF.setLifKey(lifKey);
        iscsiPortalGroupToLIF.setInterfaceName(cModeISCSIPortalGroupToLIF.getInterfaceName());
        iscsiPortalGroupToLIF.setIpAddress(cModeISCSIPortalGroupToLIF.getIpAddress());
        iscsiPortalGroupToLIF.setIpPort(cModeISCSIPortalGroupToLIF.getIpPort());
        return iscsiPortalGroupToLIF;
    }

    private ISCSISession transformISCSISession(final ISCSIPortalGroup portalGroup, final String clusterUUID, final VServer.VServerKey vServerKey, final CModeISCSISession cModeISCSISession) {
        final ISCSISession iscsiSession = portalGroup.createISCSISession(clusterUUID, vServerKey, cModeISCSISession.getPortalGroupName(), cModeISCSISession.getTargetSessionID());
        iscsiSession.setInitiatorAliasName(cModeISCSISession.getInitiatorAliasName());
        iscsiSession.setInitiatorNodeName(cModeISCSISession.getInitiatorNodeName());
        iscsiSession.setInitiatorSessionID(cModeISCSISession.getInitiatorSessionID());
        iscsiSession.setMaxConnections(cModeISCSISession.getMaxConnections());
        iscsiSession.setTargetSessionID(cModeISCSISession.getTargetSessionID());
        return iscsiSession;
    }

    private ISCSISessionConnection transformISCSISessionConnection(final ISCSISession iscsiSession, final Lif.LifKey lifKey, final CModeISCSISessionConnection cModeISCSISessionConnection) {
        final ISCSISessionConnection iscsiSessionConnection = iscsiSession.createISCSISessionConnection(cModeISCSISessionConnection.getConnectionID());
        iscsiSessionConnection.setAuthenticationStyle(cModeISCSISessionConnection.getAuthenticationMethod());
        iscsiSessionConnection.setLifKey(lifKey);
        iscsiSessionConnection.setLocalIPAddress(cModeISCSISessionConnection.getLocalIPAddress());
        iscsiSessionConnection.setLocalIPPort(cModeISCSISessionConnection.getLocalIPPort());
        iscsiSessionConnection.setRemoteIPAddress(cModeISCSISessionConnection.getRemoteIPAddress());
        return iscsiSessionConnection;
    }

    private ISCSISecurityEntry transformISCSISecurityEntry(final VServer vServer, final String clusterUUID, final CModeISCSISecurityEntry cModeISCSISecurityEntry) {
        final ISCSISecurityEntry iscsiSecurityEntry = vServer.createISCSISecurityEntry(clusterUUID, cModeISCSISecurityEntry.getInitiator());
        iscsiSecurityEntry.setAuthenticationType(cModeISCSISecurityEntry.getAuthenticationType());
        iscsiSecurityEntry.setOutboundUsername(cModeISCSISecurityEntry.getOutboundUsername());
        iscsiSecurityEntry.setUsername(cModeISCSISecurityEntry.getUsername());
        return iscsiSecurityEntry;
    }

    private License transformLicense(final Cluster parent, final CModeLicense cModeLicense, final Map<String, Node> nodesByName) {
        final License license = parent.createLicense(cModeLicense.getServiceName(), cModeLicense.getSerialNumber());
        cModeLicense.toLicense(license, nodesByName);
        return license;
    }


    private TreeQuota transformTreeQuota(final Volume parentVolume, final QTree.QTreeKey qTreeKey, final CModeQuota cModeQuota) {
        final TreeQuota.QuotaType quotaType = cModeQuota.getQuotaType();
        final TreeQuota treeQuota = parentVolume.createTreeQuota(qTreeKey, quotaType, cModeQuota.getQuotaTarget());
        populateTreeQuota(cModeQuota, treeQuota);
        // Don't include the default quota in the commit calculation.
        // Default quota has an empty qtree ID
        if (TreeQuota.QuotaType.TREE == quotaType && !treeQuota.isDefaultQuota()) {
            final Long diskLimit = treeQuota.getDiskLimit();
            if (diskLimit != null) {
                parentVolume.addToQuotaCommit(diskLimit * 1024);    // Volume reports capacities in bytes, Disk Limit is in kbytes, normalize to bytes
            }
        }
        return treeQuota;
    }

    private UserQuota transformUserQuota(final Volume parentVolume, final QTree.QTreeKey qTreeKey, final CModeQuota cModeQuota) {
        final UserQuota userQuota = parentVolume.createUserQuota(qTreeKey, cModeQuota.getQuotaType(), cModeQuota.getQuotaTarget(),
                cModeQuota.getQuotaUserHash());
        populateUserQuota(cModeQuota, userQuota);
        return userQuota;
    }

    private void populateTreeQuota(final CModeQuota cModeQuota, final TreeQuota quota) {
        final Boolean diskLimitUnlimited = cModeQuota.getDiskLimit().equals("-");
        final Boolean diskUsedUnlimited = cModeQuota.getDiskUsed().equals("-");
        final Boolean fileLimitUnlimited = cModeQuota.getFileLimit().equals("-");
        final Boolean fileUsedUnlimited = cModeQuota.getFileUsed().equals("-");
        final Boolean softDiskLimitUnlimited = cModeQuota.getSoftDiskLimit().equals("-");
        final Boolean softFileLimitUnlimited = cModeQuota.getSoftFileLimit().equals("-");
        final Boolean thresholdUnlimited = cModeQuota.getThreshold().equals("-");

        // ZAPI quota-report gives capacities in kbytes, OCIE quota view exports capacities in kbytes
        final Long diskLimit = (diskLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getDiskLimit());
        final Long diskUsed = (diskUsedUnlimited) ? 0 : Long.valueOf(cModeQuota.getDiskUsed());
        final Long fileLimit = (fileLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getFileLimit());
        final Long fileUsed = (fileUsedUnlimited) ? 0 : Long.valueOf(cModeQuota.getFileUsed());
        final Long softDiskLimit = (softDiskLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getSoftDiskLimit());
        final Long softFileLimit = (softFileLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getSoftFileLimit());
        final Long threshold = (thresholdUnlimited) ? 0 : Long.valueOf(cModeQuota.getThreshold());

        quota.setDiskLimit(diskLimit);
        quota.setDiskLimitUnlimited(diskLimitUnlimited);
        quota.setDiskUsed(diskUsed);
        quota.setDiskUsedUnlimited(diskUsedUnlimited);
        quota.setFileLimit(fileLimit);
        quota.setFileLimitUnlimited(fileLimitUnlimited);
        quota.setFileUsed(fileUsed);
        quota.setFileUsedUnlimited(fileUsedUnlimited);
        quota.setSoftDiskLimit(softDiskLimit);
        quota.setSoftDiskLimitUnlimited(softDiskLimitUnlimited);
        quota.setSoftFileLimit(softFileLimit);
        quota.setSoftFileLimitUnlimited(softFileLimitUnlimited);
        quota.setThreshold(threshold);
        quota.setThresholdUnlimited(thresholdUnlimited);
    }


    private void populateUserQuota(final CModeQuota cModeQuota, final UserQuota quota) {
        final Boolean diskLimitUnlimited = cModeQuota.getDiskLimit().equals("-");
        final Boolean diskUsedUnlimited = cModeQuota.getDiskUsed().equals("-");
        final Boolean fileLimitUnlimited = cModeQuota.getFileLimit().equals("-");
        final Boolean fileUsedUnlimited = cModeQuota.getFileUsed().equals("-");
        final Boolean softDiskLimitUnlimited = cModeQuota.getSoftDiskLimit().equals("-");
        final Boolean softFileLimitUnlimited = cModeQuota.getSoftFileLimit().equals("-");
        final Boolean thresholdUnlimited = cModeQuota.getThreshold().equals("-");

        // ZAPI quota-report gives capacities in kbytes, OCIE quota view exports capacities in kbytes
        final Long diskLimit = (diskLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getDiskLimit());
        final Long diskUsed = (diskUsedUnlimited) ? 0 : Long.valueOf(cModeQuota.getDiskUsed());
        final Long fileLimit = (fileLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getFileLimit());
        final Long fileUsed = (fileUsedUnlimited) ? 0 : Long.valueOf(cModeQuota.getFileUsed());
        final Long softDiskLimit = (softDiskLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getSoftDiskLimit());
        final Long softFileLimit = (softFileLimitUnlimited) ? 0 : Long.valueOf(cModeQuota.getSoftFileLimit());
        final Long threshold = (thresholdUnlimited) ? 0 : Long.valueOf(cModeQuota.getThreshold());

        quota.setDiskLimit(diskLimit);
        quota.setDiskLimitUnlimited(diskLimitUnlimited);
        quota.setDiskUsed(diskUsed);
        quota.setDiskUsedUnlimited(diskUsedUnlimited);
        quota.setFileLimit(fileLimit);
        quota.setFileLimitUnlimited(fileLimitUnlimited);
        quota.setFileUsed(fileUsed);
        quota.setFileUsedUnlimited(fileUsedUnlimited);
        quota.setSoftDiskLimit(softDiskLimit);
        quota.setSoftDiskLimitUnlimited(softDiskLimitUnlimited);
        quota.setSoftFileLimit(softFileLimit);
        quota.setSoftFileLimitUnlimited(softFileLimitUnlimited);
        quota.setThreshold(threshold);
        quota.setThresholdUnlimited(thresholdUnlimited);
    }


    private QuotaUser transformQuotaUser(final UserQuota userQuota, final CModeQuotaUser cModeQuotaUser) {
        final QuotaUser quotaUser = userQuota.createQuotaUser(cModeQuotaUser.getQuotaUserID());
        quotaUser.setQuotaUserName(cModeQuotaUser.getQuotaUsername());
        quotaUser.setQuotaUserType(cModeQuotaUser.getUserType());
        return quotaUser;
    }

    private Map<String, QosWorkloadDetail.QosWorkloadDetailKey> transformQosVolumeWorkloads(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKeys) {
        final Map<String, QosWorkloadDetail.QosWorkloadDetailKey> qosWorkloadNameToKey = new HashMap<String, QosWorkloadDetail.QosWorkloadDetailKey>();
        final String clusterUUID = cModeCluster.getClusterUUID();
        for (final CModeQosVolumeWorkload cModeQosWorkload : cModeCluster.getQosVolumeWorkloads().values()) {
            final String workloadUUID = cModeQosWorkload.getKey().getUUID();
            final QosWorkloadDetail.QosWorkloadDetailKey workloadKey = new QosWorkloadDetail.QosWorkloadDetailKey(clusterUUID, workloadUUID);

            final String vServerName = cModeQosWorkload.getVserverName();
            VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            String volumeName = cModeQosWorkload.getVolumeName();

            final boolean isLessThanSierraNevada = cModeCluster.isLessThanSierraNevada();
            final String qosWorkloadName = cModeQosWorkload.getName();
            if (vServerKey == null && volumeName == null) {
                if (isLessThanSierraNevada && qosWorkloadName.contains("_on_")) {
                    // Currently only for Rolling Rock
                    // Sierra Nevada uses different name convention (for SN at this point we should already have all the data we need)
                    // if vserver and volume names are null we need parse the workload name to determine the vserver and the volume
                    // split the workload name
                    final String[] splitName = cModeQosWorkload.getName().split("_on_");
                    if (splitName.length > 1) {
                        final String volumeNameSplit = splitName[0];
                        final String vServerNameSplit = splitName[1];
                        volumeName = volumeNameSplit;
                        vServerKey = vServerKeysByName.get(vServerNameSplit);
                    }
                }
            }
            // if vServer and volume are defined, we need to assert they are valid
            Volume.VolumeKey volumeKey = null;
            if (vServerKey != null) {
                if (!cluster.getVServers().containsKey(vServerKey)) {
                    logger.error(PackageContext.getLogPrefix() + " - Failed to find referenced vServer " + vServerName
                            + " for QosVolumeWorkload " + qosWorkloadName);
                    continue;
                }

                if (volumeName != null) {
                    if (volumeKeysByNameByVServerKeys.containsKey(vServerKey)) {
                        final Map<String, Volume.VolumeKey> volumeNameToKey = volumeKeysByNameByVServerKeys.get(vServerKey);
                        if (volumeNameToKey != null) {
                            volumeKey = volumeNameToKey.get(volumeName);
                            // make sure the volumeUUID was found
                            if (volumeKey == null) {
                                logger.error(PackageContext.getLogPrefix() + " - Failed to find referenced Volume " + volumeName
                                        + " for QosVolumeWorkload " + qosWorkloadName);
                                continue;
                            }else{
                                final VServer parentVServer = cluster.getVServers().get(vServerKey);
                                final Volume targetVolume = parentVServer.getVolumes().get(volumeKey);
                                if(targetVolume.isVServerRoot()){
                                    // skip the root volume
                                    if (logger.isDebugEnabled()) {
                                        logger.debug("{} - Skipping the root volume. Volume Name: {}", PackageContext.getLogPrefix(), targetVolume.getVolumeName());
                                    }
                                    continue;
                                }
                            }
                        } else {
                            // Because the vServer was validated, this should not happen
                            logger.error(PackageContext.getLogPrefix() + " - Failed to find volumes for vServer " + vServerName
                                    + " for QosVolumeWorkload " + qosWorkloadName);
                            continue;
                        }
                    }
                    else {    //if there is name but no uuid, then it is backup volume, ignore qos_volume_workload for backup volume.
                        logger.info(" {} - ignore this volume {} since it is backup volume.", new Object[]{ PackageContext.getLogPrefix(), volumeName}) ;
                        continue;
                    }
                }
            }

            // transform the workload
            final QosVolumeWorkload qosVolumeWorkload = transformQosVolumeWorkload(cluster, cModeQosWorkload, vServerKey, volumeKey);
            // add the workload
            cluster.getQosVolumeWorkloads().add(qosVolumeWorkload);
            // add the workload name and UUID to the map
            qosWorkloadNameToKey.put(cModeQosWorkload.getName(), workloadKey);
        }
        return qosWorkloadNameToKey;
    }


    private QosVolumeWorkload transformQosVolumeWorkload(final Cluster cluster, final CModeQosVolumeWorkload cModeWorkload, final VServer.VServerKey vServerKey, final Volume.VolumeKey volumeKey) {
        final QosVolumeWorkload qosVolumeWorkload = cluster.createQosVolumeWorkload(cModeWorkload.getKey().getUUID());
        qosVolumeWorkload.setName(cModeWorkload.getName());
        final EnumWrapper<QosVolumeWorkload.QosWorkloadClassType> workloadTypeWrapper = cModeWorkload.getWorkloadTypeWrapper();
        final QosVolumeWorkload.QosWorkloadClassType workloadType = workloadTypeWrapper.getEnumValue();
        final String workloadTypeRaw = workloadTypeWrapper.getRawValue();
        qosVolumeWorkload.setQosWorkloadClassType(workloadType);
        qosVolumeWorkload.setQosWorkloadClassTypeRaw(workloadTypeRaw);
        if (vServerKey != null) {
            qosVolumeWorkload.setVServerKey(vServerKey);
            if (volumeKey != null) {
                final VServer parentVServer = cluster.getVServers().get(vServerKey);
                final Volume targetVolume = parentVServer.getVolumes().get(volumeKey);
                qosVolumeWorkload.setVolumeKey(targetVolume.getKey());
            }
        }
        return qosVolumeWorkload;
    } 

    private void generateQosWorkloadNodeRelationship(final Cluster cluster) {
        if(!ClusterBuilder.isQosDiscoveryEnabled()) {
            return;
        }
        final QosWorkloadNodeRelationships qosWorkloadNodeRelationships = cluster.getQosWorkloadNodeRelationships();
        for (final QosVolumeWorkload workload : cluster.getQosVolumeWorkloads().values()) {
            for (final Node node : cluster.getNodes().values()) {
                qosWorkloadNodeRelationships.add(cluster.createQosWorkloadNodeRelationship(node.getKey(), workload.getKey()));
            }
        }
    }

    private void generateQosWorkloadDetails(final CModeCluster cModeCluster, final Cluster cluster,
                                            final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey,
                                            final Map<String, String> partnerNodeNames) {
        if (!ClusterBuilder.isQosDiscoveryEnabled()) {
            return;
        }
        StringBuffer versionString = new StringBuffer();
        if (cluster.getGeneration() != null) {
            versionString.append(cluster.getGeneration());
            if (cluster.getMajor() != null) {
                versionString.append(".").append(cluster.getMajor());
                if (cluster.getMinor() != null) {
                    versionString.append(".").append(cluster.getMinor());
                }
            }
        }
        if (VersionComparator.compareVersions(versionString.toString(), Cluster.FULL_STEAM) >= 0) {
            QosWorkloadDetailTransform.generateQosWorkloadDetailsForFSAndAbove(cModeCluster, cluster,
                    qosServiceCenterUUIDToKey, partnerNodeNames);
            return;
        }
        final boolean isSN0 = VersionComparator.compareVersions(Cluster.SIERRA_NEVADA_0, versionString.toString()) == 0;
        final boolean isSN1 = !isSN0 && (VersionComparator.compareVersions(Cluster.SIERRA_NEVADA_1, versionString.toString()) == 0);
        // Service centers (AKA CModeQosWorkloadDetail.RESOURCE_NAME) can be grouped into two primary categories:

        // Those existing before the data path hits the d-blade, including n-blade and other user space service centers
        // (collectively called "n-blade" service centers for lack of a better term).

        // And "D-blade" service centers.
        // Some "D-blade" service centers are further partitioned dynamically for each aggregate defined on the system.

        // Given our list of workloads and service centers in a given cluster, the following algorithm shall be used in
        // generating the set of workload-detail UUIDs:

        // For "volume workloads":
        // Cross the "n-blade" service centers on each node in the system.
        // Cross the "d-blade" service centers ONLY on the Nodes in the HA-Pair to which the volume's aggregate belongs
        // (so called "direct" nodes) AND only for the aggregate to which the Volume is currently assigned.

        // For "internal workloads":
        // Fully mesh the "n-blade" and "d-blade" counters across all nodes.

        // N-Blade Service Centers are combined with each discovered node
        final EnumSet<CModeQosWorkloadDetail.RESOURCE_NAME> nBladeServiceCenters = CModeQosWorkloadDetail.getNBladeResourceNames();

        // D-Blade Service Centers are combined only with Nodes that host the volume's aggregate, and only with the volume's aggregate
        final EnumSet<CModeQosWorkloadDetail.RESOURCE_NAME> dBladeServiceCenters = CModeQosWorkloadDetail.getDBladeResourceNames();

        // Collect set of all node names
        final Nodes nodes = cluster.getNodes();
        final List<String> allNodeNames = new ArrayList<String>(nodes.size());
        for (final Node node : nodes.values()) {
            allNodeNames.add(node.getName());
        }

        // Per Aggregate DELAY_CENTER_DISK_IO_ details are generated only for SN1
        final String[] dBladeDetailPrefixes = (isSN1) ?
                new String[]{"DISK_HDD_", "DISK_SSD_", "DELAY_CENTER_DISK_IO_"} :
                new String[]{"DISK_HDD_", "DISK_SSD_"};
        //set delay center disk IO suffix to "DELAY_CENTER_DISK_IO_OTHER" for SN.0, "DELAY_CENTER_DISK_IO" for all others
        final String delayCenterDiskIoSuffix = (isSN0 || isSN1) ? "DELAY_CENTER_DISK_IO_OTHER" : "DELAY_CENTER_DISK_IO";
        // Use the Volume Workloads to generate Workload Details
        for (final QosVolumeWorkload qosVolumeWorkload : cluster.getQosVolumeWorkloads().values()) {
            final VServer.VServerKey vServerKey = qosVolumeWorkload.getVServerKey();
            final Volume.VolumeKey volumeKey = qosVolumeWorkload.getVolumeKey();
            final boolean isVolumeWorkload = (volumeKey != null);

            // N-Blade service centers are reported against all nodes
            createAndAddStaticWorkloadDetails(cluster, qosVolumeWorkload, allNodeNames, nBladeServiceCenters, qosServiceCenterUUIDToKey);

            // D-Blade service centers are reported only against nodes that host the aggregate that hosts the volume
            // for volume workloads.
            // D-Blade service centers are reported against all nodes for internal workloads
            final List<String> workloadNodeNames = (isVolumeWorkload)
                    ? NetAppOCIETransformCommon.getNodeNamesForVolume(cluster, cModeCluster, vServerKey, volumeKey, partnerNodeNames)
                    : allNodeNames;
            createAndAddStaticWorkloadDetails(cluster, qosVolumeWorkload, workloadNodeNames, dBladeServiceCenters, qosServiceCenterUUIDToKey);

            // Combine each workload with workload detail "DELAY_CENTER_DISK_IO_OTHER" across all Nodes
            for (final String nodeName : workloadNodeNames) {
                createAndAddDBladeDelayCenterDiskIOOther(cluster, qosVolumeWorkload, nodeName, qosServiceCenterUUIDToKey, delayCenterDiskIoSuffix);
            }
            // Now do dynamic D-Blade workloads that include aggregate reference
            //  DISK_HDD_<aggr-name>_<aggr-uuid>
            //  DISK_SDD_<aggr-name>_<aggr-uuid>
            //  DELAY_CENTER_DISK_IO_<aggr-name>_<aggr-uuid>
            if (isVolumeWorkload) {
                final CModeAggregate cModeAggregate = NetAppOCIETransformCommon.getAggregateForVolume(cluster, cModeCluster, vServerKey, volumeKey);
                if (cModeAggregate != null) {
                    createAndAddDBladeAggregateWorkloadDetails(cluster, qosVolumeWorkload, cModeAggregate, qosServiceCenterUUIDToKey, partnerNodeNames, dBladeDetailPrefixes);
                }
            } else {
                // All aggregates
                for (final CModeAggregate cModeAggregate : cModeCluster.getAggregates().values()) {
                    createAndAddDBladeAggregateWorkloadDetails(cluster, qosVolumeWorkload, cModeAggregate, qosServiceCenterUUIDToKey, partnerNodeNames, dBladeDetailPrefixes);
                }
            }
        }
    }

    private void createAndAddStaticWorkloadDetails(final Cluster cluster, final QosVolumeWorkload qosVolumeWorkload, final List<String> volumeWorkloadNodeNames,
                                                   final EnumSet<CModeQosWorkloadDetail.RESOURCE_NAME> targetResourceNames,
                                                   final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey) {
        final String volumeWorkloadName = qosVolumeWorkload.getName();
        final QosVolumeWorkload.QosVolumeWorkloadKey qosVolumeWorkloadKey = qosVolumeWorkload.getKey();
        for (final CModeQosWorkloadDetail.RESOURCE_NAME targetResourceName : targetResourceNames) {
            for (final String nodeName : volumeWorkloadNodeNames) {
                final String nodeNamePrefix = nodeName + ":kernel:";
                final String workloadDetailName = volumeWorkloadName + "." + targetResourceName;
                final String workloadDetailUUID = nodeNamePrefix + workloadDetailName;
                final String resourceUUID = nodeNamePrefix + targetResourceName;
                if (qosServiceCenterUUIDToKey.containsKey(resourceUUID)) {
                    final QosServiceCenter.QosResourceKey qosServiceCenterKey = qosServiceCenterUUIDToKey.get(resourceUUID);
                    final QosWorkloadDetail workloadDetail = transformQosWorkloadDetail(cluster, workloadDetailName, workloadDetailUUID,
                            qosVolumeWorkloadKey, qosServiceCenterKey);
                    workloadDetail.setNodeNamePrefix(nodeName);
                    cluster.getQosWorkloadDetails().add(workloadDetail);
                }
            }
        }
    }

    private void createAndAddDBladeAggregateWorkloadDetails(final Cluster cluster, final QosVolumeWorkload qosWorkload,
                                                            final CModeAggregate cModeAggregate,
                                                            final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey,
                                                            final Map<String, String> partnerNodeNames,
                                                            final String[] dBladeDetailPrefixes) {

        // Dynamic D-Blade workloads that include aggregate reference
        //  DISK_HDD_<aggr-name>_<aggr-uuid>
        //  DISK_SSD_<aggr-name>_<aggr-uuid>
        //  DELAY_CENTER_DISK_IO_<aggr-name>_<aggr-uuid>

        final String aggregateName = cModeAggregate.getAggrName();
        final String aggregateUUID = cModeAggregate.getKey().getUUID();
        final List<String> aggregateNodeNames = NetAppOCIETransformCommon.getNodeNamesForAggregate(cModeAggregate, partnerNodeNames);
        for (final String nodeName : aggregateNodeNames) {
            for (final String dBladeDetailPrefix : dBladeDetailPrefixes) {
                createAndAddDBladeAggregateWorkloadDetail(cluster, nodeName, aggregateName, aggregateUUID, qosWorkload, dBladeDetailPrefix, qosServiceCenterUUIDToKey);
            }
        }
    }

    private void createAndAddDBladeAggregateWorkloadDetail(final Cluster cluster, final String nodeName, final String aggregateName, final String aggregateUUID,
                                                           final QosVolumeWorkload qosWorkload, final String dBladeDetailPrefix,
                                                           final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey) {

        final String workloadName = qosWorkload.getName();
        final String aggregateSuffix = new StringBuilder(aggregateName).append('_').append(aggregateUUID).toString();
        final QosVolumeWorkload.QosVolumeWorkloadKey workloadKey = qosWorkload.getKey();
        final String workloadDetailName = getDBladeAggregateWorkloadDetailName(workloadName, dBladeDetailPrefix, aggregateSuffix);
        final String workloadDetailUUID = getDBladeAggregateWorkloadDetailUUID(nodeName, workloadDetailName);
        final String serviceCenterUUID = getDBladeAggregateServiceCenterUUID(nodeName, dBladeDetailPrefix, aggregateSuffix);

        if (qosServiceCenterUUIDToKey.containsKey(serviceCenterUUID)) {
            final QosServiceCenter.QosResourceKey qosServiceCenterKey = qosServiceCenterUUIDToKey.get(serviceCenterUUID);
            final QosWorkloadDetail workloadDetail = transformQosWorkloadDetail(cluster, workloadDetailName, workloadDetailUUID,
                    workloadKey, qosServiceCenterKey);
            workloadDetail.setNodeNamePrefix(nodeName);
            workloadDetail.setAggregateUUIDSuffix(aggregateUUID);
            cluster.getQosWorkloadDetails().add(workloadDetail);
        }
    }

    private void createAndAddDBladeDelayCenterDiskIOOther(final Cluster cluster, final QosVolumeWorkload qosVolumeWorkload,
                                                          final String nodeName, final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey,
                                                          final String delayCenterDiskIoSufix) {
        final String serviceCenterUUID = getDBladeAggregateServiceCenterUUID(nodeName, delayCenterDiskIoSufix, "");
        final QosVolumeWorkload.QosVolumeWorkloadKey workloadKey = qosVolumeWorkload.getKey();
        final String workloadName = qosVolumeWorkload.getName();
        final String workloadDetailName = workloadName + '.' + delayCenterDiskIoSufix;
        final String workloadDetailUUID = getDBladeAggregateWorkloadDetailUUID(nodeName, workloadDetailName);
        if (qosServiceCenterUUIDToKey.containsKey(serviceCenterUUID)) {
            final QosServiceCenter.QosResourceKey qosServiceCenterKey = qosServiceCenterUUIDToKey.get(serviceCenterUUID);
            final QosWorkloadDetail workloadDetail = transformQosWorkloadDetail(cluster, workloadDetailName, workloadDetailUUID,
                    workloadKey, qosServiceCenterKey);
            workloadDetail.setNodeNamePrefix(nodeName);
            cluster.getQosWorkloadDetails().add(workloadDetail);
        }
    }

    private String getDBladeAggregateWorkloadDetailName(final String workloadName, final String workloadDetailPrefix,
                                                        final String aggregateSuffix) {
        return workloadName + "." + workloadDetailPrefix + aggregateSuffix;
    }

    private String getDBladeAggregateWorkloadDetailUUID(final String nodeName, final String workloadDetailName) {
        final String nodeNamePrefix = nodeName + ":kernel:";
        return nodeNamePrefix + workloadDetailName;
    }

    private String getDBladeAggregateServiceCenterUUID(final String nodeName, final String workloadDetailPrefix,
                                                       final String aggregateSuffix) {
        final String nodeNamePrefix = nodeName + ":kernel:";
        return nodeNamePrefix + workloadDetailPrefix + aggregateSuffix;
    }

    private Map<String, QosPolicyGroup.QosPolicyGroupKey> transformQosPolicyGroups(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        final Map<String, QosPolicyGroup.QosPolicyGroupKey> qosPolicyGroupsByName = new HashMap<String, QosPolicyGroup.QosPolicyGroupKey>();
        for (final CModeQosPolicyGroup cModeQosPolicyGroup : cModeCluster.getQosPolicyGroups().values()) {
            final String policyGroupName = cModeQosPolicyGroup.getPolicyGroupName();
            final String qosPolicyGroupUUID = cModeQosPolicyGroup.getKey().getUUID();
            final String vServerName = cModeQosPolicyGroup.getVServerName();
            final VServer.VServerKey vServerKey;
            if (vServerName == null) {
                vServerKey = null;
            } else {
                final VServer vServer = getVServerForChild(cluster, "QosPolicyGroup", policyGroupName, vServerName, vServerKeysByName);
                if (vServer == null) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(PackageContext.getLogPrefix() + " - Failed to vserver " + vServerName + " for qos policy group " + policyGroupName);
                    }
                    vServerKey = null;
                } else {
                    vServerKey = vServer.getKey();
                }
            }

            final QosPolicyGroup qosPolicyGroup = cluster.createQosPolicyGroup(qosPolicyGroupUUID);
            qosPolicyGroup.setMaxThroughput(cModeQosPolicyGroup.getMaxThroughput());
            qosPolicyGroup.setPgid(cModeQosPolicyGroup.getPgid());
            qosPolicyGroup.setPolicyGroupName(policyGroupName);
            qosPolicyGroup.setQosPolicyGroupClass(cModeQosPolicyGroup.getPolicyGroupClass().getEnumValue());
            qosPolicyGroup.setQosPolicyGroupClassRaw(cModeQosPolicyGroup.getPolicyGroupClass().getRawValue());
            cluster.getQosPolicyGroups().add(qosPolicyGroup);
            qosPolicyGroupsByName.put(policyGroupName, qosPolicyGroup.getKey());

            if (vServerKey != null) {
                final QosPolicyGroupVServerRelationship groupVServerRelationship = cluster.createQosPolicyGroupVServerRelationship(qosPolicyGroup.getKey(), vServerKey);
                cluster.getQosPolicyGroupVServerRelationships().add(groupVServerRelationship);
            }
        }

        // Now that QosPolicyGroups are transformed, go back and update vServer refs to QosPolicyGroups
        for (final CModeVServer cModeVServer : cModeCluster.getVServers().values()) {
            final String qosPolicyGroupName = cModeVServer.getQosPolicyGroupName();
            if (qosPolicyGroupName == null) {
                continue;
            }
            final String vServerName = cModeVServer.getName();
            if (!qosPolicyGroupsByName.containsKey(qosPolicyGroupName)) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find qos policy group " + qosPolicyGroupName
                            + " for vserver " + vServerName);
                }
                continue;
            }
            final QosPolicyGroup.QosPolicyGroupKey policyGroupKey = qosPolicyGroupsByName.get(qosPolicyGroupName);
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            if (vServerKey == null) {
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for qos policy group " + qosPolicyGroupName);
                }
                continue;
            }
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                // Shouldn't happen
                if (logger.isDebugEnabled()) {
                    logger.debug(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for qos policy group " + qosPolicyGroupName);
                }
                continue;
            }
            vServer.setQosPolicyGroupUUID(KeyBase.getOcieUUIDOrNull(policyGroupKey));
        }

        return qosPolicyGroupsByName;
    }


    private void createDelayCentersForNodes(final CModeCluster cModeCluster, final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey) {
        // add the delay centers
        final int CM_GENERATION = 8;
        final int SN_MAJOR = 2;

        final ArrayList<String> DELAY_CENTERS = new ArrayList<String>();
        DELAY_CENTERS.add("DELAY_CENTER_CLUSTER_INTERCONNECT");
        DELAY_CENTERS.add("DELAY_CENTER_NETWORK");
        DELAY_CENTERS.add("DELAY_CENTER_WAFL_SUSP_CP");
        StringBuffer versionString = new StringBuffer();
        if (cModeCluster.getGeneration() != null) {
            versionString.append(cModeCluster.getGeneration());
            if (cModeCluster.getMajor() != null) {
                versionString.append(".").append(cModeCluster.getMajor());
                if (cModeCluster.getMinor() != null) {
                    versionString.append(".").append(cModeCluster.getMinor());
                }
            }
        }
        final boolean isSN0 = VersionComparator.compareVersions(Cluster.SIERRA_NEVADA_0, versionString.toString()) == 0;
        final boolean isSN1 = !isSN0 && (VersionComparator.compareVersions(Cluster.SIERRA_NEVADA_1, versionString.toString()) == 0);

        // Check for Full Steam or higher, comparator returns 1, if versionString > FULL_STEAM(8.3)
        // returns 0, if versionString = FULL_STEAM
        // otherwise it returns -1.
        final boolean isFSorHigher = (VersionComparator.compareVersions(versionString.toString(), Cluster.FULL_STEAM) >= 0);
        /**
         *  For SN0 DELAY_CENTER_DISK_IO_OTHER
         *  For SN1 DELAY_CENTER_DISK_IO_OTHER
         *  For all other versions DELAY_CENTER_DISK_IO
         */
        if(isSN0 || isSN1) {
            DELAY_CENTERS.add("DELAY_CENTER_DISK_IO_OTHER");
        } else {
            DELAY_CENTERS.add("DELAY_CENTER_DISK_IO");
            // Add DELAY_CENTER_NVLOG_TRANSFER instance
            // DELAY_CENTER_NVLOG_TRANSFER indicate the NVLOG wait time at a delay center
            // From ONTAP - explicit "transfer" to indicate that the wait is for nvlog transfers (local and remote)
            if(isFSorHigher) {
                DELAY_CENTERS.add("DELAY_CENTER_NVLOG_TRANSFER");
            }
        }

        final List<String> RR_DELAY_CENTERS = new ArrayList();
        RR_DELAY_CENTERS.add("DELAY_CENTER_WAFL_SUSP_DISKIO");

        final String AGGR_REPLACEABLE = "<aggr_name>";
        final List<String> SN_DELAY_CENTERS = new ArrayList();
        SN_DELAY_CENTERS.add("DELAY_CENTER_QOS_LIMIT");
        if(isSN1) {
            SN_DELAY_CENTERS.add("DELAY_CENTER_DISK_IO_" + AGGR_REPLACEABLE); // AGGR_REPLACEABLE needs to be replaced with the aggregate name
        }

        final String[] SN_RESOURCES = {
            "DISK_HDD_" + AGGR_REPLACEABLE,
            "DISK_SSD_" + AGGR_REPLACEABLE,
            "CPU_dblade_background",
        };

        final List<CModeQosServiceCenter> delayCentersForNodes = new ArrayList<CModeQosServiceCenter>();

        for (final String delayCenterName : DELAY_CENTERS) {
            delayCentersForNodes.addAll(createDelayCentersForNodes(cModeCluster, delayCenterName,
                    QosServiceCenter.QosServiceCenterType.DELAY_CENTER.name()));
        }
        // additional delay centers to add
        List<String> additionalDelays = SN_DELAY_CENTERS;
        // if cluster is RR we need to add different delay centers
        if (cModeCluster.getGeneration() == CM_GENERATION && cModeCluster.getMajor() < SN_MAJOR) {
            additionalDelays = RR_DELAY_CENTERS;
        }
        // add additional delay centers
        for (final String delayCenterName : additionalDelays) {
            delayCentersForNodes.addAll(createDelayCentersForNodes(cModeCluster, delayCenterName,
                    QosServiceCenter.QosServiceCenterType.DELAY_CENTER.name()));
        }

        // add additional resources
        for (final String resourceName : SN_RESOURCES) {
            delayCentersForNodes.addAll(createDelayCentersForNodes(cModeCluster, resourceName,
                    QosServiceCenter.QosServiceCenterType.RESOURCE.name()));
        }

        // add the delay centers to the cluster
        final String clusterUUID = cModeCluster.getClusterUUID();
        for (final CModeQosServiceCenter cModeDelayCenter : delayCentersForNodes) {
            cModeCluster.getQosServiceCenters().add(cModeDelayCenter);
            final QosServiceCenter.QosResourceKey qosServiceCenterKey = new QosServiceCenter.QosResourceKey(clusterUUID, cModeDelayCenter.getKey().getName());
            qosServiceCenterUUIDToKey.put(cModeDelayCenter.getKey().getUUID(), qosServiceCenterKey);
        }
    }

    private List<CModeQosServiceCenter> createDelayCentersForNodes(final CModeCluster cluster, final String originalDelayCenterName, final String serviceCenterType) {
        final String AGGR_REPLACEABLE = "<aggr_name>";
        final List<CModeQosServiceCenter> nodeDelayCenters = new ArrayList<CModeQosServiceCenter>();

        // loop through all the nodes and create a delay center for each
        for (final CModeNode node : cluster.getCModeNodes().values()) {
            final String nodeName = node.getKey().getName();
            final String partnerNodeName = node.getPartnerName();
            String delayCenterUuid = nodeName + ":kernel:" + originalDelayCenterName;
            String delayCenterName = originalDelayCenterName;

            // special case for DELAY_CENTER_DISK_IO_<aggr_name> (and others) because we need to generate name
            // based on the aggregates local to the node
            if (originalDelayCenterName.contains(AGGR_REPLACEABLE)) {

                // loop through all the node's aggregates and create a delay delay center for each
                for (final CModeAggregate aggregate : cluster.getAggregates().values()) {

                    // For any node, only create delay centers for aggregates hosted by the node or its partner
                    if (!aggregate.getNodeNameStrings().contains(nodeName) &&
                        !aggregate.getNodeNameStrings().contains(partnerNodeName)) {
                        continue;
                    }

                    delayCenterName = getAggrDelayCenterName(originalDelayCenterName, AGGR_REPLACEABLE, aggregate);
                    delayCenterUuid = nodeName + ":kernel:" + delayCenterName;
                    nodeDelayCenters.add(createServiceCenter(cluster, delayCenterUuid, delayCenterName, serviceCenterType));
                }
            } else {
                nodeDelayCenters.add(createServiceCenter(cluster, delayCenterUuid, delayCenterName, serviceCenterType));
            }
        }
        return nodeDelayCenters;
    }

    private String getAggrDelayCenterName(final String nameTemplate, final String aggrReplaceable, final CModeAggregate aggregate) {
        final String aggregateName = aggregate.getAggrName();
        final String aggregateUUID = aggregate.getKey().getUUID();
        final String delayCenterSuffix = aggregateName + "_" + aggregateUUID;
        return nameTemplate.replace(aggrReplaceable, delayCenterSuffix);
    }

    private CModeQosServiceCenter createServiceCenter(final CModeCluster cluster, final String delayCenterUuid, final String delayCenterName, final String serviceCenterType) {
        final CModeQosServiceCenter qosServiceCenter = cluster.createQosServiceCenter(delayCenterUuid);
        qosServiceCenter.setName(delayCenterName);
        qosServiceCenter.setType(serviceCenterType);
        qosServiceCenter.setParentResourceUUID(null);
        return qosServiceCenter;
    }


    private QosWorkloadDetail transformQosWorkloadDetail(final Cluster cluster, final String workloadDetailName,
                                                         final String workloadDetailUUID,
                                                         final QosVolumeWorkload.QosVolumeWorkloadKey volumeWorkloadKey,
                                                         final QosServiceCenter.QosResourceKey qosServiceCenterKey) {
        final QosWorkloadDetail qosWorkloadDetail = cluster.createQosWorkloadDetail(workloadDetailUUID);
        qosWorkloadDetail.setName(workloadDetailName);
        qosWorkloadDetail.setQosVolumeWorkloadKey(volumeWorkloadKey);
        qosWorkloadDetail.setQosResourceKey(qosServiceCenterKey);
        return qosWorkloadDetail;
    }

    private Map<String, QosServiceCenter.QosResourceKey> transformQosServiceCenters(final CModeCluster cModeCluster, final Map<String, Node> nodeNameToNode, final Map<String, Aggregate.AggregateKey> aggrKeysByName) {
        final Map<String, QosServiceCenter.QosResourceKey> qosServiceCenterUUIDToKey = new HashMap<String, QosServiceCenter.QosResourceKey>();
        final String clusterUUID = cModeCluster.getClusterUUID();

        // First, generate hard-coded delay centers that are not available from ZAPI.
        if (ClusterBuilder.isQosDiscoveryEnabled()) {
            createDelayCentersForNodes(cModeCluster, qosServiceCenterUUIDToKey);
        }

        // setup a set of aggregateUUIDs to deal with aggrKeyByName change
        final Set<String> aggrUUIDs = new HashSet<String>(aggrKeysByName.size());
        for (final Aggregate.AggregateKey key : aggrKeysByName.values()) {
            aggrUUIDs.add(key.getID(Aggregate.AggregateKey.Field.uuid));
        }

        for (final CModeQosServiceCenter cModeQosServiceCenter : cModeCluster.getQosServiceCenters().values()) {
            final String serviceCenterName = cModeQosServiceCenter.getName();
            if (CModeQosServiceCenter.FILTER_RESOURCE_NAMES.contains(serviceCenterName)) {
                continue;
            }
            final QosServiceCenter.QosResourceKey qosResourceKey = new QosServiceCenter.QosResourceKey(clusterUUID, cModeQosServiceCenter.getKey().getName());
            qosServiceCenterUUIDToKey.put(cModeQosServiceCenter.getKey().getUUID(), qosResourceKey);
        }


        for (final CModeQosServiceCenter cModeQosServiceCenter : cModeCluster.getQosServiceCenters().values()) {
            final String serviceCenterName = cModeQosServiceCenter.getName();
            if (CModeQosServiceCenter.FILTER_RESOURCE_NAMES.contains(serviceCenterName)) {
                continue;
            }
            final String serviceCenterUUID = cModeQosServiceCenter.getKey().getUUID();
            Node node = null;
            String aggrUUID = null;

            // if the resource is disk related try to determine the aggregate it is associated with
            if (serviceCenterName.startsWith("DISK_HDD_") ||
                serviceCenterName.startsWith("DISK_SSD_") ||
                serviceCenterName.startsWith("DELAY_CENTER_DISK_IO_")) {

                // skip HDD_OTHER because they are not associated with a specific aggregate
                if (!(serviceCenterName.startsWith("DISK_HDD_OTHER") ||
                      serviceCenterName.startsWith("DISK_SSD_OTHER") ||
                      serviceCenterName.startsWith("DELAY_CENTER_DISK_IO_OTHER"))) {

                    final int start = serviceCenterName.lastIndexOf("_");
                    final String candidateAggrUUID = serviceCenterName.substring(start + 1, serviceCenterName.length());
                    if (aggrUUIDs.contains(candidateAggrUUID)) {
                        aggrUUID = candidateAggrUUID;
                    }
                }
            }

            // figure out the node
            // <node-x>:kernel:<resource name>
            final String[] splitUUID = serviceCenterUUID.split(":");
            if (splitUUID.length > 0) {
                final String nodeName = splitUUID[0];
                node = nodeNameToNode.get(nodeName);
            }
            if (node != null) {
                // try to determine the parent resource
                final String parentResourceName = cModeQosServiceCenter.getParentResourceUUID();
                String parentResourceUuid = null;
                if (parentResourceName != null) {
                    parentResourceUuid = node.getName() + ":kernel:" + parentResourceName;
                    if (!qosServiceCenterUUIDToKey.containsKey(parentResourceUuid)) {
                        logger.warn("Parent service center " + parentResourceUuid + " not found");
                        qosServiceCenterUUIDToKey.remove(serviceCenterUUID);
                        continue;
                    }
                }

                // transform the service center
                final Aggregate.AggregateKey aggregateKey = (aggrUUID != null) ? new Aggregate.AggregateKey(clusterUUID, aggrUUID) : null;
                final QosServiceCenter qosServiceCenter = transformQosServiceCenter(node, cModeQosServiceCenter, aggregateKey);
                // add the service center
                node.getQosServiceCenters().add(qosServiceCenter);

                // transform the service center parent relationship
                if (parentResourceUuid != null) {
                    final QosServiceCenter.QosResourceKey qosServiceCenterKey = new QosServiceCenter.QosResourceKey(clusterUUID, cModeQosServiceCenter.getKey().getID());
                    final QosServiceCenter.QosResourceKey parentQosServiceCenterKey = new QosServiceCenter.QosResourceKey(clusterUUID, parentResourceUuid);
                    final QosServiceCenterRelationship serviceCenterRelation = transformQosServiceCenterRelationship(node, qosServiceCenterKey, parentQosServiceCenterKey);
                    node.getQosServiceCenterRelationships().add(serviceCenterRelation);
                }
            }
        }
        return qosServiceCenterUUIDToKey;
    }

    private QosServiceCenter transformQosServiceCenter(final Node parent, final CModeQosServiceCenter cModeServiceCenter, final Aggregate.AggregateKey aggregateKey) {
        final QosServiceCenter qosServiceCenter = parent.createQosServiceCenter(cModeServiceCenter.getKey().getUUID());
        qosServiceCenter.setName(cModeServiceCenter.getName());
        qosServiceCenter.setAggregateKey(aggregateKey);
        qosServiceCenter.setType(QosServiceCenter.QosServiceCenterType.getServiceCenterType(cModeServiceCenter.getType()));

        return qosServiceCenter;
    }

    private QosServiceCenterRelationship transformQosServiceCenterRelationship(final Node parentNode,
                                                                               final QosServiceCenter.QosResourceKey qosServiceCenterKey,
                                                                               final QosServiceCenter.QosResourceKey parentQosServiceCenterKey) {
        return parentNode.createQosServiceCenterRelationship(qosServiceCenterKey, parentQosServiceCenterKey);
    }

    private void transformServiceProcessors(final CModeCluster cModeCluster, final Map<String, Node> nodesByName) {
        for (final CModeNode cModeNode : cModeCluster.getCModeNodes().values()) {
            for (final CModeServiceProcessor cModeServiceProcessor : cModeNode.getServiceProcessors().values()) {
                final String macAddress = cModeServiceProcessor.getKey().getUUID();
                final String nodeName = cModeServiceProcessor.getReturnedNodeName();
                if (!nodesByName.containsKey(nodeName)) {
                    logger.info(PackageContext.getLogPrefix() + " - Failed to find node " + nodeName + " for service processor " + macAddress);
                    continue;
                }
                final Node node = nodesByName.get(nodeName);
                final ServiceProcessor serviceProcessor = node.createServiceProcessor(macAddress);
                serviceProcessor.setIpAddress(cModeServiceProcessor.getIpAddressesAsCSV()); // many vs one?
                serviceProcessor.setIpConfigured(cModeServiceProcessor.isIpConfigured());
                serviceProcessor.setStatus(cModeServiceProcessor.getStatus().getEnumValue());
                serviceProcessor.setStatusRaw(cModeServiceProcessor.getStatus().getRawValue());
                serviceProcessor.setType(cModeServiceProcessor.getType().getEnumValue());
                serviceProcessor.setTypeRaw(cModeServiceProcessor.getType().getRawValue());
                node.getServiceProcessors().add(serviceProcessor);
            }
        }
    }

    private void transformStorageShelves(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Node> nodesByName) {
        final Map<UUIDKey, StorageShelf> reportedShelves = new HashMap<UUIDKey, StorageShelf>();
        Set<String>  localShelfIds = getLocalShelfIds(cModeCluster);
        for (final CModeNode cModeNode : cModeCluster.getCModeNodes().values()) {
            final String nodeName = cModeNode.getKey().getName();
            if (!nodesByName.containsKey(nodeName)) {
                // Shouldn't happen.
                logger.error(PackageContext.getLogPrefix() + " - Failed to find node " + nodeName + " for storage shelves.");
                continue;
            }
            final Node node = nodesByName.get(nodeName);
            final String thisNodeUUID = node.getNodeUUID();

            for (final CModeStorageShelf cModeStorageShelf : cModeNode.getStorageShelves().values()) {
                final UUIDKey shelfKey = cModeStorageShelf.getKey();
                final StorageShelf storageShelf;
                if (reportedShelves.containsKey(shelfKey)) {
                    // Second time we've seen this shelf, Update node partnerNode
                    storageShelf = reportedShelves.get(shelfKey);
                    final String thisChannel = cModeStorageShelf.getChannelName();
                    storageShelf.setPartnerNodeUUID(thisNodeUUID);
                    storageShelf.setPartnerChannelName(thisChannel);
                    
                    continue;
                }

                final String shelfUID = shelfKey.getUUID();
                storageShelf = cluster.createStorageShelf(shelfUID);
                reportedShelves.put(shelfKey, storageShelf);
                storageShelf.setShelfType(cModeStorageShelf.getShelfType().getRawValue());
                storageShelf.setChannelName(cModeStorageShelf.getChannelName());            // JSD will the channel name change
                storageShelf.setShelfID(cModeStorageShelf.getShelfID());
                storageShelf.setStackID(cModeStorageShelf.getStackId());
                if(CAMPUS.equalsIgnoreCase(cluster.getMtConfigurationType()))  {
                    setLocalAttachForShelf(storageShelf, localShelfIds);
                }
                else {
                    storageShelf.setIsLocalAttach(cModeStorageShelf.getIsLocalAttach());
                }
                storageShelf.setFirmwareRevisionA(cModeStorageShelf.getFirmwareRevA());
                storageShelf.setFirmwareRevisionB(cModeStorageShelf.getFirmwareRevB());
                storageShelf.setModule(cModeStorageShelf.getModule());
                final EnumWrapper<StorageShelf.State> stateWrapper = cModeStorageShelf.getShelfState();
                storageShelf.setShelfState(stateWrapper.getEnumValue());
                storageShelf.setShelfStateRaw(stateWrapper.getRawValue());
                storageShelf.setNodeUUID(thisNodeUUID);     // First time we've seen this shelf, only set this node.
                cluster.getStorageShelves().add(storageShelf);
            }

            // Each Shelf Channel supports multiple shelves.
            // Here we try to correlate the Storage Shelf Env info with a specific shelf.
            for (final CModeStorageShelfChannel storageShelfChannel : cModeNode.getStorageShelfChannels().values()) {
                // each channel services many shelves.
                // We don't care about the channels themselves, only the contained env info
                final String channelName = storageShelfChannel.getKey().getName();
                for (final CModeStorageShelfChannel.CModeStorageShelfEnv shelfEnv : storageShelfChannel.getStorageShelfEnvs().values()) {
                    boolean reportChannel = true;

                    // SesLogicalID identifies the ShelfUID
                    final String shelfUID = shelfEnv.getSesLogicalID();
                    final StorageShelf.StorageShelfKey shelfKey = new StorageShelf.StorageShelfKey(cModeCluster.getClusterUUID(), shelfUID);

                    final StorageShelf storageShelf = cluster.getStorageShelves().get(shelfKey);
                    if (storageShelf == null) {
                        // KFS Simulator issues
                        if (logger.isDebugEnabled()) {
                            logger.info(PackageContext.getLogPrefix() + " - Skipping Storage Shelf Channel " + channelName
                                    + ", can't find storage shelf by shelfUID " + shelfUID);
                        }
                        continue;
                    }

                    //only report for one node---burt 789341
                    if(!storageShelf.getNodeUUID().equals(thisNodeUUID))  {
                        continue;
                    }

                    // For failover, 2 channels connect to a shelf.
                    // We will discover both channels, but only update shelf attributes for the shelf's
                    // current channel.   KFS simulators don't always report channel name, so avoid null
                    // pointer exceptions
                    if (channelName == null || storageShelf.getChannelName() == null) {
                        reportChannel = false;
                    } else if (!storageShelf.getChannelName().equals(channelName)) {
                        reportChannel = false;
                    }

                    if (reportChannel) {
                        storageShelf.setSesProductID(shelfEnv.getSesProductID());

                        int coolingElementCount = 0;
                        int failedCoolingElementCount = 0;
                        for (final CModeStorageShelfChannel.CModeCoolingElement coolingElement : shelfEnv.getCoolingElements().values()) {
                            final Boolean isNotInstalled = coolingElement.isNotInstalled();
                            if (isNotInstalled != null && isNotInstalled) {
                                continue;
                            }
                            coolingElementCount++;
                            if (coolingElement.isError()) {
                                failedCoolingElementCount++;
                            }
                        }
                        storageShelf.setCoolingElementCount(coolingElementCount);
                        storageShelf.setCoolingElementFailedCount(failedCoolingElementCount);

                        int powerSupplyCount = 0;
                        int failedPowerSupplyCount = 0;
                        for (final CModeStorageShelfChannel.CModePowerSupply powerSupply : shelfEnv.getPowerSupplies().values()) {
                            final Boolean isNotInstalled = powerSupply.isNotInstalled();
                            if (isNotInstalled != null && isNotInstalled) {
                                continue;
                            }
                            powerSupplyCount++;
                            if (powerSupply.isError()) {
                                failedPowerSupplyCount++;
                            }
                        }
                        storageShelf.setPowerSupplyCount(powerSupplyCount);
                        storageShelf.setPowerSupplyFailedCount(failedPowerSupplyCount);

                        int tempSensorCount = 0;
                        int failedTempSensorCount = 0;
                        for (final CModeStorageShelfChannel.CModeTempSensor tempSensor : shelfEnv.getTempSensors().values()) {
                            final Boolean isNotInstalled = tempSensor.isNotInstalled();
                            if (isNotInstalled != null && isNotInstalled) {
                                continue;
                            }
                            tempSensorCount++;
                            if (tempSensor.isError()) {
                                failedTempSensorCount++;
                            }
                        }
                        storageShelf.setTemperatureSensorCount(tempSensorCount);
                        storageShelf.setTemperatureSensorFailedCount(failedTempSensorCount);

                        int currentSensorCount = 0;
                        int failedCurrentSensorCount = 0;
                        int voltageSensorCount = 0;
                        int failedVoltageSensorCount = 0;
                        for (final CModeStorageShelfChannel.CModeSensor sensor : shelfEnv.getSensors().values()) {
                            final Boolean isNotInstalled = sensor.isSensorNotInstalled();
                            if (isNotInstalled != null && isNotInstalled) {
                                continue;
                            }
                            final Boolean isErrorRaw = sensor.isSensorError();
                            final boolean isError = (isErrorRaw != null && isErrorRaw);
                            switch (sensor.getType()) {
                                case current:
                                    currentSensorCount++;
                                    if (isError) {
                                        failedCurrentSensorCount++;
                                    }
                                    break;

                                case voltage:
                                    voltageSensorCount++;
                                    if (isError) {
                                        failedVoltageSensorCount++;
                                    }
                                    break;
                            }
                        }
                        storageShelf.setCurrentSensorCount(currentSensorCount);
                        storageShelf.setCurrentSensorFailedCount(failedCurrentSensorCount);
                        storageShelf.setVoltageSensorCount(voltageSensorCount);
                        storageShelf.setVoltageSensorFailedCount(failedVoltageSensorCount);

                        transformStorageShelfCables(shelfEnv, storageShelf, nodesByName);
                    }


                }
            }
        }
    }

    private void transformStorageSwitches(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, Node> nodesByName) {
        // build set of all port WWNs in our local cluster
        final Set<String> allWwns = new HashSet<String>();
        for (final Node node : nodesByName.values()) {
            for (final FcpPort fcpPort : node.getFcpPorts().values()) {
                allWwns.add(fcpPort.getKey().getID(FcpPort.FcpPortKey.Field.port_name));
            }
        }

        for (final CModeStorageSwitch cModeStorageSwitch : cModeCluster.getStorageSwitches().values()) {
            for (final String peerWwn : cModeStorageSwitch.getStorageSwitchPeerWwns()) {

                // We don't want to include this switch if it's not connected to one of the nodes in our cluster
                if (allWwns.contains(peerWwn)) {
                    final EnumWrapper<StorageSwitch.Role> roleWrapper = cModeStorageSwitch.getRoleWrapper();
                    final EnumWrapper<StorageSwitch.Status> statusWrapper = cModeStorageSwitch.getStatusWrapper();

                    final StorageSwitch storageSwitch = cluster.createStorageSwitch(cModeStorageSwitch.getKey().getOcieUUID());

                    storageSwitch.setFirmwareVersion(cModeStorageSwitch.getFirmwareVersion());
                    storageSwitch.setName(cModeStorageSwitch.getName());
                    storageSwitch.setIpAddress(cModeStorageSwitch.getIpAddress());
                    storageSwitch.setModel(cModeStorageSwitch.getModel());
                    storageSwitch.setRole(roleWrapper.getEnumValue());
                    storageSwitch.setRoleRaw(roleWrapper.getRawValue());
                    storageSwitch.setStatus(statusWrapper.getEnumValue());
                    storageSwitch.setStatusRaw(statusWrapper.getRawValue());
                    storageSwitch.setVendor(cModeStorageSwitch.getVendor());
                    storageSwitch.setSensorCount(cModeStorageSwitch.getSensorCount());
                    storageSwitch.setPowerSupplyCount(cModeStorageSwitch.getPowerSupplyCount());
                    storageSwitch.setFanCount(cModeStorageSwitch.getFanCount());

                    cluster.getStorageSwitches().add(storageSwitch);
                    break;
                }
            }
        }
    }

    private void transformNodePerfUUIDs(final CModeCluster cModeCluster, final Cluster cluster) {
        cluster.getCifsNodeUUIDNameMap().putAll(cModeCluster.getCifsNodeUUIDNameMap());
        cluster.getSmb1NodeUUIDNameMap().putAll(cModeCluster.getSmb1NodeUUIDNameMap());
        cluster.getSmb2NodeUUIDNameMap().putAll(cModeCluster.getSmb2NodeUUIDNameMap());
        cluster.getNfsv3NodeUUIDNameMap().putAll(cModeCluster.getNfsv3NodeUUIDNameMap());
        cluster.getNfsv4NodeUUIDNameMap().putAll(cModeCluster.getNfsv4NodeUUIDNameMap());
        cluster.getNfsv4_1NodeUUIDNameMap().putAll(cModeCluster.getNfsv4_1NodeUUIDNameMap());
    }

    private void transformVserverPerfUUIDs(final CModeCluster cModeCluster, final Cluster cluster) {
        cluster.getCifsVserverIDNameMap().putAll(cModeCluster.getCifsVserverIDNameMap());
        cluster.getNfsv3VserverIDNameMap().putAll(cModeCluster.getNfsv3VserverIDNameMap());
        cluster.getNfsv4VserverIDNameMap().putAll(cModeCluster.getNfsv4VserverIDNameMap());
        cluster.getNfsv41VserverIDNameMap().putAll(cModeCluster.getNfsv41VserverIDNameMap());
        cluster.getLifVserverIDNameMap().putAll(cModeCluster.getLifVserverIDNameMap());
    }
    
    private void transformStorageShelfCables(final CModeStorageShelfChannel.CModeStorageShelfEnv shelfEnv, final StorageShelf storageShelf, final Map<String, Node> nodesByName) {
        for (final CModeStorageShelfChannel.CModeShelfCable cModeShelfCable : shelfEnv.getShelfCables().values()) {
            //use storageshelf's nodeUUID since these storageshelfcables are children of storageshelf
            final StorageShelfCable storageShelfCable = storageShelf.createAndAddStorageShelfCable(storageShelf.getNodeUUID(), cModeShelfCable.getShelfId(),
                    cModeShelfCable.getChannelName(), cModeShelfCable.getPartNumber(), cModeShelfCable.getSerialNumber(), cModeShelfCable.getConnectorNumber());
            storageShelfCable.setAttachedSerialNumber(cModeShelfCable.getAttachedSerialNumber());
            storageShelfCable.setShelfChannelFailure(cModeShelfCable.getShelfChannelFailure());
            storageShelfCable.setEndIdentifier(cModeShelfCable.getEndIdentifier());
            storageShelfCable.setLength(cModeShelfCable.getLength());
            storageShelfCable.setManufacturer(cModeShelfCable.getManufacturer());
            storageShelfCable.setTechnology(cModeShelfCable.getTechnology());
            storageShelfCable.setType(cModeShelfCable.getType());
            storageShelfCable.setConnectorDesignator(cModeShelfCable.getConnectorDesignator());
            storageShelfCable.setCableConnected(cModeShelfCable.getCableConnected());
            storageShelfCable.setConnectorError(cModeShelfCable.getConnectorError());
            storageShelfCable.setShelfModule(cModeShelfCable.getShelfModule());
            storageShelfCable.setSwapCount(cModeShelfCable.getSwapCount());
        }
    }

    private void transformVServerNameMappings(final CModeCluster cModeCluster, final Cluster cluster, final Map<String, VServer.VServerKey> vServerKeysByName) {
        for (final CModeVServerNameMapping cModeNameMapping : cModeCluster.getVServerNameMappings().values()) {
            final String vServerName = cModeNameMapping.getKey().getID(CModeVServerNameMapping.MappingKey.Field.vserver_name);
            if (!vServerKeysByName.containsKey(vServerName)) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Skipping Name Mapping for unknown vServer " + vServerName);
                }
                continue;
            }
            final VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
            final VServer vServer = cluster.getVServers().get(vServerKey);
            if (vServer == null) {
                if (logger.isDebugEnabled()) {
                    logger.info(PackageContext.getLogPrefix() + " - Skipping Name Mapping for unknown vServer " + vServerName);
                }
                continue;
            }
            final EnumWrapper<VServerNameMapping.Direction> directionWrapper = cModeNameMapping.getDirectionWrapper();
            final VServerNameMapping nameMapping = vServer.createVServerNameMapping(directionWrapper.getRawValue(), cModeNameMapping.getPosition());
            nameMapping.setDirectionRaw(directionWrapper.getRawValue());
            nameMapping.setDirection(directionWrapper.getEnumValue());
            nameMapping.setPattern(cModeNameMapping.getPattern());
            nameMapping.setReplacement(cModeNameMapping.getReplacement());
            vServer.getVServerNameMappings().add(nameMapping);
        }
    }

    private void transformSystemHealthAlerts(CModeCluster cModeCluster, Cluster cluster, Map<String, Node> nodesByName) {
        for(final CModeSystemHealthAlert cModeAlert : cModeCluster.getSystemHealthAlerts().values())  {
            final EnumWrapper<SystemHealthAlert.MonitorType> monitorWrapper = cModeAlert.getMonitor();
            CModeSystemHealthAlert.SystemHealthAlertKey cModeAlertKey = cModeAlert.getKey();
            String alertId = cModeAlertKey.getID(CModeSystemHealthAlert.SystemHealthAlertKey.Field.alert_id);
            String alertingResource = cModeAlertKey.getID(CModeSystemHealthAlert.SystemHealthAlertKey.Field.alerting_resource);
            final SystemHealthAlert alert = cluster.createSystemHealthAlert(alertId, alertingResource);
            String nodeUUID = nodesByName.get(cModeAlert.getNode()).getKey().getOcieUUID();

            alert.setAcknowledge(cModeAlert.getAcknowledge());
            alert.setAcknowledger(cModeAlert.getAcknowledger());
            alert.setAdditionalInfo(cModeAlert.getAdditionalInfo());
            alert.setAlertingResourceName(cModeAlert.getAlertingResourceName());
            alert.setCorrectiveActions(cModeAlert.getCorrectiveActions());
            alert.setIndicationTime(cModeAlert.getIndicationTime());
            alert.setMonitor(cModeAlert.getMonitor().getEnumValue());
            alert.setMonitorRaw(cModeAlert.getMonitor().getRawValue());
            alert.setNodeID(nodeUUID);
            alert.setPerceivedSeverity(cModeAlert.getPerceivedSeverity());
            alert.setPolicy(cModeAlert.getPolicy());
            alert.setPossibleEffect(cModeAlert.getPossibleEffect());
            alert.setProbableCause(cModeAlert.getProbableCause());
            alert.setProbableCauseDescription(cModeAlert.getProbableCauseDescription());
            alert.setSubsystem(cModeAlert.getSubsystem());
            alert.setSuppress(cModeAlert.getSuppress());
            alert.setSuppressor(cModeAlert.getSuppressor());

            cluster.getSystemHealthAlerts().add(alert);
        }
    }

    private NetworkPort.PortKey getNetworkPortKey(final CModeCluster cModeCluster, final String nodeName, final String nodeUUID, final String portName) {
        // Returns a String instead of a Key because this may represent FCP or Ethernet port.
        // Check for eth match first
        if (cModeCluster == null || nodeName == null || portName == null) {
            return null;
        }
        final String clusterUUID = cModeCluster.getClusterUUID();
        final CModeNetInterface.IFKey ethKey = new CModeNetInterface.IFKey(nodeName, portName);
        final CModeNetInterface eth = cModeCluster.getNetInterfaces().get(ethKey);
        if (eth != null) {
            // Found it as eth
            return new NetworkPort.PortKey(clusterUUID, nodeUUID, portName, eth.getMacAddress());
        } else {
            return null;
        }
    }

    private FcpPort.FcpPortKey getFcpPortKey(final CModeCluster cModeCluster, final String nodeName, final String adapterName,
                                             final Map<String, Node> nodesByName) {
        // Returns a String instead of a Key because this may represent FCP or Ethernet port.
        // Check for eth match first
        if (cModeCluster == null || nodeName == null || adapterName == null) {
            return null;
        }
        final Node node = nodesByName.get(nodeName);
        if (node == null) {
            return null;
        }
        final String nodeUUID = node.getKey().getOcieUUID();
        final String clusterUUID = cModeCluster.getClusterUUID();
        final CModeFcpAdapter.FCPKey fcpKey = new CModeFcpAdapter.FCPKey(nodeName, adapterName);
        final CModeFcpAdapter fcp = cModeCluster.getFcpAdapters().get(fcpKey);
        if (fcp != null) {

            return new FcpPort.FcpPortKey(clusterUUID, nodeUUID, fcp.getNodeWWN(), fcp.getPortWWN(), adapterName);
        } else {
            return null;
        }
    }

    /**
     *
     * Method to get vServer by vServerName
     *
     *
     * @param cluster
     * @param childType
     * @param childID
     * @param vServerName
     * @param vServerKeysByName
     * @return  VServer
     *
     */
    private VServer getVServerForChild(final Cluster cluster, final String childType, final String childID, final String vServerName, final Map<String, VServer.VServerKey> vServerKeysByName) {
        VServer.VServerKey vServerKey = vServerKeysByName.get(vServerName);
        if (vServerKey == null) {
            if (logger.isDebugEnabled()) {
                logger.debug(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerName + " for " + childType + " " + childID);
            }
            return null;
        }
        final VServer vServer = cluster.getVServers().get(vServerKey);
        if (vServer == null) {
            if (logger.isDebugEnabled()) {
                logger.debug(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerKey + " for " + childType + " " + childID);
            }
        }
        return vServer;
    }

    /**
     * Method to get vServer by vServerUuid
     *
     * @param cluster
     * @param childType
     * @param childID
     * @param vServerUuid
     * @return VServer
     *
     */
    private VServer getVServerByUUIDForChild(final Cluster cluster, final String childType, final String childID, final String vServerUuid) {
        VServer.VServerKey vServerKey =  new VServer.VServerKey(cluster.getClusterUUID(), vServerUuid);
        final VServer vServer = cluster.getVServers().get(vServerKey);
        if (vServer == null) {
            if (logger.isDebugEnabled()) {
                logger.debug(PackageContext.getLogPrefix() + " - Failed to find vserver " + vServerKey + " for " + childType + " " + childID);
            }
        }
        return vServer;
    }

    private Volume getVolumeForChild(final VServer vServer, final String childType, final String childID, final String volumeName, final Map<VServer.VServerKey, Map<String, Volume.VolumeKey>> volumeKeysByNameByVServerKey) {

        final Map<String, Volume.VolumeKey> volumeKeysByName = volumeKeysByNameByVServerKey.get(vServer.getKey());
        if (volumeKeysByName == null) {
            if (logger.isDebugEnabled()) {
                logger.debug(PackageContext.getLogPrefix() + " - Failed to find volume " + volumeName + " for " + childType + " " + childID);
            }
            return null;
        }
        final Volume.VolumeKey volumeKey = volumeKeysByName.get(volumeName);
        if (volumeKey == null) {
            if (logger.isDebugEnabled()) {
                logger.debug(PackageContext.getLogPrefix() + " - Failed to find volume " + volumeName + " for " + childType + " " + childID);
            }
            return null;
        }
        final Volume volume = vServer.getVolumes().get(volumeKey);
        if (volume == null) {
            if (logger.isDebugEnabled()) {
                logger.debug(PackageContext.getLogPrefix() + " - Failed to find volume " + volumeKey.getVolumeInstanceUUID() + " for " + childType + " " + childID);
            }
        }
        return volume;
    }

    private long getLongForSum(final Long val) {
        return (val == null) ? 0 : val;
    }

    private Aggregate.RaidType determineRaidType(final String raidStatus) {
        Aggregate.RaidType raidType = Aggregate.RaidType.NOT_MAPPED;
        if (raidStatus != null) {
            final List<String> statusTokenList = new ArrayList<String>();
            final String[] statusTokens = raidStatus.split(",");
            for (final String statusToken : statusTokens) {
                statusTokenList.add(statusToken.trim());
            }

            // According to documentation the raid type can only be one of the following
            // raid0, raid4, raid_dp, mixed_raid_type
            if (statusTokenList.contains("raid0")) {
                raidType = Aggregate.RaidType.RAID_0;
            }
            else if (statusTokenList.contains("raid4")) {
                raidType = Aggregate.RaidType.RAID_4;
            }
            else if (statusTokenList.contains("raid_dp")) {
                raidType = Aggregate.RaidType.RAID_DP;
            }
            else if (statusTokenList.contains("mixed_raid_type")) {
                raidType = Aggregate.RaidType.MIXED_RAID_TYPE;
            }
        }
        return raidType;
    }

    private CModeVServer getCModeVServerByName(CModeCluster cModeCluster, String vServerName){
        CModeVServers cModeVServers = cModeCluster.getVServers();
        for(CModeVServer cModeVServer : cModeVServers.values()) {
            if(vServerName.equals(cModeVServer.getName())) {
                return cModeVServer;
            }
        }
        return null;
    }
    
    private void addAggregateToDoNotDeleteList(Cluster cluster, String aggregateUUID) {
        DoNotDelete doNotDelete =  cluster.createDoNotDelete(UUID.randomUUID().toString());
        doNotDelete.setObjectTagName("Aggregate");
        doNotDelete.setObjectKey(aggregateUUID);
        cluster.getDoNotDeleteList().add(doNotDelete);    
    }

    /**
     * In a "2-node Campus" configuration, storage-shelf-info-iter does not return a correct value for "is-local-attach"
     * This method is to provide a correction for Shelf.
     *
     * @param shelf            the shelf to set the flag for
     * @param localShelfIds    set of lccal storage shelf id
     *
     */
    private void setLocalAttachForShelf(StorageShelf shelf, Set<String>  localShelfIds) {
        if(localShelfIds.contains(shelf.getShelfID())) {
             shelf.setIsLocalAttach(true);
        }
        else {
             shelf.setIsLocalAttach(false);
        }
    }

    /**
     * In a "2-node Campus" configuration, storage-disk-get-iter does not return a correct value for "is-local-attach"
     * This method is to provide a correction for Disk.
     *
     * @param disk             the disk to set the flag for
     * @param localShelfIds    set of lccal storage shelf id
     *
     */
    private void setLocalAttachForDisk(Disk disk, Set<String>  localShelfIds) {
        if(localShelfIds.contains(disk.getShelf())) {
            disk.setIsLocalAttach(true);
        }
        else {
            disk.setIsLocalAttach(false);
        }
    }

    /**
     *  private method to collect local shelf ids based on the following logic
     *
     *  var local_shlef_ids;
     *  var local_node_nvram_ids = call retrieve_local_node_nvram_ids() //retrieve local nodes nvram ids from system-node-get-iter
     *  for all disks in the cluster
     *      if (
     *          (disk.dr-home-node-id == null && disk.pool == 'Pool0' && local_node_nvram_ids.contains(disk.owner-node-id)) ||
     *          (disk.pool == 'Pool0' && local_node_nvram_ids.contains(disk.dr-home-node-id))
     *         ) then
     *          local_shlef_ids.add(disk.shelf)
     *      end if
     *  end for
     *
     *  For more details - https://wikid.netapp.com/w/MEG/Gordita/DS/Stretch_MCC#OCIE_dependencies
     *
     * @param   cModeCluster     current cluster
     * @return  localShelfIds
     */
    private Set<String> getLocalShelfIds(CModeCluster cModeCluster) {
        Set<String> localShelfIds = new HashSet<>();
        Set<String> localNodeNvramIds = new HashSet<>();
        for(CModeNodeDetail cModeNodeDetail : cModeCluster.getNodeDetails().values()) {
            localNodeNvramIds.add(cModeNodeDetail.getNvramID());
        }
        for(CModeDisk cModeDisk : cModeCluster.getDisks().values()) {
            if ((cModeDisk.getDrHomeNodeID() == null && cModeDisk.getPoolID() != null && 0L == cModeDisk.getPoolID()
                    && localNodeNvramIds.contains(cModeDisk.getOwnerNodeID()))
                    || (cModeDisk.getPoolID() != null && 0L == cModeDisk.getPoolID() &&
                    localNodeNvramIds.contains(cModeDisk.getDrHomeNodeID()))) {
                localShelfIds.add(cModeDisk.getShelf());
            }
        }

        return localShelfIds;
    }
    
}

